/****************************************************************************
pyvox.h - Header file to define shared objects for Pyvox

Author:  Paul Hughett

Copyright (c) 1999-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This header file defines the objects that must be shared among the different
files that define different components of the Pyvox module.

******************************************************************************/


#ifndef PYVOX_H
#define PYVOX_H

/*** FIXME: Does this do any good?  At least, it shouldn't do any harm. ***/
#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*-----------------------------------------------------------------------
pyvox_array - Python object struct to represent a voxel array

This struct forms a Python wrapper around the voxel_array struct used
in the Voxel Kit.  The voxel array itself is kept elsewhere, which
complicates memory management but makes it possible to use the voxel
kit as a C library independently of Python.
-----------------------------------------------------------------------*/

typedef struct {
    PyObject_HEAD            /* Python header for all objects */
    voxel_array *varray;     /* Pointer to the voxel array itself */
}  pyvox_array;



/*------------------------------------------------------------------------
Global variables and objects for the Pyvox module and friends
------------------------------------------------------------------------*/

extern PyTypeObject parray_type;
extern PyTypeObject kernel_type;

extern PyObject *pyvox_dict;
extern PyObject *PyvoxError;



/*------------------------------------------------------------------------
ReturnError - Set the error string and return a value

These macros set the error string in the error object PyvoxError and
return either zero or one.  The ReturnError version should be used in
methods that normally return a Python object but return zero in case
of an error; the XReturnError version is used in some other contexts
where a side effect is normal and the function returns non-zero to
indicate an error.  The SetError version simply sets the error message
and leaves it to the user to return the appropriate error value.

ALGORITHM: The do {...} while 0 format is used so that this macro can
be used wherever a statement is acceptable.
--------------------------------------------------------------------------*/

#define SetError(message) \
    do { PyErr_SetString(PyvoxError, message; } while (0)

#define ReturnError(message) \
    do { PyErr_SetString(PyvoxError, message); return 0; } while (0)

#define XReturnError(message) \
    do { PyErr_SetString(PyvoxError, message); return 1; } while (0)



/*-----------------------------------------------------------------------
kernel_object - Python object struct to represent a kernel

This struct forms a Python wrapper around the vxl_kernel struct used
in the Voxel Kit to represent a neighborhood or convolution kernel.
The vxl_kernel struct itself is kept elsewhere, which complicates
memory management but makes it possible to use the Voxel Kit as a C
library independent of Python.
-----------------------------------------------------------------------*/

typedef struct {
    PyObject_HEAD               /* Header for all Python objects */
    struct vxl_kernel *kern;    /* Pointer to the kernel itself */
} kernel_object;



/*---------------------------------------------------------------------------
Prototypes of shared functions 
----------------------------------------------------------------------------*/

extern int PyvoxDouble_Check(PyObject *ob, double *value);
extern int PyvoxLong_Check(PyObject *ob, long *value);
extern int PyvoxArray_Check(PyObject *ob, voxel_array **vob);
extern int Pyvox_ParseShape(PyObject *shape, int *rank, long *dimen);
extern int PyKernel_Check(PyObject *ob);

/* FIXME: Shared functions that should probably be made private */
extern PyObject *setup_vxl_binpt(PyObject *pysrc1, PyObject *pysrc2,
    int deftype, 
    voxel_array **src1, int *free1, voxel_array **src2, int *free2);
extern PyObject *parray_alloc(int type, int rank, long *dimen);
extern pyvox_array *parray_create(void);
extern voxel_array *PyVoxel_GetArray(PyObject *ob);


#ifdef __cplusplus
}
#endif

#endif /* PYVOX_H */
