/****************************************************************************
voxel.h - Public header file for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 1998-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging applications.
*****************************************************************************/


#ifndef VOXEL_H
#define VOXEL_H

#include <stddef.h>
#include <stdio.h>

#include "errm.h"
#include "exim.h"



/*--------------------------------------------------------------------------
Capacity parameters:  
These parameters set the capacity limits of the voxel kit; they should 
be redefinable by modifying these definitions and recompiling, within the
limits of the machine, OS, and compiler.  They are also used as a sanity
check.
--------------------------------------------------------------------------*/

#define  VXL_MAX_RANK  8    /* Maximum allowed number of dimensions */
#define  VXL_MAX_ECHO  8    /* Maximum allowed number of echoes */


#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*------------------------------------------------------------------------
vxl_header_v0 - Voxel data file header, version 0

NOTE:  The voxel file header is dead, at least for the moment.  The
images will be stored in raw format, with parameters obtained from a
parameter file or command line.

A voxel data file is designed to represent data on a rectangular grid
of points in an architecture-independent format, with optional
annotations (yet to be designed).  Two-dimensional images and medical
volume images are typical examples.  Up to 8-dimensional data may be
represented, including mutiple spectral bands.  A voxel data file
traditionally has a ".vxl" extension.  The data itself may be included
in the main file, or put into a separate file; in the latter case, the
extension of the data file is traditionally ".img".

This structure describes the external representation of the header of
a vxl data file, assuming a hypothetical machine for which long is a
4-byte big-endian two's complement integer, float is a 4-byte
big-endian IEEE-format floating point number, and void * is a 4-byte
unsigned big-endian integer identifying the byte address (as in ftell
and fseek) at which the indicated structure begins.  The header always
begins at byte 0 of the file.  It should never be read in directly
(since the native format could be different) but must be read as a
byte stream and converted to the current internal representation.

The error field, if non-zero, indicates that some error was detected
in the process generating this data and that not all of this data is
correct.  The value of the field is one of the error constants defined
above.

The type field indicates the format of each datum, and is selected
from one of the constants defined above; the nbytes field indicates
the number of bytes allocated to each datum.  Big-endian formats are
preferred (and in fact, we have yet to define formats for little- or
middle-endian data).

The length field contains the number of bytes (storage units) actually
allocated in the file for the data itself.  It must be greater than or
equal to nbytes times the product of the dimensions.

The dimen array contains the number of elements per dimension of the
data.  The most slowly varying coordinates are given first; thus, for
example, the dimensions of a three-dimensional image scanned in the
usual raster order would be slice, row, and column.

The origin array contains the physical coordinate corresponding to
index zero on each coordinate; the spacing array contains the physical 
spacing between successive planes along each coordinate, in unstated 
units.  If the spacing is zero, it means that the spacing is unknown 
or not meaningful.  (For example, suppose that successive slices are 
T1-, T2-, and proton-weighted MRI images.)

The data field points to the actual voxel data, as a byte address
(fseek and ftell) within the file; if it is zero, the data is actually
stored in a separate file with the extension ".img", starting at byte
0.  By convention, voxels are listed starting with the upper left
corner of the first slice, then right across the slice, then down
successive scanlines, then forward through successive slices, and so
on through all the dimensions.  By an independent convention, the data are
stored in row-major order; that is, the last coordinate varies most rapidly
as the data are acessed sequentially in memory.  The two conventions
combined dictate that (for example) the first coordinate is slice number,
the second row number, the third column number, and the fourth the echo
number.  The coordinate indices start at 0 for each dimension.

The auxinfo, usertype, and userinfo fields are reserved for later
definition.
--------------------------------------------------------------------------*/

#define VXL_MAGIC_V0  0x56584c30L    /* Magic number = 'VXL0' */

struct vxl_header_v0 {
    long magic;                    /* Magic number = VXL_MAGIC_V0 */
    long rank;                     /* Number of dimensions, from 0 to 8 */
    long error;                    /* Error flag */
    long type;                     /* Architecture-independent type */
    long nbytes;                   /* Number of bytes per datum */
    long length;                   /* Actual length of data area */
    long   dimen  [VXL_MAX_RANK];  /* Array of dimensions */
    double origin [VXL_MAX_RANK];  /* Array of origin values */
    double spacing[VXL_MAX_RANK];  /* Array of spacings/intervals */
    void *auxinfo;                 /* Pointer to additional information */
    long usertype;                 /* Reserved for the user */
    void *userinfo;                /* Reserved for the user */
    void *data;                    /* The actual data */
};



/*--------------------------------------------------------------------------
voxel_array - Descriptor for a multi-dimensional data array

This structure defines the current internal representation of a voxel
data set, typically read from a voxel data file or computed in core.
Note that this structure, while similar to the external representation,
is NOT guaranteed to be identical and must be properly converted from
the external representation.  In particular, the file may have been written
on a machine with different architecture and byte order, or the file may
have been written in an older version of the format.  Modulo that caveat,
the contents of this structure are essentially those of the external
header, except that all fields are in the native machine format.

A voxel array is designed to represent data on a rectangular grid of
points.  Two-dimensional images and medical volume images are typical
examples.  Up to 8-dimensional data may be represented, including
mutiple spectral bands.

The magic field is set to VXL_MAGIC to indicate that this structure is
a voxel array and has been initialized.

The error field, if non-zero, indicates that some error was detected
in the process generating this data and that not all of this data is
correct.  The value of the field is one of the error constants defined
above.

FIXME: As far as I recall, the error field is never used; consider removing
it.

The type field indicates the format of each datum, and is selected
from one of the constants defined above; the nbytes field indicates
the number of bytes allocated to each datum.

The length field contains the number of bytes (storage units) actually
allocated for the data itself.  It must be equal to nbytes times the
product of the dimensions.  (Prior to 2000-11-25, it was permitted to
be greater; as far as I know, no code actually took advantage of this
freedom.)

The dimen array contains the number of elements per dimension of the
data.  The most slowly varying coordinates are given first; thus, for
example, the dimensions of a three-dimensional image scanned in the
usual raster order would be slice, row, and column.

The origin array contains the physical coordinate corresponding to
index zero on each coordinate; the spacing array contains the physical 
spacing between successive planes along each coordinate, in unstated 
units.  If the spacing is zero, it means that the spacing is unknown 
or not meaningful.  (For example, suppose that successive slices are 
T1-, T2-, and proton-weighted MRI images.)

The data field points to the actual voxel data, or is zero to indicate
that no storage has been allocated.  By convention, voxels are listed
starting with the upper left corner of the first slice, then right
across the slice, then down successive scanlines, then forward through
successive slices, and so on through all the dimensions.  The
coordinate indices start at 0 for each dimension.  If data is
non-zero, then type, rank, dimen, nbytes, and length must be set to
the correct values

The auxinfo, usertype, and userinfo fields are reserved for later
definition.

FIXME: The auxinfo, usertype, and userinfo fields are never used, as
far as I recall, and complicate the semantics of copying a voxel array;
consider completely removing them.

FIXME: Add a vector of strides along each dimension, which will facilitate
some access operations.

FIXME: Modify to support array views which share the same data but differ
in rank and dimensions.

Basic states of a voxel_array:  A voxel array is uninitialized if the
magic word is not the correct magic value; it is initialized if the
the magic word is correct, and other fields have been set to
reasonable values.  It is unallocated if data == 0, and allocated
otherwise; the type, rank, dimen, nbytes, and length of an allocated
array must be set to the correct values, but other fields (except
magic) need not have any specific values.  The auxinfo, usertype, and
userinfo fields will be set to zero when a voxel array is initialized,
but are otherwise ignored and untouched by the voxel kit.

It is required that the data field MUST either be zero or a value
returned by malloc or its friends.  Furthermore, the lifetime of a
pointer read from data is not guaranteed to last beyond the next voxel
kit call that writes to that voxel array.


There are various special applications of the voxel_array structure,
which make additional assumptions about the format of the data
described; two of these special formats are described next.


RAW DATA FORMAT

It is sometimes necessary to read image data from an external file
without committing to any particular external representation.  For
example, an image display program might want to allow the user to
dynamically change the assumed external format while searching for the
proper interpretation of an image data file abandoned on his doorstep.
The ``raw data format'' is used for this purpose and essentially
treats the data as simply a long stream of 8-bit unsigned chars.  The
last (innermost) dimension of the voxel array is assumed to index over
the bytes in each datum.  The outer dimensions may be used to structure
the data but this is not required; it is permissible to have just a
1-d structure in which the contents are considered as simply a stream
of bytes.


RGB2D DATA FORMAT

There are those who claim that image data is meant to be seen, not
simply computed.  Preparing volume data for display usually involves
extracting a 2d section through the multidimensional data set, scaling
or otherwise manipulating the intensity range of the data to fit
within the limits of a display, and sometimes other operations; the
``rgb2d format'' is intended to capture the result of these
manipulations without making any commitment to a particular display
device.  The rgb2d is always a 3-dimensional unsigned char image in
which the last dimension has depth three, representing in order the
red, green, and blue channels of the displayable image.  (The use of
unsigned char limits the intensity range to 0..255.  This will be
adequate for most but not all display devices; it may someday be
necessary to provide a similar format with longer pixels.)

FIXME:  Should there also be a mono2d format for gray scale images?
It's probably unnecessary for display systems, but it might be
useful for making screen dumps of monochrome images as monochrome
rather than rgb images three times larger.
---------------------------------------------------------------------------*/

#define VXL_MAGIC    19182   /* The magic number */

/* Multi-dimensional array for image data */
struct vxl_array_struct {
    long magic;                    /* Magic number = VXL_MAGIC */
    long rank;                     /* Number of dimensions >= 0 */
    long error;                    /* Error flag */
    long type;                     /* Architecture-independent type */
    long nbytes;                   /* Number of bytes per datum */
    long length;                   /* Length of data area, in bytes */
    long   dimen  [VXL_MAX_RANK];  /* Array of dimensions */
    double origin [VXL_MAX_RANK];  /* Array of origin values */
    double spacing[VXL_MAX_RANK];  /* Array of spacings/intervals */
    void *auxinfo;                 /* Pointer to additional information */
    long usertype;                 /* Reserved for the user */
    void *userinfo;                /* Reserved for the user */
    void *data;                    /* The actual data */
};

typedef struct vxl_array_struct voxel_array;



/*--------------------------------------------------------------------------
struct vxl_raw_format - Descriptor for the external (raw) format of an image

Thus structure is used primarily (if not only) by the vxl_raw_to_rgb2d
function and describes the external (raw) format of data comprising
a voxel_array.  It is sometimes used within an image display program
to permit changing the data intepretation on the fly while interactively
searching for the right format.  To support this, it also includes some
information about how to select an image section and format it for
display as a 2d rgb image.

Raw image data refers to image data stored in the external (file)
format, which need not be the same as the internal (native) format of
the processor.  Within the file itself, raw image data is simply a
stream of (at least 8-bit) unsigned char bytes.  Raw image data is
stored internally as an unsigned char voxel array in which the last
(innermost) dimension is assumed to be an index over the bytes
comprising a voxel.  In the minimal case, the voxel array has only one
(very long) dimension and appears as simply a byte stream.

The xdim and ydim fields indicate which dimensions of the data are
to be extracted as the x and y coordinates of the display; it is possible
but not usually sensible for them to be equal.  

The edim field indicates which dimension of the data indexes the different 
``echoes'' that are to mapped to red, green, and blue for display.  If 
edim is set to a value in the range 0...rank-1, then the recho, gecho, 
and becho fields indicate the specific coordinate values along this 
dimension that should be used for red, green, and blue; setting them all 
equal yields a monochrome image, of course.  On the other hand, if edim 
is set to the rank of the data array, then there is only one echo and it 
is not explicitly indexed along any dimension; the recho, gecho, and becho 
fields are ignored and a monochrome image is seen.

Related functions (even if not yet implemented):

   vxl_raw_to_rgb2d - Convert raw image data to 2d rgb section
   vxl_voxel_to_rgb2d - Convert voxel array to 2d rgb section
   vxl_rgb2d_to_XImage - Convert 2d rgb image data to XImage format
   vxl_raw_to_voxel - Convert raw image data to voxel array
                         [from file or data already in core?]
   vxl_voxel_to_raw - Convert voxel array to raw image data
   vxl_read_raw_data - Read raw image data from a file
   vxl_write_image_data - Write raw image data to a file

FIXME: Description of external raw format vs reformatting for display?
Dynamic section extraction from an open file (to save memory)?

FIXME:  Define ``natural'' translations between native and raw format
that preserve as much information as possible?

FIXME:  Is there any compelling reason to use float rather than double
here, and in the voxel array struct?  Space is not really an issue.
--------------------------------------------------------------------------*/

#define VXL_RAWFORM   19185  /* Magic number for a raw format struct */

struct vxl_raw_format {
    long  magic;                  /* Magic number = VXL_RAWFORM */
    long  error;                  /* ? Error flag */

    /* Parameters describing the external (raw) format */
    long  type;                   /* Code for external data format */
    long  bigend;                 /* 1 => big-endian; 0 => little-endian */
    long  rank;                   /* Number of dimensions >= 0 */
    long  dimen[VXL_MAX_RANK];    /* Dimensions of the voxel array */

    /* Parameters describing how to reformat for 2d RGB display */
    long  current[VXL_MAX_RANK];  /* Coordinates of current section */
    long  direct[VXL_MAX_RANK];   /* Forward (>=0) or reverse (<0) scan */
    float minint[VXL_MAX_ECHO];   /* Minimum intensity value for window */
    float maxint[VXL_MAX_ECHO];   /* Maximum intensity value for window */
    unsigned char 
         *colormap[VXL_MAX_ECHO]; /* Color map for each channel of RGB */
    long  xdim, ydim;             /* Dimensions to map onto display X and Y */
    long  edim;                   /* Dimension to map onto display color */
    long  recho, gecho, becho;    /* Echo numbers to map onto RGB channels */
};



/*----------------------------------------------------------------------------
voxel_object_data - Description of an object

This structure contains some useful information on a single object
(a connected region of non-zero voxels) found in a volume image.

Most fast algorithms for finding objects in an image have the 
characteristic that they sometimes assign different id numbers to 
two apparently distinct objects that are later found to be parts 
the same object.  When this happens, one object id is taken as canonical 
and the "canon" field of the other is set to this canonical id; 
furthermore, the count field of the non-canonical object is added
to the canonical object and then set to zero.  The canonical object
number is distinguished by the fact that its "canon" field is equal to
its "ident" field.  The "point"
coordinates of the non-canonical object are left unchanged, although 
it's not clear that they are useful for anything.  The ident (and canon)
fields are limited to the size of unsigned short to facilitate later
table lookup on object idents.  If you have more than 65535 objects in
an image, you're out of luck, at least with this version of the software.  
Ident 0 is always reserved for the background.
----------------------------------------------------------------------------*/

struct voxel_object_data {
    unsigned short ident;       /* Object identification number */
    unsigned short canon;       /* Canonical object id number */
    long point[3];              /* Coordinates of first point in the object */
    long count;                 /* Count of voxels in this object */
};



/*-----------------------------------------------------------------------
vxl_kernel - Define a neighborhood or convolution kernel

This structure defines a neighborhood of a voxel by giving the coordinate
offsets (or deltas) to each neighbor relative to the coordinates of the
central voxel; it optionally gives a coefficient to go with each neighbor,
as for a convolution kernel, as well as a constant bias term for the
convolution.

If coef is zero, then there are no coefficients, and the structure
merely defines the set of voxels which constitute the neighborhood;
the bias should be ignored in this case.  Otherwise the structure
defines a convolution kernel, for which the bias and coefficients are
always of type double.

A kernel must have rank greater than 0, since convolution is not
defined for scalars; and it must include at least one neighbor, to
avoid some anomalies in determining the rank of a kernel being created
in Pyvox.

The macro VXL_MAX_NEIGH is an upper limit on the number of neighbors
that may be included in a kernel.  The true maximum value is probably
INT_MAX, but we use a smaller value to help catch blunders.  The
choice of 1024 for the number of neighbors is more or less arbitrary
but should suffice for most applications; if you have more than that
many neighbors, you probably need a better algorithm anyway.

FIXME: Consider adding an overall scale factor to this struct.  It's
unneccessary for convolution, where it can be combined with the
coefficients, but would be useful for local statistics.
-----------------------------------------------------------------------*/

#define VXL_MAX_NEIGH  1024

struct
vxl_kernel {
    /* FIXME:  Add a magic number for verification? */
    int rank;                     /* Number of dimensions */
    int count;                    /* Number of neighbors in this kernel */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate deltas for each neighbor */
    double *coef;                 /* Coefficient array, or zero */
    double bias;                  /* Constant bias term for convolution */
};



/*------------------------------------------------------------------------
Public data structures of the voxel kit
-------------------------------------------------------------------------*/

/* Standard neighborhoods and convolution kernels */
/* FIXME:  The better way to define all these kernels would be to write
   functions that create kernels with desired characteristics on demand. */
extern struct vxl_kernel neigh3d27;
extern struct vxl_kernel lowpass3d;



/*------------------------------------------------------------------------
 Public functions of the voxel kit 
------------------------------------------------------------------------*/

extern voxel_array *vxl_create_array(void);
extern voxel_array *vxl_copy_array(voxel_array *src);
extern void vxl_destroy_array(voxel_array *dest);
extern void vxl_init_array(voxel_array *dest);
extern void vxl_alloc_array(voxel_array *dest, int type, int rank,
    long *dimen);
extern voxel_array *vxl_create_scalar(int type, double value);
extern void vxl_store_scalar(voxel_array *dest, int type, double value);
extern void vxl_free_array(voxel_array *dest);
extern void vxl_reshape_array(voxel_array *dest, int rank, long *dimen);

extern double vxl_get_voxel(voxel_array *src, long *coords);
extern void vxl_set_voxel(voxel_array *dst, long *coords, double value);

extern void vxl_read_uchar_data(voxel_array *dest, FILE *src, int rank,
    long *dimen);
extern void vxl_read_raw_data(voxel_array *dest, FILE *src, 
    int rank, long *dimen, int intype, int extype, int bigend);
extern void vxl_write_uchar_data(FILE *dest, voxel_array *src);
extern void vxl_write_raw_data(FILE *dest, int extype, int bigend, 
    voxel_array *src);

extern void vxl_const(voxel_array *dest, double value);
extern void vxl_ramp (voxel_array *dest, int axis);

extern struct vxl_kernel *vxl_create_kernel(int rank, int count,
    long (*delta)[VXL_MAX_RANK], double *coef, double bias);
extern void vxl_destroy_kernel(struct vxl_kernel *kern);



/*----------------------------------------------------------------------
Reduction operations
----------------------------------------------------------------------*/

extern double vxl_amax (voxel_array *src);
extern double vxl_amin (voxel_array *src);
extern double vxl_asum (voxel_array *src);
extern double vxl_aprod(voxel_array *src);

extern double vxl_dot(voxel_array *src1, voxel_array *src2,
    voxel_array *weight);
extern double vxl_norm1(voxel_array *src1, voxel_array *src2,
    voxel_array *weight);
extern double vxl_norm2(voxel_array *src1, voxel_array *src2,
    voxel_array *weight);
extern double vxl_normsup(voxel_array *src1, voxel_array *src2,
    voxel_array *weight);

extern void vxl_mean(voxel_array *dest, voxel_array *src1, voxel_array *src2,
    int naxes, int *axes);



/*-----------------------------------------------------------------------
Standard unary operations on voxel arrays

The following functions all take two voxel arrays as arguments, the
first as destination and the second as source.  Some other code,
notably Pyvox, exploits this parallelism to simplify its code; the
typedef for vxl_unop specifies the interface and is used by Pyvox.

FIXME: Do I want to add the semantic assumption that these are
elementwise operations?
-----------------------------------------------------------------------*/

typedef void vxl_unop (voxel_array *dest, voxel_array *src);

extern void vxl_uchar (voxel_array *dest, voxel_array *src);
extern void vxl_schar (voxel_array *dest, voxel_array *src);
extern void vxl_ushort(voxel_array *dest, voxel_array *src);
extern void vxl_short (voxel_array *dest, voxel_array *src);
extern void vxl_uint  (voxel_array *dest, voxel_array *src);
extern void vxl_int   (voxel_array *dest, voxel_array *src);
extern void vxl_ulong (voxel_array *dest, voxel_array *src);
extern void vxl_long  (voxel_array *dest, voxel_array *src);
extern void vxl_float (voxel_array *dest, voxel_array *src);
extern void vxl_double(voxel_array *dest, voxel_array *src);

extern void vxl_copy  (voxel_array *dest, voxel_array *src);
extern void vxl_not   (voxel_array *dest, voxel_array *src);
extern void vxl_minus (voxel_array *dest, voxel_array *src);
extern void vxl_abs   (voxel_array *dest, voxel_array *src);
extern void vxl_floor (voxel_array *dest, voxel_array *src);
extern void vxl_ceil  (voxel_array *dest, voxel_array *src);
extern void vxl_sqrt  (voxel_array *dest, voxel_array *src);
extern void vxl_exp   (voxel_array *dest, voxel_array *src);
extern void vxl_log   (voxel_array *dest, voxel_array *src);
extern void vxl_log10 (voxel_array *dest, voxel_array *src);
extern void vxl_sin   (voxel_array *dest, voxel_array *src);
extern void vxl_cos   (voxel_array *dest, voxel_array *src);
extern void vxl_tan   (voxel_array *dest, voxel_array *src);
extern void vxl_asin  (voxel_array *dest, voxel_array *src);
extern void vxl_acos  (voxel_array *dest, voxel_array *src);
extern void vxl_atan  (voxel_array *dest, voxel_array *src);
extern void vxl_sinh  (voxel_array *dest, voxel_array *src);
extern void vxl_cosh  (voxel_array *dest, voxel_array *src);
extern void vxl_tanh  (voxel_array *dest, voxel_array *src);



/*------------------------------------------------------------------------
Standard binary operations on voxel arrays
-------------------------------------------------------------------------*/

extern void vxl_min  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_max  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_and  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_or   (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_xor  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_add  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_sub  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_mul  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_div  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_pow  (voxel_array *dest, voxel_array *src1, voxel_array *src2);
extern void vxl_atan2(voxel_array *dest, voxel_array *src1, voxel_array *src2);



/*----------------------------------------------------------------------
Non-standard operations on voxel arrays

The following operations do not fit any of the standard patterns.

FIXME:  Check whether any of these *do* fit a standard pattern and
move as necessary.  I see a few candidates already.
----------------------------------------------------------------------*/

extern void vxl_get_slice(voxel_array *dst, voxel_array *src,
    long *nidx, long **index);
extern void vxl_set_slice(voxel_array *dst, long *nidx, long **index,
    voxel_array *src);
extern void vxl_cast  (voxel_array *dest, int dtype, voxel_array *src);
extern void vxl_scale_to_uchar(voxel_array *dest, voxel_array *src,
    double minval, double maxval);
extern void vxl_scale(voxel_array *dest, voxel_array *src,
    double gain, double bias);
extern void vxl_compare(voxel_array *dest, voxel_array *src1, 
    voxel_array *src2, unsigned char lsval, unsigned char eqval,
    unsigned char gtval);

/* Graphics functions */
extern void vxl_fill_2d_contour(voxel_array *canvas, long npoint,
    double *yx, double fill);

/* Reference implementations for interpolation and resampling */
extern double vxl_linear0(voxel_array *src, double *point);
extern void vxl_affine_nearest0(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);
extern void vxl_affine_linear0(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);

/* Fast implementations of interpolation and resampling */
extern double vxl_nearest(voxel_array *src, double *point);
extern void vxl_affine_nearest(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);
extern double vxl_linear(voxel_array *src, double *point);
extern void vxl_affine_linear(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);

extern void vxl_convolve(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh, long *shrink);
extern void vxl_dilate(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh);
extern void vxl_erode(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh);
extern void vxl_lostat(voxel_array *mean, voxel_array *sdev,
    voxel_array *src, struct vxl_kernel *kern, double scale);
extern void vxl_lomean_uchar3d(voxel_array *dest, voxel_array *src);
extern void vxl_zerbv_uchar3d(voxel_array *dest, voxel_array *src,
    double stddev);

extern void vxl_logcomp_ulong(voxel_array *dest, voxel_array *src);
extern void vxl_lookup(voxel_array *dest, voxel_array *src, 
    voxel_array *table);

extern void vxl_minmax_ulong(unsigned long *minval, unsigned long *maxval,
    voxel_array *src);
extern void vxl_stat(double *mean, double *stddev, voxel_array *source,
    voxel_array *weight);
extern void vxl_moments(double *mass, double *center, double *moments,
    voxel_array *src);
extern void vxl_mop(voxel_array *result, voxel_array *moments, 
    voxel_array *array1, voxel_array *array2, voxel_array *array3);
extern void vxl_histo_uchar(voxel_array *hist, voxel_array *src,
    voxel_array *weight);
extern void vxl_bihist_uchar(voxel_array *hist, voxel_array *mar1, 
    voxel_array *mar2, voxel_array *src1, voxel_array *src2,
    voxel_array *weight);
extern void vxl_kmeans1_histo(voxel_array *hist, int nclass,
    unsigned char *cent);
extern void vxl_nnclass1_uchar(voxel_array *dest, voxel_array *src,
    int nclas, unsigned char *clas, unsigned char *cent);
extern void vxl_nnclass2_uchar(voxel_array *dest, voxel_array *src1,
    voxel_array *src2, int nclas,  unsigned char *clas,
    unsigned char *cent1, unsigned char *cent2);

extern void vxl_2d_section(voxel_array *dest, voxel_array *src, int dim0,
    int dim1, long *slice);
extern void vxl_2d_sectwin(voxel_array *dest, voxel_array *src, int dim0,
    int dim1, long *slice, int bigend, int gmin, int gmax);
extern void vxl_init_raw_format(struct vxl_raw_format *fmt);
extern void vxl_raw_to_rgb2d(voxel_array *dst, voxel_array *src,
    struct vxl_raw_format *fmt);
extern void vxl_write_ppm_image(const char *fname, voxel_array *src);

extern void vxl_chamfer(voxel_array *dest, voxel_array *src, int dtype);
extern int  vxl_find_objects_uchar3d(voxel_array *dest,
    struct voxel_object_data *oblist[], voxel_array *src);


/* Matrix operations on voxel arrays */
extern double matrix_determinant(voxel_array *matrix);
extern void matrix_diag(voxel_array *diag, voxel_array *vector);
extern void matrix_eigen_symmetric(voxel_array *values, 
    voxel_array *vectors, voxel_array *matrix);
extern void matrix_get_diag(voxel_array *diagonal,voxel_array *matrix);
extern void matrix_inverse(voxel_array *inverse, voxel_array *matrix);
extern void matrix_principal_axes(voxel_array *values, 
    voxel_array *vectors, voxel_array *matrix);
extern void matrix_solve(voxel_array *soln, voxel_array *matrix,
    voxel_array *rhs);
extern void matrix_transpose(voxel_array *trans, voxel_array *matrix);

#ifdef __cplusplus
}
#endif

#endif /* VOXEL_H */
