/****************************************************************************
dumpbblanz - Anonymize a BBL Analyze header file

Author:  Paul Hughett

Copyright (c) 2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This program reads the BBL Analyze header file named on the command
line, replaces selected fields (descrip, aux_file, generated, scannum,
patient_id, exp_date, and exp_time) with blanks, and writes the
modified header file back over the original.  It is used to remove
patient identifying information from the header before it is given to
outsiders.  The old version of the header file is saved with the added
extension .orig appended to the original file name, but an existing
backup file is never modified.

More precisely, if no backup file exists, the named file is renamed
with the backup name and is read to obtain the header information; the
appropriate fields in the header are replaced with blanks and a new
header is written under the filename given on the command line.  If a
backup file already exists, it is not modified and is used instead of
the file named on the command line to supply the header data.

FIXME: bblanz is more or less a BBL-specific perversion of the AnalyzeView
header and should be moved into a BBL-specific package.
*****************************************************************************/

static const char rcsid[] =
"$Id: anonbblanz.c,v 1.5 2002/02/05 15:54:29 hughett Exp $ Release $Name:  $";

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "bblanz.h"
#include "errm.h"
#include "exim.h"
#include "memm.h"


int main(int argc, char *argv[])
{
    char *header;                 /* Name of the header file */
    char *backup;                 /* Name of the backup file */
    struct stat info;             /* Status of backup file, if any */
    FILE *hdrin = 0;              /* Header file to be anonymized */
    FILE *hdrout;                 /* Anonymized header file */
    struct exim_mode *mode;       /* Mode struct to read/write the header */
    int bigend = 1;               /* Always read as big endian */
    struct dsr *dsr = 0;          /* Placeholder for header struct */
    char *p;
    int i, n;

    /* Read argument and create name of the backup file */
    if (argc < 2)
        fatal("The header file must be specified");
    header = argv[1];
    n = strlen(header) + 6;
    backup = mallock(n);
    strcpy(backup, header);
    strcat(backup, ".orig");

    /* If backup does not exist, rename current file and use that */
    i = stat(backup, &info);
    if (i != 0 && errno == ENOENT) {
        n = rename(header, backup);
        if (n != 0)
            fatal("Error renaming current file to backup file");
        hdrin = fopen(backup, "rb");
        if (hdrin == 0)
            fatal("Unable to open (new) backup header file");
    }

    /* If backup exists and is a regular file, use it */
    else if ( i == 0 && S_ISREG(info.st_mode) ) {
        hdrin = fopen(backup, "rb");
        if (hdrin == 0)
            fatal("Unable to open (existing) backup header file");
    }

    /* Otherwise report an error */
    else
        fatal("Error while checking for backup file");

    /* Configure exim mode to read the file */
    mode = exim_create_mode(EXIM_IMPORT, hdrin, stdout, bigend);
    if (mode == 0)
        fatal("Insufficient memory to create input mode");

    /* Read the input file, and clean up tools */
    dsr = bblanz(mode, dsr);
    exim_destroy_mode(mode);
    fclose(hdrin);

    /* Blank the desired fields */
    for (p = dsr->hist.descrip,    i = 0; i < 80; i++)   p[i] = ' ';
    for (p = dsr->hist.aux_file,   i = 0; i < 24; i++)   p[i] = ' ';
    for (p = dsr->hist.generated,  i = 0; i < 10; i++)   p[i] = ' ';
    for (p = dsr->hist.scannum,    i = 0; i < 10; i++)   p[i] = ' ';
    for (p = dsr->hist.patient_id, i = 0; i < 10; i++)   p[i] = ' ';
    for (p = dsr->hist.exp_date,   i = 0; i < 10; i++)   p[i] = ' ';
    for (p = dsr->hist.exp_time,   i = 0; i < 10; i++)   p[i] = ' ';

    /* Open the output file */
    hdrout = fopen(header, "wb");
    if (hdrout == 0)
        fatal("Unable to open the output file");

    /* Configure to write the file */
    mode = exim_create_mode(EXIM_EXPORT, hdrout, stdout, bigend);
    if (mode == 0)
        fatal("Insufficient memory to create output mode");

    /* Actually overwrite the file now */
    dsr = bblanz(mode, dsr);
    exim_destroy_mode(mode);
    fclose(hdrout);
    free(backup);

    return 0;
}
