/**************************************************************************
binnseg.c - Bivariate nearest neighbor segmentation

Author:  Paul Hughett

Copyright (c) 1999 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program takes two unsigned char images and a listing of class centroids
in feature space and finds the class centroid closest to each voxel of the
input images; it then outputs an unsigned char image containing the class
id of each centroid found.

The following parameters are taken from the param file named on the
command line:

image1 = Filename of the image containing the first feature image
image2 = Filename of the image containing the second feature image
imgout = Filename of the output image containing the class ids
classes = Name of the file containing the class ids and centroids
nslices = Number of slices in the image
nrows   = Number of rows in the image
ncols   = Number of columns in the image

The file containing the class definitions should contain one line for
each class centroid; that line contains in order the class id to be
output for that class; the centroid coordinate for the first feature;
and the centroid coordinate for the second feature.  The class ids
need not be distinct, so that a given class id may be defined by more
than one centroid.

FIXME:  This program uses a fixed-length array for the class centroid;
it would be more flexible to use a dynamically allocated array.
**************************************************************************/


static const char rcsid[] = "$Id: binnseg.c,v 1.1 1999/06/22 20:34:54 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "cdata.h"
#include "errm.h"
#include "param.h"
#include "voxel.h"

#define MAXCLS   512        /* Maximum number of class centroids */


int
main(int argc, char **argv)
{
    Param_list params;             /* Contents of param file */
    long nslices, nrows, ncols;    /* Image dimensions */
    char *stage = "binnseg";       /* Default stage name for params */
    const char *imag1fn = 0;       /* Filename for the first image */
    const char *imag2fn = 0;       /* Filename for the second image */
    const char *imoutfn = 0;       /* Filename for the histogram as an image */
    const char *classfn = 0;       /* Filename for class definitions */
    FILE *imag1in;                 /* Stream for the first image */
    FILE *imag2in;                 /* Stream for the second image */
    FILE *imgout;                  /* Stream for the histogram as an image */
    FILE *class;                   /* Stream for the class definitions */
    voxel_array *imag1  = 0;       /* Voxel array for first image */
    voxel_array *imag2  = 0;       /* Voxel array for second image */
    voxel_array *clsimg = 0;       /* Voxel array for class image */
    int nclass;                    /* Number of centroids defined */
    unsigned char clid[MAXCLS];    /* Class ids */
    unsigned char cent1[MAXCLS];   /* Centroid coords in feature 1 */
    unsigned char cent2[MAXCLS];   /* Centroid coords in feature 2 */
    int rank = 3;                  /* Number of dimensions in images */
    long dimen[3];                 /* Array of the dimensions */
    int ndat;                      /* Number of data found by cdata */
    char **data, **cmt;            /* Data and comments found by cdata */
    const char *s;

    /* Determine the parameters for this run */
    if (argc < 2)
	fatal("No param file specified");
    params   = read_param_file(argv[1]);
    imag1fn  = get_param(params, stage, "image1");
    imag2fn  = get_param(params, stage, "image2");
    imoutfn  = get_param(params, stage, "imgout");
    classfn  = get_param(params, stage, "classes");
    nslices  = ( s = get_param(params, stage, "nslices") ) ? atol(s) :  256;
    nrows    = ( s = get_param(params, stage, "nrows")   ) ? atol(s) :  256;
    ncols    = ( s = get_param(params, stage, "ncols")   ) ? atol(s) :  256;
    dimen[0] = nslices;
    dimen[1] = nrows;
    dimen[2] = ncols;

    /* Do some consistency checking on parameters */
    if (nslices < 1 || nslices > 4000)
	fatal("nslices has an implausible value");
    if (nrows < 1 || nrows > 4000)
	fatal("nrows has an implausible value");
    if (ncols < 1 || ncols > 4000)
	fatal("ncols has an implausible value");

    /* Open and read the first image file */
    if (imag1fn == 0)
	fatal("No image1 file was specified");
    imag1in = fopen(imag1fn, "r");
    if (imag1in == 0)
	fatal("Unable to open image1 file %s", imag1fn);
    imag1 = vxl_create_array();
    vxl_read_uchar_data(imag1, imag1in, rank, dimen);
    fclose(imag1in);

    /* Open and read the second image file */
    if (imag2fn == 0)
	fatal("No image2 file was specified");
    imag2in = fopen(imag2fn, "r");
    if (imag2in == 0)
	fatal("Unable to open image2 file %s", imag2fn);
    imag2 = vxl_create_array();
    vxl_read_uchar_data(imag2, imag2in, rank, dimen);
    fclose(imag2in);

    /* Read the class definitions */
    if (classfn == 0)
	fatal("No classes file was specified");
    class = fopen(classfn, "r");
    if (class == 0)
	fatal("Unable to open class defn file %s", classfn);
    for (nclass = 0; ; nclass++) {
	if (nclass >= MAXCLS)
	    fatal("Too many class centroids found");
	ndat = cdata_read(class, &data, &cmt);
	if (ndat == EOF)  break;
	if (ndat < 3)
	    fatal("Too few values for class centroid");
	clid [nclass] = atoi(data[0]);
	cent1[nclass] = atoi(data[1]);
	cent2[nclass] = atoi(data[2]);
	cdata_free(ndat, data, cmt);
    }
    fclose(class);

    /* Process the image as required */
    clsimg = vxl_create_array();
    vxl_nnclass2_uchar(clsimg, imag1, imag2, nclass, clid, cent1, cent2);

    /* Write out the processed image */
    if (imoutfn == 0)
	fatal("No imgout file was specified");
    imgout = fopen(imoutfn, "w");
    if (imgout == 0)
	fatal("Unable to open output file %s", imoutfn);
    vxl_write_uchar_data(imgout, clsimg);
    fclose(imgout);

    /* Clean up and exit */
    if (clsimg != 0) vxl_destroy_array(clsimg);
    if (imag1  != 0) vxl_destroy_array(imag1);
    if (imag2  != 0) vxl_destroy_array(imag2);

    exit(EXIT_SUCCESS);
}
