/****************************************************************************
bips.c - Basic Image Processing Subroutines

Author:  Paul Hughett

Copyright (c) 1999-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  It contains the basic image processing subroutines,
which is a relatively small set of functions that operate on
one-dimensional vectors of image data and can reasonably be optimized
by hand for maximum performance and from which higher-level
multidimensional operations can be constructed without losing too much
performance.  They are intended primarily to be called within the Voxel
Kit itself, rather than by a user of the Voxel Kit.

Unless otherwise specified, the results are undefined if the
destintion stride is zero or if the source and destination, or two
destinations, overlap (that is, if they share any elements;
interleaving with a non-zero stride should cause no problems).  The
reason for this and similar restrictions is that a good compiler can
often generate more efficient code if it can assume that pointers
always point to distinct objects; thus we require that you do the
check for aliasing before calling these functions.  The source stride
may generally be zero, but this is only occasionally useful.

Note that, for efficiency, these functions assume that all necessary
error checking has been done by the caller and do only the error
checking that they absolutely must (typically only that the requested
operation is in fact implemented).  This is one of the reasons that
they should not be called by the user directly.

As a general rule, the source and destination addresses are specified
as void or unsigned char pointers, but strides are given in elements
of the specified type.  This mixed specification might seem confusing
but seems to work reasonably well in practice.

The priorities for these functions has been portability and efficiency
when computing large (e.g. 16 million elements) image data sets;
readability has been a poor third.  This means that means that the
code looks rather ugly, since all the decisions have been hoisted
out of the main loop(s), leaving many slight variations on the same
loop, which is often defined as a parameterized macro to save typing.
Note that these macros are not general-purpose and usually intended
for use ONLY in the function that they just precede.

This file should be compiled as C only, not as C++.  It makes heavy
use of some constructs that cause a C++ compiler to choke; in any
case, the right way to do this under C++ would be to use templates
instead of macros.  There should not be any problems calling these
routines from C++.

FIXME: The "necessary" set of operations supported in bips is
beginning to seem rather large.  Is there a better approach?

FIXME: These functions in most cases simply reproduce the arithmetic
behavior of the underlying C implementation.  Perhaps there should
be some tools for testing for unexpected behaviors and warning the
users (like the paranoia package for floating point, but also checking
for integer anomalies?).

FIXME: Add other operations, e.g. log2, hypot, shifts, etc...  Add
ternary operations, like mask/mix, daxpy, fused multiply/add (w +=
x*y?).  Add an exchange operation for transpose and related array
operations.  Add an inner loop for global statistics, and IIR (AR or
ARMA?)  filters.

FIXME: Add appropriately scaled versions of the transcedental
functions for integral types, after the fashion of analog computers.
These could be useful for fast but not very accurate computation.  Of
course, you may need to explain the theory of fixed point computation,
since analog computers are so obscure these days.

FIXME: Complex types are not yet supported (and are probably not
needed until the FFT is implemented).  It might possibly make sense to
support RGB as a data type, although this conflicts with the method of
treating echo/frequency/band as the last dimension of an array.
*****************************************************************************/

static const char rcsid[] = "$Id: bips.c,v 1.36 2001/11/13 18:53:16 hughett Exp $ Release $Name:  $";

#include <float.h>
#include <limits.h>
#include <math.h>

#include "config.h"
#include "bips.h"


/* FIXME:  Choose ONE error handling convention for bips, and use it.
   This variable is currently just ignored in favor of return values.  */
int bips_error = 0;          /* Cumulative error flag for this package */


/**************************************************************************
Internal utility functions

To be defined as needed...
**************************************************************************/

/* Limit and round to an integral destination type */
#define LIMIT(x,lo,hi)   \
    ( ((x) < (lo)) ? (lo) : ((x) > (hi)) ? (hi) : floor((x) + 0.5) )



/*************************************************************************
Utility functions
*************************************************************************/


/*------------------------------------------------------------------------
bips_locate - Compute address of a vector element

This function returns the address of the vector element numbered i*ss.
No subscript range checking is done.  If the type is invalid, then a
null pointer is returned.  This function should be used only in
circumstances where the caller does most of the validity checking.
-------------------------------------------------------------------------*/

#define LOCATE(type)          \
{                             \
    type *s = (type *) src;   \
    p = (void *) (s + i);     \
}

void *
bips_locate(
    int type,              /* Type of data */
    void *src,             /* Source array */
    long i)                /* Element number to find address of */
{
    void *p;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	LOCATE(unsigned char);
	return p;

    case INTP_USHORT:
	LOCATE(unsigned short);
	return p;

    case INTP_UINT:
	LOCATE(unsigned int);
	return p;

    case INTP_ULONG:
	LOCATE(unsigned long);
	return p;

    case INTP_SCHAR:
	LOCATE(signed char);
	return p;

    case INTP_SHORT:
	LOCATE(short);
	return p;

    case INTP_INT:
	LOCATE(int);
	return p;

    case INTP_LONG:
	LOCATE(long);
	return p;

    case INTP_FLOAT:
	LOCATE(float);
	return p;

    case INTP_DOUBLE:
	LOCATE(double);
	return p;

    default:
	/* Drop through to error return */ ;

    }

    return 0;    
}




/***************************************************************************
Nullary point operations
***************************************************************************/


/*------------------------------------------------------------------------
bips_zero - Fill with zeros

NOTE: This operation could be done by bips_copy, but occurs frequently
and it's a nuisance to have to create a zero in the correct type each
time.
-------------------------------------------------------------------------*/

#define ZERO_LOOP(type)             \
{                                   \
    type *d = (type *)dst;          \
    for (i = 0; i < nelem; i++) {   \
	*d = 0;   d += ds;          \
	}                           \
}

int
bips_zero(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	ZERO_LOOP(unsigned char);
	return 0;

    case INTP_USHORT:
	ZERO_LOOP(unsigned short);
	return 0;

    case INTP_UINT:
	ZERO_LOOP(unsigned int);
	return 0;

    case INTP_ULONG:
	ZERO_LOOP(unsigned long);
	return 0;

    case INTP_SCHAR:
	ZERO_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	ZERO_LOOP(short);
	return 0;

    case INTP_INT:
	ZERO_LOOP(int);
	return 0;

    case INTP_LONG:
	ZERO_LOOP(long);
	return 0;

    case INTP_FLOAT:
	ZERO_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	ZERO_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_fill_max - Fill with maximum possible value for destination type

NOTE: This operation could be done by bips_copy, but occurs frequently
and it's a nuisance to have to create the max value in the correct
type each time.
-------------------------------------------------------------------------*/

#define FILL_MAX_LOOP(type, maxval)   \
{                                     \
    type *d = (type *)dst;            \
    for (i = 0; i < nelem; i++) {     \
	*d = maxval;   d += ds;       \
	}                             \
}

int
bips_fill_max(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	FILL_MAX_LOOP(unsigned char, UCHAR_MAX);
	return 0;

    case INTP_USHORT:
	FILL_MAX_LOOP(unsigned short, USHRT_MAX);
	return 0;

    case INTP_UINT:
	FILL_MAX_LOOP(unsigned int, UINT_MAX);
	return 0;

    case INTP_ULONG:
	FILL_MAX_LOOP(unsigned long, ULONG_MAX);
	return 0;

    case INTP_SCHAR:
	FILL_MAX_LOOP(signed char, SCHAR_MAX);
	return 0;

    case INTP_SHORT:
	FILL_MAX_LOOP(short, SHRT_MAX);
	return 0;

    case INTP_INT:
	FILL_MAX_LOOP(int, INT_MAX);
	return 0;

    case INTP_LONG:
	FILL_MAX_LOOP(long, LONG_MAX);
	return 0;

    case INTP_FLOAT:
	FILL_MAX_LOOP(float, FLT_MAX);
	return 0;

    case INTP_DOUBLE:
	FILL_MAX_LOOP(double, DBL_MAX);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_zero_range - Fill with zeros within limits

This function sets to zero the vector elements numbered first*ds,
(first+1)*ds, ..., (limit-1)*ds.  It's sort of irregular, but useful
in some specialized circumstances.

NOTE: This operation could be done by bips_copy, but it's a nuisance to 
have to create a zero in the correct type each time and to decide exactly 
where to start.
-------------------------------------------------------------------------*/

#define ZERO_RANGE(type)                \
{                                       \
    type *d = (type *)dst;              \
    d += first * ds;                    \
    for (i = first; i < limit; i++) {   \
	*d = 0;   d += ds;              \
	}                               \
}

int
bips_zero_range(
    long first,              /* First element to zero */
    long limit,              /* Limit element to zero */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	ZERO_RANGE(unsigned char);
	return 0;

    case INTP_USHORT:
	ZERO_RANGE(unsigned short);
	return 0;

    case INTP_UINT:
	ZERO_RANGE(unsigned int);
	return 0;

    case INTP_ULONG:
	ZERO_RANGE(unsigned long);
	return 0;

    case INTP_SCHAR:
	ZERO_RANGE(signed char);
	return 0;

    case INTP_SHORT:
	ZERO_RANGE(short);
	return 0;

    case INTP_INT:
	ZERO_RANGE(int);
	return 0;

    case INTP_LONG:
	ZERO_RANGE(long);
	return 0;

    case INTP_FLOAT:
	ZERO_RANGE(float);
	return 0;

    case INTP_DOUBLE:
	ZERO_RANGE(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_fill_range - Fill with given value within limits

This function sets the vector elements numbered first*ds, (first+1)*ds, 
..., (limit-1)*ds to a value given as a double.  It's sort of irregular, 
but often useful.

NOTE: This operation could be done by bips_copy, but it's a nuisance to 
have to create a value in the correct type each time and to decide exactly 
where to start.
-------------------------------------------------------------------------*/

#define FILL_RANGE(type)                \
{                                       \
    type  v = (type)value;              \
    type *d = (type *)dst;              \
    d += first * ds;                    \
    for (i = first; i < limit; i++) {   \
	*d = v;   d += ds;              \
	}                               \
}

int
bips_fill_range(
    long first,              /* First element to zero */
    long limit,              /* Limit element to zero */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in source array */
    double value)            /* Value to fill with */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	FILL_RANGE(unsigned char);
	return 0;

    case INTP_USHORT:
	FILL_RANGE(unsigned short);
	return 0;

    case INTP_UINT:
	FILL_RANGE(unsigned int);
	return 0;

    case INTP_ULONG:
	FILL_RANGE(unsigned long);
	return 0;

    case INTP_SCHAR:
	FILL_RANGE(signed char);
	return 0;

    case INTP_SHORT:
	FILL_RANGE(short);
	return 0;

    case INTP_INT:
	FILL_RANGE(int);
	return 0;

    case INTP_LONG:
	FILL_RANGE(long);
	return 0;

    case INTP_FLOAT:
	FILL_RANGE(float);
	return 0;

    case INTP_DOUBLE:
	FILL_RANGE(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_ramp - Generate a linear ramp

This BIPS function stores the sequence 0, 1, 2, ... into the
destination.  The count is generated as a (signed) long value and cast
(using the usual C rules) into the destination type; this means that
the count may wrap around or otherwise misbehave if the destination
type cannot contain the count value.  If the number of elements is not
positive, this function returns immediately.
-------------------------------------------------------------------------*/

#define RAMP_LOOP(type)            \
{                                  \
    type *d = (type *)dst;         \
    for (i = 0; i < nelem; i++) {  \
        *d = (type)i;              \
        d += ds;    }              \
}

int
bips_ramp(
    long nelem,            /* Number of elements to store */
    int type,              /* Type of the destination array */
    void *dst,             /* Destination array */
    long ds)               /* Stride in the destination */
{
    long i;

    /* Return immediately if negative number of elements */
    if (nelem < 0) return 0;

    /* Switch according to the destination type */
    switch (type) {

    case INTP_UCHAR:
	RAMP_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	RAMP_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	RAMP_LOOP(short);
	return 0;

    case INTP_USHORT:
	RAMP_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	RAMP_LOOP(int);
	return 0;

    case INTP_UINT:
	RAMP_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	RAMP_LOOP(long);
	return 0;

    case INTP_ULONG:
	RAMP_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	RAMP_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	RAMP_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/**************************************************************************
Casts and other type conversions

These functions read a sequence of values in some specified internal
format and convert them to another (internal) format specified either
by the name of the function called or as a parameter.  Casting a type
to itself is permitted for uniformity and is probably almost as fast
as a plain copy.  The source and destination may have arbitrary
strides, except that the destination stride may not be zero.  The
results are unpredictable if thesource and destination arrays overlap.
The details of the conversion (overflow, exceptions, etc) are
inherited from the underlying C implementation.

See the exim library for conversions between internal and external data
formats.
**************************************************************************/


/*------------------------------------------------------------------------
CAST_LOOP - Generic loop for a type conversion operation

This macro defines a generic loop for implementing a type conversion
operation (or copy) where the source and destination arrays may have
arbitrary strides, except that the destination stride may not be zero.
The results are undefined if the source and destination arrays overlap.
------------------------------------------------------------------------*/

#define CAST_LOOP(type)                   \
{                                         \
    type * restrict s = (type *)src;      \
    for (i = 0; i < nelem; i++ ) {        \
        *d = *s;                          \
        d += ds;  s += ss; }              \
}



/*------------------------------------------------------------------------
bips_uchar - Convert voxel array to unsigned char
-------------------------------------------------------------------------*/

int
bips_uchar(
    long nelem,             /* Number of elements to convert */
    unsigned char *dst,     /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned char * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_schar - Convert voxel array to signed char
-------------------------------------------------------------------------*/

int
bips_schar(
    long nelem,             /* Number of elements to convert */
    signed char *dst,       /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    signed char * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_ushort - Convert voxel array to unsigned short
-------------------------------------------------------------------------*/

int
bips_ushort(
    long nelem,             /* Number of elements to convert */
    unsigned short *dst,    /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned short * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_short - Convert voxel array to short
-------------------------------------------------------------------------*/

int
bips_short(
    long nelem,             /* Number of elements to convert */
    short *dst,             /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    short * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_int - Convert voxel array to int
-------------------------------------------------------------------------*/

int
bips_int(
    long nelem,             /* Number of elements to convert */
    int *dst,               /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    int * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_uint - Convert voxel array to unsigned int
-------------------------------------------------------------------------*/

int
bips_uint(
    long nelem,             /* Number of elements to convert */
    unsigned int *dst,      /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned int * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_long - Convert voxel array to long
-------------------------------------------------------------------------*/

int
bips_long(
    long nelem,             /* Number of elements to convert */
    long *dst,             /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    long * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_ulong - Convert voxel array to unsigned long
-------------------------------------------------------------------------*/

int
bips_ulong(
    long nelem,             /* Number of elements to convert */
    unsigned long *dst,     /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned long * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_float - Convert voxel array to float
-------------------------------------------------------------------------*/

int
bips_float(
    long nelem,             /* Number of elements to convert */
    float *dst,            /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    float * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_double - Convert voxel array to double
-------------------------------------------------------------------------*/

int
bips_double(
    long nelem,             /* Number of elements to convert */
    double *dst,            /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    double * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return 0;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CAST_LOOP(short);
	return 0;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return 0;

    case INTP_INT:
	CAST_LOOP(int);
	return 0;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return 0;

    case INTP_LONG:
	CAST_LOOP(long);
	return 0;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return 0;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_cast - Convert voxel array to an arbitrary type

This function reads a sequence of voxel values in some specified
internal format and converts them to another internal type specified by
a numeric type code.
-------------------------------------------------------------------------*/

int
bips_cast(
    long nelem,             /* Number of elements to convert */
    int dtype,              /* Desired output type */
    void *dst,              /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    /* Switch according to the destination type */
    switch (dtype) {

    case INTP_UCHAR:
	return bips_uchar(nelem, (unsigned char *)dst, ds, stype, src, ss);

    case INTP_SCHAR:
	return bips_schar(nelem, (signed char *)dst, ds, stype, src, ss);

    case INTP_SHORT:
	return bips_short(nelem, (short *)dst, ds, stype, src, ss);

    case INTP_USHORT:
	return bips_ushort(nelem, (unsigned short *)dst, ds, stype, src, ss);

    case INTP_INT:
	return bips_int(nelem, (int *)dst, ds, stype, src, ss);

    case INTP_UINT:
	return bips_uint(nelem, (unsigned int *)dst, ds, stype, src, ss);

    case INTP_LONG:
	return bips_long(nelem, (long *)dst, ds, stype, src, ss);

    case INTP_ULONG:
	return bips_ulong(nelem, (unsigned long *)dst, ds, stype, src, ss);

    case INTP_FLOAT:
	return bips_float(nelem, (float *)dst, ds, stype, src, ss);

    case INTP_DOUBLE:
	return bips_double(nelem, (double *)dst, ds, stype, src, ss);

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*------------------------------------------------------------------------
bips_round_cast - Round and cast voxel array to an arbitrary type

This function reads a sequence of voxel values in some specified
internal format, rounds and limits appropriately for a given
destination type, and converts to the destination type.  That is, if
the destination is of integral type, the values are rounded to the
nearest integer and limited to the representable range of the
destination type; if the destination is floating point, then no
rounding is done (or needed).  This treatment of floating point
destinations is useful for higher-level functions that compute their
results in double and must then convert them to another format.  The
only source types permitted are float and double, on the grounds that
a plain bips_cast will handle all the other cases.

FIXME: We more or less assume that casting double to float rounds
appropriately; this might not actually be true.  But does it matter
for the intended purpose of this function?
-------------------------------------------------------------------------*/

/* Round and cast an floating point type to an integral type */
#define RCAST_LOOP(dsttype, srctype, lo, hi) \
{                                            \
    dsttype * restrict d = (dsttype *)dst;   \
    srctype * restrict s = (srctype *)src;   \
    for (i = 0; i < nelem; i++)  {           \
        *d = LIMIT(*s, lo, hi);              \
        d += ds; s+= ss;  }                  \
}

int
bips_round_cast(
    long nelem,             /* Number of elements to convert */
    int dtype,              /* Desired output type */
    void *dst,              /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_FLOAT:

	/* Switch according to the destination type */
	switch (dtype) {

	case INTP_UCHAR:
	    RCAST_LOOP(unsigned char, float, 0, UCHAR_MAX);
	    return 0;

	case INTP_SCHAR:
	    RCAST_LOOP(signed char, float, SCHAR_MIN, SCHAR_MAX);
	    return 0;

	case INTP_SHORT:
	    RCAST_LOOP(short, float, SHRT_MIN, SHRT_MAX);
	    return 0;

	case INTP_USHORT:
	    RCAST_LOOP(unsigned short, float, 0, USHRT_MAX);
	    return 0;

	case INTP_INT:
	    RCAST_LOOP(int, float, INT_MIN, INT_MAX);
	    return 0;

	case INTP_UINT:
	    RCAST_LOOP(unsigned int, float, 0, UINT_MAX);
	    return 0;

	case INTP_LONG:
	    RCAST_LOOP(long, float, LONG_MIN, LONG_MAX);
	    return 0;

	case INTP_ULONG:
	    RCAST_LOOP(unsigned long, float, 0, ULONG_MAX);
	    return 0;

	case INTP_FLOAT:
	    return bips_float(nelem, (float *)dst, ds, stype, src, ss);

	case INTP_DOUBLE:
	    return bips_double(nelem, (double *)dst, ds, stype, src, ss);

	default:
	    return ERR_ILLOP;
	}

    case INTP_DOUBLE:

	/* Switch according to the destination type */
	switch (dtype) {

	case INTP_UCHAR:
	    RCAST_LOOP(unsigned char, double, 0, UCHAR_MAX);
	    return 0;

	case INTP_SCHAR:
	    RCAST_LOOP(signed char, double, SCHAR_MIN, SCHAR_MAX);
	    return 0;

	case INTP_SHORT:
	    RCAST_LOOP(short, double, SHRT_MIN, SHRT_MAX);
	    return 0;

	case INTP_USHORT:
	    RCAST_LOOP(unsigned short, double, 0, USHRT_MAX);
	    return 0;

	case INTP_INT:
	    RCAST_LOOP(int, double, INT_MIN, INT_MAX);
	    return 0;

	case INTP_UINT:
	    RCAST_LOOP(unsigned int, double, 0, UINT_MAX);
	    return 0;

	case INTP_LONG:
	    RCAST_LOOP(long, double, LONG_MIN, LONG_MAX);
	    return 0;

	case INTP_ULONG:
	    RCAST_LOOP(unsigned long, double, 0, ULONG_MAX);
	    return 0;

	case INTP_FLOAT:
	    return bips_float(nelem, (float *)dst, ds, stype, src, ss);

	case INTP_DOUBLE:
	    return bips_double(nelem, (double *)dst, ds, stype, src, ss);

	default:
	    return ERR_ILLOP;
	}

    default:
	return ERR_ILLOP;
    }

    return ERR_ILLOP;
}



/*-------------------------------------------------------------------------
bips_scale_to_uchar - Scale voxel values to unsigned char

This function reads a sequence of voxel values in some specified
internal format, scales them using a specified offset and gain, and
converts the scaled values to the range 0..255 with values outside the
range being set to the limit.  The scaling equation is

                     y = gain*(x - offset)

for an offset b and a gain g.  This particular operation is provided
as a special case because of the prevalence of the 8-bit intensity
range on display systems.  Any integer or floating point type is
permitted as input, but not a complex type.  The rescaling is given
as gain and offset rather than min and max because this takes slightly
less computation; the higher level routines mostly use min and max
because they are more intutive in most applications.

FIXME: This function might be considered too high-level for bips,
since it could be replaced by a sequences of lower-level bips
operations done blockwise for performance.  But the added efficiency
of this specialized loop might be worth it anyway.  (Although many
modern workstations can overlap float and int computations at no
added cost.)

FIXME: This could be made somewhat more efficient for some special
cases, e.g. offset = 0 and gain = 1; specific single byte of an
unsigned type.  Whether or not it's worth it is another question.

Prior to 2000-03-01, the order of arguments was ...src, stype, ss...;
it was changed for consistency with other functions in this package.
Calls made using the old order will violate the prototype and should
be detected automatically by the compiler.

ALGORITHM: Rounding in this algorithm does not require the use of
floor() since the value being rounded is always non-negative; thus
it doesn't matter if casting a float value to an integral value
truncates toward zero or minus infinity.

FIXME: Does vxl_scale() replace this function?  But it is more general
and might be a little less efficient.
-------------------------------------------------------------------------*/

#define SCALE_TO_UCHAR(type) {                   \
    type * restrict s = (type *)src;             \
    for (i = 0; i < nelem; i++ ) {               \
	v = gain * ((double)*s - offset) + 0.5;  \
	if      ( v <   0.0 )  v =   0.0;        \
	else if ( v > 255.0 )  v = 255.0;        \
	*d = v;  d += ds;  s += ss; }  }         \


int
bips_scale_to_uchar(
    long   nelem,                /* Number of elements to convert */
    unsigned char  *dst,         /* Buffer for the output data */
    long   ds,                   /* Destination stride, in bytes */
    int    stype,                /* Type of the source data */
    void   *src,                 /* Buffer containing the input data */
    long   ss,                   /* Source stride, in elements */
    double offset,               /* Offset prior to scaling */
    double gain)                 /* Multiplier for scaling */
{
    unsigned char * restrict d = dst;
    double v;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	SCALE_TO_UCHAR(unsigned char);
	return 0;

    case INTP_SCHAR:
	SCALE_TO_UCHAR(signed char);
	return 0;

    case INTP_SHORT:
	SCALE_TO_UCHAR(short);
	return 0;

    case INTP_USHORT:
	SCALE_TO_UCHAR(unsigned short);
	return 0;

    case INTP_INT:
	SCALE_TO_UCHAR(int);
	return 0;

    case INTP_UINT:
	SCALE_TO_UCHAR(unsigned int);
	return 0;

    case INTP_LONG:
	SCALE_TO_UCHAR(long);
	return 0;

    case INTP_ULONG:
	SCALE_TO_UCHAR(unsigned long);
	return 0;

    case INTP_FLOAT:
	SCALE_TO_UCHAR(float);
	return 0;

    case INTP_DOUBLE:
	SCALE_TO_UCHAR(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/**************************************************************************
Unary point operations
**************************************************************************/



/*-------------------------------------------------------------------------
UNOP_LOOP - Generic loop for a unary point operation

This macro defines a generic loop for implementing a unary point
operation with arbitrary stride over a single input array to generate
a single output array.  The input stride may be zero, and the code
will select an optimized loop in this case.  The UNOP_XLOOP variant
must be used when the destination stride may be zero; the plain
UNOP_LOOP may handle the zero-stride case incorrectly on an optimizing
compiler.  The results are undefined if both destination and source
strides are zero; you shouldn't really be using a loop in this case
anyway.

The operation to be performed is given by a macro "asgn" which is
expected to take two arguments and expand into an assignment which
assigns to the first argument the result of the operation on the
second argument; the terminal semicolon must be omitted.  For example,

#define  NOT(d,s)   d = ~(s)

would implement the bitwise not operation.

This macro assumes that the following variable definitions are in
effect:

    void *dst, *src1;       Pointers to destination and source arrays
    long ds, ss1;           Strides in destination and source
    long nelem;             Number of elements to process

FIXME: Consider modifying the using functions to check for errors in
strides.  For the plain UNOP_LOOP, ds == 0 is an error; for the
UNOP_XLOOP, ds == ss == 0 is non-sensical.  An alternative would be to
detect the bad combinations in this macro and return an error code.
Note that the chained ifs could be converted to a switch.
Unfortunately, there doesn't seem to be any fast way to check for
source/destination overlaps.
--------------------------------------------------------------------------*/

#define UNOP_LOOP(type, asgn)  {       \
    type * restrict d  = (type *)dst;  \
    type * restrict s1 = (type *)src1; \
    long i;                            \
    if (ss1 == 0)                      \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds; }                 \
    else                               \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds;  s1 += ss1; }     \
    }


#define UNOP_XLOOP(type, asgn)  {      \
    type * restrict d  = (type *)dst;  \
    type * restrict s1 = (type *)src1; \
    long i;                            \
    if (ss1 == 0)                      \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds; }                 \
    else if (ds == 0)                  \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            s1 += ss1; }               \
    else                               \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds;  s1 += ss1; }     \
    }



/*------------------------------------------------------------------------
bips_not - Bitwise not (inversion) of an integer array

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
-------------------------------------------------------------------------*/

#define NOT(w,x)  w = ~(x)

int
bips_not(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, NOT);
	return 0;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, NOT);
	return 0;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, NOT);
	return 0;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, NOT);
	return 0;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, NOT);
	return 0;

    case INTP_SHORT:
	UNOP_LOOP(short, NOT);
	return 0;

    case INTP_INT:
	UNOP_LOOP(int, NOT);
	return 0;

    case INTP_LONG:
	UNOP_LOOP(long, NOT);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_copy - Copy a voxel array
-------------------------------------------------------------------------*/

#define COPY(w,x)  w = (x)

int
bips_copy(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, COPY);
	return 0;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, COPY);
	return 0;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, COPY);
	return 0;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, COPY);
	return 0;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, COPY);
	return 0;

    case INTP_SHORT:
	UNOP_LOOP(short, COPY);
	return 0;

    case INTP_INT:
	UNOP_LOOP(int, COPY);
	return 0;

    case INTP_LONG:
	UNOP_LOOP(long, COPY);
	return 0;

    case INTP_FLOAT:
	UNOP_LOOP(float, COPY);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COPY);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_minus - Elementwise additive inverse of a voxel array
-------------------------------------------------------------------------*/

#define MINUS(w,x)  w = -(x)

int
bips_minus(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, MINUS);
	return 0;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, MINUS);
	return 0;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, MINUS);
	return 0;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, MINUS);
	return 0;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, MINUS);
	return 0;

    case INTP_SHORT:
	UNOP_LOOP(short, MINUS);
	return 0;

    case INTP_INT:
	UNOP_LOOP(int, MINUS);
	return 0;

    case INTP_LONG:
	UNOP_LOOP(long, MINUS);
	return 0;

    case INTP_FLOAT:
	UNOP_LOOP(float, MINUS);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, MINUS);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_abs - Elementwise absolute value of a voxel array

Absolute value of unsigned data is supported for uniformity, but is
a simple copy.
-------------------------------------------------------------------------*/

#define ABS(w,x)  w = ((x) < 0) ? -(x) : (x)

int
bips_abs(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, COPY);
	return 0;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, COPY);
	return 0;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, COPY);
	return 0;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, COPY);
	return 0;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, ABS);
	return 0;

    case INTP_SHORT:
	UNOP_LOOP(short, ABS);
	return 0;

    case INTP_INT:
	UNOP_LOOP(int, ABS);
	return 0;

    case INTP_LONG:
	UNOP_LOOP(long, ABS);
	return 0;

    case INTP_FLOAT:
	UNOP_LOOP(float, ABS);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ABS);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_floor - Elementwise floor of a voxel array

FIXME: The gcc/x86 implementation of this is really time-consuming; see
if there is a better way using assembler.  One possibility is to compute
round(x-0.5) which does not require changing the FPU mode.  Better yet,
suggest to the gcc folks that they generate the better assemply code.
-------------------------------------------------------------------------*/

#define FLOOR(w,x)  w = floor(x)

int
bips_floor(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, FLOOR);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, FLOOR);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_ceil - Elementwise ceil of a voxel array

FIXME: See the comments under bips_floor.
-------------------------------------------------------------------------*/

#define CEIL(w,x)  w = ceil(x)

int
bips_ceil(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, CEIL);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, CEIL);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_exp - Elementwise exp of a voxel array
-------------------------------------------------------------------------*/

#define EXP(w,x)  w = exp(x)

int
bips_exp(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, EXP);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, EXP);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_log - Elementwise log of a voxel array
-------------------------------------------------------------------------*/

#define LOG(w,x)  w = log(x)

int
bips_log(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LOG);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LOG);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_log10 - Elementwise log10 of a voxel array
-------------------------------------------------------------------------*/

#define LOG10(w,x)  w = log10(x)

int
bips_log10(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LOG10);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LOG10);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_sqrt - Elementwise sqrt of a voxel array
-------------------------------------------------------------------------*/

#define SQRT(w,x)  w = sqrt(x)

int
bips_sqrt(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SQRT);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SQRT);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_sin - Elementwise sin of a voxel array
-------------------------------------------------------------------------*/

#define SIN(w,x)  w = sin(x)

int
bips_sin(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SIN);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SIN);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_cos - Elementwise cos of a voxel array
-------------------------------------------------------------------------*/

#define COS(w,x)  w = cos(x)

int
bips_cos(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, COS);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COS);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_tan - Elementwise tan of a voxel array
-------------------------------------------------------------------------*/

#define TAN(w,x)  w = tan(x)

int
bips_tan(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, TAN);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, TAN);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_asin - Elementwise asin of a voxel array
-------------------------------------------------------------------------*/

#define ASIN(w,x)  w = asin(x)

int
bips_asin(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ASIN);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ASIN);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_acos - Elementwise acos of a voxel array
-------------------------------------------------------------------------*/

#define ACOS(w,x)  w = acos(x)

int
bips_acos(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ACOS);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ACOS);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_atan - Elementwise atan of a voxel array

See also bips_atan2 for the four-quadrant arc tanget.
-------------------------------------------------------------------------*/

#define ATAN(w,x)  w = atan(x)

int
bips_atan(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ATAN);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ATAN);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_sinh - Elementwise sinh of a voxel array
-------------------------------------------------------------------------*/

#define SINH(w,x)  w = sinh(x)

int
bips_sinh(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SINH);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SINH);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_cosh - Elementwise cosh of a voxel array
-------------------------------------------------------------------------*/

#define COSH(w,x)  w = cosh(x)

int
bips_cosh(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, COSH);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COSH);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_tanh - Elementwise tanh of a voxel array
-------------------------------------------------------------------------*/

#define TANH(w,x)  w = tanh(x)

int
bips_tanh(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, TANH);
	return 0;

    case INTP_DOUBLE:
	UNOP_LOOP(double, TANH);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/***************************************************************************
Scatter/gather operations

Read/write elements in random order.
***************************************************************************/


/*-------------------------------------------------------------------------
bips_gather - Read elements in random order

This function reads elements in an order specified by a table of
indices, and copies them to a sequential vector.  The results are
unpredictable if some index value appears more than once, or if the
destination stride is zero.  The indices are NOT checked for range or
validity, since it is assumed that this check was done prior to
calling bips_gather; the vxl_check_index function may be useful for
this purpose.  (This function is often used for slice selection, in
which case it is more efficient to check the indices once and call
bips_gather multiple times with the same index array.)
--------------------------------------------------------------------------*/

#define GATHER_LOOP(type)    {      \
    long i;                         \
    type * restrict d = dst;        \
    type * restrict s = src;        \
    for (i = 0; i < nelem; i++)  {  \
        *d = s[index[i]];           \
        d += ds;  }   }             \

int
bips_gather(
    long nelem,          /* Number of elements to process */
    int  type,           /* Type of data */
    void *dst,           /* Destination array */
    long ds,             /* Stride in the destination array */
    void *src,           /* Source array */
    long *index)         /* Indices into the source array */
{
    /* Switch according to the data type */
    switch (type) {

    case INTP_UCHAR:
        GATHER_LOOP(unsigned char);
        return 0;

    case INTP_SCHAR:
        GATHER_LOOP(signed char);
        return 0;

    case INTP_USHORT:
        GATHER_LOOP(unsigned short);
        return 0;

    case INTP_SHORT:
        GATHER_LOOP(short);
        return 0;

    case INTP_UINT:
        GATHER_LOOP(unsigned int);
        return 0;

    case INTP_INT:
        GATHER_LOOP(int);
        return 0;

    case INTP_ULONG:
        GATHER_LOOP(unsigned long);
        return 0;

    case INTP_LONG:
        GATHER_LOOP(long);
        return 0;

    case INTP_FLOAT:
        GATHER_LOOP(float);
        return 0;

    case INTP_DOUBLE:
        GATHER_LOOP(double);
        return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}


/*-------------------------------------------------------------------------
bips_scatter - Write elements in random order

This function reads elements from a sequential vector and writes them
in an order specified by a table of indices.  The results are
unpredictable if some index value occurs more than once.  The indices
are NOT checked for range or validity, since it is assumed that this
check was done prior to calling bips_scatter; the vxl_check_index
function may be useful for this purpose.  (This function is often used
for slice selection, in which case it is more efficient to check the
indices once and call bips_scatter multiple times with the same index
array.)  The source stride may be zero, which is useful for stuffing a
constant value into an array slice.
--------------------------------------------------------------------------*/

#define SCATTER_LOOP(type)  {       \
    long i;                         \
    type * restrict d = dst;        \
    type * restrict s = src;        \
    for (i = 0; i < nelem; i++)  {  \
        d[index[i]] = *s;           \
        s += ss;  }  }              \

int
bips_scatter(
    long nelem,          /* Number of elements to process */
    int  type,           /* Type of data */
    void *dst,           /* Destination array */
    long *index,         /* Indices into the destination array */
    void *src,           /* Source array */
    long ss)             /* Stride in the source array */
{
    /* Switch according to the data type */
    switch (type) {

    case INTP_UCHAR:
        SCATTER_LOOP(unsigned char);
        return 0;

    case INTP_SCHAR:
        SCATTER_LOOP(signed char);
        return 0;

    case INTP_USHORT:
        SCATTER_LOOP(unsigned short);
        return 0;

    case INTP_SHORT:
        SCATTER_LOOP(short);
        return 0;

    case INTP_UINT:
        SCATTER_LOOP(unsigned int);
        return 0;

    case INTP_INT:
        SCATTER_LOOP(int);
        return 0;

    case INTP_ULONG:
        SCATTER_LOOP(unsigned long);
        return 0;

    case INTP_LONG:
        SCATTER_LOOP(long);
        return 0;

    case INTP_FLOAT:
        SCATTER_LOOP(float);
        return 0;

    case INTP_DOUBLE:
        SCATTER_LOOP(double);
        return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/***************************************************************************
Compound assignment operations

Such as +=, *=, etc.  These follow the same pattern as unary
operations and are implemented in the same fashion.  The source and
destination must be the same type and are assumed not to overlap.  The
desination stride may be zero, but the source and destination stride
may not both be zero.

These operations may be used for updating an array in place or for
summing (or whatever) over a dimension (an operation also known as
reduction); in either case, it is the caller's responsibility to
initialize the destination to some appropriate value.

The bitwise boolean operators are supported only for integral types; the
user is reminded that bitwise operations on signed integral types are not
portable.
***************************************************************************/


/*-------------------------------------------------------------------------
bips_and_set - Bitwise AND of source and destination
-------------------------------------------------------------------------*/

#define ANDSET(w,x)   w &= (x)

int
bips_and_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, ANDSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, ANDSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, ANDSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, ANDSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, ANDSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, ANDSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, ANDSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, ANDSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_or_set - Bitwise OR of source and destination
-------------------------------------------------------------------------*/

#define ORSET(w,x)   w |= (x)

int
bips_or_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, ORSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, ORSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, ORSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, ORSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, ORSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, ORSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, ORSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, ORSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_xor_set - Bitwise XOR of source and destination
-------------------------------------------------------------------------*/

#define XORSET(w,x)   w ^= (x)

int
bips_xor_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, XORSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, XORSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, XORSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, XORSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, XORSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, XORSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, XORSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, XORSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_min_set - Minimum of source and destination
-------------------------------------------------------------------------*/

#define MINSET(w,x)  if ((x) < (w)) w = (x)

int
bips_min_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, MINSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, MINSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, MINSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, MINSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, MINSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, MINSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, MINSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, MINSET);
	return 0;

    case INTP_FLOAT:
	UNOP_XLOOP(float, MINSET);
	return 0;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, MINSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_max_set - Maximum of source and destination
-------------------------------------------------------------------------*/

#define MAXSET(w,x)  if ((x) > (w)) w = (x)

int
bips_max_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, MAXSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, MAXSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, MAXSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, MAXSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, MAXSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, MAXSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, MAXSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, MAXSET);
	return 0;

    case INTP_FLOAT:
	UNOP_XLOOP(float, MAXSET);
	return 0;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, MAXSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_add_set - Add source to destination
-------------------------------------------------------------------------*/

#define ADDSET(w,x)  w += (x)

int
bips_add_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, ADDSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, ADDSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, ADDSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, ADDSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, ADDSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, ADDSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, ADDSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, ADDSET);
	return 0;

    case INTP_FLOAT:
	UNOP_XLOOP(float, ADDSET);
	return 0;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, ADDSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_sub_set - Subtract source from destination
-------------------------------------------------------------------------*/

#define SUBSET(w,x)  w -= (x)

int
bips_sub_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, SUBSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, SUBSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, SUBSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, SUBSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, SUBSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, SUBSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, SUBSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, SUBSET);
	return 0;

    case INTP_FLOAT:
	UNOP_XLOOP(float, SUBSET);
	return 0;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, SUBSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_mul_set - Multipy destination by source
-------------------------------------------------------------------------*/

#define MULSET(w,x)  w *= (x)

int
bips_mul_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, MULSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, MULSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, MULSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, MULSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, MULSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, MULSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, MULSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, MULSET);
	return 0;

    case INTP_FLOAT:
	UNOP_XLOOP(float, MULSET);
	return 0;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, MULSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}


/*-------------------------------------------------------------------------
bips_div_set - Divide destination by source
-------------------------------------------------------------------------*/

#define DIVSET(w,x)  w /= (x)

int
bips_div_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, DIVSET);
	return 0;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, DIVSET);
	return 0;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, DIVSET);
	return 0;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, DIVSET);
	return 0;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, DIVSET);
	return 0;

    case INTP_SHORT:
	UNOP_XLOOP(short, DIVSET);
	return 0;

    case INTP_INT:
	UNOP_XLOOP(int, DIVSET);
	return 0;

    case INTP_LONG:
	UNOP_XLOOP(long, DIVSET);
	return 0;

    case INTP_FLOAT:
	UNOP_XLOOP(float, DIVSET);
	return 0;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, DIVSET);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/***************************************************************************
Binary point operations
***************************************************************************/


/*------------------------------------------------------------------------
BINOP_LOOP - Generic loop for a binary point operation

This macro defines a generic loop for implementing a binary point
operation with arbitrary stride over two input arrays to generate a
third output array.  Either input stride may be zero, and a special
optimized loop will be used in these cases; the case that both source
strides are zero is NOT handled specially, since I can't think of any
use for it.  The BINOP_XLOOP variant must be used when the destination
stride may be zero; the plain BINOP_LOOP may handle this case
incorrectly on an optimizing compiler.  The results are undefined if
the destination and both source strides are zero.

There is also a variant form BINOP_RLOOP which handles the recursive
case correctly, even on an optimizing compiler, but not the case where
the destination stride is zero.  For this case, the user must specify
ds == ss1 and d == s1 + ss1; then the running sum/product/whatever
will be generated.  The second source stride ss2 may be zero.  It is
the caller's responsibility to set d[-ss1] correctly.

The operation to be performed is given by a macro "asgn" which is
expected to take three arguments and expand into an assignment which
assigns to the first argument the result of the operation on the
second and third arguments; the terminal semicolon must be omitted.
For example,

#define  ADD(d,s1,s2)   d = (s1) + (s2)

would implement addition.

This macro assumes the following variable definitions to be in effect:

    void *dst;            Pointer to destination array
    void *src1, *src2;    Pointers to source arrays
    long ds;              Stride in destination array
    long ss1, ss2;        Strides in source arrays
    long nelem;           Number of elements to process

FIXME: Consider modifying the using functions to detect striding
errors.  For LOOP and RLOOP, it is an error if ds == 0; for XLOOP, it
is non-sensical if ds == ss1 == ss2 == 0.  An alternative would be to
detect the bad combinations in this macro and return an error code.
Note that the chained ifs could be converted to a switch.
Unfortunately, there doesn't seem to be any fast way to check for
source/destination overlaps.

FIXME: Look at the assembly code and verify that the restrict
qualifiers are doing the right thing.  Add tests to catch errors in
the restrictions?

FIXME:  Are side parameters supported cleanly?  I think so, but it needs
to be checked; type conversions might be a problem.
-------------------------------------------------------------------------*/

/* No overlaps and dest stride != 0 */
#define BINOP_LOOP(type, asgn)  {            \
    type * restrict d  = (type *)dst;        \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ss1 == 0)                            \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }


/* Recursive overlap only and dest stride != 0 */
#define BINOP_RLOOP(type, asgn)  {           \
    type * restrict d  = (type *)dst;        \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ds == ss1 && d == s1 + ss1 &&        \
        ss2 == 0)                            \
        for (i = 0; i < nelem; i++) {        \
            asgn(*d, *(d-ss1), *s2);         \
            d += ds; }                       \
    else if (ds == ss1 && d == s1 + ss1)     \
        for (i = 0; i < nelem; i++) {        \
            asgn(*d, *(d-ss1), *s2);         \
            d += ds; s2 += ss2; }            \
    else if (ss1 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }


/* No overlap but dest stride may be zero */
#define BINOP_XLOOP(type, asgn)  {           \
    type * restrict d  = (type *)dst;        \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ds == 0 && ss1 == 0)                 \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            s2 += ss2; }                     \
    else if (ds == 0 && ss2 == 0)            \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            s1 += ss1; }                     \
    else if (ss1 == 0 && ss2 == 0)           \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; }                       \
    else if (ss1 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }



/*--------------------------------------------------------------------------
bips_and - Bitwise AND on integer arrays

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
---------------------------------------------------------------------------*/

#define AND(w,x,y)  w = (x) & (y)

int
bips_and(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, AND);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, AND);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, AND);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, AND);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, AND);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, AND);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, AND);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, AND);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*--------------------------------------------------------------------------
bips_or - Bitwise OR on integer arrays

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
---------------------------------------------------------------------------*/

#define OR(w,x,y)  w = (x) | (y)

int
bips_or(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, OR);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, OR);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, OR);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, OR);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, OR);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, OR);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, OR);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, OR);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*--------------------------------------------------------------------------
bips_xor - Bitwise XOR on integer arrays

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
---------------------------------------------------------------------------*/

#define XOR(w,x,y)  w = (x) ^ (y)

int
bips_xor(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, XOR);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, XOR);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, XOR);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, XOR);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, XOR);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, XOR);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, XOR);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, XOR);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/*------------------------------------------------------------------------
bips_compare - Compare two voxel arrays

This function compares two sequences of voxel values in any
(non-complex) internal format and outputs any of three given unsigned
char values according to whether the voxel in the first sequence is
less than, equal to, or greater than the voxel in the second sequence.
The two source sequences must be of the same type.  

NOTE that this function is atypical in that the result is always unsigned
char rather than being the same type as the source arrays; as a reminder of
this, the type argument precedes the source arguments rather than the
destination argument.  A consequence of this is that the standard BINOP_LOOP 
does not work here because the destination is always unsigned char; so we 
have to use a modified loop.
--------------------------------------------------------------------------*/

#define COMP(w,x,y)                \
    if      (x < y)  w = lsval;    \
    else if (x > y)  w = gtval;    \
    else             w = eqval


#define COMP_LOOP(type, asgn)  {             \
    unsigned char * restrict d = (unsigned char *)dst; \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ss1 == 0)                            \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }

int
bips_compare(
    long   nelem,                 /* Number of elements to compare */
    unsigned char *dst,           /* Result of comparisons */
    long   ds,                    /* Destination stride, in elements */
    int    stype,                 /* Common source type */
    void  *src1,                  /* First source array */
    long   ss1,                   /* First source stride, in elements */
    void  *src2,                  /* Second source array */
    long   ss2,                   /* Second source stride, in elements */
    unsigned char  lsval,         /* Output if less than */
    unsigned char  eqval,         /* Output if equal to */
    unsigned char  gtval)         /* Output if greater than */
{
    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	COMP_LOOP(unsigned char, COMP);
	return 0;

    case INTP_USHORT:
	COMP_LOOP(unsigned short, COMP);
	return 0;

    case INTP_UINT:
	COMP_LOOP(unsigned int, COMP);
	return 0;

    case INTP_ULONG:
	COMP_LOOP(unsigned long, COMP);
	return 0;

    case INTP_SCHAR:
	COMP_LOOP(signed char, COMP);
	return 0;

    case INTP_SHORT:
	COMP_LOOP(short, COMP);
	return 0;

    case INTP_INT:
	COMP_LOOP(int, COMP);
	return 0;

    case INTP_LONG:
	COMP_LOOP(long, COMP);
	return 0;

    case INTP_FLOAT:
	COMP_LOOP(float, COMP);
	return 0;

    case INTP_DOUBLE:
	COMP_LOOP(double, COMP);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_min - Elementwise minimum of two arrays

All non-complex numeric types are supported, but the exact
characteristics and portability are inherited from the underlying C
implementation.
---------------------------------------------------------------------------*/

#define MIN(w,x,y)  w = ((x) < (y)) ? (x) : (y)

int
bips_min(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, MIN);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, MIN);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, MIN);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, MIN);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, MIN);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, MIN);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, MIN);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, MIN);
	return 0;

    case INTP_FLOAT:
	BINOP_LOOP(float, MIN);
	return 0;

    case INTP_DOUBLE:
	BINOP_LOOP(double, MIN);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_max - Elementwise maximum of two arrays

All non-complex numeric types are supported, but the exact
characteristics and portability are inherited from the underlying C
implementation.
---------------------------------------------------------------------------*/

#define MAX(w,x,y)  w = ((x) > (y)) ? (x) : (y)

int
bips_max(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, MAX);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, MAX);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, MAX);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, MAX);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, MAX);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, MAX);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, MAX);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, MAX);
	return 0;

    case INTP_FLOAT:
	BINOP_LOOP(float, MAX);
	return 0;

    case INTP_DOUBLE:
	BINOP_LOOP(double, MAX);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_add - Sum of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.

The destination stride may not be zero; use bips_addset to sum the
elements of a vector.

By special dispensation, bips_add handles the recursive case
correctly, even on an optimizing compiler.  For this case, the user
must specify ds == ss1 and dst == src1 + ss1; then the running sum
will be generated.  The second source stride ss2 may be zero.  It is
the caller's responsibility to set dst[-ss1] correctly.
---------------------------------------------------------------------------*/

#define ADD(w,x,y)  w = (x) + (y)

int
bips_add(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_RLOOP(unsigned char, ADD);
	return 0;

    case INTP_USHORT:
	BINOP_RLOOP(unsigned short, ADD);
	return 0;

    case INTP_UINT:
	BINOP_RLOOP(unsigned int, ADD);
	return 0;

    case INTP_ULONG:
	BINOP_RLOOP(unsigned long, ADD);
	return 0;

    case INTP_SCHAR:
	BINOP_RLOOP(signed char, ADD);
	return 0;

    case INTP_SHORT:
	BINOP_RLOOP(short, ADD);
	return 0;

    case INTP_INT:
	BINOP_RLOOP(int, ADD);
	return 0;

    case INTP_LONG:
	BINOP_RLOOP(long, ADD);
	return 0;

    case INTP_FLOAT:
	BINOP_RLOOP(float, ADD);
	return 0;

    case INTP_DOUBLE:
	BINOP_RLOOP(double, ADD);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_sub - Difference of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.
---------------------------------------------------------------------------*/

#define SUB(w,x,y)  w = (x) - (y)

int
bips_sub(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, SUB);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, SUB);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, SUB);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, SUB);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, SUB);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, SUB);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, SUB);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, SUB);
	return 0;

    case INTP_FLOAT:
	BINOP_LOOP(float, SUB);
	return 0;

    case INTP_DOUBLE:
	BINOP_LOOP(double, SUB);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_mul - Product of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.

The destination stride may not be zero; use bips_mulset to multiply
the elements of a vector.

By special dispensation, bips_mul handles the recursive case
correctly, even on an optimizing compiler.  For this case, the user
must specify ds == ss1 and dst == src1 + ss1; then the running product
will be generated.  The second source stride ss2 may be zero.  It is
the caller's responsibility to set dst[-ss1] correctly.
---------------------------------------------------------------------------*/

#define MUL(w,x,y)  w = (x) * (y)

int
bips_mul(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_RLOOP(unsigned char, MUL);
	return 0;

    case INTP_USHORT:
	BINOP_RLOOP(unsigned short, MUL);
	return 0;

    case INTP_UINT:
	BINOP_RLOOP(unsigned int, MUL);
	return 0;

    case INTP_ULONG:
	BINOP_RLOOP(unsigned long, MUL);
	return 0;

    case INTP_SCHAR:
	BINOP_RLOOP(signed char, MUL);
	return 0;

    case INTP_SHORT:
	BINOP_RLOOP(short, MUL);
	return 0;

    case INTP_INT:
	BINOP_RLOOP(int, MUL);
	return 0;

    case INTP_LONG:
	BINOP_RLOOP(long, MUL);
	return 0;

    case INTP_FLOAT:
	BINOP_RLOOP(float, MUL);
	return 0;

    case INTP_DOUBLE:
	BINOP_RLOOP(double, MUL);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_div - Quotient of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.
---------------------------------------------------------------------------*/

#define DIV(w,x,y)  w = (x) / (y)

int
bips_div(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, DIV);
	return 0;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, DIV);
	return 0;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, DIV);
	return 0;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, DIV);
	return 0;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, DIV);
	return 0;

    case INTP_SHORT:
	BINOP_LOOP(short, DIV);
	return 0;

    case INTP_INT:
	BINOP_LOOP(int, DIV);
	return 0;

    case INTP_LONG:
	BINOP_LOOP(long, DIV);
	return 0;

    case INTP_FLOAT:
	BINOP_LOOP(float, DIV);
	return 0;

    case INTP_DOUBLE:
	BINOP_LOOP(double, DIV);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_pow - Elementwise exponention of two arrays

Only float and double types are supported.

FIXME:  Complex and integer types are not yet supported.  It might be useful
to support exponentiation to small integral powers.
---------------------------------------------------------------------------*/

#define POW(w,x,y)  w = pow(x,y)

int
bips_pow(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	BINOP_LOOP(float, POW);
	return 0;

    case INTP_DOUBLE:
	BINOP_LOOP(double, POW);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}




/*--------------------------------------------------------------------------
bips_atan2 - Elementwise four-quadrant arc tangent of two arrays

Only float and double types are supported.

FIXME:  Complex types are not yet supported.  (Does this particular 
operation actually generalize to complex arguments?)
---------------------------------------------------------------------------*/

#define ATAN2(w,x,y)  w = atan2(x,y)

int
bips_atan2(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	BINOP_LOOP(float, ATAN2);
	return 0;

    case INTP_DOUBLE:
	BINOP_LOOP(double, ATAN2);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    
}



/***************************************************************************
Operations with an unusual format

Most of these are inner loops for heavily used functions that could be
done, more expensively, using the simpler operations above.
***************************************************************************/



/*------------------------------------------------------------------------
bips_absdiff - Absolute difference of two arrays

This function computes the absolute value of the difference between
two arrays and converts it to double; the second array may be omitted
(by supplying a zero pointer), in which case the absolute value of the
first array is computed.  Both source arrays must be of the same type.  
This function is intended to support the computation of vector norms, 
hence the somewhat different format.
-------------------------------------------------------------------------*/

#define ABSDIFF_LOOP(type)                     \
{                                              \
    type * restrict s1 = src1;                 \
    type * restrict s2 = src2;                 \
    if (s2 == 0) {                             \
        for (i = 0; i < nelem; i++)  {         \
            *d = fabs((double)*s1);            \
            d += ds; s1 += ss1; }  }           \
    else {                                     \
        for (i = 0; i < nelem; i++) {          \
            *d = fabs((double)*s1 - *s2);      \
            d += ds; s1 += ss1; s2 += ss2; } } \
}


int 
bips_absdiff(
    long nelem,         /* Number of elements to process */
    double *dst,        /* Destination array */
    long ds,            /* Stride in destination array */
    int stype,          /* Type of the source array */
    void *src1,         /* First source array */
    long ss1,           /* Stride in first source array */
    void *src2,         /* Second source array */
    long ss2)           /* Stride in second source array */
{
    double * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	ABSDIFF_LOOP(unsigned char);
	return 0;

    case INTP_USHORT:
	ABSDIFF_LOOP(unsigned short);
	return 0;

    case INTP_UINT:
	ABSDIFF_LOOP(unsigned int);
	return 0;

    case INTP_ULONG:
	ABSDIFF_LOOP(unsigned long);
	return 0;

    case INTP_SCHAR:
	ABSDIFF_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	ABSDIFF_LOOP(short);
	return 0;

    case INTP_INT:
	ABSDIFF_LOOP(int);
	return 0;

    case INTP_LONG:
	ABSDIFF_LOOP(long);
	return 0;

    case INTP_FLOAT:
	ABSDIFF_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	ABSDIFF_LOOP(double);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    


}



/*------------------------------------------------------------------------
bips_chamfer - Inner loop for chamfer distance

This functions computes the chamfer distance in one direction using
the following algorithm for each destination voxel, where DMAX is the
maximum possible value for the destination type.  If the source voxel
is non-zero, then the destination is set to zero; else if the
destination is less than the offset destination, the destination is
unchanged; else if the offset destination is DMAX, the destination is
set to DMAX; otherwise, the destination is set to the offset
destination plus 1.  Note that the most probable use of this function
requires the caller to initialize the destination to the maximum
possible distance value by hand before calling this function in each
direction starting with the second destination pixel.  If you didn't
understand all that, don't worry; it's unlikely that this code is
useful anywhere other than the vxl_chamfer function.
-------------------------------------------------------------------------*/

#define CHAMFER_LOOP(dtype, stype, dmax)    \
{                                           \
    long i;                                 \
    dtype * restrict d = dst;               \
    stype * restrict s = src;               \
    for (i = 0; i < nelem; i++) {           \
        if ( s[0] != 0 )                    \
            d[0] = 0;                       \
        else if ( d[0] <= d[doff] )         \
            /* Do nothing */ ;              \
        else if ( d[doff] == dmax )         \
            d[0] = dmax;                    \
        else                                \
            d[0] = d[doff] + 1;             \
        d += ds;   s += ss;                 \
        }                                   \
                                            \
}

int
bips_chamfer(
    long nelem,              /* Number of elements to process */
    int dtype,               /* Destination type: uchar or ushort */
    void *dst,               /* Destination array */
    long ds,                 /* Destination stride */
    long doff,               /* Offset to prev pixel */
    int stype,               /* Source type */
    void *src,               /* Source array */
    long ss)                 /* Source stride */
{

    /* Switch according to the destination type */
    switch (dtype) {

    case INTP_UCHAR:

        /* Dest is unsigned char; now switch according to the source type */
	switch (stype) {

	case INTP_UCHAR:
	    CHAMFER_LOOP(unsigned char, unsigned char, UCHAR_MAX);
	    return 0;

	case INTP_SCHAR:
	    CHAMFER_LOOP(unsigned char, signed char, UCHAR_MAX);
	    return 0;

	case INTP_USHORT:
	    CHAMFER_LOOP(unsigned char, unsigned short, UCHAR_MAX);
	    return 0;

	case INTP_SHORT:
	    CHAMFER_LOOP(unsigned char, short, UCHAR_MAX);
	    return 0;

	case INTP_UINT:
	    CHAMFER_LOOP(unsigned char, unsigned int, UCHAR_MAX);
	    return 0;

	case INTP_INT:
	    CHAMFER_LOOP(unsigned char, int, UCHAR_MAX);
	    return 0;

	case INTP_ULONG:
	    CHAMFER_LOOP(unsigned char, unsigned long, UCHAR_MAX);
	    return 0;

	case INTP_LONG:
	    CHAMFER_LOOP(unsigned char, long, UCHAR_MAX);
	    return 0;

	case INTP_FLOAT:
	    CHAMFER_LOOP(unsigned char, float, UCHAR_MAX);
	    return 0;

	case INTP_DOUBLE:
	    CHAMFER_LOOP(unsigned char, double, UCHAR_MAX);
	    return 0;

	default:
	    return ERR_ILLOP;

	}

    case INTP_USHORT:

        /* Dest is unsigned char; now switch according to the source type */
	switch (stype) {

	case INTP_UCHAR:
	    CHAMFER_LOOP(unsigned short, unsigned char, USHRT_MAX);
	    return 0;

	case INTP_SCHAR:
	    CHAMFER_LOOP(unsigned short, signed char, USHRT_MAX);
	    return 0;

	case INTP_USHORT:
	    CHAMFER_LOOP(unsigned short, unsigned short, USHRT_MAX);
	    return 0;

	case INTP_SHORT:
	    CHAMFER_LOOP(unsigned short, short, USHRT_MAX);
	    return 0;

	case INTP_UINT:
	    CHAMFER_LOOP(unsigned short, unsigned int, USHRT_MAX);
	    return 0;

	case INTP_INT:
	    CHAMFER_LOOP(unsigned short, int, USHRT_MAX);
	    return 0;

	case INTP_ULONG:
	    CHAMFER_LOOP(unsigned short, unsigned long, USHRT_MAX);
	    return 0;

	case INTP_LONG:
	    CHAMFER_LOOP(unsigned short, long, USHRT_MAX);
	    return 0;

	case INTP_FLOAT:
	    CHAMFER_LOOP(unsigned short, float, USHRT_MAX);
	    return 0;

	case INTP_DOUBLE:
	    CHAMFER_LOOP(unsigned short, double, USHRT_MAX);
	    return 0;

	default:
	    return ERR_ILLOP;

	}

    default:
	return ERR_ILLOP;
    }
}



/*------------------------------------------------------------------------
bips_moments - First and second moments of an array

This function computes the first and second non-central moments of a
one-dimensional vector; the index is taken as the space variable.  The
moments are always returned in double, regardless of the data format.
-------------------------------------------------------------------------*/

#define MOM_LOOP(type)             \
{                                  \
    type * restrict s = src;       \
    for (i = 0; i < nelem; i++) {  \
        d0 += (m  = (double)(*s)); \
        d1 += (m *= (double)i);    \
        d2 += (m *= (double)i);    \
        s += ss;  }                \
}

int
bips_moments(
    long nelem,             /* Number of elements in the array */
    double dst[3],          /* Zeroth, first, and second moments */
    int stype,              /* Type of the source array */
    void *src,              /* Source array */
    long ss)                /* Source stride */
{
    double d0, d1, d2;      /* Moments so far */
    double m;               /* Voxel value */
    long i;

    /* Initialize the moments to zero */
    d0 = d1 = d2 = 0.0;

    /* Switch on the source type */
    switch (stype) {

    case INTP_UCHAR:
	MOM_LOOP(unsigned char);
	break;

    case INTP_USHORT:
	MOM_LOOP(unsigned short);
	break;

    case INTP_UINT:
	MOM_LOOP(unsigned int);
	break;

    case INTP_ULONG:
	MOM_LOOP(unsigned long);
	break;

    case INTP_SCHAR:
	MOM_LOOP(signed char);
	break;

    case INTP_SHORT:
	MOM_LOOP(short);
	break;

    case INTP_INT:
	MOM_LOOP(int);
	break;

    case INTP_LONG:
	MOM_LOOP(long);
	break;

    case INTP_FLOAT:
	MOM_LOOP(float);
	break;

    case INTP_DOUBLE:
	MOM_LOOP(double);
	break;

    default:
	return ERR_ILLOP;
    }

    /* Save the computed moments and return */
    dst[0] = d0;
    dst[1] = d1;
    dst[2] = d2;
    return 0;
}



/*-------------------------------------------------------------------------
bips_conv1 - Inner loop of convolution

This function performs the innermost loop of a convolution.  Each source
element is converted to double, multiplied by a double coefficient, and
added to a double destination.

FIXME: Optimize for the case where the stride is one?  Or is there enough
floating point computation here that it doesn't matter?
-------------------------------------------------------------------------*/

#define CONV1_LOOP(type)             \
{                                    \
    type * restrict s = (type *)src; \
    for (i = 0; i < nelem; i++) {    \
        *d += coef*((double)*s);     \
        d += ds; s += ss;  }         \
}

int 
bips_conv1(
    long nelem,                   /* Number of elements to process */
    double *dst,                  /* Destination: convolution sum */
    long ds,                      /* Stride in destination array */
    int stype,                    /* Type of source data */
    void *src,                    /* Source array */
    long ss,                      /* Stride in source array */
    double coef)                  /* Coefficient to multiply by */
{
    double * restrict d = dst;
    long i;

    /* Switch on the source type */
    switch (stype) {

    case INTP_UCHAR:
	CONV1_LOOP(unsigned char);
	return 0;

    case INTP_USHORT:
	CONV1_LOOP(unsigned short);
	return 0;

    case INTP_UINT:
	CONV1_LOOP(unsigned int);
	return 0;

    case INTP_ULONG:
	CONV1_LOOP(unsigned long);
	return 0;

    case INTP_SCHAR:
	CONV1_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	CONV1_LOOP(short);
	return 0;

    case INTP_INT:
	CONV1_LOOP(int);
	return 0;

    case INTP_LONG:
	CONV1_LOOP(long);
	return 0;

    case INTP_FLOAT:
	CONV1_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	CONV1_LOOP(double);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	return ERR_ILLOP;    
    }
}



/*-------------------------------------------------------------------------
bips_conv2 - Finish up a convolution

This function takes a double array containing the convolution sums,
converts to the nearest valid value of a given type (rounding and
limiting as necessary), and stores the result.

FIXME: Optimize for the case where the stride is one?  Or is there enough
floating point computation here that it doesn't matter?
-------------------------------------------------------------------------*/

/* Must limit and round for integral destination types */
#define CONV2_LOOP(type,lo,hi)             \
{                                          \
    type * restrict d = (type *)dst;       \
    for (i = 0; i < nelem; i++) {          \
        *d = (type)LIMIT(*s, lo, hi);      \
        d += ds; s += ss;  }               \
}

/* No limiting or rounding for floating point destination types */
#define CONV2_FLOOP(type)             \
{                                     \
    type * restrict d = (type *)dst;  \
    for (i = 0; i < nelem; i++) {     \
        *d = (type)(*s);              \
        d += ds; s += ss;  }          \
}

int 
bips_conv2(
    long nelem,                   /* Number of elements to process */
    int dtype,                    /* Type of destination data */
    void *dst,                    /* Destination: convolution sum */
    long ds,                      /* Stride in destination array */
    double *src,                  /* Source array */
    long ss)                      /* Stride in source array */
{
    double * restrict s = src;
    long i;

    /* Switch on the destination type */
    switch (dtype) {

    case INTP_UCHAR:
	CONV2_LOOP(unsigned char, 0, UCHAR_MAX);
	return 0;

    case INTP_USHORT:
	CONV2_LOOP(unsigned short, 0, USHRT_MAX);
	return 0;

    case INTP_UINT:
	CONV2_LOOP(unsigned int, 0, UINT_MAX);
	return 0;

    case INTP_ULONG:
	CONV2_LOOP(unsigned long, 0, ULONG_MAX);
	return 0;

    case INTP_SCHAR:
	CONV2_LOOP(signed char, SCHAR_MIN, SCHAR_MAX);
	return 0;

    case INTP_SHORT:
	CONV2_LOOP(short, SHRT_MIN, SHRT_MAX);
	return 0;

    case INTP_INT:
	CONV2_LOOP(int, INT_MIN, INT_MAX);
	return 0;

    case INTP_LONG:
	CONV2_LOOP(long, LONG_MIN, LONG_MAX);
	return 0;

    case INTP_FLOAT:
	CONV2_FLOOP(float);
	return 0;

    case INTP_DOUBLE:
	CONV2_FLOOP(double);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    

}


/*-------------------------------------------------------------------------
bips_lostat1 - Inner loop of local statistics

This function performs the innermost loop of a local statistics
computation: summing the weights, values, and squared values over a
neighborhood.  The source stride is specified by the caller; the
destination stride is always 1.

FIXME: Optimize for the case where the source stride is one?  Or is
there enough floating point computation here that it doesn't matter?

ALGORITHM: The test (coef == 1.0) below is an exception to the usual
rule that you never test floats for exact equality; if coef is 1.0, it
will almost certainly have been created by assigning the value 1.0 and
will be exactly equal.  (However, computed values or values read in
from a file might not match exactly.)
-------------------------------------------------------------------------*/

#define LOSTAT1_LOOP(type)               \
{                                        \
    type * restrict s = (type *)src;     \
    if (coef == 1.0)                     \
        for (i = nelem; i > 0; i--) {    \
            sv = (double)*s;             \
            *d0++ += 1.0;                \
            *d1++ += sv;                 \
            *d2++ += sv*sv;              \
            s += ss;  }                  \
    else                                 \
        for (i = 0; i < nelem; i++) {    \
            sv = (double)*s;             \
            *d0++ += coef;               \
            *d1++ += coef*sv;            \
            *d2++ += coef*sv*sv;         \
            s += ss;  }                  \
}

int 
bips_lostat1(
    long nelem,                   /* Number of elements to process */
    double *d0,                   /* Sum of weights */
    double *d1,                   /* Sum of values */
    double *d2,                   /* Sum of squared values */
    int stype,                    /* Type of source data */
    void *src,                    /* Source array */
    long ss,                      /* Stride in source array */
    double coef)                  /* Weight to multiply by */
{
    double sv;                    /* A source value */
    long i;

    /* Switch on the source type */
    switch (stype) {

    case INTP_UCHAR:
	LOSTAT1_LOOP(unsigned char);
	return 0;

    case INTP_USHORT:
	LOSTAT1_LOOP(unsigned short);
	return 0;

    case INTP_UINT:
	LOSTAT1_LOOP(unsigned int);
	return 0;

    case INTP_ULONG:
	LOSTAT1_LOOP(unsigned long);
	return 0;

    case INTP_SCHAR:
	LOSTAT1_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	LOSTAT1_LOOP(short);
	return 0;

    case INTP_INT:
	LOSTAT1_LOOP(int);
	return 0;

    case INTP_LONG:
	LOSTAT1_LOOP(long);
	return 0;

    case INTP_FLOAT:
	LOSTAT1_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	LOSTAT1_LOOP(double);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	return ERR_ILLOP;    
    }
}



/*-------------------------------------------------------------------------
bips_lostat2 - Finish up a local statistics calculation

This function takes three double arrays containing the summed weights,
values and squared values, computes the corresponding means and
standard deviations, and stores them to the destination arrays.

FIXME: Allow the option of obtaining only the mean or variance, to
support applications that need only one of them?  The obvious way to
signal this is to use NULL for the appropriate destination.

FIXME: Optimize for the case where the destination stride is one?  Or
is there enough floating point computation here that it doesn't
matter?
-------------------------------------------------------------------------*/

/* Must limit and round for integral destination types */
#define LOSTAT2_LOOP(type,lo,hi)        \
{                                       \
    type * restrict d1 = (type *)dst1;  \
    type * restrict d2 = (type *)dst2;  \
    for (i = 0; i < nelem; i++) {       \
        w = *s0;                        \
        m = *s1 / w;                    \
        *d1 = (type)LIMIT(m,lo,hi);     \
        v = scale*sqrt( *s2/w - m*m );  \
        *d2 = (type)LIMIT(v,lo,hi);     \
        s0++; s1++; s2++;               \
        d1 += ds; d2 += ds; }           \
}



/* No limiting or rounding needed for floating point destination types */
#define LOSTAT2_FLOOP(type)               \
{                                         \
    type * restrict d1 = (type *)dst1;    \
    type * restrict d2 = (type *)dst2;    \
    for (i = 0; i < nelem; i++) {         \
        w = *s0;                          \
        *d1 = m = *s1 / w ;               \
        *d2 = scale*sqrt( *s2/w - m*m );  \
        s0++; s1++; s2++;                 \
        d1 += ds; d2 += ds; }             \
}

int 
bips_lostat2(
    long nelem,                   /* Number of elements to process */
    int dtype,                    /* Type of destination data */
    void *dst1,                   /* Destination for mean values */
    void *dst2,                   /* Destination for standard deviations */
    long ds,                      /* Stride in destination array */
    double *s0,                   /* Source for sums of weights */
    double *s1,                   /* Source for sums of values */
    double *s2,                   /* Source for sums of squared values */
    double scale)                 /* Multiplier to scale std deviations */
{
    double w, m, v;
    long i;

    /* Switch on the destination type */
    switch (dtype) {

    case INTP_UCHAR:
	LOSTAT2_LOOP(unsigned char, 0, UCHAR_MAX);
	return 0;

    case INTP_USHORT:
	LOSTAT2_LOOP(unsigned short, 0, USHRT_MAX);
	return 0;

    case INTP_UINT:
	LOSTAT2_LOOP(unsigned int, 0, UINT_MAX);
	return 0;

    case INTP_ULONG:
	LOSTAT2_LOOP(unsigned long, 0, ULONG_MAX);
	return 0;

    case INTP_SCHAR:
	LOSTAT2_LOOP(signed char, SCHAR_MIN, SCHAR_MAX);
	return 0;

    case INTP_SHORT:
	LOSTAT2_LOOP(short, SHRT_MIN, SHRT_MAX);
	return 0;

    case INTP_INT:
	LOSTAT2_LOOP(int, INT_MIN, INT_MAX);
	return 0;

    case INTP_LONG:
	LOSTAT2_LOOP(long, LONG_MIN, LONG_MAX);
	return 0;

    case INTP_FLOAT:
	LOSTAT2_FLOOP(float);
	return 0;

    case INTP_DOUBLE:
	LOSTAT2_FLOOP(double);
	return 0;

    case INTP_COMPLEX:
	return ERR_UNIMP;

    case INTP_DCOMPLEX:
	return ERR_UNIMP;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;    

}



/*--------------------------------------------------------------------------
bips_lookup - Process through lookup table

This function takes each element of the source array (which must be either
unsigned char or unsigned short), uses it as an index into a table of
values (which may be any type), and puts that entry from the table into
the output.  Neither the source nor the destination stride may be zero;
the source and destination may not overlap.

FIXME:  Should this function report a fatal error if the index is outside
the table, or just silently provide zero?
--------------------------------------------------------------------------*/

#define LUT_LOOP(type, xtype)                    \
{                                                \
    type  * restrict d = (type *)dst;            \
    type  * restrict t = (type *)table;          \
    xtype * restrict s = (xtype *)src;           \
    long i;                                      \
    for (i = 0; i < nelem; i++) {                \
        if (*s >= ntable)                        \
            panic("Index too large for table");  \
	*d = t[*s]; d += ds; s += ss; }          \
}

int
bips_lookup(
    long nelem,                 /* Number of elements to process */
    int type,                   /* Type of table and result */
    int xtype,                  /* Type of index array */
    void *dst,                  /* Destination array */
    long ds,                    /* Stride in destination array */
    void *src,                  /* Source array of index values */
    long ss,                    /* Stride in source array */
    void *table,                /* Table of values */
    long ntable)                /* Number of elements in the table */
{
    /* Check for valid strides */
    if (ds == 0 || ss == 0)
	panic("Source or destination stride of zero is not permitted");

    /* If unsigned char index, switch on the table type */
    if (xtype == INTP_UCHAR)
	switch (type) {
	    
	case INTP_UCHAR:
	    LUT_LOOP(unsigned char, unsigned char);
	    return 0;

	case INTP_SCHAR:
	    LUT_LOOP(signed char, unsigned char);
	    return 0;

	case INTP_USHORT:
	    LUT_LOOP(unsigned short, unsigned char);
	    return 0;

	case INTP_SHORT:
	    LUT_LOOP(short, unsigned char);
	    return 0;

	case INTP_UINT:
	    LUT_LOOP(unsigned int, unsigned char);
	    return 0;

	case INTP_INT:
	    LUT_LOOP(int, unsigned char);
	    return 0;

	case INTP_ULONG:
	    LUT_LOOP(unsigned long, unsigned char);
	    return 0;

	case INTP_LONG:
	    LUT_LOOP(long, unsigned char);
	    return 0;

	case INTP_FLOAT:
	    LUT_LOOP(float, unsigned char);
	    return 0;

	case INTP_DOUBLE:
	    LUT_LOOP(double, unsigned char);
	    return 0;

	default:
	    return ERR_UNIMP;

	}

    /* If unsigned short index, switch on the table type */
    else if (xtype == INTP_USHORT)
	switch (type) {

	case INTP_UCHAR:
	    LUT_LOOP(unsigned char, unsigned short);
	    return 0;

	case INTP_SCHAR:
	    LUT_LOOP(signed char, unsigned short);
	    return 0;

	case INTP_USHORT:
	    LUT_LOOP(unsigned short, unsigned short);
	    return 0;

	case INTP_SHORT:
	    LUT_LOOP(short, unsigned short);
	    return 0;

	case INTP_UINT:
	    LUT_LOOP(unsigned int, unsigned short);
	    return 0;

	case INTP_INT:
	    LUT_LOOP(int, unsigned short);
	    return 0;

	case INTP_ULONG:
	    LUT_LOOP(unsigned long, unsigned short);
	    return 0;

	case INTP_LONG:
	    LUT_LOOP(long, unsigned short);
	    return 0;

	case INTP_FLOAT:
	    LUT_LOOP(float, unsigned short);
	    return 0;

	case INTP_DOUBLE:
	    LUT_LOOP(double, unsigned short);
	    return 0;

	default:
	    return ERR_UNIMP;

	}

    /* All other index types are invalid */
    else
	return ERR_ILLOP;
}



/*--------------------------------------------------------------------------
bips_alpha - Generate integer and fractional parts of arithmetic sequence

This function generates an arithmetic sequence, separates each element
into its integer and fractional parts, and stores them in two arrays.
It is a useful operation for image resampling.  The integer part is
defined to be the floor, and the fractional part is defined as the
original number minus the floor.  The type of the fractional parts
stored must be either float or double; the type of the integer parts
stored is always long.  The base and increment of the sequence are
always double, since the added precision is almost always necessary
(and the cost is minor--if not negative--on most modern workstations).
The name of this function comes from the common use of the Greek
letter alpha to represent the fractional part that is computes.

FIXME: This doesn't seem to be the most useful form of this function; 
the affine resampling algorithms actually use something else.

FIXME: The floor() function is remarkably slow for gcc/i386; could
something faster be used instead?

FIXME: Write the more general (if less useful?) function to obtain the
integer and fractional parts of the elements of a vector.
---------------------------------------------------------------------------*/

#define ALPHA_LOOP(type)                 \
{                                        \
    type * restrict df = (type *)dsti;   \
    for (n = 0; n < nelem; n++) {        \
        ival = floor(val);               \
        *di  = (long)ival;               \
        *df  = val - ival;               \
        val += incr;                     \
        df  += dsf;  di += dsi; }        \
}

int
bips_alpha(
    long   nelem,           /* Number of elements to store (in each dest) */
    long  *dsti,            /* Array of integer parts */
    long   dsi,             /* Stride in the integer array */
    int    type,            /* Type to store in the fraction array */
    void  *dstf,            /* Array of fractional parts */
    long   dsf,             /* Stride in the fraction array */
    double base,            /* First value in the arithmetic sequence */
    double incr)            /* Increment in the arithmetic sequence */
{
    double val = base, ival;   /* Value and its integer part */
    long * restrict di = dsti; /* Location in integer array */
    long n;                    /* Index over elements */

    /* Switch according to the destination type */
    switch (type) {

    case INTP_FLOAT:
	ALPHA_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	ALPHA_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*--------------------------------------------------------------------------
bips_lerp - Linear interpolation 

This function performs the linear interpolation

   dst[n] += alpha[n]*(src[n]-dst[n]),

including a optional cast from the original type of the source to the
type of the destination.  The destination is NOT initialized; this
turns out to be more convenient for multi-dimensional interpolation.
The destination and alpha arrays must be the same type, either float
or double; the source may be of any type.  The results are
unpredictable if any of the destination, alpha, and source arrays
overlap.
--------------------------------------------------------------------------*/

#define LERP_INNER(dtype)                    \
{                                            \
    dtype * restrict d = (dtype *)dst;       \
    dtype * restrict a = (dtype *)alf;       \
    for (n = 0; n < nelem; n++) {            \
        d[0] += a[0]*((dtype)s[0] - d[0]);   \
        d += ds; a += as; s += ss;  }        \
}

#define LERP_LOOP(stype)               \
{                                      \
    stype * restrict s = (stype *)src; \
    if (datype == INTP_DOUBLE)         \
        LERP_INNER(double)             \
    else if (datype == INTP_FLOAT)     \
        LERP_INNER(float)              \
    else                               \
        return ERR_ILLOP;              \
}

int
bips_lerp(
    long nelem,            /* Number of elements to process */
    int datype,            /* Type of the destination and alpha arrays */
    void *dst,             /* Destination array */
    long ds,               /* Destination stride */
    void *alf,             /* Alpha array */
    long as,               /* Alpha stride */
    int stype,             /* Source type */
    void *src,             /* Source array */
    long ss)               /* Source stride */
{
    int n;                 /* Index over elements */

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	LERP_LOOP(unsigned char);
	return 0;

    case INTP_USHORT:
	LERP_LOOP(unsigned short);
	return 0;

    case INTP_UINT:
	LERP_LOOP(unsigned int);
	return 0;

    case INTP_ULONG:
	LERP_LOOP(unsigned long);
	return 0;

    case INTP_SCHAR:
	LERP_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	LERP_LOOP(short);
	return 0;

    case INTP_INT:
	LERP_LOOP(int);
	return 0;

    case INTP_LONG:
	LERP_LOOP(long);
	return 0;

    case INTP_FLOAT:
	LERP_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	LERP_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}



/*--------------------------------------------------------------------------
bips_glerp - Gather values and do linear interpolation 

This function gathers neighboring voxels from an n-dimensional
array, casts to double, and performs the linear interpolation

   dst[n] = (1.0-alpha[n])*src[index[n]] + alpha[n]*src[index[n]+1] .

This admittedly odd computation turns out to be useful in a fast
algorithm for N-dimensional linear interpolation (and may in fact be
useful only there!).  The results are unpredictable if any of the
destination, alpha, and source arrays overlap.
--------------------------------------------------------------------------*/

#define GLERP_LOOP(stype)                 \
{                                         \
    stype * restrict s = (stype *)src;    \
    stype * restrict si;                  \
    for (n = 0; n < nelem; n++) {         \
        si = s + index[n];                \
        s0 = si[0];                       \
        s1 = si[1];                       \
        d[0] = s0 + a[0] * (s1-s0);       \
        d += ds; a += as;  }              \
}

int
bips_glerp(
    long nelem,            /* Number of elements to process */
    double *dst,           /* Destination array */
    long ds,               /* Destination stride */
    double *alf,           /* Alpha array */
    long as,               /* Alpha stride */
    int stype,             /* Source type */
    void *src,             /* Source array */
    long *index)           /* Element indices into the source array */
{
    double * restrict d = dst;   /* Pointer to current destination */
    double * restrict a = alf;   /* Pointer to current alpha */
    int n;                       /* Index over elements */
    double s0, s1;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	GLERP_LOOP(unsigned char);
	return 0;

    case INTP_USHORT:
	GLERP_LOOP(unsigned short);
	return 0;

    case INTP_UINT:
	GLERP_LOOP(unsigned int);
	return 0;

    case INTP_ULONG:
	GLERP_LOOP(unsigned long);
	return 0;

    case INTP_SCHAR:
	GLERP_LOOP(signed char);
	return 0;

    case INTP_SHORT:
	GLERP_LOOP(short);
	return 0;

    case INTP_INT:
	GLERP_LOOP(int);
	return 0;

    case INTP_LONG:
	GLERP_LOOP(long);
	return 0;

    case INTP_FLOAT:
	GLERP_LOOP(float);
	return 0;

    case INTP_DOUBLE:
	GLERP_LOOP(double);
	return 0;

    default:
	/* Drop through to error return */ ;

    }

    return ERR_ILLOP;
}
