/***************************************************************************
cdata.c - Functions to read/modify/write commented data files

Author:  Paul Hughett

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


The functions in this file are intended to support programs that read,
modify, or write commented data files.

Commented data is a generic file format, or a set of rules for
formatting a file, which are intended to construct files that can be
easily read by either people or computers, copied from one computer
platform to another, and most difficult, modified by computers without
losing the information intended for people.

The computer-readable part of commented data consists of space- or
tab-delimited fields, with line breaks as a higher-order separation
between blocks of data.  The content of the data fields is arbitrary,
except that they may not contain spaces, tabs, newlines, backslashes,
or number signs; there is currently no quoting mechanism to include
these special characters in data fields, but one may be added later if
it seems useful enough.  A number sign # begins a comment that
continues up to the next line break.  A backslash, if the last
non-blank non-comment character on a line, hides the following line
break, effectively turning it into a space; backslashes within
comments have no special meaning.  The juxtaposition \# without an
intervening space should be avoided, since it will eventually mean an
escaped number sign; use \ # instead to mean a line continuation
followed by a comment.

The decomment(1) program can be used to remove comments and paste
together continuation lines and thus to prepare a commented data file
for SAS and other programs that do not understand commented data.

FIXME: The future probably holds the use of quote marks to delimit
data fields containing embedded spaces and the use of backslash to
escape individual characters, including quotes, backslashes, and
number signs; you should avoid writing any commented data files that
would break if these features are added.  But note that not all of
these can be successfully decommented.

NOTE:  The idiom isspace((int)c) which appears several times in this
file is used to prevent a spurious warning message "subscript has type
`char'" generated by gcc installed under Solaris.  The problem is that
isspace is defined as a macro which does a table lookup.

*****************************************************************************/

static const char rcsid[] = "$Id: cdata.c,v 1.9 2000/06/26 20:12:34 hughett Exp $ Release $Name:  $";

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "errm.h"
#include "memm.h"

#include "cdata.h"



/*---------------------------------------------------------------------------
read_long_line - Read an arbitrarily long line from a stream

This function reads an arbitrarily long line from a stream, allocates
enough memory to store it, forces the line to terminate with a NL in
the case where an EOF is seen first, and returns the line to the
caller as a string.  It is the caller's responsibility to free the
memory when finished.  If an error occurs or if no characters can be
read because the file is already at EOF, then NULL is returned.  A
maximum character count is provided, mostly to prevent filling up all
of memory with garbage in case of bugs or attempting to read a binary
file; an input line longer than this limit causes a fatal error.  If
the caller provides zero, then this function is permitted to keep
reading characters until malloc runs out of memory.

FIXME:  This function might prove useful elsewhere.
---------------------------------------------------------------------------*/

static char *             /* String terminated by NL and NUL */
read_long_line(
    FILE *input,          /* Stream to read from */
    size_t maxlen)        /* Maximum number of chars permitted, or zero */
{
    size_t lbuf = 32;     /* (Initial) length of line buffer */
    char *buf = 0;        /* Line buffer */
    char *next;           /* Place in buf to read the next block */
    size_t lnext;         /* Space left in the buffer */
    size_t len;           /* Total chars read so far */
    char *s = 0;

    /* If at EOF, just return NULL */
    if (feof(input)) return 0;

    /* Allocate an initial buffer */
    buf   = mallock(lbuf);
    next  = buf;
    lnext = lbuf;

    /* Read a chunk and enlarge buffer until large enough */
    while (1) {
	s = fgets(next, lnext, input);
	len = strlen(buf);

	/* Return NULL if no chars available before error or EOF */
	if (s == 0) {
	    free(buf);
	    return 0;
	}

	/* If buffer ends with NL, we're done; trim the string
	 and return it */
	if ( buf[len-1] == '\n') {
	    buf = reallock(buf, len+1);
	    return buf;
	}

	/* If we're now at EOF, insert an NL and return; this
	   might require enlarging the buffer. */
	if ( feof(input) ) {
	    if (len + 2 > lbuf)
		buf = reallock(buf, len+2);
	    buf[len  ] = '\n';
	    buf[len+1] = '\0';
	    return buf;
	}

	/* Otherwise enlarge the buffer and read another chunk */
	lbuf = 2*lbuf;
	if (maxlen > 0 && lbuf >= maxlen)
	    fatal("read_long_line: input line longer than %d chars", maxlen);
	buf  = reallock(buf, lbuf);
	next  =  buf + len;
	lnext = lbuf - len;
	continue;
    }

}




/*---------------------------------------------------------------------------
cdata_read - Read next computer line of a commented data file

This function reads the next line of a commented data file plus as
many continuation lines as are needed, separates the text into data
and comment fields, and returns all the various fields to the caller.
Memory is allocated as needed for and is expected to be freed by the
caller; cdata_free can be used for this purpose.

This function returns the number of data fields found; or zero to
indicate that a single comment field but no data was found; or EOF to
indicate either end of file or i/o error.  (The current version
signals a fatal error for an i/o error but this might change in the
future; write your code to do the right thing in either case.)

The semantics of this function can be described by the following 
algorithm (although the actual implementation is slightly different).  
Successive lines are read from the input and processed until at least 
one line containing non-blank data has been found and the last line 
does not contain a continuation marker.  If a number sign # is found 
on the line, it and all following characters including the terminating 
newline are marked as comments.  If a backslash \ is the last 
non-whitespace character in the non-comment portion of the line,
another line will always be read; the backslash and any following 
characters are also marked as comments.  Any whitespace characters 
preceding the backslash, number sign, or end of the line are also 
marked as comments.  The remainder of the line (if any) is divided 
into fields separated by spaces or tabs; the spaces or tabs are 
associated with the following non-blank data.  Once all the necesary 
lines have been, they can be seen as a sequence of comment and data 
fields and which both starts and ends with a comment field.  Adjacent 
comment fields are merged into a single comment field; any internal 
newlines are retained.  Data fields may contain leading but not trailing 
whitespace; they may not contain internal newlines.

If both backslash and number sign are used, they should be separated
by at least one whitespace character; the sequence "\#" may someday
become a quoted number sign rather than a continuation followed by
a comment.

Note that this function returns one more comment field than it does
data fields, because the text reads both begins and ends with a
comment field.  The number of data fields is the value returned, so
that the inevitable error will not cause a segfault or worse.

FIXME:  A data field may not be continued onto another line.  The line
break separates two data fields even if a continuation marker is present.
Unfortunately, this probably cannot be fixed without breaking the ability
to preserve comments while modifying the data.

FIXME:  The arrays allocated for comment and data fields are reallocated
for each new field read; this is simpler to program but might be rather
inefficient.

---------------------------------------------------------------------------*/


int                       /* Returns: Number of data found, or 0, or EOF */
cdata_read(
    FILE *input,          /* Stream to read arguments from */
    char **dat[],         /* Array of pointers to the data */
    char **cmt[])         /* Comment preceding each data field */
{
    int  ndat;            /* Number of data fields found */
    size_t ldat = 0;      /* Allocated slots for data fields */
    size_t lstr = 0;      /* Allocated space for current comment */
    char *str = 0;        /* Current comment being built */
    int  more;            /* Continuation market seen on this line */
    char *p;              /* Pointer to a line read from the input */
    char *p1;             /* Pointer to the beginning of a comment */
    char *pd, *pe;        /* Pointers to beginning and end of a data field */
    char *s;
    int len;

    /* Preset returned values to no data, no comments found*/
    ndat  = 0;
    *dat  = 0;
    *cmt  = 0;

    /* Return immediately if at EOF; die if error on this input */
    if (feof(input))
	return EOF;
    if (ferror(input))
	fatal("cdata_read: I/O error reading input file");


    /* Create the initial cmt and dat arrays */
    *cmt = mallock(sizeof(char **));
    (*cmt)[0] = 0;
    ldat = 1;
    *dat = mallock(ldat * sizeof(char **));

    /* Read and process lines until done */
    for ( more = 1; more || ndat == 0; ) {
	more = 0;

	/* Read an input line */
	p = read_long_line(input, 1024);

	/* Handle EOF or I/O error */
	if (p == NULL && feof(input))  /* EOF => done */
	    return EOF;
	else if (p == NULL)            /* I/O error is fatal */
	    fatal("cdata_read: I/O error reading input file");

	/* Find the comment on this line, if any */
	for ( p1 = p; *p1 != 0 && *p1 != '#'; p1++ ) ;

	/* Find preceding whitespace and the continuation marker, if any */
	for ( ; p1 > p && isspace((int)p1[-1]); p1-- ) ;
	if (p1[-1] == '\\')  {
	    p1--;
	    more = 1; 
	}
	for ( ; p1 > p && isspace((int)p1[-1]); p1-- ) ;

	/* Separate the beginning, if any, into data fields */
	for ( pd = p; pd < p1; pd = pe ) {

	    /* Find beginning and end of next data field */
	    for ( pe = pd; pe < p1 && isspace((int)*pe); pe++ ) ;
	    for ( ; pe < p1 && !isspace((int)*pe); pe++ ) ;
	    len = pe - pd;

	    /* If any data is actually present... */
	    if ( len > 0 && !isspace((int)*(pe-1)) ) {

		/* Save this data field for the caller */
		s = mallock(len+1);
		strncpy(s, pd, len);
		s[len] = 0;
		*dat = reallock(*dat, (++ldat)*sizeof(char *));
		(*dat)[ndat++] = s;

		/* Append an empty comment to the list */
		*cmt = reallock(*cmt, (ldat+1)*sizeof(char *));
		(*cmt)[ndat] = 0;
	    }

	}

	/* If there's a comment on this line, append it */
	if ( *p1 != 0 ) {

	    /* Create a new string buffer if necessary */
	    if ( (*cmt)[ndat] == 0 ) {
		lstr = strlen(p1);
		str = mallock(lstr+1);
		(*cmt)[ndat] = str; 
		strcpy(str, p1);
	    }

	    /* Otherwise enlarge the old buffer */
	    else {
		lstr += strlen(p1);
		(*cmt)[ndat] = str = reallock(str, lstr+1);
		strcat(str, p1);
	    }

	}

	/* Free the buffer used for the input line */
	free(p);
    }

    return ndat;
}


/*---------------------------------------------------------------------------
cdata_free - Free the storage allocated by cdata_read

This function frees the storage allocated by cdata_read.
---------------------------------------------------------------------------*/
void
cdata_free(
    int  ndat,           /* Number of arguments found */
    char *dat[],         /* Array of pointers to the data */
    char *cmt[])         /* Comment preceding each datum */
{
    int i;

    if (dat != 0) {
	for ( i = 0; i < ndat; i++ ) 
	    if (dat[i] != 0)  free( dat[i] );
	free (dat);
    }

    if (cmt != 0) {
	for ( i = 0; i<= ndat; i++ )
	    if (cmt[i] != 0 ) free (cmt[i] );
	free (cmt);
    }

    return;	
}



/*---------------------------------------------------------------------------
cdata_write - Write next computer line of a commented data file

----------------------------------------------------------------------------*/

void
cdata_write(
    FILE *stream,         /* Stream to read arguments from */
    int  ndat,            /* Number of arguments found */
    char *dat[],          /* Array of pointers to the data */
    char *cmt[])          /* Comment preceding each datum */
{
    int i;

    if (cmt && cmt[0])       fputs(cmt[0],   stream);
    for ( i = 0; i < ndat; i++ ) {
	if (dat && dat[i]  ) fputs(dat[i],   stream);
	if (cmt && cmt[i+1]) fputs(cmt[i+1], stream);
    }

    return;
}



/*---------------------------------------------------------------------------
cdata_decomment - Remove comments and write next line of a commented data file

This function writes out the next line of a commented data file, as read by
cdata_read, but without any of the comments.  It is intended primarily to
convert commented data files into a format that can easily be read by older
programs that don't understand commented data files.
----------------------------------------------------------------------------*/

void
cdata_decomment(
    FILE *stream,         /* Stream to write arguments to */
    int  ndat,            /* Number of arguments found */
    char *dat[],          /* Array of pointers to the data */
    char *cmt[])          /* Comment preceding each datum */
{
    int i;

    if (dat == 0) return;

    for ( i = 0; i < ndat; i++ )
	if (dat[i]) {
	    if (!isspace((int)dat[i][0])) fputs(" ", stream);
	    fputs(dat[i], stream);
	}
    fputs("\n", stream);

    return;
}



/*---------------------------------------------------------------------------
cdata_print - Print results of cdata_read, usually for debugging
---------------------------------------------------------------------------*/

void
cdata_print(
    FILE *stream,         /* Stream to print arguments to */
    int  ndat,            /* Number of arguments found */
    char *dat[],          /* Array of pointers to the data */
    char *cmt[])          /* Comment preceding each datum */
{
    int i;

    fprintf(stream, "\nResults of cdata_read:\n");
    fprintf(stream, "ndat = %d\n", ndat);
    fprintf(stream, "cmt[ 0] = |%s|\n", cmt[0]);
    for (i = 0; i < ndat; i++) {
	fprintf(stream, "dat[%2d] = |%s|\n", i,   dat[i]);
	fprintf(stream, "cmt[%2d] = |%s|\n", i+1, cmt[i+1]);
    }

    return;
}
