/****************************************************************************
clap.c - C wrapper functions for the LAPACK library

Author:  Paul Hughett

Copyright (c) 2000-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file provides C wrapper functions for the LAPACK Fortran subroutines
used by the Voxel Kit.  (There is no reason why other LAPACK subroutines
could not be included, but no reason to, until they are needed.)  These
functions have been gathered into a separate file since calling Fortran
from C is rather messy, especially when supporting multiple platforms.
Having all the mess in the same place should make porting to additional
platforms a little bit easier.

For purposes of describing the C functions, we will suppose that the
successive elements of a matrix are written out from left to right,
and then top to bottom; typographical rows are called "rows", and
columns are called "columns."  Thus the usual C statement for
initializing a matrix will show rows and columns in the natural way
simply by inserting returns at the end of each row.  Upper and lower
triangle matrices are defined in the usual fashion relative to this
typographical representation.

The nomenclature "leading dimension" is retained, even though it
appears as a trailing dimension in C.  Actually, we could change that
to "last dimension" and retain the acronym; there are now a few
instances of that usage in the code below.

This initial implementation is designed to support the LAPACK and BLAS
rpm packages provided on the RedHat Powertools disk for Linux.  Other
platforms will be added as they become necessary.

Likewise, the selection of LAPACK subroutines supported is driven by
current needs of the BBLimage project; other subroutines will be added
as needed.

The likely issues to deal with in porting this to another platform are
the following:

1. The Fortran implementation may mangle subroutine names.  For example,
the LAPACK subroutine DSYEV may appear to C as dsyev_, with an added
trailing underscore.  Other implementations are alleged to use a leading
underscore, or do no mangling.

2.  The sizes of integer and floating point types may differ between C
and Fortran.  Here we assume (temporarily) that int <-> integer, 
float <-> real, and double <-> double precision.  Scalars are no problem;
arrays are harder because they may require the allocation of temporary
storage for the conversion.

3.  The semantics of strings differ between C and Fortran, and suitable
mangling may be required to pass the CHARACTER*1 values that LAPACK
interprets to control its operation.  Here we assume (temporarily) that
it suffices for C to pass a pointer to char.

4.  LAPACK and BLAS do not use COMMON, so we don't have to figure out
how to translate _that_ between C and Fortran.


Some other issues for interfacing C and Fortran are, as far as I know,
independent of the platform.

1.  Fortran stores arrays in column-major order, with the first subscript
varying most rapidly as you consider successive elements in memory.  C
stores arrays in row-major order, with the last subscript varying most
rapidly.  In effect, passing an array between languages transposes it.

2.  Fortran passes arguments by reference (with the possible exception
of string arguments, which are sometimes passed by descriptor); C
normally passes arguments by value, but can achieve the effect of
passing by reference by passing a pointer to the actual value.

4. Some implementations of Fortran (e.g. f2c) expect a Fortran main
program to link to.  If the main program is in C, it may be necessary
to provide a dummy Fortran main program.

5. Different implementations of Fortran may require that different
libraries be linked into the program.


Some issues concerning the wrappers and how thick they are:

1.  Fortran name mangling.

2.  Differences between C and Fortran types, and the required conversions
(which may need to be applied to arrays and strings).

3.  Hiding the allocation of work storage from the C programmer, without
being too inefficient.

4.  Hiding transposition of arrays.  (Rewriting descriptions and doing limited
adaptations within the functions.)

5.  Arrays taken raw, or with voxel array headers.

6.  Type of data encoded in function name or passed a parameter a la BIPS?

7.  Provide raw LAPACK subroutines or nicer but less efficient versions?

8.  Use special compressed forms like LU in a single array or return separate
arrays?  Support special matrix formats like symmetric, banded, etc, or 
just plain arrays?

9.  Fortran origin is 1, C origin is zero.


The current thinking is to provide one wrapper layer which is just thick
enough to be portable across platforms for the efficiency-minded, and 
another layer which does everything nicely for the simplicity-minded.

There are also some LAPACK-specific issues:

1.  On encountering an error, most LAPACK subroutines call the subroutine
XERBLA.  The standard library implementation of this subroutine prints
and error message and aborts the program.  This is not acceptable for
many LAPACK users, so it may be necessary to override the normal XERBLA
implementation; whether or not this is possible, and how it can be done
may be platform-dependent.

2.  Many LAPACK functions require working storage; FORTRAN's limitations
require that this storage be allocated by the caller and passed to the
LAPACK subroutine.  Implementing this in C requires that the C wrapper
allocate the necessary storage via malloc, or automatic allocation, which
is a nuisance and possibly inefficient.


Other functions to add: matrix multiply, inverse, condition estimates,
svd, LU, solution of equations, ...

FIXME: This code assumes the memm memory management functions.  Is this
appropriate?  If not, what is?

FIXME: Try this code on the Suns and Tru64 and make any changes needed
to support those platforms as well.  Don't forget to install LAPACK
and BLAS first.  Modify the configure script to detect the platform
and do the right thing.
*****************************************************************************/

static const char rcsid[] = "$Id: clap.c,v 1.7 2001/10/05 21:23:14 hughett Exp $ Release $Name:  $";


#include <stdio.h>
#include <stdlib.h>

#include "clap.h"
#include "memm.h"


/*-------------------------------------------------------------------------
Dummy MAIN__

The f2c library (used on Linux, at least) expect to find a function
MAIN__ which is the Fortran main program.  We have the main program
in C, but must still provide this function.
---------------------------------------------------------------------------*/

int MAIN__(void);       /* A prototype, just to placate fussy compilers */

int
MAIN__()
{
    fprintf(stderr, "Called the dummy Fortran MAIN program\n");
    abort();
}


/*------------------------------------------------------------------------
XERBLA replacement

Tested this on Linux; it seems to override the library-supplied XERBLA
without any problems!!

FIXME: Does it work on other platforms?

FIXME: Get the arguments right, and add the prototype to clap.h.  Note that
this is a call from Fortran to C, reversing the usual case in this file.
-------------------------------------------------------------------------*/

void xerbla_(char *, int *);   /* A prototype for fussy compilers */

void
xerbla_(char *sbname, int *info)
{
    fprintf(stderr, "Called the substitute xerbla!\n");
    return;
}


/***************************************************************************
BLAS Wrappers
***************************************************************************/


/*---------------------------------------------------------------------------
clap_dgemm - Compute product of two general matrices
clap_sgemm - Compute product of two general matrices

This wrapper function calls the BLAS routine DGEMM or SGEMM to compute
the matrix expression

    C = alpha*op(A)*op(B) + beta*C

where op() may be either the matrix or its inverse.  All of A, B, and
C are general matrices stored in the C row-major format; the right
magic is done inside this function to translate to the FORTRAN
conventions.  op(A) is assumed to be m by k; op(B) is assumed to be k
by n; and C is assumed to be m by n.

ALGORITHM: The translation into FORTRAN conventions is accomplished
simply by reversing the order of A and B in the multiplication.

FIXME: Rewrite and move to matrix.c.
----------------------------------------------------------------------------*/

int
clap_dgemm(
    int transa,           /* Transpose A? */
    int transb,           /* Transpose B? */
    int m,                /* First dimension of A and C */
    int n,                /* Last dimension of B and C */
    int k,                /* Last dimen of A and first dimen of B */
    double alpha,         /* Scalar coefficient for A*B */
    double *a,            /* First matrix */
    int lda,              /* Leading dimension of A */
    double *b,            /* Second matrix */
    int ldb,              /* Leading dimension of B */
    double beta,          /* Scalar coefficient for C */
    double *c,            /* Destination/third matrix */
    int ldc)              /* Leading dimension of C */
{
    fchar1 ftra;              /* Transpose A? */
    fchar1 ftrb;              /* Transpose B? */
    fint fm, fn, fk;          /* Matrix dimensions */
    fint flda, fldb, fldc;    /* Leading dimensions */
    fdouble falpha, fbeta;    /* Coefficients */

    /* Translate C arguments into Fortran arguments */
    ftra = transa ? 'T' : 'N';
    ftrb = transb ? 'T' : 'N';
    fm = m;
    fn = n;
    fk = k;
    flda = lda;
    fldb = ldb;
    fldc = ldc;
    falpha = alpha;
    fbeta  = beta;

    /* Call BLAS to do the real work (Reverse order of A and B) */
    dgemm_(&ftrb, &ftra, &fn, &fm, &fk, &falpha, b, &fldb, a, &flda,
	  &fbeta, c, &fldc);

    /* Always return 0 for success */
    return 0;    
}

int
clap_sgemm(
    int transa,           /* Transpose A? */
    int transb,           /* Transpose B? */
    int m,                /* First dimension of A and C */
    int n,                /* Last dimension of B and C */
    int k,                /* Last dimen of A and first dimen of B */
    float alpha,          /* Scalar coefficient for A*B */
    float *a,             /* First matrix */
    int lda,              /* Leading dimension of A */
    float *b,             /* Second matrix */
    int ldb,              /* Leading dimension of B */
    float beta,           /* Scalar coefficient for C */
    float *c,             /* Destination/third matrix */
    int ldc)              /* Leading dimension of C */
{
    fchar1 ftra;              /* Transpose A? */
    fchar1 ftrb;              /* Transpose B? */
    fint fm, fn, fk;          /* Matrix dimensions */
    fint flda, fldb, fldc;    /* Leading dimensions */
    freal falpha, fbeta;      /* Coefficients */

    /* Translate C arguments into Fortran arguments */
    ftra = transa ? 'T' : 'N';
    ftrb = transb ? 'T' : 'N';
    fm = m;
    fn = n;
    fk = k;
    flda = lda;
    fldb = ldb;
    fldc = ldc;
    falpha = alpha;
    fbeta  = beta;

    /* Call BLAS to do the real work (Reverse order of A and B) */
    sgemm_(&ftrb, &ftra, &fn, &fm, &fk, &falpha, b, &fldb, a, &flda,
	  &fbeta, c, &fldc);

    /* Always return 0 for success */
    return 0;    
}



/**************************************************************************
LAPACK wrappers
**************************************************************************/


/*---------------------------------------------------------------------------
clap_dsyev - Compute eigenvalues and vectors of real symmetric matrix
clap_ssyev - Compute eigenvalues and vectors of real symmetric matrix

This wrapper function calls DSYEV or SSYEV to compute the eigenvalues
and optionally the eigenvectors of a real symmetric matrix.  If both
is zero, only the eigenvalues are computed; otherwise both eigenvalues
and eigenvectors are computed.  The matrix to be solved is given as
argument a.  If up is non-zero, the upper triangle of a is used;
otherwise the lower triangle is used.  If both is non-zero, then the
eigenvectors are returned as the rows of A.  If not, then either the
upper or lower triangle of A is destroyed; the other triangle is
untouched.  The eigenvalues of A are returned in the array W.  The
return value is the INFO value returned by DSYEV.  A zero value
indicates success, non-zero values indicate failure; for a further
explanation of the possible non-zero values, see the DSYEV
documentation.

FIXME: I believe that this function is now obsolete, having been replaced
by matrix_eigen_symmetric.

FIXME: Verify that the various minor details of the description have
been correctly translated from Fortran to C.

FIXME: It appears that, on exit, the rows of A contain the
eigenvectors of the original A.  Confirm this.  Also check if the
eigenvector matrix is consistently proper or improper.

FIXME: Can the real version be generated from the double version, or
possibly both from some more concise representation?
----------------------------------------------------------------------------*/

int
clap_dsyev(
    int both,           /* 0 => eigenvalues only; 1 => both */
    int up,             /* 0 => lower triangle; 1 => upper triangle */
    int n,              /* Order of the matrix A */
    double *a,          /* The matrix A, with dimensions [n, lda] */
    int lda,            /* Leading dimension of A */
    double *w)          /* Eigenvalues of A, with dimension [n] */
{
    fchar1 jobz;        /* Job selector */
    fchar1 uplo;        /* Upper or lower triangle */
    fint fn;            /* Order of matrix A */
    fint flda;          /* Leading dimension of A */
    fint info;          /* Status returned from LAPACK */
    fint lwork;         /* Size of working storage */
    fdouble *work;      /* Working storage for LAPACK */
    fdouble wsize;      /* Recommended size of working storage */

    /* Convert C arguments into Fortran arguments */
    /* FIXME: Almost certainly platform-specific */
    /* FIXME: Definitely assumes that double <-> double precision */
    jobz = both ? 'V' : 'N';
    uplo = up   ? 'L' : 'U';    /* Remember the transpose! */
    fn   = n;
    flda = lda;

    /* Allocate working storage */
    lwork = -1;
    dsyev_(&jobz, &uplo, &fn, a, &flda, w, &wsize, &lwork, &info);
    lwork = (fint)wsize;
    work = mallock(lwork * sizeof(fdouble));

    /* Call the LAPACK subroutine */
    dsyev_(&jobz, &uplo, &fn, a, &flda, w, work, &lwork, &info);

    /* Free working storage and return */
    free(work);

    return info;
}

int
clap_ssyev(
    int both,           /* 0 => eigenvalues only; 1 => both */
    int up,             /* 0 => lower triangle; 1 => upper triangle */
    int n,              /* Order of the matrix A */
    float *a,           /* The matrix A, with dimensions [n, lda] */
    int lda,            /* Leading dimension of A */
    float *w)           /* Eigenvalues of A, with dimension [n] */
{
    fchar1 jobz;        /* Job selector */
    fchar1 uplo;        /* Upper or lower triangle */
    fint fn;            /* Order of matrix A */
    fint flda;          /* Leading dimension of A */
    fint info;          /* Status returned from LAPACK */
    fint lwork;         /* Size of working storage */
    freal *work;        /* Working storage for LAPACK */
    freal wsize;        /* Recommended size of working storage */


    /* Convert C arguments into Fortran arguments */
    /* FIXME: Almost certainly platform-specific */
    /* FIXME: Definitely assumes that double <-> double precision */
    jobz = both ? 'V' : 'N';
    uplo = up   ? 'L' : 'U';    /* Remember the transpose! */
    fn   = n;
    flda = lda;

    /* Allocate working storage */
    lwork = -1;
    ssyev_(&jobz, &uplo, &fn, a, &flda, w, &wsize, &lwork, &info);
    lwork = (fint)wsize;
    work = mallock(lwork * sizeof(freal));

    /* Call the LAPACK subroutine */
    ssyev_(&jobz, &uplo, &fn, a, &flda, w, work, &lwork, &info);

    /* Free working storage and return */
    free(work);

    return info;
}



