/**************************************************************************
conseg.c - Concordance of two segmented images

Author:  Paul Hughett

Copyright (c) 1999 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This program computes the concordance of two segmented images from
their bivariate histogram; the concordance is defined as the fraction
of voxels that have the same classification in both images.  The
bivariate histogram is read from the file named on the command line or
from standard input and must be in the format produced by vibibhist,
optionally with comments according to the commented data file format.
The output from this program is in the commented data format and looks
like the following:

256335     93.0        # Total number of voxels and percent same
193284     85.3        # Same, excluding voxels zero in either image

**************************************************************************/

static const char rcsid[] = "$Id: conseg.c,v 1.4 1999/01/23 21:05:42 hughett Exp $ Release $Name:  $";

#include <stdio.h>
#include <stdlib.h>

#include "cdata.h"
#include "errm.h"
#include "memm.h"


int
main(int argc, char *argv[])
{
    FILE *input = stdin;      /* File to read, or standard input */
    int ndat;                 /* Number of data on current input line */
    char **dat;               /* Array of data fields */
    char **cmt;               /* Array of comment fields */
    int irow, icol;           /* Count rows and columns of the histogram */
    int ncol;                 /* Number of columns in histogram */
    int rgray;                /* Gray level for current row of histogram */
    int *cgray;               /* Gray level for each column of histogram */
    long cell;                /* Value in a cell of the histogram */
    long count;               /* Sum of all cells in histogram */
    long diag;                /* Sum of diagonal cells in histogram */
    long count0;              /* Sum of all cells with gray == 0 */
    long diag0;               /* Sum of diag cells with gray == 0 */

    /* Open the input file; if none, use standard input */
    if (argc >= 2) {
	input = fopen(argv[1], "r");
	if (input == 0)
	    fatal("Unable to open input file %s", argv[1]);
    }

    /* Read the column headings; rmbr gray level for each column */
    ndat = cdata_read(input, &dat, &cmt);
    if (ndat < 0)
	fatal("No data found in histogram file");
    ncol = ndat;
    cgray = mallock(ncol*sizeof(int));
    cgray[0] = -1;       /* Trap refs to column containing gray values */
    for (icol = 1; icol < ndat; icol++)
	cgray[icol] = atoi(dat[icol]);
	
    /* Read the rows of the histogram and add up the counts */
    count = diag = count0 = diag0 = 0;
    for (irow = 1; !feof(input); irow++) {
	ndat = cdata_read(input, &dat, &cmt);
	if (ndat == 0) break;
	if (ndat != ncol)
	    fatal("Too many or too few columns in histogram row %d", irow);
	rgray = atoi(dat[0]);
	for (icol = 1; icol < ncol; icol++) {
	    cell = atol(dat[icol]);
	    count += cell;
	    if (rgray == cgray[icol])
		diag += cell;
	    if (rgray == 0 || cgray[icol] == 0)
		count0 += cell;
	    if (rgray == 0 && rgray == cgray[icol])
		diag0 += cell;
	}
	cdata_free(ndat, dat, cmt);
	
    }

    /* Compute and output the concordances */
    printf("%8ld  %8.3f\n", count, (double)diag/count);
    printf("%8ld  %8.3f\n", count-count0,
	   (double)(diag-diag0)/(count-count0));

    exit(EXIT_SUCCESS);
}
