/**************************************************************************
dstring.c - Functions for dynamically allocated string

Author:  Paul Hughett

Copyright (c) 1999 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


These functions provide a rudimentary capability within C for strings
with dynamically allocated memory, so that the programmer does not
have to constantly concern himself with malloc and friends.  However,
the implementation here has never really been tested or debugged and
you use it at your own risk.  (I originally wrote it to support
cdata.c but it turned out to be just about as easy to do the memory
allocation explicitly in that code.)  There is also a risk is that
when I take this up again for some other application, I will change
all the definitions and calling sequences.

**************************************************************************/

static const char rcsid[] = "$Id: dstring.c,v 1.3 1999/01/21 19:00:42 hughett Exp $ Release $Name:  $";

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "errm.h"
#include "memm.h"

#include "dstring.h"



/*---------------------------------------------------------------------------
dstrcreate - Create an empty dynamic string.

This function creates a new dynamics string that has no storage allocated
to it and which contains nothing.  It returns a pointer to the dstring.
----------------------------------------------------------------------------*/

dstring *
dstrcreate()
{
    dstring *s;

    s = mallock(sizeof(dstring));
    s->alloc  = 0;
    s->string = 0;

    return s;
}


/*---------------------------------------------------------------------------
dstrdestroy - Destroy a dynamic string

This function destroys a dynamic string and returns all its memory to
the heap.  It is permissible to destroy an empty string but it is a
fatal error if the dynamic string has never been allocated.
---------------------------------------------------------------------------*/

void
dstrdestroy(
    dstring *s)           /* The string to be destroyed */
{
    if (s == 0)
        fatal("dstrdestroy: Attempt to destroy a non-existent string");

    if (s->string != 0)
        free(s->string);
    free(s);

    return;
}



/*---------------------------------------------------------------------------
dstralloc - Allocate space in a dynamic string

This function allocates sufficient space in a dynamic string to
contain at least the specified number of characters, plus a
terminating NUL.  If the current space allocation exceeds the space
requested, the allocation is reduced.  It is a fatal error if the
requested space is not available.
---------------------------------------------------------------------------*/

void
dstralloc(
    dstring *s,              /* String to allocate space in */
    size_t n)                /* Number of characters to allocate */
{

    if (s == 0)
        fatal("dstralloc:  Attempt to modify a non-existent string");

    if (n == 0 && s->string != 0) {
        free(s->string);
	s->alloc  = 0;
	s->string = 0;
    }
    else if (n > 0) {
        s->string = reallock(s->string, n);
    }

    return;
}



/*---------------------------------------------------------------------------
dstrextract - Extract the string from a dynamic string

This function returns the string contained within a dynamic string, and
sets the contents of the dynamic string to be empty.  It is useful for
passing a string which has been built up in a dstring to other routines,
and making the dstring available for other uses.  If the dstring is empty,
a null pointer is returned; if it contains an empty string, that empty
string is returned.  It is a fatal error if the dstring has never been
created.
---------------------------------------------------------------------------*/

char *
dstrextract(
    dstring *s)                /* Dynamic string to extract from */
{
    char *t;

    if (s == 0)
        fatal("dstrextract: Attempt to access a non-existent string");

    t = s->string;
    s->alloc  = 0;
    s->string = 0;

    return t;
}
   


/*---------------------------------------------------------------------------
dstrcat - Concatenate two dynamic strings

This function concatenates two strings, rellocating space in the
destination as necessary.  The dstring type is used.

FIXME:  Allow the src to be a plain (char *) string instead?
---------------------------------------------------------------------------*/

dstring *
dstrcat(
    dstring *dest,
    const dstring *src)
{
    size_t len;

    /* FIXME:  Suppose that src or dest is null, or points to an
       unallocated string.  Suppose that realloc fails. */

    /* Count memory needed and enlarge string area if needed */
    len = strlen(dest->string) + strlen(src->string) + 1;
    if (dest->alloc < len) {
	dest->string = reallock(dest->string, len);
	dest->alloc  = len;
    }

    /* Actually concatentate the strings */
    strcat(dest->string, src->string);

    return dest;
}




/*---------------------------------------------------------------------------
augment_string_array - Add a string to an array of strings

This function adds a string to an array of strings, rellocating the
array if necessary to make room.  The argument array is a pointer to a
cell containing a pointer to an array of pointers to char arrays; for
example, if foo is declared as char *foo[32], then the appropriate
argument would be &foo.  If array contains NULL, then a short array
will be allocated.

WARNING: This function has never been used, tested, or debugged.  (I
thought that it would be useful in cdata_read, but it turned out to be
just as easy to do the right things in line.)  Use at your own risk.
It might also be useful elsewhere (once debugged, of course).
---------------------------------------------------------------------------*/

void
augment_string_array(
    char   **array[],        /* Pointer to array of pointers to strings */
    size_t *nstring,         /* Number of strings currently in array */
    size_t *nalloc,          /* Number of strings allocated */
    char   *string)          /* New string to be added to array */
{
    /* If no array allocated, allocate enough space for two strings */
    if (*array == 0) {
	*nalloc = 2;
	*array = mallock((*nalloc)*sizeof(char *));
	*nstring = 0;
    }

    /* Reallocate array if necessary to make room */
    else if (*nstring >= *nalloc) {
	*nalloc += 1;
	*array = reallock(*array, (*nalloc)*sizeof(char *));
    }

    /* Now add the string to the array */
    *array[*nstring] = string;
    *nstring += 1;

    return;
}
