/****************************************************************************
errm.c - Error manager

Author:  Paul Hughett

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file provides a standard interface for reporting program errors
within software written at the Brain Behavior Laboratory at the
University of Pennsylvania.  Three levels of error are provided.
A warning indicates that the operation is anomalous but that continuation
is possible.  A fatal error indicates that the operation has failed and
cannot continue but that there is no apparent reason that some other
operation could not be attempted instead.  A panic indicates that there
appears to be damage to static data structures and that the prudent user
will save his files and exit as soon as possible.

Reasonable responses to an error:
Ignore it.
Print a message on stderr and continue.
Print a message and abort the program, possibly dumping core for debugging.
Do a longjump to the main command loop to try another interactive command.
Convert to a Python exception and unwind the stack.

FIXME: Add the capability for the user to define his own handler for
an error; pre-define some common options as listed above.

FIXME:  Include a machine-interpretable error tag to indicate the error
type to some recovery mechanism?

FIXME:  There doesn't seem to be a portable and robust way to convert
a fatal() error into a Python exception without passing error flags
all the way back up the call stack.
*****************************************************************************/

static const char rcsid[] = "$Id: errm.c,v 1.2 2000/01/26 18:57:02 hughett Exp $ Release $Name:  $";


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "errm.h"


/*------------------------------------------------------------------------
global flags for errm
------------------------------------------------------------------------*/

int errm_abort_flag = 1;       /* Abort (and core dump) on error? */



/*------------------------------------------------------------------------
warning - Indicate a non-fatal error condition

Generically speaking, this function is used to indicate a non-fatal
error condition, that does not prohibit continuation of the involved
function.  This specific implementation prints a warning message and
returns to the caller; other implementations might find it useful to
treat such errors differently.

FIXME:  As a debugging option, abort on warnings so that they can be
easily found and debugged.
-------------------------------------------------------------------------*/

void
warning(
    const char *message,      /* Message as a format string */
    ...)                      /* Values as needed */
{
    va_list ap;

    va_start(ap, message);
    fprintf(stderr, "Warning: ");
    vfprintf(stderr, message, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    return;
}



/*------------------------------------------------------------------------
fatal - Report a fatal error condition

Generically speaking, this function is used to indicate an error
condition which makes it impossible for the calling function to
proceed.  This specific implementation treats such errors by printing
an error message and either aborting the program (and dumping core) or
terminating with a non-zero flag.  Other implementations, such as in a
command-driven program, might prefer to abort only the current command
but allow the user to continue with a new command.

FIXME: Not fully implemented yet; this version just screams and dies.
It ought in general to be a long jump to the main command loop, so
that the current command fails, but the program can continue.
-------------------------------------------------------------------------*/

void
fatal(
    const char *message,      /* Message as a format string */
    ...)                      /* Values as needed */
{
    va_list ap;

    va_start(ap, message);
    fprintf(stderr, "Error: ");
    vfprintf(stderr, message, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    if (errm_abort_flag)
	abort();
    else
	exit(EXIT_FAILURE);
}



/*------------------------------------------------------------------------
panic - Report a panic error condition

Generically speaking, this function is used to indicate that permanent
data structures appear to be incorrect and that it may be dangerous to
continue program operation.  This particular implementation prints an
error message and either aborts the program (usually causing a core
dump) or terminates with a non-zero flag.  Other implementations might
prefer to do some clean-up or emergency shutdown procedures first.
-------------------------------------------------------------------------*/

void
panic(
    const char *message,      /* Message as a format string */
    ...)                      /* Values as needed */
{
    va_list ap;

    va_start(ap, message);
    fprintf(stderr, "Panic: ");
    vfprintf(stderr, message, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    if (errm_abort_flag)
	abort();
    else
	exit(EXIT_FAILURE);
}
