/****************************************************************************
exim.c - Export or import data in various external representations

Author:  Paul Hughett

Copyright (c) 2000-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file defines functions for converting between various internal
and external data representations; it is part of the Voxel Kit, a set
of general-purpose functions for volume image processing, with a bias
toward medical imaging applications.

FIXME: Think though, implement, and document the error handling
policy for these functions.

FIXME: The exim code probably has many other applications and should
be made easy to use independently.  Maybe even make into its own
package.

FIXME: Consider specifying the endianness as part of the external type
by adding a large constant (e.g. 256) to one of these codes to indicate
big-endian.

FIXME: Add other traits for types, including integral?, floating-point?,
complex?, min and max values, accumulation type, ...

FIXME: Possibly add functions to indicate which conversions are lossy.
Possibly provide some way to indicate the "natural" endianness of
external types such as VAX formats.  Add conversions for such things
as ANSI float, VAX float, etc as needed.  Consider adding specific
conversions of single items between given types; this will be handy
for people who need to convert only single items or who don't care
about efficiency.

FIXME: Add exim_label to cleanly handle labels for vector fields?  Add
exim_struct to nicely handle allocation for structs, as distinct from
vectors?  Include macros in exim.h so that file descriptions are short
enough to make nice tables in an 80-column line? (See bblanz for an
example.)  Add a string function to handle null-terminated strings?
Test this code on the NRIA header format, which is a lot messier than
bblanz.
*****************************************************************************/

static const char rcsid[] = "$Id: exim.c,v 1.21 2001/11/14 16:25:28 hughett Exp $ Release $Name:  $";

#include <config.h>         /* Local configuration per autoconf */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "errm.h"
#include "exim.h"
#include "memm.h"

#define MAXBYTES  16      /* Max bytes per data element */


/* Internal functions */
int exim_imdump(FILE *dump, void *indata, int intype, char *fmt,
    unsigned char *exdata, int extype, int bigend);



/*--------------------------------------------------------------------------
external_types - Table of external type characteristics

NOTES:  The entries in this table will often be found by indexing with
the absolute value of the type code; it is *essential* that the order
of entries match the order of the codes.  A prudent programmer using
the index will also verify the code against the intype value.

This table *must* be kept in sync with the type code definitions in
exim.h.

FIXME: The natural internal type has been chosen assuming that the
native signed integral formats use the two's complement format.  It
will require modification for other platforms.

NOTE:  Nextypes and Nintypes have type size_t, so a type code must be
converted to avoid warning about comparing signed and unsigned; this is
sort of ugly, but seems to be unavoidable.
--------------------------------------------------------------------------*/

static struct extype {
    int extype;                /* Numeric code for type */
    char *name;                /* Short name of type */
    int size;                  /* Number of bytes per element */
    int natural;               /* "Natural" internal type to convert to */
    char *desc;                /* One-line description of type */
}
external_types[] = {
    { EXTP_NONE,      "(none)",     0, INTP_NONE,
      "undefined or unspecified type" },
    { EXTP_UINT1,     "uint1",      1, INTP_UCHAR,
      "1-byte unsigned integer" },
    { EXTP_UINT2,     "uint2",      2, INTP_USHORT,
      "2-byte unsigned integer" },
    { EXTP_UINT4,     "uint4",      4, INTP_ULONG,
      "4-byte unsigned integer" },
    { EXTP_INT1,      "int1",       1, INTP_SCHAR,
      "1-byte two's complement integer" },
    { EXTP_INT2,      "int2",       2, INTP_SHORT,
      "2-byte two's complement integer" },
    { EXTP_INT4,      "int4",       4, INTP_LONG,
      "4-byte two's complement integer" },
    { EXTP_FLOAT4,    "float4",     4, INTP_FLOAT,
      "4-byte IEEE-754 float" },
    { EXTP_FLOAT8,    "float8",     8, INTP_DOUBLE,
      "8-byte IEEE-754 float" },
    { EXTP_COMPLEX8,  "complex8",   8, INTP_COMPLEX,
      "complex as two 4-byte IEEE-754 floats" },
    { EXTP_COMPLEX16, "complex16", 16, INTP_DCOMPLEX,
      "complex as two 8-byte IEEE-754 floats" },
};

#define Nextypes  (sizeof(external_types) / sizeof(external_types[0]))



/*--------------------------------------------------------------------------
internal_types - Table of internal type characteristics

NOTE:  The entries in this table will often be found by indexing with
the absolute value of the type code; it is *essential* that the order
of entries match the order of the codes.  A prudent programmer using
the index will also verify the code against the intype value.  

NOTE that the natural external type of an int or unsigned int will be
2 bytes long on a platform where sizeof(int) == 2 and 4 bytes long
otherwise.

This table *must* be kept in sync with the type code definitions in
exim.h.
--------------------------------------------------------------------------*/

static struct intype {
    int intype;                /* Numeric code for type */
    char *name;                /* Short name of type */
    int size;                  /* Number of bytes per element */
    int natural;               /* "Natural" external type to convert to */
    char *desc;                /* One-line description of type */
}
internal_types[] = {
    { INTP_NONE,      "(none)",     0, EXTP_NONE,
      "undefined or unspecified type" },
    { INTP_UCHAR,    "uchar",    sizeof(unsigned char),  EXTP_UCHAR,
      "unsigned char" },
    { INTP_USHORT,   "ushort",   sizeof(unsigned short), EXTP_UINT2,
      "unsigned short" },
#if SIZEOF_INT == 2
    { INTP_UINT,     "uint",     sizeof(unsigned int),   EXTP_UINT2,
      "unsigned int" },
#else
    { INTP_UINT,     "uint",     sizeof(unsigned int),   EXTP_UINT4,
      "unsigned int" },
#endif
    { INTP_ULONG,    "ulong",    sizeof(unsigned long),  EXTP_UINT4,
      "unsigned long" },
    { INTP_SCHAR,    "schar",    sizeof(signed char),    EXTP_INT1,
      "signed char" },
    { INTP_SHORT,    "short",    sizeof(short),          EXTP_INT2,
      "short" },
#if SIZEOF_INT == 2
    { INTP_INT,      "int",      sizeof(int),            EXTP_INT2,
      "int" },
#else
    { INTP_INT,      "int",      sizeof(int),            EXTP_INT4,
      "int" },
#endif
    { INTP_LONG,     "long",     sizeof(long),           EXTP_INT4,
      "long" },
    { INTP_FLOAT,    "float",    sizeof(float),          EXTP_FLOAT4,
      "float" },
    { INTP_DOUBLE,   "double",   sizeof(double),         EXTP_FLOAT8,
      "double" },
    { INTP_COMPLEX,  "complex",  2*sizeof(float),        EXTP_COMPLEX8,
      "float complex" },
    { INTP_DCOMPLEX, "dcomplex", 2*sizeof(double),       EXTP_COMPLEX16,
      "double complex"},
};

#define Nintypes  (sizeof(internal_types) / sizeof(internal_types[0]))



/*-------------------------------------------------------------------------
exim_all_extypes - Return array of all defined external type codes

This function returns an array which includes all the external type
codes currently defined.  The functions exim_sizeof_type,
exim_nameof_type, exim_describe_type, etc can be used to determine
the characterstics of the defined types.  The actual number of defined
codes is returned as the value of the function, even if this is more
than the length of the destination array provided by the caller; the
number of values stored is never more than the length of the array.
Thus calling with an array size of zero can be used to determine how
much space needs to be allocated; calling with array == 0 has the same
effect.
--------------------------------------------------------------------------*/

int                             /* Returns number of type codes defined */
exim_all_extypes(
    int *array,                 /* Destination array provided by caller */
    int size)                   /* Length of destination array */
{
    int nsize;                  /* Smaller of size and actual count */
    int i;

    nsize = Nextypes;
    if (size < nsize) nsize = size;

    if ( array != 0 && nsize > 0 )
	for (i = 0; i < nsize; i++) {
	    array[i] = external_types[i].extype;
	}

    return Nextypes;
}



/*-------------------------------------------------------------------------
exim_all_intypes - Return array of all defined internal type codes

This function returns an array which includes all the internal type
codes currently defined.  The functions exim_sizeof_type,
exim_nameof_type, exim_describe_type, etc can be used to determine
the characteristics of the defined types.  The actual number of defined
codes is returned as the value of the function, even if this is more
than the length of the destination array provided by the caller; the
number of values stored is never more than the length of the array.
Thus calling with an array size of zero can be used to determine how
much space needs to be allocated; calling with array == 0 has the same
effect.
--------------------------------------------------------------------------*/

int                             /* Returns number of type codes defined */
exim_all_intypes(
    int *array,                 /* Destination array provided by caller */
    int size)                   /* Length of destination array */
{
    int nsize;                  /* Smaller of size and actual count */
    int i;

    nsize = Nintypes;
    if (size < nsize) nsize = size;

    if ( array != 0 && nsize > 0 )
	for (i = 0; i < nsize; i++) {
	    array[i] = internal_types[i].intype;
	}

    return Nintypes;
}



/*-------------------------------------------------------------------------
exim_sizeof_type - Bytes per datum for a given exim data type

Given the numeric code for a data type, this function returns the
number of bytes per datum used for that data type.  If the type is not
recognized, zero is returned.  Both internal and external types are
handled.  Either of the functions exim_sizeof_intype or
exim_sizeof_extype will verify that the type code is a valid
internal/external code and return its size; they thus provide better error
checking when only internal or external types are expected.
--------------------------------------------------------------------------*/

int
exim_sizeof_type(
    int type)
{
    /* External types have positive numeric codes */
    if (type > 0 && (size_t)type < Nextypes ) {
	if (external_types[type].extype != type)
	    panic("error in external_types table");
	else
	    return external_types[type].size;
    }

    /* Internal types have negative numeric codes */
    else if ( type < 0 && (size_t)(-type) < Nintypes) {
	if (internal_types[-type].intype != type)
	    panic("error in internal_types table");
	else
	    return internal_types[-type].size;
    }

    /* Return zero for any type not recognized */
    else
	return 0;

    return 0;     /* Unnecessary, but keeps compiler happy */
}




/*-------------------------------------------------------------------------
exim_sizeof_intype - Bytes per datum for a given internal data type

Given the numeric code for an internal data type, this function
returns the number of bytes per datum used for that data type.  If the
type is not recognized, zero is returned.  Only internal types are
recognized, so this function doubles as a validity check for an
internal type code.
--------------------------------------------------------------------------*/

int
exim_sizeof_intype(
    int type)
{
    /* Internal types have negative numeric codes */
    if ( type < 0 && (size_t)(-type) < Nintypes) {
	if (internal_types[-type].intype != type)
	    panic("error in internal_types table");
	else
	    return internal_types[-type].size;
    }

    /* Return zero for any type not recognized */
    else
	return 0;

    return 0;     /* Unnecessary, but keeps compiler happy */
}




/*-------------------------------------------------------------------------
exim_sizeof_extype - Bytes per datum for a given external data type

Given the numeric code for an external data type, this function
returns the number of bytes per datum used for that data type.  If the
type is not recognized, zero is returned.  Only external types are
recognized, so this function doubles as a validity check for an
internal type code.
--------------------------------------------------------------------------*/

int
exim_sizeof_extype(
    int type)
{
    /* External types have positive numeric codes */
    if (type > 0 && (size_t)type < Nextypes ) {
	if (external_types[type].extype != type)
	    panic("error in external_types table");
	else
	    return external_types[type].size;
    }

    /* Return zero for any type not recognized */
    else
	return 0;

    return 0;     /* Unnecessary, but keeps compiler happy */
}



/*------------------------------------------------------------------------
exim_integral_type - Check if an exim type is an integral type

Given the numeric code of a type defined by exim, this function
returns 1 if the type permits only integral values, and 0 otherwise.
If an invalid type code is provided, then a fatal error occurs.  Note
that an unsigned long (and on 16-bit platforms, an unsigned int)
cannot be stored in a long variable, so this function cannot be used
to test for that.

FIXME: Currently supports only internal types.  This may actually be all
that is required in applications; let's wait and see.
--------------------------------------------------------------------------*/

int
exim_integral_type(
    int type)            /* Type code to check */
{
    switch (type) {

    case INTP_UCHAR:
    case INTP_SCHAR:
    case INTP_USHORT:
    case INTP_SHORT:
    case INTP_UINT:
    case INTP_INT:
    case INTP_ULONG:
    case INTP_LONG:
	return 1;

    case INTP_FLOAT:
    case INTP_DOUBLE:
    case INTP_COMPLEX:
    case INTP_DCOMPLEX:
	return 0;

    default:
	fatal("Illegal type code in exim_integral_type");
	return 0;    /* Not needed but keeps compiler happy */
    }
}



/*------------------------------------------------------------------------
exim_floating_type - Check if an exim type is a floating point type

Given the numeric code of a type defined by exim, this function
returns 1 if the type uses a floating point representation, and 0
otherwise.  If an invalid type code is provided, then a fatal error
occurs.

FIXME: Decide whether complex types are counted as floating point types
sometime before complex types are actually implemented.  For the moment
they are, but it's not clear that that's the right choice.

FIXME: Currently supports only internal types.  This may actually be all
that is required in applications; let's wait and see.

FIXME: Add the corresponding exim_complex_type() function.
--------------------------------------------------------------------------*/

int
exim_floating_type(
    int type)            /* Type code to check */
{
    switch (type) {

    case INTP_UCHAR:
    case INTP_SCHAR:
    case INTP_USHORT:
    case INTP_SHORT:
    case INTP_UINT:
    case INTP_INT:
    case INTP_ULONG:
    case INTP_LONG:
	return 0;

    case INTP_FLOAT:
    case INTP_DOUBLE:
    case INTP_COMPLEX:
    case INTP_DCOMPLEX:
	return 1;

    default:
	fatal("Illegal type code in exim_floating_type");
	return 0;    /* Not needed but keeps compiler happy */
    }
}



/*-------------------------------------------------------------------------
exim_nameof_type - Return short name of a given exim data type

Given the numeric code of a data type defined by exim, this function
returns a null-terminated string containing a one-word human-readable
name for the type.  The string "(invalid)" is returned if the type
is not recognized.
--------------------------------------------------------------------------*/

const char *
exim_nameof_type(
    int type)
{
    /* External types have positive numeric codes */
    if (type > 0 && (size_t)type < Nextypes ) {
	if (external_types[type].extype != type)
	    panic("error in external_types table");
	else
	    return external_types[type].name;
    }

    /* Internal types have negative numeric codes */
    else if ( type < 0 && (size_t)(-type) < Nintypes) {
	if (internal_types[-type].intype != type)
	    panic("error in internal_types table");
	else
	    return internal_types[-type].name;
    }

    /* "Unspecified" type has code 0 */
    else if ( type == 0 )
	return external_types[type].name;

    /* Return error string for any type not recognized */
    else
	return "(invalid)";

    return 0;     /* Unnecessary, but keeps compiler happy */
}



/*-------------------------------------------------------------------------
exim_describe_type - Return description of a given exim data type

Given the numeric code of a data type defined by exim, this function
returns a null-terminated string containing a one-line human-readable
description of the type.  The string "(invalid type code)" is returned
if the type is not recognized.
--------------------------------------------------------------------------*/

const char *
exim_describe_type(
    int type)
{
    /* External types have positive numeric codes */
    if (type > 0 && (size_t)type < Nextypes ) {
	if (external_types[type].extype != type)
	    panic("error in external_types table");
	else
	    return external_types[type].desc;
    }

    /* Internal types have negative numeric codes */
    else if ( type < 0 && (size_t)(-type) < Nintypes) {
	if (internal_types[-type].intype != type)
	    panic("error in internal_types table");
	else
	    return internal_types[-type].desc;
    }

    /* Undefined/unspecified type has zero code */
    else if ( type == 0 )
	return external_types[type].desc;

    /* Return error string for any type not recognized */
    else
	return "(invalid type code)";

    return 0;     /* Unnecessary, but keeps compiler happy */
}



/*-------------------------------------------------------------------------
exim_natural_intype - Return natural internal type for an external type

Given the numeric code of an external data type defined by exim, this function
returns the "natural" internal type to which it should normally be
converted.  Note that the natural internal type for a given external
type may differ between platforms, or between different installations
on the same platform.  (In other words, use this function to get the
internal type each time you need it; don't write code that depends on the
internal type remaining constant.)
--------------------------------------------------------------------------*/

int
exim_natural_intype(
    int type)
{
    /* External types have postive numeric codes */
    if (type > 0 && (size_t)type < Nextypes ) {
	if (external_types[type].extype != type)
	    panic("error in external_types table");
	else
	    return external_types[type].natural;
    }

    /* Return undefined type as error for any type not recognized */
    else
	return INTP_NONE;

    return 0;     /* Unnecessary, but keeps compiler happy */
}



/*-------------------------------------------------------------------------
exim_natural_extype - Return natural external type for an internal type

Given the numeric code of an internal data type defined by exim, this
function returns the "natural" external type to which it should
normally be converted.  Note that the natural external type for a
given internal type may differ between platforms, or between different
installations on the same platform.  (In other words, use this
function to get the external type each time you need it; don't write
code that depends on the external type remaining constant.)

It should be noted that it can be very dangerous to let the external
type be chosen automatically; you are likely to wind up with a data
file that you don't know the format of and cannot read back correctly.
--------------------------------------------------------------------------*/

int
exim_natural_extype(
    int type)
{
    /* Internal types have negative numeric codes */
    if (type < 0 && (size_t)(-type) < Nintypes ) {
	if (internal_types[-type].intype != type)
	    panic("error in internal_types table");
	else
	    return internal_types[-type].natural;
    }

    /* Return undefined type as error for any type not recognized */
    else
	return EXTP_NONE;

    return 0;     /* Unnecessary, but keeps compiler happy */
}



/*---------------------------------------------------------------------
exim_error - Check for and report any errors

This function checks the err status returned from exim_scalar or
exim_vector and takes the appropriate action if it's not zero.  In
general, values that cannot be successfully be converted are reported
as warnings, but errors that appear to be programming errors
(including unimplemented operations) are handled as fatal errors and
terminate the program.  The field argument need be non-null only when
a particular, named field is being converted.
---------------------------------------------------------------------*/

void
exim_error(
    char *field,               /* Name of the field being processed, or zero */
    int err)                   /* Error code returned from exim */
{
    switch (err) {

    /* No error, no action needed */
    case ERR_OKAY:
	break;

    /* Warning if data element cannot be converted */
    case ERR_RANGE:
        if (field != 0)
            warning("Unable to convert contents of %s field", field);
        else 
            warning("Unable to convert contents of (unspecified) field");
	break;

    /* Fatal error for unimplemented operations */
    case ERR_UNIMP:
	fatal("Unimplemented operation");
	break;

    /* Fatal error for end of file or i/o error */
    case ERR_EOF:
    case ERR_FILE:
	fatal("I/O error or unexpected end of file");
	break;

    /* Fatal error for illegal operation */
    case ERR_ILLOP:
	fatal("Illegal operation requested");
	break;

    /* Catch-all for anything missed above */
    default:
	fatal("Error number %d in exim", err);
    }

    return;
}



/*------------------------------------------------------------------------
exim_get_value - Get array element converted to a double

This function gets a specified element from an array of values in
some specified internal type, converts it to double, and returns
it.  The internal type may not be complex.  No error checking is
done except to verify that the given internal type is a valid type.

FIXME:  Can we do something better than dying for an invalid type?

FIXME:  This might be a good candidate for inlining, although it does
seem a little long for that optimization.  If so, its definition
probably needs to be moved into the header file.
------------------------------------------------------------------------*/

#define EXIM_GET_VALUE(type)   { type *s = src; v = s[n]; } 

double
exim_get_value(
    void *src,      /* Array to extract element from */
    int   stype,    /* Type of data in array */
    long  n)        /* Element number to get (from 0) */
{
    double v;

    switch (stype) {

    case INTP_UCHAR:
	EXIM_GET_VALUE(unsigned char);
	return v;

    case INTP_USHORT:
	EXIM_GET_VALUE(unsigned short);
	return v;

    case INTP_UINT:
	EXIM_GET_VALUE(unsigned int);
	return v;

    case INTP_ULONG:
	EXIM_GET_VALUE(unsigned long);
	return v;

    case INTP_SCHAR:
	EXIM_GET_VALUE(signed char);
	return v;

    case INTP_SHORT:
	EXIM_GET_VALUE(short);
	return v;

    case INTP_INT:
	EXIM_GET_VALUE(int);
	return v;

    case INTP_LONG:
	EXIM_GET_VALUE(long);
	return v;

    case INTP_FLOAT:
	EXIM_GET_VALUE(float);
	return v;

    case INTP_DOUBLE:
	EXIM_GET_VALUE(double);
	return v;

    default:
	/* Fall through to error handling */ ;
    }

    fatal("Invalid internal type");
    return ERR_ILLOP;  /* Not used but keeps compiler from complaining */
}



/*-------------------------------------------------------------------------
exim_set_value - Convert from double and store into an array element

This function converts a double value into a value of some specified
internal type and stores that value into an array; if the destination
is an integral type, the value is rounded to the nearest integer.  The
value is not checked against the range of the destination type.  The
internal type may not be complex.  No error checking is done except to
verify that the given internal type is a valid type.

Prior to 2000-12-05, this function was called exim_put_value; the name
was changed for consistency with the practice in Python.

FIXME: Should we check against the range of the destination type?

FIXME:  Can we do something better than dying for an invalid type?

FIXME:  This might be a good candidate for inlining, although it does
seem a little long for that optimization.  If so, the definition needs
to be moved into the header file.
-------------------------------------------------------------------------*/

#define EXIM_SET_RVALUE(type)  { type *d = dst; d[n] = value; } 
#define EXIM_SET_IVALUE(type)  { type *d = dst; d[n] = floor(value + 0.5); } 

void
exim_set_value(
    void *dst,             /* Array to place value into */
    int   dtype,           /* Type of the destination array */
    long  n,               /* Element number to set (from 0) */
    double value)          /* Value to set to */
{
    switch (dtype) {

    case INTP_UCHAR:
	EXIM_SET_IVALUE(unsigned char);
	break;

    case INTP_USHORT:
	EXIM_SET_IVALUE(unsigned short);
	break;

    case INTP_UINT:
	EXIM_SET_IVALUE(unsigned int);
	break;

    case INTP_ULONG:
	EXIM_SET_IVALUE(unsigned long);
	break;

    case INTP_SCHAR:
	EXIM_SET_IVALUE(signed char);
	break;

    case INTP_SHORT:
	EXIM_SET_IVALUE(short);
	break;

    case INTP_INT:
	EXIM_SET_IVALUE(int);
	break;

    case INTP_LONG:
	EXIM_SET_IVALUE(long);
	break;

    case INTP_FLOAT:
	EXIM_SET_RVALUE(float);
	break;

    case INTP_DOUBLE:
	EXIM_SET_RVALUE(double);
	break;

    default:
	fatal("Invalid internal type");
    }

    return;
}



/*-----------------------------------------------------------------
exim_swap_bytes - Change byte order

Given a byte string containing count blocks of nbin bytes each, this
function writes out those blocks with nbout bytes each by rearranging
the input bytes and possibly inserting zero bytes or omitting input
bytes.  Note that this is a generic function for an arbitrary new byte
order; in many common cases, there may be a specific byte-swapping
routine that will be faster.  The order of the bytes in the output is
specified by an integer array inbyte in which the ith entry specifies
the source of the ith output byte in each block.  If the ith entry in
this array is zero or positive, then it is the number of the input
byte which should be used as the ith output byte; if it is negative,
then the value of ctab[1-inbyte[i]] is used for the ith output byte.
By special dispensation, supplying a null pointer for ctab will set
the output byte to zero.

The results are unpredictable if the source and destination buffers
overlap or are the same.

TECHNICAL NOTES:  We assume that the compiler is smart enough to do
strength reduction and the other simple loop optimizations and write
the loops to be understood by a human.  For efficiency, we loop over
the output byte numbers in the outer loop, and over blocks in the
inner loop.  The inner loop should not be too long, or the cache will
miss and slow things down.  

FIXME: Block the outer loop to approximate the cache size so that the
caller doesn't have to.
-------------------------------------------------------------------*/

void
exim_swap_bytes(
    long count,            /* Number of blocks to process */
    unsigned char *dest,   /* Destination buffer */   
    unsigned char *src,    /* Source buffer */
    int nbout,             /* Number of bytes per output block */
    int nbin,              /* Number of bytes per input block */
    int *inbyte,           /* Input byte number for each output byte */
    unsigned char *ctab)   /* Table of constant values */
{
    int i;                 /* Index over output bytes per block */
    long k;                /* Index over output blocks */
    unsigned char c1;
    int c2;

    for (i = 0; i < nbout; i++) {
	if (inbyte[i] < 0) {
	    c1 = (ctab == 0) ? 0 : ctab[1-inbyte[i]];
	    for (k = 0; k < count; k++)
		dest[k*nbout+i] = c1;
	}
	else {
	    c2 = inbyte[i];
	    for (k = 0; k < count; k++)
		dest[k*nbout+i] = src[k*nbin+c2];
	}
    }

    return;
}



/*-------------------------------------------------------------------------
exim_ulong_to_bytes - Convert unsigned long into bytes

This function converts an array of unsigned long values into an array
of unsigned char values, taking a given number of bytes from each
ulong value and arranging them in big- or little-endian byte order.
If the number of bytes to be written for each unsigned long datum is
less than 4, the high order bytes of the unsigned long data are
ignored.  The count is the number of unsigned long data to be
processed, not the number of usigned char bytes to be produced.

One common use of this routine is to convert pixel values computed as
unsigned long to the byte order and number required for a particular X
server.

ALGORITHM: For efficiency, this routine dispatches on the number of
bytes and byte order and has a custom loop for each possible
combination (but it is assumed that the compiler will optimize the
details).
--------------------------------------------------------------------------*/

void
exim_ulong_to_bytes(
    long count,           /* Number of data to process */
    unsigned char *dest,  /* Output buffer */
    unsigned long *src,   /* Input buffer */
    int nbytes,           /* Number of bytes per datum: 1, 2, 3, or 4 */
    int bigendian)        /* Big (1) or little (0) endian */
{
    long i;

    /* NOTE: This code exploits the fact that, in ANSI C, assigning an
       unsigned long to an unsigned char simply drops the high-order
       bits to avoid explicitly masking with 0xff. If your compiler is
       non conformant in this respect, you may want to add the
       explicit mask. */

    if (nbytes < 1 || nbytes > 4) {
	fatal("Invalid nbytes in exim_ulong_to_uchar"); }
 
    else if (nbytes == 4 && bigendian) {
	for ( i = 0; i < count; i++) {
	    dest[4*i+3] = src[i];
	    dest[4*i+2] = src[i] >>  8;
	    dest[4*i+1] = src[i] >> 16;
	    dest[4*i  ] = src[i] >> 24;
	}
    }

    else if (nbytes == 4) {
	for ( i = 0; i < count; i++) {
	    dest[4*i  ] = src[i];
	    dest[4*i+1] = src[i] >>  8;
	    dest[4*i+2] = src[i] >> 16;
	    dest[4*i+3] = src[i] >> 24;
	}
    }

    else if (nbytes == 3 && bigendian) {
	for ( i = 0; i < count; i++) {
	    dest[3*i+2] = src[i];
	    dest[3*i+1] = src[i] >>  8;
	    dest[3*i  ] = src[i] >> 16;
	}
    }

    else if (nbytes == 3) {
	for ( i = 0; i < count; i++) {
	    dest[3*i  ] = src[i];
	    dest[3*i+1] = src[i] >>  8;
	    dest[3*i+2] = src[i] >> 16;
	}
    }

    else if (nbytes == 2 && bigendian) {
	for ( i = 0; i < count; i++) {
	    dest[2*i+1] = src[i];
	    dest[2*i  ] = src[i] >>  8;
	}
    }

    else if (nbytes == 2) {
	for ( i = 0; i < count; i++) {
	    dest[2*i  ] = src[i];
	    dest[2*i+1] = src[i] >>  8;
	}
    }

    else if (nbytes == 1) {
	for ( i = 0; i < count; i++) {
	    dest[i] = src[i];
	}
    }

    return;
}



/*-------------------------------------------------------------------------
exim_export_float4B - Convert double to IEEE single big-endian

This function converts a double value to the four bytes of its IEEE 754
single precision representation, in big-endian order.

FIXME: Handle NaN, Inf, and denormal representations.
--------------------------------------------------------------------------*/

inline static void
exim_export_float4B(
    double value,              /* Value to convert */
    unsigned char *d)          /* Where to put the result */
{
    double v;                  /* Significand */
    int e;                     /* Exponent */
    int n;                     /* Truncated quotients */

    if ( value == 0.0) {
	d[0] = d[1] = d[2] = d[3] = 0;
    }
    else {
	v = 256.0*(fabs(frexp(value, &e)) - 0.5);
	e += 126;
	d[0] = ((value < 0) ? 0x80 : 0) + ((e >> 1) & 0x7f);
	n = floor(v);
	d[1] = ((e % 2) << 7) + n;
	v = 256.0*(v-n);
	d[2] = n = floor(v);
	v = 256.0*(v-n);
	d[3] = floor(v);
    }

    return;
}



/*-------------------------------------------------------------------------
exim_export_float4L - Convert double to IEEE single little-endian

This function converts a double value to the four bytes of its IEEE 754
single precision representation, in little-endian order.

FIXME: Handle NaN, Inf, and denormal representations.
--------------------------------------------------------------------------*/

inline static void
exim_export_float4L(
    double value,              /* Value to convert */
    unsigned char *d)          /* Where to put the result */
{
    double v;                  /* Significand */
    int e;                     /* Exponent */
    int n;                     /* Truncated quotients */

    if (value == 0.0) {
	d[3] = d[2] = d[1] = d[0] = 0;
    }
    else {
	v = 256.0*(fabs(frexp(value, &e)) - 0.5);
	e += 126;
	d[3] = ((value < 0) ? 0x80 : 0) + ((e >> 1) & 0x7f);
	n = floor(v);
	d[2] = ((e % 2) << 7) + n;
	v = 256.0*(v-n);
	d[1] = n = floor(v);
	v = 256.0*(v-n);
	d[0] = floor(v);
    }

    return;
}



/*-------------------------------------------------------------------------
exim_export_float8B - Convert double to IEEE double big-endian

This function converts a double value to the eight bytes of its IEEE 754
double precision representation, in big-endian order.

FIXME: Handle NaN, Inf, and denormal representations.
--------------------------------------------------------------------------*/

inline static void
exim_export_float8B(
    double value,              /* Value to convert */
    unsigned char *d)          /* Where to put the result */
{
    double v;                  /* Significand */
    int e;                     /* Exponent */
    int n;                     /* Truncated quotients */

    if ( value == 0.0) {
	d[0] = d[1] = d[2] = d[3] = d[4] = d[5] = d[6] = d[7] = 0;
    }
    else {
	v = 32.0*(fabs(frexp(value, &e)) - 0.5);
	e += 1022;
	d[0] = ((value < 0) ? 0x80 : 0) + ((e >> 4) & 0x7f);
	n = floor(v);
	d[1] = ((e & 15) << 4) + n;
	v = 256.0*(v-n);
	d[2] = n = floor(v);
	v = 256.0*(v-n);
	d[3] = n = floor(v);
	v = 256.0*(v-n);
	d[4] = n = floor(v);
	v = 256.0*(v-n);
	d[5] = n = floor(v);
	v = 256.0*(v-n);
	d[6] = n = floor(v);
	v = 256.0*(v-n);
	d[7] = floor(v);
    }

    return;
}



/*-------------------------------------------------------------------------
exim_export_float8L - Convert double to IEEE double little-endian

This function converts a double value to the eight bytes of its IEEE 754
double precision representation, in little-endian order.

FIXME: Handle NaN, Inf, and denormal representations.
--------------------------------------------------------------------------*/

inline static void
exim_export_float8L(
    double value,              /* Value to convert */
    unsigned char *d)          /* Where to put the result */
{
    double v;                  /* Significand */
    int e;                     /* Exponent */
    int n;                     /* Truncated quotients */

    if ( value == 0.0) {
	d[7] = d[6] = d[5] = d[4] = d[3] = d[2] = d[1] = d[0] = 0;
    }
    else {
	v = 32.0*(fabs(frexp(value, &e)) - 0.5);
	e += 1022;
	d[7] = ((value < 0) ? 0x80 : 0) + ((e >> 4) & 0x7f);
	n = floor(v);
	d[6] = ((e % 16) << 4) + n;
	v = 256.0*(v-n);
	d[5] = n = floor(v);
	v = 256.0*(v-n);
	d[4] = n = floor(v);
	v = 256.0*(v-n);
	d[3] = n = floor(v);
	v = 256.0*(v-n);
	d[2] = n = floor(v);
	v = 256.0*(v-n);
	d[1] = n = floor(v);
	v = 256.0*(v-n);
	d[0] = floor(v);
    }

    return;
}



/*-------------------------------------------------------------------------
exim_export - Convert from internal to external format

This function takes a stream of elements in some specified internal
(native) data format, converts them to some specified external data
format, and saves the results as a byte stream.  Narrowing conversions
(e.g.  longs to 2-byte int) are not permitted; nor are conversions
between signed and unsigned types.  The stride may be negative (but
not zero).  The results are unpredictable if the source and
destination arrays overlap.  NOTE that the destination stride is in
BYTES, not elements; this is very easy to get wrong and will generate
really confusing results.  This function will return an error flag chosen 
from those in errm.h, or zero to indicate success.

This function is written to be portable, but there cases in which it
will produce unpredictable results.  If the native signed integral types
support both positive and negative zero, both will be exported as the
plain zero of the two's complement representation.  Integral or floating 
point values outside the range representable in the external format will 
yield unpredictable values.  On platforms which do not natively support 
IEEE 754 floating point, exporting HUGE_VAL will produce unpredictable
results.  If the native floating point types possess greater precision
than can be represented in IEEE 754, the excess precision will be lost
on export.  No warning will be given in any of these cases, partly 
because we don't have any non-IEEE, non-two's complement platforms to
test on, and don't need to handle those cases anyway.

FIXME: This version will convert unsigned to unsigned, signed to
signed, and floating point to floating point, but will not mix these
classes.  Is this sufficient?

ALGORITHM: The conversion of signed integral types into two's complement
is done using floating point division because the division of negative
integers in ANSI C is so underspecified as to be useless, and the div and
ldiv facilities round the wrong direction (toward zero rather than negative
infinity).
-------------------------------------------------------------------------*/

int
exim_export(
    long  nelem,                 /* Number of elements to convert */
    unsigned char  *dst,         /* Buffer for the output data */
    int   dtype,                 /* Desired external data format */
    long  ds,                    /* Destination stride, in BYTES */
    int   bigend,                /* Is destination to be big endian? */
    void  *src,                  /* Buffer containing the input data */
    int   stype,                 /* Desired internal data format */
    long  ss)                    /* Source stride, in ELEMENTS */
{
    unsigned char *d = dst;
    long i;

    /* Verify that the source and destination are non-zero */
    if (dst == 0 || src == 0)
	fatal("Zero source or destination address");

    /* Switch according to the source type */
    switch (stype) {

    /*** Source is unsigned char; switch on destination type ***/
    case INTP_UCHAR:
	switch (dtype) {
	    unsigned char *s;

	case EXTP_UINT1:
	    for ( s = src, i = 0; i < nelem; i++ ) {
		d[0] = (*s) & 0xff;
		s += ss; d += ds; }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = 0;
                    d[1] = (*s) & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
                    d[0] = (*s) & 0xff;
		    d[1] = 0;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_UINT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = d[1] = d[2] = 0;
                    d[3] = (*s) & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
                    d[0] = (*s) & 0xff;
		    d[1] = d[2] = d[3] = 0;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is unsigned short; switch on destination type ***/
    case INTP_USHORT:
	switch (dtype) {
	    unsigned short *s;

	case EXTP_UINT2:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = ((*s) >> 8) & 0xff;
                    d[1] = (*s) & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
                    d[0] = (*s) & 0xff;
		    d[1] = ((*s) >> 8) & 0xff;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_UINT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = d[1] = 0;
		    d[2] = ((*s) >> 8) & 0xff;
                    d[3] = (*s) & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
                    d[0] = (*s) & 0xff;
		    d[1] = ((*s) >> 8) & 0xff;
		    d[2] = d[3] = 0;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is unsigned int; switch on destination type ***/
    case INTP_UINT:
	switch (dtype) {
	    unsigned int *s;

#if SIZEOF_INT <= 2     /* Otherwise an int won't fit in 2 bytes */
	case EXTP_UINT2:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = ((*s) >> 8) & 0xff;
                    d[1] = (*s) & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
                    d[0] = (*s) & 0xff;
		    d[1] = ((*s) >> 8) & 0xff;
		    s += ss;  d += ds;
		}
		return 0;
	    }
#endif

	case EXTP_UINT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = *s >> 24 & 0xff;
		    d[1] = *s >> 16 & 0xff;
		    d[2] = *s >>  8 & 0xff;
		    d[3] = *s & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = *s & 0xff;
		    d[1] = *s >>  8 & 0xff;
		    d[2] = *s >> 16 & 0xff;
		    d[3] = *s >> 24 & 0xff;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is unsigned long; switch on destination type ***/
    case INTP_ULONG:
	switch (dtype) {
	    unsigned long *s;

	case EXTP_UINT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = *s >> 24 & 0xff;
		    d[1] = *s >> 16 & 0xff;
		    d[2] = *s >>  8 & 0xff;
		    d[3] = *s & 0xff;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = *s & 0xff;
		    d[1] = *s >>  8 & 0xff;
		    d[2] = *s >> 16 & 0xff;
		    d[3] = *s >> 24 & 0xff;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is (signed) char; switch on destination type ***/
    case INTP_SCHAR:
	switch (dtype) {
	    signed char *s;

	case EXTP_INT1:
	    for ( s = src, i = 0; i < nelem; i++ ) {
		d[0] = (unsigned char)s[0];
		s += ss;  d += ds;	}

	case EXTP_INT2:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] =  ( *s < 0) ? -1 : 0;
		    d[1] = (unsigned char)s[0];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[1] = ( *s < 0) ? -1 : 0;    
		    d[0] = (unsigned char)s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_INT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[0] = d[1] = d[2] = ( *s < 0) ? -1 : 0;
		    d[3] = (unsigned char)s[0];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    d[3] = d[2] = d[1] = ( *s < 0) ? -1 : 0;
		    d[0] = (unsigned char)s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is (signed) short; switch on destination type ***/
    case INTP_SHORT:
	switch (dtype) {
	    short *s;
	    int n;
	    double v;

	case EXTP_INT2:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    n = (int)floor((double)s[0]/256.0);
		    d[0] = (unsigned char)n;
		    d[1] = s[0] - 256*n;
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    n = (int)floor((double)s[0]/256.0);
		    d[1] = (unsigned char)n;
		    d[0] = s[0] - 256*n;
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_INT4:
	    /* FIXME: Assume here that short is only 2 bytes? */
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    v = ((double)*s) / 16777216.0;
		    d[0] = n = floor(v);
		    v = 256.0*(v-n);
		    d[1] = n = floor(v);
		    v = 256.0*(v-n);
		    d[2] = n = floor(v);
		    v = 256.0*(v-n);
		    d[3] = n = floor(v);
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    v = ((double)*s) / 16777216.0;
		    d[3] = n = floor(v);
		    v = 256.0*(v-n);
		    d[2] = n = floor(v);
		    v = 256.0*(v-n);
		    d[1] = n = floor(v);
		    v = 256.0*(v-n);
		    d[0] = n = floor(v);
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is (signed) int; switch on destination type ***/
    case INTP_INT:
	switch (dtype) {
	    int *s;
	    int n;
	    double v;

	case EXTP_INT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    v = ((double)*s) / 16777216.0;
		    d[0] = n = floor(v);
		    v = 256.0*(v-n);
		    d[1] = n = floor(v);
		    v = 256.0*(v-n);
		    d[2] = n = floor(v);
		    v = 256.0*(v-n);
		    d[3] = n = floor(v);
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    v = ((double)*s) / 16777216.0;
		    d[3] = n = floor(v);
		    v = 256.0*(v-n);
		    d[2] = n = floor(v);
		    v = 256.0*(v-n);
		    d[1] = n = floor(v);
		    v = 256.0*(v-n);
		    d[0] = n = floor(v);
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Source is (signed) long; switch on destination type ***/
    case INTP_LONG:
	switch (dtype) {
	    long *s;
	    int n;
	    double v;

	case EXTP_INT4:
	    if (bigend) {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    v = ((double)*s) / 16777216.0;
		    d[0] = n = floor(v);
		    v = 256.0*(v-n);
		    d[1] = n = floor(v);
		    v = 256.0*(v-n);
		    d[2] = n = floor(v);
		    v = 256.0*(v-n);
		    d[3] = n = floor(v);
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( s = src, i = 0; i < nelem; i++ ) {
		    v = ((double)*s) / 16777216.0;
		    d[3] = n = floor(v);
		    v = 256.0*(v-n);
		    d[2] = n = floor(v);
		    v = 256.0*(v-n);
		    d[1] = n = floor(v);
		    v = 256.0*(v-n);
		    d[0] = n = floor(v);
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

	/*** Source is float; switch on destination type ***/
    case INTP_FLOAT:
	switch (dtype) {
	    float *s;

	case EXTP_FLOAT4:
	    s = src;
	    if (bigend) {
		for (i = 0; i < nelem; i++) {
		    exim_export_float4B(*s, d);
		    d += ds; s += ss;
		}
		return 0;
	    }
	    else {
		for (i = 0; i < nelem; i++) {
		    exim_export_float4L(*s, d);
		    d += ds; s += ss;
		}
		return 0;
	    }

	case EXTP_FLOAT8:
	    s = src;
	    if (bigend) {
		for (i = 0; i < nelem; i++) {
		    exim_export_float8B(*s, d);
		    d += ds; s += ss;
		}
		return 0;
	    }
	    else {
		for (i = 0; i < nelem; i++) {
		    exim_export_float8L(*s, d);
		    d += ds; s += ss;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

	/*** Source is double; switch on destination type ***/
    case INTP_DOUBLE:
	switch (dtype) {
	    float *s;

	case EXTP_FLOAT8:
	    s = src;
	    if (bigend) {
		for (i = 0; i < nelem; i++) {
		    exim_export_float8B(*s, d);
		    d += ds; s += ss;
		}
		return 0;
	    }
	    else {
		for (i = 0; i < nelem; i++) {
		    exim_export_float8L(*s, d);
		    d += ds; s += ss;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    default:
	/* Other source types have not yet been implemented */
	return ERR_UNIMP;
	
    }  /* End switch on source type */

    /* Control never reaches here; each switch case returns directly */
}





/*-------------------------------------------------------------------------
exim_import - Convert from external to internal format

This function interprets a byte stream as a specified external data
format, converts it to some desired internal (native) format, and
saves the converted data.  Narrowing conversions (e.g.  4-byte ints to
short, or signed int to unsigned int) are not permitted and must be
done later in core where it is easier to check for errors.  The stride
may be negative (but not zero).  The results are unpredictable if the
source and destination arrays overlap.  NOTE that the source stride is
in BYTES, not elements; this is very easy to get wrong and will
generate really confusing results.  This function will return an error
flag chosen from those in errm.h, or zero to indicate success.

This function attempts to be portable, but there are some definite
limits on that portability, especially since we don't have platforms
available for testing that exemplify some of the odder behaviors
permitted by the 1989 ANSI C standard.  The code used here should
correctly import any external value that can be validly represented on
the native platform.  However, the results are unpredicatable if you
attempt to import the minimum possible two's complement value into a
signed integer of the same width on a platform which is not natively
two's complement; importing into a wider native type should be safe.
Importing a NaN, Inf, or denormal onto a platform which does not
natively support IEEE 754 may produce unpredicatble results, as may
importing values which are outside the range representable on the
native platform.  Low-order bits of floating point values may be lost
if imported into native formats with insufficient resolution.  No
warnings or errors will be given in any of these cases, mostly because
we don't have a non-IEEE, non-two's complement platform handy for
testing that the testing code will work properly (and thus don't
actually need to handle such cases).  If unsign3d chars on your
platform are more than 8 bits wide, you are completely on your own; I
haven't even attempted to guess how foreign files get read on such
systems.  It is illegal to attempt to import UINT4 to UINT, or INT4 to
INT, on a machine on which the int and unsigned int types are less
than four bytes long.

FIXME: The current version of this code does not handle Inf, NaN, or
denormal floats on input.
-------------------------------------------------------------------------*/

int
exim_import(
    long  nelem,                 /* Number of elements to convert */
    void  *dst,                  /* Buffer for the output data */
    int   dtype,                 /* Desired internal data format */
    long  ds,                    /* Destination stride, in elements */
    unsigned char *src,          /* Buffer containing the input data */
    int   stype,                 /* Desired external data format */
    long  ss,                    /* Source stride, in bytes */
    int   bigend)                /* Is source big-endian? */
{
    unsigned char  *s = src;
    long i;

    /* Verify that the source and destination are non-zero */
    if (dst == 0 || src == 0)
	fatal("Zero source or destination address");

    /* Switch according to the destination type */
    switch (dtype) {

    /*** Destination is unsigned char; switch on source type ***/
    case INTP_UCHAR:
	switch (stype) {
	    unsigned char *d;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds; }
	    return 0;

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is unsigned short; switch on source type ***/
    case INTP_USHORT:
	switch (stype) {
	    unsigned short *d;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[0] << 8) + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[1] << 8) + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is unsigned int; switch on source type ***/
    case INTP_UINT:
	switch (stype) {
	    unsigned int *d;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[0] << 8) + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[1] << 8) + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

#if SIZEOF_INT >= 4
	case EXTP_UINT4:
	    if (sizeof(int) < 4)
		return ERR_ILLOP;
	    else if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = ((unsigned int)s[0] << 24)
                       + ((unsigned int)s[1] << 16) 
                       + ((unsigned int)s[2] <<  8) + s[3];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = ((unsigned int)s[3] << 24)
                       + ((unsigned int)s[2] << 16) 
                       + ((unsigned int)s[1] <<  8) + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }
#endif

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is unsigned long; switch on source type ***/
    case INTP_ULONG:
	switch (stype) {
	    unsigned long *d;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[0] << 8) + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[1] << 8) + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_UINT4:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = ((unsigned long)s[0] << 24)
                       + ((unsigned long)s[1] << 16) 
                       + ((unsigned long)s[2] <<  8) + s[3];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = ((unsigned long)s[3] << 24)
                       + ((unsigned long)s[2] << 16) 
                       + ((unsigned long)s[1] <<  8) + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is signed char; switch on source type ***/
    case INTP_SCHAR:
	switch (stype) {
            signed char *d;

	case EXTP_INT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = (*s & 0x80) ? (*s - 256) : *s;
                s += ss;   d += ds;  }
	    return 0;

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is signed short; switch on source type ***/
    case INTP_SHORT:
	switch (stype) {
          signed short *d, d1;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = (*s & 0x80) ? (*s - 256) : *s;
                s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = d1*256 + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[1] & 0x80) ? (s[1] - 256) : s[1];
                    *d = d1*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is signed int; switch on source type ***/
    case INTP_INT:
	switch (stype) {
          signed int *d, d1;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = (*s & 0x80) ? (*s - 256) : *s;
                s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = d1*256 + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[1] & 0x80) ? (s[1] - 256) : s[1];
                    *d = d1*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

#if SIZEOF_INT >= 4
	case EXTP_INT4:
	    if (sizeof(int) < 4)
		return ERR_ILLOP;
	    else if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = ((d1*256 + s[1])*256 + s[2])*256 + s[3];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[3] & 0x80) ? (s[3] - 256) : s[3];
                    *d = ((d1*256 + s[2])*256 + s[1])*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }
#endif

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is signed long; switch on source type ***/
    case INTP_LONG:
	switch (stype) {
          signed long *d, d1;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    *d = (s[0] << 8) + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    *d = (s[1] << 8) + s[0];
  		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_INT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = (*s & 0x80) ? (*s - 256) : *s;
                s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = d1*256 + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[1] & 0x80) ? (s[1] - 256) : s[1];
                    *d = d1*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_INT4:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = ((d1*256 + s[1])*256 + s[2])*256 + s[3];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[3] & 0x80) ? (s[3] - 256) : s[3];
                    *d = ((d1*256 + s[2])*256 + s[1])*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is float; switch on source type ***/
    case INTP_FLOAT:
	switch (stype) {
	    float *d;
            int m1;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[0] << 8) + s[1];
		    s += ss;  d += ds;  }
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[1] << 8) + s[0];
		    s += ss;  d += ds;	}
		return 0;
	    }

	case EXTP_INT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = (*s & 0x80) ? (*s - 256) : *s;
                s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    m1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = m1*256 + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    m1 = (s[1] & 0x80) ? (s[1] - 256) : s[1];
                    *d = m1*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_FLOAT4:
	    if (bigend) {
		int ex;
                long m1;
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    if ( s[0] == 0 && s[1] == 0 && s[2] == 0 && s[3] == 0 ) {
			*d = 0.0;
		    }
		    else {
			ex = ((s[0] & 0x7f) <<  1) + (s[1] >> 7);
			m1 = ((long)(0x80 | s[1]) << 16)
			    + ((long)        s[2]  <<  8) + s[3];
			if ( s[0] & 0x80 )   m1 = -m1;
			*d = ldexp((double)m1, ex-150);
		    }
		    s += ss;  d += ds;  }
		return 0;
	    }
	    else {
		int ex;
                long m1;
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    if (s[3] == 0 && s[2] == 0 && s[1] == 0 && s[0] == 0) {
			*d = 0.0;
		    }
		    else {
			ex = ((s[3] & 0x7f) <<  1) + (s[2] >> 7);
			m1 = ((long)(0x80 | s[2]) << 16)
			    + ((long)        s[1]  <<  8) + s[0];
			if ( s[3] & 0x80 )   m1 = -m1;
			*d = ldexp((double)m1, ex-150);
		    } 
		    s += ss;  d += ds;	}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    /*** Destination is double; switch on source type ***/
    case INTP_DOUBLE:
	switch (stype) {
	    double *d;

	case EXTP_UINT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = *s;   s += ss;   d += ds;  }
	    return 0;

	case EXTP_UINT2:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[0] << 8) + s[1];
		    s += ss;  d += ds;  }
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = (s[1] << 8) + s[0];
		    s += ss;  d += ds;	}
		return 0;
	    }

	case EXTP_UINT4:
	    if (bigend) {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = ((unsigned long)s[0] << 24)
                       + ((unsigned long)s[1] << 16)
                       + ((unsigned long)s[2] <<  8) + s[3];
		    s += ss;  d += ds;  }
		return 0;
	    }
	    else {
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    *d = ((unsigned long)s[3] << 24)
                       + ((unsigned long)s[2] << 16)
                       + ((unsigned long)s[1] <<  8) + s[0];
		    s += ss;  d += ds;	}
		return 0;
	    }

	case EXTP_INT1:
	    for ( d = dst, i = 0; i < nelem; i++ ) {
		*d = (*s & 0x80) ? (*s - 256) : *s;
                s += ss;   d += ds;  }
	    return 0;

	case EXTP_INT2:
	    if (bigend) {
                long d1;
                for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = d1*256 + s[1];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
                long d1;
                for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[1] & 0x80) ? (s[1] - 256) : s[1];
                    *d = d1*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_INT4:
	    if (bigend) {
                long d1;
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[0] & 0x80) ? (s[0] - 256) : s[0];
                    *d = ((d1*256 + s[1])*256 + s[2])*256 + s[3];
		    s += ss;  d += ds;	}
		return 0;
	    }
	    else {
                long d1;
		for ( d = dst, i = 0; i < nelem; i++ ) {
                    d1 = (s[3] & 0x80) ? (s[3] - 256) : s[3];
                    *d = ((d1*256 + s[2])*256 + s[1])*256 + s[0];
		    s += ss;  d += ds;
		}
		return 0;
	    }

	case EXTP_FLOAT4:
	    if (bigend) {
		int ex;
                long m1;
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    if (s[0] == 0 && s[1] == 0 && s[2] == 0 && s[3] == 0) {
			*d = 0.0;
		    }
		    else {
			ex = ((s[0] & 0x7f) <<  1) + (s[1] >> 7);
			m1 = ((long)(0x80 | s[1]) << 16)
			  + ((long)        s[2]  <<  8) + s[3];
			if ( s[0] & 0x80 )   m1 = -m1;
			*d = ldexp((double)m1, ex-150); 
		    }
		    s += ss;  d += ds;  }
		return 0;
	    }
	    else {
		int ex;
                long m1;
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    if (s[3] == 0 && s[2] == 0 && s[1] == 0 && s[0] == 0) {
			*d = 0.0;
		    }
		    else {
			ex = ((s[3] & 0x7f) <<  1) + (s[2] >> 7);
			m1 = ((long)(0x80 | s[2]) << 16)
			  + ((long)        s[1]  <<  8) + s[0];
			if ( s[3] & 0x80 )   m1 = -m1;
			*d = ldexp((double)m1, ex-150); 
		    }
		    s += ss;  d += ds;	}
		return 0;
	    }

	case EXTP_FLOAT8:
	    if (bigend) {
		int ex;
                double m1, m2;
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    if (s[0] == 0 && s[1] == 0 && s[2] == 0 && s[3] == 0 &&
			s[4] == 0 && s[5] == 0 && s[6] == 0 && s[7] == 0) {
			*d = 0.0;
		    }
		    else {
			ex = ((s[0] & 0x7f) <<  4) + (s[1] >> 4);
			m1 = ((long)(0x10 | (s[1] & 0xf)) << 16)
			   + ((long)s[2]  <<  8) + s[3];
			m2 = ((long)s[4] << 24)
			   + ((long)s[5] << 16)
			   + ((long)s[6] <<  8) + s[7];
			m1 = ldexp(m1,ex-1043) + ldexp(m2,ex-1075);
			*d = (s[0] & 0x80) ? -m1 : m1;
		    }
		    s += ss;  d += ds;  }
		return 0;
	    }
	    else {
		int ex;
                double m1, m2;
		for ( d = dst, i = 0; i < nelem; i++ ) {
		    if (s[7] == 0 && s[6] == 0 && s[5] == 0 && s[4] == 0 &&
			s[3] == 0 && s[2] == 0 && s[1] == 0 && s[0] == 0) {
			*d = 0.0;
		    }
		    else {
			ex = ((s[7] & 0x7f) <<  4) + (s[6] >> 4);
			m1 = ((long)(0x10 | (s[6] & 0xf)) << 16)
			   + ((long)s[5]  <<  8) + s[4];
			m2 = ((long)s[3] << 24)
			   + ((long)s[2] << 16)
			   + ((long)s[1] <<  8) + s[0];
			m1 = ldexp(m1,ex-1043) + ldexp(m2,ex-1075);
			*d = (s[7] & 0x80) ? -m1 : m1;
		    }
		    s += ss;  d += ds;	}
		return 0;
	    }

	default:
	    return ERR_ILLOP;
	}

    default:
	/* Other destination types have not yet been implemented */
	return ERR_UNIMP;
	
    }  /* End switch on destination type */

    /* Control never reaches here; each switch case returns directly */
}



/*----------------------------------------------------------------------
exim_imdump - Import and print a single external datum

This function converts a single external datum stored in a buffer to a
specified internal type, optionally saves the converted value to an
internal data address, and prints the converted value using a
user-specified formatting string.  The formatting string may be a null
pointer, in which case no value is printed.  The internal data address
may be a null pointer, in which case no value is stored.

This function is intended primarily for use in exim_scalar and its
relatives, and is mostly a wrapper for exim_import that arranges to
keep the converted value around long enough to print even if it's not
permanently stored.
-----------------------------------------------------------------------*/

int
exim_imdump(
    FILE *dump,                  /* File to write the ASCII to */
    void *indata,                /* Location of internal data element */
    int  intype,                 /* Internal data format */
    char *fmt,                   /* Formatting string for fprintf */
    unsigned char *exdata,       /* External data element to print */
    int  extype,                 /* External data format */
    int  bigend)                 /* Big endian? */
{
    switch (intype) {

    case INTP_UCHAR: {
	unsigned char v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(unsigned char *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_USHORT: {
	unsigned short v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(unsigned short *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_UINT: {
	unsigned int v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(unsigned int *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_ULONG: {
	unsigned long v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(unsigned long *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_SCHAR: {
	signed char v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(signed char *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_SHORT: {
	short v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(short *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_INT: {
	int v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(int *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_LONG: {
	long v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(long *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_FLOAT: {
	float v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(float *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    case INTP_DOUBLE: {
	double v;
	exim_import(1, &v, intype, 0, exdata, extype, 0, bigend);
	if (indata != 0)   *(double *)indata = v;
	fprintf(dump, fmt, v);
	return ERR_OKAY; }

    default:
	return ERR_ILLOP;

    }
    
    /* Never reaches here; each arm of switch returns directly */
}



/*----------------------------------------------------------------------
exim_create_mode - Create and set mode for exim_scalar and relatives

This function creates a struct exim_mode and initializes it to values
specified by the user; a pointer to this struct may be passed as an
argument to exim_scalar and its relatives to define the parameters
which do not depend on individual fields to be converted.  Once you've
created a mode struct, you can modify its setting by directly
modifying the structure; the primary purpose of this function is to
ensure that any new parameters get initialized to reasonable values
even if they were added after you wrote your code.

FIXME: The current version does little or no checking of the arguments.
-----------------------------------------------------------------------*/

struct exim_mode *
exim_create_mode(
    int dir,             /* Import, export, or dump? */
    FILE *exfile,        /* File containing external data */
    FILE *dump,          /* File to print values to */
    int bigend)          /* Big endian? */
{
    struct exim_mode *mode;   /* Mode struct being created */

    mode = mallock(sizeof(struct exim_mode));
    mode->dir    = dir;
    mode->exfile = exfile;
    mode->dump   = dump;
    mode->bigend = bigend;
    return mode;
}



/*----------------------------------------------------------------------
exim_destroy_mode - Free an exim mode structure

This function destroys a struct exim_mode and frees its memory.  It is
a serious error to attempt to free a mode structure which was never
created, except that supplying a null pointer will do nothing.
-----------------------------------------------------------------------*/

void
exim_destroy_mode(
    struct exim_mode *mode)      /* The mode to be destroyed */
{
    if (mode != 0)
	free(mode);
    return;
}



/*-----------------------------------------------------------------------
exim_scalar - Import/export/dump a scalar value

This function imports, exports, or dumps a single scalar value; the
particular operation is selected by the mode argument, which is a
structure defined in exim.h.  The reason for this odd modularization
is to permit the external data structure to be defined by a single
piece of code, which can then be used either to import data from an
external file, export data to an external file, or dump the contents
of an external file in human-readable form.

The SCALAR macro defined in exim.h provides a convenient way of
encapsulating calls to this function so that they can be read as
tables.

If the direction specified in the mode is EXIM_IMPORT or EXIM_IMDUMP,
a single data value will be read in the specified external format from
the external file, converted to the specified internal format, and
stored to the internal data address.  The internal address may be a
null pointer, in which case the value is silently discarded.

If the direction is EXIM_IMDUMP, then the value is also printed to the
stream also specified by the mode argument, using the given format
string in fprintf; if any label for the contents is desired, it may be
included in this string.  If the format string argument is a null
pointer, then the value will not be written.

If the direction is EXIM_EXPORT, then a single value is read from the
internal data address in the specified internal format, converted to
the specified external format, and written to the output file specified
in the mode.  The format string is ignored, but it is a fatal error
if the internal data address is a null pointer.  (NOTE that this mode
has not yet been implemented.)  

FIXME: The EXDUMP mode is not yet implemented.  The utility of the
EXDUMP mode is becoming less clear with time, and it may never be
implemented.
-----------------------------------------------------------------------*/

int
exim_scalar(
    struct exim_mode *mode,  /* Field-independent parameters */
    void *indata,            /* Internal data value */
    int intype,              /* Internal data type */
    int extype,              /* External data type */
    char *format)            /* Dump format */
{
    unsigned char
	buff[MAXBYTES];       /* Buffer for a single external value */
    int exlen;                /* Length of a single external datum */
    int err, cnt;

    /* Determine length of the external data type */
    exlen = exim_sizeof_extype(extype);
    if (exlen == 0)
	return ERR_ILLOP;
    if (exlen > MAXBYTES)
	fatal("Buffer is too short for datum");

    /* Switch on the direction of transfer */
    switch (mode->dir) {
	
    /* Read data from a file into memory, and possibly print as ASCII */
    case EXIM_IMPORT:
    case EXIM_IMDUMP:
	cnt = fread(buff, exlen, 1, mode->exfile);
	if (cnt != 1)
	    return ERR_FILE;
	if ( mode->dir == EXIM_IMPORT && indata != 0 ) {
	    err = exim_import(1, indata, intype, 0, buff, extype, 0, 
			      mode->bigend); }
	else if ( mode->dir == EXIM_IMDUMP && format != 0 ) {
	    err = exim_imdump(mode->dump, indata, intype, format, buff,
			      extype, mode->bigend); }
	else
	    err = 0;
	return err;

    /* Write data to a file from memory */
    case EXIM_EXPORT:
        if (indata == 0)
            return ERR_ILLOP;
        err = exim_export(1, buff, extype, 0, mode->bigend, indata, intype, 0);
        cnt = fwrite(buff, exlen, 1, mode->exfile);
        if (err == 0 && cnt != 1)
            err = ERR_FILE;
	return err;

    /* Read data from memory and print in ASCII form */
    case EXIM_EXDUMP:
	return ERR_UNIMP;

    /* Invalid mode switch */
    default:
	return ERR_ILLOP;
    }

    /* Never reaches here; each branch of switch returns directly */
}



/*-----------------------------------------------------------------------
exim_vector - Import/export/dump an array of values

This function imports, exports, or dumps an array of values; the
particular operation is selected by the mode argument, which has the
possible values defined in exim.h.  The reason for this odd
modularization is to permit the external data structure to be defined
by a single piece of code, which can then be used either to import
data from an external file, export data to an external file, or dump
the contents of an external file in human-readable form.

The VECTOR macro defined in exim.h provides a convenient way of
encapsulating calls to this function so that they can be read as
tables.

If the direction specified in the mode is EXIM_IMPORT or EXIM_IMDUMP,
successive data values will be read in the specified external format
from the external file, converted to the specified internal format,
and stored sequentially to the internal data address.  The internal
address may be a null pointer, in which case the value is silently
discarded.

If the direction is EXIM_IMDUMP, then the values are also printed to
the stream specified by the mode argument, using the given format
string in fprintf; if any label for the contents is desired, it may be
included in this string.  If the format string argument is a null
pointer, then the value will not be written.  It is frequently desired
that the values should be preceded by some starting text and followed
by some ending text.  To support this, the format string may be in the
format "...[...]...", containing exactly one matched pair of square
brackets.  The text before "[" is used in a call to fprintf before
any element is printed; the text between the brackets is used to print
each element; and the text following "]" is used after all the
elements have been printed.  For example, the format string "dimen:
[%d, ]\n" could be used to print out the dimensions of an array on a
single line, preceded by a label and followed by a newline.

If the direction is EXIM_EXPORT, then the values are read successively
from the internal data address in the specified internal format,
converted to the specified external format, and written to the output
file specified in the mode.  The format string is ignored, but it is a
fatal error if the internal data address is a null pointer.  (NOTE
that this mode has not yet been implemented.)

FIXME: The first non-zero error code is returned but may not be the
most serious error.

FIXME: The EXDUMP mode is not yet implemented, and may never be.
-----------------------------------------------------------------------*/

int
exim_vector(
    struct exim_mode *mode,  /* Field-independent parameters */
    size_t nelem,            /* Number of data elements to process */
    void *indata,            /* First internal data value */
    int intype,              /* Internal data type */
    int extype,              /* External data type */
    char *format)            /* Dump format */
{
    unsigned char
	buff[MAXBYTES];          /* Buffer for a single external value */
    int exlen;                   /* Length of a single external datum */
    int inlen;                   /* Length of a single internal datum */
    unsigned char *d = indata;   /* Location of current data element */
    int err = 0, err1 = 0, cnt = 0;
    int i;

    /* Verify internal and external types */
    exlen = exim_sizeof_type(extype);
    if (exlen == 0)
	return ERR_ILLOP;
    if (exlen > MAXBYTES)
	panic("Buffer is too short for datum");
    inlen = exim_sizeof_type(intype);
    if (inlen == 0)
	return ERR_ILLOP;

    /* Switch on the transfer direction */
    switch (mode->dir) {
	
    /* Read data from a file into memory */
    case EXIM_IMPORT:
	for (i = 0; i < nelem; i++) {
	    cnt = fread(buff, exlen, 1, mode->exfile);
	    if (cnt != 1)
		return ERR_FILE;
	    if ( d != 0) {
		err1 = exim_import(1, d, intype, 0, buff, extype, 0, 
				   mode->bigend);
		if (err1 != 0 && err == 0) err = err1;
		d += inlen; }
	}
	return err;

    /* Read data from a file and print in ASCII form */
    case EXIM_IMDUMP: {
        char fmt[80];               /* Buffer for format substrings */
        char *f1=0, *f2=0, *f3=0;   /* The substrings themselves */

        /* Find the start, middle, and end format strings */
        if (strlen(format) > 80-1)
            fatal("Buffer too small for given format string");
	if (format != 0) {
	    strcpy(fmt, format);
	    f1 = f2 = fmt;
	    while ( *f2 && *f2 != '[') f2++;
	    if ( *f2 == 0 ) {
		f1 = f3 = 0;
		f2 = fmt; }
	    else {
		*f2 = 0;
		f3 = f2 += 1;
		while ( *f3 && *f3 != ']') f3++;
		if (*f3 == 0)   f3   = 0;
		else           *f3++ = 0;
	    }
	}

        /* Print the start string */
        if (f1 != 0) fprintf(mode->dump, f1);

        /* Print each element */
	for (i = 0; i < nelem; i++) {
	    cnt = fread(buff, exlen, 1, mode->exfile);
	    if (cnt != 1)
		return ERR_FILE;
	    if ( format != 0) {
		err1 = exim_imdump(mode->dump, d, intype, f2,
 			 	   buff, extype, mode->bigend);
		if (err1 != 0 && err == 0) err = err1;
		if (d != 0)   d += inlen; }
	}

        /* Print the end string and exit */
        if (f3 != 0) fprintf(mode->dump, f3);
	return err;
    }

    /* Write data to a file from memory */
    case EXIM_EXPORT:
        if (d == 0)
            return ERR_ILLOP;
        for (i = 0; i < nelem; i++) {
            err = exim_export(1, buff, extype, 0, mode->bigend, d, intype, 0);
            cnt = fwrite(buff, exlen, 1, mode->exfile);
            if (err == 0 && cnt != 1)
                err = ERR_FILE;
            d += inlen;
        }
	return err;

    /* Read data from memory and print in ASCII form */
    case EXIM_EXDUMP:
	return ERR_UNIMP;

    /* Invalid direction */
    default:
	return ERR_ILLOP;
    }

    /* Never reaches here; each branch of switch returns directly */
}



/*-----------------------------------------------------------------------
exim_alloc - Allocate space for an array or structure

This function allocates memory for an array or structure as determined
by the dir field in the mode structure.  The array is sized to contain
nelem elements of the size determined by the internal type; storage
for a structure can be allocated by setting nelem to sizeof(struct)
and the internal type to INTP_UCHAR.  If the direction is EXIM_IMPORT
or EXIM_IMDUMP, storage is allocated for an array or struct and a
pointer to the beginning of that block is returned; the storage is
initialized as by calloc.  If the direction is EXIM_EXPORT, the value
of indata is returned.  Zero is returned if the required storage
cannot be allocated, if an illegal direction is requested, or if the
internal type is not recognized.
------------------------------------------------------------------------*/

void *
exim_alloc(
    struct exim_mode *mode,        /* Field-independent parameters */
    void *indata,                  /* First internal data value */
    size_t nelem,                  /* Number of elements */
    int intype)                    /* Internal data type */
{
    size_t inlen;                  /* Size of an internal datum */

    switch (mode->dir) {

    case EXIM_IMPORT:
    case EXIM_IMDUMP:
	inlen  = exim_sizeof_intype(intype);
	indata = callock(nelem, inlen);
	return indata;

    case EXIM_EXPORT:
	return indata;

    default:
	fatal("Invalid direction requested");
	return 0;          /* Keeps compiler from complaining */
    }
}
