/****************************************************************************
imstack.c - Stack individual slices to make an image file

Author:  Paul Hughett

Copyright (c) 1998-1999 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


See the man page for operating instructions.

imstack [-f first] [-l last] [-b] [-k skip] [-s bytes] [-0] [-n nfmt] 
   pattern outfile

*****************************************************************************/

static const char rcsid[] = "$Id: imstack.c,v 1.2 2000/06/28 16:09:25 hughett Exp $ Release $Name:  $";


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#define MAXFILENAME  256

/* Global */
extern void fatal(const char *message, ...);
extern void *mallock( size_t size );

/* Local */
void slice_fn(char *fname, int len, char *pattern, char *sfmt, int  sliceno);



/*---------------------------------------------------
main
---------------------------------------------------*/

int
main(int argc, char *argv[])
{
    /* Parameters provided by user */
    int first=1, last=256;         /* First and last slice number */
    char *pattern="foo#.img";      /* Generic filename for slices */
    char *outfile;                 /* Image filename */
    char *nfmt="%d";               /* Format for slice numbers */
    long skip=0;                   /* Nmbr bytes to skip */
    int  fromend = 0;              /* Skip bytes at beginning or end? */
    long length=2*256*256;         /* Nmbr data bytes per slice */
    int swap = 0;                  /* Swap bytes? */

    /* Other variables */
    char slicefn[MAXFILENAME];     /* Filename for current slice */
    unsigned char *sbuff, uc;      /* Slice buffer */
    FILE *in, *out;                /* Input and output image files */
    int islice, c, err;
    size_t len;
    long l;

    /* Get the options from the command line */
    while ( (c = getopt(argc, argv, "0bf:ek:l:s:n:")) != EOF) 
	switch (c) {
	case '0':      /* That's zero, folks! */
	    nfmt = "%03d"; break;
	case 'b':
	    swap = 1; break;
	case 'e':
	    fromend = 1; break;
	case 'f':
	    first = atoi(optarg); break;
	case 'k':
	    skip = atol(optarg); break;
	case 'l':
	    last = atoi(optarg); break;
	case 's':
	    length = atol(optarg); break;
	case 'n':
	    nfmt = optarg; break;
	}

    /* Get the other arguments from the command line */
    if (argc - optind < 2)
	fatal("Usage:  imstack [options] pattern outfile");
    pattern = argv[optind];
    outfile = argv[optind+1];

    /* FIXME: Do sanity check on arguments */

    /* DEBUG:  Show options and arguments */
    if (0) {
	fprintf(stderr, "Slices %d-%d\n", first, last);
	fprintf(stderr, "Skip = %ld, slice length = %ld\n", skip, length);
	fprintf(stderr, "Byte swap = %d\n", swap);
	fprintf(stderr, "Pattern is %s\n", pattern);
	fprintf(stderr, "Number format %s\n", nfmt);
	fprintf(stderr, "Output file is %s\n", outfile);
    }       

    /* Open the output image file */
    out = fopen(outfile, "wb");
    if (out == 0) 
	fatal("Unable to open output file %s", outfile);

    /* Allocate a buffer for the slice images */
    sbuff = mallock(length);

    /* Loop over the slice numbers */
    fprintf(stderr, "\n");
    for (islice = first; islice <= last; islice++) {

	/* Generate filename for current slice */
	slice_fn(slicefn, MAXFILENAME, pattern, nfmt, islice);
	fprintf(stderr, "Reading %s\r", slicefn);

	/* Open file containing the current slice */
	in = fopen(slicefn, "rb");
	if (in == 0 && islice == first)
	    fatal("Unable to open input slice file %s", slicefn);
	else if (in == 0)
	    break;

	/* Skip over the file header, if any */
	if (fromend)
	    err = fseek(in, -skip-length, SEEK_END);
	else
	    err = fseek(in, skip, SEEK_SET);
	if (err)
	    fatal("Seek failed on slice file %s", slicefn);
	
	/* Read the slice data */
	len = fread(sbuff, 1, length, in);
	if (len != length)
	    fatal("Unable to read complete slice %s", slicefn);
	
	/* Swap bytes if desired */
	if (swap)
	    for (l = 0; l < length-1; l += 2) {
		uc = sbuff[l+1];
		sbuff[l+1] = sbuff[l];
		sbuff[l] = uc; 
	    }

	/* Copy slice contents to output  */
	len = fwrite(sbuff, 1, length, out);
	if (len != length)
	    fatal("Unable to write complete slice %s", slicefn);

	fclose(in);
    }
    
    /* Clean up and exit successfully */
    fprintf(stderr, "Done...processed slices %d through %d.\n",
	    first, islice-1);
    fclose(out);
    free(sbuff);

    exit(EXIT_SUCCESS);
}


/*----------------------------------------------------------------------
slice_fn - Generate filename for a given slice number

FIXME:  Promote to a library function?
----------------------------------------------------------------------*/

void
slice_fn(
    char *fname,     /* Buffer to contain generated file name */
    int len,         /* Length of the buffer */
    char *pattern,   /* Filename, with # at position of slice number */
    char *sfmt,      /* Format for slice number.  E.g. "%03d" */
    int  sliceno)    /* Slice number */
{
    char *pos;       /* Position of slice number */
    char *p2;        /* Position just after slice number */
    int mls;         /* Maximum length of slice number */

    /* Find postion of slice number */
    pos = strchr(pattern, '#');
    if (pos == NULL)
	fatal("Slice pattern contains no # character");
    for (p2 = pos; *p2 == '#'; p2++) {}
    mls = len - (pos - pattern) - strlen(p2) - 1;

    /* Build up the output string */
    /* FIXME:  Potential buffer overflow here */
    strncpy(fname, pattern, pos-pattern);
    sprintf(fname+(pos-pattern), sfmt, sliceno);
    strcat(fname, p2);

    return;
}
