/****************************************************************************
inleav2.c - Interleave slices from 2 image files

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program reads slices alternately from two raw image files and
concatenates them into a single file.  More precisely, it reads
a given number of bytes alternately from the two files and outputs
them to the output file.

See the man page for a full description.

*****************************************************************************/

static const char rcsid[] = "$Id: inleav2.c,v 1.1 1999/05/28 18:22:20 hughett Exp $ Release $Name:  $";

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>       /* For getopt and friends */

#include "errm.h"


int
main(int argc, char *argv[])
{
    FILE *in1, *in2;                /* The input files */
    FILE *out = stdout;             /* The output file */
    size_t nbytes = 131072;         /* Number of bytes per slice */
    unsigned char *buf1, *buf2;     /* Input/output buffers */
    size_t n1, n2, n;
    int c;

    /* Get the command options, if any */
    while ( (c = getopt(argc, argv, "s:")) != EOF )
	switch (c) {
	case 's':
	    nbytes = strtoul(optarg,0,0);
	    break;
	default:
	    fatal("Option %c was not recognized.", c);
	    break;
	}

    /* Get the input file names, and open the files */
    if (argc-optind < 2)
	fatal("Fewer than 2 input files provided.");
    in1 = fopen(argv[optind], "rb");
    if (in1 == 0)
	fatal("Unable to open input file %s", argv[optind]);
    in2 = fopen(argv[optind+1], "rb");
    if (in2 == 0)
	fatal("Unable to open input file %s", argv[optind+1]);

    /* Get the output file name, if any, and open it */
    if (argc-optind >= 3) {
	out = fopen(argv[optind+2], "wb");
	if (out == 0)
	    fatal("Unable to open output file %s", argv[optind+2]);
    }

    /* Allocate buffer areas for input and output */
    buf1 = malloc(nbytes);
    buf2 = malloc(nbytes);
    if (buf1 == 0 || buf2 == 0)
	fatal("Unable to allocate i/o buffer");

    /* Read slices alternately and write sequentially */
    while (1) {

	/* Read a slice from each input file */
	n1 = fread(buf1, 1, nbytes, in1);
	n2 = fread(buf2, 1, nbytes, in2);

	/* Done if now at EOF on both input files */
	if (n1 == 0 && feof(in1) && n2 == 0 && feof(in2))
	    break;

	/* Check for possible i/o errors */
	if (n1 == 0 && ferror(in1))
	    fatal("Error reading first input file");
	if (n2 == 0 && ferror(in2))
	    fatal("Error reading second input file");

	/* Check for incomplete slices */
	if (n1 < nbytes)
	    fatal("Incomplete final slice on first input file");
	if (n2 < nbytes)
	    fatal("Incomplete final slice on second input file");

	/* Write out the two slices */
	n = fwrite(buf1, 1, nbytes, out);
	if (n < nbytes)
	    fatal("Error writing to output file");
	n = fwrite(buf2, 1, nbytes, out);
	if (n < nbytes)
	    fatal("Error writing to output file");
    }

    return EXIT_SUCCESS;
}
