/****************************************************************************
lovar.c - Compute local variance of a 3d unsigned char image

Author:  Paul Hughett

Copyright (c) 1999-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program computes the local variance over a 3x3x3 neighborhood
of a 3d unsigned char image and outputs it as the standard deviation
scaled to fit into a unsigned char image.  This is a simple form of
edge detector.

The following parameters are taken from the param file named on the
command line:

image   = Filename of the image to be segmented
imgout  = Name of the output file to be created
debug   = Level of debugging output (0-3)

The debugging levels are as follows:  0 => silent execution; 1 =>
report beginning and end of each step; 2 => briefly report results
of each step; 3 => write output files showing intermediate results
in detail.  Not all of these have necessarily been implemented.

NOTE: This program was modified on 2000-12-20 to use vxl_lostat rather
than vxl_lovar_uchar3d; the newer function uses an improved algorithm
but will not produce exactly the same values as the old one.

FIXME: Now that we have lostat implemented within Pyvox, it's not
clear that this file-based program retains any utility.  Consider
removing it or replacing it with a Pyvox script.  (Replacing with a
script would require teaching Pyvox how to read parameter files--a
dubious accomplishment.)

FIXME:  If image slices are widely separated, then doing the variance
only within each slice and not between slices should give better results.
**************************************************************************/


static const char rcsid[] = "$Id: lovar.c,v 1.2 2000/12/20 20:34:30 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "cdata.h"
#include "errm.h"
#include "param.h"
#include "voxel.h"




int
main(int argc, char **argv)
{
    Param_list params;             /* Contents of param file */
    char *stage = "lovar";         /* Default stage name for params */
    int debug = 3;                 /* Level of debugging output (Not used)*/
    const char *imagefn = 0;       /* Filename for the input image */
    const char *imoutfn = 0;       /* Filename for the output image */
    FILE *imgin;                   /* Stream for the first image */
    FILE *imgout;                  /* Stream for the output image */
    voxel_array *image  = 0;       /* Voxel array for input image */
    voxel_array *lomean = 0;       /* Voxel array for the local mean */
    voxel_array *lovar  = 0;       /* Voxel array for the local variance */
    int rank = 3;                  /* Number of dimensions in image */
    long nslices, nrows, ncols;    /* Image dimensions */
    long dimen[3];                 /* Array of the dimensions */
    const char *s;

    /* Determine the parameters for this run */
    if (argc < 2)
	fatal("No param file specified");
    if (debug >= 1)
	fprintf(stderr, "Reading parameters...");
    params   = read_param_file(argv[1]);
    imagefn  = get_param(params, stage, "image");
    imoutfn  = get_param(params, stage, "imgout");
    debug    = (s = get_param(params, stage, "debug"))   ? atol(s) : debug;
    nslices  = (s = get_param(params, stage, "nslices")) ? atol(s) :  256;
    nrows    = (s = get_param(params, stage, "nrows")  ) ? atol(s) :  256;
    ncols    = (s = get_param(params, stage, "ncols")  ) ? atol(s) :  256;
    dimen[0] = nslices;
    dimen[1] = nrows;
    dimen[2] = ncols;

    /* Do some consistency checking on parameters */
    if (nslices < 1 || nslices > 4000)
	fatal("nslices has an implausible value");
    if (nrows < 1 || nrows > 4000)
	fatal("nrows has an implausible value");
    if (ncols < 1 || ncols > 4000)
	fatal("ncols has an implausible value");
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Open and read the image file */
    if (imagefn == 0)
	fatal("No input image file was specified");
    if (debug >= 1)
	fprintf(stderr, "Reading input image...");
    imgin = fopen(imagefn, "rb");
    if (imgin == 0)
	fatal("Unable to open image file %s", imagefn);
    image = vxl_create_array();
    vxl_read_uchar_data(image, imgin, rank, dimen);
    fclose(imgin);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Compute the local variance */
    if (debug >= 1)
        fprintf(stderr, "Computing local mean and variance...");
    lomean = vxl_create_array();
    lovar  = vxl_create_array();
    vxl_lostat(lomean, lovar, image, &neigh3d27, 2.0);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Write out the processed image */
    if (debug >= 1)
        fprintf(stderr, "Writing the variance image to a file...");
    if (imoutfn == 0)
	fatal("No imgout file was specified");
    imgout = fopen(imoutfn, "w");
    if (imgout == 0)
	fatal("Unable to open output file %s", imoutfn);
    vxl_write_uchar_data(imgout, lovar);
    fclose(imgout);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Clean up and exit */
    if (lomean != 0) vxl_destroy_array(lomean);
    if (lovar  != 0) vxl_destroy_array(lovar);
    if (image  != 0) vxl_destroy_array(image);

    exit(EXIT_SUCCESS);
}
