/****************************************************************************
matrix.c - Matrix calculations on voxel arrays

Author:  Paul Hughett

Copyright (c) 2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file contains functions for doing matrix calculations on voxel
arrays, mostly using the LAPACK and BLAS Fortran libraries.  It is
part of the Voxel Kit, a set of general-purpose functions for volume
image processing, with a bias toward medical imaging applications.

The contents of this file are still somewhat experimental.  It's not
yet entirely clear how to portably interface between C and Fortran and
how to package the numerical linear algebra operations for use by
Voxel Kit users.

Tentative policy: Make these functions simple and natural to use, even
if it takes extra time to transpose or otherwise convert from C to
Fortran.  The assumption here is that matrix operations are minor
compared to the true image operations.  Q: As long as you need to 
copy most everything anyway, convert to double and save the need to
provide both real and double versions of everything?

Tentative policy: Assume that long <-> integer, float <-> real, and
double <-> double precision, but arrange for compile errors if these
are false.

FIXME: Complete the minimal set of matrix operations: transpose
(generalize for n-dimen arrays); diag and get diag; least squares; 
SVD; pseudoinverse; and ???

FIXME: Include these functions in voxel.h or add a new header matrix.h?
For the moment, leave them in voxel.h while I decide if this is the
right approach.

FIXME: The repeated duplication of nearly identical code for double and
float versions just cries out for C++ templates.  (The repeated code
is a little too long to lend itself to macros.)  Automatic generation
from a more concise representation is another possibility.

FIXME:  Identify common validity checking steps and isolate as functions
or macros.  Q:  What are the efficiency tradeoffs between the two?
*****************************************************************************/

static const char rcsid[] = "$Id: matrix.c,v 1.8 2001/10/24 15:41:19 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"
#include "bips.h"
#include "clap.h"



/*-------------------------------------------------------------------------
matrix_check - Verify that voxel array represents a matrix
-------------------------------------------------------------------------*/

static inline int
matrix_check(
    voxel_array *matrix)
{
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	return 0;
    if ( matrix->type != INTP_FLOAT && matrix->type != INTP_DOUBLE )
	return 0;
    return 1;
}



/*-------------------------------------------------------------------------
matrix_check_square - Verify that voxel array represents a square matrix
-------------------------------------------------------------------------*/

static inline int
matrix_check_square(
    voxel_array *matrix)
{
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	return 0;
    if ( matrix->type != INTP_FLOAT && matrix->type != INTP_DOUBLE )
	return 0;
    if ( matrix->dimen[0] != matrix->dimen[1] )
	return 0;
    return 1;
}



/*--------------------------------------------------------------------------
matrix_determinant - Determinant of a voxel array

This function computes the determinant of a voxel array considered as a
matrix.

ALGORITHM: The documentation for the pivot vector returned by xGETRF
is somewhat confusing.  It appears that it is NOT a permutation per
se, but a list of the row interchanges that were done.  Thus the
vector 1 3 3 indicates that row 1 was untouched, row 2 was
interchanged with 3, and the new(!) row 3 was untouched.  Assuming
that this interpretation is correct, the sign of the permutation can
be determined simply by counting the number of interchanges; how to
compute the permutation matrix is less clear.

FIXME: Allocating new storage each time for mdata and permute is
inefficient if many determinants must be calculated.  Storage allocated
on the stack for sufficiently small arrays would be better.

FIXME: Maintaining parallel double and float versions is likely to be
tedious and error-prone.  Is there a better method?
---------------------------------------------------------------------------*/

double
matrix_determinant(
    voxel_array *matrix)       /* Matrix to take determinant of */
{
    double det = 1.0;          /* Determinant value */
    fint  *pivot  = 0;         /* Pivot vector */
    fint m, n, lda;            /* Dimensions of matrix */
    fint info;                 /* Info return from LAPACK */
    int i;

    /* Verify rank, shape, and type of array */
    if ( !matrix_check_square(matrix) )
	fatal("Not a matrix or not square");

    /* Do type-independent set up for LAPACK */
    m = n = lda = matrix->dimen[0];
    pivot = mallock(m*sizeof(fint));

    /* Switch on array type to computational function */
    switch (matrix->type) {

    case INTP_FLOAT:
	{
	    float  *data  = 0;     /* Data from the voxel array */
	    freal  *mdata = 0;     /* Matrix data */

	    /* Copy matrix data into working array */
	    data  = matrix->data;
	    mdata = mallock(m*n*sizeof(freal));
	    for (i = 0; i < m*n; i++)
		mdata[i] = data[i];

	    /* Call LAPACK for LU decomposition */
	    sgetrf_(&m, &n, mdata, &lda, pivot, &info);
	    if (info < 0)
		fatal("Error calling xGETRF");

	    /* Compute determinant from diagonal of LU decomposition */
	    if (info > 0)       /* Matrix is exactly singular */
		return 0.0;
	    det = 1.0;
	    for (i = 0; i < m*n; i += n+1)
		det *= mdata[i];
	    free(mdata);

	    break;
	}

    case INTP_DOUBLE:
	{
	    double  *data  = 0;     /* Data from the voxel array */
	    fdouble *mdata = 0;     /* Matrix data */

	    /* Copy matrix data into working array */
	    data  = matrix->data;
	    mdata = mallock(m*n*sizeof(fdouble));
	    for (i = 0; i < m*n; i++)
		mdata[i] = data[i];

	    /* Call LAPACK for LU decomposition */
	    dgetrf_(&m, &n, mdata, &lda, pivot, &info);
	    if (info < 0)
		fatal("Error calling xGETRF");

	    /* Compute determinant from diagonal of LU decomposition */
	    if (info > 0)       /* Matrix is exactly singular */
		return 0.0;
	    det = 1.0;
	    for (i = 0; i < m*n; i += n+1)
		det *= mdata[i];
	    free(mdata);

	    break;
	}

    default:
	fatal("Array type is neither float nor double");
	return 0.0;
    }
    
    /* Multiply determinant by sign of permutation */
    for (i = 0; i < m; i++) {
	if (pivot[i] != i+1)          /* Fortran origin is 1 */
	    det = -det;
    }
    free(pivot);

    return det;
}



/*-------------------------------------------------------------------------
matrix_diag - Construct diagonal matrix from given elements

This function creates a diagonal matrix from a vector of given values.
Unlike the Matlab or Octave function diag, it will not extract the
diagonal elements from a matrix; the function matrix_get_diag is used
for that purpose.

FIXME: This function is no longer used in Pyvox, so beware of errors.
--------------------------------------------------------------------------*/

void
matrix_diag(
    voxel_array *diag,      /* (out) Diagonal matrix */
    voxel_array *vector)    /* (in) Vector of diagonal elements */
{
    int type;               /* Type of data in matrix */
    long n, dimen[2];       /* Dimensions of result */

    /* Verify rank */
    if ( vector == 0 || vector->magic != VXL_MAGIC ||
	 vector->rank != 1 || vector->data == 0 )
	fatal("Argument is not a valid vector");
    type = vector->type;

    /* Allocate voxel array for result */
    n = dimen[0] = dimen[1] = vector->dimen[0];
    vxl_alloc_array(diag, type, 2, dimen);

    /* Fill result with zeros */
    bips_zero(n*n, type, diag->data, 1);

    /* Fill diagonal with given elements */
    bips_copy(n, type, diag->data, n+1, vector->data, 1);

    return;
}


/*--------------------------------------------------------------------------
matrix_eigen_symmetric - Solve real symmetric eigensystem

This function computes the eigenvalues and eigenvectors of a voxel
array considered as a real symmetric matrix.  The eigenvalues are
returned as a rank-1 voxel array.  The eigenvectors are returned as a
square rank-2 voxel array, the rows of which are the eigenvectors.
The matrix of eigenvectors is guaranteed to be orthogonal but not
necessarily to be proper. (If a proper rotation is necessary, use
matrix_principal_axes instead.)

FIXME: Check for compatibility between C and Fortran types.
---------------------------------------------------------------------------*/

void
matrix_eigen_symmetric(
    voxel_array *values,          /* Vector of eigenvalues */
    voxel_array *vectors,         /* Matrix of eigenvectors */
    voxel_array *matrix)          /* Matrix to solve */
{
    int type;                     /* Type of matrix */
    long dimen[2];                /* Dimensions of matrix */
    /* FIXME: int or long for next 2 varibs? */
    fint n;                      /* Dimension of matrix */
    fint lda;                    /* Last dimension of matrix */
    fchar1 jobz;                 /* Job selector */
    fchar1 uplo;                 /* Upper or lower triangle? */
    fint info;                   /* Status returned from LAPACK */
    fint lwork;                  /* Size of working storage */
    long i;

    /* Verify rank, shape, and type of matrix */
    if ( !matrix_check_square(matrix) )
	fatal("Not a matrix or not square");
    type = matrix->type;
    dimen[0] = dimen[1] = matrix->dimen[0];

    /* Allocate destination arrays */
    vxl_alloc_array(values,  type, 1, dimen);
    vxl_alloc_array(vectors, type, 2, dimen);

    /* Set up type-independent arguments for LAPACK */
    n = lda = matrix->dimen[0];
    jobz = 'V';           /* Both eigenvalues and vectors */
    uplo = 'U';           /* Lower triangle (in C!) */

    /* Switch on array type to computational function */
    switch (type) {

    case INTP_FLOAT:
	{
	    freal *work;        /* Working storage for LAPACK */
	    freal wsize;        /* Recommended size of working storage */
	    float *matdata;     /* Matrix data */
	    float *valdata;     /* Eigenvalue data */
	    float *vecdata;     /* Eigenvector data */

	    /* Set up remaining LAPACK arguments */
	    matdata = matrix->data;
	    valdata = values->data;
	    vecdata = vectors->data;

	    /* Copy matrix into eigenvector array for computation in place */
	    for ( i = 0; i < n*n; i++)
		vecdata[i] = matdata[i];

	    /* Allocate working storage */
	    lwork = -1;
	    ssyev_(&jobz, &uplo, &n, vecdata, &lda, valdata, &wsize, 
		   &lwork, &info);
	    lwork = (fint)wsize;
	    work = mallock(lwork*sizeof(float));

	    /* Call LAPACK to do the work */
	    ssyev_(&jobz, &uplo, &n, vecdata, &lda, valdata, work, 
		   &lwork, &info);

	    /* Free allocated storage */
	    free(work);
	    break;
	}

    case INTP_DOUBLE:
	{
	    fdouble *work;       /* Working storage for LAPACK */
	    fdouble wsize;       /* Recommended size of working storage */
	    double *matdata;     /* Matrix data */
	    double *valdata;     /* Eigenvalue data */
	    double *vecdata;     /* Eigenvector data */

	    /* Set up remaining LAPACK arguments */
	    matdata = matrix->data;
	    valdata = values->data;
	    vecdata = vectors->data;

	    /* Copy matrix into eigenvector array for computation in place */
	    for ( i = 0; i < n*n; i++)
		vecdata[i] = matdata[i];

	    /* Allocate working storage */
	    lwork = -1;
	    dsyev_(&jobz, &uplo, &n, vecdata, &lda, valdata, &wsize, 
		   &lwork, &info);
	    lwork = (fint)wsize;
	    work = mallock(lwork*sizeof(double));

	    /* Call LAPACK to do the work */
	    dsyev_(&jobz, &uplo, &n, vecdata, &lda, valdata, work, 
		   &lwork, &info);

	    /* Free allocated storage */
	    free(work);
	    break;
	}

    default:
	fatal("Array type is neither float nor double");
	return;
    }

    /* Check for LAPACK errors */
    if (info < 0)
	fatal("Error calling xSYEV");
    if (info > 0)
	fatal("Eigenvalue calculation failed to converge");

    return;
}



/*----------------------------------------------------------------------
matrix_get_diag - Extract diagonal elements of a matrix

This function extracts the diagonal elements of a rectangular matrix
and returns them as a vector of the same type as the matrix.
-----------------------------------------------------------------------*/

void
matrix_get_diag(
    voxel_array *diagonal,        /* Diagonal elements */
    voxel_array *matrix)          /* Matrix to get diagonal of */
{
    int type;                     /* Type of data in matrix */
    long *dimen;                  /* Dimensions of matrix */
    long ss;                      /* Stride between diagonal elements */
    long n;                       /* Dimension of vector result */

    /* Check rank, type, and dimensions of the matrix */
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	fatal("Argument is not a valid matrix");
    type  = matrix->type;
    dimen = matrix->dimen;

    /* Allocate voxel array for the vector result */
    n = dimen[0];
    if ( n > dimen[1] )
	n = dimen[1];
    vxl_alloc_array(diagonal, type, 1, &n);

    /* Copy diagonal elements to result */
    ss = dimen[1] + 1;
    bips_copy(n, type, diagonal->data, 1, matrix->data, ss);

    return;
}



/*-----------------------------------------------------------------------
matrix_inverse - Inverse of a matrix

This function computes the inverse of a non-singular square matrix
represented as a voxel array of type float or double.  It is a fatal
error if the matrix is singular.

FIXME: Do something better than dying if the matrix turns out to be
singular.
-------------------------------------------------------------------------*/

void
matrix_inverse(
    voxel_array *inverse,        /* Inverse of matrix */
    voxel_array *matrix)         /* Matrix to find inverse of */
{
    int type;                    /* Type of matrix */
    long dimen[2], n;            /* Matrix dimensions */
    fint *pivot = 0;             /* Pivot vector */
    fint lda;                    /* Last dimension of matrix */
    fint lwork;                  /* Length of working storage */
    fint info;                   /* Info return from LAPACK */
    long i;

    /* Verify rank, shape, and type of matrix */
    if ( !matrix_check_square(matrix) )
	fatal("Not a matrix or not square");
    type = matrix->type;
    n = dimen[0] = dimen[1] = matrix->dimen[0];

    /* Allocate destination array */
    vxl_alloc_array(inverse, type, 2, dimen);

    /* Do type-independent set up for LAPACK */
    lda   = matrix->dimen[0];
    pivot = mallock(n*sizeof(fint));

    /* Switch on matrix data type */
    switch (type) {

    case INTP_DOUBLE:
	{
	    double  *mdata = 0;          /* Matrix data */
	    double  *idata = 0;          /* Inverse data */
	    fdouble *work  = 0;          /* Working storage */
	    fdouble wsize;               /* Recommended size */

	    /* Copy matrix data */
	    mdata = matrix ->data;
	    idata = inverse->data;
	    for (i = 0; i < n*n; i++)
		idata[i] = mdata[i];

	    /* Compute LU decomposition */
	    dgetrf_(&n, &n, idata, &lda, pivot, &info);
	    if ( info < 0 )
		fatal("Error calling xGETRF");
	    if ( info > 0 )
		fatal("Matrix is singular");

	    /* Allocate working storage for inverse */
	    lwork = -1;
	    dgetri_(&n, idata, &lda, pivot, &wsize, &lwork, &info);
	    lwork = wsize;
	    work = mallock(lwork*sizeof(fdouble));

	    /* Compute inverse */
	    dgetri_(&n, idata, &lda, pivot, work, &lwork, &info);
	    if ( info != 0 )
		fatal("Error calling xGETRI");
	    free(work);

	    break;
	}

    case INTP_FLOAT:
	{
	    float *mdata = 0;          /* Matrix data */
	    float *idata = 0;          /* Inverse data */
	    freal *work  = 0;          /* Working storage */
	    freal wsize;               /* Recommended size */

	    /* Copy matrix data */
	    mdata = matrix ->data;
	    idata = inverse->data;
	    for (i = 0; i < n*n; i++)
		idata[i] = mdata[i];

	    /* Compute LU decomposition */
	    sgetrf_(&n, &n, idata, &lda, pivot, &info);
	    if ( info < 0 )
		fatal("Error calling xGETRF");
	    if ( info > 0 )
		fatal("Matrix is singular");

	    /* Allocate working storage for inverse */
	    lwork = -1;
	    sgetri_(&n, idata, &lda, pivot, &wsize, &lwork, &info);
	    lwork = wsize;
	    work = mallock(lwork*sizeof(freal));

	    /* Compute inverse */
	    sgetri_(&n, idata, &lda, pivot, work, &lwork, &info);
	    if ( info != 0 )
		fatal("Error calling xGETRI");
	    free(work);

	    break;
	}

    default: 
	fatal("Array type is neither float nor double");
	return;
    }

    /* Free temporary storage and return */
    free(pivot);
    return;
}



/*--------------------------------------------------------------------------
matrix_principal_axes - Principal axes decomposition of symmetric matrix

This function computes the principal axes decomposition of a voxel array
considered as a real symmetric matrix; it is equivalent to solving the
real symmetric eigensystem using matrix_eigen_symmetric except that
the matrix of eigenvectors is adjusted if necessary to ensure that it
defines a proper rotation.

FIXME: Perhaps define row operations, to avoid that ugly jump into
type-specific operations at the end.
---------------------------------------------------------------------------*/

void
matrix_principal_axes(
    voxel_array *values,      /* Eigenvalues, as vector */
    voxel_array *vectors,     /* Proper orthogonal matrix of eigenvectors */
    voxel_array *matrix)      /* Matrix to decompose */
{
    double det;        /* Determinant of matrix of eigenvectors */
    int type;          /* Type of matrix */
    long n, i;

    /* Solve the eigensystem */
    matrix_eigen_symmetric(values, vectors, matrix);

    /* Compute determinant of rotation matrix */
    det = matrix_determinant(vectors);

    /* If det == -1, invert last row of rotation matrix */
    if (det < 0) {
	n = matrix->dimen[0];
	type = matrix->type;
	switch (type) {

	case INTP_FLOAT:
	    {
		float *data;
		data = vectors->data;
		for (i = 0; i < n; i++) {
		    data[n*(n-1)+i] = -data[n*(n-1)+i];
		}
		break;
	    }

	case INTP_DOUBLE:
	    {
		double *data;
		data = vectors->data;
		for (i = 0; i < n; i++) {
		    data[n*(n-1)+i] = -data[n*(n-1)+i];
		}
		break;
	    }

	default:
	    fatal("Matrix is neither float nor double");

	} /* end of switch */

    } /* end of if */

    return;
}



/*-----------------------------------------------------------------------
matrix_solve - Solve a linear system AX = B

This function uses the LU decomposition to solve the linear system AX
= B for compatible float or double matrices A and B.  It is a fatal
error if the matrix is singular.

FIXME: Do something better than dying if the matrix turns out to be
singular.

ALGORITHM: Both A and B must be transposed to Fortran conventions before
being passed to LAPACK; the solution X must be transposed back to C
conventions before being returned.
-------------------------------------------------------------------------*/

void
matrix_solve(
    voxel_array *soln,           /* Solution X of system */
    voxel_array *matrix,         /* Matrix A on left-hand side */
    voxel_array *rhs)            /* Matrix B on right-hand side */
{
    int type;                    /* Commmon type of matrices */
    long adim[2], bdim[2];       /* Dimensions of A and B matrices */
    fint *pivot = 0;             /* Pivot vector */
    fint lda, ldb;               /* Last dimension of matrix */
    fint info;                   /* Info return from LAPACK */
    fchar1 tsel = 'N';           /* Do not transpose A in LAPACK */
    long i, j;

    /* Verify shape and type of input matrices */
    if ( !matrix_check_square(matrix) )
	fatal("Left-hand side not a square matrix");
    lda = adim[0] = adim[1] = matrix->dimen[0];
    type = matrix->type;
    if ( !matrix_check(rhs) )
	fatal("Right-hand side not a matrix");
    if ( type != rhs->type )
	fatal("Types of A and B matrices do not match");
    bdim[0] = rhs->dimen[0];
    bdim[1] = rhs->dimen[1];
    ldb     = bdim[0];
    if ( adim[1] != bdim[0] )
        fatal("A and B matrices have incompatible shapes");
        
    /* Allocate destination array */
    vxl_alloc_array(soln, type, 2, bdim);

    /* Do type-independent set up for LAPACK */
    pivot = mallock(adim[0]*sizeof(fint));

    /* Switch on matrix data type */
    switch (type) {

    case INTP_DOUBLE:
	{
	    double  *mdata = 0;          /* Matrix data A */
	    double  *rdata = 0;          /* RHS data B */
            double  *sdata = 0;          /* Solution data X */
	    fdouble *adata = 0;          /* Transposed copy of A */
            fdouble *bdata = 0;          /* Transposed copy of B or X */

	    /* Copy/transpose A to temporary storage */
	    adata = mallock(adim[0]*adim[1]*sizeof(fdouble));
	    mdata = matrix->data;
	    for (i = 0; i < adim[0]; i++)
		for (j = 0; j < adim[1]; j++)
		    adata[j*adim[0]+i] = mdata[i*adim[1]+j];

	    /* Copy/transpose B to temporary storage */
	    bdata = mallock(bdim[0]*bdim[1]*sizeof(fdouble));
	    rdata = rhs->data;
	    for (i = 0; i < bdim[0]; i++)
		for (j = 0; j < bdim[1]; j++)
		    bdata[j*bdim[0]+i] = rdata[i*bdim[1]+j];

	    /* Compute LU decomposition */
	    dgetrf_(&adim[0], &adim[0], adata, &lda, pivot, &info);
	    if ( info < 0 )
		fatal("Error calling xGETRF");
	    if ( info > 0 )
		fatal("Matrix is singular");

	    /* Use LU decomposition to solve system */
	    dgetrs_(&tsel, &adim[0], &bdim[1], adata, &lda, pivot, 
		    bdata, &ldb, &info);
	    if ( info < 0 )
		fatal("Error calling xGETRS");

	    /* Copy/transpose the solution X from temporary storage */
	    sdata = soln->data;
	    for (i = 0; i < bdim[0]; i++)
		for (j = 0; j < bdim[1]; j++)
		    sdata[i*bdim[1]+j] = bdata[j*bdim[0]+i];

	    /* Free temporary storage */
	    free(bdata);
	    free(adata);

	    break;
	}

    case INTP_FLOAT:
	{
	    float  *mdata = 0;          /* Matrix data A */
	    float  *rdata = 0;          /* RHS data B */
            float  *sdata = 0;          /* Solution data X */
	    freal  *adata = 0;          /* Transposed copy of A */
            freal  *bdata = 0;          /* Transposed copy of B or X */

	    /* Copy/transpose A to temporary storage */
	    adata = mallock(adim[0]*adim[1]*sizeof(freal));
	    mdata = matrix->data;
	    for (i = 0; i < adim[0]; i++)
		for (j = 0; j < adim[1]; j++)
		    adata[j*adim[0]+i] = mdata[i*adim[1]+j];

	    /* Copy/transpose B to temporary storage */
	    bdata = mallock(bdim[0]*bdim[1]*sizeof(freal));
	    rdata = rhs->data;
	    for (i = 0; i < bdim[0]; i++)
		for (j = 0; j < bdim[1]; j++)
		    bdata[j*bdim[0]+i] = rdata[i*bdim[1]+j];

	    /* Compute LU decomposition */
	    sgetrf_(&adim[0], &adim[0], adata, &lda, pivot, &info);
	    if ( info < 0 )
		fatal("Error calling xGETRF");
	    if ( info > 0 )
		fatal("Matrix is singular");

	    /* Use LU decomposition to solve system */
	    sgetrs_(&tsel, &adim[0], &bdim[1], adata, &lda, pivot, 
		    bdata, &ldb, &info);
	    if ( info < 0 )
		fatal("Error calling xGETRS");

	    /* Copy/transpose the solution X from temporary storage */
	    sdata = soln->data;
	    for (i = 0; i < bdim[0]; i++)
		for (j = 0; j < bdim[1]; j++)
		    sdata[i*bdim[1]+j] = bdata[j*bdim[0]+i];

	    /* Free temporary storage */
	    free(bdata);
	    free(adata);

	    break;
	}

    default: 
	fatal("Array type is neither float nor double");
	return;
    }

    /* Free temporary storage and return */
    free(pivot);
    return;
}



/*--------------------------------------------------------------------------
matrix_transpose - Transpose of a matrix

This function computes the transpose of a matrix.

FIXME: Generalize to do any permutation of coordinates on an array of
any rank, and to reverse the order of indices on an axis.  Also
generalize to work on arrays of any type.

FIXME: Block for better cache performance.
----------------------------------------------------------------------------*/

void
matrix_transpose(
    voxel_array *trans,     /* Transposed matrix */
    voxel_array *matrix)    /* Matrix to be transposed */
{
    int type;               /* Type of data in matrix */
    int nbytes;             /* Bytes per data element */
    long tdimen[2];         /* Dimensions of transposed matrix */
    long nsrows;            /* Src rows = dest cols, in elements */
    long nscols;            /* Src cols = dest rows, in elements */
    unsigned char *src;     /* Source row addr, in bytes */
    unsigned char *dst;     /* Destination column addr, in bytes */
    int info;               /* Status return from BIPS */
    long i;

    /* Verify rank and type */
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	fatal("Argument is not a matrix");
    type   = matrix->type;
    nbytes = matrix->nbytes;   

    /* Allocate voxel array for transposed matrix */
    tdimen[1] = nsrows = matrix->dimen[0];
    tdimen[0] = nscols = matrix->dimen[1]; 
    vxl_alloc_array(trans, type, 2, tdimen);

    /* Copy rows into columns */
    dst = trans ->data;
    src = matrix->data;
    for (i = 0; i < nsrows; i++) {
	info = bips_copy(nscols, type, dst, nsrows, src, 1);
	if (info != 0)
	    fatal("Error calling bips_copy");
	dst += nbytes;
	src += nbytes * nscols;
    }

    return;
}
