/****************************************************************************
memm.c - Memory management

Author:  Paul Hughett

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file provides a standard interface for memory allocation and
deallocation within software written at the Brain Behavior Laboratory
at the University of Pennsylvania.  The current version consists of versions
of malloc, calloc, and realloc which check whether memory was successfully
allocated and report an error if not.  Others may be added in the future,
such as the ability to maintain a fast pool of small blocks of fixed size.
*****************************************************************************/

static const char rcsid[] = "$Id: memm.c,v 1.1 2000/03/10 20:28:55 hughett Exp $ Release $Name:  $";


#include <stdlib.h>

#include "errm.h"
#include "memm.h"



/*-------------------------------------------------------------------------
callock - Allocate and clear memory, with error checking

This function calls the standard library routine calloc and generates
a fatal error if it cannot reallocate a block of the requested size.
If the requested block size is non-positive, a warning is generated and
zero is returned.
---------------------------------------------------------------------------*/

void *
callock( size_t nelem,
	 size_t elsize )
{
  void * retval;

  if (nelem <= 0 || elsize <= 0) {
      warning("callock: Block requested with non-positive length.");
      return 0; }

  retval = calloc( nelem, elsize );
  if (retval == NULL) 
      fatal("callock:  Unable to allocate %d bytes.", nelem*elsize);

  return retval;
}



/*---------------------------------------------------------------------------
mallock - Allocate memory, with error message if not available.

This function calls the standard library routine malloc and generates
a fatal error if it cannot allocate a block of the requested size.
To help catch memory errors, attempting to allocate a block with
non-positive length generates a warning message but gets a zero
pointer.
----------------------------------------------------------------------------*/

void *
mallock( size_t size )
{
  void * retval;

  if (size <= 0) {
      warning("mallock: Block requested with non-positive length");
      return 0; }

  retval = malloc( size );
  if ( retval == NULL ) 
      fatal("mallock:  Unable to allocate %d bytes.", size);

  return retval;
}



/*---------------------------------------------------------------------------
reallock - Reallocate memory, with error message if not available.

This function calls the standard library routine realloc and generates
a fatal error if it cannot reallocate a block of the requested size.
To help catch memory errors, attempting to reallocate a block with
negative length generates a warning message but gets a zero
pointer.
----------------------------------------------------------------------------*/

void *
reallock( void *ptr, size_t size )
{
  void * retval;

  if (size < 0) {
      warning("reallock: Block requested with negative length");
      return 0; }

  retval = realloc( ptr, size );
  if ( retval == NULL ) 
      fatal("reallock:  Unable to reallocate %d bytes.", size);

  return retval;
}
