/****************************************************************************
neighbor.c - Voxel Kit routines for neighborhood operations

Author:  Paul Hughett

Copyright (c) 1998-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file contains code for neighborhood
operations and related functions.

Neighborhood operations are those in which each output voxel depends on
a neighborhood of the corresponding input voxel.  Some examples include
local mean, local variance, erosion, dilation, and convolution; some
less obvious examples include the Gibbs energy, surface area by counting
methods, the median filter, and ICM (iterated conditional modes).

These functions all face similar algorithmic problems in choosing the
access patterns to use a multi-level cache efficiently.  Several
different algorithms currently coexist in this section, while I
experiment with different trade-offs between implementation time,
readability, and efficiency; the trend over time can be expected to be
toward improving efficiency, even at the cost of implementation time
and readability, since these functions are likely to be major
contributors to the total run time.  A further trend will be to
support neighborhoods of arbitrary shape in images of arbitrary rank.

The vxl_kernel struct is used to define a neighborhood in any number
of dimensions; see its definition for the precise details.  Generally,
it specifies the neighbors of a voxel by giving coordinate deltas;
a convolution coefficient may also be specified for each neighbor.

Note that the algorithms used here are designed to maximize locality
within a multilevel cache architecture, thereby decreasing the run
time.  If you don't understand what that means, you probably shouldn't
touch this code, under penalty of a possible order-of-magnitude
increase in run time.  The code is also optimized for the case that
most voxels are interior rather than boundary voxels.
*****************************************************************************/

static const char rcsid[] = "$Id: neighbor.c,v 1.2 2001/10/19 18:47:01 hughett Exp $ Release $Name:  $";

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>             /* For memset */

#include "config.h"
#include "bips.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"



/*------------------------------------------------------------------------
                        Common Data Structures

FIXME: The better way to define all these kernels would be to write
functions that create kernels with desired characteristics on demand.
--------------------------------------------------------------------------*/

/* The 9 in-plane neighbors of a voxel in 3d */
static long n3d9ip[][VXL_MAX_RANK] = {
    { 0, -1, -1}, { 0, -1,  0}, { 0, -1, +1},
    { 0,  0, -1}, { 0,  0,  0}, { 0,  0, +1},
    { 0, +1, -1}, { 0, +1,  0}, { 0, +1, +1},
};

struct vxl_kernel neigh3d9ip = {3, 9, n3d9ip, NULL, 0.0};



/* The 27 neighbors of a voxel in 3D */
static long n3d27[][VXL_MAX_RANK] = {
    {-1, -1, -1}, {-1, -1,  0}, {-1, -1, +1},
    {-1,  0, -1}, {-1,  0,  0}, {-1,  0, +1},
    {-1, +1, -1}, {-1, +1,  0}, {-1, +1, +1},
    { 0, -1, -1}, { 0, -1,  0}, { 0, -1, +1},
    { 0,  0, -1}, { 0,  0,  0}, { 0,  0, +1},
    { 0, +1, -1}, { 0, +1,  0}, { 0, +1, +1},
    {+1, -1, -1}, {+1, -1,  0}, {+1, -1, +1},
    {+1,  0, -1}, {+1,  0,  0}, {+1,  0, +1},
    {+1, +1, -1}, {+1, +1,  0}, {+1, +1, +1},
};

/* Coefficients for a 27-element rectangular window */
static double rect27[27] = {
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
};

struct vxl_kernel neigh3d27 = {3, 27, n3d27, rect27, 0.0};

/* The 13 neighbors of a voxel in 3D which precede it in raster order */
static long n3d13[][VXL_MAX_RANK] = {
    {-1, -1, -1}, {-1, -1,  0}, {-1, -1, +1},
    {-1,  0, -1}, {-1,  0,  0}, {-1,  0, +1},
    {-1, +1, -1}, {-1, +1,  0}, {-1, +1, +1},
    { 0, -1, -1}, { 0, -1,  0}, { 0, -1, +1},
    { 0,  0, -1}, 
};

struct vxl_kernel neigh3d13 = {3, 13, n3d13, NULL, 0.0};

/* Coefficients for lowpass filter to block Nyquist frequency */
static double lpcoefs[27] = {
    1.0/64.0, 1.0/32.0, 1.0/64.0,
    1.0/32.0, 1.0/16.0, 1.0/32.0,
    1.0/64.0, 1.0/32.0, 1.0/64.0,
    1.0/32.0, 1.0/16.0, 1.0/32.0,
    1.0/16.0, 1.0/ 8.0, 1.0/16.0,
    1.0/32.0, 1.0/16.0, 1.0/32.0,
    1.0/64.0, 1.0/32.0, 1.0/64.0,
    1.0/32.0, 1.0/16.0, 1.0/32.0,
    1.0/64.0, 1.0/32.0, 1.0/64.0,
};

/* Lowpass filter which precisely blocks Nyquist frequency in x, y, and z */
struct vxl_kernel lowpass3d = {3, 27, n3d27, lpcoefs, 0.0};



/*------------------------------------------------------------------------
vxl_create_kernel - Create a convolution kernel

This function creates a vxl_kernel structure from given neighborhood
coordinates, coefficients, and bias; the coefficients and bias are
optional.  Some basic consistency checks are made on the requested
kernel.  The function vxl_destroy_kernel can be used to free the
kernel when it is no longer necessary.

For safety, the delta and coef arrays are copied to newly allocated
storage; the originals may be recycled or freed by the caller upon
return from vxl_create_kernel.
-------------------------------------------------------------------------*/

struct vxl_kernel *
vxl_create_kernel(
    int rank,                      /* Number of dimensions */
    int count,                     /* Number of neighbors in the kernel */
    long (*delta)[VXL_MAX_RANK],   /* Coordinate deltas for each neightbor */
    double *coef,                  /* Coefficient array, or zero */
    double bias)                   /* Constant bias term for convolution */
{
    struct vxl_kernel *kern;       /* Kernel under construction */
    long (*ndelta)[VXL_MAX_RANK];  /* Copy of delta array */
    double *ncoef = 0;             /* Copy of coef array */
    int n;                         /* Index over neighbors */
    int d;                         /* Index over dimensions */

    /* Do some basic consistency checks */
    if (rank < 1 || rank > VXL_MAX_RANK)
	fatal("Invalid rank requested");
    if (count < 1 || count > VXL_MAX_NEIGH)
	fatal("Invalid neighbor count requested");

    /* Copy the delta array to newly allocated storage */
    ndelta = mallock(count*VXL_MAX_RANK*sizeof(long));
    for (n = 0; n < count; n++) 
	for (d = 0; d < VXL_MAX_RANK; d++)
	    ndelta[n][d] = (d < rank) ? delta[n][d] : 0;

    /* Copy the coef array, if present, to newly allocated storage */
    ncoef = 0;
    if (coef != 0) {
	ncoef  = mallock(count*sizeof(double));
	for (n = 0; n < count; n++)
	    ncoef[n] = coef[n];
    }

    /* Create and initialize the kernel */
    kern = mallock(sizeof(struct vxl_kernel));
    kern->rank  = rank;
    kern->count = count;
    kern->delta = ndelta;
    kern->coef  = ncoef;
    kern->bias  = bias;

    return kern;
}



/*------------------------------------------------------------------------
vxl_destroy_kernel - Free all storage associate with a kernel struct

This function frees the delta array, coef array (if any), and struct
storage associated with a kernel.  It is a serious error to attempt
to destroy a kernel that was not created with vxl_create_kernel.
-------------------------------------------------------------------------*/

void
vxl_destroy_kernel(
    struct vxl_kernel *kern)         /* The kernel to be destroyed */
{
    /* Do some consistency checks to help detect misteaks */
    if ( kern == 0 || kern->delta == 0 ||
         kern->rank  < 1 || kern->rank  > VXL_MAX_RANK ||
	 kern->count < 0 || kern->count > VXL_MAX_NEIGH )
	fatal("Attempting to destroy an invalid kernel");

    /* Looks okay, so destroy this kernel */
    if (kern->coef != 0)
	free(kern->coef);
    free(kern->delta);
    free(kern);    

    return;
}



/*--------------------------------------------------------------------------
vxli_rowout - Determine if an image row is outside the image

This function determines if the coordinates of an image row plus delta
coordinates are outside the boundaries of the image; this is a common
operation in the implementation of the neighborhood operations.  Error
checking is assumed to be done by the caller.
---------------------------------------------------------------------------*/

static inline int 
vxli_rowout(
    int rank,                  /* Rank of the voxel array */
    long *dimen,               /* Dimensions of the voxel array */
    long *coords,              /* Row coordinates to be checked */
    long *delta)               /* Offsets to be added before checking */
{
    int i;                     /* Index over dimensions */
    long y;                    /* Coordinate value along some dimension */

    for ( i = rank-2; i >= 0; i-- ) {
	y = coords[i] + delta[i];
	if ( y < 0 || y >= dimen[i] )
	    return 1;
    }

    return 0;
}



/*--------------------------------------------------------------------------
vxl_convolve - Do direct convolution on a image of any rank

This function performs a direct convolution and optional subsampling
on an image of any rank and data type, using a neighborhood and
coefficients defined by a vxl_kernel of the same rank.  The rank must
be one or greater; convolution is not defined on images of rank zero.
The coefficients are always of type double.  The type of the output
image is always the same as the type of the input image.  Intermediate
calculations are always done in type double and rounded back to the
final type; if a result voxel cannot be represented within range of
the result type, the result is set to the appropriate limit of the
range.  Incomplete neighborhoods at the boundary of the image are
handled correctly; that is, the missing voxels are simply omitted from
the neighborhood.  The shrink argument may be set to NULL, in which
case no subsampling is done.  In case you were wondering, the
subsampling is done so that the unneeded values are never computing,
which is faster than a plain convolution followed by a plain
subsampling.

FIXME: Implement functions to create convolution kernels of various
types, or to convert from a voxel array plus a center.  Provide
Pyvox bindings to these functions.
---------------------------------------------------------------------------*/

void
vxl_convolve(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern,   /* The convolution kernel to use */
    long *shrink)              /* Shrink factor along each dimension */
{
    unsigned char *dd, *sd;    /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    double *sum;               /* One row of convolution sums */
    double coef;               /* Coefficient from the convolution kernel */
    int type;                  /* (Internal) type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long ldsrc;                /* Nmbr elements on last dimen of source */
    long lddst;                /* Nmbr elements on last dimen of destination */
    long shrunk[VXL_MAX_RANK]; /* Local array of shrink factors */
    long lshrink;              /* Shrink factor along last dimension */
    int nbytes;                /* Length of an element, in bytes */
    long scoord[VXL_MAX_RANK]; /* Coordinates of a source row (scanline) */
    long dcoord[VXL_MAX_RANK]; /* Coordinates of a destination row */
    long *sdimen;              /* Dimensions of the source image */
    long ddimen[VXL_MAX_RANK]; /* Dimensions of the destination image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *ldelta;              /* Delta along last dimen, for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    long xmin, xmax;           /* Min and max valid column indices */
    long anel;                 /* Adjusted number of elements to process */
    double *asum;              /* Adjusted starting position in sum array */
    unsigned char *asd;        /* Adjusted starting position in source row */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;
    int i;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	fatal("Convolution is undefined for images of rank 0");
    last   = rank - 1;
    sdimen = src->dimen;
    ldsrc  = sdimen[last];
    type   = src->type;
    sd     = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	fatal("No kernel specified");
    else if (kern->rank != rank)
	fatal("Source and kernel must have the same rank");
    else if (kern->coef == 0)
	fatal("Convolution kernel has no coefficients");

    /* Default to no subsampling */
    if (shrink == 0) {
	shrink = shrunk;
        for (i = 0; i < rank; i++)
	    shrink[i] = 1;
    }
    lshrink = shrink[last];

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    for (i = 0; i < rank; i++)
        ddimen[i] = ceil((double)sdimen[i]/shrink[i]);
    lddst = ddimen[last];
    vxl_alloc_array(dest, type, rank, ddimen);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor; get delta along last
       dimension for each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    ldelta = mallock(nn*sizeof(long));
    for (in = 0; in < nn; in++) {
	offset[in] = nbytes*vxli_delta(src, delta[in]);
	ldelta[in] = delta[in][last];
    }

    /* Allocate one row of working storage for intermediate sums */
    sum = mallock(lddst*sizeof(double));

    /* Loop over all rows of the destination */
    vxli_firstrow(rank, dcoord);
    do {

	/* Locate current row in source and destination */
        for (i = 0; i < rank; i++)
	    scoord[i] = dcoord[i] * shrink[i];
	csd = sd + nbytes*vxli_offset2(rank, sdimen, scoord, 1);
	cdd = dd + nbytes*vxli_offset2(rank, ddimen, dcoord, 1);

	/* Clear the intermediate sums */
	err = bips_copy(lddst, INTP_DOUBLE, sum, 1, &kern->bias, 0);
	if (err != 0)
	    fatal("Error in calling BIPS function");

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip rows that are completely outside the source */
	    if ( vxli_rowout(rank, sdimen, scoord, delta[in]) )
		continue;

	    /* Adjust interation limits for source and destination limits */
	    xmin = ceil((double)(-ldelta[in])/lshrink);
	    if (xmin < 0)
		xmin = 0;
	    xmax = ceil((double)(ldsrc-ldelta[in])/lshrink);
            if (xmax > lddst)
		xmax = lddst;

	    /* Process into the destination row */
	    anel = xmax - xmin;
            asum = sum + xmin;
	    asd  = csd + offset[in] + xmin*lshrink*nbytes;
	    coef = ((double *)(kern->coef))[in];
	    err  = bips_conv1(anel, asum, 1, type, asd, lshrink, coef);
	    if (err != 0)
		fatal("Error in calling BIPS function");

	}

	/* Convert intermediate sums to give result */
	err = bips_conv2(lddst, type, cdd, 1, sum, 1);
	if (err != 0)
	    fatal("Error in calling BIPS function");
    }  while ( vxli_nextrow(rank, ddimen, dcoord) ) ;

    /* Free temporary storage */
    free(sum);
    free(ldelta);
    free(offset);

    return;
}


/* Old, non-shrinking version of vxl_convolve 
   FIXME: Use this as a guide to modify erode, dilate, and perhaps other 
   neighborhood operations to permit subsampling; then delete it. */
void
vxl_convolve_old(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd, *sd;    /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    double *sum;               /* One row of convolution sums */
    double *asum;              /* Address adjusted for missing neighbors */
    double coef;               /* Coefficient from the convolution kernel */
    int type;                  /* (Internal) type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a scanline (aka row) */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	fatal("Convolution is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	fatal("No kernel specified");
    else if (kern->rank != rank)
	fatal("Source and kernel must have the same rank");
    else if (kern->coef == 0)
	fatal("Convolution kernel has no coefficients");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, src->type, src->rank, src->dimen);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Allocate one row of working storage for intermediate sums */
    sum = mallock(nlast*sizeof(double));

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
	csd = sd + nbytes*vxli_offset2(rank, dimen, coord, 1);
	cdd = dd + nbytes*vxli_offset2(rank, dimen, coord, 1);

	/* Clear the intermediate sums */
	err = bips_copy(nlast, INTP_DOUBLE, sum, 1, &kern->bias, 0);
	if (err != 0)
	    fatal("Error in calling BIPS function");

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd  = csd + offset[in];
	    asum = sum;
	    anl  = nlast;
	    if (delta[in][last] < 0) {
		anl  += delta[in][last];
		asum -= delta[in][last];
		asd  -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    coef = ((double *)(kern->coef))[in];
	    err = bips_conv1(anl, asum, 1, type, asd, 1, coef);
	    if (err != 0)
		fatal("Error in calling BIPS function");

	}

	/* Convert intermediate sums to give result */
	err = bips_conv2(nlast, type, cdd, 1, sum, 1);
	if (err != 0)
	    fatal("Error in calling BIPS function");
    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(sum);
    free(offset);

    return;
}



/*--------------------------------------------------------------------------
vxl_dilate - Do morphological dilation on a image of any rank

This function performs a morphological dilation operation on an
unsigned volume image of any rank one or more; each voxel of the
output is the bitwise OR of the neighbors specified in the kernel.
The type of the input must be an unsigned integral type; the type of
the output is the same as the type of the input.  Incomplete
neighborhoods at the boundary of the image are handled correctly; that
is, voxels outside the image are simply omitted from the OR.  Any
bias and coefficients provided in the kernel are ignored; it is needed
just to define the neighborhood or structuring element of the dilation.

NOTE: The generic version of this function seems to be about 1% slower
than the old, specialized 3d version, even with inlining; that is
probably a reasonable price for most applications.  Hence the old
version has been deleted.

FIXME: Implement functions to create useful neighborhoods, and provide
Pyvox bindings to these functions.

FIXME: Add capability to subsample; see convolve for the algorithm.

FIXME: The dilate and erode operations are identical except for the
choice of operator from AND and OR and the choice of initial value for
the destination; figure out how to share the code.  Also consider
whether substituting other operations for AND and OR creates useful
operations.  In particular, min and max seem like they might be
useful.
---------------------------------------------------------------------------*/

void
vxl_dilate(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd,  *sd;   /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    unsigned char *add;        /* Current dest row adjusted for boundaries */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a row */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	fatal("Convolution is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	fatal("No kernel specified");
    else if (kern->rank != rank)
	fatal("Source and kernel must have the same rank");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
	csd = sd + nbytes*vxli_offset2(rank, dimen, coord, 1);
	cdd = dd + nbytes*vxli_offset2(rank, dimen, coord, 1);

	/* Clear the destination */
	memset(cdd, 0, nlast*nbytes);

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd = csd + offset[in];
	    add = cdd;
	    anl = nlast;
	    if (delta[in][last] < 0) {
		anl += delta[in][last];
		add -= delta[in][last] * nbytes;
		asd -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    err = bips_or_set(anl, type, add, 1, asd, 1);
	    if (err != 0)
		fatal("Error in calling BIPS function");

	}

    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(offset);

    return;
}



/*--------------------------------------------------------------------------
vxl_erode - Do morphological erosion on a image of any rank

This function performs a morphological erosion operation on an
unsigned volume image of any rank one or more; each voxel of the
output is the bitwise AND of the neighbors specified in the kernel.
The type of the input must be an unsigned integral type; the type of
the output is the same as the type of the input.  Incomplete
neighborhoods at the boundary of the image are handled correctly; that
is, voxels outside the image are simply omitted from the neighborhood.
Any bias and coefficients provided in the kernel are ignored; it is
needed just to define the neighborhood or structuring element.

NOTE: The generic version of this function seems to be about 1% slower
than the old, specialized 3d version, even with inlining; that is
probably a reasonable price for most applications.  The old 3d-only
version has therefore been deleted.

FIXME: Add capability to subsample; see convolve for the algorithm.

FIXME: Implement functions to create useful neighborhoods, and provide
Pyvox bindings to these functions.
---------------------------------------------------------------------------*/

void
vxl_erode(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd,  *sd;   /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    unsigned char *add;        /* Current dest row adjusted for boundaries */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a row */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	fatal("Convolution is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	fatal("No kernel specified");
    else if (kern->rank != rank)
	fatal("Source and kernel must have the same rank");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
	csd = sd + nbytes*vxli_offset2(rank, dimen, coord, 1);
	cdd = dd + nbytes*vxli_offset2(rank, dimen, coord, 1);

	/* Initialize the destination */
	/* FIXME:  Is -1 below truly portable? */
	memset(cdd, -1, nlast*nbytes);

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd = csd + offset[in];
	    add = cdd;
	    anl = nlast;
	    if (delta[in][last] < 0) {
		anl += delta[in][last];
		add -= delta[in][last] * nbytes;
		asd -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    err = bips_and_set(anl, type, add, 1, asd, 1);
	    if (err != 0)
		fatal("Error in calling BIPS function");

	}

    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(offset);

    return;
}



/*--------------------------------------------------------------------------
vxl_lostat - Compute local (neighborhood) statistics on a image of any rank

This function computes the local mean and standard deviation within
each neighborhood of an image of any rank and returns the results as
local mean and local standard deviation images of the same rank,
dimensions, and type as the original image.  The neighborhood is
defined by a convolution kernel; the kernel coefficients define the
weights assigned to each neighbor.  The rank must be one or greater;
while the local statistics are sort of defined on a rank zero image,
they are not really useful to compute and it simplifies the algorithm
not to have to treat this special case.  To avoid division by zero,
the neighborhood must include the central (0,0,...) voxel.
Intermediate calculations are done in type double and rounded back to
the final type; if a result voxel cannot be represented within range
of the result type, the result is set to the appropriate limit of the
range.  The standard deviation image is multiplied by the given scale
factor before conversion; this feature is probably useful only for
images of integral type, to match the expected range of values to the
range supported by the type.  Incomplete neighborhoods at the boundary
of the image are handled correctly; that is, the missing voxels are
simply omitted from the neighborhood, so that the mean and standard
deviation are computed correctly for the incomplete neighborhood; note
that the standard error of the mean will be increased accordingly for
neighborhoods on the boundary of the image.

NOTE: This function (with a suitable kernel) replaces the now-omitted
function vxl_lovar_uchar3d.  The newer function will handle images of
any rank, dimensions, or type, and permits the user to specify the
neighborhood; it also handles incomplete neighborhoods at the boundary
of the image correctly.  If you're converting old code, note that
vxl_lostat rounds the standard deviation where vxl_lovar_uchar3d
truncates, and that the scale factor 2.0 is probably better than the
scale factor (255.0 / 127.4125) used in vxl_lovar_uchar3d.  These
differences mean that the output values will be slightly different
between the two algorithms.

FIXME: Possibly provide a way of returning only the mean or variance
image?  There is probably little or no saving of time, but the space
saving might be significant in some applications that need only one
of the return values.

FIXME: Is there an application for the local skew or higher moments?

FIXME: Add capability to subsample; see convolve for the algorithm.
---------------------------------------------------------------------------*/

void
vxl_lostat(
    voxel_array *mean,         /* Local mean image (output) */
    voxel_array *sdev,         /* Standard deviation image (output) */
    voxel_array *src,          /* Image to take statistics of (input) */
    struct vxl_kernel *kern,   /* The convolution kernel to use */
    double scale)              /* Multiplier for the std dev image */
{
    double *sum, *sumx, *sumxx;      /* Sums of weights, values, squares */
    double *asum, *asumx, *asumxx;   /* Same, adjusted for missing neighbors */
    unsigned char *md, *dd, *sd;     /* Actual data for mean, dev, and src */
    unsigned char *cmd, *cdd, *csd;  /* Current row in each image */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    double coef;               /* Coefficient from the kernel */
    double zero = 0.0;         /* Zero constant for clearing arrays */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a scanline (aka row) */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int count;                 /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err, err1, err2, err3;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	fatal("Operation is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	fatal("No kernel specified");
    else if (kern->rank != rank)
	fatal("Source and kernel must have the same rank");
    else if (kern->coef == 0)
	fatal("Kernel has no coefficients");

    /* Check validity of destinations; allocate memory */
    if (mean == 0 || mean->magic != VXL_MAGIC)
	fatal("Invalid destination array for mean");
    vxl_alloc_array(mean, src->type, src->rank, src->dimen);
    md = mean->data;
    if (sdev == 0 || sdev->magic != VXL_MAGIC)
	fatal("Invalid destination array for sdev");
    vxl_alloc_array(sdev, src->type, src->rank, src->dimen);
    dd = sdev->data;

    /* Compute offset in bytes to each neighbor */
    count = kern->count;
    delta = kern->delta;
    offset = mallock(count*sizeof(long));
    for (in = 0; in < count; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Allocate one row of working storage for intermediate sums */
    sum   = mallock(nlast*sizeof(double));
    sumx  = mallock(nlast*sizeof(double));
    sumxx = mallock(nlast*sizeof(double));

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination arrays */
	csd = sd + nbytes*vxli_offset2(rank, dimen, coord, 1);
	cmd = md + nbytes*vxli_offset2(rank, dimen, coord, 1);
	cdd = dd + nbytes*vxli_offset2(rank, dimen, coord, 1);

	/* Clear the intermediate sums */
	err1 = bips_copy(nlast, INTP_DOUBLE, sum,   1, &zero, 0);
	err2 = bips_copy(nlast, INTP_DOUBLE, sumx,  1, &zero, 0);
	err3 = bips_copy(nlast, INTP_DOUBLE, sumxx, 1, &zero, 0);
	if (err1 != 0 || err2 != 0 || err3 != 0)
	    fatal("Error in calling BIPS function");

	/* Loop over the neighbors */
	for (in = 0; in < count; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd    = csd + offset[in];
	    asum   = sum;
	    asumx  = sumx;
            asumxx = sumxx;
	    anl    = nlast;
	    if (delta[in][last] < 0) {
		anl    += delta[in][last];
		asum   -= delta[in][last];
		asumx  -= delta[in][last];
		asumxx -= delta[in][last];
		asd    -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    coef = kern->coef[in];
	    err = bips_lostat1(anl, asum, asumx, asumxx, type, asd, 1, coef);
	    if (err != 0)
		fatal("Error in calling BIPS function");

	}

	/* Convert intermediate sums to give result */
	err = bips_lostat2(nlast, type, cmd, cdd, 1, sum, sumx, sumxx, scale);
	if (err != 0)
	    fatal("Error in calling BIPS function");
    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(sumxx);
    free(sumx);
    free(sum);
    free(offset);

    return;
}



/*-------------------------------------------------------------------------
vxl_zerbv_uchar3d - Zero boundary voxels in 3d image

Given a 3d unsigned char volume image and a variance threshold value
(expressed as a standard deviation), this function computes the local
variance using a 3x3x3 neighborhood and copies the original image to
another image, zeroing any voxels for which the local variance exceeds
the threshold.  The intent is that the output image should contain
only voxels in the interior of regions with approximately uniform
tone, and that voxels on the boundary between regions should be
zeroed; this should make some histogram-based segmentation algorithms
more robust by eliminating partial volume effects from the analysis.
Voxels at the edges of the image are boundary voxels by definition, to
avoid the problems of having an incomplete neighborhood.  The
threshold value is given as the standard deviation, or the square root
of the threshold variance, and is a compromise between the intensity
variance of uniform regions and the difference in gray level between
uniform regions.  As a rough rule of thumb, take the threshold to be 2
or 3*sigma, where sigma is the standard deviation of the intensity
variation within a uniform region; if this value is comparable to or
larger than the intensity differences between regions, this algorithm
probably won't produce any useful results.  The minimum and maximum
possible values of the standard deviation are approximately 0.19 and
127.5 for an unsigned char image.

FIXME: This function might be obsolete, now that vxl_lostat works.
Consider rewriting for arbitrary rank, type and kernel, following the
methods used for convolution; also consider using the vxl_lostat
function to do most of the real work.
-------------------------------------------------------------------------*/

void
vxl_zerbv_uchar3d(
    voxel_array *dest,         /* Output volume image */
    voxel_array *src,          /* Input volume image */
    double stddev)             /* Threshold value, as standard deviation */
{
    unsigned char *dd, *sd;    /* Actual data array for dest and source */
    double mean, var;          /* Local mean and variance */
    double varth;              /* Variance theshold */
    long nz, ny, nx;           /* Number of slices, rows, columns */
    long iz, iy, ix;           /* Indices over slices, rows, columns */
    long dz, dy, dx;           /* Stride in slices, rows, columns */
    long cv;                   /* Index to current voxel */
    unsigned char *csv;        /* Pointer to current source voxel */
    double vox;                /* Value of current voxel */
    long dn[27], *pdn;         /* Offsets to neighbors of current voxel */
    int in;                    /* Index over neighbors */

    /* Check validity of threshold. */
    if (stddev < 0 || stddev > 127.501)
	fatal("vxl_zerbv_uchar3d: Invalid variance threshold");
    varth = stddev * stddev;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 || 
	src->type != INTP_UCHAR || src->rank != 3)
	fatal("vxl_zerbv_uchar3d: Invalid or empty source array");
    sd = src->data;
    nz = src->dimen[0];
    ny = src->dimen[1];
    nx = src->dimen[2];

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("vxl_zerbv_uchar3d: Invalid destination array");
    vxl_alloc_array(dest, INTP_UCHAR, 3, &(src->dimen[0]));
    dd = dest->data;

    /* Compute step sizes in each dimension */
    dz = vxli_step(src, 0);
    dy = vxli_step(src, 1);
    dx = vxli_step(src, 2);

    /* Compute offsets to neighbors */
    pdn = dn;
    for (iz = 0; iz < 3; iz++)
	for (iy = 0; iy < 3; iy++)
	    for (ix = 0; ix < 3; ix++)
		*pdn++ = (iz-1)*dz + (iy-1)*dy + (ix-1)*dx;

    /* Zero first slice of destination */
    memset(dd, 0, nx*ny);
   
    /* Loop over intermediate slices */
    for (iz = 1; iz < nz-1; iz++) {

	/* Zero first row of destination */
	memset(dd+iz*dz, 0, nx);

	/* Loop over intermediate rows of destination */
	for (iy = 1; iy < ny-1; iy++) {
	    cv = iz*dz + iy*dy;

	    /* Zero first column in current row */
	    dd[cv++] = 0;

	    /* Loop over intermediate columns of destination */
	    for (ix = 1; ix < nx-1; ix++) {
		csv = sd + cv;

		/* Compute mean and variance of neighborhood */
		mean = var = 0.0;
		for (in = 0; in < 27; in++) {
		    vox = (double)csv[dn[in]];
		    mean += vox;
		    var  += vox*vox; }
		mean = mean/27.0;
		var  = var/27.0 - mean*mean;

		/* Copy or zero dest voxel according to variance */
		dd[cv++] = (var < varth) ? *csv : 0;
	    }

	    /* Zero last column in current row */
	    dd[cv] = 0;
	}

	/* Zero last row of destination */
	memset(dd+iz*dz+(ny-1)*dy, 0, nx);
    }

    /* Zero last slice of destination */
    memset(dd+(nz-1)*dz, 0, ny*nx);

    return;
}
