/**************************************************************************
param.c - Parameter file/list management functions

Author:  Paul Hughett

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


The functions in this source file provide some facilties to manage the
many parameters that enter into an automated image analysis protocol.
Simply putting all the parameters for each program into its command
line is not a good solution, since the command line becomes very long
and error-prone.  Instead, we take advantage of the fact that most of
the parameters are fixed for a given analysis protocol or class of
images; thus we can put these parameters in a parameter file specific
to the protocol, and put only those that pertain to a specific image
(for example, the name of the image file) into the command line.  Most
protocols will require the combined efforts of several programs, so we
choose to specify all their parameters in a single file, rather than
having a separate file for each program; this facilitates documenting
the protocol and keeping it up to date as needed.

A typical line in a parameter file looks like this:

stage.param = value     # comment

where stage is the name of a processing stage (which often defaults to
the name of the program that executes it), param is the name of the
particular parameter to specify for that program, value is the value
for that parameter, and comment is an optional comment.  The stage
name is optional, in which case the separating period should be
omitted.  The value actually passed to the program will be the
characters following the equal sign up to the number sign or end of
line, with leading and trailing blanks removed; embedded blanks will
be untouched.  A line containing only whitespace or whitespace
preceding a number sign will be ignored.

A program may be used in more than one step of an analysis protocol,
in which case different stage names must be used to distinguish their
individual parameter values.

FIXME:  It might be useful to add functions that add or modify an entry
in the param_list, or to write it out as a parameter.  This could be
a convenient way to keep track of various intermediate results that
don't really need a whole file to themselves.  But doing this would
probably also require nested parameter lists.
***************************************************************************/

static const char rcsid[] = "$Id: param.c,v 1.8 2000/01/24 21:30:41 hughett Exp $ Release $Name:  $";

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
#include <unistd.h>

#include "errm.h"
#include "param.h"


#define MAXCHAR  256          /* Max chars per input line */




/*-------------------------------------------------------------------------
read_parameter_file - Read and preprocess a bblimage parameter file

This function reads a parameter file for bblimage in the format described
above, preprocesses it for later searching, and returns a pointer to a
data structure with the file contents.  This pointer can be given to
the function get_param later, to find the value of a given parameter.
If the parameter file cannot be opened or read successfully, this function
produces a warning message and returns a NULL pointer.

TECHNICAL NOTE: The parameters read are currently stored as follows: A
series of null-terminated strings, followed by an additional NUL
character to end the series.  Each string has the form
"stage.param=value" with all non-significant spaces removed.  This
format is subject to change and may be changed to something completely
different in future releases; you depend on this particular structure
at your peril.
-------------------------------------------------------------------------*/


Param_list
read_param_file(
    const char *filename)       /* Name of the file to read */
{
    char *parms;                /* Contents of file, for later search */
    FILE *file;                 /* File to read from */
    struct stat buf;            /* File status */
    char line[MAXCHAR];         /* Input buffer */
    int ln;                     /* Line number in parameter file */
    int flag;                   /* Return status from called function */
    char *p, *q;                /* Next char in parms and line */
    char *ps;                   /* Saved p for error backtracking */
    char *plim;                 /* End of the allocated param_list */

    /* Allocate space for parameters structure */
    /* FIXME:  This is too generous because space is allocated for
       both payload and comments. */
    flag = stat(filename, &buf);
    if (flag != 0) {
	warning("Unable to stat parameter file %s", filename);
	return (void*)0; }
    p = parms = malloc(buf.st_size+3);  /* FIXME: Is 3 right??? */
    plim = p + buf.st_size+3;
    if (parms == 0) {
	warning("Unable to allocate space for parameters");
	return (void*)0; }

    /* Open the parameter file */
    file = fopen(filename, "r");
    if (file == 0)  {
	warning("Unable to open parameter file %s", filename);
	return (void *)0; }

    /* Read parameter file line by line */
    ln = 0;
    while (!feof(file)) {

	/* Read the next line */
	fgets(line, MAXCHAR, file);
	ln++;

	/* Remove any trailing white space and comment */
	q = line;
	while (*q != 0 && *q != '\n' && *q != '#') q++;
	if (q == line) continue;
	for (q--; q > line && isspace((int)*q); q--)  ;
	if (q == line) continue;
	*(q+1) = 0;

	/* Go to beginning of line; rmbr location in parms */
	q = line;
	ps = p;

	/* Skip over leading white space */
	while (isspace((int)*q))  q++;
	if ( !isalnum((int)*q) && *q != '.' ) {
	    warning("Invalid parameter name at line %d of parameter file", ln);
	    continue; }
	
	/* Copy the program and parameter names */
	while (isalnum((int)*q) || *q == '.')  *p++ = *q++;

	/* There should be an equals sign next, possibly with white space */
	while (isspace((int)*q)) q++;
	if (*q++ != '=') {
	    warning("Missing = at line %d of parameter file", ln);
	    p = ps;
	    continue; }
	*p++ = '=';
	while (isspace((int)*q)) q++;

	/* Copy the parameter value, followed by a NUL */
	while (*q != 0)  *p++ = *q++;
	*p++ = 0;
	*p = 0;   /* Add the final NUL, but overwrite with next param */

	/* Check for overrun */
	if (p >= plim)
	    fatal("read_param_file: Overrun of param_list allocated.");

    }
    
    /* Close the file and return parameter structure */
    fclose(file);

    return (const void *)parms;
}



/*-------------------------------------------------------------------------
dump_param_list - Dump a parameters list, for debugging

Given a param list, this function writes out its contents on a given
device in human-readable form.  There is one line for each entry in
the table, and the whole thing is followed by a newline.  The only
likely use for this function is in debugging.
-------------------------------------------------------------------------*/

void
dump_param_list(
    Param_list parms,       /* Param list to be dumped */
    FILE *out)              /* Stream to dump it onto */
{
    char *p;                /* Current location in params list */

    /* Allow caller to provide a null pointer */
    if (parms == 0) {
	fprintf(out, "(null)\n");
	return; }

    /* Print out each entry in the list */
    for ( p = (char *)parms; *p != 0; p += strlen(p)+1 )
	fprintf(out, "%s\n", p);

    return;
}



/*--------------------------------------------------------------------------
get_param - Get value of a parameter from a params structure

This function concatenates the given program (or stage) name and the
variable name, separated by a period, and searches the params list for
the name thus obtained.  If either the stage name or the variable name
is empty, the period is omitted.  If that name is found, its value is
returned as a string (possibly the null string); otherwise NULL is
returned.  If either the stage name or the parameter name is NULL or an
empty string, the period is omitted.
---------------------------------------------------------------------------*/

const char *
get_param(
    Param_list parms,       /* Parameters structure */
    const char *stage,      /* Name of program or stage */
    const char *param)      /* Name of variable to get */
{
    char buf[MAXCHAR];      /* String buffer */
    char empty[] = "";      /* An empty string */
    int len;                /* Length of combined names */
    char *p;

    /* If params list is empty, it doesn't contain any params */
    if (parms == 0)  return NULL;

    /* Concatentate the stage and param names, if any */
    if (stage == 0)  stage = empty;
    if (param == 0)  param = empty;
    len = strlen(stage) + strlen(param) + 1;
    if (len >= MAXCHAR)
	fatal("Parameter name exceeds buffer length");
    strcpy(buf, stage);
    if (*stage != 0  && *param != 0)
	strcat(buf, ".");
    strcat(buf, param);
    len = strlen(buf);  /* Recompute in case period was omitted */

    /* Search the params list for that stage/param combination;
       if found, return the string value */
    for ( p = (char *)parms; *p != 0; p += strlen(p) + 1) {
	if (strncmp(p, buf, len) == 0 && p[len] == '=')
	    return p+len+1;
    }

    /* If no such label, return NULL */
    return NULL;
}





/*-------------------------------------------------------------------------
Quick and dirty test program for the params list functions
-------------------------------------------------------------------------*/

#if 0

int
main(int argc, char *argv[])
{
    Param_list parms;
    const char *s;

    /* Read and then dump a handy params file */
    parms = read_param_file("foo.par");
    dump_param_list(parms, stdout);

    /* Look up an existing and an absent param */
    s = get_param(parms, "a", "out");
    printf("a.out is %s\n", s);
    s = get_param(parms, "c", "parm");
    printf("c.parm is %s\n", s);
    s = get_param(parms, "end.of.line", 0);
    printf("end.of.line is %s\n", s);
    s = get_param(parms, "", "parm");
    printf("parm is %s\n", s);
    s = get_param(parms, "foo", "bar");
    printf("foo.bar is %s\n", s);

    return 0;
}

#endif

