/****************************************************************************
parray.c - Pyvox array type for Python

Author:  Paul Hughett

Copyright (c) 1999-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains code to implement the Pyvox array type (often
abbreviated as parray) for the Pyvox module, which provides a Python
interface to the Voxel Kit.

Most functions in this file have the prefix "parray_" to indicate
that they define the Pyvox array type.  The obsolete prefix "voxel_" 
means the same thing, but not all instances have been converted yet.

Note that this code is still rather experimental and subject to
change.  YOU HAVE BEEN WARNED!!!


FIXME: Create versions of const, array, read, and ramp (and add fill
from list) that store data into existing Pyvox arrays without changing
their type or shape.  Provide a means for creating a Pyvox array
without initializing its contents; this would be useful for
efficiently using the functions just aforesaid.  One possible method
would be to supply None as the data value, and warn the user that if
he tries to access the array without filling it, he might get
anything.  Even better, perhaps, is that pyvox.array() should create
but not allocate and initialize the array (thus varray->data==0
indicates an uninitialized array); then any array operation that fills
the entire array can just do so; but any array operation that affects
only individual pixels must set the entire array to zero first.
(array.eval() should do the fill) Actually, there are only a few
operations that are sensible for an uninitialized array; perhaps the
others should just generate an error.

FIXME:  The ordering of declarations and definitions in this file seems
to be rather delicate and in some places counter-intuitive.  Look for
a better way to do things.

FIXME: It would be really nice if parrays were sufficiently like sequences 
that you could simply iterate over their elements, rather than needing ugly 
artifices like

    for i in range(len(foo)) :
        do something with foo[i]

or

    for x in foo.list() :
        do something with x

Unfortunately, Python (1.5.2) doesn't seem to permit that; it insists on
using the mapping methods rather than the sequence methods when they exist
(with some awkward exceptions for which we have to provide sequence methods).

FIXME: PyObjects may also be pyvox_arrays or kernel_objects in
disguise; the current practice is to pass arguments as PyObjects and
cast as needed inside the function but this is sometimes confusing and
eliminates some error checking.  Is there a better way?

FIXME: The Voxel Kit and its subsidiaries call fatal(), etc when they
see an error; using the current version of errm.c, this causes an
immediate abort (and possibly a core dump).  Improve the error manager
to support the conversion of errors into Python exceptions that can be
caught and reasonably handled.  It's probably useful to permit the
user to choose between immediate abort (for debugging Pyvox) and
a Python error message (for using Pyvox).
*****************************************************************************/

static const char rcsid[] = "$Id: parray.c,v 1.42 2002/02/28 20:57:36 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <stdio.h>

#include "Python.h"

#include "config.h"
#include "clap.h"
#include "exim.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"       /* For vxli_count */
#include "pyvox.h"      /* Must follow voxel.h */

#define DEBUG_ALLOC  0     /* Trace array allocation/freeing? */


/* Forward declarations to facilitate logical organization of this file;
   the staticforward keyword is used to avoid certain compiler bugs--see
   the definition of staticforward in (python1.5) object.h for details. */
staticforward struct PyMethodDef  parray_methods[];



/*-----------------------------------------------------------------------
Common error strings

FIXME:  Gather the common error strings into this code block rather than
defining them individually for each function.  Better yet, create nice
Python exceptions for each common error.
------------------------------------------------------------------------*/

static const char cent_error[] =   /* Error: Invalid centroid argument */
	"Initial centroids must be a sequence of values in 0..255";
static const char clid_error[] =   /* Error: Invalid class ids argument */
	"Class ids must be a sequence of values in 0..255";
static const char match_error[] =  /* Error: Clids and centroids don't match */
	"Different numbers of class ids and centroids";
static const char uchar_error[] =  /* Error: Only uchar implemented */
	"Sorry, only unsigned char has been implemented yet.";
static const char voxel_error[] =  /* Error: not a Pyvox array */
	"Input is not a Pyvox array";



/*************************************************************************
             Utility Functions for Pyvox Arrays

FIXME: These should probably be renamed to the standard parray_ prefix;
the attempt to use Python style names seems to have been misguided.
(Misguidoed?)
*************************************************************************/



/*------------------------------------------------------------------------
PyvoxArray_Check - Test whether a Python object is a voxel array

If the given object is a voxel array (varray), this function returns 1
and stores a pointer to the voxel_array struct in the second argument;
otherwise it returns 0.
-------------------------------------------------------------------------*/

int
PyvoxArray_Check(
    PyObject *ob,
    voxel_array **vob)
{
    if (ob != 0 && ob->ob_type == &parray_type) {
	if (vob != 0)
	    *vob = ((pyvox_array *)ob)->varray;
	return 1;
    }
    else
	return 0;
}



/*-------------------------------------------------------------------------
PyVoxel_GetArray - Verify and get voxel array from a Python object

This function verifies that a given PyObject is in fact a voxel array
object and returns a pointer to the voxel array it refers to.  If the
PyObject is not a voxel array object, zero is returned.

FIXME: Is it really necessary to have both this and PyvoxArray_Check?
---------------------------------------------------------------------------*/

voxel_array *
PyVoxel_GetArray(
    PyObject *ob)           /* Alleged voxel array object */
{
    if (ob != 0 && ob->ob_type == &parray_type)
	return ((pyvox_array *)ob)->varray;
    else
	return 0;
}



/*-------------------------------------------------------------------------
PyEmptyObject_Check - Test if a Python object is an empty object

This function returns TRUE if the object is a list, tuple, or voxel
array containing zero elements.  The empty object is treated specially
in some contexts, often selecting a default value.

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
--------------------------------------------------------------------------*/

#if 0
static int
PyEmptyObject_Check(
    PyObject *ob)
{
    if ( ob == 0 )
        return 0;
    else if ( PyTuple_Check(ob) && PyTuple_Size(ob) == 0 )
        return 1;
    else if ( PyList_Check(ob) && PyList_Size(ob) == 0 )
        return 1;
    else if ( PyvoxArray_Check(ob, 0) && PyVoxel_Count(ob) == 0 )
        return 1;
    else 
        return 0;
}
#endif



/*-------------------------------------------------------------------------
PyScalar_Check - Test whether a Python object is a scalar

This function returns TRUE if the object is either a voxel array object
of rank zero (that is, a scalar) or a number (which can be converted into
a scalar).

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
-------------------------------------------------------------------------*/

#if 0
static int
PyScalar_Check(
    PyObject *ob)
{
    voxel_array *var;

    if ( ob == 0 )
        return 0;
    else if ( ob->ob_type == &parray_type &&
              (var = ((pyvox_array *)ob)->varray) != 0 &&
              var->rank == 0 )
	return 1;
    else if ( PyNumber_Check(ob) )
	return 1;
    else
	return 0;
}
#endif



/*-------------------------------------------------------------------------
PyNonScalarArray_Check - Test if a Python object is a non-scalar voxel array

This function returns TRUE if the object is a voxel array object
of rank strictly greater than zero.

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
-------------------------------------------------------------------------*/

#if 0
static int
PyNonScalarArray_Check(
    PyObject *ob)
{
    voxel_array *var;

    if ( ob != 0
	 && ob->ob_type == &parray_type
	 && (var = ((pyvox_array *)ob)->varray) != 0
	 && var->rank > 0 )
	return 1;
    else
	return 0;
}
#endif



/*--------------------------------------------------------------------------
PyVoxel_Type - Return internal type code of a voxel array

Returns zero if the Pyvox array is invalid or empty.
---------------------------------------------------------------------------*/

static int
PyVoxel_Type(
    PyObject *ob)
{
    voxel_array *var;

    if ( ob != 0
	 && ob->ob_type == &parray_type
	 && (var = ((pyvox_array *)ob)->varray) != 0 )
	return var->type;
    else
	return 0;
}



/*--------------------------------------------------------------------------
PyScalar_FromNumber - Create a scalar voxel array from a Python number
---------------------------------------------------------------------------*/

static PyObject *
PyScalar_FromNumber(
    PyObject *number,          /* The number to convert */
    int intype)                /* Internal type code for the array */
{
    pyvox_array *ob;          /* Scalar being created */
    double val;

    if ( !PyvoxDouble_Check(number, &val) )
	ReturnError("Cannot create scalar from a non-number");
    ob  = parray_create();
    vxl_store_scalar(ob->varray, intype, val);
    return (PyObject *)ob;
}



/*-------------------------------------------------------------------------
PyNumSeq_AsLong - Convert a numeric sequence into a C array of longs

This function takes a Python tuple or list containing numbers and
converts it into a C array of long ints.  The length of the sequence
must exactly match the specified length of the array.  As a special
case, a naked number will be treated as a singleton sequence.  The
address of the array is returned, unless the sequence cannot be
correctly converted; in that case NULL is returned.  It is the user's
responsibility to generate any necessary error message.

FIXME:  Empty sequences are permitted for now, but this might change in
the future.

FIXME: Also accept Pyvox arrays of rank 1.

FIXME: Rename as PyvoxLongSeq_Check to use to both check and convert,
a la PyvoxLong_Check?
-------------------------------------------------------------------------*/

static long *
PyNumSeq_AsLong(
    PyObject *ob,               /* The object to convert */
    long *values,               /* Where to put values extracted from object */
    int  n)                     /* Number of values to extract */
{
    int      nelem;             /* Number of elements in the sequence */
    PyObject *elem;             /* An element of the sequence */
    int i;

    if ( PySequence_Check(ob) ) {
        nelem = PyObject_Length(ob);
        if (nelem != n)
            return 0;
        for (i = 0; i < nelem; i++) {
            elem = PySequence_GetItem(ob, i);
	    if ( ! PyvoxLong_Check(elem, &values[i]) )
                return 0;
        }
        return values;
    }
    else if ( n == 1 && PyvoxLong_Check(ob, &values[0]) ) {
        return values;
    }
    else
        return 0;
}



/*----------------------------------------------------------------------
PyvoxDoubleSeq_Check - Check for double sequence and convert

This utility function checks whether or not a Python object is a list,
tuple, or Pyvox array of some specified length which contains only
numbers.  If it is, it is converted to a C array of doubles and the
function returns 1; otherwise the function returns 0.  As a special
case, a naked number will be treated as a singleton sequence.  It is
the user's responsibility to generate any necessary error message.

FIXME: Generalize to also accept a Pyvox array, unless a Pyvox array passes
PySequence_Check, in which case it already does.
-----------------------------------------------------------------------*/

static int
PyvoxDoubleSeq_Check(
    PyObject *ob,            /* The object to convert */
    double *values,          /* Where to put values extracted from object */
    int  n)                  /* Number of values to extract */
{
    int      nelem;          /* Number of elements in the sequence */
    PyObject *elem;          /* An element of the sequence */
    int i;

    if ( PySequence_Check(ob) ) {
        nelem = PyObject_Length(ob);
        if (nelem != n)
            return 0;
        for (i = 0; i < nelem; i++) {
            elem = PySequence_GetItem(ob, i);
	    if ( ! PyvoxDouble_Check(elem, &values[i]) )
                return 0;
        }
        return 1;
    }
    else if ( n == 1 && PyvoxDouble_Check(ob, &values[0]) ) {
        return 1;
    }
    else
        return 0;
}



/*------------------------------------------------------------------------
Pyvox_CheckAxes - Interpret axes argument for reduction operator

This function checks that the axes argument for a Pyvox reduction
operator is in one of the acceptable forms and converts it to the
(naxes, axes) form expected by the Voxel Kit.  The acceptable forms
are an integer in the range [0, rank-1] or [-1, -rank], meaning the
single axis specified, counting either from the left or the right; a
tuple or list of such integers; or None, meaning all axes.  This function 
returns non-zero if the argument is valid, and zero otherwise.

ALGORITHM: The interpretation of the axis arguments themselves is delegated
to the vxli_reduce_setup() function, which will be called by the reduction
function.  FIXME: The check is duplicated here, so we get a Python exception 
rather than a fatal error in C.  
---------------------------------------------------------------------------*/

static int
Pyvox_CheckAxes(
    PyObject *ob,               /* The purported Pyvox axes argument */
    int rank,                   /* The rank of the array being reduced */
    int *naxes,                 /* (Out) The number of axes to reduce on */
    int *axes)                  /* (Out) The axes to reduce on */
{
    int nax;                    /* Number of elements in a tuple or list */
    PyObject *ax;               /* An axis, as a Python object */
    int n;                      /* An axis, as a C number */
    int i;

    if ( ob == 0 || ob == Py_None ) {
	*naxes = -1;     /* Tells vxl_reduce_setup() to do all axes */
	return 1;
    }
    else if ( PyInt_Check(ob) ) {
	n = PyInt_AsLong(ob);
	if (n >= -rank && n < rank) {
	    *naxes = 1;
	    axes[0] = n;
	    return 1;
	}
	else
	    return 0;

    }
    else if ( PyList_Check(ob) ) {
	*naxes = nax = PyList_Size(ob);
	for (i = 0; i < nax; i++) {
	    ax = PyList_GetItem(ob, i);
	    if ( !PyInt_Check(ax) )
		return 0;
	    n = PyInt_AsLong(ax);
	    if ( n >= -rank && n < rank)
		axes[i] = n;
	    else
		return 0;
	}
	return 1;
    }
    else if ( PyTuple_Check(ob) ) {
	*naxes = nax = PyTuple_Size(ob);
	for (i = 0; i < nax; i++) {
	    ax = PyTuple_GetItem(ob, i);
	    if ( !PyInt_Check(ax) )
		return 0;
	    n = PyInt_AsLong(ax);
	    if ( n >= -rank && n < rank)
		axes[i] = n;
	    else
		return 0;
	}
	return 1;
    }
    else
	return 0;
}




/************************************************************************
Attributes of the Pyvox Array Type

The functions must precede the definition of the parray_getattr and
parray_setattr functions.
*************************************************************************/



/*------------------------------------------------------------------------
parray_refcnt - Report reference count of a Pyvox array

This instance method returns the current reference count of the instance;
it is probably useful only for debugging.

FIXME:  Redefine to use with other types of objects?
-------------------------------------------------------------------------*/

static PyObject *
parray_refcnt(
    PyObject *self)
{
    return PyInt_FromLong(self->ob_refcnt);
}

	

/*-----------------------------------------------------------------------
parray_origin - Coordinate origin of a Pyvox array

This function returns the physical coordinate origin of a voxel array
object as a list of Python floats.

FIXME:  There is probably some code here that could be shared with
parray_spacing or other methods.
-----------------------------------------------------------------------*/

static PyObject *
parray_origin(
    PyObject *self)           /* Pyvox array to find the origin of */
{
    PyObject *origin;         /* List containing the dimensions */
    PyObject *dim;            /* One of the dimensions */
    voxel_array *varray;      /* The voxel array itself */
    int rank, i;

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, &varray) )
	fatal("This isn't a Pyvox array!");
    rank = varray->rank;

    /* Construct a list of the origin values */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the list containing them, but
       the Python API manual is somewhat obscure on this point. */
    origin = PyList_New(rank);
    for (i = 0; i < rank; i++) {
        dim = PyFloat_FromDouble(varray->origin[i]);
	/* FIXME:  The Python C API manual does not specify the value
	   returned by PyList_SetItem; I am assuming that it returns
	   zero on success, and non-zero otherwise. */
	if ( PyList_SetItem(origin, i, dim) )
	    fprintf(stderr, "PyList_SetItem is unhappy\n");
    }

    return origin;
}



/*----------------------------------------------------------------------
parray_set_origin - Set origin of a Pyvox array

This function sets the origin attribute of a voxel array.
-----------------------------------------------------------------------*/

static int
parray_set_origin(
    PyObject *self,           /* Object to set origin of */
    PyObject *value)          /* Value to set */
{
    voxel_array *vself;       /* The voxel array itself */
    PyObject *pov;            /* One of the origin values */
    double v[VXL_MAX_RANK];   /* New origin values */
    int rank, i;

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, &vself) )
	fatal("This isn't a Pyvox array!");
    rank  = vself->rank;

    /* A plain number is acceptable for a rank-1 array */
    if ( rank == 1 && PyvoxDouble_Check(value, &v[0] )) {
	vself->origin[0] = v[0];
    }

    /* Otherwise a list of numbers is required */
    else if ( PySequence_Check(value) && rank == PyObject_Length(value) ) {
	for (i = 0; i < rank; i++) {
	    pov = PySequence_GetItem(value, i);
	    if ( !PyvoxDouble_Check(pov, &v[i]) )
		XReturnError("New origin values must be numbers");
	}
	for (i = 0; i < rank; i++)
	    vself->origin[i] = v[i];
    }

    /* Anything else is an error */
    else {
	XReturnError("New origin must be a list of numbers");
    }
    
    return 0;
}



/*----------------------------------------------------------------------
parray_set_spacing - Set spacing of a Pyvox array

This function sets the spacing attribute of a voxel array.
-----------------------------------------------------------------------*/

static int
parray_set_spacing(
    PyObject *self,           /* Object to set spacing of */
    PyObject *value)          /* Value to set */
{
    voxel_array *vself;       /* The voxel array itself */
    PyObject *pov;            /* One of the spacing values */
    double v[VXL_MAX_RANK];   /* New spacing values */
    int rank, i;

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, &vself) )
	fatal("This isn't a Pyvox array!");
    rank  = vself->rank;

    /* A plain number is acceptable for a rank-1 array */
    if ( rank == 1 && PyvoxDouble_Check(value, &v[0] )) {
	vself->spacing[0] = v[0];
    }

    /* Otherwise a list of numbers is required */
    else if ( PySequence_Check(value) && rank == PyObject_Length(value) ) {
	for (i = 0; i < rank; i++) {
	    pov = PySequence_GetItem(value, i);
	    if ( !PyvoxDouble_Check(pov, &v[i]) )
		XReturnError("New spacing values must be numbers");
	}
	for (i = 0; i < rank; i++)
	    vself->spacing[i] = v[i];
    }

    /* Anything else is an error */
    else {
	XReturnError("New spacing must be a list of numbers");
    }
    
    return 0;
}

	

/*-----------------------------------------------------------------------
parray_size - Dimensions of a Pyvox array

This function returns the dimensions of a Pyvox array as a list of
integers and is used to implement the instance.size member of a voxel
object.  The rank is obviously the length of this list.
-----------------------------------------------------------------------*/

static PyObject *
parray_size(
    PyObject *self)           /* Pyvox array to find the size of */
{
    PyObject *size;           /* List containing the dimensions */
    PyObject *dim;            /* One of the dimensions */
    voxel_array *varray;      /* The voxel array itself */
    int rank, i;

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, &varray) )
	fatal("This isn't a Pyvox array!");
    rank = varray->rank;

    /* Construct a list of the dimensions */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the list containing them, but
       the Python API manual is somewhat obscure on this point. */
    size = PyList_New(rank);
    for (i = 0; i < rank; i++) {
        dim = PyInt_FromLong(varray->dimen[i]);
	/* FIXME:  The Python C API manual does not specify the value
	   returned by PyList_SetItem; I am assuming that it returns
	   zero on success, and non-zero otherwise. */
	if ( PyList_SetItem(size, i, dim) )
	    fprintf(stderr, "PyList_SetItem is unhappy\n");
    }

    return size;
}

	

/*-----------------------------------------------------------------------
parray_spacing - Coordinate spacings of a Pyvox array

This function returns the physical coordinate spacing of a voxel array
object as a list of Python floats.

FIXME:  There is probably some code here that could be shared with
parray_origin or other methods.
-----------------------------------------------------------------------*/

static PyObject *
parray_spacing(
    PyObject *self)           /* Pyvox array to find the spacing of */
{
    PyObject *spacing;        /* List containing the dimensions */
    PyObject *dim;            /* One of the dimensions */
    voxel_array *varray;      /* The voxel array itself */
    int rank, i;

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, &varray) )
	fatal("This isn't a Pyvox array!");
    rank = varray->rank;

    /* Construct a list of the spacing values */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the list containing them, but
       the Python API manual is somewhat obscure on this point. */
    spacing = PyList_New(rank);
    for (i = 0; i < rank; i++) {
        dim = PyFloat_FromDouble(varray->spacing[i]);
	/* FIXME:  The Python C API manual does not specify the value
	   returned by PyList_SetItem; I am assuming that it returns
	   zero on success, and non-zero otherwise. */
	if ( PyList_SetItem(spacing, i, dim) )
	    fprintf(stderr, "PyList_SetItem is unhappy\n");
    }

    return spacing;
}



/**************************************************************************
Standard methods for the Pyvox array type

The following functions implement core methods for the Pyvox array
type and are listed in the type object.  There are also a few closely
related functions that are not actually listed in the type object.
**************************************************************************/



/*-----------------------------------------------------------------------
parray_create - Create an instance of a Pyvox array

This function creates a brand-new instance of a Pyvox array and
initializes it to contain a voxel array of rank 0, containing no data.
This function is preferred over parray_alloc when the created array
will promptly be used as a destination by the Voxel Kit, which will do
its own allocation.  The reference count of the created array is set
to 1.  Note that if Py_BuildValue is used to created a result to be
returned, the "N" code should be used rather than "O", which
increments the reference count.

FIXME: It's probably better to have this function return its result as
type PyObject *, but there's a lot of code in this file that assumes
the existing type.

FIXME: While you're fixing the first bug, also arrange for this
function to return the voxel_array via a pointer argument.
------------------------------------------------------------------------*/

pyvox_array *
parray_create(void)
{
    pyvox_array *self;        /* Pyvox array being created */
    voxel_array *varray;      /* The voxel array that it contains */

    /* Create a voxel array */
    varray = vxl_create_array();

    /* Create a Pyvox array that contains it */
    self = (pyvox_array *)PyObject_NEW(pyvox_array, &parray_type);
    if (self == 0)
        ReturnError("Creation of Pyvox array failed");
    self->varray = varray;

    /* Report creation of a new Pyvox array */
    if (DEBUG_ALLOC) {
	fprintf(stderr, "parray create:  %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)self)->ob_refcnt); 
    }

    return self;
}



/*-----------------------------------------------------------------------
parray_destroy - Destroy an instance of a Pyvox array

This function destroys an instance of a Pyvox array by freeing the
storage associated with both the voxel array itself and the Python
object wrapper that refers to it.  It is normally called automatically
by the Python garbage collector when the reference count of the object
goes to zero.
-----------------------------------------------------------------------*/

static inline
void
parray_destroy(
    pyvox_array *self)
{
    voxel_array *varray;          /* The voxel array in the object */

    if (DEBUG_ALLOC)
        fprintf(stderr, "parray destroy: %p\n", (void *)self);

    varray = self->varray;
    vxl_destroy_array(varray);
    PyMem_DEL(self);

    return;
}



/*-----------------------------------------------------------------------
parray_alloc - Create and allocate an instance of a voxel array object

This function creates a brand-new instance of a Pyvox array,
initializes it to a given type, rank, and dimensions, and allocates
storage for the contents.  The contents of the array are left
undefined, on the perhaps dubious assumption that the user will
immediately fill them in.  This function should be used only when it
is necessary to actually allocate and initialize the voxel array; if
the new array is used as a destination for a Voxel Kit function which
allocates the array (which is the usual case), then parray_create
should be used instead to avoid a redundant allocation.  The reference
count of the created array is set to 1.  Note that if Py_BuildValue
is used to created a result to be returned, the "N" code should be
used rather than "O", which increments the reference count.
------------------------------------------------------------------------*/

PyObject *
parray_alloc(
    int type,                   /* Type of voxel array to create */
    int rank,                   /* Rank of voxel array to create */
    long *dimen)                /* Dimensions of voxel array */
{
    pyvox_array *self;          /* Pyvox array being created */
    voxel_array *varray;        /* The voxel array that it contains */

    /* Create and allocate a voxel array */
    varray = vxl_create_array();
    vxl_alloc_array(varray, type, rank, dimen);

    /* Create a Pyvox array that contains it */
    self = (pyvox_array *)PyObject_NEW(pyvox_array, &parray_type);
    if (self == 0)
        ReturnError("creation of Pyvox array failed");
    self->varray = varray;

    /* Report creation of a new Pyvox array */
    if (DEBUG_ALLOC) {
	fprintf(stderr, "parray alloc:   %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)self)->ob_refcnt); 
    }

    return (PyObject *)self;
}



/*------------------------------------------------------------------------
parray_copy - Create a copy of a Pyvox array

This Pyvox array method creates a new Pyvox array with the same contents
and attributes of an existing array, but which shares no storage with it.
-------------------------------------------------------------------------*/

static PyObject *
parray_copy(
    PyObject *self, 
    PyObject *args)
{
    pyvox_array *in,  *out;         /* Input and output as Pyvox arrays */
    voxel_array *vin, *vout;        /* Input and output as voxel arrays */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in  = (pyvox_array *)self;
    vin = in->varray;

    /* Copy the voxel array */
    vout = vxl_copy_array(vin);

    /* Create a new Pyvox array to contain it */
    out = (pyvox_array *)PyObject_NEW(pyvox_array, &parray_type);
    if (out == 0)
        ReturnError("Creation of Pyvox array failed");
    out->varray = vout;

    /* Return the new Pyvox array */
    return (PyObject *)out;
}



/*------------------------------------------------------------------------
parray_print - Print (informal) representation of a Pyvox array

This function prints out an informal representation of a Pyvox array
on a specified FILE.  This representation contains enough information
to distinguish different objects, but not to reconstruct the object
when read back in.  (It was thought undesireable to print out a 10 MB
object in full, without at least warning the poor unsuspecting user).

FIXME: The format could use a lot of improvement yet.

FIXME: Provide a means for printing out the whole array, for those who
really, really want it.
-------------------------------------------------------------------------*/

static int
parray_print(
    PyObject *self,              /* The object being printed */
    FILE *out,                   /* Where to print it */
    int flags)                   /* FIXME:  What's this for? */
{
    voxel_array *vself;          /* Voxel array struct */
    int rank;                    /* Rank of the voxel array */
    long *dimen;                 /* Dimensions of the array */
    long coord[VXL_MAX_RANK];    /* Index into the array */
    long count;                  /* Number of elements in array */
    const char *name;            /* Short name of type */
    int i, j;

    /* Get characteristics of the array */
    vself = PyVoxel_GetArray(self);
    rank  = vself->rank;
    dimen = vself->dimen;
    name  = exim_nameof_type(vself->type);
    count = vxli_count(vself);

    /* Print summary characteristics */
    fprintf(out, "<Pyvox voxel array: %s (", name);
    for (i = 0; i < rank-1; i++)
	fprintf(out, "%ld,", vself->dimen[i]);
    if (rank > 0)
	fprintf(out, "%ld", vself->dimen[rank-1]);
    fprintf(out, ") at %p>", self);

    /* If short, print all elements */
    coord[0] = coord[1] = 0;
    fprintf(out, "\n");
    if (rank == 0) {
	fprintf(out, "  %f", vxl_get_voxel(vself, coord));
    }
    else if(rank == 1 && count <= 100) {
	for (i = 0; i < dimen[0]; i++) {
	    coord[0] = i;
	    fprintf(out, "  %f", vxl_get_voxel(vself, coord));
	}
    }
    else if (rank == 2 && count <= 100) {
	for (i = 0; i < dimen[0]; i++) {
	    coord[0] = i;
	    for (j = 0; j < dimen[1]; j++) {
		coord[1] = j;
		fprintf(out, "  %f", vxl_get_voxel(vself, coord));
	    }
	    fprintf(out, "\n");
	}
    }
    else {
	fprintf(out, 
		"  [%ld elements not printed to save electrons]",
		count);
    }
    fprintf(out, "\n");

    return 0;
}



/*----------------------------------------------------------------------
parray_getattr - Get an attribute of a Pyvox array

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  Convert to a
table-driven approach, perhaps with hashing.

FIXME: Is it safe to use foo.type to get the internal type of the data
in a voxel array, or does this somehow conflict with the type()
built-in of Python?
----------------------------------------------------------------------*/

static PyObject *
parray_getattr(
    PyObject *self,          /* Pyvox array to get attribute of */
    char *attr)              /* Name of the attribute to get */
{
    voxel_array *varray;     /* The voxel array itself */

    /* Check for some errors */
    if ( !PyvoxArray_Check(self, &varray) )
	fatal("This isn't a Pyvox array!");

    /* Return value of any requested member of the instance */
    if (strcmp(attr, "origin") == 0)
	return parray_origin(self);
    else if (strcmp(attr, "rank") == 0)
	return Py_BuildValue("i", varray->rank);
    else if (strcmp(attr, "refcnt") == 0)
	return parray_refcnt(self);
    else if (strcmp(attr, "spacing") == 0)
	return parray_spacing(self);
    else if (strcmp(attr, "size") == 0)
	return parray_size(self);
    else if (strcmp(attr, "type") == 0)
	return Py_BuildValue("i", varray->type);

    /* Provide list of members available from this instance */
    if (strcmp(attr, "__members__") == 0)
	return Py_BuildValue("[sssss]", "origin",  "rank",
			     "spacing", "size", "type");

    /* Use the Python API to find methods associated with this type */
    return Py_FindMethod(parray_methods, self, attr);
}



/*----------------------------------------------------------------------
parray_setattr - Set an attribute of a Pyvox array

This method sets a given attribute of a Pyvox array.

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  Convert to a
table-driven approach, perhaps with hashing.

FIXME:  Does this handle inherited attributes correctly?
----------------------------------------------------------------------*/

static int
parray_setattr(
    PyObject *self,          /* Pyvox array to set attribute of */
    char *attr,              /* Name of the attribute to get */
    PyObject *value)         /* Value to set attribute to */
{
    voxel_array *varray;     /* The voxel array itself */

    /* Check for some errors */
    if ( !PyvoxArray_Check(self, &varray) )
	fatal("This isn't a Pyvox array!");

    /* Return value of any requested member of the instance */
    if (strcmp(attr, "origin") == 0)
	return parray_set_origin(self, value);
    else if (strcmp(attr, "spacing") == 0)
	return parray_set_spacing(self, value);
    else
	return 0;
}



/***********************************************************************
Standard Mapping Methods for the Pyvox Array Type
***********************************************************************/



/*-----------------------------------------------------------------------
parray_convert_slice - Convert a Python slice object to C values

This function takes a Python slice object and converts the initial,
final, and step values to C long values.

FIXME: Handle Python's negative number convention for indices relative
to the end of the sequence.

FIXME: Generate an error if the contents of the slice object are not
numbers.
------------------------------------------------------------------------*/

static void
parray_convert_slice(
    PyObject *ob,                /* Python slice object to convert */
    long deflimit,               /* Default value for limit */
    long *init,                  /* (out) initial value from slice ob */
    long *limit,                 /* (out) limit value from slice ob */
    long *step)                  /* (out) step value from slice ob */
{
    PyObject *val;

    val = PyObject_GetAttrString(ob, "start");
    if ( val == Py_None )
	*init = 0;
    else
	*init = PyInt_AsLong(PyNumber_Int(val));
    val = PyObject_GetAttrString(ob, "stop");
    if ( val == Py_None )
	*limit = deflimit;
    else
	*limit = PyInt_AsLong(PyNumber_Int(val));
    val = PyObject_GetAttrString(ob, "step");
    if ( val == Py_None )
	*step = 1;
    else
	*step = PyInt_AsLong(PyNumber_Int(val));
    
    return;
}



/*------------------------------------------------------------------------
parray_prep_slice - Convert slice description from Python to Voxel Kit

This function takes a slice description in the style used within
Python (or Pyvox) and converts it to the style used within the Voxel
Kit.  The subscr argument may be a naked number, which is interpreted
as a (single) index value for a rank-1 array; a naked slice object,
which applies to the only axis of a rank-1 array; a naked ellipsis,
which is interpreted to mean that the desired slice is (a copy of) the
entire array; or a list or tuple of subscripts.  Each subscript may be
a number; a slice object; a tuple or list of numbers; or an ellipsis.
As a special case, a rank-0 (scalar) array will accept (and ignore)
any number of subscripts, returning its single element as a number;
this is mostly to work around the fact that Python syntax does not
allow you to provide an empty list of subscripts to an array.
Attempting to access a non-existent voxel raises an error.

There are three possible values for the returned value.  If the
conversion is unsuccessful (and not handled by fatal), zero is
returned.  If the pre-allocated space provided as the temp argument is
sufficiently large or not used at all, its value is return and must
not be freed.  If the pre-allocated space is inadequate, or temp was
zero, then a pointer to space allocated by this function is returned,
and must be freed by the caller.

NOTE: The usual assignment "bar = foo" in Python does a shallow copy,
thus causing bar and foo to refer to the same array.  To copy the
contents (a deep copy), you must use the form "bar = foo[...]".

FIXME: Error management is not handled consistently.

FIXME: Possibly allow a rank-1 array in place of a tuple or list; or a
rank-0 array in place of a number?  The right way to do this is
probably to get int() and float() to return a Python int or float for
a singleton array; then PyvoxLong should do the right thing.

FIXME: Extend this function to permit things like foo[:] to return all
the elements as a vector; NewAxis as used in NumPy; and the negative
number convention used in Python to select elements relative to the
end of a vector.

ALGORITHM: The possibility of ellipses means that the number of items
in the subsrc argument and the number of dimensions in the self array
may not be equal.  To keep them straight, we refer to the first as
subscripts and the second as coordinates.  The actual values to be used
as coordinates are referred to as indices.

The subscript list is scanned twice, once to determine how much temporary
storage is needed and once to actually process it.  This appears to yield
the simplest code and should not be unreasonably inefficient compared to
the time taken to actually extract the desired slice.
------------------------------------------------------------------------*/

#define PATEMP 64      /* Size of pre-allocated space */

#if (PATEMP < VXL_MAX_RANK) || (PATEMP < 3)
#error "Defined value of PATEMP violates its constraints"
#endif

static long *                   /* Temp space to be freed after use */
parray_prep_slice(
    voxel_array *sarray,        /* Voxel array to slice from */
    PyObject *subscr,           /* Subscripts for desired voxel or slice */
    long *nidx,                 /* (out) How to handle each axis */
    long **index,               /* (out) Index values along each axis */
    long *temp)                 /* Preallocated temp space, length PATEMP */
{
    int rank;                   /* Rank of the voxel array */
    PyObject *sub;              /* Current subscript */
    int nsub, isub;             /* Number of items in subscr, and index */
    int icor;                   /* Index over coordinates in voxel array */
    int nell;                   /* Number of ellipses seen */
    long ntemp = 0;             /* Temp space needed for coords and slices */
    long *p, nelem, s;
    int i;

    /* DEBUG: Show input subscript */
    if (0)
        PyObject_Print(subscr, stderr, 0);

    /* Handle rank 0 as a special case */
    rank = sarray->rank;
    if (rank == 0) {
        return temp; }

    /* Handle naked ellipsis as a special case */
    if ( subscr == Py_Ellipsis ) {
	for (i = 0; i < rank; i++)
	    nidx[i] = 0;
	return temp;
    }

    /* Handle naked number for rank-1 array as a special case */
    if ( rank == 1 && PyvoxLong_Check(subscr, &temp[0]) ) {
	nidx[0]  = 1;
	index[0] = &temp[0];
	return temp;
    }

    /* Handle naked slice object for rank-1 array as special case */
    if ( rank == 1 && PySlice_Check(subscr) ) {
	nidx[0]  = -1;
	index[0] = &temp[0];
	parray_convert_slice(subscr, sarray->dimen[0], temp, temp+1, temp+2);
	return temp;
    }

    /* Otherwise the subscr argument must be a sequence object */
    if ( !PySequence_Check(subscr) )
	fatal("Invalid subscripts to an array");

    /* Prescan the subscripts to determine the total number of
       ellipsis and non-ellipsis subscripts, and the space that must
       be allocated for indices. */
    nsub  = PyObject_Length(subscr); 
    nell  = 0;
    ntemp = 0;
    for (isub = 0; isub < nsub; isub++) {
        sub = PySequence_GetItem(subscr, isub);
        if ( PyNumber_Check(sub) ) {
            /* Subscript is a number */
            ntemp += 1;
        }
        else if ( PySequence_Check(sub) ) {
            /* Subscript is a list of numbers */
            nelem = PyObject_Length(sub);
            if ( nelem == 0 )
                fatal("Empty list of indices is not permitted");
            for (i = 0; i < nelem; i++)
                if ( ! PyNumber_Check(PySequence_GetItem(sub, i)) )
                     fatal("Index value is not a number");
            ntemp += nelem;
        }
        else if ( PySlice_Check(sub) ) {
            /* Subscript is a slice object */
            ntemp += 3;
        }
        else if ( sub == Py_Ellipsis ) {
            /* Subscript is an ellipsis */
            if (nell > 0)
                fatal("More than one ellipsis in subscript list");
            nell = 1;
        }
        else {
            fatal("Illegal object in subscript list");
        }
    }
    if ( (nell == 0 && nsub != rank) || (nell > 0 && nsub > rank) )
        fatal("Too many or too few subscripts");

    /* Allocate space for index arrays, if needed */
    if ( ntemp > PATEMP )  
        temp = mallock(ntemp*sizeof(long));

    /* Convert slice objects to sequences, scalars and lists to index
       arrays, and a single ellipsis to the missing coordinates */
    icor = 0;
    p = temp;
    for ( isub = 0; isub < nsub; isub++) {
        sub = PySequence_GetItem(subscr, isub);
	if ( PyvoxLong_Check(sub, &s) ) {
            /* Subscript is a number */
            nidx [icor] = 1;
            index[icor++] = p;
            *p++ = s;
        }
        else if ( PySequence_Check(sub) ) {
            /* Subscript is a list of numbers */
            nelem = PyObject_Length(sub);
            nidx [icor] = nelem;
            index[icor++] = PyNumSeq_AsLong(sub, p, nelem);
            p += nelem;
        }
        else if ( PySlice_Check(sub) ) {
            /* Subscript is a slice object */
            nidx [icor] = -1;
            index[icor] = p;
	    parray_convert_slice(sub, sarray->dimen[icor], p, p+1, p+2);
	    p += 3;
            icor++;
        }
        else if ( sub == Py_Ellipsis ) {
            /* Subscript is an ellipsis */
            for (i = nsub-1; i < rank; i++)
                nidx[icor++] = 0;
        }
        else {
            panic("Internal error: Unexpected type in subscripts");
        }
    }
    if ( icor > rank || p > temp + ntemp )
        panic("Internal error: overran allocated space");

    return temp;
}



/*------------------------------------------------------------------------
parray_getitem - Return the voxel value(s) at given coordinates

This function returns the voxel values at a given position or in a
given slice.  See parray_prep_slice for a description of the subscript
list.  Attempting to retrieve a non-existent voxel raises an error.

FIXME: Remove axes for which the subscript is a naked number?

FIXME: Accessing a non-existent voxel crashes Python; I need to
convert fatal() into a Python exception instead.
------------------------------------------------------------------------*/

static PyObject *
parray_getitem(
    PyObject *self,             /* Voxel array object to fetch from */
    PyObject *subscr)           /* Subscripts for desired voxel or slice */
{
    voxel_array *sarray;        /* Voxel array indicated by self */
    int rank;                   /* Rank of self as voxel array */
    pyvox_array *result;        /* The extracted subarray */
    voxel_array *rarray;        /* ...and its voxel array */
    long patemp[PATEMP];        /* Pre-allocated temp space */
    long *temp = patemp;        /* Actual temp space used */
    long nidx[VXL_MAX_RANK];    /* Methods arg for vxl_get_slice */
    long *index[VXL_MAX_RANK];  /* Indices arg for vxl_get_slice */

    /* DEBUG: print the arguments */
    if (0) {
	PyObject_Print(self,  stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	PyObject_Print(subscr, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
    }

    /* Get the voxel array itself, plus its rank */
    sarray = ((pyvox_array *)self)->varray;
    rank   = sarray->rank;

    /* If array is rank 0 or all subscripts are numbers, return a number */
    if ( rank == 0 || PyNumSeq_AsLong(subscr, temp, rank) ) {
	return PyFloat_FromDouble( vxl_get_voxel(sarray, temp) ); }

    /* Convert slice description for use by Voxel Kit */
    temp = parray_prep_slice(sarray, subscr, nidx, index, patemp);
    if (temp == 0)
	return 0;

    /* Allocate space for the result */
    result = parray_create();
    rarray = result->varray;

    /* Call the Voxel Kit to do the actual work */
    vxl_get_slice(rarray, sarray, nidx, index);

    /* Discard temporary storage and return the result */
    if (temp != patemp)   free(temp);
    return (PyObject *)result;
}



/*------------------------------------------------------------------------
parray_length - Return the length of a Pyvox array

This function returns the length of a Pyvox array and is used to
implement the Pythen len(x) function.

FIXME:  Python expects this function to return int, but array lengths
can be long instead; this will be a bug on platforms for which int and
long are different.  Perhaps check that the count is <= INT_MAX and
abort if not.

FIXME:  Is any other error checking necessary here?
------------------------------------------------------------------------*/

static int
parray_length(
    PyObject *self)
{
    return vxli_count(((pyvox_array *)self)->varray);
}



/*------------------------------------------------------------------------
parray_setitem - Set voxels in a given slice

This function sets the values of the voxels at a given position or in
a given slice.  The source may be either a Python numeric value or
rank 0 voxel array, which is converted to the type of the voxel array
according to the rules of the underlying C implementation and
broadcast to all the voxels contained in the slice, or a voxel array
containing exactly the same number of elements as the specified slice,
which are converted to the type of the destination array and copied
into it.  See parray_prep_slice for a description of the subscript
list.  Attempting to set a non-existent voxel raises an error.


FIXME: This might be somewhat inefficient for the case where all the
subscripts are numbers and vxl_set_voxel could be used.

FIXME: Extend this to work for the case "foo[...] = [...]"; that is,
for assigning the contents of a sequence into a subarray, or the
entire array without changing its shape or type.

FIXME: Also support sequences (that is, Python slice objects) of
numbers as the value to be assigned?

FIXME: Accessing a non-existent voxel crashes Python; I need to
convert fatal() into a Python exception instead.
------------------------------------------------------------------------*/

static int                      /* Returns status: 0 => ok */
parray_setitem(
    PyObject *self,             /* Voxel array object to set values in */
    PyObject *subscr,           /* Subscripts for desired voxel or slice */
    PyObject *value)            /* Value(s) to set voxel or slice to */
{
    voxel_array *darray;        /* Voxel array destination indicated by self */
    int drank;                  /* Rank of destination */
    voxel_array *sarray;        /* Source array */
    long patemp[PATEMP];        /* Pre-allocated temp space */
    long *temp = patemp;        /* Actual temp space used */
    long nidx[VXL_MAX_RANK];    /* Methods arg for vxl_set_slice */
    long *index[VXL_MAX_RANK];  /* Indices arg for vxl_set_slice */
    PyObject *sval;             /* Scalar array containing constant value */

    /* DEBUG: print the arguments */
    if (0) {
	PyObject_Print(self,  stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	PyObject_Print(subscr, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	PyObject_Print(value, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
    }

    /* Get the destination voxel array itself, plus its rank */
    darray = ((pyvox_array *)self)->varray;
    drank  = darray->rank;

    /* Get the value(s) to assign, wrapping in a scalar array if needed */
    if ( PyvoxArray_Check(value, &sarray) ) {
        sval   = 0;
    }
    else if ( (sval = PyScalar_FromNumber(value, darray->type)) ) {
        sarray = ((pyvox_array *)sval)->varray;
    }
    else {
        PyErr_SetString(PyvoxError, 
                        "Can assign to slice only from array or number");
        return 1;
    }

    /* Call the Voxel Kit to do the real work */
    temp = parray_prep_slice(darray, subscr, nidx, index, patemp);
    if (temp == 0)
	return 0;
    vxl_set_slice(darray, nidx, index, sarray);

    /* Free temporary objects and return */
    if (sval != 0)   {
        Py_DECREF(sval);
    }
    if (temp != patemp)  free(temp);

    return 0;
}



/*-----------------------------------------------------------------------
parray_as_mapping - Mapping protocol functions for the voxel array type
-----------------------------------------------------------------------*/

static PyMappingMethods
parray_as_mapping = {
    (inquiry)       parray_length,    /* mp_length:        len(x)   */
    (binaryfunc)    parray_getitem,   /* mp_subscript:     x[y]     */
    (objobjargproc) parray_setitem,   /* mp_ass_subscript: x[y] = v */
};



/************************************************************************
Standard sequence methods for the Pyvox array type

Almost all the subscripting methods are handled by the mapping methods,
but for 1-dimensional arrays, Python calls the sequence methods instead.
So we must provide methods suitable for 1-d arrays here.

FIXME: The C functions implementing the Python constructs len(x), x[i],
x[i:j], x[i] = v, and x[i:j] = v could be taken from either the mapping
or sequence method tables when i and j are integers.  This is a very
fertile ground for confusion, so watch carefully for errors.  Note also
that the choice made in the ambiguous case might change in later versions
of Python.
************************************************************************/



/*------------------------------------------------------------------------
parray_newslice - Create a new slice object

This function creates a new slice object given its start, limit, and
step values.  It is probably used only in parray_getslice and
parray_setslice to work around some limits of the Python
implementation.
-------------------------------------------------------------------------*/

static PyObject *
parray_newslice(
    long start,           /* Start value */
    long limit,           /* Limit value */
    long step)            /* Step  value */
{
    PyObject *pystart, *pylimit, *pystep;   /* Values as Python objects */

    pystart = PyInt_FromLong(start);
    pylimit = PyInt_FromLong(limit);
    pystep  = PyInt_FromLong(step);

    return PySlice_New(pystart, pylimit, pystep);
}



/*------------------------------------------------------------------------
parray_getslice - Get given slice of a 1-d array

This function returns a Pyvox array containing the voxel values in a
given slice of a 1-dimensional array; arrays with other numbers of
dimensions are handled through the mapping methods.  Attempting to
retrieve a non-existent voxel raises an error.

ALGORITHM: We take the easy way out here, and just construct a slice
object and pass it to parray_getitem to do the real work.  The
argument is that we are just working around a bug in Python.
------------------------------------------------------------------------*/

static PyObject *
parray_getslice(
    PyObject *self,             /* Voxel array object to fetch from */
    int start,                  /* Index of first element in slice */
    int limit)                  /* Index of limit element in slice */
{
    PyObject *subscr;           /* Indices as a slice object */
    PyObject *result;           /* Resulting slice */  

    /* Convert indices into a slice object and call parray_getitem */
    subscr = parray_newslice(start, limit, 1);
    result = parray_getitem(self, subscr);

    Py_DECREF(subscr);
    return result;
}



/*------------------------------------------------------------------------
parray_setslice - Set given slice of a 1-d array

This function sets the contents of a given slice of a 1-dimensional
array from another 1-d array; arrays with other numbers of dimensions
are handled through the mapping methods.  Attempting to set a
non-existent voxel raises an error.

ALGORITHM: We take the easy way out here, and just construct a slice
object and pass it to parray_getitem to do the real work.  The
argument is that we are just working around a bug in Python.
------------------------------------------------------------------------*/

static int
parray_setslice(
    PyObject *self,             /* Voxel array object to fetch from */
    int start,                  /* Index of first element in slice */
    int limit,                  /* Index of limit element in slice */
    PyObject *value)            /* Values to be set */
{
    PyObject *subscr;           /* Indices as a slice object */
    int result;                 /* Result returned from parray_setitem */

    /* Convert indices into a slice object and call parray_getitem */
    subscr = parray_newslice(start, limit, 1);
    result = parray_setitem(self, subscr, value);

    Py_DECREF(subscr);
    return result;
}



/*----------------------------------------------------------------------
parray_as_sequence - Table of sequence methods for the Pyvox array type
----------------------------------------------------------------------*/

static PySequenceMethods 
parray_as_sequence = {
    (inquiry)          0,                  /* sq_length:     len(x)     */
    (binaryfunc)       0,                  /* sq_concat:     x + y      */
    (intargfunc)       0,                  /* sq_repeat:     n * x      */
    (intargfunc)       0,                  /* sq_item:       x[i]       */
    (intintargfunc)    parray_getslice,    /* sq_slice:      x[i:j]     */
    (intobjargproc)    0,                  /* sq_ass_item:   x[i] = v   */
    (intintobjargproc) parray_setslice,    /* sq_ass_slice:  x[i:j] = v */
};



/************************************************************************
Standard Numeric Methods for the Pyvox Array Type

FIXME: Many of these functions could be written using a standard wrapper
function; the result would be more maintainable.
************************************************************************/



/*----------------------------------------------------------------------
parray_coerce - Coerce voxel arrays to common type if possible

This function takes two Python objects that may be voxel arrays,
scalars, or numbers and coerces them to a common type so that
arithmetic may be performed upon them.  The converted objects replace
the original objects.  Zero is returned if conversion is successful,
and a non-zero value if it fails.  Checking for compatibility
of types and sizes is done by the arithmetic routines themselves.
The references passed by the caller are assumed to be borrowed, and
the reference counts will be incremented if the object is retained
without conversion.

FIXME:  Generalize to convert two voxel arrays of different types
when needed.  This will have to wait until the cast functions in
bips are implemented.

FIXME: The exact semantics of this function don't seem to be defined
in the Python documentation, and I have guessed at it from studying a
similar routine in Numerical Python.  Some verification might be in
order.  Some questions:  Can the case "voxel sees voxel" be used to
coerce incompatible voxel types?  Is the case "number sees voxel"
actually used, or does Python convert automatically to the symmetric
case "voxel sees number"?  Are the reference counts managed correctly?
------------------------------------------------------------------------*/

static int
parray_coerce(
    PyObject **src1,         /* First object */
    PyObject **src2)         /* Second object */
{
    int intype;              /* Internal type of voxel array */

    if ( PyvoxArray_Check(*src1, 0) && PyvoxArray_Check(*src2, 0) ) {
	if (0) fprintf(stderr, "Voxel sees voxel\n");
	Py_INCREF(*src1);
	Py_INCREF(*src2);
	return 0; }
    else if ( PyvoxArray_Check(*src1, 0) && PyNumber_Check(*src2) ) {
	if (0) fprintf(stderr, "Voxel sees number\n");
	intype = PyVoxel_Type(*src1);
	*src2 = PyScalar_FromNumber(*src2, intype);
	Py_INCREF(*src1);
	return 0; }
    else if ( PyNumber_Check(*src1) && PyvoxArray_Check(*src2, 0) ) {
	if (0) fprintf(stderr, "Number sees voxel\n");
	intype = PyVoxel_Type(*src2);
	*src1 = PyScalar_FromNumber(*src1, intype);
	Py_INCREF(*src2);
	return 0; }
    else
	return -1;          /* Cannot make compatible */
	
}



/*-------------------------------------------------------------------------
parray_abs1 - Elementwise absolute value of a voxel array

This deprecated function is invoked by the Python abs(array); the
recommended syntax array.abs() is consistent with other univariate
array operations in Pyvox and is implemented by parray_abs2.

FIXME:  The functions parray_abs1, parray_minus, and parray_not are all
essentially the same except for the name of the Voxel Kit function to
be executed; perhaps they should all be implemented using the same
wrapper function.
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_abs1(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    voxel_array  *src1;     /* Operand as voxel array */
    PyObject *pyres;        /* Result to be */
    pyvox_array *result;    /* Result as a Pyvox array */

    /* Check that the argument is a voxel array */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_abs(result->varray, src1);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_minus - Elementwise additive inverse of a voxel array
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_minus(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    voxel_array  *src1;     /* Operand as voxel array */
    PyObject *pyres;        /* Result to be */
    pyvox_array *result;   /* Result as a Pyvox array */

    /* Check that the argument is a voxel array */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_minus(result->varray, src1);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_not - Bitwise NOT of a voxel array
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_not(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    voxel_array  *src1;     /* Operand as voxel array */
    PyObject *pyres;        /* Result to be */
    pyvox_array *result;    /* Result as a Pyvox array */

    /* Check that the argument is a voxel array */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_not(result->varray, src1);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_plus - Complement to additive inverse of a voxel array

This function is invoked to compute "+V" when V is a voxel array;
it does nothing except to pass through its argument, without even
copying it or doing any error checking.  It is provided only for
symmetry with "-V".
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_plus(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    return pysrc1;
}



/*-------------------------------------------------------------------------
parray_and - Bitwise AND of two voxel arrays

FIXME: All the binary numeric operations use very similar code; work out
a way to use a common wrapper function for them.
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_and(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_and(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_or - Bitwise OR of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_or(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_or(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_xor - Bitwise XOR of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_xor(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_xor(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_add -  Sum of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_add(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_add(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_sub - Difference of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_sub(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_sub(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_mul - Product of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_mul(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_mul(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*-------------------------------------------------------------------------
parray_div - Quotient of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_div(
    PyObject *pysrc1,       /* Arg 1: Left operand */
    PyObject *pysrc2)       /* Arg 2: Right operand */
{
    voxel_array  *src1, *src2;         /* Operands as voxel arrays */
    PyObject *pyres;                   /* Result to be */
    pyvox_array *result;               /* Result as a Pyvox array */

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, &src1) )
	ReturnError("Invalid first argument");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Invalid second argument");

    /* Call Voxel Kit to do the actual computation */
    result = parray_create();
    vxl_div(result->varray, src1, src2);
    pyres = (PyObject *)result;

    return pyres;
}



/*----------------------------------------------------------------------
parray_as_number - Number protocol functions for the voxel array type

FIXME: Arrange for float(), int(), and long() to return a number for
singleton arrays and an error otherwise.  This will go partway toward
making singleton arrays act like numbers in the appropriate contexts.
-----------------------------------------------------------------------*/

static PyNumberMethods
parray_as_number = {
	(binaryfunc)  parray_add,       /* nb_add */
	(binaryfunc)  parray_sub,       /* nb_subtract */
	(binaryfunc)  parray_mul,       /* nb_multiply */
	(binaryfunc)  parray_div,       /* nb_divide */
	(binaryfunc)  0,                /* nb_remainder */
	(binaryfunc)  0,                /* nb_divmod */
	(ternaryfunc) 0,                /* nb_power */
	(unaryfunc)   parray_minus,     /* nb_negative */
	(unaryfunc)   parray_plus,      /* nb_positive */
	(unaryfunc)   parray_abs1,      /* nb_absolute */
	(inquiry)     0,                /* nb_nonzero */
	(unaryfunc)   parray_not,       /* nb_invert */
	(binaryfunc)  0,                /* nb_lshift */
	(binaryfunc)  0,                /* nb_rshift */
	(binaryfunc)  parray_and,       /* nb_and */
	(binaryfunc)  parray_xor,       /* nb_xor */
	(binaryfunc)  parray_or,        /* nb_or */
	(coercion)    parray_coerce,    /* nb_coerce */
	(unaryfunc)   0,                /* nb_int */
	(unaryfunc)   0,                /* nb_long */
	(unaryfunc)   0,                /* nb_float */
	(unaryfunc)   0,                /* nb_oct */
	(unaryfunc)   0,                /* nb_hex */
};



/**************************************************************************
Simple Unary Point Operations on Pyvox Arrays

These operations correspond directly to Voxel Kit operations and can
be handled via simple wrapper functions.

FIXME: This is a somewhat elastic category and may need some refinement.
Note also that some methods may still need to be modified to use the
standard wrapper functions.
**************************************************************************/



/*-------------------------------------------------------------------------
parray_vxl_unop - Generic unary operation on voxel arrays

This wrapper function is used to translate a standard unary Voxel Kit
function into a Pyvox array method.  More precisely, this function
applies a given function of type vxl_unop from the Voxel Kit to (the
voxel array contained in) self and returns the resulting voxel array
as a Pyvox array.  The Pyvox method takes no arguments.
-------------------------------------------------------------------------*/

static inline
PyObject *
parray_vxl_unop(
    vxl_unop *funct,          /* Voxel Kit function to apply */
    PyObject *self,           /* Self given to wrapper function */
    PyObject *args)           /* Args given to wrapper function */
{
    voxel_array *vin, *vout;  /* Input and output as voxel arrays */
    pyvox_array *out;         /* Output as (Pyvox) Pyvox arrays */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;
    if ( !PyvoxArray_Check(self, &vin) )
	ReturnError("Self is not a Pyvox array");
  
    /* Compute and return the result */
    out  = parray_create();
    vout = out->varray;
    funct(vout, vin);

    return (PyObject *)out;
}



/*-----------------------------------------------------------------------
Type conversions

These Pyvox array methods convert self, which is a Pyvox array any
type into a new Pyvox array of the specified type.  The details of the
conversion are inherited from the underlying C implementation.  A
Pyvox array can be converted to another Pyvox array of the same type;
this is simply a copy (and is probably almost as fast as a plain
copy).

The code for parray_uchar is treated specially, to provide the capability
to scale before conversion, since this operation is used so heavily.
-------------------------------------------------------------------------*/

static PyObject *
parray_schar(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_schar, self, args); }

static PyObject *
parray_ushort(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_ushort, self, args); }

static PyObject *
parray_short(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_short, self, args); }

static PyObject *
parray_uint(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_uint, self, args); }

static PyObject *
parray_int(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_int, self, args); }

static PyObject *
parray_ulong(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_ulong, self, args); }

static PyObject *
parray_long(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_long, self, args); }

static PyObject *
parray_float(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_float, self, args); }

static PyObject *
parray_double(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_double, self, args); }



/*-------------------------------------------------------------------------
pyvox.uchar - Convert to unsigned char, with optional rescaling

This Pyvox function converts self, which is a Pyvox array of any type
into a Pyvox array of unsigned.  The caller may optionally specify the
lower and upper limits of the range to be linearly compressed to fit
the 0..255 range of unsigned char voxels; if no limits are specified,
then the input range 0..255 will be converted to the output range
0..255.
-------------------------------------------------------------------------*/

static PyObject *
parray_uchar(
    PyObject *self,
    PyObject *args)
{
    double minval =   0.0;    /* Arg 1: Lower limit of input range */
    double maxval = 255.0;    /* Arg 2: Upper limit of input range */

    voxel_array *vin;         /* Input as a voxel array struct */
    pyvox_array *out;         /* Output as a Pyvox array */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|dd", &minval, &maxval) )
	return NULL;
    if ( !PyvoxArray_Check(self, &vin) )
	ReturnError("Self is not a Pyvox array");
  
    /* Compute and return the result */
    out = parray_create();
    vxl_scale_to_uchar(out->varray, vin, minval, maxval);
    return (PyObject *)out;

}



/*------------------------------------------------------------------------
Unary operations on voxel arrays, in standard format

The following functions all take a single Pyvox voxel array as input,
generate a single Pyvox voxel array as output, and call upon a function
of type vxl_unop from the Voxel Kit.  This uniformity means that theyd
can all be implemented in the same fashion using the wrapper function
parray_vxl_unop.  Each definition *could* be reduced to a single line
by using an appropriate macro, but this would make them harder to find
using emacs tags and to understand.
-------------------------------------------------------------------------*/

static PyObject *
parray_abs2(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_abs, self, args); }

static PyObject *
parray_acos(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_acos, self, args); }

static PyObject *
parray_asin(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_asin, self, args); }

static PyObject *
parray_atan(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_atan, self, args); }

static PyObject *
parray_ceil(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_ceil, self, args); }

static PyObject *
parray_cos(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_cos, self, args); }

static PyObject *
parray_cosh(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_cosh, self, args); }

static PyObject *
parray_exp(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_exp, self, args); }

static PyObject *
parray_floor(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_floor, self, args); }

static PyObject *
parray_log(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_log, self, args); }

static PyObject *
parray_log10(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_log10, self, args); }

static PyObject *
parray_sin(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_sin, self, args); }

static PyObject *
parray_sinh(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_sinh, self, args); }

static PyObject *
parray_sqrt(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_sqrt, self, args); }

static PyObject *
parray_tan(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_tan, self, args); }

static PyObject *
parray_tanh(PyObject *self, PyObject *args)
{   return parray_vxl_unop(vxl_tanh, self, args); }



/*-------------------------------------------------------------------------
parray_logcomp - Log-compress an unsigned long volume image

This Pyvox array method compresses the dynamic range of an unsigned
long Pyvox array using the transformation y = A*log(1+x), where A is
chosen such that the largest voxel value actually present in the image
is converted to the value 255; the compressed image is returned as
unsigned char.
-------------------------------------------------------------------------*/

static 
PyObject *
parray_logcomp(
    PyObject *self,
    PyObject *args)
{
    pyvox_array *in, *out;    /* Input and output as Pyvox arrays */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = (pyvox_array *)self;
    if (in->varray->type != INTP_ULONG)
	ReturnError("Sorry, only ulong has been implemented yet.");
  
    /* Compute and return the result */
    out = parray_create();
    vxl_logcomp_ulong(out->varray, in->varray);
    return (PyObject *)out;

}



/*-------------------------------------------------------------------------
parray_lookup - Process self through a lookup table

This Pyvox array method takes each voxel in self and uses it as the
index into a lookup table, saving the result of the lookup as the
destination image.  Self must be either unsigned char or unsigned
short but may be of any shape.  The table image must be rank 1 and
contain at least as many entries as the largest voxel value in the
source image; but it (and thus the destination) may be of any type.
-------------------------------------------------------------------------*/

static PyObject *
parray_lookup(
    PyObject *self,         /* Self: Source image */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 1: Lookup table */
    PyObject *pydest;       /* Result: Result of operation */

    voxel_array  *src1, *src2;           /* Sources as voxel arrays */
    pyvox_array *vodest;                 /* Result as a Pyvox array */
    voxel_array  *dest;                  /* Result as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &pysrc2) )
	return NULL;
    if ( !PyvoxArray_Check(self, &src1) )
	ReturnError("Self is not a voxel array");
    if ( !PyvoxArray_Check(pysrc2, &src2) )
	ReturnError("Argument 1 is not a voxel array");

    /* Call Voxel Kit to do the actual computation */
    vodest = parray_create();
    dest   = vodest->varray;
    vxl_lookup(dest, src1, src2);
    pydest = (PyObject *)vodest;

    return pydest;
}




/**************************************************************************
Simple Binary Point Operations on Pyvox Arrays 

These operations correspond directly to Voxel Kit operations and can
be handled via simple wrapper functions.

FIXME: This is a somewhat elastic category and may need some refinement.
Note also that some methods may still need to be modified to use the
standard wrapper functions.
**************************************************************************/



/*-------------------------------------------------------------------------
setup_vxl_binpt - Set up a binary point operation in the Voxel Kit

Given two Python objects which may be either numbers or voxel array
objects, this function promotes numbers into scalar voxel array
objects so that the Voxel Kit may be called in all cases.  It is an
error to provide two voxel arrays of different types; casting is NOT
automatic as it is in C.  A number is promoted to a scalar array of
the same type as the other object; if two numbers are provided, then
they are both promoted to a default type specified as an argument.
Shape compatibility will be checked within the Voxel Kit itself.
Promoting a number yields a temporary scalar array which must be freed
once the operation is complete; the free1 and free2 flags indicate
which must be freed.  This function returns a pointer to the first
Python object on success (which is meaningless but non-zero), and zero
on failure.  

NOTE that this function is inappropriate for those arithmetic
operations for which the Python type coercion method is automatically
invoked; see parray_and for an example.

FIXME: When used in a pyvox class method, this function permits either
operand to be a scalar.  This is no longer possible when used as a
Pyvox array method.  Do I need to provide a means to compute binary
operations when the first operand is a number?  If so, how?  One idea
is to add a method pyvox.scalar which take a number and converts it to
a scalar array; however, this might not coerce to the right type.
The usage might be something like "pyvox.scalar(0.39).atan2(array)".
Another option might be for the methods to take either one or two
arguments: if one arg, then self and arg 1 are combined; if two args,
then args 1 and 2 are combined.

FIXME:  Exactly how should errors be handled?  The current system doesn't
really seem right somehow.
------------------------------------------------------------------------*/

PyObject *
setup_vxl_binpt(
    PyObject *pysrc1,            /* Left operand, as Python object */
    PyObject *pysrc2,            /* Right operand, as Python object */
    int deftype,                 /* Default type if both numbers */
    voxel_array **src1,          /* Left operand, as voxel array */
    int *free1,                  /* Should src1 be freed after use? */
    voxel_array **src2,          /* Right operand, as voxel array */
    int *free2)                  /* Should src2 be freed after use? */
{
    int type = 0;                /* Type of voxel arrays */
    double val;

    /* Case 1: Both arguments are Pyvox arrays; set up to call Voxel Kit */
    if ( PyvoxArray_Check(pysrc1, src1) &&
	 PyvoxArray_Check(pysrc2, src2) ) {
	if (0) fprintf(stderr, "Voxel meets voxel\n");
	*free1 = *free2 = 0;
    }

    /* Case 2: First argument is Pyvox array, second is number; convert
       second to scalar array and set up to call Voxel Kit */
    else if ( PyvoxArray_Check(pysrc1, src1) &&
	      PyNumber_Check(pysrc2) ) {
	if (0) fprintf(stderr, "Voxel meets number\n");
	type  = (*src1)->type;
	val = PyFloat_AsDouble(PyNumber_Float(pysrc2));
	*src2 = vxl_create_scalar(type, val);
	*free1 = 0;
	*free2 = 1;
    }

    /* Case 3: First argument is a number, second is Pyvox array; convert
       first to scalar array and set up to call Voxel Kit */
    else if ( PyNumber_Check(pysrc1) &&
	      PyvoxArray_Check(pysrc2, src2) ) {
	if (0) fprintf(stderr, "Number meets voxel\n");
	type  = (*src2)->type;
	val = PyFloat_AsDouble(PyNumber_Float(pysrc1));
	*src1 = vxl_create_scalar(type, val);
	*free1 = 1;
	*free2 = 0;
    }

    /* Case 4: Both arguments are numbers; convert both to default type */
    else if ( PyNumber_Check(pysrc1) && PyNumber_Check(pysrc2) ) {
	if (0) fprintf(stderr, "Number meets number\n");
	type = deftype;
	val = PyFloat_AsDouble(PyNumber_Float(pysrc1));
	*src1 = vxl_create_scalar(type, val);
	val = PyFloat_AsDouble(PyNumber_Float(pysrc2));
	*src2 = vxl_create_scalar(type, val);
	*free1 = *free2 = 1;
    }

    /* Case 5:  Anything else is an error */
    else
	ReturnError("Incompatible arguments to voxel array computation");

    return pysrc1;
}



/*-------------------------------------------------------------------------
parray_atan2 - Four-quadrant arc tangent of two voxel arrays

This Pyvox function computes the elementwise four-quadrant arc tangent
of two numeric voxel arrays; that is, it computes dest = atan2(x, y)
for each element.  The first array is self; the second is passed as
the method argument.
-------------------------------------------------------------------------*/

static 
PyObject *
parray_atan2(               /* Result: Combination of the operands */
    PyObject *self,         /* Self: First operand */
    PyObject *args)
{
    PyObject *arg1;         /* Arg 1: Second operand */

    voxel_array  *src1, *src2;       /* Operands as voxel arrays */
    int free1 = 0, free2 = 0;        /* Should src1 & src2 be freed? */
    pyvox_array  * dest;             /* Result as a Pyvox array */
    voxel_array  *vdest;             /* Result as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &arg1) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, arg1, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	return NULL;

    /* Call Voxel Kit to do the actual computation */
    dest  = parray_create();
    vdest = dest->varray;
    vxl_atan2(vdest, src1, src2);

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    return (PyObject *)dest;
}



/*-------------------------------------------------------------------------
parray_compare - Compare self with another Pyvox array

This Pyvox array method compares self element by element with another
Pyvox array and returns a new unsigned char voxel array object
containing the results of the comparison; each element of the result
is set to one of three given values according to whether the result of
the comparison is less than, equal to, are greater than.

FIXME: Generalize to let lsval, eqval, and gtval be arrays as well as
numbers.
-------------------------------------------------------------------------*/

static PyObject *
parray_compare(
    PyObject *self,         /* Self: First image to compare */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 1: Second image to compare */
    int lsval;              /* Arg 2: Output value if less than */
    int eqval;              /* Arg 3: Output value if equal */
    int gtval;              /* Arg 4: Output value if greater than */
    PyObject *pycomp;       /* Result: Unsigned char array of comparisons */

    voxel_array  *src1, *src2;           /* Sources as voxel arrays */
    int free1 = 0, free2 = 0;            /* Should src1 & src2 be freed? */
    pyvox_array *vocomp;                 /* Result as a Pyvox array */
    voxel_array  *comp;                  /* Result as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "Oiii", &pysrc2, 
			  &lsval, &eqval, &gtval) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, pysrc2, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	return NULL;

    /* Call Voxel Kit to do the actual comparison */
    vocomp = parray_create();
    comp   = vocomp->varray;
    vxl_compare(comp, src1, src2, lsval, eqval, gtval);
    pycomp = (PyObject *)vocomp;

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    return pycomp;
}



/*-------------------------------------------------------------------------
parray_max - Elementwise maximum of self and another Pyvox array

This Pyvox array method computes the elementwise maximum of self and
another Pyvox arrays; that is, it computes dest = max(x, y) for each
element.
-------------------------------------------------------------------------*/

static PyObject *
parray_max(
    PyObject *self,         /* Self: First operand */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 1: Second operand */
    PyObject *pydest;       /* Result: Result of operation */

    voxel_array  *src1, *src2;           /* Sources as voxel arrays */
    int free1 = 0, free2 = 0;            /* Should src1 & src2 be freed? */
    pyvox_array *vodest;                 /* Result as a Pyvox array */
    voxel_array  *dest;                  /* Result as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &pysrc2) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, pysrc2, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	return NULL;

    /* Call Voxel Kit to do the actual computation */
    vodest = parray_create();
    dest   = vodest->varray;
    vxl_max(dest, src1, src2);
    pydest = (PyObject *)vodest;

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    return pydest;
}



/*-------------------------------------------------------------------------
parray_min - Elementwise minimum of self and another Pyvox array

This Pyvox array method computes the elementwise minimum of self and
another Pyvox array; that is, it computes dest = min(x, y) for each
element.
-------------------------------------------------------------------------*/

static PyObject *
parray_min(
    PyObject *self,         /* Self: First operand */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 2: Second operand */
    PyObject *pydest;       /* Result: Result of operation */

    voxel_array  *src1, *src2;           /* Sources as voxel arrays */
    int free1 = 0, free2 = 0;            /* Should src1 & src2 be freed? */
    pyvox_array *vodest;                 /* Result as a Pyvox array */
    voxel_array  *dest;                  /* Result as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &pysrc2) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, pysrc2, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	return NULL;

    /* Call Voxel Kit to do the actual computation */
    vodest = parray_create();
    dest   = vodest->varray;
    vxl_min(dest, src1, src2);
    pydest = (PyObject *)vodest;

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    return pydest;
}



/*-------------------------------------------------------------------------
parray_pow - Power function (exponentiation) of two voxel arrays

This Pyvox array method computes the elementwise power function of
self and another Pyvox array; that is, it computes dest = pow(x, y)
for each element.
-------------------------------------------------------------------------*/

static PyObject *
parray_pow(
    PyObject *self,         /* Self: First operand */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 2: Second operand */
    PyObject *pydest;       /* Result: Result of operation */

    voxel_array  *src1, *src2;           /* Sources as voxel arrays */
    int free1 = 0, free2 = 0;            /* Should src1 & src2 be freed? */
    pyvox_array *vodest;                 /* Result as a Pyvox array */
    voxel_array  *dest;                  /* Result as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &pysrc2) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, pysrc2, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	return NULL;

    /* Call Voxel Kit to do the actual computation */
    vodest = parray_create();
    dest   = vodest->varray;
    vxl_pow(dest, src1, src2);
    pydest = (PyObject *)vodest;

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    return pydest;
}



/*************************************************************************
Other Instance Methods for Pyvox Arrays

FIXME: Add functions to change the type of a voxel array.
This is, of course, not guaranteed to be portable.
*************************************************************************/



/*-------------------------------------------------------------------------
parray_amax - Find maximum value in voxel array


This method finds the largest value in the voxel array.

See also rmax to find the maximum in each row, and max to find the
elementwise maximum of two arrays.
-------------------------------------------------------------------------*/

static PyObject *
parray_amax(
    PyObject *self,           /* Voxel array to find maximum element of */
    PyObject *args)           /* Arguments: none expected */
{
    PyObject *result;         /* Result: maximum value */

    voxel_array *vself;       /* Input volume image */
    double maxval;            /* Maximum value found */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    vself  = PyVoxel_GetArray(self);
  
    /* Compute and return the result */
    maxval = vxl_amax(vself);
    result = Py_BuildValue("d", maxval);
    return result;
}



/*-------------------------------------------------------------------------
parray_amin - Find minimum value in a voxel array


This method finds the smallest value in the voxel array.

See also rmin to find the minimum in each row, and min to find the
elementwise minimum of two arrays.
-------------------------------------------------------------------------*/

static PyObject *
parray_amin(
    PyObject *self,           /* Voxel array to find minimum element of */
    PyObject *args)           /* Arguments: none expected */
{
    PyObject *result;         /* Result: minimum value */

    voxel_array *vself;       /* Input volume image */
    double minval;            /* Minimum value found */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    vself  = PyVoxel_GetArray(self);
  
    /* Compute and return the result */
    minval = vxl_amin(vself);
    result = Py_BuildValue("d", minval);
    return result;
}



/*-------------------------------------------------------------------------
parray_asum - Find sum of elements in voxel array


This method finds the sum of the element values in the voxel array.
-------------------------------------------------------------------------*/

static PyObject *
parray_asum(
    PyObject *self,           /* Voxel array to sum */
    PyObject *args)           /* Arguments: none expected */
{
    PyObject *result;         /* Result: sum of elements */

    voxel_array *vself;       /* Input volume image */
    double res;               /* Result as a number */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    vself  = PyVoxel_GetArray(self);
  
    /* Compute and return the result */
    res = vxl_asum(vself);
    result = Py_BuildValue("d", res);
    return result;
}



/*-------------------------------------------------------------------------
parray_aprod - Find product of elements in a Pyvox array


This method finds the product of the elements in the Pyvox array.
-------------------------------------------------------------------------*/

static PyObject *
parray_aprod(
    PyObject *self,           /* Voxel array to find product of */
    PyObject *args)           /* Arguments: none expected */
{
    PyObject *result;         /* Result: product of elements */

    voxel_array *vself;       /* Input volume image */
    double res;               /* Result as a number */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    vself  = PyVoxel_GetArray(self);
  
    /* Compute and return the result */
    res = vxl_aprod(vself);
    result = Py_BuildValue("d", res);
    return result;
}



/*-------------------------------------------------------------------------
parray_bigob - Remove all but the biggest object in a binary image

This instance method for Pyvox arrays aggregates all the non-zero
voxels of self into objects (maximal connected sets of non-zero voxels),
counts the number of voxels in each object, and returns an unsigned
char mask covering only the largest object.  (FIXME: Suppose there are
two or more biggest objects?)

NOTE:  This is a temporary version for testing and interim use.  It will
probably be replaced by several functions for doing individual steps of
the process, once I figure out exactly how to map object lists and 
operations on them into Python.
-------------------------------------------------------------------------*/

static PyObject *        /* Result: Volume image of biggest object only */
parray_bigob(
    PyObject *self,      /* Self: Volume image to process */
    PyObject *args)
{
    int lab1 = 255;      /* Arg 1: (Optional) label for biggest object */
    int lab2 = 0;        /* Arg 2: (Optional) label for all other objects */

    pyvox_array *in, *out;    /* Input and output as Pyvox arrays */
    voxel_array  *vin, *vout; /* Input and output as voxel arrays */
    voxel_array  *labels;     /* Image of voxel labels */
    voxel_array  *table;      /* Look up table to convert labels to mask */
    unsigned char *td;        /* Data contained in table */
    struct voxel_object_data *oblist;   /* Table of objects found */
    long nobs;                /* Number of objects found */
    int ibrain;               /* Canonical object id of brain */
    long vbrain;              /* Number of voxels in brain object */
    int i;
#if 0      /* Needed only for debugging output */
    int ncan;                 /* Number of canonical objects found */
    int n;
#endif

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|ii", &lab1, &lab2) )
	return NULL;

    /* Find and verify the input voxel array itself */
    if ( !PyvoxArray_Check(self, &vin) )
	ReturnError("Self is not a Pyvox array");
    in = (pyvox_array *)self;

    /* Scan the input to do preliminary labels and create the object table */
    labels = vxl_create_array();
    nobs = vxl_find_objects_uchar3d(labels, &oblist, vin);

    /* DEBUG: Report the first 50 objects found */
#if 0
    fprintf(stderr, "\nFound %ld objects in this image\n", nobs);
    for (i = 0; i < 50 && i < nobs; i++) {
	fprintf(stderr, "%6hu (%6hu): %10ld voxels at (%4ld, %4ld, %4ld)\n",
		oblist[i].ident, oblist[i].canon, oblist[i].count,
		oblist[i].point[0], oblist[i].point[1], oblist[i].point[2]);
    }
#endif

    /* DEBUG: Report the first 50 large canonical objects found. */
#if 0
    ncan = 0;
    for (i = 0; i < nobs; i++)  
	if (oblist[i].canon == i)  ncan ++;
    fprintf(stderr, "\nFound %d canonical objects in this image\n",
	    ncan);
    for ( n = 0, i = 0;
	  n < 50 && i < nobs;
	  i++ ) {
	if (oblist[i].canon == i && oblist[i].count > 100) {
	    n++;
	    fprintf(stderr, 
		    "%6hu (%6hu): %10ld voxels at (%4ld, %4ld, %4ld)\n",
		    oblist[i].ident, oblist[i].canon, oblist[i].count,
		    oblist[i].point[0], oblist[i].point[1],
		    oblist[i].point[2]);
	}
    }
#endif

    /* Find the largest non-background object and call it brain.
       FIXME:  Suppose there are *two* largest objects? */
    ibrain = 0;
    vbrain = 0;
    for (i = 1; i < nobs; i++)
	if (oblist[i].count > vbrain) {
	    ibrain = oblist[i].canon;
	    vbrain = oblist[i].count;
	}
    if (ibrain == 0)
	fatal("Couldn't find any non-background objects");
    if (0)  /* DEBUG */
	fprintf(stderr, "The brain object has index %d\n", ibrain);

    /* Create a lookup table to convert brain to 255, others to 0 */
    table = vxl_create_array();
    vxl_alloc_array(table, INTP_UCHAR, 1, &nobs);
    td = table->data;
    td[0] = 0;
    for (i = 1; i < nobs; i++) {
	if (oblist[i].canon == ibrain)
	    td[i] = lab1;
	else
	    td[i] = lab2;
    }

    /* DEBUG:  Print first 50 entries of the lookup table */
#if 0
    for (i = 0; i < 50; i++)
	fprintf(stderr, "%5d -> %3d\n", i, td[i]);
#endif
    
    /* Create mask by processing labels image through the lookup table */
    out = parray_create();
    vout = out->varray;
    vxl_lookup(vout, labels, table);

    /* Free temporary storage */
    vxl_destroy_array(table);
    free(oblist);
    vxl_destroy_array(labels);

    return (PyObject *)out;
}



/*-------------------------------------------------------------------------
parray_bihist - Compute bivariate histogram of two unsigned char images

This instance method for Pyvox arrays computes the bivariate histogram
and marginal histograms of self and another Pyvox array, both of which
must be the same shape and type unsigned char.  An optional unsigned
char weight image may be provided, which must the the same shape as
the source image.  The bivariate histogram is returned as a
two-dimensional unsigned long voxel arrays; the two marginal
histograms (which are the univariate histograms of the individual
images) are returned as one-dimensional unsigned long voxel arrays.
These results are returned as a tuple containing the bivariate
histogram and the two univariate histograms.
-------------------------------------------------------------------------*/

static PyObject *
parray_bihist(
    PyObject *self,             /* Self: First image */
    PyObject *args)
{
    PyObject *src2;             /* Arg 1: Second image */
    PyObject *weight = 0;       /* Arg 2: Weight image */

    pyvox_array *hist;          /* Result 1: Bivariate histogram */
    pyvox_array *mar1;          /* Result 2: Univariate histogram of 1st img */
    pyvox_array *mar2;          /* Result 3: Univariate histogram of 2nd img */

    voxel_array *s1, *s2, *w = 0;  /* Arguments as voxel arrays */
    PyObject *result;              /* Python tuple containing the results */

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "O|O", &src2, &weight) )
	return NULL;
    if ( !PyvoxArray_Check(self, &s1) || s1->type != INTP_UCHAR )
	ReturnError("Self is not an unsigned char voxel array");
    if ( !PyvoxArray_Check(src2, &s2) || s2->type != INTP_UCHAR)
	ReturnError("Source 2 is not an unsigned char voxel array");
    if ( weight != 0 && weight != Py_None ) {
	if ( !PyvoxArray_Check(weight, &w) || w->type != INTP_UCHAR )
	    ReturnError("Weight is not an unsigned char voxel array");
    }

    /* Create Pyvox arrays for the output */
    hist = parray_create();
    mar1 = parray_create();
    mar2 = parray_create();

    /* Compute the histograms */
    vxl_bihist_uchar(hist->varray, mar1->varray, mar2->varray, s1, s2, w);

    /* Form results into a tuple and return them */
    result = Py_BuildValue("(NNN)", hist, mar1, mar2);
    return result;
}



/*-------------------------------------------------------------------------
parray_chamfer - Compute chamfer distance transform of an image

This instance method computes the chamfer distance image of the given
image.  The chamfer distance is defined to be zero wherever the original
image is non-zero, and to be equal to the taxicab (L2) distance to the
nearest non-zero voxel otherwise.  The original image may be of any
rank greater than zero, and any size and type.  The distance image may
be either unsigned char or unsigned short, as specified by the caller, or
may be defaulted to the shortest type able to contain the maximum possible
distance.
---------------------------------------------------------------------------*/

static PyObject *
parray_chamfer(
    PyObject *self,             /* Self: the original image */
    PyObject *args)
{
    int dtype = INTP_NONE;      /* Arg 1: Requested destination type */
    PyObject *result;           /* Result: Chamfer distance image */

    voxel_array *vself;         /* Self as a voxel array */
    voxel_array *vres;          /* Result as a voxel array */

    /* Fetch the argument, if any */
    if ( !PyArg_ParseTuple(args, "|i", &dtype) )
	return NULL;

    /* Find the input voxel array */
    if ( !PyvoxArray_Check(self, &vself) )
	ReturnError("Self is not a Pyvox array");

    /* Compute and return the result */
    result = (PyObject *)parray_create();
    vres   = ((pyvox_array *)result)->varray;
    vxl_chamfer(vres, vself, dtype);

    return result;
}



/*-------------------------------------------------------------------------
parray_convolve - Convolve an image with a convolution kernel

This instance method for Pyvox arrays convolves self, which may be of
any non-zero rank, dimensions, or type, with a convolution kernel of
the same rank, and optionally subsamples the resulting image using a
given integral shrink factor along each dimension.  (Pixel values not
included in the subsampled image are never calculated, of course.)
The calculations are done in double precision and converted back to
the original image type; if the original type cannot represent the
values, the nearest representable value is used instead.

FIXME: Add some other useful filters, such as high pass, gradient
(Sobel) and Laplacian.  (Note that the derivative filters require a
bias that depends on the data type.)  Add functions to construct
various useful kernels.
-------------------------------------------------------------------------*/

static PyObject *
parray_convolve(
    PyObject *self,             /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern;           /* Arg 1: Convolution kernel to use */
    PyObject *pyshrink=0;       /* Arg 2: Shrink factors */
    PyObject *output;           /* Result of convolution */

    voxel_array *in, *out;      /* The voxel arrays themselves */
    struct vxl_kernel *kern;    /* The kernel struct itself */
    long shrink[VXL_MAX_RANK];  /* Shrink factors */
    int i;

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O|O", &pykern, &pyshrink) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = ((pyvox_array *)self)->varray;

    /* Find the kernel struct itself */
    if (pykern->ob_type != &kernel_type)
	ReturnError("Kernel is not a Pyvox kernel object");
    kern = ((kernel_object *)pykern)->kern;
    if (in->rank != kern->rank)
	ReturnError("Image and kernel have different ranks");

    /* Find the shrink factors, if any */
    if ( pyshrink == 0 ) {
        for (i = 0; i < in->rank; i++)
            shrink[i] = 1;
    }
    else if ( PyvoxLong_Check(pyshrink, &shrink[0]) ) {
        for (i = 1; i < in->rank; i++)
            shrink[i] = shrink[0];
    }
    else if ( PyNumSeq_AsLong(pyshrink, shrink, in->rank) )
        /* Do nothing */ ;
    else
        ReturnError("Unable to interpret shrink argument");
  
    /* Compute and return the result */
    output = (PyObject *)parray_create();
    out = ((pyvox_array *)output)->varray;
    vxl_convolve(out, in, kern, shrink);

    return output;
}



/*-----------------------------------------------------------------------
parray_count - Total number of voxels in a Pyvox array

This Pyvox array method returns the total number of voxels in a Pyvox
array.
-------------------------------------------------------------------------*/

static
PyObject *
parray_count(
    PyObject *self,          /* Pyvox array */
    PyObject *args)          /* Arguments (none) */
{
    voxel_array *vself;      /* Self as a voxel array */
    long count;              /* Number of voxels */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if ( !vself )
	ReturnError("Self is not a Pyvox array");

    /* Determine and return the count */
    count = vxli_count(vself);
    return Py_BuildValue("l", count);
}



/*------------------------------------------------------------------------
parray_det - Determinant of a Pyvox array (considered as a matrix)

This Pyvox array method returns the determinant of a Pyvox array
considered as a matrix.
-------------------------------------------------------------------------*/

static
PyObject *
parray_det(
    PyObject *self,            /* Pyvox array */
    PyObject *args)            /* Arguments (none) */
{
    voxel_array *vself;        /* Self as a voxel array */
    double det;                /* Determinant of self */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");

    /* Compute determinant and package as a Python object */
    det = matrix_determinant(vself);
    return Py_BuildValue("d", det);
}



/*------------------------------------------------------------------------
parray_diag - Extract diagnonal elements of a rank-2 array

This instance method constructs a new rank-1 Pyvox array containing
the diagonal elements of the given rank-2 array; the type of the
result is the same as the type of the original array.  The inverse
operation of constructing a diagonal matrix from its diagonal elements
may be done using pyvox.diag().

NOTE: Prior to 2001-10-23 this function constructed a diagonal matrix
from a vector of the elements to be placed on the diagonal.  This is
an incompatible change, so beware of old users; the argument types,
however, are incompatible, so old uses should be detected automatically.
-------------------------------------------------------------------------*/

static
PyObject *
parray_diag(
    PyObject *self,        /* Pyvox array */
    PyObject *args)        /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    PyObject *res;          /* Result as a Python array */
    voxel_array *vres;      /* Result as a voxel array */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");

    /* Create Pyvox array for result */
    res  = (PyObject *)parray_create();
    vres = PyVoxel_GetArray(res);

    /* Construct the diagonal matrix */
    matrix_get_diag(vres, vself);

    /* Return the result as a Pyvox array */
    return Py_BuildValue("N", res);
}



/*-------------------------------------------------------------------------
parray_dilate - Morphological dilation of a volume image

This instance method for Pyvox arrays performs a morphological
dilation on self, which must be a volume image of non-zero rank and
any unsigned integral type.  The neighborbood is specified by a kernel
object and defaults to a centered 3x3x3 neighborhood.  The bias and
coefficients of the kernel, if any, are ignored.

FIXME:  Default to a neighborhood of the correct rank, computing it as
needed.
-------------------------------------------------------------------------*/

static PyObject *
parray_dilate(
    PyObject *self,               /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern = 0;         /* Arg 1: Kernel specifying neighborhood */

    pyvox_array *in, *out;        /* Input and output as Pyvox arrays */
    struct vxl_kernel *kern;      /* The kernel struct itself */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pykern) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = (pyvox_array *)self;

    /* Find the kernel struct, if specified */
    kern = &neigh3d27;
    if ( pykern != 0 && PyKernel_Check(pykern) ) {
	kern = ((kernel_object *)pykern)->kern;
    }
    if (kern->rank != in->varray->rank)
	ReturnError("Image and kernel must have the same rank");

    /* Compute and return the result */
    out = parray_create();
    vxl_dilate(out->varray, in->varray, kern);

    return (PyObject *)out;
}



/*--------------------------------------------------------------------------
parray_dot - Compute weighted vector dot product of two arrays

This function computes the vector dot product of two arrays,
optionally weighted by a pixelwise weight.  The source and weight
arrays may be of any type and are converted to double for this
computation; they must, however, be the same shape.  The weight array
is optional, defaulting to one.
---------------------------------------------------------------------------*/

static PyObject *
parray_dot(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    double result;           /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OO", &other, &weight) )
	return NULL;

    /* Extract voxel arrays */
    if ( !PyvoxArray_Check(self, &slf) )
	ReturnError("Self is not a Pyvox array");
    if ( other != 0 && other != Py_None && !PyvoxArray_Check(other, &oth) )
	ReturnError("Other is not a Pyvox array");
    if ( weight != 0 && weight != Py_None && !PyvoxArray_Check(weight, &wate) )
	ReturnError("Weight is not a Pyvox array");

    /* Call Voxel Kit to do the real work; return result */
    result = vxl_dot(slf, oth, wate);
    return Py_BuildValue("d", result);
}



/*-------------------------------------------------------------------------
parray_erode - Morphological erosion of a volume image

This instance method for Pyvox arrays performs a morphological erosion
on self, which must be a volume image of any non-zero rank and any
unsigned integral type, using an optional neighborhood specified as a
kernel object.  If omitted, the neighborhood defaults to a centered
3x3x3 neighborhood.  The bias and coefficients of the kernel, if any,
are ignored.

FIXME:  Default to a neighborhood of the correct rank, computing it as
needed.
-------------------------------------------------------------------------*/

static PyObject *
parray_erode(
    PyObject *self,                /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern = 0;          /* Arg 1: Neighborhood to use */

    pyvox_array *in, *out;         /* Input and output as Pyvox arrays */
    struct vxl_kernel *kern;       /* The kernel struct */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pykern) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = (pyvox_array *)self;
  
    /* Find the kernel struct, if specified */
    kern = &neigh3d27;
    if ( pykern != 0 && PyKernel_Check(pykern) ) {
	kern = ((kernel_object *)pykern)->kern;
    }
    if (kern->rank != in->varray->rank)
	ReturnError("Image and kernel must have the same rank");
  
    /* Compute and return the result */
    out = parray_create();
    vxl_erode(out->varray, in->varray, kern);

    return (PyObject *)out;
}



/*----------------------------------------------------------------------
parray_eigsy - Solve real symmetric eigensystem

This Pyvox array method returns a pair (val, vec) containing the
eigenvalues and eigenvectors of the Pyvox array considered as a real
symmetric matrix.
-----------------------------------------------------------------------*/

static 
PyObject *
parray_eigsy(
    PyObject *self,            /* Pyvox array */
    PyObject *args)            /* Arguments (none) */
{
    voxel_array *vself;        /* Self as a voxel array */
    PyObject *val;             /* Eigenvalues */
    voxel_array *vval;         /* Eigenvalues as voxel array */
    PyObject *vec;             /* Eigenvectors */
    voxel_array *vvec;         /* Eigenvectors as voxel array */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");

    /* Create voxel arrays for values and vectors */
    val  = (PyObject *)parray_create();
    vval = PyVoxel_GetArray(val);
    vec  = (PyObject *)parray_create();
    vvec = PyVoxel_GetArray(vec);

    /* Solve the eigensystem */
    matrix_eigen_symmetric(vval, vvec, vself);

    /* Return pair of eigenvalues and eigenvectors */
    return Py_BuildValue("(NN)", val, vec);
}



/*------------------------------------------------------------------------
parray_fill2d - Fill contour in a 2D image

This instance method for two-dimensional Pyvox arrays fills with a
specified value all the voxels that lie within a contour given by a
set of vertices.

FIXME: The format for the vertices is currently very restrictive; generalize
it to include a variety of useful formats.
-------------------------------------------------------------------------*/

static PyObject *
parray_fill2d(
    PyObject *self,           /* Self: Array to draw into */
    PyObject *args)
{
    PyObject *ppoints;        /* Arg 1: Vertices defining the contour */
    PyObject *pvalue;         /* Arg 2: Value to fill with */

    voxel_array *canvas;      /* Voxel array to draw into */
    voxel_array *points;      /* Vertices defining the contour */
    long npoints;             /* Number of vertices */
    double value;             /* Value to fill with */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "OO", &ppoints, &pvalue) )
	return NULL;

    /* Find the voxel array to draw into */
    if ( !PyvoxArray_Check(self, &canvas) )
	ReturnError("Not a Pyvox array");

    /* Find the vertices defining the contour */
    if  ( !PyvoxArray_Check(ppoints, &points) ||
	  points == 0 || points->magic != VXL_MAGIC ||
	  points->data == 0 || points->type != INTP_DOUBLE ||
	  points->rank != 2 || points->dimen[1] != 2
	)
	ReturnError("Vertices must be given as an N by 2 Pyvox array");
    npoints = points->dimen[0];

    /* Find the value to fill with */
    if ( !PyvoxDouble_Check(pvalue, &value) )
	ReturnError("Fill value is incompatible type");

    /* Call the Voxel Kit to do the real work */
    vxl_fill_2d_contour(canvas, npoints, points->data, value);

    /* Return the modified array */
    Py_INCREF(self);
    return self;
}



/*-------------------------------------------------------------------------
parray_histo - Compute univariate histogram of an unsigned char image

This instance method for Pyvox arrays computes the univariate
histogram of self, which must be an unsigned char image, and returns
it as a one-dimensional unsigned long voxel array.  An optional weight
array may be provided, which must be the same shape and type as self.

FIXME:  Implemented only for unsigned char images.
-------------------------------------------------------------------------*/

static PyObject *
parray_histo(
    PyObject *self,             /* Self: Image to take histogram of */
    PyObject *args)
{
    PyObject *weight = 0;       /* Arg 1: Weight image */
    PyObject *hist;             /* Result: Univariate histogram */

    voxel_array *vself;         /* Self as a voxel array */
    voxel_array *vweight = 0;   /* Weight as a voxel array */
    voxel_array *vhist;         /* Histogram as a voxel array */

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &weight) )
	return NULL;
    if ( !PyvoxArray_Check(self, &vself) || vself->type != INTP_UCHAR )
	ReturnError("Self is not an unsigned char voxel array");
    if ( weight != 0 && weight != Py_None )
	if ( !PyvoxArray_Check(weight, &vweight) ||
	     vweight->type != INTP_UCHAR )
	    ReturnError("Weight is not an unsigned char voxel array");

    /* Compute and return the histogram */
    hist = (PyObject *)parray_create();
    vhist = ((pyvox_array *)hist)->varray;
    vxl_histo_uchar(vhist, vself, vweight);

    return (PyObject *)hist;
}



/*-------------------------------------------------------------------------
parray_i2p - Compute the index-to-physical transform

This function returns the affine transform which converts from index to
physical coordinates for this array, based on the origin and spacing
attributes.

ALGORITHM: This function makes an upcall to the Python level, where the
affine transform class is defined.
-------------------------------------------------------------------------*/

static PyObject *
parray_i2p(
    PyObject *self,             /* Array to get transform of */
    PyObject *args)             /* Args: None used */
{
    voxel_array *vself;         /* Self as a voxel array */
    int rank;                   /* Rank of the array */
    long dimen[2];              /* Dimensions of matrix and offset */
    PyObject *mtx;              /* Matrix of the transform */
    voxel_array *vmtx;          /* Matrix, as a voxel array */
    double *mdata;              /* Data of the matrix */
    PyObject *off;              /* Offset of the transform */
    voxel_array *voff;          /* Offset, as a voxel array */
    double *odata;              /* Data of the offset */
    PyObject *affine;           /* Affine class object */
    PyObject *result;           /* The transform to be returned as result */
    int i;

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the array itself */
    if ( !PyvoxArray_Check(self, &vself) )
	ReturnError("Not a Pyvox array");
    rank = vself->rank;

    /* Construct arrays for the matrix and offset of the transform */
    dimen[0] = dimen[1] = rank;
    mtx   = parray_alloc(INTP_DOUBLE, 2, dimen);
    vmtx  = PyVoxel_GetArray(mtx);
    mdata = (double *)vmtx->data;
    dimen[1] = 1;
    off   = parray_alloc(INTP_DOUBLE, 2, dimen);
    voff  = PyVoxel_GetArray(off);
    odata = (double *)voff->data;

    /* Fill matrix and offset with the right values */
    for (i = 0; i < rank*rank; i++)
	mdata[i] = 0.0;
    for (i = 0; i < rank; i++) {
	odata[i] = vself->origin[i];
	mdata[i*rank+i] = vself->spacing[i];
    }

    /* Construct and return the affine transform */
    affine = PyDict_GetItemString(pyvox_dict, "affine");
    result = PyObject_CallFunction(affine, "OO", mtx, off);

    Py_INCREF(result);
    return result;
}



/*--------------------------------------------------------------------------
parray_kmeans1 - Compute univariate K-means from a histogram

Assuming that self is the univariate histogram of an unsigned char
image and given initial guesses at class centroids, this instance
method for Pyvox arrays uses the K-means algorithm to refine the class
centroids.  The centroids thus computed can be used later by the
array.nnclass1 function to do the actual segmentation.  If initial
guesses are not available for the centroids, they should all be set to
zero.  The number of classes found is set by the number of centroids
provided.

FIXME: The initial centroids could be made optional, but only if I
provide some other means of specifying the number of classes to be
found.  One possibility is to consider a naked number rather than
as list as the number of classes to be found.

FIXME:  It's probably reasonable to allow the histogram to be something
other than a Pyvox array.  A numeric sequence, for instance.  For that
matter, a rank-1 Pyvox array could also reasonably be regarded as a
sequence and support all the usual sequence operations.
---------------------------------------------------------------------------*/

static PyObject *
parray_kmeans1(
    PyObject *self,                 /* Self: Image histogram */
    PyObject *args)
{
    PyObject *centin = 0;           /* Arg 1: Initial centroids */
    PyObject *centout;              /* Result: Final centroids */

    const char centroid_error[] =   /* Error: Invalid centroid argument */
	"Initial centroid list must be a sequence of values in 0..255";
    pyvox_array *vohist;            /* hist as a Pyvox array */
    voxel_array  *vahist;           /* Voxel array that hist refers to */
    unsigned char *cent;            /* Working copy of centroids */
    int nclass;                     /* Number of classes to find */
    int i;
    long val;
    PyObject *ob;

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "O", &centin) )  
	return NULL;

    /* Find the voxel array that the hist argument refers to */
    vohist = (pyvox_array *)self;
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    vahist = vohist->varray;

    /* Convert the centroids to a format usable in C */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(centin) )
	ReturnError(centroid_error);
    nclass = PyObject_Length(centin);
    cent = mallock(nclass*sizeof(unsigned char));
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(centin, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnError(centroid_error);
	if ( val < 0 || val > 255)
	    ReturnError(centroid_error);
	cent[i] = val;
    }

    /* Run the K-means algorithm */
    vxl_kmeans1_histo(vahist, nclass, cent);

    /* Convert the updated centroids to a Python object and return it. */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the tuple containing them, but
       the Python API manual is somewhat obscure on this point. */
    centout = PyList_New(nclass);
    for (i = 0; i < nclass; i++) {
        ob = PyInt_FromLong(cent[i]);
	PyList_SetItem(centout, i, ob); }

    free(cent);
    return centout;
}



/*------------------------------------------------------------------------
parray_inverse - Compute inverse of a matrix

This instance method computes the inverse of a non-singular square
matrix represented as a Pyvox array of type float or double.

FIXME: Generate a Python exception if the matrix is singular.
-------------------------------------------------------------------------*/

static
PyObject *
parray_inverse(
    PyObject *self,         /* Pyvox array */
    PyObject *args)         /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    PyObject *inv;          /* Inverse as a Python array */
    voxel_array *vinv;      /* Inverse as a voxel array */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");

    /* Create Pyvox array for result */
    inv  = (PyObject *)parray_create();
    vinv = PyVoxel_GetArray(inv);

    /* Compute the inverse */
    matrix_inverse(vinv, vself);

    /* Return the result as a Pyvox array */
    return Py_BuildValue("N", inv);
}



/*------------------------------------------------------------------------
parray_linear0 - Linear interpolation between pixels (reference version)

This instance method for Pyvox arrays computes the approximate pixel
intensity between sample values using linear interpolation.  Pixel
values outside the array are assumed to be zero.  This implementation
is slow but simple enough to understand; it is intended as a reference
version for testing faster but less comprehensible implementations.
-------------------------------------------------------------------------*/

static PyObject *
parray_linear0(
    PyObject *self,                 /* Pyvox array to interpolate in */
    PyObject *args)                 /* Other arguments */
{
    PyObject *poynt;                /* Arg 1: Point to evaluate at */

    voxel_array *vself;             /* Self as voxel array */
    int rank;                       /* Rank of self */
    double point[VXL_MAX_RANK];     /* Point to evaluate at */
    double value = 0.0;             /* Value at the given point */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");
    rank = vself->rank;

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "O", &poynt) )  
	return NULL;
    if ( !PyvoxDoubleSeq_Check(poynt, point, rank) )
	 ReturnError("Dimensions of array and point do not match");

    /* Evaluate at the given point */
    value = vxl_linear0(vself, point);

    /* Return the result as a Python float */
    return Py_BuildValue("d", value);
}



/*------------------------------------------------------------------------
parray_linear - Linear interpolation between pixels

This instance method for Pyvox arrays computes the approximate pixel
intensity between sample values using linear interpolation.  Pixel
values outside the array are assumed to be zero.

This calls the fast implementation of the linear interpolation function.
-------------------------------------------------------------------------*/

static PyObject *
parray_linear(
    PyObject *self,                 /* Pyvox array to interpolate in */
    PyObject *args)                 /* Other arguments */
{
    PyObject *poynt;                /* Arg 1: Point to evaluate at */

    voxel_array *vself;             /* Self as voxel array */
    int rank;                       /* Rank of self */
    double point[VXL_MAX_RANK];     /* Point to evaluate at */
    double value = 0.0;             /* Value at the given point */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");
    rank = vself->rank;

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "O", &poynt) )  
	return NULL;
    if ( !PyvoxDoubleSeq_Check(poynt, point, rank) )
	 ReturnError("Dimensions of array and point do not match");

    /* Evaluate at the given point */
    value = vxl_linear(vself, point);

    /* Return the result as a Python float */
    return Py_BuildValue("d", value);
}



/*--------------------------------------------------------------------------
parray_list - Return array elements as a list

This instance method for Pyvox arrays returns the elements of the array
(in row-major order) as a Python list of floats, ints, or longs.  See also 
parray_tuple, which returns a tuple rather than a list.
--------------------------------------------------------------------------*/

static PyObject *
parray_list(
    PyObject *self,            /* Pyvox array to enlist */
    PyObject *args)            /* Arguments (none in this case) */
{
    PyObject *result;          /* List to return as result */
    voxel_array *vself;        /* Self as a voxel array */
    int type;                  /* Type of voxel array */
    void *data;                /* Data contained in voxel array */
    long count;                /* Number of elements in array */
    double value;
    long i;

    /* Find voxel array of self; count number of elements */
    vself = PyVoxel_GetArray(self);
    if ( !vself )
	ReturnError("Not a Pyvox array");
    type  = vself->type;
    data  = vself->data;
    count = vxli_count(vself);

    /* Check other arguments; there should be none */
    if ( !PyArg_ParseTuple(args, "") )  
	return NULL;

    /* If an integral array, construct a list of Python ints and longs */
    result = PyList_New(count);
    if ( exim_integral_type(type) ) {
	for (i = 0; i < count; i++) {
	    value = exim_get_value(data, type, i);
	    if ( value <= (double)LONG_MAX )
		PyList_SetItem(result, i, PyInt_FromLong((long)value));
	    else
		PyList_SetItem(result, i, PyLong_FromDouble(value));
	}
    }

    /* Otherwise construct a list of Python floats */
    else {
	for (i = 0; i < count; i++) {
	    value = exim_get_value(data, type, i);
	    PyList_SetItem(result, i, PyFloat_FromDouble(value));
	}
    }

    return result;
}



/*-------------------------------------------------------------------------
parray_lostat - Compute local statistics of a volume image

This instance method for Pyvox arrays computes the local mean and
standard deviation within each 3x3x3 neighborhood of self and returns
them as a list containing two new images of the same shape and type;
the first is the local mean and the second is the local standard
deviation.  Boundary voxels are handled correctly.  The standard
deviation is multiplied by 2.0 for images of type exim.uchar, to
better match the possible range of values to the range supported by
unsigned char; the scale is left at 1.0 for all other data types.

FIXME:  The underlying voxel kit function supports any type, rank,
kernel, and scale; generalize this function to access those capabilities.

FIXME: Provide means to support users who want only the local mean or
local variance?  This might not be particularly useful, though.  The
local mean could be computed as a convolution, but the local variance
requires that the local mean be computed (but perhaps never stored?).
-------------------------------------------------------------------------*/

static PyObject *
parray_lostat(
    PyObject *self,           /* Self: Volume image to take statistics of */
    PyObject *args)
{
    PyObject *result;         /* Result: List containing two volume images */

    PyObject *mean, *sdev;    /* The mean and std dev images */
    voxel_array *in, *m, *d;  /* Voxel arrays inside the Pyvox objects */

    struct vxl_kernel *kern;  /* Neighborhood definition */
    double scale;             /* Scale factor for std dev */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = ((pyvox_array *)self)->varray;
    if (in->rank != 3)
	ReturnError("Sorry, only 3d has been implemented yet.");
  
    /* Create Pyvox arrays for the output */
    mean = (PyObject *)parray_create();
    sdev = (PyObject *)parray_create();
    m = ((pyvox_array *)mean)->varray;
    d = ((pyvox_array *)sdev)->varray;

    /* Set parms needed by Voxel Kit but not yet supported by Pyvox */
    kern = &neigh3d27;
    if (in->type == INTP_UCHAR)
	scale = 2.0;
    else
	scale = 1.0;

    /* Compute and return the result */
    vxl_lostat(m, d, in, kern, scale);
    result = Py_BuildValue("[NN]", mean, sdev);
    return result;
}



/*-------------------------------------------------------------------------
parray_lowpass - Lowpass filter a volume image

This instance method for Pyvox arrays lowpass filters and optionally
subsamples self, which must be a 3d volume image of any type.  The
kernel is a 3x3x3 convolution kernel of the form 2^-(3+|x|+|y|+|z|);
this kernel will completely surpress the Nyqist frequency along any of
the three principal axes.  If the shrink argument is provided, then
pixel values are computed for only the subsample with that spacing.

FIXME: Make the default bandwidth depend on the subsampling interval.

FIXME: Future versions of this function may operate in any number of
dimensions, and permit any desired bandwidth.

FIXME: Should this function be removed, in lieu of a function that
generates a lowposs kernel in any specified number of dimensions?
-------------------------------------------------------------------------*/

static PyObject *
parray_lowpass(
    PyObject *self,             /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pyshrink = 0;      /* Arg 1: Desired shrink factors */

    pyvox_array *in, *out;       /* Input and output as Pyvox arrays */
    long shrink[VXL_MAX_RANK];   /* Shrink factors */
    int i;

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pyshrink) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = (pyvox_array *)self;
  
    /* Find the shrink factors, if any */
    if ( pyshrink == 0 ) {
        for (i = 0; i < in->varray->rank; i++)
            shrink[i] = 1;
    }
    else if ( PyvoxLong_Check(pyshrink, &shrink[0]) ) {
        for (i = 1; i < in->varray->rank; i++)
            shrink[i] = shrink[0];
    }
    else if ( PyNumSeq_AsLong(pyshrink, shrink, in->varray->rank) )
        /* Do nothing */ ;
    else
        ReturnError("Unable to interpret shrink argument");
  
    /* Compute and return the result */
    out = parray_create();
    vxl_convolve(out->varray, in->varray, &lowpass3d, shrink);

    return (PyObject *)out;

}



/*--------------------------------------------------------------------------
parray_mean - Compute mean along specified axes

This function computes the mean along specified axes to yield an array or
scalar of reduced rank.  The source and wight arrays may be of any type;
the result is always of type double.

FIXME: Check here that arrays are the same shape, so we get a Python
exception rather than a C crash?  Better yet, array for fatal() to
generate a Python exception.
---------------------------------------------------------------------------*/

static PyObject *
parray_mean(
    PyObject *self,             /* Self: array to take means of */
    PyObject *args)
{
    PyObject *pweight = 0;      /* Arg 1: weight array */
    PyObject *paxes   = 0;      /* Arg 2: axes to reduce on */

    pyvox_array *mean = 0;      /* Result: mean along given axes */

    voxel_array *vself  = 0;    /* Self as a voxel array */
    voxel_array *weight = 0;    /* Weight as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|OO", &pweight, &paxes) )
	return NULL;

    /* Extract voxel arrays */
    if  ( ! PyvoxArray_Check(self, &vself) )
	ReturnError("Self is not a Pyvox array");
    if  ( ! (pweight == 0) &&
	  ! (pweight == Py_None) && 
	  ! PyInt_Check(pweight) &&
	  ! PyFloat_Check(pweight) &&
	  ! PyvoxArray_Check(pweight, &weight) )
	ReturnError("Weight is not a Pyvox array");

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnError("Invalid axes specification");

    /* Call Voxel Kit to do the real work and return result */
    mean = parray_create();
    vxl_mean(mean->varray, vself, weight, naxes, axes);
    if ( mean->varray->rank > 0 )
	return (PyObject *)mean;
    else 
	return Py_BuildValue("d", vxl_get_voxel(mean->varray, 0));
}



/*----------------------------------------------------------------------
parray_mmul - Multiply self by another matrix

This Pyvox array method returns a new array which is the product of
self and another matrix.  As a special case, if the other matrix is a
rank-1 vector of the appropriate length, it will be treated as a
column vector.

FIXME: The current version requires that both arrays be the same type.

FIXME: This peeks inside the voxel array structure; decide if this is
okay.
-----------------------------------------------------------------------*/

static 
PyObject *
parray_mmul(
    PyObject *self,            /* Voxel array */
    PyObject *args)            /* Arguments */
{
    PyObject *other;           /* Arg 1: Matrix to multiply self by */
    PyObject *result;          /* Result: Product of matrices */

    voxel_array *vself;        /* Self as a voxel array */
    voxel_array *vother;       /* Other as a voxel array */
    voxel_array *vresult;      /* Result as a voxel array */
    int type1, type2;          /* Types of the two voxel arrays */
    long dimen1[2];            /* Dimensions of self */
    long dimen2[2];            /* Dimensions of other */
    long dimen3[2];            /* Dimensions of result */
    int m, n, k;               /* Matrix dimensions */
    int lda, ldb, ldc;         /* Matrix last dimensions */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a voxel array");

    /* Check that self is 2-D float or double */
    if (vself->rank != 2)
	ReturnError("Rank of alleged matrix is not 2");
    type1 = vself->type;
    if (type1 != INTP_FLOAT && type1 != INTP_DOUBLE)
	ReturnError("Type of matrix is neither float nor double");
    dimen1[0] = vself->dimen[0];
    dimen1[1] = vself->dimen[1];

    /* Check that other is compatible with self */
    if ( !PyArg_ParseTuple(args, "O", &other) )
	return NULL;
    vother = PyVoxel_GetArray(other);
    if ( !vother )
	ReturnError("Other is not a voxel array");
    type2 = vother->type;
    if (type2 != INTP_FLOAT && type2 != INTP_DOUBLE)
	ReturnError("Type of matrix is neither float nor double");
    if (type1 != type2)
	ReturnError("Array types are not compatible");
    if ( vother->rank == 2 && vother->dimen[0] == dimen1[1] ) {
	dimen2[0] = vother->dimen[0];
	dimen2[1] = vother->dimen[1]; }
    else if ( vother->rank == 1 && vother->dimen[0] == dimen1[1] ) {
	dimen2[0] = vother->dimen[0];
	dimen2[1] = 1; }
    else
	ReturnError("Matrix shapes are not compatible");
    
    /* Create voxel array for result */
    dimen3[0] = dimen1[0];
    dimen3[1] = dimen2[1];
    result = parray_alloc(type1, 2, dimen3);
    vresult = PyVoxel_GetArray(result);

    /* Multiply the matrices */
    ldc =       m = dimen1[0];
    ldb = ldc = n = dimen2[1];
    lda =       k = dimen1[1];
    switch (type1) {

    case INTP_FLOAT:
	clap_sgemm(0, 0, m, n, k, 1.0, vself->data, lda,
		   vother->data, ldb, 0.0, vresult->data, ldc);
	break;

    case INTP_DOUBLE:
	clap_dgemm(0, 0, m, n, k, 1.0, vself->data, lda,
		   vother->data, ldb, 0.0, vresult->data, ldc);
	break;

    default:
	ReturnError("Matrix multiplication supported in float or double only");

    }

    /* Return product of matrices */
    return Py_BuildValue("N", result);
}



/*-------------------------------------------------------------------------
parray_moments - Compute moments of a volume image

This Pyvox function computes the total mass, center of gravity, and
second central moments of a volume image; the value returned is a tuple
containing the total mass (as a Python float), the center of gravity
(as a voxel array), and the moments (as a voxel array).  The center of 
gravity and moments are in physical units, as defined by the origin and 
spacing of the volume image.

FIXME: This is just a temporary version; the final version will also
compute the principal axes and return values through a class object
rather than a tuple.  Well, maybe.  After using this version for a
while, it seems a perfectly natural interface.
-------------------------------------------------------------------------*/

static PyObject *
parray_moments(
    PyObject *self,           /* Voxel array to find moments of */
    PyObject *args)           /* Arguments: none expected */
{
    PyObject *result;         /* Result: List of mass, cg, and moments */

    voxel_array *vself;       /* Input volume image */
    PyObject *center;         /* Center of gravity */
    PyObject *moments;        /* Central moments */
    voxel_array *vcg, *vcm;   /* Center of gravity and moments */
    double mass, *cg, *cm;    /* Total mass, center of gravity, moments */
    int rank;                 /* Rank of the input voxel array */
    long dim[2];              /* Dimensions of moments array */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    vself  = PyVoxel_GetArray(self);
    rank   = vself->rank;
    dim[0] = dim[1] = rank;
  
    /* Create Pyvox arrays for the output */
    /* FIXME: Verify that allocation succeeds? */
    center  = parray_alloc(INTP_DOUBLE, 1, dim);
    moments = parray_alloc(INTP_DOUBLE, 2, dim);
    vcg = ((pyvox_array *)center) ->varray;
    vcm = ((pyvox_array *)moments)->varray;
    cg = vcg->data;
    cm = vcm->data;

    /* Compute and return the result */
    vxl_moments(&mass, cg, cm, vself);
    result = Py_BuildValue("[dNN]", mass, center, moments);
    return result;
}



/*-----------------------------------------------------------------------
parray_mop - Arbitrary moments of a product of arrays

This instance methed computes arbitrary moments of the elementwise product
of one, two, or three arrays.
-------------------------------------------------------------------------*/

static PyObject *
parray_mop(
    PyObject *self,             /* Self: First array */
    PyObject *args)             /* Other arguments */
{
    PyObject *pmoments=0;       /* Arg 1: moments requested */
    PyObject *parray2=0;        /* Arg 2: second array */
    PyObject *parray3=0;        /* Arg 3: third array */

    pyvox_array *presult=0;     /* Result: computed moments */

    voxel_array *array1=0;      /* First array */
    voxel_array *array2=0;      /* Second array */
    voxel_array *array3=0;      /* Third array */
    voxel_array *moments=0;     /* Moments requested */
    voxel_array *result=0;      /* Result */

    /* Find the voxel array contained in self */
    if ( !PyvoxArray_Check(self, &array1) )
	ReturnError("Self isn't a Pyvox array");

    /* Parse the argument tuple */
    if ( !PyArg_ParseTuple(args, "O|OO", &pmoments, &parray2, &parray3) )
	return NULL;

    /* Extract the voxel arrays contained in the arguments */
    if ( !PyvoxArray_Check(pmoments, &moments) )
	ReturnError("Requested moments not a Pyvox array");
    if ( parray2 != 0 && !PyvoxArray_Check(parray2, &array2) )
	ReturnError("Array 2 is not a Pyvox array");
    if ( parray3 != 0 && !PyvoxArray_Check(parray3, &array3) )
	ReturnError("Array 3 is not a Pyvox array");

    /* Create a new Pyvox array for the result */
    presult = parray_create();
    result  = presult->varray;

    /* Call the Voxel Kit to do the real work */
    vxl_mop(result, moments, array1, array2, array3);

    /* Return the computed moments */
    return (PyObject *)presult;
}



/*------------------------------------------------------------------------
parray_nearest - Nearest neighbor interpolation within an array

This instance method for Pyvox arrays computes the approximate pixel
intensity between sample values using nearest neighbor interpolation.
Pixel values outside the array are assumed to be zero.
-------------------------------------------------------------------------*/

static PyObject *
parray_nearest(
    PyObject *self,                 /* Pyvox array to interpolate in */
    PyObject *args)                 /* Other arguments */
{
    PyObject *poynt;                /* Arg 1: Point to evaluate at */

    voxel_array *vself;             /* Self as voxel array */
    int rank;                       /* Rank of self */
    double point[VXL_MAX_RANK];     /* Point to evaluate at */
    double value = 0.0;             /* Value at the given point */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");
    rank = vself->rank;

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "O", &poynt) )  
	return NULL;
    if ( !PyvoxDoubleSeq_Check(poynt, point, rank) )
	 ReturnError("Dimensions of array and point do not match");

    /* Evaluate at the given point */
    value = vxl_nearest(vself, point);

    /* Return the result as a Python float */
    return Py_BuildValue("d", value);
}



/*------------------------------------------------------------------------
parray_nnclass1 - Univariate nearest neighbor classification

Assuming that self is an unsigned char volume image and given a set of
class centroids, this instance method for Pyvox arrays finds the
nearest class centroid for each voxel position and outputs the
corresponding class id to yield a class image.  Different centroids
may be assigned to the same class number.  There must be exactly as
many class ids as class centroids.

FIXME: The current calling sequence was adopted because it was fairly
easy to implement quickly, but it's not clear that it's the best
option; so it might change in the future.
-------------------------------------------------------------------------*/

static PyObject *
parray_nnclass1(
    PyObject *self,             /* Self: Feature image to classify */
    PyObject *args)
{
    PyObject *pyclid;           /* Arg 1: Class ids as Python sequence */
    PyObject *pycent;           /* Arg 2: Class centroids as Python sequence */
    PyObject *pyclass;          /* Result: Image of voxel classes */
    const char cent_error[] =   /* Error: Invalid centroid argument */
	"Initial centroids must be a sequence of values in 0..255";
    const char clid_error[] =   /* Error: Invalid class ids argument */
	"Class ids must be a sequence of values in 0..255";
    const char match_error[] =  /* Error: Clids and centroids don't match */
	"Different numbers of class ids and centroids";

    pyvox_array *vosrc;         /* Feature image as a Pyvox array */
    voxel_array  *vasrc;        /* Feature image as a voxel array */
    pyvox_array *voclass;       /* Class image as a Pyvox array */
    voxel_array  *vaclass;      /* Class image as a voxel array */
    int nclass;                 /* Number of classes */
    unsigned char *clid;        /* Array of class ids */
    unsigned char *cent;        /* Array of class centroids */

    int i;
    long val;
    PyObject *ob;

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "OO", &pyclid, &pycent) )
	return NULL;

    /* Verify the source image */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    vosrc = (pyvox_array *)self;
    vasrc = vosrc->varray;
    if ( vasrc->type != INTP_UCHAR)
	ReturnError("Sorry, only uchar has been implemented yet.");

    /* Convert the class ids */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pyclid) )
	ReturnError(clid_error);
    nclass = PyObject_Length(pyclid);
    clid = mallock(nclass*sizeof(unsigned char));
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pyclid, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnError(clid_error);
	if ( val < 0 || val > 255)
	    ReturnError(clid_error);
	clid[i] = val;
    }

    /* Convert the class centroids */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pycent) )
	ReturnError(cent_error);
    if ( PyObject_Length(pycent) != nclass )
	ReturnError(match_error);
    cent = mallock(nclass*sizeof(unsigned char));
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pycent, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnError(cent_error);
	if ( val < 0 || val > 255)
	    ReturnError(cent_error);
	cent[i] = val;
    }

    /* Call Voxel Kit to do the real work */
    voclass = parray_create();
    vaclass = voclass->varray;
    vxl_nnclass1_uchar(vaclass, vasrc, nclass, clid, cent);

    /* Free working storage and return the result */
    free(clid);
    free(cent);
    pyclass = (PyObject *)voclass;
    return pyclass;
}



/*------------------------------------------------------------------------
parray_nnclass2 - Bivariate nearest neighbor classification

Assuming that self is a unsigned char volume image and given another
unsigned char volume image and a set of class centroids, this instance
method for Pyvox arrays finds the nearest class centroid for each
voxel position and outputs the corresponding class id to yield a class
image.  Different centroids may be assigned to the same class id.
There must be exactly as many class ids as class centroids.

The current calling sequence was adopted because it was fairly easy to
implement quickly, but it's not clear that it's the best option; so it
might change in the future.
-------------------------------------------------------------------------*/

static PyObject *
parray_nnclass2(
    PyObject *self,            /* Arg 1: First feature image */
    PyObject *args)
{
    PyObject *pysrc2;           /* Arg 2: Second feature image */
    PyObject *pyclid;           /* Arg 3: Class ids as Python sequence */
    PyObject *pycent1;          /* Arg 4: Class centroids in 1st image */
    PyObject *pycent2;          /* Arg 5: Class centroids in 2nd image */
    PyObject *pyclass;          /* Result: Image of voxel classes */

    pyvox_array *vosrc1, *vosrc2;    /* Feature images as Pyvox arrays */
    voxel_array  *vasrc1, *vasrc2;   /* Feature images as a voxel array */
    pyvox_array *voclass;            /* Class image as a Pyvox array */
    voxel_array  *vaclass;           /* Class image as a voxel array */
    int nclass;                      /* Number of classes */
    unsigned char *clid;             /* Array of class ids */
    unsigned char *cent1, *cent2;    /* Arrays of class centroids */

    int i;
    long val;
    PyObject *ob;

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &pysrc2, 
			   &pyclid, &pycent1, &pycent2) )
	return NULL;

    /* Verify the source images */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError(voxel_error);
    vosrc1 = (pyvox_array *)self;
    vasrc1 = vosrc1->varray;
    if ( vasrc1->type != INTP_UCHAR)
	ReturnError(uchar_error);

    if ( !PyvoxArray_Check(pysrc2, 0) )
	ReturnError(voxel_error);
    vosrc2 = (pyvox_array *)pysrc2;
    vasrc2 = vosrc2->varray;
    if ( vasrc2->type != INTP_UCHAR)
	ReturnError(uchar_error);

    /* Convert the class ids */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pyclid) )
	ReturnError(clid_error);
    nclass = PyObject_Length(pyclid);
    clid = mallock(nclass*sizeof(unsigned char));
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pyclid, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnError(clid_error);
	if ( val < 0 || val > 255)
	    ReturnError(clid_error);
	clid[i] = val;
    }

    /* Convert the class centroids for first feature image */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pycent1) )
	ReturnError(cent_error);
    if ( PyObject_Length(pycent1) != nclass )
	ReturnError(match_error);
    cent1 = mallock(nclass*sizeof(unsigned char));
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pycent1, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnError(cent_error);
	if ( val < 0 || val > 255)
	    ReturnError(cent_error);
	cent1[i] = val;
    }

    /* Convert the class centroids for second feature image */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pycent2) )
	ReturnError(cent_error);
    if ( PyObject_Length(pycent2) != nclass )
	ReturnError(match_error);
    cent2 = mallock(nclass*sizeof(unsigned char));
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pycent2, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnError(cent_error);
	if ( val < 0 || val > 255)
	    ReturnError(cent_error);
	cent2[i] = val;
    }

    /* Call Voxel Kit to do the real work */
    voclass = parray_create();
    vaclass = voclass->varray;
    vxl_nnclass2_uchar(vaclass, vasrc1, vasrc2, nclass, clid, cent1, cent2);

    /* Free working storage and return the result */
    free(clid);
    free(cent1);
    free(cent2);
    pyclass = (PyObject *)voclass;
    return pyclass;
}



/*--------------------------------------------------------------------------
parray_norm1 - Compute vector 1-norm of an array

This function computes the vector 1-norm of an array or the difference
between two arrays, optionally weighted by a pixelwise weight.  The
two arrays must be the same type, but the weight array may be of any
type.  The other and weight arrays are optional, defaulting to zero
and one respectively; if provided, they must be the same shape as the
first array.  As a special case, a scalar zero may be provided as the
value of other; otherwise, scalars may not be provided as arguments.
---------------------------------------------------------------------------*/

static PyObject *
parray_norm1(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    double result;           /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OO", &other, &weight) )
	return NULL;

    /* Extract voxel arrays */
    if ( !PyvoxArray_Check(self, &slf) )
	ReturnError("Self is not a Pyvox array");
    if ( other == 0 )
	/* Do nothing */ ;
    else if ( PyvoxArray_Check(other, &oth) )
	/* Do nothing */ ;
    else if ( PyArg_Parse (other, "d",  &result) && result == 0.0 )
	/* Do nothing */ ;
    else
	ReturnError("Other must be a Pyvox array");
    if ( weight == 0 || weight == Py_None )
        wate = 0;
    else if ( !PyvoxArray_Check(weight, &wate) )
	ReturnError("Weight must be a Pyvox array");

    /* Call Voxel Kit to do the real work; return result */
    result = vxl_norm1(slf, oth, wate);
    return Py_BuildValue("d", result);
}



/*--------------------------------------------------------------------------
parray_norm2 - Compute vector 2-norm of an array

This function computes the vector 2-norm of an array or the difference
between two arrays, optionally weighted by a pixelwise weight.  The
two arrays must be the same type, but the weight array may be of any
type.  The other and weight arrays are optional, defaulting to zero
and one respectively; if provided, they must be the same shape as the
first array.  As a special case, a scalar zero may be provided as the
value of other; otherwise, scalars may not be provided as arguments.
---------------------------------------------------------------------------*/

static PyObject *
parray_norm2(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    double result;           /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OO", &other, &weight) )
	return NULL;

    /* Extract voxel arrays */
    if ( !PyvoxArray_Check(self, &slf) )
	ReturnError("Self is not a Pyvox array");
    if ( other == 0 )
	/* Do nothing */ ;
    else if ( PyvoxArray_Check(other, &oth) )
	/* Do nothing */ ;
    else if ( PyArg_Parse (other, "d",  &result) && result == 0.0 )
	/* Do nothing */ ;
    else
	ReturnError("Other must be a Pyvox array");
    if ( weight == 0 || weight == Py_None )
        wate = 0;
    else if ( !PyvoxArray_Check(weight, &wate) )
	ReturnError("Weight must be a Pyvox array");

    /* Call Voxel Kit to do the real work; return result */
    result = vxl_norm2(slf, oth, wate);
    return Py_BuildValue("d", result);
}



/*--------------------------------------------------------------------------
parray_normsup - Compute vector sup-norm of an array

This function computes the vector sup-norm of an array or the
difference between two arrays, optionally weighted by a pixelwise
weight.  The two arrays must be the same type; the weight array may be
of any type.  The other and weight arrays are optional, defaulting to
zero and one respectively; if provided, they must be the same shape as
the first array.  As a special case, a scalar zero may be provided as
the value of other; otherwise, scalars may not be provided as
arguments.
---------------------------------------------------------------------------*/

static PyObject *
parray_normsup(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    double result;           /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OO", &other, &weight) )
	return NULL;

    /* Extract voxel arrays */
    if ( !PyvoxArray_Check(self, &slf) )
	ReturnError("Self is not a Pyvox array");
    if ( other == 0 )
	/* Do nothing */ ;
    else if ( PyvoxArray_Check(other, &oth) )
	/* Do nothing */ ;
    else if ( PyArg_Parse (other, "d",  &result) && result == 0.0 )
	/* Do nothing */ ;
    else
	ReturnError("Other must be a Pyvox array");
    if ( weight == 0 || weight == Py_None )
        wate = 0;
    else if ( !PyvoxArray_Check(weight, &wate) )
	ReturnError("Weight must be a Pyvox array");

    /* Call Voxel Kit to do the real work; return result */
    result = vxl_normsup(slf, oth, wate);
    return Py_BuildValue("d", result);
}



/*-------------------------------------------------------------------------
parray_p2i - Compute the physical-to-index transform

This function returns the affine transform which converts from
physical to index coordinates for this array, based on the origin and
spacing attributes.

ALGORITHM: This function makes an upcall to the Python level, where the
affine transform class is defined.
-------------------------------------------------------------------------*/

static PyObject *
parray_p2i(
    PyObject *self,             /* Array to get transform of */
    PyObject *args)             /* Args: None used */
{
    voxel_array *vself;         /* Self as a voxel array */
    int rank;                   /* Rank of the array */
    long dimen[2];              /* Dimensions of matrix and offset */
    PyObject *mtx;              /* Matrix of the transform */
    voxel_array *vmtx;          /* Matrix, as a voxel array */
    double *mdata;              /* Data of the matrix */
    PyObject *off;              /* Offset of the transform */
    voxel_array *voff;          /* Offset, as a voxel array */
    double *odata;              /* Data of the offset */
    PyObject *affine;           /* Affine class object */
    PyObject *result;           /* The transform to be returned as result */
    int i;

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the array itself */
    if ( !PyvoxArray_Check(self, &vself) )
	ReturnError("Not a Pyvox array");
    rank = vself->rank;

    /* Construct arrays for the matrix and offset of the transform */
    dimen[0] = dimen[1] = rank;
    mtx   = parray_alloc(INTP_DOUBLE, 2, dimen);
    vmtx  = PyVoxel_GetArray(mtx);
    mdata = (double *)vmtx->data;
    dimen[1] = 1;
    off   = parray_alloc(INTP_DOUBLE, 2, dimen);
    voff  = PyVoxel_GetArray(off);
    odata = (double *)voff->data;

    /* Fill matrix and offset with the right values */
    for (i = 0; i < rank*rank; i++)
	mdata[i] = 0.0;
    for (i = 0; i < rank; i++) {
	mdata[i*rank+i] = 1.0 / vself->spacing[i];
	odata[i] = - vself->origin[i] / vself->spacing[i];
    }

    /* Construct and return the affine transform */
    affine = PyDict_GetItemString(pyvox_dict, "affine");
    result = PyObject_CallFunction(affine, "OO", mtx, off);

    Py_INCREF(result);
    return result;
}



/*----------------------------------------------------------------------
parray_prinaxes - Principal axes transformation of real symmetric matrix

This Pyvox array method returns a pair (val, vec) containing the
eigenvalues and eigenvectors of the Pyvox array considered as a real
symmetric matrix.  The matrix of eigenvalues is guaranteed to be proper.
-----------------------------------------------------------------------*/

static 
PyObject *
parray_prinaxes(
    PyObject *self,            /* Pyvox array */
    PyObject *args)            /* Arguments (none) */
{
    voxel_array *vself;        /* Self as a voxel array */
    PyObject *val;             /* Eigenvalues */
    voxel_array *vval;         /* Eigenvalues as voxel array */
    PyObject *vec;             /* Eigenvectors */
    voxel_array *vvec;         /* Eigenvectors as voxel array */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");

    /* Create voxel arrays for values and vectors */
    val  = (PyObject *)parray_create();
    vval = PyVoxel_GetArray(val);
    vec  = (PyObject *)parray_create();
    vvec = PyVoxel_GetArray(vec);

    /* Compute principal axes transformation */
    matrix_principal_axes(vval, vvec, vself);

    /* Return pair of eigenvalues and eigenvectors */
    return Py_BuildValue("(NN)", val, vec);
}



/*----------------------------------------------------------------------
parray_reshape - Change the shape (dimensions) of a Pyvox array

This instance method changes the shape of a Pyvox array without
modifying the values contained therein.  It is not permitted to change
the total number of elements by this method; that is, the product of
the new dimensions must equal the product of the old dimensions.
Similarly, it is not permitted to change the type of the data.
-----------------------------------------------------------------------*/

static PyObject *
parray_reshape(
    PyObject *self,              /* Pyvox array to reshape */
    PyObject *args)              /* New dimensions */
{
    PyObject *shape;             /* Arg 1: Tuple or list with new dimens */

    voxel_array *vself;          /* Self as a voxel array */
    int  newrank;                /* New rank */
    long newdim[VXL_MAX_RANK];   /* New dimensions */
    long oldcnt, newcnt;         /* Old and new number of elements */
    int i;

    /* Convert self to a voxel array */
    vself = ((pyvox_array *)self)->varray;

    /* Fetch the arguments */
    if ( ! PyArg_ParseTuple(args, "O", &shape) ||
	 ! Pyvox_ParseShape(shape, &newrank, newdim) )
	return NULL;

    /* Verify that the number of elements is unchanged */
    newcnt = 1;
    for (i = 0; i < newrank; i++)
	newcnt *= newdim[i];
    oldcnt = 1;
    for (i = 0; i < vself->rank; i++)
	oldcnt *= vself->dimen[i];
    if ( newcnt != oldcnt )
	ReturnError("Reshape may not change number of elements");

    /* Call Voxel Kit to actually change the shape */
    vxl_reshape_array(vself, newrank, newdim);

    /* Return the modified array */
    Py_INCREF(self);
    return self;
}



/*----------------------------------------------------------------------
parray_scale - Scale without changing type

This instance method multiples each element by a given constant, adds
another constant, and returns the results at a new array of the same
type as the original, limiting to the range of the destination as
necessary.
------------------------------------------------------------------------*/

static 
PyObject *
parray_scale(
    PyObject *self,             /* Pyvox array */
    PyObject *args)             /* Arguments */
{
    double gain = 1.0;          /* Arg 1: Constant to multiply by */
    double bias = 0.0;          /* Arg 2: Constant to add */
    PyObject *result;           /* Result: New scaled array */

    voxel_array *vself;         /* Self as a voxel array */
    voxel_array *vresult;       /* Result as a voxel array */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a voxel array");

    /* Convert the other arguments */
    if ( !PyArg_ParseTuple(args, "|dd", &gain, &bias) )
	return NULL;

    /* Create voxel array for result */
    result  = (PyObject *)parray_create();
    vresult = PyVoxel_GetArray(result);

    /* Call Voxel Kit to do the real work */
    vxl_scale(vresult, vself, gain, bias);
    
    /* Return the result */
    return Py_BuildValue("N", result);
}



/*------------------------------------------------------------------------
parray_solve - Solve the linear system AX = B

This instance method solves the linear system AX = B for a square matrix
A and general matrix B.  Both A and B must be float or double, and both
must be the same type.

FIXME: Generate a Python exception if the matrix is singular.
-------------------------------------------------------------------------*/

static
PyObject *
parray_solve(
    PyObject *self,         /* Pyvox array = A */
    PyObject *args)         /* Arguments */
{
    PyObject *rhs;          /* Arg 1: Right hand side = B */

    voxel_array *vself;     /* Self A as a voxel array */
    voxel_array *vrhs;      /* RHS B as a voxel array */
    PyObject *soln;         /* Solution X as a Python array */
    voxel_array *vsoln;     /* Solution X as a voxel array */

    /* Parse arguments */
    if ( !PyArg_ParseTuple(args, "O", &rhs) )
	return NULL;

    /* Find voxel array of self and rhs */
    if ( !PyvoxArray_Check(self, &vself) )
	ReturnError("Not a Pyvox array");
    if ( !PyvoxArray_Check(rhs,  &vrhs) )
	ReturnError("RHS is not a Pyvox array");

    /* Create Pyvox array for result */
    soln  = (PyObject *)parray_create();
    vsoln = PyVoxel_GetArray(soln);

    /* Compute the inverse */
    matrix_solve(vsoln, vself, vrhs);

    /* Return the result as a Pyvox array */
    return Py_BuildValue("N", soln);
}



/*-----------------------------------------------------------------------
parray_stat - Compute mean and standard deviation of array elements

This function computes the mean and standard deviation of the elements
of the array, optionally weighted by the elements of another array.
The results are returned as a tuple containing the mean and standard
deviation.
------------------------------------------------------------------------*/

static PyObject *
parray_stat(
    PyObject *self,             /* Self: Source array */
    PyObject *args)
{
    PyObject *weight = 0;       /* Arg 1: Optional weight array */

    voxel_array *vself = 0;     /* Self as a voxel array */
    voxel_array *vweight = 0;   /* Weight as a voxel array */
    double mean=0, stddev=0;    /* Mean and standard deviation */

    /* Parse the Python arguments and extract voxel arrays */
    if ( !PyArg_ParseTuple(args, "|O", &weight) )
	return NULL;
    if ( !PyvoxArray_Check(self, &vself) )
	ReturnError("Self is not a Pyvox array");
    if ( weight == 0 || weight == Py_None )
        vweight = 0;
    else if ( !PyvoxArray_Check(weight, &vweight) )
	ReturnError("Weight must be a Pyvox array");

    /* Call Voxel Kit to do the real work; return results */
    vxl_stat(&mean, &stddev, vself, vweight);
    return Py_BuildValue("(dd)", mean, stddev);
}
    


/*------------------------------------------------------------------------
parray_trans - Transpose of a matrix

This instance method computes the transpose of a matrix represented as
a Pyvox array of any type.

FIXME: Generalize to allow any rank, any permutation of axes, or reversal
of indices along an axis.
-------------------------------------------------------------------------*/

static
PyObject *
parray_trans(
    PyObject *self,        /* Pyvox array */
    PyObject *args)        /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    PyObject *res;          /* Result as a Python array */
    voxel_array *vres;      /* Result as a voxel array */

    /* Find voxel array of self */
    vself = PyVoxel_GetArray(self);
    if (!vself)
	ReturnError("Not a Pyvox array");

    /* Create Pyvox array for result */
    res  = (PyObject *)parray_create();
    vres = PyVoxel_GetArray(res);

    /* Compute the inverse */
    matrix_transpose(vres, vself);

    /* Return the result as a Pyvox array */
    return Py_BuildValue("N", res);
}



/*--------------------------------------------------------------------------
parray_tuple - Return array elements as a tuple

This instance method for Pyvox arrays returns the elements of the array
(in row-major order) as a Python tuple of floats, ints, or longs.  See also 
parray_list, which returns a list rather than a tuple.
--------------------------------------------------------------------------*/

static PyObject *
parray_tuple(
    PyObject *self,            /* Pyvox array to enlist */
    PyObject *args)            /* Arguments (none in this case) */
{
    PyObject *result;          /* Tuple to return as result */
    voxel_array *vself;        /* Self as a voxel array */
    int type;                  /* Type of voxel array */
    void *data;                /* Data contained in voxel array */
    long count;                /* Number of elements in array */
    double value;
    long i;

    /* Find voxel array of self; count number of elements */
    vself = PyVoxel_GetArray(self);
    if ( !vself )
	ReturnError("Not a Pyvox array");
    type  = vself->type;
    data  = vself->data;
    count = vxli_count(vself);

    /* Check other arguments; there should be none */
    if ( !PyArg_ParseTuple(args, "") )  
	return NULL;

    /* If an integral array, construct a tuple of Python ints and longs */
    result = PyTuple_New(count);
    if ( exim_integral_type(type) ) {
	for (i = 0; i < count; i++) {
	    value = exim_get_value(data, type, i);
	    if ( value <= (double)LONG_MAX )
		PyTuple_SetItem(result, i, PyInt_FromLong((long)value));
	    else
		PyTuple_SetItem(result, i, PyLong_FromDouble(value));
	}
    }

    /* Otherwise construct a tuple of Python floats */
    else {
	for (i = 0; i < count; i++) {
	    value = exim_get_value(data, type, i);
	    PyTuple_SetItem(result, i, PyFloat_FromDouble(value));
	}
    }

    return result;
}



/*--------------------------------------------------------------------------
parray_write - Write contents of a Pyvox array to an external file

This instance method writes the contents of a Pyvox array to an
external file in some specified external format.  If a string is given
as the file argument, then that file is opened for writing, the data
are written to it, and the file is closed; if a file object is given,
then the data are written to the file starting at its current position
but it is not closed nor repositioned after writing.  If no external
data type is specified, the data are written in the "natural" data
type corresponding to the type of the voxel array; note that careless
use of this feature can produce output files that you don't know how
to read back in.

FIXME: Allowing the user to default the external data type is rather
dangerous; how will he know what type to use to read it back in?
Perhaps the external data type should be mandatory?

FIXME:  Should the external type be specified as an integer, or as some
more complex but useful object?
---------------------------------------------------------------------------*/

static PyObject *
parray_write(
    PyObject *self,              /* Pyvox array to write */
    PyObject *args)              /* Arguments describing the output */
{
    PyObject *file;              /* Arg 1: Filename or open file */
    int format = EXTP_NONE;      /* Arg 2: External format (optional) */
    int bigend = 1;              /* Arg 3: Bigendian? (default to yes) */

    voxel_array *varray;         /* Voxel array that self contains */
    char *filename;              /* Name of the image file for output */
    FILE *outfile;               /* Open file for output */
    
    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O|ii", &file, &format, &bigend) )
	return NULL;

    /* Convert a filename or open file into a C FILE pointer */
    if ( PyString_Check(file) ) {
        filename = PyString_AsString(file);
	outfile = fopen(filename, "wb");
	if (outfile == 0)
	    ReturnError("Unable to open the given filename");
    }
    else if ( PyFile_Check(file) )
        outfile = PyFile_AsFile(file);
    else
        ReturnError("Argument 1 isn't a filename or an open file");

    /* Choose the external format and write the data */
    varray = ((pyvox_array *)self)->varray;
    if (format == EXTP_NONE)
	format = exim_natural_extype(varray->type);
    vxl_write_raw_data(outfile, format, bigend, varray);

    /* Close the file, if specified as a filename */
    if ( PyString_Check(file) ) {
	fclose(outfile);
    }

    /* Return the Python object None */
    Py_INCREF(Py_None);
    return Py_None;
}



/*--------------------------------------------------------------------------
parray_writeppm - Write contents of a Pyvox array as a ppm file

This instance method writes the contents of a Pyvox array to an
external file in the binary Portable Pixel Map format.  The array
must be unsigned char, rank 2, and the last dimension must be 3.
---------------------------------------------------------------------------*/

static PyObject *
parray_writeppm(
    PyObject *self,              /* Pyvox array to write */
    PyObject *args)              /* Arguments describing the output */
{
    PyObject *file;              /* Arg 1: Filename */

    voxel_array *varray;         /* Voxel array that self contains */
    char *filename;              /* Name of the image file for output */
    
    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O", &file) )
	return NULL;
    if ( ! PyString_Check(file) )
        ReturnError("Argument isn't a filename");
    filename = PyString_AsString(file);

    /* Choose the external format and write the data */
    varray = ((pyvox_array *)self)->varray;
    vxl_write_ppm_image(filename, varray);

    /* Return the Python object None */
    Py_INCREF(Py_None);
    return Py_None;
}



/*-------------------------------------------------------------------------
parray_zerbv - Zero boundary voxels in a 3d image

Given a variance threshold value (expressed as a standard deviation),
this instance method for Pyvox arrays computes the local variance of
self using a 3x3x3 neighborhood and copies the original image to
another image, zeroing any voxels for which the local variance exceeds
the threshold.  The intent is that the output image should contain
only voxels in the interior of regions with approximately uniform
tone, and that voxels on the boundary between regions should be
zeroed; this should make histogram-based segmentation algorithms more
robust by eliminating partial volume effects from the analysis.
Voxels at the edges of the image are boundary voxels by definition, to
avoid the problems of having an incomplete neighborhood.  The
threshold value is given as the standard deviation, or the square root
of the threshold variance, and is a compromise between the intensity
variance of uniform regions and the difference in gray level between
uniform regions.  As a rough rule of thumb, take the threshold to be 2
or 3*sigma, where sigma is the standard deviation of the intensity
variation within a uniform region; if this value is comparable to or
larger than the intensity differences between regions, this algorithm
probably won't produce any useful results.  The minimum and maximum
possible values of the standard deviation are approximately 0.19 and
127.5 for a 3d unsigned char image.

FIXME:  This is currently implemented for 3d unsigned char images only.

FIXME:  It might make more sense to generate a mask image; then it would
be possible to apply the mask to some differently processed base image.
-------------------------------------------------------------------------*/

static PyObject *
parray_zerbv(
    PyObject *self,        /* Self: Volume image to process */
    PyObject *args)
{
    double stddev;         /* Arg 2: Threshold value, as standard deviation */

    pyvox_array *in, *out;   /* Input and output as Pyvox arrays */

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "d", &stddev) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, 0) )
	ReturnError("Self is not a Pyvox array");
    in = (pyvox_array *)self;
    if (in->varray->type != INTP_UCHAR || in->varray->rank != 3)
	ReturnError("Sorry, only 3d uchar has been implemented yet.");
  
    /* Compute and return the result */
    out = parray_create();
    vxl_zerbv_uchar3d(out->varray, in->varray, stddev);
    return (PyObject *)out;

}



/**************************************************************************
Method Table for Pyvox Array Type
**************************************************************************/



/*--------------------------------------------------------------------------
parray_methods - Table of instance methods for getattr
--------------------------------------------------------------------------*/

static struct PyMethodDef
parray_methods [] = {
    { "abs",      (PyCFunction)parray_abs2,     1, "Absolute value" },
    { "acos",     (PyCFunction)parray_acos,     1, "Arc cosine" },
    { "amax",     (PyCFunction)parray_amax,     1, "Maximum in array" },
    { "amin",     (PyCFunction)parray_amin,     1, "Minimum in array" },
    { "aprod",    (PyCFunction)parray_aprod,    1, "Product of elements" },
    { "asin",     (PyCFunction)parray_asin,     1, "Arc sine" },
    { "asum",     (PyCFunction)parray_asum,     1, "Sum of elements" },
    { "atan",     (PyCFunction)parray_atan,     1, "Arc tangent" },
    { "atan2",    (PyCFunction)parray_atan2,    1, "Arc tangent of two args" },
    { "bigob",    (PyCFunction)parray_bigob,    1, "Biggest object in image" },
    { "bihist",   (PyCFunction)parray_bihist,   1, "Bivariate histogram" },
    { "ceil",     (PyCFunction)parray_ceil,     1, "Ceiling" },
    { "chamfer",  (PyCFunction)parray_chamfer,  1, "Chamfer distance" },
    { "compare",  (PyCFunction)parray_compare,  1, "Comparison" },
    { "convolve", (PyCFunction)parray_convolve, 1, "Convolution" },
    { "copy",     (PyCFunction)parray_copy,     1, "Copy of array" },
    { "cos",      (PyCFunction)parray_cos,      1, "Cosine" },
    { "cosh",     (PyCFunction)parray_cosh,     1, "Hyperbolic cosine" },
    { "count",    (PyCFunction)parray_count,    1, "Count of voxels in image"},
    { "det",      (PyCFunction)parray_det,      1, "Determinant" },
    { "diag",     (PyCFunction)parray_diag,     1, "Diagonal of matrix" },
    { "dilate",   (PyCFunction)parray_dilate,   1, "Morphological dilation" },
    { "dot",      (PyCFunction)parray_dot,      1, "Dot product" },
    { "double",   (PyCFunction)parray_double,   1, "Cast to double" },
    { "eigsy",    (PyCFunction)parray_eigsy,    1, "Symmetric eigensystem" },
    { "erode",    (PyCFunction)parray_erode,    1, "Morphological erosion" },
    { "exp",      (PyCFunction)parray_exp,      1, "Exponential" },
    { "fill2d",   (PyCFunction)parray_fill2d,   1, "Fill 2D contour" },
    { "float",    (PyCFunction)parray_float,    1, "Cast to float" },
    { "floor",    (PyCFunction)parray_floor,    1, "Floor" },
    { "histo",    (PyCFunction)parray_histo,    1, "Histogram" },
    { "i2p",      (PyCFunction)parray_i2p,      1, "Index-to-physical xform" },
    { "int",      (PyCFunction)parray_int,      1, "Cast to int" },
    { "inverse",  (PyCFunction)parray_inverse,  1, "Inverse of matrix" },
    { "kmeans1",  (PyCFunction)parray_kmeans1,  1, "K-means in 1D" },
    { "linear",   (PyCFunction)parray_linear,   1, "Linear interpolation" },
    { "linear0",  (PyCFunction)parray_linear0,  1, "Linear interpolation" },
    { "list",     (PyCFunction)parray_list,     1, "List of elements" },
    { "log",      (PyCFunction)parray_log,      1, "Logarithm base e" },
    { "log10",    (PyCFunction)parray_log10,    1, "Logarithm base 10" },
    { "logcomp",  (PyCFunction)parray_logcomp,  1, "Logarithmic compression" },
    { "long",     (PyCFunction)parray_long,     1, "Cast to long" },
    { "lookup",   (PyCFunction)parray_lookup,   1, "Lookup table" },
    { "lostat",   (PyCFunction)parray_lostat,   1, "Local mean and variance" },
    { "lowpass",  (PyCFunction)parray_lowpass,  1, "Lowpass filter" },
    { "max",      (PyCFunction)parray_max,      1, "Elementwise maximum" },
    { "mean",     (PyCFunction)parray_mean,     1, "Mean along given axes" },
    { "min",      (PyCFunction)parray_min,      1, "Elementwise minimum" },
    { "mmul",     (PyCFunction)parray_mmul,     1, "Matrix multiplication" },
    { "moments",  (PyCFunction)parray_moments,  1, "Image moments" },
    { "mop",      (PyCFunction)parray_mop,      1, "Moments of image product"},
    { "nearest",  (PyCFunction)parray_nearest,  1, "NN interpolation" },
    { "nnclass1", (PyCFunction)parray_nnclass1, 1, "NN classifier in 1D" },
    { "nnclass2", (PyCFunction)parray_nnclass2, 1, "NN classifier in 2D" },
    { "norm1",    (PyCFunction)parray_norm1,    1, "Vector 1-norm" },
    { "norm2",    (PyCFunction)parray_norm2,    1, "Vector 2-norm" },
    { "normsup",  (PyCFunction)parray_normsup,  1, "Vector sup-norm" },
    { "p2i",      (PyCFunction)parray_p2i,      1, "Physical-to-index xform" },
    { "pow",      (PyCFunction)parray_pow,      1, "Power function" },
    { "prinaxes", (PyCFunction)parray_prinaxes, 1, "Principal axes" },
    { "reshape",  (PyCFunction)parray_reshape,  1, "Change shape of array" },
    { "scale",    (PyCFunction)parray_scale,    1, "Scale by gain and bias" },
    { "schar",    (PyCFunction)parray_schar,    1, "Cast to signed char" },
    { "short",    (PyCFunction)parray_short,    1, "Cast to short" },
    { "sin",      (PyCFunction)parray_sin,      1, "Sine" },
    { "sinh",     (PyCFunction)parray_sinh,     1, "Hyperbolic sine" },
    { "solve",    (PyCFunction)parray_solve,    1, "Solve linear system" },
    { "sqrt",     (PyCFunction)parray_sqrt,     1, "Square root" },
    { "stat",     (PyCFunction)parray_stat,     1, "Mean and std deviation" },
    { "tan",      (PyCFunction)parray_tan,      1, "Tangent" },
    { "tanh",     (PyCFunction)parray_tanh,     1, "Hyperbolic tangent" },
    { "trans",    (PyCFunction)parray_trans,    1, "Matrix transpose" },
    { "tuple",    (PyCFunction)parray_tuple,    1, "Tuple of elements" },
    { "uchar",    (PyCFunction)parray_uchar,    1, "Cast to unsigned char" },
    { "uint",     (PyCFunction)parray_uint,     1, "Cast to unsigned int" },
    { "ulong",    (PyCFunction)parray_ulong,    1, "Cast to unsigned long" },
    { "ushort",   (PyCFunction)parray_ushort,   1, "Cast to unsigned short" },
    { "write",    (PyCFunction)parray_write,    1, "Write in raw format" },
    { "writeppm", (PyCFunction)parray_writeppm, 1, "Write in PPM format"},
    { "zerbv",    (PyCFunction)parray_zerbv,    1, "Zero boundary voxels" },
    { NULL,       NULL }
};



/*************************************************************************
Pyvox array type object
*************************************************************************/



/*------------------------------------------------------------------------
parray_type - Python type descriptor for the voxel array type
------------------------------------------------------------------------*/

PyTypeObject parray_type = {

    /* Type head */
    PyObject_HEAD_INIT(&PyType_Type)
    0,                              /* Object size */
    "voxel",                        /* Name of type */
    sizeof(pyvox_array),            /* Basic size of type */
    0,                              /* Item size */

    /* Standard methods */
    (destructor)  parray_destroy,   /* Instance destructor */
    (printfunc)   parray_print,     /* "print x" */
    (getattrfunc) parray_getattr,   /* "instance.attr" */
    (setattrfunc) parray_setattr,   /* "instance.attr = foo" */
    (cmpfunc)     0,                /* compare instances */
    (reprfunc)    0,                /* string representation */

    /* Method suites for standard classes */
    (PyNumberMethods *)   &parray_as_number,    /* tp_as_number */
    (PySequenceMethods *) &parray_as_sequence,  /* tp_as_sequence */
    (PyMappingMethods *)  &parray_as_mapping,   /* tp_as_mapping */

    /* Other methods */
    (hashfunc) 0,                   /* Hash function of an instance */
    (ternaryfunc) 0,                /* "instance()" */
    (reprfunc) 0,                   /* "str(instance)" */

    /* See Include/object.h for still other methods */
};
