/****************************************************************************
pyexim.c - Python bindings for the export/import toolkit

Author:  Paul Hughett

Copyright (c) 2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains code to implement the data export/import package
exim as a Python extension.  The current version is extremely
experimental and subject to change.  YOU HAVE BEEN WARNED!!!

The only objects defined so far are names for the various internal
and external types; methods may be added later as needed.

FIXME: Include a list of all the internal and external types.

FIXME: Is it reasonable to somehow embed the endianness in the type?

FIXME: Create object classes for intypes and extypes; give them attributes
for such things as name, description, size, natural extype and intype, ...
A possible organization:

exim.intypes = a list of defined intypes
exim.extypes = a list of defined extypes
exim.import(???) = import a data stream
exim.export(???) = export a data stream
exim.type(code) = type object corresponding to a numeric code
exim.bigend  = make an extype bigendian?
exim.littleend = make an extype little endian?

where type is an extype or intype:
type.size = bytes per element
type.name = name of the type
type.description = description of the type
type.code = numeric type code used in exim.c
intype.natural = natural extype for this intype
intype.min = minimum possible value
intype.max = maximum possible value
extype.natural = natural intype for this extype
extype.min, max = min/max values, if possible to define

*****************************************************************************/

static const char rcsid[] = "$Id: pyexim.c,v 1.2 2000/12/06 16:27:41 hughett Exp $ Release $Name:  $";

#include "Python.h"

#include "exim.h"
#include "memm.h"


/* Stop gcc from grumbling about missing prototypes */
void initexim(void);



/*------------------------------------------------------------------------
exim.error - Unique error object for Exim to signal an exception
------------------------------------------------------------------------*/

static PyObject *EximError;

#define RaiseError(message) \
    { PyErr_SetString(EximError, message); return NULL; }



/*-----------------------------------------------------------------------
init_external_formats - Initialize Python names of external formats

This function initializes the names which are used to specify external
data formats.  These names happen to be integers this month; there is
no guarantee that they will still be integers next month, or that the
correspondence between names and values will be unchanged.  In short,
use the names, NOT the values!  Endianness is conveyed separately.

FIXME:  The code to skip the EXTP_NONE type relies on the fact that
that type has code zero; this may not be very robust.

FIXME:  This code assumes that PyDict_SetItemString() does not modify
the string pointed to by name; this is probably true, but not actually
guaranteed in the documentation.
------------------------------------------------------------------------*/

static void
init_external_formats(
    PyObject *dict)        /* Dictionary to add the names to */
{
    PyObject *format;      /* A format being defined */
    int nex;               /* Number of external types */
    int *extypes;          /* List of external types */
    const char *name;      /* Name of an external type */
    int i;

    /* Get the list of defined extypes from the exim module */
    nex = exim_all_extypes(0,0);
    extypes = mallock(nex*sizeof(int));
    (void)exim_all_extypes(extypes, nex);

    /* Add each type to the dictionary; magically ignore the none type
       (which has code 0) */
    for (i = 0; i < nex; i++) 
	if (extypes[i] != 0) {
	    format = Py_BuildValue("i", extypes[i]);
	    name = exim_nameof_type(extypes[i]);
	    PyDict_SetItemString(dict, (char *)name, format);
	}

    free(extypes);
    return;
}


/*-----------------------------------------------------------------------
init_internal_formats - Initialize Python names of internal formats

This function initializes the names which are used to specify internal
data formats.  These names happen to be integers this month; there is
no guarantee that they will still be integers next month, or that the
correspondence between names and values will be unchanged.  In short,
use the names, NOT the values!  Endianness is always the native
endianness for the platform.

FIXME:  The code to skip the INTP_NONE type relies on the fact that
that type has code zero; this may not be very robust.

FIXME:  This code assumes that PyDict_SetItemString() does not modify
the string pointed to by name; this is probably true, but not actually
guaranteed in the documentation.
------------------------------------------------------------------------*/

static void
init_internal_formats(
    PyObject *dict)        /* Dictionary to add the names to */
{
    PyObject *format;      /* A format being defined */
    int nin;               /* Number of internal types */
    int *intypes;          /* List of internal types */
    const char *name;      /* Name of an internal type */
    int i;

    /* Get the list of defined intypes from the exim module */
    nin = exim_all_intypes(0,0);
    intypes = mallock(nin*sizeof(int));
    (void)exim_all_intypes(intypes, nin);

    /* Add each type to the dictionary; magically ignore the none type
       (which has code 0) */
    for (i = 0; i < nin; i++) 
	if (intypes[i] != 0) {
	    format = Py_BuildValue("i", intypes[i]);
	    name = exim_nameof_type(intypes[i]);
	    PyDict_SetItemString(dict, (char *)name, format);
	}

    free(intypes);
    return;
}



/*-------------------------------------------------------------------------
exim_methods - Methods implemented by the Exim module
-------------------------------------------------------------------------*/

static struct PyMethodDef
exim_methods[] = {
    { NULL, NULL }
};



/*--------------------------------------------------------------------------
initexim - Initialize the Exim module
--------------------------------------------------------------------------*/

void
initexim()
{
    PyObject *module;        /* The Exim module being created */
    PyObject *dict;          /* ...and its dictionary */

    module = Py_InitModule("exim", exim_methods);
    dict = PyModule_GetDict(module);
    EximError = Py_BuildValue("s", "exim.error");
    PyDict_SetItemString(dict, "error", EximError);
    init_external_formats(dict);
    init_internal_formats(dict);
    if ( PyErr_Occurred() )
        Py_FatalError("Can't initialize exim module");

    return;
}
