/****************************************************************************
pyvox.c - Python bindings for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 1999-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains code to provide the Voxel Kit as an extension to
Python, somewhat along the lines of PIL or Numeric Python.  The current
version is extremely experimental and subject to change.  YOU HAVE
BEEN WARNED!!!

NOTE that, while this file is named pyvox.c, the module that it
creates is named pyvoxC; the module pyvox refers to a Python-language
upper level that imports the pyvoxC module.

FIXME:  The ordering of declarations and definitions in this file seems
to be rather delicate and in some places counter-intuitive.  Look for
a better way to do things.

FIXME: PyObjects may also be pyvox_arrays or kernel_objects in
disguise; the current practice is to pass arguments as PyObjects and
cast as needed inside the function but this is sometimes confusing and
eliminates some error checking.  Is there a better way?

FIXME: The Voxel Kit and its subsidiaries call fatal(), etc when they
see an error; using the current version of errm.c, this causes an
immediate abort (and possibly a core dump).  Improve the error manager
to support the conversion of errors into Python exceptions that can be
caught and reasonably handled.  It's probably useful to permit the
user to choose between immediate abort (for debugging Pyvox) and
a Python error message (for using Pyvox).

FIXME:  This file is starting to get too large for comfort.  Consider
splitting into voxel arrays and everything else; the everything else
can be split later if it seems advisable.
*****************************************************************************/

static const char rcsid[] = "$Id: pyvox.c,v 1.58 2002/02/05 15:49:32 hughett Exp $ Release $Name:  $";

#include <stdio.h>
#include <string.h>

#include "Python.h"

#include "bblanz.h"    /* For bblanz header formats */
#include "clap.h"
#include "exim.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"       /* For vxli_count */
#include "pyvox.h"      /* Must follow voxel.h */


/* Stop gcc from grumbling about missing prototypes */
void initpyvoxC(void);



/*************************************************************************
Some basic objects and functions

FIXME: Some of these functions are likely to be useful in more than
just Pyvox and should be considered for promotion to a more general
library file.
*************************************************************************/



/*-----------------------------------------------------------------------
pyvox_dict - Dictionary for the pyvox module

This global variable is set to point to the dictionary of the pyvox
module (the Python level, not the C level in pyvoxC) when that module
is initialized.  It can be used from the C level to find and call
functions and methods that are defined on the Python level.
-------------------------------------------------------------------------*/

PyObject *pyvox_dict = 0;


/*------------------------------------------------------------------------
PyvoxError - Unique error object for Pyvox to signal an exception
------------------------------------------------------------------------*/

PyObject *PyvoxError = 0;




/*------------------------------------------------------------------------
get_sequence_as_long - Convert a suitable Python sequence to a long array

This function verifies that a Python object is a sequence with a
specified number of elements and that each element is a Python
integer; it then translates the verified sequence into a C array of
long integers.  The requested number of elements may be zero, in which
case the C array is unchanged.  As a special case, a naked number will
be accepted in lieu of a sequence with only one element.  This
function returns a pointer to the array, or NULL if the object cannot
be converted.

NOTE: This function has been replaced by the functions PyNumSeq_Check 
and PyNumSeq_AsLong and is now deprecated.
------------------------------------------------------------------------*/

#if 0               /* No longer used; code kept temporarily for safety */
static long *
get_sequence_as_long(
    long *elem,                /* Elements of sequence */
    int  nelem,                /* Expected number of elements */
    PyObject *seq)             /* The sequence to be converted */
{
    PyObject *item;            /* An element in the sequence */
    int i;

    /* Convert a sequence of the right length. */
    if ( PySequence_Check(seq) && PyObject_Length(seq) == nelem ) { 
	for (i = 0; i < nelem; i++) {
	    item = PySequence_GetItem(seq, i);
	    if ( PyNumber_Check(item) )
		elem[i] = PyInt_AsLong(PyNumber_Int(item));
	    else
		ReturnError("Not a sequence of numbers");
	}
    }

    /* Convert a naked number as if a singleton sequence */
    else if ( nelem == 1 && PyNumber_Check(seq) ) {
	elem[0] = PyInt_AsLong(PyNumber_Int(seq));
    }

    /* Everything else is illegal */
    else
	ReturnError("Not a sequence of numbers");

    /* DEBUG:  Print out the converted array */
    if (0) {
	fprintf(stderr, "Sequence is: ");
	for (i = 0; i < nelem; i++)  
	    fprintf(stderr, " %ld", elem[i]);
	fprintf(stderr, "\n");
    }

    return elem;
}
#endif



/*-----------------------------------------------------------------------
PyvoxLong_Check - Check if PyObject is convertible to a C long value

This Pyvox internal function attempts to convert a Python object into
a C long value.  If successful, it returns 1 and stores the value at
the location specified by the caller.  If not, it returns 0 and sets
an error message in the Python error object.  Note that the calling
sequence differs significantly from the built-in PyNumber_Check
functions; this form is more useful for testing for the right type in
if statements.

FIXME: Maybe also convert singleton arrays to numbers?  What about
singleton lists or tuples?

FIXME:  It is not clear what happens if one of the numbers provided is
outside the range that can be represented by a C long; investigate!
It does appear that neither PyInt_AsLong nor PyNumber_Int generates
an error.
------------------------------------------------------------------------*/

int
PyvoxLong_Check(
    PyObject *ob,                  /* Python object to convert */
    long *value)                   /* Where to put value, or zero */
{
    PyObject *val;                 /* Value converted to Python int */

    if ( PyInt_Check(ob) ) {
	if (value) 
	    *value = PyInt_AsLong(ob);
	return 1; }
    else if ( PyFloat_Check(ob) ) {
	if (value)
	    *value = (long)PyFloat_AsDouble(ob);
	return 1; }
    else if ( PyNumber_Check(ob) && (val = PyNumber_Int(ob)) ) {
	if (value)
	    *value = PyInt_AsLong(val);
	return 1;
    }
    else {
	return 0; }
}



/*-----------------------------------------------------------------------
PyvoxDouble_Check - Check if PyObject is convertible to a C double value

This Pyvox internal function attempts to convert a Python object into
a C double value.  If successful, it returns 1 and stores the value at
the location specified by the caller.  If not, it returns 0 and sets
an error message in the Python error object.  Note that the calling
sequence differs significantly from the built-in PyNumber_Check
functions; this form is more useful for testing for the right type in
if statements.

FIXME: Maybe also convert singleton arrays to numbers?  What about
singleton lists or tuples?
------------------------------------------------------------------------*/

int
PyvoxDouble_Check(
    PyObject *ob,                  /* Python object to convert */
    double *value)                 /* Where to put value, or zero */
{
    PyObject *val;                 /* Value converted to Python float */

    if ( PyInt_Check(ob) ) {
	if (value) 
	    *value = (double)PyInt_AsLong(ob);
	return 1; }
    else if ( PyFloat_Check(ob) ) {
	if (value)
	    *value = PyFloat_AsDouble(ob);
	return 1; }
    else if ( PyNumber_Check(ob) && (val = PyNumber_Float(ob)) ) {
	if (value)
	    *value = PyFloat_AsDouble(val);
	return 1; }
    else {
	return 0; }
}



/*-------------------------------------------------------------------------
PyNumSeq_Check - Test for a list or tuple containing only numbers

This function tests whether the given Python object is a list or tuple
which contains only numbers.  As a special case, a naked number will
be treated as a singleton sequence.  The function returns 1 if the
sequence qualifies and 0 otherwise.

FIXME: It's not clear whether or not an empty sequence should be
considered as a valid numeric sequence.  It is currently, but this
might change in the future.

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
-------------------------------------------------------------------------*/

#if 0
static int
PyNumSeq_Check(
    PyObject *ob)
{
    int      nelem;             /* Number of elements in the sequence */
    PyObject *elem;             /* An element of the sequence */
    int i;

    if ( PySequence_Check(ob) ) {
        nelem = PyObject_Length(ob);
        for (i = 0; i < nelem; i++) {
            elem = PySequence_GetItem(ob, i);
            if ( ! PyNumber_Check(elem) )
                return 0;
        }
        return 1;
    }
    else if ( PyNumber_Check(ob) )
        return 1;
    else
        return 0;
}
#endif



/*------------------------------------------------------------------------
Pyvox_ParseShape - Parse a Pyvox shape description into C

This Pyvox internal function takes a Pyvox shape description--that is,
a tuple or list of numbers--and converts it into the rank and
dimensions that are useful within C.  An empty list is permitted and
indicates a rank-0 or scalar array.  As a special case, a plain number
will be treated as a list containing that one number to specify a rank
1 array.  If the parse is successful, it returns 1; otherwise, it sets
an error message and returns 0.  It is the responsibility of the
caller to return NULL to Python as necessary to indicate that the
operation was successful.
-------------------------------------------------------------------------*/

int                                 /* Returns 1 if successful, else 0 */
Pyvox_ParseShape(
    PyObject *shape,                /* List or tuple to parse */
    int *rank,                      /* (out) Specified rank */
    long *dimen)                    /* (out) Specified dimensions */
{
    PyObject *d;                    /* An element from the shape */
    int i;

    /* A naked number specifies a vector, or rank-1 array */
    if ( PyvoxLong_Check(shape, &dimen[0]) ) {
	*rank = 1;
	return 1;
    }

    /* A list or tuple can specify a multi-dimensional object */
    else if ( PySequence_Check(shape) ) {
	*rank =  PyObject_Length(shape);
	for (i = 0; i < *rank; i++) {
	    d = PySequence_GetItem(shape, i);
	    if ( (d == 0) || ! PyvoxLong_Check(d, &dimen[i]) )
		ReturnError("Some dimension is not a number");
	    if (dimen[i] <= 0)
		ReturnError("Some dimension is zero or negative");
	}
	return 1;
    }

    /* Anything else is an error */
    else {
	ReturnError("Shape must be a number or list of positive numbers");
    }
}



/**********************************************************************
Affine transform functions

Most of the code for the affine transform class is written in Python,
but a few functions are written in C for efficiency.  Those functions
are provided to the Python code as functions defined in the PyvoxC 
module.
************************************************************************/



/*----------------------------------------------------------------------
affine_nearest0 - Affine transform and resample with NN interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using nearest
neighbor interpolation, and returns the result as a new Pyvox array.
Note that the affine transform must be the (reverse)
destination-to-source transform, NOT the forward transform you might
expect.  The type of the result is always the same as the type of the
source.

This version calls the reference implementation in the voxel kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_nearest0(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    PyObject *result;         /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    vmatrix = PyVoxel_GetArray(matrix);
    voffset = PyVoxel_GetArray(offset);
    vsource = PyVoxel_GetArray(source);

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(vsource->type, rank, dimen);
    if (!result)
	ReturnError("Unable to allocate new voxel array");
    vresult = PyVoxel_GetArray(result);

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_nearest to take the voxel arrays instead. */
    vxl_affine_nearest0(vresult, vsource->type, rank, dimen, vsource,
                        vmatrix->data, voffset->data);    
    
    return result;
}



/*----------------------------------------------------------------------
affine_nearest - Affine transform and resample with NN interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using nearest
neighbor interpolation, and returns the result as a new Pyvox array.
Note that the affine transform must be the (reverse)
destination-to-source transform, NOT the forward transform you might
expect.  The type of the result is always the same as the type of the
source.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_nearest(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    PyObject *result;         /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    vmatrix = PyVoxel_GetArray(matrix);
    voffset = PyVoxel_GetArray(offset);
    vsource = PyVoxel_GetArray(source);

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(vsource->type, rank, dimen);
    if (!result)
	ReturnError("Unable to allocate new voxel array");
    vresult = PyVoxel_GetArray(result);

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_nearest to take the voxel arrays instead. */
    vxl_affine_nearest(vresult, vsource->type, rank, dimen, vsource,
		       vmatrix->data, voffset->data);    
    
    return result;
}



/*----------------------------------------------------------------------
affine_linear0 - Affine transform and resample with linear interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using linear
interpolation, and returns the result as a new Pyvox array.  Note that
the affine transform must be the (reverse) destination-to-source
transform, NOT the forward transform you might expect.  The type of
the result is always the same as the type of the source.

This version calls the reference implementation in the voxel kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_linear0(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    PyObject *result;         /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    vmatrix = PyVoxel_GetArray(matrix);
    voffset = PyVoxel_GetArray(offset);
    vsource = PyVoxel_GetArray(source);

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(vsource->type, rank, dimen);
    if (!result)
	ReturnError("Unable to allocate new voxel array");
    vresult = PyVoxel_GetArray(result);

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_linear0 to take the voxel arrays instead. */
    vxl_affine_linear0(vresult, vsource->type, rank, dimen, vsource,
		       vmatrix->data, voffset->data);    
    
    return result;
}



/*----------------------------------------------------------------------
affine_linear - Affine transform and resample with linear interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using linear
interpolation, and returns the result as a new Pyvox array.  Note that
the affine transform must be the (reverse) destination-to-source
transform, NOT the forward transform you might expect.  The type of
the result is always the same as the type of the source.

This version calls the fast implementation in the Voxel Kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_linear(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    PyObject *result;         /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    vmatrix = PyVoxel_GetArray(matrix);
    voffset = PyVoxel_GetArray(offset);
    vsource = PyVoxel_GetArray(source);

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(vsource->type, rank, dimen);
    if (!result)
	ReturnError("Unable to allocate new voxel array");
    vresult = PyVoxel_GetArray(result);

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_linear to take the voxel arrays instead. */
    vxl_affine_linear(vresult, vsource->type, rank, dimen, vsource,
                      vmatrix->data, voffset->data);    
    
    return result;
}




/*************************************************************************
Kernel objects

A kernel object descibes a voxel neighborhood, optionally specifying a
convolution coefficient for each neighbor and a constant bias term
for the convolution sum.  The rank of a kernel is the number of
dimensions; the count is the number of neighbors of any given voxel.
Each neighbor is specified by a set of coordinate deltas relative to
the given voxel.  A kernel can be created with the pyvox.kernel
function; it cannot (currently) be modified or destroyed once created.

FIXME:  Provide functions to save a kernel to a file, or read one from
a file.
*************************************************************************/



/*-------------------------------------------------------------------------
PyKernel_Check - Test whether a Python object is a kernel object

FIXME: Rename to PyvoxKernel_Check.
-------------------------------------------------------------------------*/

int
PyKernel_Check(
    PyObject *ob)
{
    if (ob != 0 && ob->ob_type == &kernel_type)
	return 1;
    else
	return 0;
}



/*----------------------------------------------------------------------
kernel_coef - Return kernel coefficients

This function returns a list containing the coefficients of the given
kernel object.  Note that the list is a copy; changing it will have no
effect on the kernel.  If no coefficients have been defined, the list
will be empty; this may happen even if the neighborhood is non-empty.

FIXME:  Should the list construction loop here be upgraded into a
common library function?
------------------------------------------------------------------------*/

static PyObject *
kernel_coef(
    PyObject *self)
{
    struct vxl_kernel *kern;      /* The kernel struct itself */
    PyObject *pycoef;             /* List containing the coefficients */
    double *coef;                 /* Array of coefficients */
    PyObject *c;                  /* A coefficient */
    int count;                    /* Number of neighbors */
    int n;                        /* Index over neighbors */

    /* Is this really a kernel object? If so, get relevant values. */
    if ( !PyKernel_Check(self) )
	ReturnError("this is not a kernel object");
    kern  = ((kernel_object *)self)->kern;
    coef  = kern->coef;
    count = (coef == 0) ? 0 : kern->count;

    /* Construct list with the coefficients */
    pycoef = PyList_New(count);
    for (n = 0; n < count; n++) {
	c = PyFloat_FromDouble(coef[n]);
	if ( PyList_SetItem(pycoef, n, c) )
	    fprintf(stderr, "PyList_SetItem is unhappy\n");
    }

    return pycoef;	
}



/*-----------------------------------------------------------------------
kernel_create - Create an instance of a kernel object

This function creates a brand-new instance of a kernel object and
initializes it to contain a kernel with given rank, count, neighbors,
coefficients, and bias.

FIXME:  Should this take the neighbors and coefs as Python objects can
convert them internally?  Which approach is more useful?
------------------------------------------------------------------------*/

static PyObject *
kernel_create(
    int rank,                      /* Number of dimensions */
    int count,                     /* Number of neighbors */
    long (*delta)[VXL_MAX_RANK],   /* Coordinate delta to each neighbor */
    double *coef,                  /* Coefficient for each neighbor */
    double bias)                   /* Constant offset for convolution */
{
    kernel_object *self;           /* Kernel object being created */
    struct vxl_kernel *kern;       /* The kernel struct that it contains */

    /* Create a new kernel struct with given characteristics */
    kern = vxl_create_kernel(rank, count, delta, coef, bias);

    /* Create a Python kernel object that contains it */
    self = (kernel_object *)PyObject_NEW(kernel_object, &kernel_type);
    if (self == 0)
        ReturnError("creation of kernel object failed");
    self->kern = kern;

    /* DEBUG:  Report creation of a new kernel object */
    if (0) {
	fprintf(stderr, "Created new kernel object at %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)self)->ob_refcnt); 
    }

    return (PyObject *)self;
}




/*----------------------------------------------------------------------
kernel_delta - Return coordinate deltas of a kernel

This function returns a list of lists containing the coordinate deltas
of a kernel object.  The outer list contains one entry for each
neighbor; each inner list contains the deltas for each coordinate.
Note that the list is a copy; changing it will have no effect on the
kernel.  It is possible to have an empty neighborhood; if so, the list
returned is empty.
------------------------------------------------------------------------*/

static PyObject *
kernel_delta(
    PyObject *self)
{
    struct vxl_kernel *kern;      /* The kernel struct itself */
    PyObject *outer, *inner;      /* Inner and outer lists */
    int count;                    /* Number of neighbors */
    int rank;                     /* Number of dimensions */
    long (*delta)[VXL_MAX_RANK];  /* Array of coord deltas */
    PyObject *del;                /* A single delta value */
    int n, d;                     /* Index over neighbors and dimensions */

    /* Is this really a kernel object? If so, get useful values. */
    if ( !PyKernel_Check(self) )
	ReturnError("this is not a kernel object");
    kern  = ((kernel_object *)self)->kern;
    rank  = kern->rank;
    count = kern->count;
    delta = kern->delta;

    /* Construct the desired list of coordinate deltas */
    outer = PyList_New(count);
    for (n = 0; n < count; n++) {
	inner = PyList_New(rank);
	for (d = 0; d < rank; d++ ) {
	    del = PyInt_FromLong(delta[n][d]);
	    if ( PyList_SetItem(inner, d, del) )
		fprintf(stderr, "PyList_SetItem is unhappy");
	}
	if ( PyList_SetItem(outer, n, inner) )
	    fprintf(stderr, "PyList_SetItem is unhappy");
    }

    return outer;	
}



/*-----------------------------------------------------------------------
kernel_destroy - Destroy an instance of a kernel object

This function destroys an instance of a kernel object by freeing the
storage associated with both the kernel struct itself and the Python
object wrapper that refers to it.  It is normally called automatically
by the Python garbage collector when the reference count of the object
goes to zero.
-----------------------------------------------------------------------*/

static void
kernel_destroy(
    kernel_object *self)
{
    struct vxl_kernel *kern;      /* The kernel struct of the object */

    if (0)  /* DEBUG */
        fprintf(stderr, "Destroyed a kernel object at %p\n", (void *)self);

    kern = self->kern;
    vxl_destroy_kernel(kern);
    PyMem_DEL(self);

    return;
}


/*----------------------------------------------------------------------
kernel_getattr - Get an attribute of a kernel object

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  In other
words, design and use an attribute table for this purpose.
----------------------------------------------------------------------*/

/* Declare the kernel_methods array; it will be defined later. */
static struct PyMethodDef kernel_methods[];

static PyObject *
kernel_getattr(
    PyObject *self,       /* Kernel object to get attribute of */
    char *attr)           /* Name of the attribute to get */
{
    struct vxl_kernel *kern;     /* The kernel struct itself */

    /* Catch things that aren't kernel objects */
    if ( !PyKernel_Check(self) )
	ReturnError("this is not a kernel object");
    kern = ((kernel_object *)self)->kern;

    /* Return value of any requested member of the instance */
    if (strcmp(attr, "bias") == 0)
	return Py_BuildValue("d", kern->bias);
    else if (strcmp(attr, "coef") == 0)
	return kernel_coef(self);
    else if (strcmp(attr, "count") == 0)
	return Py_BuildValue("i", kern->count);
    else if (strcmp(attr, "delta") == 0)
	return kernel_delta(self);
    else if (strcmp(attr, "rank") == 0)
	return Py_BuildValue("i", kern->rank);

    /* Provide list of members available from this instance */
    if (strcmp(attr, "__members__") == 0)
	return Py_BuildValue("[sssss]",
			     "bias", "coef", "count", "delta", "rank");

    /* Use the Python API to find methods associated with this type */
    return Py_FindMethod(kernel_methods, self, attr);
}



/*------------------------------------------------------------------------
kernel_print - Print (informal) representation of a kernel object

This function prints out an informal representation of a kernel object
on a specified FILE.  This representation contains enough information
to distinguish different objects, but not to reconstruct the object
when read back in.
-------------------------------------------------------------------------*/

static int
kernel_print(
    kernel_object *self,         /* The object being printed */
    FILE *out,                   /* Where to print it */
    int flags)                   /* FIXME:  What's this for? */
{
    fprintf(out, "<Pyvox kernel object at %p>", self);
    return 0;
}




/*------------------------------------------------------------------------
kernel_type - Python type descriptor for the kernel type
------------------------------------------------------------------------*/

/* static */      /* FIXME: Can this be made static? Should it be? */
PyTypeObject kernel_type = {

    /* Type head */
    PyObject_HEAD_INIT(&PyType_Type)
    0,                              /* Object size */
    "kernel",                       /* Name of type */
    sizeof(kernel_object),          /* Basic size of type */
    0,                              /* Item size */

    /* Standard methods */
    (destructor)  kernel_destroy,   /* Instance destructor */
    (printfunc)   kernel_print,     /* "print x" */
    (getattrfunc) kernel_getattr,   /* "instance.attr" */
    (setattrfunc) 0,                /* "instance.attr = foo" */
    (cmpfunc)     0,                /* compare instances */
    (reprfunc)    0,                /* string representation */

    /* Method suites for standard classes */
    (PyNumberMethods *)   0,        /* tp_as_number */
    (PySequenceMethods *) 0,        /* tp_as_sequence */
    (PyMappingMethods *)  0,        /* tp_as_mapping */

    /* Other methods */
    (hashfunc) 0,                   /* Hash function of an instance */
    (ternaryfunc) 0,                /* "instance()" */
    (reprfunc) 0,                   /* "str(instance)" */

    /* See Include/object.h for still other methods */
};



/*--------------------------------------------------------------------------
kernel_methods - Table of kernel instance methods for getattr

(None are actually defined yet.)

FIXME: It's apparently possible to add documentation strings to this
table (and probably to the other method tables) and would be a good
idea in general.
--------------------------------------------------------------------------*/

static struct PyMethodDef
kernel_methods [] = {
    { NULL,      NULL }
};



/*************************************************************************
Methods belonging to the Pyvox Module

The following objects and methods are part of the Pyvox module (rather
than belonging to the Pyvox array type or to instances of the voxel
type.
*************************************************************************/



/*------------------------------------------------------------------------
pyvox_arraym - Create a voxel array containing given data

This method constructs a voxel array with given dimensions and type, and
fills it with data from a list.  The given data may be a scalar value, in
which case the entire array is filled with that value.

ALGORITHM: The function is named pyvox_arraym in C to avoid a conflict
with the pyvox_array struct, but has the name pyvox.array in Python.

FIXME: Allow the data argument to be a file name or file object, in
which case this method should read the file to initialize the array.

FIXME: If the shape is omitted, create a scalar of type double?

FIXME: If the data is None, create the array but don't initialize it?
Warn the user in the strongest possible terms that he uses this option
at his own risk.

FIXME: Allow the first dimension to be -1, to indicate it's to be filled
in later, possibly by read?

FIXME: Leaks memory if the operation fails.

FIXME: Looks inside the voxel array structure.
-------------------------------------------------------------------------*/

static PyObject *
pyvox_arraym(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;          /* Arg 1: Tuple or list of dimensions */
    int intype = INTP_DOUBLE; /* Arg 2: Desired internal type */
    PyObject *data = 0;       /* Arg 3: Data to fill array with */

    int rank;                 /* Desired rank */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions */
    PyObject *result;         /* The new varray  object */
    voxel_array *vres;        /* and its voxel array */
    double value = 0.0;       /* Value to be stored in result */
    int nelem, n;             /* Number of data values supplied */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|iO", &shape, &intype, &data) )
        return NULL;

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(intype, rank, dimen);
    if (!result)
	ReturnError("Unable to allocate new voxel array");
    vres = PyVoxel_GetArray(result);

    /* Case 1: Data is a single number, or omitted */
    if ( data == 0 || PyvoxDouble_Check(data, &value) ) {
	vxl_const(vres, value);
    }

    /* Case 2: Data is a sequence of numbers */
    else if ( PySequence_Check(data) ) {
	nelem = PyObject_Length(data);
	if ( nelem != vxli_count(vres) )
	    ReturnError("Length of data must match length of array");
	for (n = 0; n < nelem; n++) {
	    if ( !PyvoxDouble_Check(PySequence_GetItem(data, n), &value) )
		ReturnError("Data must be a number or list of numbers");
	    exim_set_value(vres->data, intype, n, value);
	}
    }

    /* Case 3: Anything else is an error */
    else
	ReturnError("Data must be a number or list of numbers");

    /* Return the result */
    return result;
}



/*--------------------------------------------------------------------------
pyvox_const - Create constant array of given type and dimensions

This pyvox function creates a voxel array of some specified type,
rank, and dimensions, and fills it with a specified constant value.
If the type is omitted, it defaults to unsigned char. (It's not really
clear that this is a useful default, but what the heck.)  If the value
is omitted, it defaults to zero.

FIXME: The more reasonable default for the type is probably double rather
than unsigned char; it's less surprising when you store any value greater
than 255.
--------------------------------------------------------------------------*/

static PyObject *
pyvox_const(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    int intype = INTP_UCHAR;    /* Arg 2: Desired internal type */
    double value = 0.0;         /* Arg 3: Desired value */

    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    pyvox_array *dest;          /* The new voxel array, as Python object */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|id", &shape, &intype, &value) )
        return NULL;

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;
    
    /* Create and initialize the destination array */
    dest = parray_create();
    vxl_alloc_array(dest->varray, intype, rank, dimen);
    vxl_const(dest->varray, value);

    return (PyObject *)dest;
}



/*--------------------------------------------------------------------------
pyvox_ramp - Create array containing coordinate values

This pyvox function creates a voxel array of some specified type,
rank, and dimensions, and fills each voxel with the value of one of
the coordinate axes.  If the type is omitted, it defaults to double.
If the axis is omitted, it defaults to the last axis.
--------------------------------------------------------------------------*/

static PyObject *
pyvox_ramp(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    int intype = INTP_DOUBLE;   /* Arg 2: Desired internal type */
    int axis = -1;              /* Arg 3: Desired axis to store */

    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    pyvox_array *dest;          /* The new voxel array, as Python object */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|ii", &shape, &intype, &axis) )
        return NULL;

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;
    if (axis == -1)
	axis = rank - 1;
    
    /* Create and initialize the destination array */
    dest = parray_create();
    vxl_alloc_array(dest->varray, intype, rank, dimen);
    vxl_ramp(dest->varray, axis);

    return (PyObject *)dest;
}



/*-------------------------------------------------------------------------
pyvox_kernel - Construct a convolution kernel

This Pyvox function constructs a convolution kernel given the
neighbors (defined by coordinate offsets or deltas), the coefficient
for each neighbor, and the constant bias for the convolution.  The
coefficients are optional; if omitted, the kernel defines a
neighborhood only.  The bias is also optional; if omitted, it defaults
to zero.  Once constructed, the kernel cannot be changed, although
this restriction may be lifted in later versions.  The number of
neighbors is determined by the length of the outer delta list; the
length of the coefficient list must be the same, or zero.  The rank is
determined by the length of the inner delta lists; it is an error if
these are not all the same.  Kernels containing no neighbors are
invalid by decree, since it is not possible to determine the rank in
that case.

FIXME:  Should the various list conversion loops below be made into
generic library functions?

FIXME:  Treat empty kernels as a special case, in which the rank is
always zero and the result of convolving is always zero?  I think that
all the functions that use kernels will have no neighbors without any
modification.
------------------------------------------------------------------------*/

static PyObject *
pyvox_kernel(
    PyObject *self,
    PyObject *args)
{
    PyObject *pydelta;         /* Arg 1: Coordinate deltas */
    PyObject *pycoef = 0;      /* Arg 2: Convolution coefficients */
    double bias = 0.0;         /* Arg 3: Constant bias for convolution */

    PyObject *pykern;              /* Kernel being constructed */
    int rank;                      /* Number of dimensions */
    int count;                     /* Number of neighbors, from delta array */
    int cnt2 = 0;                  /* Number of neighbors, from coef array */
    long (*delta)[VXL_MAX_RANK];   /* Coordinate deltas */
    double *coef = 0;              /* Convolution coefficients */
    PyObject *cf;                  /* A coefficient value */
    PyObject *inner;               /* Inner list of deltas */
    PyObject *del;                 /* A single coordinate delta */
    int n, d;                      /* Index over neighbors and dimensions */

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|Od", &pydelta, &pycoef, &bias) )
	return NULL;

    /* Determine number of neighbors */
    count = PyObject_Length(pydelta);
    if (count < 1)
	ReturnError("Empty kernels are not permitted.");
    if (pycoef != 0) {
	cnt2  = PyObject_Length(pycoef);
	if ( cnt2 != count && cnt2 != 0)
	    ReturnError("Delta and coef arrays have incompatible lengths");
    }

    /* Construct coefficient array */
    if (cnt2 != 0) {
	coef = mallock(count*sizeof(double));
	for (n = 0; n < count; n++) {
	    cf = PySequence_GetItem(pycoef, n);
	    if ( !PyvoxDouble_Check(cf, &coef[n]) )
		ReturnError("Some coefficient is not a number");
	}
    }

    /* Determine rank; construct delta array */
    delta = mallock(count*VXL_MAX_RANK*sizeof(long));
    inner = PySequence_GetItem(pydelta, 0);
    rank = PyObject_Length(inner);
    if (rank > VXL_MAX_RANK)
	ReturnError("Requested rank exceeds max supported rank");
    for (n = 0; n < count; n++) {
	inner = PySequence_GetItem(pydelta, n);
	if (rank != PyObject_Length(inner) )
	    ReturnError("Inconsistent dimensions in coordinate deltas");
	for (d = 0; d < rank; d++) {
	    del = PySequence_GetItem(inner, d);
	    if ( PyInt_Check(del) )
		delta[n][d] = PyInt_AsLong(del);
	    else
		ReturnError("Some coordinate delta is not an integer");
	}
    }

    /* Construct the kernel itself */
    pykern = kernel_create(rank, count, delta, coef, bias);

    /* Free temporary storage */
    free(delta);
    free(coef);

    return pykern;
}



/*------------------------------------------------------------------------
pyvox_read - Read raw image data from an external file into a voxel array

This Pyvox function reads raw image data from an external file in some
specified external format, converts it to the most natural internal
format, and returns it as a voxel array.  The arguments are the file
to read from, the desired shape of the array (as a tuple or list), the
external type (which defaults to unsigned char), and an optional flag
to mark bigendian (which defaults to bigendian).

FIXME: Consider renaming this as readraw, since raw data is the only
sort of data that it will read.  Similarly, use writeraw instead of write?

FIXME: Let the shape default to rank 1 and however many elements can
be read from the file?

FIXME: Let the outermost dimension (e.g. number of slices) be
determined by how many elements can be read from the file?

FIXME: Are you sure that you want to use ints to represent the
possible external formats?
------------------------------------------------------------------------*/

static PyObject *
pyvox_read(
    PyObject *self,
    PyObject *args)
{
    PyObject *file;            /* Arg 1: A filename or an open file */
    PyObject *shape;           /* Arg 2: Tuple or list giving dimensions */
    int extype = EXTP_UCHAR;   /* Arg 3: External format (default uchar) */
    int bigend = 1;            /* Arg 4: Bigendian? (defaults to 1) */

    char *filename;            /* Name of an image file to open */
    FILE *infile;              /* Stream for the image file */
    int intype;                /* Natural internal format for the data */
    int rank;                  /* Assumed rank of the image data */
    long dimen[VXL_MAX_RANK];  /* Assumed dimensions of the image data */
    pyvox_array *obj;          /* A new Pyvox Array to hold the data */

    /* Fetch all the arguments */
    if ( !PyArg_ParseTuple(args, "OO|ii", &file, &shape, &extype, &bigend) )
        return NULL;

    /* Translate a filename or open file into a C FILE pointer */
    if ( PyString_Check(file) ) {
        filename = PyString_AsString(file);
	infile = fopen(filename, "rb");
	if (infile == 0)
	    ReturnError("Unable to open the given filename");
    }
    else if ( PyFile_Check(file) )
        infile = PyFile_AsFile(file);
    else
        ReturnError("Argument 1 isn't a filename or an open file");

    /* Determine the assumed rank and dimensions of the data */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) )
	return NULL;

    /* Choose the "natural" internal format for this data. */
    if ( exim_sizeof_type(extype) == 0 )
	ReturnError("Invalid external type");
    intype = exim_natural_intype(extype);
    if ( intype == 0 )
	ReturnError("Cannot find natural internal format for this data");
    
    /* Create a new Pyvox array */
    obj = parray_create();

    /* Read the data into the voxel array */
    vxl_read_raw_data(obj->varray, infile, rank, dimen,
		      intype, extype, bigend);

    /* Return the filled and wrapped voxel array to the caller */
    return (PyObject *)obj;

}



/*------------------------------------------------------------------------
init_bblanz - Read bblanz header and copy info into bblanz instance

This function reads a bblanz header and copies selected information
from that header into attributes of a bblanz instance.  It is designed
to be called only the the pyvox.bblanz method defined in the Python
layer.  The header file is assumed to have the extension ".hdr", which
will be replaced by ".img" to obtain the image file.

FIXME: Check endianness of the header file, rather than assuming bigendian
(even if it is standard at BBL).

FIXME: BBL does not actually set the datatype field, so we have to guess
wildly.

FIXME: Whenever you get around to real extypes rather than ints, change
this code to match.

FIXME: bblanz is more or less a BBL-specific perversion of the AnalyzeView
header and should be moved into a BBL-specific package.
--------------------------------------------------------------------------*/

static PyObject *
init_bblanz(
    PyObject *self,
    PyObject *args)
{
    PyObject *inst;           /* Arg 1: bblanz instance to write to */
    PyObject *file;           /* Arg 2: A filename */

    char *filename = 0;       /* Name of the header file to open */
    FILE *hdrin = 0;          /* Stream for the header file */
    int bigend = 1;           /* Assume bigendian order */
    struct exim_mode *mode;   /* Mode struct to read the header */
    struct dsr *dsr = 0;      /* The header itself */
    PyObject *value;          /* Attribute value as a Python object */
    PyObject *vitem;          /* Item within an attribute tuple */
    int extype = 0;           /* Type of external data */
    int rank;                 /* Number of dimensions */
    int i;

    /* Read the arguments */
    if ( !PyArg_ParseTuple(args, "OO", &inst, &file) )
	return NULL;

    /* Open the indicated header file */
    if ( PyString_Check(file) ) {
	filename = PyString_AsString(file);
	hdrin = fopen(filename, "rb");
	if (hdrin == 0)
	    ReturnError("Unable to open the given header file");
    }

    /* Verify the name of the header file */
    if ( strlen(filename) < 4 || 
	 strcmp(filename+strlen(filename)-4, ".hdr") != 0 )
	ReturnError("Header file does not have '.hdr' extension");

    /* Read the header file */
    mode = exim_create_mode(EXIM_IMPORT, hdrin, stdout, bigend);
    if (mode == 0)
	ReturnError("Insufficient memory to create header input mode");
    dsr = bblanz(mode, dsr);
    exim_destroy_mode(mode);
    fclose(hdrin);

    /* Create hdrfile and imgfile attributes */
    /* FIXME: Check allocation and ref counting; or for memory leakage */
    filename = strdup(filename);
    value = PyString_FromString(filename);
    PyObject_SetAttrString(inst, "hdrfile", value);
    strcpy(filename+strlen(filename)-4, ".img");
    value = PyString_FromString(filename);
    PyObject_SetAttrString(inst, "imgfile", value);

    /* Create extype and bigend attributes, guessing as necessary */
    if (dsr->dime.bitpix == 8)
	extype = EXTP_UINT1;
    else if (dsr->dime.bitpix == 16)
	extype = EXTP_UINT2;
    else
	ReturnError("Data type was not recognized");
    PyObject_SetAttrString(inst, "extype", PyInt_FromLong(extype));
    PyObject_SetAttrString(inst, "bigend", PyInt_FromLong(bigend));

    /* Create rank attribute */
    rank = dsr->dime.dim[0];
    PyObject_SetAttrString(inst, "rank", PyInt_FromLong(rank));

    /* Create size attribute */
    value = PyTuple_New(rank);
    for (i = 0; i < rank; i++) {
	vitem = PyInt_FromLong(dsr->dime.dim[rank-i]);
	PyTuple_SetItem(value, i, vitem);
    }
    PyObject_SetAttrString(inst, "size", value);

    /* Create spacing attribute */
    value = PyTuple_New(rank);
    for (i = 0; i < rank; i++) {
	vitem = PyFloat_FromDouble(dsr->dime.pixdim[rank-i]);
	PyTuple_SetItem(value, i, vitem);
    }
    PyObject_SetAttrString(inst, "spacing", value);

    /* Free working storage */
    free(filename);
    free(dsr);

    /* Return None */
    Py_INCREF(Py_None);
    return Py_None;
}



/*------------------------------------------------------------------------
hello_world - Test some Python implementation details

This function is an experiment to determine what value of self is
provided in a call to the C level and no other useful function.

FIXME: Has this function served its function sufficiently that it could
be removed?
-------------------------------------------------------------------------*/

static PyObject *
hello_world(
    PyObject *self,
    PyObject *args)
{
    /* Print a friendly greeting and show who self is */
    fprintf(stderr, "Greetings from ");
    if (self)
	PyObject_Print(self, stderr, 0);
    else
	fprintf(stderr, "a null pointer");
    fprintf(stderr, "\n");

    /* Return the None value */
    Py_INCREF(Py_None);
    return Py_None;
}



/*-------------------------------------------------------------------------
set_pyvox_dict - Remember the dictionary for the python layer of pyvox

This function is called by the pyvox module as it is being
initialized, and is used to provide the dictionary of the pyvox module
to the Pyvox C level; the dictionary is used to find the callable
objects corresponding to specific functions and methods so that the
corresponding Python objects may be invoked from the C level.

NOTE: The value of self supplied by the caller is 0.  I suppose that
this indicates that this is not being called as a class method but as
a function defined within a modules.

FIXME: Can we also define new functions by inserting them directly in
this dictionary, rather than having to pass calls through the python
level?
--------------------------------------------------------------------------*/

static PyObject *
set_pyvox_dict(
    PyObject *self,
    PyObject *args)
{

    /* Save the dictionary */
    if ( !PyArg_ParseTuple(args, "O", &pyvox_dict) )
	return 0;

    /* Return the Python object None */
    Py_INCREF(Py_None);
    return Py_None;
}



/*-----------------------------------------------------------------------
upcall_function - Test upcalls to a function defined on the Python level

This function is a simple example of how to call a pyvox function
defined in Python, using the dictionary pyvox_dict.  It may be invoked
from python by "pyvox.pyvoxC.upcall_function()" and takes one optional
argument: the name of the person to be greeted.  The function called
is defined in pyvox.py and is called "hello"; note how it must be
looked up in the pyvox dictionary.
------------------------------------------------------------------------*/

static PyObject *
upcall_function(
    PyObject *self,
    PyObject *args)
{
    PyObject *name = 0;       /* Name of person to greet */
    PyObject *hello;          /* Function to make upcall to */
    PyObject *result;         /* Result of upcall */

    /* Get the optional name from the argument list */
    if ( !PyArg_ParseTuple(args, "|O", &name) )
	return NULL;

    /* Find the hello function in the pyvox dictionary */
    hello = PyDict_GetItemString(pyvox_dict, "hello");

    /* Call the hello function, possibly with name */
    if (name)
	result = PyObject_CallFunction(hello, "O", name);
    else
	result = PyObject_CallFunction(hello, "");

    /* Return result */
    Py_INCREF(result);          /* FIXME: Is this right? */
    return result;
}



/*-----------------------------------------------------------------------
upcall_method - Test upcalls to a class method on the Python level

This function is a simple example of how to call a class method
written in Python and defined in the file pyvox.py.  It can be invoked
from Python by "pyvox.pyvoxC.upcall_method()" and takes two arguments:
a instance of the affine class, and an optional name of the person to
be greeted.
------------------------------------------------------------------------*/

static PyObject *
upcall_method(
    PyObject *self,
    PyObject *args)
{
    PyObject *class;          /* Class instance to upcall a method of */
    PyObject *name = 0;       /* Name of person to greet */
    PyObject *result;         /* Result of upcall */

    /* Get the class instance and optional name from the argument list */
    if ( !PyArg_ParseTuple(args, "O|O", &class, &name) )
	return NULL;

    /* Call the hello method, possibly with name */
    if (name)
	result = PyObject_CallMethod(class, "hello", "O", name);
    else
	result = PyObject_CallMethod(class, "hello", "");

    /* Return result */
    Py_INCREF(result);          /* FIXME: Is this right? */
    return result;
}



/*-------------------------------------------------------------------------
pyvox_methods - Methods implemented by the Pyvox module
-------------------------------------------------------------------------*/

static struct PyMethodDef
pyvox_methods[] = {
    { "array",            pyvox_arraym,    1, "Create new Pyvox array" },
    { "const",            pyvox_const,     1, "New array of constant value" },
    { "kernel",           pyvox_kernel,    1, "Create new Pyvox kernel" },
    { "ramp",             pyvox_ramp,      1, "New array filled with ramp" },

    /* Hidden methods for the Python layer to use */
    { "affine_linear",    affine_linear,   1 },    /* Fast version */
    { "affine_linear0",   affine_linear0,  1 },    /* Reference version */
    { "affine_nearest",   affine_nearest,  1 },
    { "affine_nearest0",  affine_nearest0, 1 },    /* Reference version */
    { "hello_world",      hello_world,     1 },    /* For testing upcalls */
    { "init_bblanz",      init_bblanz,     1 },    /* Read bblanz header */
    { "read",             pyvox_read,      1 },    /* Called by rawread */
    { "set_pyvox_dict",   set_pyvox_dict,  1 },
    { "upcall_function",  upcall_function, 1 },    /* For testing upcalls */
    { "upcall_method",    upcall_method,   1 },    /* For testing upcalls */
    { NULL, NULL }
};



/*--------------------------------------------------------------------------
initpyvoxC - Initialize the Pyvox module
--------------------------------------------------------------------------*/

void
initpyvoxC()
{
    PyObject *module;        /* The Pyvox module being created */
    PyObject *dict;          /* ...and its dictionary */

    module = Py_InitModule("pyvoxC", pyvox_methods);
    dict = PyModule_GetDict(module);
    PyvoxError = Py_BuildValue("s", "pyvox.error");
    PyDict_SetItemString(dict, "error", PyvoxError);
    PyDict_SetItemString(dict, "ArrayType",  (PyObject *)&parray_type);
    PyDict_SetItemString(dict, "KernelType", (PyObject *)&kernel_type);
    if ( PyErr_Occurred() )
        Py_FatalError("Can't initialize pyvox module");

    return;
}
