/****************************************************************************
qdv.c - Quick diagnostic viewer for volume images

Author:  Paul Hughett

Copyright (c) 1998-2002 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


See the man page for operating instructions.


FIXME:  Allow the user to specify a color table to be applied after
the image is converted to unsigned char.

FIXME:  Permit intensity windowing on a per-echo basis.

FIXME: Use sliders rather than numeric fields to specify the intensity
window, with instant image updates as the intensity limits change.

FIXME: Image format changes made using the b keystroke are not
propagated to any open format dialog; furthermore, an open dialog box
is not switched to the new image if spacebar is used.  These don't seem
to be what this user expects.

FIXME:  Use the window system for errors and status messages.  Translate
warning, fatal, and panic into the appropriate dialog boxes.  Restart
rather than abort for fatal errors.

FIXME:  The current version of this program is likely to break for
dynamic changes in the size of the image being displayed.  Figure out
how it should work and make it work that way.

FIXME: Now that the size can be set via menu, remove the command line
option?

FIXME: Figure out why the arrow buttons do not have the desired
effects.  Following keys don't seem to get translated into events and
keysyms: arrow key group, home, tab

FIXME:  This program does not produce reasonable error messages for
files that are not volume images.

FIXME:  Handle window manager's requests for icon size.

FIXME: Add index lines to show the current position in the other
coordinates.

FIXME:  Scan through slices should stop as soon as the key is released.

FIXME:  Make compatible with the radiological convention, if possible.
This may require establishing a convention for the image files themselves.

FIXME:  Modularize and create library routines, so that this can be used
as a basis for other display-based programs.

*****************************************************************************/

static const char rcsid[] = "$Id: qdv.c,v 1.36 2002/02/25 21:25:35 hughett Exp $ Release $Name:  $";


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>

#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Label.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>

#include "config.h"

#include "errm.h"
#include "exim.h"
#include "memm.h"
#include "voxel.h"

#include "../bitmaps/qdv_icon.h"

#define BITMAPDEPTH 1

#define MAX_MAPPED_STRING_LENGTH 10  /* Max result from XLookupString */

#define MAX_ZOOM  2                  /* Maximum zoom factor supported */


/*------------------------------------------------------------------------
VoxelViewRec - current parameters describing how an image is displayed

This structure contains information describing a single volume image
(stored either on disk or in memory), an XImage which contains some
selected slice of that image, and the parameters which describe how
to select and present that slice.  The window or drawable to receive the
image is NOT defined in this structure and must be specified independently.

FIXME:  This claims to support external image formats other than unsigned
char and short; create some such files and test that claim.
------------------------------------------------------------------------*/

struct VoxelViewRec {
    const char *imgname;   /* Local name for the image:  e.g. A/B */
    const char *imgfile;   /* Name of image file from command line*/
    FILE *imgin;           /* Image input */
    struct stat imgstat;   /* Information on the image file */
    struct vxl_raw_format  /* How to decode and display this data */
        rawform;
    voxel_array volimg;    /* (Optional) Complete volume image contents */
    voxel_array sect;      /* Current slice being displayed */
    long length;           /* Total number of bytes in image data */
    int curformat;         /* Current data format; see encode_format */
    int curview;           /* Current viewing axes; see encode_axis */
    XImage *ximage;        /* XImage structure */
};



/* Local prototypes */
void showSlice (struct VoxelViewRec *image, Drawable win);
void showStatus(struct VoxelViewRec *image, Drawable win);
struct VoxelViewRec *prepareVoxelView(const char *imgfile,
     const char *imgname);
void freeVoxelView(struct VoxelViewRec *view);
void redrawView(void);
Widget CreateMenuBar (Widget parent);
void CreateFileMenu  (Widget parent);
void CreateFormatMenu(Widget parent);
void ApplyImageFormat(Widget format);
void decode_format(struct vxl_raw_format *fmt, int ifmt);
int  encode_format(struct vxl_raw_format *fmt);
int  incr_format(int ifmt);
char *describe_format(int ifmt);
void decode_axis(struct vxl_raw_format *fmt, int ifmt);
int  encode_axis(struct vxl_raw_format *fmt);
int  incr_axis(int ifmt);


/* Interface to X Windows (some of which may get moved to another file) */
XVisualInfo bestVisual(Display *display);
void mask2shift(unsigned long mask, int nbits, int *right, int *left);
void describeVisual(XVisualInfo *visual);
void describeXImage(XImage *ximage);
void mono2d_to_XImage(XImage *ximage, voxel_array *section);
void rgb2d_to_XImage(XImage *ximage, voxel_array *section, int zoom);
void getGC(Window win, GC *gc, XFontStruct *font_info);
void load_font(XFontStruct **font_info);

/* X Callback functions */
void FormatCallback     (Widget w, XtPointer clientData, XtPointer callData);
void OpenCallback       (Widget w, XtPointer clientData, XtPointer callData);
void PPMCallback        (Widget w, XtPointer clientData, XtPointer callData);
void OpenOKCallback     (Widget w, XtPointer clientData, XtPointer callData);
void OpenCancelCallback (Widget w, XtPointer clientData, XtPointer callData);
void QuitCallback       (Widget w, XtPointer clientData, XtPointer callData);


/* ImageFormatDialog and friends */
void CreateImageFormatDialog(Widget parent, struct VoxelViewRec *image);
void DestroyImageFormatDialog(Widget format);
void FormatOkCallback    (Widget w, XtPointer clientData, XtPointer callData);
void FormatApplyCallback (Widget w, XtPointer clientData, XtPointer callData);
void FormatCancelCallback(Widget w, XtPointer clientData, XtPointer callData);


/* Action routines that respond to specific X events */
#define XACTION(funct)  \
void funct(Widget w, XEvent *ev, String *params, Cardinal *numParams)

XACTION(anyKey);            /* For debugging */
XACTION(cleanExit);
XACTION(ConfigureAction);
XACTION(ExposeAction);
XACTION(incrCurAxis);
XACTION(nextEndian);
XACTION(nextImage);
XACTION(nextLock);
XACTION(nextView);
XACTION(nextZoom);

/* Symbolic names for the action routines */
static XtActionsRec actionsTable [] = {
    { "anyKey",          anyKey},
    { "cleanExit",       cleanExit },
    { "ConfigureAction", ConfigureAction },
    { "ExposeAction",    ExposeAction },
    { "incrCurAxis",     incrCurAxis },
    { "nextEndian",      nextEndian },
    { "nextImage",       nextImage },
    { "nextLock",        nextLock },
    { "nextView",        nextView },
    { "nextZoom",        nextZoom },
};

/* Bind keys to action routines */
static char defaultTranslations[] =
    /* "#override\n"  */
    "<Configure>:    ConfigureAction()\n"
    "<Expose>:       ExposeAction()\n"
    "<Key>A:         nextView()\n"
    "<Key>B:         nextEndian()\n"
    "<Key>L:         nextLock()\n"
    "<Key>Q:         cleanExit()\n"
    "<Key>Z:         nextZoom()\n"
    "<Key>space:     nextImage()\n"
    "<Key>+:         incrCurAxis(+1)\n"
    "<Key>-:         incrCurAxis(-1)\n"
    "<Key>Right:     incrCurAxis(+1)\n"   /* FIXME:  These don't work */
    "<Key>Left:      incrCurAxis(-1)\n"
    "<Key>KP_Right:  incrCurAxis(+1)\n"   /* These don't work on Suns */
    "<Key>KP_Left:   incrCurAxis(-1)\n"
    "<Key>:          anyKey()\n"  ;



/* Global information about the X window configuration in general  */
XtAppContext app;
Display *display;           /* Display on which the app shell appears */
Screen  *screen;            /* Screen  on which the app shell appears */
XFontStruct *font_info;     /* The standard font for this application */

/* Global information about the image viewer */
/* FIXME: Following globals might be specific to the image display */
XVisualInfo visinfo;        /* Visual selected for image display */
Colormap colormap;          /* Color map for that visual */
GC viewer_gc;               /* Graphics context */

/* Global information about the status window */
GC status_gc;               /* Graphics context for drawing text */

/* Some of the Widget hierarchy */
Widget approot;       /* Invisible root shell for this application */
Widget shell;         /* TopLevelShell for the image viewer itself */
Widget canvas;        /* DrawingArea for the image viewer */
Widget statwin;       /* TopLevelShell for a menu and status display */
Widget statmgr;       /* Manager widget for the status and menu widgets */
Widget status;        /* Label to display the current status */
Widget menu;          /* Menu bar for the menu/status window */


/* Images opened and available for display */
#define NImage 2           /* Number of images that can be defined at once */

struct VoxelViewRec *Image[NImage];  /* The available images */
struct VoxelViewRec *CurrentImage;   /* The image currently displayed, or 0 */



/* Parameters gathered from the command line (and usually copied into
   the VoxelViewRec for each image) */
char *imgfileA = 0;         /* Names of the image files */
char *imgfileB = 0;
int nrows=256, ncols=256;   /* Image dimensions */


/* Miscellaneous global information */
static char *progname;      /* Name this program was invoked under */
static int  lockAB = 0;     /* Force A and B to same coordinate positon? */
static int  zoomAB = 1;     /* Zoom factor, applied to both A and B */



/*--------------------------------------------------------------------------
main - main program
---------------------------------------------------------------------------*/

int
main(int argc, char *argv[])
{
    XtTranslations transTable;    /* Default translations */
    char *window_name = "Quick Diagnostic Volume Viewer";
    /* char *icon_name = "qdv"; */  /* FIXME */
    Pixmap icon_pixmap;
    int c;

    /* Initialize the X Toolkit and create an invisible application shell */
    approot = XtVaAppInitialize ( &app, "QDV",  NULL, 0, &argc, argv, NULL,
				  XmNmappedWhenManaged, FALSE,
				  NULL);
    display = XtDisplay(approot);
    screen = XtScreen(approot);
    XtRealizeWidget(approot);

    /* Handle non-X options and arguments on command line */
    while ((c = getopt( argc, argv, "s:t")) != EOF) {
	switch (c) {
	case 's':
	    nrows = atoi(optarg);
	    break;
	}
    }

    /* Handle command line arguments */
    progname = argv[0];
    if (optind < argc)
	imgfileA = argv[optind];
    window_name = "Quick Image Viewer";
    if (optind + 1 < argc) 
	imgfileB = argv[optind+1];

    /* Do some sanity checks on the arguments and options */
    ncols = nrows;
    if (nrows <= 0 || nrows > 1024) 
	fatal("Invalid image size");

    /* Prepare actions and default translations */
    XtAppAddActions ( app, actionsTable, XtNumber ( actionsTable ) );
    transTable =  XtParseTranslationTable ( defaultTranslations );

    /* Setup a menu/status window */
    statwin = XtVaCreatePopupShell ("statwin", topLevelShellWidgetClass,
				    approot,
				    XmNallowShellResize, 1,
				    NULL);
    statmgr = XtCreateManagedWidget("statmgr", xmRowColumnWidgetClass,
				    statwin, NULL, 0);
    menu = CreateMenuBar(statmgr);
    status = XtVaCreateManagedWidget("status", xmDrawingAreaWidgetClass,
				     statmgr,
				     XmNtranslations, transTable,
				     XmNbackground, XBlackPixelOfScreen(screen),
				     XmNheight, 192,
				     XmNwidth,  256,
				     NULL);
    XtPopup (statwin, XtGrabNone);

    /* Prepare an icon pixmap for this app */
    /* FIXME:  Could this go sooner in the code? */
    icon_pixmap = XCreateBitmapFromData(display, XtWindow(approot), 
					qdv_icon_bits, qdv_icon_width,
					qdv_icon_height);
    XtVaSetValues(approot, XmNiconPixmap, icon_pixmap, NULL);

    /* Find the best visual and create suitable color map */
    visinfo = bestVisual(display);
    colormap = XCreateColormap(display, XRootWindowOfScreen(screen),
       visinfo.visual, AllocNone);

    /* Create a visible popup shell as an image viewer */
    shell = XtVaCreatePopupShell ("viewer", topLevelShellWidgetClass, approot,
				  XmNallowShellResize, 1,
				  NULL);

    /* Tell the image viewer shell to use the best visual */
    XtVaSetValues(shell,
		  XmNvisual,   visinfo.visual,
		  XmNdepth,    visinfo.depth,
		  XmNcolormap, colormap,
		  NULL);

    /* Read the image files and set up VoxelViews for them */
    CurrentImage = 0;
    if (imgfileB != 0)
	CurrentImage = Image[1] = prepareVoxelView(imgfileB, "B");
    if (imgfileA != 0)
	CurrentImage = Image[0] = prepareVoxelView(imgfileA, "A");
    
    /* Create the drawing area widget */
    canvas = XtVaCreateManagedWidget( "message", 
				      xmDrawingAreaWidgetClass, shell, 
				      XmNheight,       256,
				      XmNwidth,        256,
				      XmNresizePolicy, XmRESIZE_ANY,
				      XmNtranslations, transTable,
				      XmNvisual,       visinfo.visual,
				      XmNdepth,        visinfo.depth,
				      XmNcolormap,     colormap,
				      NULL );

    /* Prepare translations for the drawing area widget */
    /* XtAugmentTranslations ( canvas, transTable ); */

    /* Realize the image viewer */
    XtPopup (shell, XtGrabNone);

    /* Set up fonts and graphics context (FIXME:  Sooner?) */
    load_font(&font_info);
    getGC(XtWindow(canvas), &viewer_gc, font_info);
    getGC(XtWindow(status), &status_gc, font_info);

    /* Enter the event loop */
    XtAppMainLoop ( app );          /* Never returns */

    return 0;                       /* ...but this keeps gcc happy */
}



/**************************************************************************

        Action routines to respond to various X events

**************************************************************************/


/*--------------------------------------------------------------------------
anyKey - examine a keypress event and describe it

This function analyzes a keypress event and outputs to stderr an
description of its contents.  It is intended for debugging rather
than ordinary use.
----------------------------------------------------------------------------*/

void
anyKey(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    char buffer[MAX_MAPPED_STRING_LENGTH];  /* For XLookupString */
    int bufsize = MAX_MAPPED_STRING_LENGTH;
    KeySym keysym;
    XComposeStatus compose;
    int count;

    count = XLookupString(&(ev->xkey), buffer, bufsize,
			  &keysym, &compose);

    if (0) fprintf(stderr, "keysym = 0x%x\n", (unsigned int)keysym);

    return;
}


/*--------------------------------------------------------------------------
cleanExit - Clean up and exit from program
---------------------------------------------------------------------------*/

void
cleanExit(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    exit(EXIT_SUCCESS);
}


/*--------------------------------------------------------------------------
ConfigureAction - Respond to a Configure event

Just a placeholder for the moment.
---------------------------------------------------------------------------*/

void
ConfigureAction(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
#if 0
    int newwidth, newheight, new;

    newwidth  = ev->xconfigure.width;
    newheight = ev->xconfigure.height;

    fprintf(stderr, "Received a ConfigureNotify event:\n");
    fprintf(stderr, "New height = %d, new width = %d\n", newwidth, newheight);
#endif

    return;
}



/*-----------------------------------------------------------------------
ExposeAction - Redraw as needed after an expose

This function redraws the current slice on the drawing canvas when
it is exposed.  Since the entire image must be redrawn, all but
the last of several contiguous exposes is ignored.
------------------------------------------------------------------------*/

void
ExposeAction(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    if (ev->xexpose.count == 0)
	redrawView();

    return;
}



/*--------------------------------------------------------------------------
nextEndian - Select the next possible length/endian combination

This function selects the next possible combination of length and
endianness for the image data and displays the image using that
combination.
----------------------------------------------------------------------------*/

void
nextEndian(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    int ifmt;

    ifmt = incr_format(encode_format(&CurrentImage->rawform));
    CurrentImage->curformat = ifmt;
    decode_format(&CurrentImage->rawform, ifmt);

    redrawView();

    return;
}


/*--------------------------------------------------------------------------
nextImage - Select the next available image for display

This function selects the next of the images available for display
and displays it.  The selection logic is as follows:  if the next
image is defined, choose it; choose the first defined image; or set
the current image to none.  If lockAB is enabled, then the view and
coordinates are copied from the old image to the new.
----------------------------------------------------------------------------*/

void
nextImage(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    int i;

    if (CurrentImage == Image[0] && Image[1] != 0) {
	CurrentImage = Image[1];
        if ( lockAB ) {
            Image[1]->curview = Image[0]->curview;
            decode_axis(&Image[1]->rawform, Image[1]->curview);
            for (i = 0; i < 3; i++)
                Image[1]->rawform.current[i] = Image[0]->rawform.current[i];
        }
    }
    else if (CurrentImage == Image[1] && Image[0] != 0) {
	CurrentImage = Image[0];
        if ( lockAB ) {
            Image[0]->curview = Image[1]->curview;
            decode_axis(&Image[0]->rawform, Image[0]->curview);
            for (i = 0; i < 3; i++)
                Image[0]->rawform.current[i] = Image[1]->rawform.current[i];
        }
    }
    else if (Image[0] != 0)
	CurrentImage = Image[0];
    else if (Image[1] != 0)
	CurrentImage = Image[1];
    else
	CurrentImage = 0;
	
    redrawView();

    return;
}



/*-----------------------------------------------------------------------
nextLock - Select the next available image lock

This function toggles the state of the lockAB variable, which indicates
whether or not changes in the current coordinate position and axis of
one image are to be copied into the other.
------------------------------------------------------------------------*/

void
nextLock(
    Widget   w,
    XEvent   *ev,
    String   *params,
    Cardinal *numParams)
{
    lockAB = !lockAB;
    redrawView();
    return;
}



/*-----------------------------------------------------------------------
nextView - Advance to the next possible view of the image

This function advances to the next possible view (e.g. cycling through
axial, coronal, and sagittal) and displays the new current view.
------------------------------------------------------------------------*/

void
nextView(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    int i;

    i = incr_axis(encode_axis(&CurrentImage->rawform));
    CurrentImage->curview = i;
    decode_axis(&CurrentImage->rawform, i);
    redrawView();

    return;
}



/*--------------------------------------------------------------------------
nextZoom - Select next zoom factor for the current image

This function selects the next possible zoom factor for the current
image, wrapping around after all have been seen once.
----------------------------------------------------------------------------*/

void
nextZoom(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams )
{
    /* Set the new zoom factor */
    if (zoomAB == MAX_ZOOM)
        zoomAB = 1;
    else
        zoomAB += 1;

    /* Redraw the view and return */
    redrawView();
    return;
}
    



/*--------------------------------------------------------------------------
incrCurAxis - Increment coordinate value along current axis

This function increments or decrements the coordinate position along
the current axis and displays the new slice.
----------------------------------------------------------------------------*/

void
incrCurAxis(
    Widget    w, 
    XEvent   *ev,
    String   *params, 
    Cardinal *numParams)
{
    int incr = +1;
    long *dimen;               /* Dimensions of the image */
    long *current;             /* Coordinates of the current point */
    int  curview;              /* Current view (axis) selection */
    
    /* Determine the increment requested */
    if (*numParams > 0)
	incr = atoi(params[0]);

    /* Locate current position and image dimensions */
    dimen   = CurrentImage->rawform.dimen;
    current = CurrentImage->rawform.current;
    curview = CurrentImage->curview;

    /* Optional debugging report */
    if (0) fprintf(stderr, "view = %d  incr = %d  old = %ld",
		   CurrentImage->curview, incr, current[curview]);

    /* Increment the coordinate appropriate to the current view */
    if (0 <= curview && curview <= 2) {
	current[curview] += incr;
	if (current[curview] <  0)
	    current[curview] = dimen[curview]-1;
	if (current[curview] >= dimen[curview])
	    current[curview] = 0;
    }
    else {
	warning("Current axis is invalid");
	CurrentImage->curview = 0;
    }

    /* Optional debugging information */
    if (0)
	fprintf(stderr, "  new = %ld\n", current[curview]);

    /* Show the slice which is now selected */
    redrawView();

    return;
}



/*-----------------------------------------------------------------------
redrawView - Redraw the image after any modification

This function should be called to redraw the current slice and status 
whenever it has been changed.

FIXME:  Should this also take care of updating the format dialog, if
any?
------------------------------------------------------------------------*/

void
redrawView(void)
{
    showSlice (CurrentImage, XtWindow(canvas));
    showStatus(CurrentImage, XtWindow(status));

    return;
}


/**************************************************************************
File selection clients

These functions and data structures are called by the File Menu to open
and read or write various data files and call in turn upon the generic
capabilities for file name selection provided by OpenCallback and its
friends.  The FSClientDataRec structure is used to pass client-specific
information between layers.
*************************************************************************/


/*--------------------------------------------------------------------------
FSClientDataRec - data for a specific user of a file selection dialog
--------------------------------------------------------------------------*/

typedef void openfunc(const char *fname, void *info);

struct FSClientDataRec {
    openfunc *openfunc;/* Function to execute the actual open command */
    void *info;            /* For client-specific information */
};



/*--------------------------------------------------------------------------
OpenImageAB - (Menu) Read image A or B from a selected file

This function is called from the File Menu to open and read a selected
image file as Image A or B.  The following data structures provide the
information that specifies whether image A or B is to be read.
--------------------------------------------------------------------------*/

static void
OpenImageAB(
    const char *filename,   /* Name of the file to open */
    void *info)             /* Pointer to Image[0] or Image[1] */
{
    struct VoxelViewRec **image = info;   /* Image A or B */
    const char *name=0;                   /* Image name "A" or "B" */

    /* Choose image name corresponding to the selection */
    /* FIXME:  This is rather ugly code */
    if      (*image == Image[0]) name = "A";
    else if (*image == Image[1]) name = "B";
    else    fatal("(Internal) Invalid image selection");

    /* Free the old image and read the new one */
    freeVoxelView(*image);
    CurrentImage = *image = prepareVoxelView(filename, name);

    /* Redisplay the current view */
    redrawView();

    return;
}

struct FSClientDataRec 
    OpenImageAData = {OpenImageAB, (void *)&Image[0]},
    OpenImageBData = {OpenImageAB, (void *)&Image[1]};



/*-------------------------------------------------------------------------
WritePPM - Save the current display as a PPM image file
-------------------------------------------------------------------------*/

static void 
WritePPM(
    const char *filename,   /* Name of the file to open */
    void *info)             /* Not used */
{
    /* Complain if there is no current image displayed */
    if (CurrentImage == 0) {
	warning("No current image to write as a PPM file");
	return; }

    /* Otherwise write the current slice as a file */
    vxl_write_ppm_image(filename, &CurrentImage->sect);

    return;
}

struct FSClientDataRec WritePPMData = {WritePPM, 0};




/**************************************************************************
Menu functions for qdv

These functions create and handle the menus used in qdv.
**************************************************************************/


/*-------------------------------------------------------------------------
CreateMenuBar - create the menu bar for qdv
--------------------------------------------------------------------------*/

Widget
CreateMenuBar(
    Widget parent)
{
    Widget menubar;

    menubar = XmCreateMenuBar(parent, "menu", NULL, 0);

    CreateFileMenu(menubar);
    CreateFormatMenu(menubar);

    XtManageChild(menubar);

    return menubar;
}



/*--------------------------------------------------------------------------
CreateFileMenu - create the file menu pane for qdv

Technical note: The Open A and Open B buttons are distinguished in the
callback routine by the clientData, which is a pointer to an int which
is either 0 or 1.
---------------------------------------------------------------------------*/

void
CreateFileMenu(
    Widget parent)
{
    Widget cascade, submenu, openA, openB, writePPM, quit;

    /* Create the "File: button and its cascade menu */
    submenu = XmCreatePulldownMenu(parent, "filemenu", NULL, 0);
    cascade = XtVaCreateManagedWidget("File", xmCascadeButtonWidgetClass,
				      parent,
				      XmNsubMenuId, submenu,
				      NULL);

    /* Create buttons on the cascading menu */
    openA = XtCreateManagedWidget("Open A", xmPushButtonWidgetClass, submenu,
				  NULL, 0);
    XtAddCallback(openA, XmNactivateCallback, OpenCallback, 
		  (XtPointer)&OpenImageAData);
    openB = XtCreateManagedWidget("Open B", xmPushButtonWidgetClass, submenu,
				  NULL, 0);
    XtAddCallback(openB, XmNactivateCallback, OpenCallback, 
		  (XtPointer)&OpenImageBData);
    writePPM = XtCreateManagedWidget("Write PPM", xmPushButtonWidgetClass, 
				     submenu, NULL, 0);
    XtAddCallback(writePPM, XmNactivateCallback, OpenCallback,
		  (XtPointer)&WritePPMData);
    quit = XtCreateManagedWidget("Quit", xmPushButtonWidgetClass, submenu,
				 NULL, 0);
    XtAddCallback(quit, XmNactivateCallback, QuitCallback, NULL);    

    return;
}


/*-------------------------------------------------------------------------
OpenCallback - open an image file for display
-------------------------------------------------------------------------*/


void 
OpenCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    Widget fsDialog;

    fsDialog = XmCreateFileSelectionDialog(w, "openFileDialog", NULL, 0);
    XtAddCallback(fsDialog, XmNokCallback,     OpenOKCallback, clientData);
    XtAddCallback(fsDialog, XmNcancelCallback, OpenCancelCallback, NULL);
    XtManageChild(fsDialog);

    return;
}


/*-------------------------------------------------------------------------
OpenOKCallback - Handle OK button for an open file selection
--------------------------------------------------------------------------*/

void
OpenOKCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    XmFileSelectionBoxCallbackStruct *cbs;
    char *filename;
    struct FSClientDataRec *client;

    XtUnmanageChild(w);

    /* Find the filename string */
    cbs = (XmFileSelectionBoxCallbackStruct *)callData;
    XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &filename);

    /* Invoke the client's open function to do the real open */
    client = (struct FSClientDataRec *)clientData;
    (client->openfunc)(filename, client->info);

    return;
}


/*-------------------------------------------------------------------------
OpenCancelCallback - Handle cancel button for an open file selection
--------------------------------------------------------------------------*/

void
OpenCancelCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    XtUnmanageChild(w);
    return;
}



/*-------------------------------------------------------------------------
QuitCallback - Quit qdv when appropriate menu button activated
-------------------------------------------------------------------------*/


void 
QuitCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    exit(EXIT_SUCCESS);
}





/*--------------------------------------------------------------------------
CreateFormatMenu - create the format menu pane for qdv

FIXME:  Separate this function into qdv-specific and generic image format
manipulation stuff.
---------------------------------------------------------------------------*/

void
CreateFormatMenu(
    Widget parent)
{
    Widget cascade, submenu, format;

    /* Create the "File: button and its cascade menu */
    submenu = XmCreatePulldownMenu(parent, "formatmenu", NULL, 0);
    cascade = XtVaCreateManagedWidget("Format", xmCascadeButtonWidgetClass,
				      parent,
				      XmNsubMenuId, submenu,
				      NULL);

    /* Create buttons on the cascading menu */
    format = XtVaCreateManagedWidget("Format", xmPushButtonWidgetClass, 
				     submenu,
				     NULL);
    XtAddCallback(format, XmNactivateCallback, FormatCallback, NULL); 

    return;
}


/*-----------------------------------------------------------------------
FormatCallback - Handle Format button
-----------------------------------------------------------------------*/

void 
FormatCallback(
    Widget w,              /* Currently ignored */
    XtPointer clientData,  /* Currently ignored */
    XtPointer callData)    /* Currently ignored */
{
    CreateImageFormatDialog(w, CurrentImage);
    return;
}



/*************************************************************************
CreateRadioValueInt - create radio box to choose a value

FIXME:  These should probably be pulled into a generic library.
*************************************************************************/

struct RadioValueIntRec {
    char *label;      /* Label to use on button, or zero to mark end */
    int value;        /* Corresponding value */
};

Widget CreateRadioValueInt(Widget w,  char *name, int *varib,
    struct RadioValueIntRec *pairs);

static void RadioValueIntCallback(Widget w, XtPointer clientData,
    XtPointer callData);


/*-------------------------------------------------------------------------
CreateRadioValueInt - radio buttons to control value of a variable

This function creates a RadioBox which allows the user to
interactively control the value of a variable, with all callbacks
handled automatically.

------

A pointer to the variable to be controlled is recorded in the userData
resource of the RadioBox; a pointer to each possible value is recorded
in the clientData argument for the corresponding button callback.
-------------------------------------------------------------------------*/

Widget
CreateRadioValueInt(
    Widget w,           /* Parent widget */
    char *name,         /* Resource class name */
    int *varib,         /* Value to be set */
    struct RadioValueIntRec *pairs)
{
    Widget rb;          /* RadioBox */
    Widget button;      /* One of its buttons */
    char *label;        /* Label for the button */
    int value;          /* Corresponding value */
    int i;

    rb = XmCreateRadioBox(w, name, NULL, 0);
    XtVaSetValues(rb, XmNuserData, (XtPointer)varib);
    XtManageChild(rb);

    for (i = 0; pairs[i].label != NULL; i++) {
	label = pairs[i].label;
	value = pairs[i].value;
	button = XtCreateManagedWidget(label, xmToggleButtonWidgetClass,
				       rb, NULL, 0);
	if (*varib == value)
	    XtVaSetValues(button, XmNset, TRUE, NULL);
	XtAddCallback(button, XmNarmCallback,
		      RadioValueIntCallback, (XtPointer)&pairs[i].value);

    }

    return rb;
}


/*------------------------------------------------------------------------
RadioValueIntCallback - callback routine for CreateRadioValueInt
------------------------------------------------------------------------*/

static void
RadioValueIntCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    XtPointer user;    /* User data */
    int *varib;        /* Variable to modify */
    int *value;        /* Value to store there */

    XtVaGetValues(XtParent(w), XmNuserData, &user);
    varib = (int *)user;
    value = (int *)clientData;
    *varib = *value;

    return;
} 



/*************************************************************************
ImageFormatDialog - Interactively manipulate image format parameters

This set of functions permits the user to interactively modify the
format (size, numeric format, byte order, intensity window) of an
image and view the results.

FIXME: Although the voxel kit supports intensity windowing per echo,
here we use the values for echo 0 for all the echoes.  Also, they are
double in the vxl_raw_format struct but only int here.

FIXME: Think about the layout of the dialog and see if it can be
improved.
*************************************************************************/


struct ImageFormatDialogRec {
    struct VoxelViewRec *image;    /* Image being reformatted */
    long old_nslices;              /* Initial number of slices */
    long old_nrows;                /* Initial number of rows */
    long old_ncols;                /* Initial number of columns */
    long old_mingray;              /* Initial minimum for intensity window */
    long old_maxgray;              /* Initial maximum for intensity window */
    int  old_curformat;            /* Initial image format (type, endian) */
    int  new_curformat;            /* Value set from RadioBox */
    Widget dialog;                 /* Dialog controlling image format */
    Widget nrows_text;             /* TextField for number of rows */
    Widget ncols_text;             /* TextField for number of columns */
    Widget gmin_text;              /* TextField for window minimum */
    Widget gmax_text;              /* TextField for window maximum */
    Widget rcform;                 /* RadioBox for image format */
};



/*-------------------------------------------------------------------------
CreateImageFormatDialog - Interactively modify image file format

This function is normally invoked by the "Format" menu button and pops
up a dialog box to solicit parameters that describe the file format
used for the current image.  These parameters include the number of
rows, number of columns, intensity window, numeric format
(signed/unsigned, char/short, float,...) and byte order (endianness).

This function is still somewhat qdv-centric, but it would be useful to
make it into a library function for other programs to use.  (Need to
get the current image in an application-independent fashion.)

FIXME: If the current image is changed under the format dialog, the
dialog will continue to affect the old current image rather than the
new current image.

FIXME: Find some way to insure that a long image file name is not truncated
to fit inside a narrow dialog.
-------------------------------------------------------------------------*/

/* Labels for buttons to control number format */
struct RadioValueIntRec stform[] = 
{
    {"unsigned char",  0},
    {"unsigned short, big-endian", 1},
    {"unsigned short, little-endian", 2},
    {"unsigned char, rgb", 3},
    {NULL,0}
};

#define MAXSTRING   80    /* Space to allocate for string buffer */

void 
CreateImageFormatDialog(
    Widget parent,                /* Parent for the dialog */
    struct VoxelViewRec *image)   /* Image to be handled */
{
    Widget dialog;        /* Dialog box to enter image file format */
    Widget apply;         /* .."Apply" button */
    Widget rc;            /* ..RowColumn widget to manage work area */
    Widget title;         /* ....Label naming image and file */
    Widget sep1;          /* ....Separator */
    Widget rcsize;        /* ....RowColumn for image size */
    Widget nrlbl, nclbl;  /* ......Labels for next two fields */
    Widget nrows, ncols;  /* ......TextFields for number of rows/cols */
    Widget mnlbl, mxlbl;  /* ......Labels for the next two fields */
    Widget gmin, gmax;    /* ......TextFields for intensity window limits */
    Widget sep2;          /* ....Separator */
    Widget rcform;        /* ....RadioBox for number format */
    struct ImageFormatDialogRec *record;  /* Record dialog info */
    XmString xmstr;       /* General-purpose compound string variable */
    char string[MAXSTRING];       /* General-purpose string buffer */

    /* Verify that there really is an image to format */
    if (image == 0)
	fatal("Attempting to set format for a non-existent image");

    /* Create a user data structure to remember dialog info and status */
    record = mallock(sizeof(struct ImageFormatDialogRec));
    record->image = image;
    record->old_nslices = image->rawform.dimen[0];
    record->old_nrows   = image->rawform.dimen[1];
    record->old_ncols   = image->rawform.dimen[2];
    record->old_mingray = image->rawform.minint[0];
    record->old_maxgray = image->rawform.maxint[0];
    record->old_curformat = encode_format(&image->rawform);
    record->new_curformat = record->old_curformat;

    /* Create a dialog box to collect image format parameters */
    dialog = XmCreateMessageDialog(parent, "Set image file format", NULL, 0);
    record->dialog = dialog;
    /* XtVaSetValues(dialog, XmNmwmDecorations, -1, NULL); */
    XtVaSetValues(dialog,
		  XmNautoUnmanage, False, 
		  XmNuserData, (XtPointer)record,
		  NULL);
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_SYMBOL_LABEL));
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_MESSAGE_LABEL));

    /* Add callbacks for the usual action buttons (Apply is different) */
    XtAddCallback(dialog,   XmNokCallback,       FormatOkCallback,     NULL);
    XtAddCallback(dialog,   XmNcancelCallback,   FormatCancelCallback, NULL);

    /* Add an Apply button and its callback */
    xmstr = XmStringCreate("Apply", XmFONTLIST_DEFAULT_TAG);
    apply = XtVaCreateManagedWidget("Apply", xmPushButtonWidgetClass, dialog,
				    XmNlabelString, xmstr,
				    NULL);
    XmStringFree(xmstr);
    XtAddCallback(apply, XmNactivateCallback, FormatApplyCallback,  NULL);

    /* Manage the contents with a RowColumn widget */
    rc = XtVaCreateManagedWidget("rowcol", xmRowColumnWidgetClass, dialog,
				 XmNnumColumns, 1,
				 XmNpacking, XmPACK_TIGHT,
				 XmNorientation, XmVERTICAL,
				 XmNspacing, 0,
				 NULL);

    /* Create a title describing the file being considered */
    snprintf(string, MAXSTRING, "Image %s: %s",
	     image->imgname, image->imgfile);
    xmstr = XmStringCreate(string, XmFONTLIST_DEFAULT_TAG);
    title = XtVaCreateManagedWidget("title", xmLabelWidgetClass, rc,
				    XmNlabelString, xmstr,
				    NULL);
    XmStringFree(xmstr);

    /* Add a separator */
    sep1 = XtVaCreateManagedWidget("sep1", xmSeparatorWidgetClass, rc,
				   XmNseparatorType, XmSHADOW_ETCHED_IN,
				   XmNheight, 20,
				   NULL);

    /* Create RowColumn to manage size of image and intensity window */
    /* FIXME:  Restrict to numeric values within reasonable range. */
    rcsize = XtVaCreateManagedWidget("rcsize", xmRowColumnWidgetClass, rc,
				     XmNnumColumns, 2,
				     XmNpacking, XmPACK_COLUMN,
				     XmNorientation, XmHORIZONTAL,
				     NULL);

    /* Create label and field Widgets for the number of rows */
    xmstr = XmStringCreate("Number of rows ",  XmFONTLIST_DEFAULT_TAG);
    nrlbl = XtVaCreateManagedWidget("nrlbl", xmLabelWidgetClass, rcsize,
				    XmNlabelString, xmstr, NULL);
    XmStringFree(xmstr);
    snprintf(string, MAXSTRING, "%ld", image->rawform.dimen[1]);
    nrows = XtVaCreateManagedWidget("nrows", xmTextFieldWidgetClass, rcsize,
				    XmNvalue, string,
				    XmNcolumns, 5,
				    XmNmaxLength, 5,
				    NULL);
    record->nrows_text = nrows;

    /* Create label and field Widgets for the intensity window minimum */
    xmstr = XmStringCreate("Intensity minimum ",  XmFONTLIST_DEFAULT_TAG);
    mnlbl = XtVaCreateManagedWidget("mnlbl", xmLabelWidgetClass, rcsize,
				    XmNlabelString, xmstr, NULL);
    XmStringFree(xmstr);
    snprintf(string, MAXSTRING, "%d", (int)(image->rawform.minint[0]));
    gmin = XtVaCreateManagedWidget("gmin", xmTextFieldWidgetClass, rcsize,
				    XmNcolumns, 5,
				    XmNmaxLength, 5,
				    XmNvalue, string,
				    NULL);
    record->gmin_text = gmin;

    /* Create label and field Widgets for the number of columns */
    xmstr = XmStringCreate("Number of columns ",  XmFONTLIST_DEFAULT_TAG);
    nclbl = XtVaCreateManagedWidget("nclbl", xmLabelWidgetClass, rcsize,
				    XmNlabelString, xmstr, NULL);
    XmStringFree(xmstr);
    snprintf(string, MAXSTRING, "%ld", image->rawform.dimen[2]);
    ncols = XtVaCreateManagedWidget("ncols", xmTextFieldWidgetClass, rcsize,
				    XmNvalue, string,
				    XmNcolumns, 5,
				    XmNmaxLength, 5,
				    NULL);
    record->ncols_text = ncols;

    /* Create label and field Widgets for the intensity window maximum */
    xmstr = XmStringCreate("Intensity maximum ",  XmFONTLIST_DEFAULT_TAG);
    mxlbl = XtVaCreateManagedWidget("mxlbl", xmLabelWidgetClass, rcsize,
				    XmNlabelString, xmstr, NULL);
    XmStringFree(xmstr);
    snprintf(string, MAXSTRING, "%d", (int)(image->rawform.maxint[0]));
    gmax = XtVaCreateManagedWidget("gmax", xmTextFieldWidgetClass, rcsize,
				   XmNcolumns, 5,
				   XmNmaxLength, 5,
				   XmNvalue, string,
				   NULL);
    record->gmax_text = gmax;

    /* Add a separator */
    sep2 = XtVaCreateManagedWidget("sep2", xmSeparatorWidgetClass, rc,
				   XmNseparatorType, XmSHADOW_ETCHED_IN,
				   XmNheight, 20,
				   NULL);

    /* Create RadioBox to control type, endian, rgb */
    rcform = CreateRadioValueInt(rc, "rcform", &(record->new_curformat),
				 stform);
    record->rcform = rcform;

    /* Map the dialog box to the screen */
    XtManageChild(dialog);

    return;
}



/*-------------------------------------------------------------------------
DestroyImageFormatDialog - Destroy an ImageFormatDialog and descendents

This function destroys the specified ImageFormatDialog widget and all of
its descendents, including the ImageFormatDialogRec in which the widget
stores some of its state.  The voxel_array (that is, image) that the
dialog refers to is unchanged.
--------------------------------------------------------------------------*/

void
DestroyImageFormatDialog(
    Widget format)             /* ImageFormatDialog widget to be destroyed */
{
    XtPointer userData;
    struct ImageFormatDialogRec *record;

    /* Find the ImageFormatDialogRec and free it */
    XtVaGetValues(format, XmNuserData, &userData, NULL);
    record = (struct ImageFormatDialogRec *)userData;
    free (record);

    /* Destroy the dialog widget itself */
    XtDestroyWidget(format);

    return;
}



/*-------------------------------------------------------------------------
encode_format & decode_format - Support subset of image formats

Qdv supports only a subset of the possible combinations of external
data type, endianness, and number of echoes.  The particular
combinations are denoted by small integers as list below; the order is
chosen so that the initial value and its immediate successors are the
most used possibilities.  These functions translate between the
encoded integers and the decoded forms used in a vxl_raw_format
struct.  The encodement is rather lazy; if the combination is not one
of the implemented combinations, it will be interpreted as one of them
anyway.  

Also provided is a function to increment the encoded form, with
wraparound at the end of the range, and a function to generate from
the code a one-line description suitable for use in the status window.

      0 => mono uchar
      1 => mono uint2, big endian
      2 => mono uint2, little endian
      3 => rgb  uchar
--------------------------------------------------------------------------*/

void
decode_format(
    struct vxl_raw_format *fmt,     /* Where to put decoded parameters */
    int ifmt)                       /* Integer to decode */
{
    switch (ifmt) {

    case 0:
    default:
	fmt->type = EXTP_UCHAR;
	fmt->bigend = 0;
	fmt->dimen[3] = 1;
	fmt->recho = 0;
	fmt->gecho = 0;
	fmt->becho = 0;
	return;

    case 1:
	fmt->type = EXTP_UINT2;
	fmt->bigend = 1;
	fmt->dimen[3] = 1;
	fmt->recho = 0;
	fmt->gecho = 0;
	fmt->becho = 0;
	return;

    case 2:
	fmt->type = EXTP_UINT2;
	fmt->bigend = 0;
	fmt->dimen[3] = 1;
	fmt->recho = 0;
	fmt->gecho = 0;
	fmt->becho = 0;
	return;

    case 3:
	fmt->type = EXTP_UCHAR;
	fmt->bigend = 0;
	fmt->dimen[3] = 3;
	fmt->recho = 0;
	fmt->gecho = 1;
	fmt->becho = 2;
	return;

    }
}


int 
encode_format(
    struct vxl_raw_format *fmt)    /* Format to encode as an integer */
{
    if (fmt->dimen[3] == 3)
	return 3;
    else if (fmt->type == EXTP_UCHAR)
	return 0;
    else if (fmt->bigend == 1)
	return 1;
    else
	return 2;
}


int
incr_format(
    int ifmt)          /* Encoded format to increment */
{
    if (ifmt >= 3)  return 0;
    else            return ifmt+1;
}

char *
describe_format(
    int ifmt)         /* Encoded format to describe */
{
    switch (ifmt) {
    case 0:  return "Format: uint 1"; 
    case 1:  return "Format: uint 2 B";
    case 2:  return "Format: uint 2 L";
    case 3:  return "Format: uint 1, rgb";
    default: return "Format not recognized";
    }
}




/*-------------------------------------------------------------------------
encode_axis & decode_axis - Support subset of viewable sections

Qdv supports only a subset of the possible combinations of coordinate
axis pairs that define a view to be displayed.  The particular
combinations are denoted by small integers as list below; the order is
chosen so that the initial value and its immediate successors are the
most used possibilities.  These functions translate between the
encoded integers and the decoded forms used in a vxl_raw_format
struct.  The encodement is rather lazy; if the combination is not one
of the implemented combinations, it will be interpreted as one of them
anyway.

Also provided is a function to increment the encoded form, with
wraparound at the end of the range.

      0 => xy => xdim=2, ydim=1
      1 => xz => xdim=2, ydim=0
      2 => yz => xdim=1, ydim=0

19991223: The encoding was 0 => xy; 1 => yz; 2 => xz.  Changed it so that
the code number 0-2 maps directly to the missing axis; that is, the one
that must be incremented by incrCurAxis.  -PWH
--------------------------------------------------------------------------*/

void
decode_axis(
    struct vxl_raw_format *fmt,     /* Where to put decoded parameters */
    int ifmt)                       /* Integer to decode */
{
    switch (ifmt) {

    case 0:
    default:
	fmt->xdim = 2;
	fmt->ydim = 1;
	return;

    case 1:
	fmt->xdim = 2;
	fmt->ydim = 0;
	return;

    case 2:
	fmt->xdim = 1;
	fmt->ydim = 0;
	return;
    }
}


int 
encode_axis(
    struct vxl_raw_format *fmt)    /* Format to encode as an integer */
{
    if (fmt->ydim == 1)
	return 0;
    else if (fmt->xdim == 2)
	return 1;
    else
	return 2;
}


int
incr_axis(
    int ifmt)          /* Encoded format to increment */
{
    if (ifmt >= 2)  return 0;
    else            return ifmt+1;
}




/*-------------------------------------------------------------------------
ApplyImageFormat - copy format info from the dialog into the image

This function copies formatting information from an ImageFormatDialogRec
to the image itself and then redraws the image in the new format.  It
is invoked by the OK and Apply buttons of the Image Format Dialog.
-------------------------------------------------------------------------*/

void
ApplyImageFormat(
    Widget format)
{
    XtPointer userData;
    struct ImageFormatDialogRec *record;
    struct VoxelViewRec *image;
    long slen;        /* Number of bytes in a slice */
    char *s;
    int i, n;

    /* Find the image being reformatted */
    XtVaGetValues(format, XmNuserData, &userData, NULL);
    record = (struct ImageFormatDialogRec *)userData;
    image = record->image;

    /* Copy the number of rows */
    errno = 0;
    s = XmTextFieldGetString(record->nrows_text);
    i = strtol(s, NULL, 10);
    if (errno || i < 1 || i > 2048)
	fatal("Invalid number of rows (%s)", s);
    image->rawform.dimen[1] = i;

    /* Copy the number of columns */
    errno = 0;
    s = XmTextFieldGetString(record->ncols_text);
    i = strtol(s, NULL, 10);
    if (errno || i < 1 || i > 2048)
	fatal("Invalid number of columns (%s)", s);
    image->rawform.dimen[2] = i;

    /* Decode format to type, endianness, and nechoes */
    decode_format(&image->rawform, record->new_curformat);
    image->curformat = record->new_curformat;

    /* Determine the number of slices (Must follow determination of
       ncols, nrows, and nechoes) */
    slen = image->rawform.dimen[1] * image->rawform.dimen[2] *
	image->rawform.dimen[3] * exim_sizeof_type(image->rawform.type);
    image->rawform.dimen[0] = image->length / slen;

    /* Copy the minimum intensity for the window */
    errno = 0;
    s = XmTextFieldGetString(record->gmin_text);
    i = strtol(s, NULL, 10);
    if (errno || i < 0)
	fatal("Invalid minimum intensity (%s)", s);
    for (n = 0; n < VXL_MAX_ECHO; n++)
	image->rawform.minint[n] = i;

    /* Copy the maximum window intensity */
    errno = 0;
    s = XmTextFieldGetString(record->gmax_text);
    i = strtol(s, NULL, 10);
    if (errno || i <= image->rawform.minint[0])
	fatal("Invalid maximum intensity (%s)", s);
    for (n = 0; n < VXL_MAX_ECHO; n++)
	image->rawform.maxint[n] = i;

    /* Redraw the image in the new format */
    redrawView();

    return;
}


/*--------------------------------------------------------------------------
FormatOkCallback - Process Format/Okay

This function implements the OK button on the Image Format Dialog.
--------------------------------------------------------------------------*/

void
FormatOkCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    /* Copy the new format info to the image itself */
    ApplyImageFormat(w);

    /* Remove the dialog from the screen and destroy it */
    DestroyImageFormatDialog(w);

    return;
}


/*--------------------------------------------------------------------------
FormatApplyCallback - Process Format/Okay
--------------------------------------------------------------------------*/

void
FormatApplyCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    Widget format;

    /* Find the format dialog widget */
    format = XtParent(w);

    /* Apply the new format information to the image */
    ApplyImageFormat(format);

    return;
}


/*--------------------------------------------------------------------------
FormatCancelCallback - Process Format/Cancel

This function implements the Cancel button on the Image Format Dialog.
It restores the image format saved when the dialog was created and redraws
the image using the old format.
--------------------------------------------------------------------------*/

void
FormatCancelCallback(
    Widget w,
    XtPointer clientData,
    XtPointer callData)
{
    XtPointer userData;
    struct ImageFormatDialogRec *record;
    struct VoxelViewRec *image;
    int n;

    /* Find the image being reformatted */
    XtVaGetValues(w, XmNuserData, &userData, NULL);
    record = (struct ImageFormatDialogRec *)userData;
    image = record->image;

    /* Restore saved values of format parameters */
    image->rawform.dimen[0]  = record->old_nslices;
    image->rawform.dimen[1]  = record->old_nrows;
    image->rawform.dimen[2]  = record->old_ncols;
    decode_format(&image->rawform, record->old_curformat);
    for (n = 0; n < VXL_MAX_ECHO; n++) {
	image->rawform.minint[n] = record->old_mingray;
	image->rawform.maxint[n] = record->old_maxgray; }

    /* Redraw the image using the old parameters */
    redrawView();

    /* Destroy Format dialog */
    DestroyImageFormatDialog(w);

    return;
}


/**************************************************************************
Utility routines for the X user interface

These routines support the operation of the top-level user interface in
Motif or faciltate the use of X without being specifically related to the
payload functionality.
**************************************************************************/


/*--------------------------------------------------------------------------
mask2shift - Compute shifts to convert gray value to RGB

Given a mask for one component of an RGB value, this function
computes the right and left shifts that should be applied to
convert an n-bit unsigned int gray value to that component
of an unsigned long RGB value.  For example, suppose that the
gray value is an unsigned char (8 bits) and the green mask
is 0x01f0; then the desired component is 5 bits long, the
first (right) shift is 3 bits, and the second (left) shift is
4 bits; thus the desired expression would be

   rgb = (gray >> 3) << 4 + other components.

If the number of bits in the gray value is less than the width
of the field defined by the mask, this function will return
the correct (negative) value, but the expression above will not
portably compute the desired RGB value.  If the mask is all
zero, then this function will return a right shift of n bits,
causing the gray value to be discarded.
-----------------------------------------------------------------------------*/

void
mask2shift(
    unsigned long mask,       /* Mask for an RGB component */
    int nbits,                /* Number of bits in gray value */
    int *right,               /* Right shift needed */
    int *left)                /* Left shift needed */
{
    int i;

    /* Avoid infinite loop if mask is all zero */
    if (mask == 0) {
	*right = nbits;
	*left = 0;  
	return; }

    /* Otherwise compute values for a non-zero mask */
    for (i = 0; (mask & 1) == 0; i++) mask >>= 1;
    *left = i;
    for (i = 0; (mask & 1) == 1; i++) mask >>= 1;
    *right = nbits - i;

    return;
} 


/*-------------------------------------------------------------------------
describeVisual - Display the characteristics of a Visual on stderr

Mostly for debugging.
---------------------------------------------------------------------------*/

void
describeVisual(XVisualInfo *visual)
{
    fprintf (stderr, "\nDescribing an XVisualInfo at %p\n", visual);
    fprintf (stderr, "   depth = %d\n",         visual->depth);
    fprintf (stderr, "   class = %d\n",         visual->class);
    fprintf (stderr, "   red_mask = %lx\n",     visual->red_mask);
    fprintf (stderr, "   green_mask = %lx\n",   visual->green_mask);
    fprintf (stderr, "   blue_mask = %lx\n",    visual->blue_mask);
    fprintf (stderr, "   colormap_size = %d\n", visual->colormap_size);
    fprintf (stderr, "   bits_per_rgb = %d\n",  visual->bits_per_rgb);

    return;
}


/*--------------------------------------------------------------------------
describeXImage - Describe an XImage on stderr

Mostly for debugging.
---------------------------------------------------------------------------*/
void
describeXImage(XImage *ximage)
{
    fprintf(stderr, "\nDescribing an XImage at %p\n", ximage);
    fprintf(stderr, "   width  = %d\n", ximage->width);
    fprintf(stderr, "   height = %d\n", ximage->height);
    fprintf(stderr, "   format = %d\n", ximage->format);
    fprintf(stderr, "   data  -> %p\n", (void *)(ximage->data));
    fprintf(stderr, "   byte_order  = %d\n", ximage->byte_order);
    fprintf(stderr, "   bitmap_unit = %d\n", ximage->bitmap_unit);
    fprintf(stderr, "   bitmap_bit_order = %d\n", ximage->bitmap_bit_order);
    fprintf(stderr, "   bitmap_pad = %d\n", ximage->bitmap_pad);
    fprintf(stderr, "   depth = %d\n", ximage->depth);
    fprintf(stderr, "   bytes_per_line = %d\n", ximage->bytes_per_line);
    fprintf(stderr, "   bits_per_pixel = %d\n", ximage->bits_per_pixel);
    fprintf(stderr, "   red_mask   = %lx\n", ximage->red_mask);
    fprintf(stderr, "   green_mask = %lx\n", ximage->green_mask);
    fprintf(stderr, "   blue_mask  = %lx\n", ximage->blue_mask);
    fprintf(stderr, "   obdata -> %p\n", (void *)(ximage->obdata));

    return;
}




/*---------------------------------------------------------------------------
bestVisual - Find the deepest TrueColor visual on the X server

This function examines the visuals supported by the given display on
the X server and chooses (if possible) the TrueColor visual with the
greatest depth.

----------------------------------------------------------------------------*/

XVisualInfo
bestVisual(
    Display *display)
{
    XVisualInfo *visuals;          /* Visuals supported by server */
    XVisualInfo best;              /* Best visual for our purposes */
    int nvisuals;
    int i, ibest;

    /* Let's see what visuals are available */
     visuals = XGetVisualInfo(display, 0L, &visinfo, &nvisuals);

    /* Print available visuals for debugging */
    if (0) {
        printf ("nvisuals = %d\n", nvisuals);
        for (i = 0; i < nvisuals; i++) {
            fprintf(stderr, "Visual number %d:\n", i);
	    describeVisual(&visuals[i]);
	}
    }

    /* Look for the TrueColor visual with the largest depth */
    ibest = -1;
    for (i = 0; i < nvisuals; i++) {
	if ( visuals[i].class == TrueColor &&
             (ibest < 0 || visuals[i].depth > visuals[ibest].depth) )
	    ibest = i;
    }
    if (ibest < 0) 
        fatal("Sorry, server doesn't provide TrueColor");
    best = visuals[ibest];
    if (0) describeVisual(&best);
    XFree(visuals);

    return best;
}


/*-------------------------------------------------------------------------
getGC - Create a graphics context for this program

Note that the background and foreground are set to black and white for
the default color map and thus are probably incorrect for anyone who uses
something other than the default.
--------------------------------------------------------------------------*/

void getGC(
   Window win,
   GC *gc,
   XFontStruct *font_info)
{
	unsigned long valuemask = 0; /* ignore XGCvalues and use defaults */
	XGCValues values;
	unsigned int line_width = 6;
	int line_style = LineOnOffDash;
	int cap_style = CapRound;
	int join_style = JoinRound;
	int dash_offset = 0;
	static char dash_list[] = {12, 24};
	int list_length = 2;

	/* Create default Graphics Context */
	*gc = XCreateGC(display, win, valuemask, &values);

	/* specify font */
	XSetFont(display, *gc, font_info->fid);

	/* Set foreground to white and background to black */
	XSetForeground(display, *gc, XWhitePixelOfScreen(screen));
	XSetBackground(display, *gc, XBlackPixelOfScreen(screen));

	/* set line attributes */
	XSetLineAttributes(display, *gc, line_width, line_style, 
			cap_style, join_style);

	/* set dashes */
	XSetDashes(display, *gc, dash_offset, dash_list, list_length);

	return;
}

/*-------------------------------------------------------------------------
load_font - Load the font to be used in this program
--------------------------------------------------------------------------*/

void load_font(
   XFontStruct **font_info)
{
	char *fontname = "9x15";

	/* Load font and get font information structure. */
	*font_info = XLoadQueryFont(display,fontname);
	if (*font_info == NULL)
	    fatal("%s: Cannot open 9x15 font", progname);

	return;
}





/************************************************************************
VoxelView functions

These functions support the display of a volume image under X.
*************************************************************************/


/*----------------------------------------------------------------------
prepareVoxelView - prepare a voxel view from an image file

This function examines the image file to determine the number of
slices that it contains and reads the whole thing into memory for fast
access.  This function also creates an XImage that will be used to
display the specified VoxelView image.  The parameters needed to
convert any given slice of the image for display are stored in the
created VoxelImageRec.

FIXME: (?) The parameters nrows, and ncols are copied from the command
line via the global variables of the same name.

FIXME: The temporary uncompressed file is always put into /tmp, which
might or might not be the right place for it.  For example, it may be
small on many systems.  An it will not exist on non-Unix systems.

FIXME: Provide workarounds for the various not-quite-portable functions
required below.
----------------------------------------------------------------------*/

#define MAXCMD  255                /* Buffer size for command */

#if !HAVE_STRDUP
#error "The prepareVoxelView function currently requires strdup()"
#endif

struct VoxelViewRec *
prepareVoxelView(
    const char *imgfile,           /* Filename to read image from */
    const char *imgname)           /* Name to display with the image */
{
    char *testfile = 0;            /* Possible name of compressed file */
    struct stat teststat;          /* Status info about testfile */
    char tempfn[] = "/tmp/qdvXXXXXX";   /* Temporary uncompressed filename */
    int  tempfd = -1;              /* File descriptor for tempfn */
    int rank;                      /* Number of dimensions in voxel array */
    long dimen[4];                 /* Dimensions of voxel array */
    size_t slen;                   /* Length of something or another */
    struct VoxelViewRec *image;    /* Image View parameter record */
    unsigned char *xidata;         /* Image data for an TrueColor XImage */
    XImage *ximage;                /* An XImage structure */
    long nx, ny;                   /* Number of columns and rows */
    int status, n;

    /* Tell user what's going on */
    fprintf(stderr, "Reading image...");

    /* Allocate space for an image view parameter record */
    image = mallock(sizeof(struct VoxelViewRec));

    /* Remember the name of the input file */
    /* FIXME:  strdup is common but not actually ANSI */
    image->imgfile = strdup(imgfile);

    /* Determine true name of the (possibly compressed) image file */
    n = strlen(imgfile);
    testfile = mallock(n+4);
    strcpy(testfile, imgfile);
    if  ( stat(testfile, &teststat) == 0 )
        imgfile = testfile;
#if HAVE_GUNZIP && HAVE_MKSTEMP && HAVE_SNPRINTF && HAVE_SYSTEM
    else if (( strcpy(testfile+n, ".Z" ) && stat(testfile, &teststat) == 0 ) ||
             ( strcpy(testfile+n, ".gz") && stat(testfile, &teststat) == 0 ) )
        imgfile = testfile;
#endif
    else
        fatal("Unable to stat the requested image file");

    /* Uncompress a compressed file into a temporary file; the temporary
       file replaces the original file for the rest of this function */
    /* FIXME: the mkstemp() function is not supported on all systems,
       and may permit only 25 successive calls on others.  */
    /* FIXME: Some systems that don't have gunzip might have uncompress */
#if HAVE_GUNZIP && HAVE_MKSTEMP && HAVE_SNPRINTF && HAVE_SYSTEM
    n = strlen(imgfile);
    if ( strcmp( &imgfile[n-2], ".Z" ) == 0 ||
         strcmp( &imgfile[n-3], ".gz") == 0 ) {
        char cmd[MAXCMD+1];            /* Command to uncompress image file */
        char cmdfmt[] = "gunzip -c %s > %s";   /* Command format */
        tempfd = mkstemp(tempfn);
        if (tempfd < 0)
            fatal("Unable to create temporary uncompressed file");
        n = snprintf(cmd, MAXCMD, cmdfmt, imgfile, tempfn);
        if ( n < 0 || n > MAXCMD )
            fatal("Insufficient buffer to construct gunzip command");
        n = system(cmd);
        if ( n != 0 )
            fatal("The gunzip command failed");
        imgfile = tempfn;
    }
#endif

    /* Determine status of the image file; the number of slices is
       computed in showSlice because it depends on nbytes */
    /* FIXME:  failure to stat the file shouldn't kill the program */
    status = stat(imgfile, &image->imgstat);
    if (status != 0) 
        fatal("Stat failed on image file.");

    /* Open the image file (or temporary uncompressed file) for reading */
    image->imgin = fopen(imgfile, "rb");
    if (image->imgin == 0) 
        fatal("Unable to open image file");

    /* Arrange for a temporary uncompressed file to automatically disappear; 
       this must be done AFTER the file has been opened to be read */
    if ( tempfd >= 0 )
        unlink(tempfn);

    /* Set the default (initial) shape of the image; copy some parameters
       (nrows, ncols) from the command line */
    /* FIXME:  Get nrows, ncols from elsewhere than globals? */
    image->imgname   = imgname;
    image->curview   = 0;
    image->curformat = 0;
    vxl_init_raw_format(&image->rawform);
    /* Next line sets type, bigend, and nechoes = dimen[3] */
    decode_format(&image->rawform, image->curformat);
    image->length = image->imgstat.st_size;
    slen = nrows*ncols*exim_sizeof_type(image->rawform.type);
    image->rawform.rank = 4;
    image->rawform.dimen[0] = image->length / slen;
    image->rawform.dimen[1] = nrows;
    image->rawform.dimen[2] = ncols;
    image->rawform.dimen[3] = 1;
    image->rawform.current[0] = 0;
    image->rawform.current[1] = 0;
    image->rawform.current[2] = 0;
    image->rawform.current[3] = 0;
    for (n = 0; n < VXL_MAX_ECHO; n++) {
	image->rawform.minint[n] = 0;
	image->rawform.maxint[n] = 255; }
    image->rawform.xdim = 2;
    image->rawform.ydim = 1;
    image->rawform.edim = 3;

    /* Allocate voxel arrays */
    vxl_init_array(&image->volimg);
    vxl_init_array(&image->sect);

    /* Read the entire image file into a voxel array.  The image is
       always read as a long 1-dimensional string of bytes, to ensure
       that the entire image file is read, even if the assumed number
       of rows, columns, or bytes per voxel is incorrect.  Then the
       assumed shape of the image is set, which simply ignores any
       fractional voxels, rows, or slices.*/
    rank = 1;                /* Read simply as a byte string */
    dimen[0] = image->imgstat.st_size;
    vxl_read_uchar_data(&image->volimg, image->imgin, rank, dimen);

    /* Allocate memory for the X representation of the image.
       Allow 4 bytes per voxel for XImage. */
    /* FIXME: This will break for dynamic changes in nrows and ncols
       or the image display window size. */
    /* FIXME: Do a better job of computing the slice size */
    nx = image->rawform.dimen[2] * MAX_ZOOM;
    ny = image->rawform.dimen[1] * MAX_ZOOM;
    slen = 4 * nx * ny * sizeof(unsigned char);
    xidata  = malloc(slen);
    if (xidata == 0) 
        fatal("Unable to allocate image memory");
    ximage = XCreateImage(display, visinfo.visual, visinfo.depth,
	ZPixmap, 0, xidata, nx, ny, 16, 0);
    image->ximage = ximage;
    if (0) describeXImage(ximage);

    /* Free temporary memory and exit */
    free(testfile);
    fprintf(stderr, "done.\n");
    return image;
}



/*------------------------------------------------------------------------
freeVoxelView - Free memory associated with a VoxelViewRec

This function frees the memory associated with a VoxelViewRec,
including all subordinate memory and the VoxelViewRec itself.  It also
closes the open file, if any.

NOTE:  The memory pointed at by view->data should be freed by the call
to XDestroyImage; in any case, we will not do it here.
-------------------------------------------------------------------------*/

void
freeVoxelView(
    struct VoxelViewRec *view)
{

    /* If no view, just return silently */
    if (view == 0) return;

    /* Otherwise free all the contents */
    if (view->ximage)  XDestroyImage(view->ximage);
    if (view->imgin)   fclose(view->imgin);
    if (view->imgfile) free((char *)(view->imgfile));
    vxl_free_array(&view->sect);
    vxl_free_array(&view->volimg);

    /* Free the VoxelViewRec itself */
    free(view);
    return;
}



/*----------------------------------------------------------------------------
mono2d_to_XImage - Convert monochrome 2d uchar voxel array to an XImage

Convert a monochrome two-dimensional unsigned char voxel array to X
format, correcting for number of bytes per voxel and endianness of the
XImage; convert 0 to a dark blue.  The conversion is done one scanline
at a time, as a compromise between loop overhead and keeping all the
working data in the cache.  Unmatched voxels in the XImage will be
silently set to dark blue (which indicates zero); unmatched voxels in
the voxel array will be silently ignored.

FIXME: This function should be generalized to convert an arbitrary 2D
image into an XImage using a specified color map, zoom, and possibly
flips and subimages.  It should also handle the cases where the the
image is translated relative to the XImage, or when less than the full
XImage needs to be defined(?).

FIXME:  This function is no longer specific to qdv and should be
moved into some more appropriate library code; perhaps an X interface
to the voxel kit?

FIXME:  It might be faster to use a color table than to compute the
TrueColor values on the fly.

FIXME: Generalize to support zooming, as does rgb2d_to_XImage().

PRIOR to 20000122, the scanline buffer pixel was declared static and
retained between calls; allocating and freeing the buffer each time as
is done now is slightly slower but thread safe.
-----------------------------------------------------------------------------*/

void
mono2d_to_XImage(
    XImage *ximage,          /* XImage to convert to */
    voxel_array *section)    /* Voxel array to convert from */
{
    unsigned long *pixel;    /* One scanline of pixel values for XImage */
    unsigned char *rci;      /* Index over input voxels */
    unsigned char *data;     /* Contents of the XImage */
    int width, height;       /* Size of XImage */
    int nbxi;                /* Nmbr bytes per pixel in the XImage */
    int bigxi;               /* Is the XImage big or little endian? */
    int redr, redl,          /* Shifts to convert gray value to RGB: */
	grnr, grnl,          /* ...shift gray right by redr bits then */
	blur, blul;          /* ...left by redl bits; ditto for grn and blu. */
    long n0, n1;             /* Size of current section */
    long nn0, nn1;           /* Lesser size of section and image */
    long i0, i1;             /* Index into current section */
    unsigned long dkblue;    /* Dark blue to mark zero voxels */
    unsigned char gray;

    /* Extract needed information about the destination XImage */
    data   = (unsigned char *)ximage->data;
    width  =  ximage->width;
    height =  ximage->height;
    nbxi   =  ximage->bits_per_pixel / 8;
    bigxi  = (ximage->byte_order == MSBFirst);

    /* Allocate buffer to use in conversion process */
    pixel = malloc(width*sizeof(unsigned long));
    if (pixel == 0)
	fatal("Unable to allocate pixel buffer");

    /* Determine the translation from gray values in the input
       image to TrueColor values for X.  This is rather messy
       in the general case, requiring that we shift the gray
       value right by so many bits to remove excess bits on the
       right, then shift left to position properly within the
       word.  This section computes the necessary shifts (but
       works only for 8-bit gray values. */
    mask2shift(visinfo.red_mask,   8, &redr, &redl);
    mask2shift(visinfo.green_mask, 8, &grnr, &grnl);
    mask2shift(visinfo.blue_mask,  8, &blur, &blul);

    /* Get information about the voxel array; find intersection */
    /* FIXME:  Next few lines look inside the voxel array structure */
    n0  = section->dimen[0];
    nn0 = (n0 < height) ? n0 : height;
    n1  = section->dimen[1];
    nn1 = (n1 < width) ? n1 : width;
    dkblue = (0x80 >> blur) << blul;   /* Dark blue */

    /* Loop over scanlines in this slice of the image */
    for (i0 = 0; i0 < nn0; i0++) {
	rci = (unsigned char *)section->data + i0*n1;

	/* Convert gray values from file to RGB pixel values */
	for (i1 = 0; i1 < nn1; i1++) {
	    
	    /* Convert each gray value to the corresponding RGB pixel 
	       value, or dark blue if it's BACKGROUND (0) */
	    gray = *rci++;            
	    if (gray != 0)
		pixel[i1] = ((gray >> redr) << redl) +
		            ((gray >> grnr) << grnl) +
	                    ((gray >> blur) << blul) ;
	    else
		pixel[i1] = dkblue;
	}

	/* Pad the row with zeros as needed */
	for (i1 = n1; i1 < width; i1++)
	    pixel[i1] = dkblue;      /* Dark blue */

	/* Convert pixel values to the X server size and byte order */
	exim_ulong_to_bytes(width, data+i0*nbxi*width, pixel, nbxi, bigxi);
    }

    /* Set unmatched rows of the XImage to dark blue */
    if (n0 < height) {
	for (i1 = 0; i1 < width; i1++)  pixel[i1] = dkblue;
	for (i0 = nn0; i0 < height; i0++)
	    	exim_ulong_to_bytes(width, data+i0*nbxi*width,
                                    pixel, nbxi, bigxi);
    }

    free(pixel);
    return;
}



/*----------------------------------------------------------------------------
rgb2d_to_XImage - Convert 2d rgb voxel array to an XImage

This function converts a two-dimensional unsigned char rgb voxel array
to an XImage suitable for display on an X server.  The conversion is
done one scanline at a time, as a compromise between loop overhead and
keeping all the working data in the cache.  Unmatched voxels in the
voxel array will be silently ignored; unmatched voxels in the XImage
will be silently set to zero.

FIXME: This function should be generalized to convert an arbitrary 2D
image into an XImage using a specified zoom (DONE), and possibly flips
and subimages.  It should also handle the cases where the the image is
translated relative to the XImage, or when less than the full XImage
needs to be defined(?).

FIXME: This function is not specific to qdv and should be moved into
some more appropriate library code; perhaps an X interface to the
voxel kit?

PRIOR to 20000122, the scanline buffer pixel was declared static and
retained between calls; allocating and freeing each time as is done
now is slightly slower but thread safe.
-----------------------------------------------------------------------------*/

void
rgb2d_to_XImage(
    XImage *ximage,          /* XImage to convert to */
    voxel_array *section,    /* Voxel array to convert from */
    int zoom)                /* Zoom factor to use */
{
    unsigned long *pixel;    /* One scanline of pixel values for XImage */
    unsigned char *rci;      /* Index over input voxels */
    unsigned char *data;     /* Contents of the XImage */
    int width, height;       /* Size of XImage */
    int nbxi;                /* Nmbr bytes per pixel in the XImage */
    int bigxi;               /* Is the XImage big or little endian? */
    int redr, redl,          /* Shifts to convert gray value to RGB: */
	grnr, grnl,          /* ...shift gray right by redr bits then */
	blur, blul;          /* ...left by redl bits; ditto for grn and blu. */
    long n0, n1;             /* Size of current section */
    long nn0, nn1;           /* Lesser size of section and image */
    long i0, i1;             /* Index into current section */
    unsigned long pixval;    /* Value of current pixel */
    long iz;                 /* Index over replicated pixels for zoom */

    /* Extract needed information about the destination XImage */
    data   = (unsigned char *)ximage->data;
    width  =  ximage->width;
    height =  ximage->height;
    nbxi   =  ximage->bits_per_pixel / 8;
    bigxi  = (ximage->byte_order == MSBFirst);

    /* Check that the requested zoom is reasonable */
    if ( zoom < 1 || zoom > 3 )
        fatal("Requested zoom is too large or too small");

    /* Allocate buffer to use in conversion process */
    pixel = malloc(width*sizeof(unsigned long));
    if (pixel == 0)
	fatal("Unable to allocate pixel buffer");

    /* Determine the translation from rgb values in the input
       image to TrueColor values for X.  This is rather messy
       in the general case, requiring that we shift the red
       value right by so many bits to remove excess bits on the
       right, then shift left to position properly within the
       word; and similarly for green and blue.  This section 
       computes the necessary shifts (but works only for 8-bit 
       rgb values. */
    mask2shift(visinfo.red_mask,   8, &redr, &redl);
    mask2shift(visinfo.green_mask, 8, &grnr, &grnl);
    mask2shift(visinfo.blue_mask,  8, &blur, &blul);

    /* Get information about the voxel array; find intersection */
    /* FIXME:  Next few lines look inside the voxel array structure */
    n0  = section->dimen[0];
    nn0 = (n0*zoom < height) ? n0 : height/zoom;
    n1  = section->dimen[1];
    nn1 = (n1*zoom < width) ? n1 : width/zoom;

    /* Loop over source scanlines in this slice of the image */
    for (i0 = 0; i0 < nn0; i0++) {
	rci = (unsigned char *)section->data + 3*i0*n1;

	/* Convert gray values from file to RGB pixel values */
	for (i1 = 0; i1 < nn1; i1++) {
	    
	    /* Convert each gray value to the corresponding RGB pixel */
            pixval = ((rci[0] >> redr) << redl) +
		     ((rci[1] >> grnr) << grnl) +
		     ((rci[2] >> blur) << blul) ;

            /* Replicate pixels in X as needed */
            for (iz = 0; iz < zoom; iz++)
                pixel[i1*zoom+iz] = pixval;

            /* Step to the next source pizel */
	    rci += 3;
	}

	/* Pad the row with zeros as needed */
	for (i1 = n1*zoom; i1 < width; i1++)
	    pixel[i1] = 0;

	/* Convert pixel values to the X server size and byte order */
        for (iz = 0; iz < zoom; iz++)
            exim_ulong_to_bytes(width, data+(i0*zoom+iz)*nbxi*width, 
                                pixel, nbxi, bigxi);  
    }

    /* Set unmatched rows of the XImage to zeros */
    if (n0 < height) {
	for (i1 = 0; i1 < width; i1++)  pixel[i1] = 0;
	for (i0 = nn0*zoom; i0 < height; i0++)
	    	exim_ulong_to_bytes(width, data+i0*nbxi*width,
                                    pixel, nbxi, bigxi);
    }

    free(pixel);
    return;
}



/*---------------------------------------------------------------------------
showSlice - Display the selected slice in an X window

This function reads the current slice, converts it to the right format
for X, and displays it in an X window.
----------------------------------------------------------------------------*/

void
showSlice(
    struct VoxelViewRec *image,   /* Description of image and view */
    Drawable win)                 /* Window or pixmap to draw into */
{   
    voxel_array *vol;             /* Volume image to extract a 2d slice from */
    voxel_array *slc;             /* A 2d rgb slice to be displayed */
    struct vxl_raw_format *fmt;   /* How to format image for display */
    int width, height;            /* Size of image or window */
    size_t slen;                  /* Length of something or another */
    int iaxis;                    /* Current axis/view */

    /* If no current image, just clear the window and return */
    if (image == 0) {
	XClearWindow(display, win);
	return;
    }

    /* Find the structures that will be needed */
    slc = &image->sect;
    vol = &image->volimg;
    fmt = &image->rawform;

    /* Recompute the number of slices in the image.
       FIXME:  Should be redundant, since this is computed elsewhere
       when the # of rows, cols, echoes, or data format changes. */
    slen = fmt->dimen[1] * fmt->dimen[2] * fmt->dimen[3] * 
	exim_sizeof_type(fmt->type);
    fmt->dimen[0] = image->length / slen;

    /* If outside the image, just clear the window and return */
    iaxis = image->curview;
    if ( fmt->current[iaxis] < 0 || 
         fmt->current[iaxis] >= fmt->dimen[iaxis] ) {
	XClearWindow(display, win);
	return;
    }

    /* Adjust the window size to match the current zoom factor */
    XtVaSetValues(canvas,
                  XmNheight, 256*zoomAB,
                  XmNwidth,  256*zoomAB,
                  NULL);

    /* Extract the desired slice from the voxel array in memory. */
    vxl_raw_to_rgb2d(slc, vol, fmt);

    /* Convert the current section to an XImage and write to display */
    /* FIXME:  Allow window to be smaller or larger than image */
    rgb2d_to_XImage(image->ximage, &image->sect, zoomAB);
    width  = image->ximage->width;
    height = image->ximage->height;
    XPutImage(display, win, viewer_gc, image->ximage, 0, 0, 0, 0,
	      width, height);

    return;
}




/*---------------------------------------------------------------------------
showStatus - Display status for the selected slice in an X window

This function displays some status information about the current slice.

----------------------------------------------------------------------------*/

#define MAXSTAT  100        /* Max length of status string */

void
showStatus(
    struct VoxelViewRec *image,   /* Description of image and view */
    Drawable win)                 /* Window or pixmap to draw into */
{ 
    int font_height;            /* Height of font used for label */
    char *nbe;                  /* N bytes per voxel and endianness */
    char *view;                 /* Description of view displayed */
    char status[MAXSTAT+1]; 
    size_t len;                 /* Length of something or another */
    const char *code;           /* Code name for current image */
    int sx, sy;                 /* Position for current string */
    const char *s, *s1;

    /* Prepare to display new status information */
    XClearWindow(display, win);
    font_height = font_info->ascent + font_info->descent;
    sx = 9;
    sy = font_height;

    /* Show whether A or B image, or none, is being displayed */
    if (image != 0)   code = image->imgname;
    else              code = "No current image";
    len = snprintf(status, MAXSTAT, "Image: %s", code);
    XDrawString(display, win, status_gc, sx, sy, status, len);
    sy += font_height;

    /* If no current image, skip everything else */
    if (image == 0)  return;

    /* Show the file name, skipping the directory path */
    s = s1 = image->imgfile;
    for ( ; *s; s++) if (*s == '/') s1 = s+1;
    len = snprintf(status, MAXSTAT, "File: %s", s1);
    XDrawString(display, win, status_gc, sx, sy, status, len);
    sy += font_height;

    /* Show the file format used to interpret the data file */
    nbe = describe_format(image->curformat);
    len = strlen(nbe);
    XDrawString(display, win, status_gc, sx, sy, nbe, len);
    sy += font_height;

    /* Show intensity window */
    /* FIXME: min and max gray is per echo now? */
    len = snprintf(status, MAXSTAT, "Window:  %d-%d", 
		   (int)(image->rawform.minint[0]),
		   (int)(image->rawform.maxint[0]));
    XDrawString(display, win, status_gc, sx, sy, status, len);
    sy += font_height;

    /* Show image view being used */
    if (image->curview == 0)
	view = "View: +X +Y";
    else if (image->curview == 1) 
	view = "View: +X +Z";
    else if (image->curview == 2) 
	view = "View: +Y +Z";
    else /* Impossible */
	view = "Invalid view";
    len = strlen(view);
    XDrawString(display, win, status_gc, sx, sy, view, len);
    sy += font_height;

    /* Show current slice being displayed */
    if (image->curview == 0)
	len = snprintf(status, MAXSTAT, "Slice:  Z = %ld/%ld",
		       image->rawform.current[0], image->rawform.dimen[0]);
    else if (image->curview == 1) 
	len = snprintf(status, MAXSTAT, "Slice: Y = %ld/%ld",
		       image->rawform.current[1], image->rawform.dimen[1]);
    else if (image->curview == 2) 
	len = snprintf(status, MAXSTAT, "Slice: X = %ld/%ld",
		       image->rawform.current[2], image->rawform.dimen[2]);
    else /* Impossible */
	len = 0;
    XDrawString(display, win, status_gc, sx, sy, status, len);
    sy += font_height;

    /* Show the current zoom for this image */
    len = snprintf(status, MAXSTAT, "Zoom: %d", zoomAB);
    XDrawString(display, win, status_gc, sx, sy, status, len);
    sy += font_height;

    /* Show image locking, if any */
    if (lockAB)
        len = snprintf(status, MAXSTAT, "Lock: A/B");
    else
        len = snprintf(status, MAXSTAT, "Lock: None");
    XDrawString(display, win, status_gc, sx, sy, status, len);
    sy += font_height;

    return;
}
