/***************************************************************************
rand.c - Functions for random deviates and related stuff

Author:  Paul Hughett

Copyright (c) 1998-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


The functions in this file are used to generate random numbers of
various sorts.  The theory under which they are designed is discussed
in Press et al [1], but the implementation is new.

[1] William H. Press, Saul A. Teukolsky, William T. Vetterling and
Brian P. Flannery, Numerical Recipes in C: The Art of Scientific
Computing, 2nd Edition, Cambridge University Press, 1992.

*****************************************************************************/

static const char rcsid[] = "$Id: rand.c,v 1.5 2001/10/05 21:22:20 hughett Exp $ Release $Name:  $";

/* Following two header files should not be needed here but are; they cannot
   be included within image_inc.h because that excites a bug in SunOS.  */

#include <stdio.h>
#include <stdlib.h>

#include "errm.h"
#include "memm.h"
#include "rand.h"


/***************************************************************************
ran1 - A high-quality shuffled minimal standard random number generator

This set of functions iplements the minimal standard random number
generator of Park and Miller, with Bays-Durham shuffling, as
transmitted via Press et al [1].  Multiple independent generators may
be used simultaneously; each generator is instantiated by a state
vector created by calling ran1_init.  A seed value provided to
ran1_init selects the starting point in the sequence; any value
between 1 and RAN1_MAX is permitted.  These routines are thread-safe
provided that no state vector is shared between threads.  The period
of the generator is 2^31-2, or about 2.1E9 and it is believed to be
free of serial correlations over any interval much less than the
period.  The output of a generator may be obtained as a uniform float
variate in the range 0 to 1 exclusive, as a uniform long variate in
the range 1..RAN1_MAX inclusive, or as a equiprobable choice over the
integers 0..N-1 for some N chosen by the user; these options may be
freely intermixed.

FIXME: Add other useful distributions, such as normal.
***************************************************************************/

/* Constants for the generator and shuffler */
#define A 16807         /* Multiplier */
#define M 2147483647    /* Modulus */
#define Q 127773        /* Approximate quotient M/A */
#define R 2836          /* Remainder of quotient M/A */
#define N 32            /* Number of values in shuffle array */

/* The basic multiplicative congruential generator */
#define RAN1(x)  (x = A*(x%Q) - R*(x/Q), x = (x < 0) ? x + M : x)


/* The state vector is an array of N+2 longs.  The first element state[0]
   is the current seed value for the generator; the second element
   state[1] is the index of the next entry in the shuffle array to be
   output; and the remaining N values are the shuffled random values
   themselves. */



/*-------------------------------------------------------------------------
ran1_init - Create initial state for a random number generator
--------------------------------------------------------------------------*/

void *
ran1_init(
    long seed)             /* Seed value to select starting place */
{
    long *state;           /* State vector for random sequence */
    int i;

    /* Error if seed is invalid */
    if (seed <= 0 || seed >= M)
	fatal("ran1_init: Invalid seed value");

    /* Allocate a state vector */
    state = mallock((N+2)*sizeof(long));

    /* Compute and throw away a few warmup values */
    state[0] = seed;
    for (i = 0; i < 8; i++) RAN1(state[0]);

    /* Preload the shuffling array and set initial index */
    state[1] = RAN1(state[0]) % N;
    for (i = 0; i < N; i++)   state[i+2] = RAN1(state[0]);

    return (void *)state;
}



/*------------------------------------------------------------------------
ran1_free - Free the state vector for a random number generator
------------------------------------------------------------------------*/

void
ran1_free(
    void *state)               /* State vector for this sequence */
{
    if (state != 0)
        free(state);
    return;
}



/*-------------------------------------------------------------------------
ran1_long - Return a uniformly random number in long format
--------------------------------------------------------------------------*/

long
ran1_long(
    void *state)              /* State vector for this sequence */
{
    long *s = (long *)state;  /* The state vector as a long array */
    long x;                   /* The random number to be output */
    long i;                   /* Slot to replace in shuffle array */

    i = s[1];
    if (i < 0 || i >= N)
	fatal("ran1_long: index out of range");
    x = s[i+2];
    s[i+2] = RAN1(s[0]);
    s[1] = x % N;         /* FIXME:  Use high-order bits */

    return x;
}



/*---------------------------------------------------------------------------
ran1_choose - Return a random choice from integers 0...N-1

Given a integer N, this function returns one of the integers 0..N-1
with equal probability.  The range N is typed as an int, since the
ran1 generator used probably won't give good results for very large
values of N.

FIXME:  Let N be as large as the period?
---------------------------------------------------------------------------*/

int
ran1_choose(
    void *state,           /* State vector for this sequence */
    int  n)                /* Number of ints to choose from */
{
    long div, lim, x;

    div = (M-1)/n;
    lim = div*n;

    do {
	x = ran1_long(state);
    }
    while (x > lim);

    return (x-1)/div;
}



/*---------------------------------------------------------------------------
ran1_float - Return a uniform random variate as a float value

The value returned is in the range [0, 1]; note that the granularity
might be less than the float epsilon.

FIXME: Test that this works correctly and produces good randomness.
---------------------------------------------------------------------------*/

float ran1_float(
    void *state)              /* State vector for this sequence */
{
    return (ran1_long(state)-1) / (float)(RAN1_MAX-1);
}



/*---------------------------------------------------------------------------
ran1_double - Return a uniform random variate as a double value

The value returned is in the range [0, 1]; note that the granularity
might be less than the double epsilon.

FIXME: Test that this works correctly and produces good randomness.
---------------------------------------------------------------------------*/

double ran1_double(
    void *state)              /* State vector for this sequence */
{
    return (ran1_long(state)-1) / (double)(RAN1_MAX-1);
}
