/****************************************************************************
reduce.c - Voxel Kit routines for reduction operations

Author:  Paul Hughett

Copyright (c) 1998-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file contains code for reduction
operations and related functions.

Reduction operations on arrays are defined as operations that process
along selected axes of an array to obtain a scalar or an array of
lower rank; some functions which consist of reduction plus additional
operations are also included here.  Both unary and binary operations
are permitted.  Many of these functions (should) permit reduction
along specified axes only, to yield an array of lower rank rather than
a scalar.  Examples of reduction operations include array sum (as well
as row and column sums), array product, array minimum, array maximum,
norms, dot product, mean, variance, covariance, and correlation.

These functions all face similar algorithmic problems in choosing the
access patterns to use a multi-level cache efficiently.  Several
different algorithms may currently coexist in this file, while I
experiment with different trade-offs between implementation time,
readability, and efficiency; the trend over time can be expected to be
toward improving efficiency, even at the cost of implementation time
and readability, since these functions are likely to be major
contributors to the total run time.

Note that the algorithms used here are designed to maximize locality
within a multilevel cache architecture, thereby decreasing the run
time.  If you don't understand what that means, you probably shouldn't
touch this code, under penalty of a possible order-of-magnitude
increase in run time.  

FIXME: Think through the right way to implement these functions to
keep as much common as possible.  Should the result array(s) always be
double, or is it necessary to allow other result types?  Can we assume
that the result is normally sufficiently low dimension that efficient
access is not a serious performance issue?  Note that the expensive
part of dot, norm2, (weighted) mean, variance, covariance, and
correlation consists of sums of powers or products of arrays.  For
example, computing the weighted mean requires the sum of the weights
and the sum of the weights times the data.  So a possible set of core
operations might include sumx, sumxx, sumxy, sumxxy(for weighted
variance), sumxyz, amin, amax.  It is probably convenient if compound
operations such as sumxy also optionally generate sumx and sumy.  Then

    dot      uses  sumxy  or  sumwxy
    norm2    uses  sumxx  or  sumwxx
    mean     uses  sumx   or  sumwx
    var      uses  sumxx  or  sumwxx, plus mean
    cov      uses  sumxy  or  sumwxy, plus mean and var
    corr     uses  cov, var

FIXME: The vxl_mean() function permits reduction on selected axes
only; modify the other functions (where sensible) to do the same.
Propagate the changes up into Pyvox.

FIXME: Add Pyvox test cases for these functions.
*****************************************************************************/

static const char rcsid[] = "$Id: reduce.c,v 1.2 2002/02/28 20:57:37 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "bips.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"



/*-------------------------------------------------------------------------
vxli_reduce_setup - Initialize reduction loop over source rows

This function determines the dimensions of the destination array in a
reduction operation, plus a few other useful parameters.  It is used in
conjunction with vxli_reduce_first() and vxli_reduce_next() to implement
reduction operations over a specified set of axes.  The standard usage is:

    vxli_reduce_setup(...);
    ...allocate the destination(s)
    vxli_reduce_first(...);
    do {
        ...operate on current source row
    }  while ( vxli_reduce_next(...) );


Setting naxes to a negative number means to reduce on all axes;
setting naxes to zero means to reduce on no axes, which is legal but
usually devolves into a expensive copy; setting naxes to the rank or
greater is legal but means that axes contains duplicates.  The
elements in axes may be in any order, and may be repeated without
harm.  Setting one of these elements to a value in [0, rank-1] means
to reduce that axis, counting from the left; setting to [-1, -rank]
means to reduce that axis, counting from the right.  (That is, -1 is
equivalent to rank-1, or the last axis.)  All other values are
illegal.

See vxl_mean and other functions in this file for examples to copy.
--------------------------------------------------------------------------*/

void static inline
vxli_reduce_setup(
    int   naxes,         /* In: Number of axes to reduce along */
    int  *axes,          /* In: Axes to reduce along */
    int   srank,         /* In: Rank of source */
    long *sdimen,        /* In: Dimensions of source */
    int  *drank,         /* Out: Rank of destination */
    long *ddimen,        /* Out: Dimensions of destination */
    int  *dimmap,        /* Out: Map dest axis to corresp src axis */
    long *dstep,         /* Out: Step size along last destination axis */
    long *sperd)         /* Out: Number of src elements per dest element */
{
    int reduce[VXL_MAX_RANK];   /* Reduce this axis? */
    int i, iaxis, irank;
    long spd = 1;

    /* Convert axes definition to tabular format */
    if ( naxes < 0 )
	for (i = 0; i < VXL_MAX_RANK; i++)
	    reduce[i] = 1;
    else {
	for (i = 0; i < VXL_MAX_RANK; i++)
	    reduce[i] = 0;
	for (i = 0; i < naxes; i++) {
	    iaxis = axes[i];
	    if  (0 <= iaxis && iaxis < srank)
		reduce[iaxis] = 1;
	    else if ( -srank <= iaxis && iaxis < 0 )
		reduce[srank+iaxis] = 1;
	    else
		fatal("Reduction axis exceeds rank of array");
	}
    }

    /* Determine destination dimensions */
    irank = 0;
    for (i = 0; i < srank; i++)
	if ( !reduce[i] ) {
	    dimmap[irank] = i;
	    ddimen[irank++] = sdimen[i];
	}
	else {
	    spd *= sdimen[i];
	}
    *drank = irank;
    *sperd = spd;
    if ( reduce[srank-1] )
	*dstep = 0;
    else
	*dstep = 1;

    return;
}



/*-------------------------------------------------------------------------
vxli_reduce_first - Initialize reduction loop over source rows

This function initializes a loop over rows of the source images(s) to
compute some reduction operation; it works in conjunction with
vxli_reduce_setup and vxli_reduce_next.  See vxli_reduce_setup for the
typical usage.
--------------------------------------------------------------------------*/

void static inline
vxli_reduce_first(
    int   srank,         /* (In) Rank of source */
    long *scoord,        /* (Out) Coords of first row in source */
    int   drank,         /* (In) Rank of destination */
    long *dcoord)        /* (Out) Coords of first row in destination */
{
    int i;

    /* Zero the source and destination row coordinates */
    for (i = 0; i < srank; i++)
	scoord[i] = 0;
    for (i = 0; i < drank; i++)
	dcoord[i] = 0;

    return;
}



/*-------------------------------------------------------------------------
vxli_reduce_next - Step to next row in a reduction loop

This function steps to the next source row in a reduction loop.  See the
function vxli_reduce_first for usage instructions.
-------------------------------------------------------------------------*/

int static inline
vxli_reduce_next(
    int srank,             /* (In) Rank of the source image */
    long *sdimen,          /* (In) Dimensions of the source image */
    long *scoord,          /* (Out) Current row in the source image */
    int drank,             /* (In) Rank of the destination image */
    int  *dimmap,          /* (In) Map dest axis to corresp source axis */
    long *dcoord)          /* (Out) Current row in the destination image */
{
    int more = 0;          /* More rows to process? */
    int i;

    /* Step to next source row */
    for (i = srank-2; i >= 0; i--) {
	if ( ++scoord[i] < sdimen[i] ) {
	    more = 1;
	    break;
	}
	else
	    scoord[i] = 0;
    }

    /* Find corresponding destination row */
    for (i = 0; i < drank; i++)
	dcoord[i] = scoord[dimmap[i]];

    /* Return 1 until last row has been done */
    return more;
}



/*-------------------------------------------------------------------------
vxl_amax - Find maximum value in a voxel array

This function finds the maximum value contained in a voxel array and returns
it as a double value.

FIXME: Assumes that type double is no shorter than any other supported
internal type.
---------------------------------------------------------------------------*/

double
vxl_amax(
    voxel_array *src)           /* Array to find max of */
{
    int type = 0;               /* Data type */
    long ds = 0, ss = 1;        /* Dest and source strides */
    long count;                 /* Total number of elements */
    int err;                    /* Error flag */
    unsigned char               /* Running maximum */
	 val[sizeof(double)];

    /* Check validity of source; count elements */
    if ( src == 0 || src->magic != VXL_MAGIC ||
	 src->data == 0 ||
	 exim_sizeof_intype((type = src->type)) == 0 )
	fatal("Invalid source array");
    count = vxli_count(src);
    if (count < 1)
	fatal("Source array is empty");

    /* Initialize running max with first value */
    err = bips_copy(1, type, val, ds, src->data, ss);
    if (err)
	fatal("Error reported by bips_copy");

    /* Compute the array maximum */
    err = bips_max_set(count, type, val, ds, src->data, ss);
    if (err)
	fatal("Error reported by bips_max_set");

    /* Convert maximum to double and return it */
    return exim_get_value(val, type, 0);
}



/*-------------------------------------------------------------------------
vxl_amin - Find minimum value in a voxel array

This function finds the minimum value contained in a voxel array and
returns it as a double value.

FIXME: Assumes that type double is no shorter than any other supported
internal type.
---------------------------------------------------------------------------*/

double
vxl_amin(
    voxel_array *src)           /* Array to find min of */
{
    int type = 0;               /* Data type */
    long ds = 0, ss = 1;        /* Dest and source strides */
    long count;                 /* Total number of elements */
    int err;                    /* Error flag */
    unsigned char               /* Running minimum */
	 val[sizeof(double)];

    /* Check validity of source; count elements */
        if ( src == 0 || src->magic != VXL_MAGIC ||
	     src->data == 0 ||
	     exim_sizeof_intype((type = src->type)) == 0 )
	fatal("Invalid source array");
    count = vxli_count(src);
    if (count < 1)
	fatal("Source array is empty");

    /* Initialize running min with first value */
    err = bips_copy(1, type, val, ds, src->data, ss);
    if (err)
	fatal("Error reported by bips_copy");

    /* Compute the array minimum */
    err = bips_min_set(count, type, val, ds, src->data, ss);
    if (err)
	fatal("Error reported by bips_min_set");

    /* Convert minimum to double and return it */
    return exim_get_value(val, type, 0);
}



/*-------------------------------------------------------------------------
vxl_aprod - Compute product of all voxels in a voxel array

This function computes the product of all elements in a voxel array and
returns the result as a double value.

FIXME: Assumes that type double is no shorter than any other supported
internal type.  This will cause problems when we add complex values.

FIXME: Make this faster by rewriting to use bips.  Compute by rows for
better round-off properties?
---------------------------------------------------------------------------*/

double
vxl_aprod(
    voxel_array *src)         /* Array to take product of */
{
    int type = 0;             /* Data type */
    long count;               /* Total voxels in the array */
    double res;               /* Result */
    long i;

    /* Check validity of source; count elements */
        if ( src == 0 || src->magic != VXL_MAGIC ||
	     src->data == 0 ||
	     exim_sizeof_intype((type = src->type)) == 0 )
	fatal("Invalid source array");
    count = vxli_count(src);
    if (count < 1)
	fatal("Source array is empty");

    /* Sum all the voxel values */
    res = 1.0;
    for (i = 0; i < count; i++) {
	res *= exim_get_value(src->data, src->type, i);
    }

    return res;
}



/*-------------------------------------------------------------------------
vxl_asum - Compute sum of all voxels in a voxel array

This function sums all the voxel values in an array and returns the sum
as a double value.

FIXME: Assumes that type double is no shorter than any other supported
internal type.  This will cause problems when we add complex values.

FIXME: Make this faster by rewriting to use bips.  Avoid roundoff
error by summing over rows or slices first.
---------------------------------------------------------------------------*/

double
vxl_asum(
    voxel_array *src)         /* Array to take sum of */
{
    int type = 0;             /* Data type */
    long count;               /* Total voxels in the array */
    double sum;               /* Sum of voxels */
    long i;

    /* Check validity of source; count elements */
        if ( src == 0 || src->magic != VXL_MAGIC ||
	     src->data == 0 ||
	     exim_sizeof_intype((type = src->type)) == 0 )
	fatal("Invalid source array");
    count = vxli_count(src);
    if (count < 1)
	fatal("Source array is empty");

    /* Sum all the voxel values */
    sum = 0.0;
    for (i = 0; i < count; i++) {
	sum += exim_get_value(src->data, src->type, i);
    }

    return sum;
}



/*-------------------------------------------------------------------------
vxl_dot - Compute weighted vector dot product of two voxel arrays

This function computes the weighted vector dot product of two voxel
arrays.  If weight is a null pointer, then a uniformly weighted dot
product is computed; otherwise, weight supplies the weights for each
voxel.  If src2 is a null pointer, then a simple weighted sum is
computed.  The src1 argument may not be zero.  The arrays src1, src2,
and weight must be the same rank and dimensions.  The arrays src1,
src2, and weight may be of any type and will be converted to double
for the computation.
--------------------------------------------------------------------------*/

double
vxl_dot(
    voxel_array *src1,          /* First array */
    voxel_array *src2,          /* Second array, or zero */
    voxel_array *weight)        /* Voxel weights, or zero */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int rank;                   /* Rank of the arrays */
    double res, result = 0.0;   /* Norm of array */
    long coord[VXL_MAX_RANK];   /* Coordinates of each row */
    double *abuf=0, *bbuf=0;    /* Working storage for a row */
    unsigned char *s1=0, *s2=0; /* Pointers to beginning of each source row */
    unsigned char *w=0;         /* Pointer to beginning of each weight row */
    long nelem;                 /* Number of elements in a row */
    int err;                    /* Error indicator */
    long i;

    /* Check validity of source 1 */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	 exim_sizeof_intype((stype1 = src1->type)) == 0 )
	fatal("Invalid source 1 array");
    rank  = src1->rank;
    nelem = src1->dimen[rank-1];

    /* Check validity of source 2 */
    if ( src2 != 0 ) {
	if ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	     exim_sizeof_intype((stype2 = src2->type)) == 0 )
	    fatal("Invalid source 2 array");
	if ( !vxli_same_shape(src1, src2) )
	    fatal("Incompatible source 1 and 2 arrays");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     exim_sizeof_intype((wtype = weight->type)) == 0 )
	    fatal("Invalid weight array");
	if ( !vxli_same_shape(weight, src1) )
	    fatal("Source 1 and weight arrays are incompatible");
    }

    /* Allocate working storage for one row */
    abuf = mallock(nelem*sizeof(double));
    bbuf = mallock(nelem*sizeof(double));

    /* Loop over rowss */
    vxli_firstrow(rank, coord);
    do {

	/* Convert this row of src1 to type double */
	s1 = vxli_locate(src1, coord, 1);
        err = bips_double(nelem, abuf, 1, stype1, s1, 1);
        if (err)
	    fatal("Error calling BIPS");

	/* Multiply by source 2, if provided */
	if (src2 != 0) {
	    s2 = vxli_locate(src2, coord, 1);
	    err = bips_double(nelem, bbuf, 1, stype2, s2, 1);
	    if (err)   fatal("Error calling BIPS");
	    for (i = 0; i < nelem; i++)
		abuf[i] *= bbuf[i];
	}
	
	/* Multiply by weights, if provided */
	if (weight != 0) {
	    w = vxli_locate(weight, coord, 1);
            err = bips_double(nelem, bbuf, 1, wtype, w, 1);
	    if (err)   fatal("Error calling BIPS");
	    for (i = 0; i < nelem; i++)
		abuf[i] *= bbuf[i];
	}

	/* Sum over elements in this row */
	res = 0.0;
	for (i = 0; i < nelem; i++)
	    res += abuf[i];
	result += res;

    } while (vxli_nextrow(rank, src1->dimen, coord));

    /* Free working storage and return result */
    free(abuf);
    free(bbuf);
    return result;
}



/*-------------------------------------------------------------------------
vxl_mean - Compute mean along selected axes of an image

This function computes weighted mean along specified axes of a voxel
array, returning the results as a double array of lesser dimension.
Setting the number of axes to zero means to reduce along all axes;
that is, to compute the weighted mean of the entire array.  If the
weights sum to zero for some destination element, then you get whatever
your platform provides for division by zero; this should be nan on
IEEE-754 platforms.

FIXME: Division by zero is possible and may cause problems if the platform
does not support IEEE-754 and nans.

FIXME: Provide sum of weights as optional output?

FIXME: Is there any reason to allow the option to return a result of type
other than double?
--------------------------------------------------------------------------*/

void
vxl_mean(
    voxel_array *dest,           /* Destination array */
    voxel_array *src1,           /* Source array */
    voxel_array *src2,           /* Weight array, or zero */
    int naxes,                   /* Number of axes to reduce along, or zero */
    int *axes)                   /* Axes to reduce along */
{
    voxel_array *sumw = 0;         /* Sums of weights on given axes */
    int stype1=0, stype2=0;        /* Types of the source arrays */
    int srank, drank;              /* Rank of source and destination */
    long *sdimen;                  /* Dimensions of source arrays */
    int  dimmap[VXL_MAX_RANK];     /* Map dest dimen to corres src dimen */
    long ddimen[VXL_MAX_RANK];     /* Dimensions of destination array */
    long scoord[VXL_MAX_RANK];     /* Coordinates of current row in source */
    long dcoord[VXL_MAX_RANK];     /* Coordinates of current row in dest */
    long nelem;                    /* Number of elements per source row */
    long dstep;                    /* Step size in destination */
    long sperd;                    /* Source elements per dest element */
    long ndest;                    /* Number of elements in destination */
    double *work1, *work2;         /* Two rows of working memory */
    unsigned char *srow1, *srow2;  /* Location of current src 1 and 2 rows */
    double *drow;                  /* Location of current dest row */
    double *wrow;                  /* Location of current row in sumw */
    int i, err;
    long k;

    /* Validate the first source array (data to average) */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	 exim_sizeof_intype((stype1 = src1->type)) == 0 )
	fatal("Invalid source 1 array");
    srank  = src1->rank;
    sdimen = src1->dimen;
    nelem  = sdimen[srank-1];

    /* Validate the second source array (weights) */
    if (src2 != 0) {
	if ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	     exim_sizeof_intype((stype2 = src2->type)) == 0 )
	    fatal("Invalid weight array");
	if ( !vxli_same_shape(src1, src2) )
	    fatal("Incompatible source and weight arrays");
    }

    /* Determine dimensions and other properties of destination */
    vxli_reduce_setup(naxes, axes, srank, sdimen,
		      &drank, ddimen, dimmap, &dstep, &sperd);

    /* Allocate and clear the destination array */
    vxl_alloc_array(dest, INTP_DOUBLE, drank, ddimen);
    vxl_const(dest, 0.0);
    ndest = vxli_count(dest);

    /* Allocate two rows of working storage */
    work1 = mallock(2*nelem*sizeof(double));
    work2 = work1 + nelem;

    /* Case 1: Weights are provided */
    if (src2) {

	/* Allocate voxel array to store sums of weights */
	sumw = vxl_create_array();
	vxl_alloc_array(sumw, INTP_DOUBLE, drank, ddimen);
	vxl_const(sumw, 0.0);

	/* Loop over source and destination rows */
	vxli_reduce_first(srank, scoord, drank, dcoord);
	do {
	    
	    /* Copy source row to buffer, casting to double */
	    srow1 = vxli_locate(src1, scoord, 1);
	    err = bips_double(nelem, work1, 1, stype1, srow1, 1);
	    if (err != 0)
		fatal("Error calling BIPS");

	    /* Copy weight row to buffer, casting to double */
	    srow2 = vxli_locate(src2, scoord, 1);
	    err   = bips_double(nelem, work2, 1, stype2, srow2, 1);
	    if (err != 0)
		fatal("Error calling BIPS");

	    /* Accumulate weights, and source times weights */
	    wrow = (double *)vxli_locate(sumw, dcoord, 1);
	    drow = (double *)vxli_locate(dest, dcoord, 1);
	    for (i = 0; i < nelem; i++) {
		wrow[i*dstep] += work2[i];
		drow[i*dstep] += work1[i] * work2[i];
	    }

	} while ( vxli_reduce_next(srank, sdimen, scoord, 
				   drank, dimmap, dcoord) );

	/* Divide by sum of weights for each dest element */
	drow = dest->data;
	wrow = sumw->data;
	for (k = 0; k < ndest; k++)
	    drow[k] /= wrow[k];

	/* Free the array used for sums of weights */
	vxl_destroy_array(sumw);

    }

    /* Case 2: No weights are provided */
    else {

	/* Loop over source and destination rows */
	vxli_reduce_first(srank, scoord, drank, dcoord);
	do {
	    
	    /* Copy source row to buffer */
	    srow1 = vxli_locate(src1, scoord, 1);
	    err = bips_double(nelem, work1, 1, stype1, srow1, 1);
	    if (err != 0)
		fatal("Error calling BIPS");

	    /* Accumulate source */
	    drow = (double *)vxli_locate(dest, dcoord, 1);
	    for (i = 0; i < nelem; i++)
		drow[i*dstep] += work1[i];

	} while ( vxli_reduce_next(srank, sdimen, scoord, 
				   drank, dimmap, dcoord) );

	/* Divide by number of source elements per dest element */
	drow = dest->data;
	for (k = 0; k < ndest; k++)
	    drow[k] /= (double)sperd;
    }

    /* Free working memory and return */
    free(work1);
    return;
}



/*-------------------------------------------------------------------------
vxl_minmax_ulong - Find min and max values in a ulong volume image

This function scans a unsigned long volume image to find the minimum
and maximum voxel values contained therein.  One purpose of this
function is to determine the scale factors to fit the actual data
range into a shorter type.

FIXME:  This should be generalized for arbitrary types of arrays.
-------------------------------------------------------------------------*/

void
vxl_minmax_ulong(
    unsigned long *minval,     /* Minimum value found */
    unsigned long *maxval,     /* Maximum value found */
    voxel_array *src)          /* Image to find min/max of */
{
    unsigned long *sdata;      /* Image data itself */
    long count;                /* Count of voxels in the image */
    unsigned long minv = ULONG_MAX;
    unsigned long maxv = 0;
    long i;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_ULONG)
	fatal("vxl_minmax_ulong: Invalid or empty source array");

    /* Scan for min/max */
    sdata = src->data;
    count = vxli_count(src);
    for (i = 0; i < count; i++) {
	if (sdata[i] < minv) minv = sdata[i];
	if (sdata[i] > maxv) maxv = sdata[i]; }

    /* Save the results */
    if (minval != 0) *minval = minv;
    if (maxval != 0) *maxval = maxv;

    return;
}



/*------------------------------------------------------------------------
vxl_moments - Compute central moments of an image

This function computes the total mass, center of gravity, and (2nd)
central moments of an image.  The moments are computed in physical
coordinates, as defined by the origin and spacing values.  It is the
caller's responsibility to allocate the necessary space for mass (one
double value), center (rank values), and moments (rank*rank values).

FIXME: What happens if one of the spacing values is zero?

FIXME: It's not particularly useful to compute moments of a rank-0
array, so we just outlaw it for the moment.  (Pun?  What pun?)
------------------------------------------------------------------------*/

void
vxl_moments(
    double *mass,        /* Total mass of the image; length 1 */
    double *center,      /* Center of gravity; length rank */
    double *moments,     /* Central moments; length rank*rank */
    voxel_array *src)    /* Image to take moments of */
{
    int rank;                   /* Rank of the image */
    int type;                   /* Type of the image */
    long *dimen;                /* Dimensions of the image */
    long ncols;                 /* Number of elements along last dimension */
    long coord[VXL_MAX_RANK];   /* Run over coordinates of the image */
    double mom[3];              /* Row moments from bips */
    void *row;                  /* Start address of an image row */
    int info;                   /* Status returned by bips */
    int i, j;

    /* Validate the arguments */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->data == 0 || src->rank < 1)
	fatal("Invalid or empty source array");
    rank  = src->rank;
    type  = src->type;
    dimen = src->dimen;
    ncols = dimen[rank-1];
    if (mass == 0 || center == 0 || moments == 0)
	fatal("Null destination for moments");

    /* Initialize moment arrays: Use center and moments initially to
       store the first and second non-central moments. */
    *mass = 0.0;
    for (i = 0; i < rank; i++) {
	center[i] = 0.0;
	for (j = 0; j < rank; j++) {
	    moments[i*rank+j] = 0.0;
	}
    }

    /* Sum row moments over the rows of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Compute moments for the current row */
	row = vxli_locate(src, coord, 1);
	info = bips_moments(ncols, mom, type, row, 1);
	if (info != 0)
	    fatal("Failure when calling BIPS");

	/* Accumulate moments over all rows; by symmetry, we need only
	   the lower triangle of the moments array. */
	*mass += mom[0];
	for (i = 0; i < rank-1; i++) {
	    center[i] += coord[i]*mom[0];
	    for (j = 0; j <= i; j++)
		moments[i*rank+j] += coord[i]*coord[j]*mom[0];
	}
	center[rank-1] += mom[1];
	for (j = 0; j < rank-1; j++)
	    moments[(rank-1)*rank + j] += coord[j]*mom[1];
	moments[rank*rank-1] += mom[2];
    }
    while ( vxli_nextrow(rank, dimen, coord) );

    /* Convert to central moments (in lower triangle only) */
    for (i = 0; i < rank; i++) {
	center[i] /= *mass;
	for (j = 0; j <= i; j++)
	    moments[i*rank+j] -= (*mass)*center[i]*center[j];
    }

    /* Translate to physical coordinates and reflect across diagonal */
    for (i = 0; i < rank; i++) {
	center[i] = src->spacing[i] * center[i] + src->origin[i];
	moments[i*rank+i] *= src->spacing[i] * src->spacing[i];
	for (j = 0; j < i; j++) {
	    moments[j*rank+i] = moments[i*rank+j]
		*= src->spacing[i] * src->spacing[j];
	}
    }

    return;
}



/*-------------------------------------------------------------------------
vxl_mop - Arbitrary moments of the elementwise product of three arrays

This function computes arbitrary moments of an array or the
elementwise product of two or three arrays.  The arrays must be of the
same rank and shape, but the rank must be at least 2; the arrays may
be of any type or mixture of types.  The moments calculation is always
done in double, regardless of the types of the arrays.  The moments to
be computed are given by an K by M int array, where K is the common
rank of the arrays and M is the number of moments to be computed.
Each row specifies one moment to be computed; each element of that row
gives the power to which the corresponding coordinate is to be raised.
The requested moments are returned in a one-dimensional array, in the
same order as the rows defining them.
---------------------------------------------------------------------------*/

void
vxl_mop(
    voxel_array *result,         /* Values of requested moments */
    voxel_array *moments,        /* Table of desired moments */
    voxel_array *array1,         /* First array */
    voxel_array *array2,         /* Second array, or zero if none */
    voxel_array *array3)         /* Third array, or zero */
{
    int rank;                    /* Common rank of the arrays */
    long *dimen;                 /* Common dimension of the arrays */
    long ncols;                  /* Number of elements along last dimension */
    int type1 = 0;               /* Types of the arrays */
    int type2 = 0, type3 = 0;
    long coord[VXL_MAX_RANK];    /* Run over coordinates of the arrays */
    long momcnt;                 /* Number of moments requested */
    int *momdef;                 /* Actual data area for moments definition */
    int maxpow;                  /* Maximum power requested in x */
    double *work1 = 0;           /* Working storage for two source rows */
    double *work2 = 0;
    double *xramp = 0;           /* Row of x coordinate values */
    double *rowmom = 0;          /* X-moments of a row */
    void *row1, *row2, *row3;    /* Start of a row in each source array */
    double *rdata;               /* Data area for result array */
    double mom;                  /* Partially computed moment value */
    long imom;                   /* Index over requested moments */
    long icol;                   /* Index over elements of a row */
    int  icor;                   /* Index over coordinate axes */
    int  ipow;                   /* Index over powers of a coordinate */
    int  info;                   /* Return value from BIPS */

    /* Validate first source array */
    if ( array1 == 0 || array1->magic != VXL_MAGIC ||
         array1->data == 0 || array1->rank < 2 )
	fatal("Array 1 is invalid");
    rank  = array1->rank;
    type1 = array1->type;
    dimen = array1->dimen;
    ncols = dimen[rank-1];

    /* Validate other source arrays, if provided */
    if (array2 != 0) {
	if ( array2->magic != VXL_MAGIC || array2->data == 0 ||
	     !vxli_same_shape(array1, array2) )
	    fatal("Array 2 is invalid or incompatible with array 1");
        type2 = array2->type;
    }
    if (array3 != 0) {
	if ( array3->magic != VXL_MAGIC || array3->data == 0 ||
	     !vxli_same_shape(array1, array3) )
	    fatal("Array 3 is invalid or incompatible with array 1");
        type3 = array3->type;
    }

    /* Validate the moments definition array */
    if ( moments == 0 || moments->magic != VXL_MAGIC ||
	 moments->rank != 2 || moments->dimen[1] != rank ||
	 moments->type != INTP_INT )
	fatal("Moments definition array is invalid");
    momcnt = moments->dimen[0];
    momdef = moments->data;

    /* Allocate and initialize result array for moments */
    vxl_alloc_array(result, INTP_DOUBLE, 1, &momcnt);
    rdata = result->data;
    for (imom = 0; imom < momcnt; imom++)
	rdata[imom] = 0.0;

    /* Determine highest power requested in x; allocate space for x-moments */
    maxpow = 0;
    for (imom = 0; imom < momcnt; imom++)
	if (maxpow < momdef[imom*rank+rank-1])
	    maxpow = momdef[imom*rank+rank-1];
    rowmom = mallock((maxpow+1)*sizeof(double));

    /* Allocate three rows of working storage for array product */
    work1 = mallock(3*ncols*sizeof(double));
    work2 = &work1[ncols];
    xramp = &work1[2*ncols];

    /* Construct row of x coordinate values */
    /* FIXME: Use BIPS instead? */
    for (icol = 0; icol < ncols; icol++)
        xramp[icol] = icol;

    /* Loop over the rows of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Convert first source array to double */
	row1 = vxli_locate(array1, coord, 1);
	info = bips_double(ncols, work1, 1, type1, row1, 1);
	if (info != 0)
	    fatal("Error calling BIPS");

	/* Multiply by second source array, if present */
	if (array2 != 0) {
	    row2 = vxli_locate(array2, coord, 1);
	    info = bips_double(ncols, work2, 1, type2, row2, 1);
	    if (info != 0)
		fatal("Error calling BIPS");
	    info = bips_mul_set(ncols, INTP_DOUBLE, work1, 1, work2, 1);
	    if (info != 0)
		fatal("Error calling BIPS");
	}

	/* Multiply by third source array, if present */
	if (array3 != 0) {
	    row3 = vxli_locate(array3, coord, 1);
	    info = bips_double(ncols, work2, 1, type3, row3, 1);
	    if (info != 0)
		fatal("Error calling BIPS");
	    info = bips_mul_set(ncols, INTP_DOUBLE, work1, 1, work2, 1);
	    if (info != 0)
		fatal("Error calling BIPS");
	}
	    
	/* Loop over powers of x to compute row moments in x */
	for (ipow = 0; ipow <= maxpow; ipow++)
	    rowmom[ipow] = 0.0;
	info = bips_add_set(ncols, INTP_DOUBLE, &rowmom[0], 0, work1, 1);
	if (info != 0)
	    fatal("Error calling BIPS");
	for (ipow = 1; ipow <= maxpow; ipow++) {
	    info = bips_mul_set(ncols, INTP_DOUBLE, work1, 1, xramp, 1);
	    if (info != 0)
		fatal("Error calling BIPS");
	    info = bips_add_set(ncols, INTP_DOUBLE, 
				&rowmom[ipow], 0, work1, 1);
	    if (info != 0)
		fatal("Error calling BIPS");
	}

	/* Loop over the requested moments */
	/* FIXME: Find a faster way to compute the powers?  For example,
	   sort the moments into some order and reshuffle before exit.
	   And/or precompute powers of the outermost coordinate. */
	for (imom = 0; imom < momcnt; imom++) {

	    /* Get appropriate X moment for this row */
	    mom = rowmom[momdef[imom*rank+rank-1]];
	    
	    /* Multiply by powers of remaining coordinates */
	    for (icor = 0; icor < rank-1; icor++) {
		for (ipow = 0; ipow < momdef[imom*rank+icor]; ipow++)
		    mom *= (double)coord[icor];
	    }

	    /* Add this row moment to the accumulated array moments */
	    rdata[imom] += mom;

	}
    }
    while ( vxli_nextrow(rank, dimen, coord) );

    /* Free working storage and return */
    free(work1);
    free(rowmom);
    return;
}



/*-------------------------------------------------------------------------
vxl_norm1 - Compute weighted vector 1-norm of a voxel array

This function computes the weighted vector 1-norm of the given array,
or of the difference of two arrays.  If src2 is a non-null pointer,
the norm of the difference is computed; otherwise, the norm of src1 is
computed.  If weight is a null pointer, then a uniformly weighted norm
is computed; otherwise, weight supplies the weights for each voxel.
The arrays src1, src2, and weight must be the same rank and dimensions.
The arrays src1 and src2 must be the same type; the weight array may
be of any type and will be converted to double for the computation.
--------------------------------------------------------------------------*/

double
vxl_norm1(
    voxel_array *src1,          /* First or only array */
    voxel_array *src2,          /* Second array, or zero */
    voxel_array *weight)        /* Voxel weights, or zero */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int rank;                   /* Rank of the arrays */
    long count;                 /* Total number of elements */
    double res, result = 0.0;   /* Norm of array */
    long coord[VXL_MAX_RANK];   /* Coordinates of each row */
    double *diff=0;             /* Absolute difference of sources */
    double *coef=0;             /* Weights converted to double */
    unsigned char *s1=0, *s2=0; /* Pointers to beginning of each source row */
    unsigned char *w=0;         /* Pointer to beginning of each weight row */
    long nelem;                 /* Number of elements in a row */
    int err;                    /* Error indicator */

    /* Check validity of source 1 */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	 exim_sizeof_intype((stype1 = src1->type)) == 0 )
	fatal("Invalid source 1 array");
    rank  = src1->rank;
    count = vxli_count(src1);
    nelem = src1->dimen[rank-1];

    /* Check validity of source 2 */
    if ( src2 != 0 ) {
	if ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	     exim_sizeof_intype((stype2 = src2->type)) == 0 )
	    fatal("Invalid source 2 array");
	if ( stype2 != stype1 || !vxli_same_shape(src1, src2) )
	    fatal("Incompatible source 1 and 2 arrays");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     exim_sizeof_intype((wtype = weight->type)) == 0 )
	    fatal("Invalid weight array");
	if ( !vxli_same_shape(weight, src1) )
	    fatal("Source 1 and weight arrays are incompatible");
    }

    /* Allocate working storage */
    diff = mallock(nelem*sizeof(double));
    coef = mallock(nelem*sizeof(double));

    /* Loop over rows of the sources */
    vxli_firstrow(src1->rank, coord);
    do {

	/* Compute absolute differences */
	s1 = vxli_locate(src1, coord, 1);
        if (src2 != 0)
	    s2 = vxli_locate(src2, coord, 1);
	err = bips_absdiff(nelem, diff, 1, stype1, s1, 1, s2, 1);
        if (err)
	    fatal("Error calling BIPS");
	
	/* Multiply by weights */
	if (weight != 0) {
	    w = vxli_locate(weight, coord, 1);
            err = bips_double(nelem, coef, 1, wtype, w, 1);
	    if (err)
		fatal("Error calling BIPS");
	    err = bips_mul_set(nelem, INTP_DOUBLE, diff, 1, coef, 1);
	    if (err)
		fatal("Error calling BIPS");
	}

	/* Sum up the differences in this row */
	res = 0.0;
	err = bips_add_set(nelem, INTP_DOUBLE, &res, 0, diff, 1);
	if (err)
	    fatal("Error calling BIPS");
	result += res;

    } while (vxli_nextrow(src1->rank, src1->dimen, coord));

    /* Free working storage and return result */
    free(diff);
    free(coef);
    return result;
}



/*-------------------------------------------------------------------------
vxl_norm2 - Compute weighted vector 2-norm of a voxel array

This function computes the weighted vector 2-norm of the given array,
or of the difference of two arrays.  If src2 is a non-null pointer,
the norm of the difference is computed; otherwise, the norm of src1 is
computed.  If weight is a null pointer, then a uniformly weighted norm
is computed; otherwise, weight supplies the weights for each voxel.
The arrays src1, src2, and weight must be the same rank and dimensions.
The arrays src1 and src2 must be the same type; the weight array may
be of any type and will be converted to double for the computation.

FIXME: There are several places here where BIPS is used for values
that are always type double, and which could be done with simple
loops.  Should BIPS or the simple loops be used?  Ditto for 1-norm
and sup-norm.
--------------------------------------------------------------------------*/

double
vxl_norm2(
    voxel_array *src1,          /* First or only array */
    voxel_array *src2,          /* Second array, or zero */
    voxel_array *weight)        /* Voxel weights, or zero */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int rank;                   /* Rank of the arrays */
    long count;                 /* Total number of elements  FIXME: Unused? */
    double res, result = 0.0;   /* Norm of array */
    long coord[VXL_MAX_RANK];   /* Coordinates of each row */
    double *diff=0;             /* Absolute difference of sources */
    double *coef=0;             /* Weights converted to double */
    unsigned char *s1=0, *s2=0; /* Pointers to beginning of each source row */
    unsigned char *w=0;         /* Pointer to beginning of each weight row */
    long nelem;                 /* Number of elements in a row */
    int err;                    /* Error indicator */
    long i;

    /* Check validity of source 1 */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	 exim_sizeof_intype((stype1 = src1->type)) == 0 )
	fatal("Invalid source 1 array");
    rank  = src1->rank;
    count = vxli_count(src1);
    nelem = src1->dimen[rank-1];

    /* Check validity of source 2 */
    if ( src2 != 0 ) {
	if ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	     exim_sizeof_intype((stype2 = src2->type)) == 0 )
	    fatal("Invalid source 2 array");
	if ( stype2 != stype1 || !vxli_same_shape(src1, src2) )
	    fatal("Incompatible source 1 and 2 arrays");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     exim_sizeof_intype((wtype = weight->type)) == 0 )
	    fatal("Invalid weight array");
	if ( !vxli_same_shape(weight, src1) )
	    fatal("Source 1 and weight arrays are incompatible");
    }

    /* Allocate working storage */
    diff = mallock(nelem*sizeof(double));
    coef = mallock(nelem*sizeof(double));

    /* Loop over rows of the sources */
    vxli_firstrow(src1->rank, coord);
    do {

	/* Compute absolute differences and square them */
	s1 = vxli_locate(src1, coord, 1);
        if (src2 != 0)
	    s2 = vxli_locate(src2, coord, 1);
	err = bips_absdiff(nelem, diff, 1, stype1, s1, 1, s2, 1);
        if (err)
	    fatal("Error calling BIPS");
	for (i = 0; i < nelem; i++)
	    diff[i] *= diff[i];
	
	/* Multiply by weights */
	if (weight != 0) {
	    w = vxli_locate(weight, coord, 1);
            err = bips_double(nelem, coef, 1, wtype, w, 1);
	    if (err)
		fatal("Error calling BIPS");
	    err = bips_mul_set(nelem, INTP_DOUBLE, diff, 1, coef, 1);
	    if (err)
		fatal("Error calling BIPS");
	}

	/* Sum up the differences in this row */
	res = 0.0;
	err = bips_add_set(nelem, INTP_DOUBLE, &res, 0, diff, 1);
	if (err)
	    fatal("Error calling BIPS");
	result += res;

    } while (vxli_nextrow(src1->rank, src1->dimen, coord));

    /* Free working storage and return result */
    free(diff);
    free(coef);
    return sqrt(result);
}



/*-------------------------------------------------------------------------
vxl_normsup - Compute weighted vector sup-norm of a voxel array

This function computes the weighted vector sup-norm of the given array,
or of the difference of two arrays.  If src2 is a non-null pointer,
the norm of the difference is computed; otherwise, the norm of src1 is
computed.  If weight is a null pointer, then a uniformly weighted norm
is computed; otherwise, weight supplies the weights for each voxel.
The arrays src1, src2, and weight must be the same rank and dimensions.
The arrays src1 and src2 must be the same type; the weight array may
be of any type and will be converted to double for the computation.

FIXME: There are several places here where BIPS is used for values
that are always type double, and which could be done with simple
loops.  Should BIPS or the simple loops be used?  Ditto for 1-norm
and sup-norm.
--------------------------------------------------------------------------*/

double
vxl_normsup(
    voxel_array *src1,          /* First or only array */
    voxel_array *src2,          /* Second array, or zero */
    voxel_array *weight)        /* Voxel weights, or zero */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int rank;                   /* Rank of the arrays */
    long count;                 /* Total number of elements */
    double result = 0.0;        /* Norm of array */
    long coord[VXL_MAX_RANK];   /* Coordinates of each row */
    double *diff=0;             /* Absolute difference of sources */
    double *coef=0;             /* Weights converted to double */
    unsigned char *s1=0, *s2=0; /* Pointers to beginning of each source row */
    unsigned char *w=0;         /* Pointer to beginning of each weight row */
    long nelem;                 /* Number of elements in a row */
    int err;                    /* Error indicator */
    long i;

    /* Check validity of source 1 */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	 exim_sizeof_intype((stype1 = src1->type)) == 0 )
	fatal("Invalid source 1 array");
    rank  = src1->rank;
    count = vxli_count(src1);
    nelem = src1->dimen[rank-1];

    /* Check validity of source 2 */
    if ( src2 != 0 ) {
	if ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	     exim_sizeof_intype((stype2 = src2->type)) == 0 )
	    fatal("Invalid source 2 array");
	if ( stype2 != stype1 || !vxli_same_shape(src1, src2) )
	    fatal("Incompatible source 1 and 2 arrays");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     exim_sizeof_intype((wtype = weight->type)) == 0 )
	    fatal("Invalid weight array");
	if ( !vxli_same_shape(weight, src1) )
	    fatal("Source 1 and weight arrays are incompatible");
    }

    /* Allocate working storage */
    diff = mallock(nelem*sizeof(double));
    coef = mallock(nelem*sizeof(double));

    /* Loop over rows of the sources */
    vxli_firstrow(src1->rank, coord);
    do {

	/* Compute absolute differences and square them */
	s1 = vxli_locate(src1, coord, 1);
        if (src2 != 0)
	    s2 = vxli_locate(src2, coord, 1);
	err = bips_absdiff(nelem, diff, 1, stype1, s1, 1, s2, 1);
        if (err)
	    fatal("Error calling BIPS");
	
	/* Multiply by weights */
	if (weight != 0) {
	    w = vxli_locate(weight, coord, 1);
            err = bips_double(nelem, coef, 1, wtype, w, 1);
	    if (err)
		fatal("Error calling BIPS");
	    err = bips_mul_set(nelem, INTP_DOUBLE, diff, 1, coef, 1);
	    if (err)
		fatal("Error calling BIPS");
	}

	/* Find the maximum over all elements */
	for (i = 0; i < nelem; i++)
	    if (diff[i] > result) result = diff[i];

    } while (vxli_nextrow(src1->rank, src1->dimen, coord));

    /* Free working storage and return result */
    free(diff);
    free(coef);
    return result;
}



/*------------------------------------------------------------------------
vxl_stat - Compute mean and standard deviation for an array

This function computes the mean and standard deviation of the elements
of an array, with optional element weights.  The source and weight
arrays may be of any type or shape, but must be the same shape.

FIXME: Technically, this should use the sampling estimate of the variance,
but I don't know how that's computed for a weighted mean and variance.

FIXME: Do the sums rowwise, to reduce the rounding error.
-------------------------------------------------------------------------*/

void
vxl_stat(
    double *mean,               /* Mean value */
    double *stddev,             /* Standard deviation */
    voxel_array *source,        /* Source array */
    voxel_array *weight)        /* Weight array, or zero */
{
    int stype=0, wtype=0;       /* Type of source and weight arrays */
    int rank;                   /* Rank of the arrays */
    long *dimen;                /* Dimensions of the arrays */
    long coord[VXL_MAX_RANK];   /* Coordinates of a row */
    double sumw  = 0.0;         /* Sum of weights */
    double sumww = 0.0;         /* Sum of squared weights */
    double sumx  = 0.0;         /* Weighted sum of elements */
    double sumxx = 0.0;         /* Weighted sum of squared elements */
    double *sbuff, *wbuff;      /* Row buffers for source and weights */
    unsigned char *srow = 0;    /* Pointer to beginning of a source row */
    unsigned char *wrow = 0;    /* Pointer to beginning of a weight row */
    long nelem;                 /* Number of elements in a row */
    int err;                    /* Error indicator */
    double x, w;                /* Element and its weight */
    double m, v;                /* Mean and variance */
    double c;                   /* Correction for degrees of freedom */
    long i;

    /* Validate the source array */
    if ( source == 0 || source->magic != VXL_MAGIC || source->data == 0 ||
	 exim_sizeof_intype((stype = source->type)) == 0 )
	fatal("Invalid source array");
    rank  = source->rank;
    dimen = source->dimen;
    nelem = dimen[rank-1];

    /* Validate the weight array */
    if (weight != 0) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     exim_sizeof_intype((wtype = weight->type)) == 0 )
	    fatal("Invalid weight array");
	if ( !vxli_same_shape(source, weight) )
	    fatal("Incompatible source and weight arrays");
    }

    /* Allocate working storage */
    sbuff = mallock(2*nelem*sizeof(double));
    wbuff = sbuff + nelem;

    /* Loop over rows of the source to sum elements */
    vxli_firstrow(rank, coord);
    do {

	/* Copy source row into buffer, converting to double */
	srow = vxli_locate(source, coord, 1);
	err  = bips_double(nelem, sbuff, 1, stype, srow, 1);
	if (err)
	    fatal("Error calling BIPS");

	/* If weights provided, accumulate weighted sums */
	if (weight) {
	    wrow = vxli_locate(weight, coord, 1);
	    err  = bips_double(nelem, wbuff, 1, wtype, wrow, 1);
	    if (err)
		fatal("Error calling BIPS");
	    for (i = 0; i < nelem; i++) {
		x = sbuff[i];
		w = wbuff[i];
		sumw  += w;
		sumww += w * w;
		sumx  += w * x;
		sumxx += w * x * x;
	    }
	}

	/* Otherwise just accumulate unweighted sums */
	else {
	    for (i = 0; i < nelem; i++) {
		x = sbuff[i];
		sumx  += x;
		sumxx += x * x;
	    }
	}

    } while ( vxli_nextrow(rank, dimen, coord) );

    /* Free working storage */
    free(sbuff);

    /* Compute mean and standard deviation with weights */
    if (weight) {
	if (sumw == 0.0)
	    fatal("Weights sum to zero");
	m = sumx / sumw;
	c = (1.0 - sumww / (sumw * sumw));
	if (c == 0.0)
	    fatal("Not enough data to compute a variance");
	v = (sumxx / sumw - m * m) / c;
    }

    /* Compute mean and standard deviation without weights */
    else {
	sumw = vxli_count(source);
	if (sumw < 2.0)
	    fatal("Not enough data to compute a variance");
	m = sumx / sumw;
	v = (sumxx / sumw - m * m) * sumw / (sumw - 1.0);
    }

    /* Save results and return */
    *mean = m;
    *stddev = sqrt(v);
    return;    
}
