/****************************************************************************
resample.c - Voxel Kit routines for image resampling

Author:  Paul Hughett

Copyright (c) 1998-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file contains code for image resampling
and related functions.
*****************************************************************************/

static const char rcsid[] = "$Id: resample.c,v 1.4 2001/11/01 19:12:48 hughett Exp $ Release $Name:  $";

#include <float.h>
#include <math.h>
#include <stdlib.h>

#include "config.h"
#include "bips.h"
#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"



/*------------------------------------------------------------------------
next_vertex - Find next vertex in n-dimensional cubic neighborhood

This function computes the next vertex in an n-dimensional cubic
neighborhood; it returns zero after the last vertex has been
traversed, and one otherwise.  The number of elements on each side of
the neighborhood is specified by range; the coordinates themselves, in
delta, vary from 0 to range-1, and must be initialized to zero.  See
vxl_linear0 for an example using this function.
-------------------------------------------------------------------------*/

static inline
int
next_vertex(
    int rank,                /* Number of dimensions */
    long *delta,             /* (Relative) coordinates of vertex */
    long range)              /* Number of elements along each side */
{
    int i;

    for (i = rank-1; i >= 0; i--) {
	if ( ++delta[i] < range )
	    return 1;
	else
	    delta[i] = 0; }
    return 0;
}



/*-----------------------------------------------------------------------
vxli_tweak_transform - Convert transform from physical to index coords

This function modifies the matrix and offset of an affine transform 
expressed in physical coordinates to yield the matrix and offset of
the same transform expressed in index coordinates, using the origin
and spacing of the source and destination arrays.  Note that the
transform is assumed to map from destination coordinates backwards
into source coordinates; while this is counter-intuitive, it is in fact
the most convenient way to express the transform for resampling.

This function is designed for use in resampling functions that have
already checked the validity of their arguments and does not do much
further checking; in particular, it assumes that the given rank is
correct for both the source and destination arrays.

FIXME: Having the resampling functions account automatically for the
difference between index and physical coordinates seemed like a good
idea at the time but turned out to introduce some messy new problems;
I am letting this function--which was written specifically for that
purpose--languish while I reconsider the issue.
------------------------------------------------------------------------*/

#if 0
static void
vxli_tweak_transform(
    int rank,                     /* Rank of source and destination arrays */
    double *omatrix,              /* (out) Matrix of affine transform */
    double *ooffset,              /* (out) Offset of affine transform */
    double *imatrix,              /* (in) Matrix of affine transform */
    double *ioffset,              /* (in) Offset of affine transform */
    voxel_array *dest,            /* Destination array */
    voxel_array *src)             /* Source array */
{
    double *dorigin, *dspace;     /* Origin and spacing in destination array */
    double *sorigin, *sspace;     /* Origin and spacing in source array */
    int i, j;

    /* Find parameters of source and destination arrays */
    dorigin = dest->origin;
    dspace  = dest->spacing;
    sorigin = src ->origin;
    sspace  = src ->spacing;

    /* Compute composite matrix and offset */
    for (i = 0; i < rank; i++) {
	ooffset[i] = ioffset[i] - sorigin[i];
	for (j = 0; j < rank; j++) {
	    omatrix[i*rank+j] = dspace[j] * imatrix[i*rank+j] / sspace[i];
	    ooffset[i] += imatrix[i*rank+j] * dorigin[j];
	}
	ooffset[i] /= sspace[i];
    }

}
#endif



/*------------------------------------------------------------------------
vxli_sharp_clip - Clip a line segment against a sharp boundary

This function is used in the fast resampling functions with nearest
neighbor interpolation to clip the image of an ouput row against the
limits of the input image.  The lower and upper limit arguments are
the minimum and maximum values for which source data exists, typically
zero and the array dimension MINUS ONE.  The output array didx
contains indices in the output row corresponding to intersections with
the source array; they are to be interpreted as follows:

     0       <= i < didx[0]    Point i is outside
     didx[0] <= i < didx[1]    Point i is inside
     didx[1] <= i < npoint     Point i is outside

Points outside the source may simply be set to zero; points inside the
source may be interpolated by a fast algorithm that need not check
each vertex against the source limits.  If the segment completely
misses the input image, then both didx[0] and didx[1] are set to
npoint.
-------------------------------------------------------------------------*/

static void
vxli_sharp_clip(
    int rank,                  /* Number of dimensions */
    long npoint,               /* Number of points in output row */
    double *origin,            /* Image of first point */
    double *delta,             /* Increment to next point */
    double *lower,             /* Lower limit in source array */
    double *upper,             /* Upper limit in source array */
    long *didx)                /* (Out) Row intersections with source */
{
    double x0, x1;             /* Row intersections on current axis */
    double y0, y1;             /* Tightest row intersections */
    double ymax;               /* Upper bound on intersection */
    double invdel;
    int i;

    /* Initialize tightest intersections so far */
    y0 = 0.0;
    y1 = ymax = (double)(npoint-1);

    /* Loop over axes and update intersections */
    for (i = 0; i < rank; i++) {

	/* Case 0: segment is constant on this axis */
	if (delta[i] == 0.0) {
	    if  ( origin[i] < lower[i] || origin[i] > upper[i] ) {
		didx[0] = didx[1] = npoint;
		return;
	    }
	}

	/* Cases 1 and 2: segment is not constant */
	else {

	    /* Compute intersections */
	    invdel = 1.0 / delta[i];
	    x0 = (lower[i] - origin[i]) * invdel;
	    x1 = (upper[i] - origin[i]) * invdel;

	    /* Case 1: segment directed toward plus axis */
	    if (delta[i] > 0.0) {
		if (x0 > ymax || x1 < 0.0)  {
		    didx[0] = didx[1] = npoint;
		    return; 
		}
		if ( x0 > y0 ) y0 = x0;
		if ( x1 < y1 ) y1 = x1;
	    }

	    /* Case 2: segment directed toward minus axis */
	    else if (delta[i] < 0.0) {
		if (x1 > ymax || x0 < 0.0)  {
		    didx[0] = didx[1] = npoint;
		    return; 
		}
		if ( x1 > y0 ) y0 = x1;
		if ( x0 < y1 ) y1 = x0;
	    }
	}
    }

    /* Save results and return */
    if ( y0 > ymax || y1 < 0 || y0 > y1 )  {
	didx[0] = didx[1] = npoint;
    }
    else {
	didx[0] = ceil (y0);
	didx[1] = floor(y1) + 1;
    }
    if (0)                                   /* DEBUG */
	fprintf(stderr, "boundaries: %f %f => %ld %ld\n", 
		y0, y1, didx[0], didx[1]);
    return;
}



/*------------------------------------------------------------------------
vxli_fuzzy_clip - Clip a line segment against a fuzzy boundary

This function is used in the fast resampling functions with linear or
higher order interpolation to clip the image of an ouput row against
the limits of the input image; the fuzziness comes in because image
points near the boundary must be handled specially since only some
neighbors exist.  The width must be the half-width of the boundary or
radius (in the Lsup norm) of the interpolation neighborhood.  This
function is believed to work correctly for the exact value of the
width; no error allowance for rounding should be necessary.  The
output array didx contains indices in the output row corresponding to
intersections with the source array; they are to be interpreted as
follows:

     0       <= i < didx[0]    Point i is completely outside (clean miss)
     didx[0] <= i < didx[1]    Point i is near the boundary  (penumbra)
     didx[1] <= i < didx[2]    Point i is completely inside  (umbra)
     didx[2] <= i < didx[3]    Point i is near the boundary  (penumbra)
     didx[3] <= i < npoint     Point i is completely outside (clean miss)

There are two special cases of interest:  If the row cleanly misses both
the umbra and penumbra, then didx[0] = 0 and didx[1] = didx[2] = didx[3]
= npoint.  If the row misses the umbra but hits the penumbra, then didx[0]
is the start of the penumbra and didx[1] = didx[2] = didx[3] are set to
the end of the penumbra.

Points completely outside the source may simply be set to zero; points
completely inside the source may be interpolated by a fast algorithm
that need not check each vertex against the source limits; points near
the boundary must be interpolated by a careful algorithm that actually
checks each vertex against the limits of the source.
-------------------------------------------------------------------------*/

static void
vxli_fuzzy_clip(
    int rank,                  /* Number of dimensions */
    long npoint,               /* Number of points in output row */
    double *origin,            /* Image of first point */
    double *delta,             /* Increment to next point */
    double *lower,             /* Lower limit in source array */
    double *upper,             /* Upper limit in source array */
    double width,              /* Width of fuzzy border */
    long *didx)                /* (Out) Row intersections with source */
{
    double x0, x1, x2, x3;     /* Row intersections on current axis */
    double y0, y1, y2, y3;     /* Tightest row intersections */
    double ymax;               /* Upper limit of row */
    double invdel;             /* Reciprocal of delta[i] */
    int penumbra = 0;          /* Entirely within penumbra for some axis? */
    int miss = 0;              /* Clean miss on some axis? */
    int i;

    /* Initialize tightest intersections */
    ymax = (double)(npoint-1);
    y0 = y1 = 0.0;
    y2 = y3 = ymax;

    /* Loop over axes and update intersections */
    for (i = 0; i < rank; i++) {

	/* Case 0: segment is constant on this axis */
	if (delta[i] == 0.0) {

	    /* Subcase 0.1: Clean miss on this axis => all axes */
	    if  ( origin[i] < lower[i] - width ||
		  origin[i] > upper[i] + width ) {
		miss = 1;
		break;
	    }
	    /* Subcase 0.2: Within penumbra on this axis */
            else if  ( origin[i] < lower[i] + width ||
		  origin[i] > upper[i] - width )
		penumbra = 1;

	    /* Subcase 0.3: Within umbra on this axis */
	    else
		/* Do nothing */ ;
	}

	/* Cases 1 and 2: segment is not constant */
	else {

	    /* Compute intersection, checking for clean misses */
	    invdel = 1.0 / delta[i];
	    x0 = (lower[i] - width - origin[i]) * invdel;
	    x1 = (lower[i] + width - origin[i]) * invdel;
	    x2 = (upper[i] - width - origin[i]) * invdel;
	    x3 = (upper[i] + width - origin[i]) * invdel;

	    /* Case 1: segment directed toward plus axis */
	    if (delta[i] > 0.0) {
		if ( x0 > y0 ) y0 = x0;
		if ( x1 > y1 ) y1 = x1;
		if ( x2 < y2 ) y2 = x2;
		if ( x3 < y3 ) y3 = x3;
	    }

	    /* Case 2: segment directed toward minus axis */
	    else if (delta[i] < 0.0) {
		if ( x3 > y0 ) y0 = x3;
		if ( x2 > y1 ) y1 = x2;
		if ( x1 < y2 ) y2 = x1;
		if ( x0 < y3 ) y3 = x0;
	    }
	}
    }

    /* Convert results to indices:  The rounding of didx[1] and didx[2] 
       is chosen such that an exact integer value does interpolation as 
       part of the umbra */
    if ( miss || y0 > ymax || y3 < 0.0 ) {
	didx[0] = didx[1] = didx[2] = didx[3] = npoint;
    }
    else if ( penumbra || y1 > y2 ) {
	didx[0] = ceil (y0);
	didx[3] = floor(y3) + 1;
	didx[1] = didx[2] = didx[3];
    }
    else {
	didx[0] = ceil (y0);
	didx[1] = ceil (y1);
	didx[2] = floor(y2) + 1;
	didx[3] = floor(y3) + 1;
    }

    return;
}



/*------------------------------------------------------------------------
vxl_nearest - Nearest neighbor interpolation in a voxel array

This function evaluates the value of a voxel array at a given
position, using nearest neighbor interpolation; voxels outside the
array are assumed to be zero.  The returned value is always double,
regardless of the type of the array.
---------------------------------------------------------------------------*/

double                           /* Value interpolated in array */
vxl_nearest(
    voxel_array *src,            /* Voxel array to interpolate in */
    double *point)               /* Point to interpolate at */
{
    int   rank;                  /* Rank of the array */
    long *dimen;                 /* Dimensions of the array */
    long  coord[VXL_MAX_RANK];   /* Rounded coordinates of voxel */
    int   i;

    /* Check validity of the array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank  = src->rank;
    dimen = src->dimen;

    /* If rank 0, just return the only value in the array */
    if (rank < 1)
	return vxl_get_voxel(src, 0);

    /* Round coordinate to get index on each axis; compare to boundaries */
    for (i = 0; i < rank; i++) {
	coord[i] = floor(point[i] + 0.5);
	if  ( coord[i] < 0 || coord[i] >= dimen[i] )
	    return 0.0;
    }

    /* Extract and return requested value from array */
    return vxl_get_voxel(src, coord);
}



/*------------------------------------------------------------------------
vxl_linear0 - Linear interpolation in a voxel array (reference version)

This function evaluates the value of a voxel array at a given
position, interpolating linearly between sample values; voxels outside
the array are assumed to be zero.  The returned value is always double,
regardless of the type of the array.  This implementation is slow but
simple to understand; it is best used as a reference implementation to
test faster but more complex algorithms.
---------------------------------------------------------------------------*/

double                              /* Value interpolated in array */
vxl_linear0(
    voxel_array *src,               /* Voxel array to interpolate in */
    double *point)                  /* Point to interpolate at */
{
    int rank;                       /* Rank of the array */
    long   *dimen;                  /* Dimensions of the array */
    long   index[VXL_MAX_RANK];     /* Index of lowest corner */
    double alpha[VXL_MAX_RANK];     /* Fraction part of coordinate */
    long   delta[VXL_MAX_RANK];     /* Index increment to each corner */
    long   coord[VXL_MAX_RANK];     /* Coordinates for a corner */
    int    valid;                   /* Is corner inside array? */
    double value = 0.0;             /* Value at the given position */
    double vc;                      /* Voxel value at a corner */
    int i;

    /* Check validity of the array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank  = src->rank;
    dimen = src->dimen;

    /* If rank 0, just return the only value in the array */
    if (rank < 1)
	return vxl_get_voxel(src, 0);

    /* Compute index and fraction on each axis;
       initialize counter over neighbors */
    for (i = 0; i < rank; i++) {
	index[i] = floor(point[i]);
        alpha[i] = point[i] - index[i];
        delta[i] = 0;
    }

    /* Loop over the neighbors to compute weighted sum */
    do {

	/* Compute coordinates of current corner; check for boundaries */
	valid = 1;
	for (i = 0; i < rank; i++) {
	    coord[i] = index[i] + delta[i];
	    if ( coord[i] < 0 || coord[i] >= dimen[i] ) {
		valid = 0;
		break;
	    }
	}
	if ( !valid )
	    continue;

	/* Fetch voxel value at the neighbor */
	vc = vxl_get_voxel(src, coord);

	/* Multiply by fraction along each dimension and add to value */
	for (i = 0; i < rank; i++) {
	    if (delta[i] == 0)
		vc *= 1.0 - alpha[i];
	    else
		vc *= alpha[i];
	}
	value += vc;

    } while (next_vertex(rank, delta, 2));

    return value;
}



/*------------------------------------------------------------------------
vxl_linear - Linear interpolation in a voxel array

This function evaluates the value of a voxel array at a given
position, interpolating linearly between sample values; voxels outside
the array are assumed to be zero.  The returned value is always
double, regardless of the type of the array.  NOTE that this function
is intended to interpolate at isolated points, but has too much
overhead to be used in a loop to resample an entire image; a function
such as vxl_affine_linear is much better for resampling since it
spreads the overhead over the entire image to be resampled.

This implementation is written to be reasonably fast but not
necessarily comprehensible; it is verified against the vxl_linear0
implementation.
---------------------------------------------------------------------------*/

double                              /* Value interpolated in array */
vxl_linear(
    voxel_array *src,               /* Voxel array to interpolate in */
    double *point)                  /* Point to interpolate at */
{
    int type;                       /* Type of the array */
    int rank;                       /* Rank of the array */
    long   *dimen;                  /* Dimensions of the array */
    void   *data;                   /* Data contained in the array */
    long   index[VXL_MAX_RANK];     /* Index of lowest corner */
    double alpha[VXL_MAX_RANK][2];  /* Fraction part of coordinate */
    long   delta[VXL_MAX_RANK];     /* Index increment to each corner */
    long   step [VXL_MAX_RANK];     /* Step size in each coordinate */
    long   offset;                  /* Corner location in vectorized array */
    int    valid;                   /* Is corner inside array? */
    double value = 0.0;             /* Value at the given position */
    double vc;                      /* Voxel value at a corner */
    long coord;                     /* Coordinate value */
    int i;

    /* Check validity of the array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    rank  = src->rank;
    dimen = src->dimen;
    type  = src->type;
    data  = src->data;

    /* If rank 0, just return the only value in the array */
    if (rank < 1)
	return vxl_get_voxel(src, 0);

    /* Compute index and fraction on each axis; initialize counter
       over neighbors; check for clean miss.  The reason for treating
       the interval (-1,0) specially is that floor() is very slow in
       gcc/i386, but casting to long always truncates toward zero
       rather than minus infinity. */
    for (i = 0; i < rank; i++) {
	delta[i] = 0;
        if (point[i] >= 0.0 && point[i] < dimen[i]) {
            index[i] = (long)point[i];
            alpha[i][1] = point[i] - (double)index[i];
            alpha[i][0] = 1.0 - alpha[i][1];
        }
        else if (point[i] <= -1.0 || point[i] >= dimen[i]) {
            return 0.0;
        }
        else {
            alpha[i][1] =  point[i] + 1.0;
            alpha[i][0] = -point[i];
            index[i] = -1;
        }
    }

    /* Precompute step size along each coordinate */
    step[rank-1] = 1;
    for (i = rank-1; i > 0; i--)
	step[i-1] = step[i] * dimen[i];

    /* Loop over the neighbors to compute weighted sum */
    do {

	/* Compute coordinates of current corner; check for boundaries */
	/* FIXME: Fuse this with the next_vertex calculation and eliminate
	   some high-order coordinate changes. Also reduce strength. */
	valid  = 1;
        offset = 0;
	for (i = rank-1; i >= 0; i--) {
	    coord   = index[i] + delta[i];
            offset += coord * step[i];
	    if ( coord < 0 || coord >= dimen[i] ) {
		valid = 0;
		break;
	    }
	}
	if ( !valid )   continue;

	/* Fetch voxel value at the neighbor */
	/* FIXME: Convert this to inline for speed; is there a way to
	   do the switch on type only once? */
	vc = exim_get_value(data, type, offset);

	/* Multiply by fraction along each dimension and add to value */
	if ( vc != 0.0 ) {
	    for (i = 0; i < rank; i++) {
		vc *= alpha[i][delta[i]];
	    }
	    value += vc;
	}

    } while (next_vertex(rank, delta, 2));

    return value;
}



/*------------------------------------------------------------------------
vxli_affine_exact - Affine transform and resample without interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates without interpolation, and stores the
results as a newly allocated voxel array.  The elements of the affine
matrix and offset are required to be integers, in which case the
coordinates in the source arrays are always integers and interpolation
is never required; in addition, the source and destination must be the
same type.  This special case occurs frequently as image flips,
rotations by multiples of pi/2, and transposition and can be handled
more efficiently than the general nearest neighbor or linear
interpolation.  Since this function is intended to be called from
vxl_affine_nearest or vxl_affine_linear to handle this special case
more efficiently, it is declared as static and has an atypical calling
sequence: It first checks whether the given affine transform satisfies
the requirements.  If so, it proceeds to do the transform and returns
1; otherwise, it does nothing and returns 0.

The rank of the destination need not match the rank of the source;
this permits projections into a lower dimensional space and liftings
into a higher dimensional one.  However, the rank of the source must
be at least one.  The transform must be specified as the
destination-to-source transformation, since it is used to transform
each destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The number of rows in the matrix is assumed to be the same as
the source rank, and the number of columns is assumed to be the
destination rank; the C ordering of matrix elements is assumed.  The
type of the destination must be the same of the source; this differs
for linear interpolation, in which the types may be different.

The coreps parameter determines how close to integers the coefficients
of the transform must be to use the exact algorithm.  To use this
algorithm only when the transform coefficients are exactly integers,
set coreps = 0.0; otherwise the recommended value is a few times
DBL_EPSILON.  The purpose of this test is to permit a a transform such
as rotation of a multiple of pi/2 to be treated as a fast transform
even though it cannot be computed exactly as integers.

ALGORITHM: This is a fast implementation that clips the back transform
of each row against the limits of the source to save boundary tests in
the inner loop and does the base address calculations in a fast loop.
-------------------------------------------------------------------------*/

#define VXL_MAX_MATRIX (VXL_MAX_RANK*VXL_MAX_RANK)

static int
vxli_affine_exact(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset,              /* Constant part of the transform */
    double coreps)               /* Permissible error in source coords */
{
    int srank;                   /* Rank of the source */
    long imatrix[VXL_MAX_MATRIX];/* Transform matrix approximated by longs */
    long ioffset[VXL_MAX_RANK];  /* Transform offset approximated by longs */
    double corerr;               /* Error of approximation, per coord */
    long   coord[VXL_MAX_RANK];  /* Voxel coords of destination row */
    long   coors[VXL_MAX_RANK];  /* Voxel coords in source, as long */
    double point[VXL_MAX_RANK];  /* Voxel coords in source, as double */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double lower[VXL_MAX_RANK];  /* Lower bounds in source */
    double upper[VXL_MAX_RANK];  /* Upper bounds in source */
    long sstep;            /* Source stride along a given source axis */
    long sslast;           /* Source stride for unit step in last dest axis */
    long ncol;             /* Number of columns in dest row */
    long didx[2];          /* Clipped limits in dest row */
    int  isrc, idst;       /* Indices over source and destination axes */
    unsigned char *drow;   /* Start of a destination row */
    int  flag;             /* BIPS status returned */
    long nelem;            /* Number of non-zero elements in dest row */
    void *dst0;            /* First element to copy to in dest row */
    void *src0;            /* First element to copy from in src row */
    long i;

    /* Validate the source array */
    /* FIXME: Probably duplicates tests in calling function */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	fatal("Invalid or empty source array");
    srank  = src->rank;

    /* Check that source and destination types match */
    if (src->type != type)
	fatal("Source and destination types do not match");

    /* Approximate transform by longs and test if good enough; the
       comparison between corerr and coreps handles equality so that
       setting coreps = 0.0 will use the exact algorithm if and only
       if the transform contains only exact integers. */
    for (idst = 0; idst < rank; idst++) {
	for (isrc = 0; isrc < srank; isrc++) {
	    i = isrc*rank + idst;
	    imatrix[i] = floor(matrix[i] + 0.5);
	    corerr = fabs(matrix[i] - (double)imatrix[i]);
	    if (corerr > coreps)
		return 0;
	}
        ioffset[idst] = floor(offset[idst] + 0.5);
	corerr = fabs(offset[idst] - (double)ioffset[idst]);
        corerr = corerr / (double)dimen[idst];
        if (corerr > coreps)
	    return 0;
    }

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);
    ncol = dimen[rank-1];

    /* Prepare point increments along last dimension of destination,
       plus lower and upper limits for clipper */
    for (isrc = 0; isrc < srank; isrc++) {
	delta[isrc] = imatrix[isrc*rank+rank-1];
	lower[isrc] = 0.0;
	upper[isrc] = (double)(src->dimen[isrc] - 1);
    }

    /* Compute the source stride that corresponds to a unit step along 
       the last axis of the destination */
    sstep  = 1;
    sslast = imatrix[srank*rank-1];
    for (isrc = srank-2; isrc >= 0; isrc--) {
	sstep  *= src->dimen[isrc+1];
        sslast += sstep * imatrix[isrc*rank+rank-1];
    }

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute the source coordinates that correspond to the first
           point in the current row of the destination */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = ioffset[isrc];
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += imatrix[isrc*rank+idst] * coord[idst];
	    }
	}

	/* Clip source coordinates against source boundary */
	vxli_sharp_clip(srank, ncol, point, delta, lower, upper, didx);

	/* Zero destination voxels outside the source */
	drow = vxli_locate(dest, coord, 1);
	flag = bips_zero_range(0, didx[0], type, drow, 1);
	if (flag)  fatal("Error calling bips");
	flag = bips_zero_range(didx[1], ncol, type, drow, 1);
	if (flag)  fatal("Error calling bips");

        /* Copy desired source pixels, if any, into the destination */
        nelem = didx[1] - didx[0];
        if ( nelem > 0 ) {

            /* Find coords for first voxel actually present in source */
            for (isrc = 0; isrc < srank; isrc++)
                coors[isrc] = point[isrc] +
                    didx[0] * imatrix[isrc*rank+rank-1];

            /* Copy desired source voxels into the destination row */
            dst0  = bips_locate(type, drow, didx[0]);
            src0  = vxli_locate(src, coors, 1);
            flag  = bips_copy(nelem, type, dst0, 1, src0, sslast);
            if (flag)  fatal("Error calling bips");

        }

    } while (vxli_nextrow(rank, dimen, coord));

    return 1;
}



/*------------------------------------------------------------------------
vxl_affine_nearest0 - Affine transform and resample with NN interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using nearest neighbor interpolation,
and stores the results as a newly allocated voxel array.  The rank of
the destination need not match the rank of the source; this permits
projections into a lower dimensional space and liftings into a higher
dimensional one.  However, the rank of the source must be at least
one.  The transform must be specified as the destination-to-source
transformation, since it is used to transform each destination voxel
coordinate back into the source; this is the opposite of what you
might naively expect, so be careful to do it right.  The number of
rows in the matrix is assumed to be the same as the source rank, and
the number of columns is assumed to be the destination rank; the C
ordering of matrix elements is assumed.  The type of the destination
need not be the same of the source; this costs nothing extra since all
the intermediate calculations are done in double anyway.

This is a reference implementation.
-------------------------------------------------------------------------*/

void
vxl_affine_nearest0(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int srank;                   /* Rank of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double *value;               /* Row of interpolated values */
    long ncol;                   /* Number of columns in dest row */
    int isrc, idst;              /* Indices over source and destination axes */
    int icol;                    /* Index over columns of a dest row */
    unsigned char *drow;         /* Start of a destination row */
    int flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	fatal("Invalid or empty source array");
    srank  = src->rank;

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);

    /* Allocate memory for a row of interpolated values */
    ncol = dimen[rank-1];
    value = mallock(ncol*sizeof(double));

    /* Prepare point increments along last dimension of destination */
    for (isrc = 0; isrc < srank; isrc++)
	delta[isrc] = matrix[isrc*rank+rank-1];

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute source point corresponding to first column in row */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Loop over the columns of the row */
	for (icol = 0; icol < ncol; icol++) {

	    /* Interpolate in source and save in row array */
	    value[icol] = vxl_nearest(src, point);

	    /* Update position to next column */
	    for (isrc = 0; isrc < srank; isrc++)
		point[isrc] += delta[isrc];
	}

	/* Copy row to destination, converting to destination type */
	drow = vxli_locate(dest, coord, 1);
	flag = bips_cast(ncol, type, drow, 1, INTP_DOUBLE, value, 1);
	if (flag)
	    fatal("Error calling BIPS");

    } while (vxli_nextrow(rank, dimen, coord));

    /* Free temporary storage and return */
    free(value);
    return;
}



/*------------------------------------------------------------------------
vxl_affine_nearest - Affine transform and resample with NN interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using nearest neighbor interpolation,
and stores the results as a newly allocated voxel array.  The rank of
the destination need not match the rank of the source; this permits
projections into a lower dimensional space and liftings into a higher
dimensional one.  However, the rank of the source must be at least
one.  The transform must be specified as the destination-to-source
transformation, since it is used to transform each destination voxel
coordinate back into the source; this is the opposite of what you
might naively expect, so be careful to do it right.  The number of
rows in the matrix is assumed to be the same as the source rank, and
the number of columns is assumed to be the destination rank; the C
ordering of matrix elements is assumed.  The type of the destination
must be the same of the source; this differs for linear interpolation,
in which the types may be different.

This is a fast implementation that clips the back transform
of each row against the limits of the source to save boundary tests in
the inner loop and does the base address calculations in a fast loop.

FIXME: Check the various 0.5 offsets and exact boundary calculations
to make sure they're right.

FIXME: Allow the source and destination types to differ, using a
temporary double array for this special case?  Or gather to a
temporary array and then cast to the destination?
-------------------------------------------------------------------------*/

void
vxl_affine_nearest(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int srank;                   /* Rank of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords of destination row */
    long   sstep[VXL_MAX_RANK];  /* Source stride along each axis */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double lower[VXL_MAX_RANK];  /* Lower bounds in source */
    double upper[VXL_MAX_RANK];  /* Upper bounds in source */
    long   *index;               /* Source index to each voxel in a row */
    long ncol;                   /* Number of columns in dest row */
    long didx[2];                /* Clipped limits in dest row */
    int isrc, idst;              /* Indices over source and destination axes */
    int icol;                    /* Index over columns of a dest row */
    unsigned char *drow;         /* Start of a destination row */
    int flag;                    /* BIPS status returned */
    long nelem;                  /* Number of non-zero elements in dest row */
    long *idx0;
    void *dst0;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	fatal("Invalid or empty source array");
    srank  = src->rank;

    /* Check that source and destination types match */
    /* FIXME: Modify to permit source and dest to have different types? */
    if (src->type != type)
	fatal("Source and destination types do not match");

    /* If possible, use exact resampling algorithm; this requires that
       all transform coefficients are close enough to integers */
    flag = vxli_affine_exact(dest, type, rank, dimen,
			     src, matrix, offset, 5.0*DBL_EPSILON);
    if (flag == 1)
	return;

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);

    /* Allocate working memory */
    ncol  = dimen[rank-1];
    index = mallock(ncol*sizeof(long));

    /* Compute source stride along each axis */
    sstep[srank-1] = 1;
    for (isrc = srank-2; isrc >= 0; isrc--) {
	sstep[isrc] = sstep[isrc+1] * src->dimen[isrc+1];
    }

    /* Prepare point increments along last dimension of destination;
       also lower and upper limits for clipper */
    for (isrc = 0; isrc < srank; isrc++) {
	delta[isrc] = matrix[isrc*rank+rank-1];
	lower[isrc] = 0.0;
	upper[isrc] = (double)(src->dimen[isrc]);
    }

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute source coordinates of first point in the current row;
	   offset by 0.5 so that truncation will round to nearest neighbor */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc] + 0.5;
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Clip source coordinates against source boundary */
	vxli_sharp_clip(srank, ncol, point, delta, lower, upper, didx);

	/* Zero destination voxels outside the source */
	drow = vxli_locate(dest, coord, 1);
	flag = bips_zero_range(0, didx[0], type, drow, 1);
	if (flag)  fatal("Error calling bips");
	flag = bips_zero_range(didx[1], ncol, type, drow, 1);
	if (flag)  fatal("Error calling bips");

	/* Compute indices into the source for this row; note that clipping
           insures that all coordinates here are non-negative, so it's safe
           to cast to long rather than use floor() */
        for (icol = didx[0]; icol < didx[1]; icol++)
            index[icol] = 0;
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] += didx[0] * delta[isrc];
	    for (icol = didx[0]; icol < didx[1]; icol++) {
		index[icol] += sstep[isrc] * (long)(point[isrc]);
		point[isrc] += delta[isrc];
	    }
	}

	/* Gather desired source voxels into the destination row */
	nelem = didx[1] - didx[0];
	dst0  = bips_locate(type, drow, didx[0]);
	idx0  = index + didx[0];
	flag  = bips_gather(nelem, type, dst0, 1, src->data, idx0);
	if (flag)  fatal("Error calling bips");

    } while (vxli_nextrow(rank, dimen, coord));

    /* Free temporary storage and return */
    free(index);
    return;
}



/*------------------------------------------------------------------------
vxl_affine_linear0 - Affine transform and resample with linear interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using linear interpolation between
voxels, and stores the results as a newly allocated voxel array.  The
rank of the destination need not match the rank of the source; this
permits projections into a lower dimensional space and liftings into a
higher dimensional one.  However, the rank of the source must be at
least one.  The transform must be specified as the destination-
to-source transformation, since it is used to transform each
destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The number of rows in the matrix is assumed to be the same as
the source rank, and the number of columns is assumed to be the
destination rank; the C ordering of matrix elements is assumed.  The
type of the destination need not be the same of the source; this costs
nothing extra since all the intermediate calculations are done in
double anyway.

This implementation is simple but slow; it is intended primarily as a
reference implementation to verify faster but less comprehensible
implementations.

FIXME: This function should probably round rather than simply casting
for destinations of integral type.
-------------------------------------------------------------------------*/

void
vxl_affine_linear0(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int srank;                   /* Rank of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double value;                /* Interpolated value in source */
    int isrc, idst;              /* Indices over source and destination axes */

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	fatal("Invalid or empty source array");
    srank  = src->rank;

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);

    /* Loop over the destination voxels */
    vxli_firstvoxel(rank, coord);
    do {

	/* Compute corresponding position in source */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Interpolate in source and save in destination */
	value = vxl_linear0(src, point);
	vxl_set_voxel(dest, coord, value);

    } while (vxli_nextvoxel(rank, dimen, coord) );

    return;
}



/*------------------------------------------------------------------------
vxli_linear_fast - Linear interpolation in a voxel array

This function evaluates the value of a voxel array at a given
position, interpolating linearly between sample values; voxels outside
the array are assumed to be zero.  The returned value is always
double, regardless of the type of the array.  This version checks the
given point against the source boundaries but assumes that the source
has already been checked for validity and that the step sizes have
already been computed; it is intended specifically for use by
vxl_affine_linear rather than as a general-purpose function.  Use
vxl_linear() as the general purpose function.
---------------------------------------------------------------------------*/

static
double                              /* Value interpolated in array */
vxli_linear_fast(
    voxel_array *src,               /* Voxel array to interpolate in */
    double *point,                  /* Point to interpolate at */
    long   *step)                   /* Precomputed step size for each axis */
{
    int type;                       /* Type of the array */
    int rank;                       /* Rank of the array */
    long   *dimen;                  /* Dimensions of the array */
    void   *data;                   /* Data contained in the array */
    long   index[VXL_MAX_RANK];     /* Index of lowest corner */
    double alpha[VXL_MAX_RANK][2];  /* Fraction part of coordinate */
    long   delta[VXL_MAX_RANK];     /* Index increment to each corner */
    long   offset;                  /* Corner location in vectorized array */
    int    valid;                   /* Is corner inside array? */
    double value = 0.0;             /* Value at the given position */
    double vc;                      /* Voxel value at a corner */
    long coord;                     /* Coordinate value */
    int i;

    /* Get some parameters from the source */
    rank  = src->rank;
    dimen = src->dimen;
    type  = src->type;
    data  = src->data;

    /* Compute index and fraction on each axis; initialize counter
       over neighbors; check for clean miss.  The interval (-1,0) is
       handled specially because floor() is slow and casting to long
       truncates toward zero. */
    for (i = 0; i < rank; i++) {
	delta[i] = 0;
        if (point[i] >= 0.0 && point[i] < dimen[i]) {
            index[i] = (long)point[i];
            alpha[i][1] = point[i] - (double)index[i];
            alpha[i][0] = 1.0 - alpha[i][1];
        }
        else if (point[i] <= -1.0 || point[i] >= dimen[i]) {
            return 0.0;
        }
        else {
            alpha[i][1] =  point[i] + 1.0;
            alpha[i][0] = -point[i];
            index[i] = -1;
        }
    }

    /* Loop over the neighbors to compute weighted sum */
    /* FIXME: Would vxli_first_neighbor and _next_neighbor be
       appropriate here?  Elsewhere in this file? */
    do {

	/* Compute coordinates of current corner; check for boundaries */
	valid  = 1;
        offset = 0;
	for (i = rank-1; i >= 0; i--) {
	    coord   = index[i] + delta[i];
            offset += coord * step[i];
	    if ( coord < 0 || coord >= dimen[i] ) {
		valid = 0;
		break;
	    }
	}
	if ( !valid )   continue;

	/* Fetch voxel value at the neighbor */
	/* FIXME: Convert this to inline for speed; is there a way to
	   do the switch on type only once? */
	vc = exim_get_value(data, type, offset);

	/* Multiply by fraction along each dimension and add to value */
	if ( vc != 0.0 ) {
	    for (i = 0; i < rank; i++) {
		vc *= alpha[i][delta[i]];
	    }
	    value += vc;
	}

    } while (next_vertex(rank, delta, 2));

    return value;
}



/*------------------------------------------------------------------------
vxl_affine_linear - Affine transform and resample with linear interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using linear interpolation between
voxels, and stores the results as a newly allocated voxel array.  The
rank of the destination need not match the rank of the source; this
permits projections into a lower dimensional space and liftings into a
higher dimensional one.  However, the rank of the source must be at
least one.  (Allowing rank 0 adds no new capability and would have to
be specially handled.)  The transform must be specified as the
destination-to-source transformation, since it is used to transform
each destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The number of rows in the matrix is assumed to be the same as
the source rank, and the number of columns is assumed to be the
destination rank; the C ordering of matrix elements is assumed.  The
type of the destination need not be the same of the source; this costs
nothing extra since all the intermediate calculations are done in
double anyway.

This implementation is written to be fast but not necessarily
comprehensible; it is verified against the linear0 implementation.


ALGORITHM:  You are not expected to understand this.

All voxels in the destination can be divided into three sets: an umbra
of voxels whose interpolation neighborhoods in the source lie
completely inside the boundaries of the source; a penumbra of voxels
whose interpolation neighborhoods straddle the boundary; and all other
voxels (the anumbra?), whose neighborhoods completely miss the source.
Voxels in the umbra can be computed with a fast algorithm that does
not need to check each neighbor against the boundary.  Voxels in the
penumbra must be computed with an algorithm that does check against
the boundary; such voxels are relatively uncommon.  All other voxels
can simply be set to zero.

Voxels in the umbra are relatively common and are computed using a
fast iterative vectorized algorithm which simulates the obvious
recursive algorithm for pairwise mixing (along each axis) of intensity
values at corners of the interpolation neighborhood.

FIXME: The interpolated value is arguably guaranteed to be within the
representable range of the destination type, but bips_round_cast checks
against the limits anyway.  The extra cost is about 4% on a Pentium II,
but you could eliminate it if you were really pushing for a little extra
performance.
-------------------------------------------------------------------------*/

void
vxl_affine_linear(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int   srank;                 /* Rank of the source */
    int   stype;                 /* Type of the source */
    void *sdata;                 /* Data array of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    long   sstep[VXL_MAX_RANK];  /* Step size along each axis of source */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double lower[VXL_MAX_RANK];  /* Lower limits of source array */
    double upper[VXL_MAX_RANK];  /* Upper limits of source array */
    long   *first;               /* (0,0,0) corner of neighborhood in src */
    double *alpha[VXL_MAX_RANK]; /* Interpolation fractions, per axis */
    double *stack[VXL_MAX_RANK]; /* Stack of partially mixed rows */
    double *value;               /* Row of interpolated values */
    int depth;                   /* Number of items in the stack */
    long neigh[VXL_MAX_RANK];    /* Relative indices of current neighbor */
    long  didx[4];               /* Dest indices for clipped source */
    long ncol;                   /* Number of columns in dest row */
    int isrc, idst;              /* Indices over source and destination axes */
    long icol;                   /* Index over columns of a dest row */
    unsigned char *drow;         /* Start of a destination row */
    double *stk0, *stk1, *alf;
    long idx, nelem, ngh, *fst;
    int flag, i;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	fatal("Invalid or empty source array");
    srank = src->rank;
    stype = src->type;
    sdata = src->data;

    /* Verify that source and destination have the same rank */
    if (rank != srank)
        fatal("Source and destination ranks are different");

    /* If possible, use exact resampling algorithm; this requires that
       all transform coefficients are close enough to integers */
    flag = vxli_affine_exact(dest, type, rank, dimen,
			     src, matrix, offset, 5.0*DBL_EPSILON);
    if (flag == 1)
	return;

    /* Precompute step size along each source coordinate */
    sstep[rank-1] = 1;
    for (i = rank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * src->dimen[i];

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, type, rank, dimen);

    /* Allocate memory for a stack of partially mixed rows, the
        interpolation fractions, and the (0,0,0) corner of each
        neighborhood */
    ncol  = dimen[rank-1];
    first = mallock(ncol*sizeof(unsigned char *));
    stack[0] = mallock(rank*ncol*sizeof(double));
    alpha[0] = mallock(rank*ncol*sizeof(double));
    value = stack[0];
    for (i = 1; i < rank; i++) {
	stack[i] = stack[i-1] + ncol;
	alpha[i] = alpha[i-1] + ncol;
    }

    /* Prepare point increments along last dimension of destination,
       plus lower and upper limits of source */
    for (isrc = 0; isrc < srank; isrc++) {
	delta[isrc] = matrix[isrc*rank+rank-1];
	lower[isrc] = 0.0;
	upper[isrc] = src->dimen[isrc] - 1.0;
    }

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute source point corresponding to first column in row */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Clip to determine pixels within the source */
	vxli_fuzzy_clip(srank, ncol, point, delta, lower, upper, 1.0, didx);

	/* Zero points completely outside the source */
	for (icol = 0; icol < didx[0]; icol++)
	    value[icol] = 0.0;
	for (icol = didx[3]; icol < ncol; icol++)
	    value[icol] = 0.0;

	/* Use boundary-checking method for first penumbra */
	for (isrc = 0; isrc < srank; isrc++)
	    point[isrc] += didx[0] * delta[isrc];
	for (icol = didx[0]; icol < didx[1]; icol++) {
	    value[icol] = vxli_linear_fast(src, point, sstep);
	    for (isrc = 0; isrc < srank; isrc++)
		point[isrc] += delta[isrc];
	}

	/* Compute first corner and fractions for the umbra */
	for (icol = didx[1]; icol < didx[2]; icol++)
	    first[icol] = 0;
	for (isrc = 0; isrc < srank; isrc++) {
	    for (icol = didx[1]; icol < didx[2]; icol++) {
		idx = (long)point[isrc];
		first[icol] += sstep[isrc] * idx;
		alpha[isrc][icol] = point[isrc] - idx;
		point[isrc] += delta[isrc];
	    }
	}

        /* Use fancy vectorized algorithm to interpolate rows */
        nelem = didx[2] - didx[1];
        for (isrc = 0; isrc < srank; isrc++)
            neigh[isrc] = 0;
        isrc  = srank - 1;
        depth = 0;
        while (1) {

            if (isrc == srank-1) {

                /* Gather/mix at leaf nodes */
                stk0 = stack[depth] + didx[1];
                depth++;
                alf  = alpha[isrc]  + didx[1];
                fst  = first + didx[1];
                for (ngh = 0, i = 0; i < srank; i++)
                    ngh += neigh[i] * sstep[i];
                stk1 = bips_locate(stype, sdata, ngh);
                flag = bips_glerp(nelem, stk0, 1, alf, 1, stype, stk1, fst);
                if (flag)
                    fatal("Error calling bips");

                /* Step to next node to process */
                neigh[isrc] = 0;
                if (isrc == 0)           /* For rank == 1 */
                    break;
                else if ( neigh[isrc-1] == 0 )
                    neigh[isrc-1] = 1;
                else
                    neigh[--isrc]++;
            }

            else if (neigh[isrc] < 2) {
                isrc++;
            }

            else if (neigh[isrc] == 2) {

                /* Mix on an interior node */
                depth--;
                stk1 = stack[depth  ] + didx[1];
                stk0 = stack[depth-1] + didx[1];
                alf  = alpha[isrc]    + didx[1];
                flag = bips_lerp(nelem,  INTP_DOUBLE, stk0, 1,
                                 alf, 1, INTP_DOUBLE, stk1, 1);
                if (flag)
                    fatal("Error calling bips");

                /* Step to next node to process */
                if (isrc == 0)      /* Done if just mixed the root */
                    break;
                neigh[isrc] = 0;
                neigh[--isrc]++;
                if (neigh[isrc] < 2)
                    isrc = srank - 1;
            }

            else
                fatal("Fast interpolation algorith is broken");
        }

	/* Use boundary-checking method for second penumbra */
	for (icol = didx[2]; icol < didx[3]; icol++) {
	    value[icol] = vxli_linear_fast(src, point, sstep);
	    for (isrc = 0; isrc < srank; isrc++)
		point[isrc] += delta[isrc];
	}

	/* Copy row to destination, converting to destination type */
	drow = vxli_locate(dest, coord, 1);
	flag = bips_round_cast(ncol, type, drow, 1, INTP_DOUBLE, value, 1);
	if (flag)
	    fatal("Error calling BIPS");

    } while (vxli_nextrow(rank, dimen, coord));

    /* Free temporary storage and return */
    free(alpha[0]);
    free(stack[0]);
    free(first);
    return;
}
