/**************************************************************************
rpsamp.c - Select random sample of points in an image

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program selects a random sample of points in a volume image of
given size, possibly limited to points where the mask image is
non-zero, and writes them to standard output, one per line, in the
order slice, row, column (all zero-origin).  The mask image is assumed
to be unsigned char; other options may be added later.

The parameters are taken from the param file named on the command line.
The following parameters are understood:

   nslices = Number of slices in the image
   nrows   = Number of rows in the image
   ncols   = Number of columns in the image
   npoints = Number of random points to generate
   seed    = Seed to use for the random number generator (see ran1_init     
             for the permitted values).  Defaults to 129.
   mask    = Name of an image file to be used as mask

FIXME: The algorithm used here is not efficient when the mask covers
only a small part of the total image volume.

***************************************************************************/

static const char rcsid[] = "$Id: rpsamp.c,v 1.6 2000/12/20 20:59:42 hughett Exp $ Release $Name:  $";

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "errm.h"
#include "param.h"
#include "rand.h"
#include "voxel.h"


int
main(int argc, char **argv)
{
    Param_list params;             /* Contents of param file */
    long nslices, nrows, ncols;    /* Image dimensions */
    long npoints;                  /* Number of random points */
    char *stage = "rpsamp";        /* Default stage name for params */
    const char *maskfn = 0;        /* Filename for mask image, or zero */
    FILE *maskin;                  /* Stream for the mask image */
    voxel_array *mask = 0;         /* The mask image (if any) */
    int rank = 3;                  /* Number of dimensions in mask image */
    long dimen[3];                 /* Array of the dimensions */
    long point[3];                 /* A random point */
    long seed;                     /* Seed value for random number generator */
    void *rstate;                  /* State for random number generator */
    long i;
    const char *s;

    /* Determine the parameters for this run */
    if (argc < 2)
	fatal("No param file specified");
    params = read_param_file(argv[1]);
    nslices = ( s = get_param(params, stage, "nslices") ) ? atol(s) :  256;
    nrows   = ( s = get_param(params, stage, "nrows")   ) ? atol(s) :  256;
    ncols   = ( s = get_param(params, stage, "ncols")   ) ? atol(s) :  256;
    npoints = ( s = get_param(params, stage, "npoints") ) ? atol(s) : 1000;
    seed    = ( s = get_param(params, stage, "seed")    ) ? atol(s) :  129;
    maskfn  = get_param(params, stage, "mask");
    dimen[0] = nslices;
    dimen[1] = nrows;
    dimen[2] = ncols;

    /* Do some consistency checking on parameters */
    if (nslices < 1 || nslices > 4000)
	fatal("nslices has an implausible value");
    if (nrows < 1 || nrows > 4000)
	fatal("nrows has an implausible value");
    if (ncols < 1 || ncols > 4000)
	fatal("ncols has an implausible value");
    if (npoints < 1 || npoints > 1000000)
	fatal("npoints has an implausible value");

    /* Open and read the mask image file, if any */
    if (maskfn != 0) {
	maskin = fopen(maskfn, "r");
	if (maskin == 0)
	    fatal("Unable to open mask image %s", maskfn);
	mask = vxl_create_array();
	vxl_read_uchar_data(mask, maskin, rank, dimen);
        fclose(maskin);
    }

    /* Generate random points, checking against the mask */
    rstate = ran1_init(seed);
    for (i = 0; i < npoints; i++) {
	do {
	    point[2] = ran1_choose(rstate, dimen[2]);
	    point[1] = ran1_choose(rstate, dimen[1]);
	    point[0] = ran1_choose(rstate, dimen[0]);
	} while (mask && vxl_get_voxel(mask, point) == 0.0);
	printf("%5ld  %5ld  %5ld\n", point[0], point[1], point[2]);
    }

    /* Clean up and exit */
    if (mask != 0) vxl_destroy_array(mask);
    exit(EXIT_SUCCESS);
}
