/****************************************************************************
skmiv.c - Segmentation using shaded K-means on interior voxels

Author:  Paul Hughett

Copyright (c) 1999 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This experimental program segments a single-echo unsigned char image
using the skmiv algorithm (shaded K-means on interior voxels).
Boundary voxels are defined to be those voxels on the edges of the
image or for which the local variance in a 3x3x3 neighborhood exceeds
a threshold set by the user; this threshold is specified as a standard
deviation and represents a compromise between the variance of a
uniform region and the difference in gray level between different
regions.  An initial guess at the centroids must also be provided and
allows the user to define class ids for the different tissue classes.
The file defining the initial centroids is a commented data file
containing one line for each class; that line contains the desired
class id and the coordinate of the initial centroid.  It is
permissible for more than one centroid to share the same class id.

The following parameters are taken from the param file named on the
command line:

image   = Filename of the image to be segmented
imgout  = Name of the output file to be created
thresh  = Threshold value for variance
classes = Filename of the file containing the initial centroids
nslices = Number of slices in the image
nrows   = Number of rows in the image
ncols   = Number of columns in the image
debug   = Level of debugging output (0-3)

The debugging levels are as follows:  0 => silent execution; 1 =>
report beginning and end of each step; 2 => briefly report results
of each step; 3 => write output files showing intermediate results
in detail.

FIXME:  The "shaded" feature has yet to be implemented.

FIXME:  This program uses a fixed-length array for the class centroid;
it would be more flexible to use a dynamically allocated array.
**************************************************************************/


static const char rcsid[] = "$Id: skmiv.c,v 1.2 2001/11/14 16:30:07 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "cdata.h"
#include "errm.h"
#include "param.h"
#include "voxel.h"

#define MAXCLS   512        /* Maximum number of class centroids */



int
main(int argc, char **argv)
{
    Param_list params;             /* Contents of param file */
    char *stage = "skmiv";         /* Default stage name for params */
    double thresh;                 /* Variance threshold */
    int debug = 3;                 /* Level of debugging output */
    const char *imagefn = 0;       /* Filename for the input image */
    const char *imoutfn = 0;       /* Filename for the output image */
    const char *classfn = 0;       /* Filename for class definitions */
    FILE *imgin;                   /* Stream for the first image */
    FILE *imgout;                  /* Stream for the histogram as an image */
    FILE *clsin;                   /* Stream for the class definitions */
    voxel_array *image  = 0;       /* Voxel array for input image */
    voxel_array *inter  = 0;       /* Voxel array for the interior voxels */
    voxel_array *hist   = 0;       /* Histogram of interior voxels */
    voxel_array *segmnt = 0;       /* Voxel array for class image */
    int nclass;                    /* Number of centroids defined */
    unsigned char clid[MAXCLS];    /* Class ids */
    unsigned char cent[MAXCLS];    /* Centroid coords */
    int rank = 3;                  /* Number of dimensions in image */
    long nslices, nrows, ncols;    /* Image dimensions */
    long dimen[3];                 /* Array of the dimensions */
    int ndat;                      /* Number of data found by cdata */
    char **data, **cmt;            /* Data and comments found by cdata */
    const char *s;
    int i;

    /* Determine the parameters for this run */
    if (debug >= 1)
	fprintf(stderr, "Reading parameters...");
    if (argc < 2)
	fatal("No param file specified");
    params   = read_param_file(argv[1]);
    imagefn  = get_param(params, stage, "image");
    imoutfn  = get_param(params, stage, "imgout");
    classfn  = get_param(params, stage, "classes");
    thresh   = (s = get_param(params, stage, "thresh"))  ? strtod(s, 0) : 10.0;
    debug    = (s = get_param(params, stage, "debug"))   ? atol(s) : debug;
    nslices  = (s = get_param(params, stage, "nslices")) ? atol(s) :  256;
    nrows    = (s = get_param(params, stage, "nrows")  ) ? atol(s) :  256;
    ncols    = (s = get_param(params, stage, "ncols")  ) ? atol(s) :  256;
    dimen[0] = nslices;
    dimen[1] = nrows;
    dimen[2] = ncols;

    /* Do some consistency checking on parameters */
    if (nslices < 1 || nslices > 4000)
	fatal("nslices has an implausible value");
    if (nrows < 1 || nrows > 4000)
	fatal("nrows has an implausible value");
    if (ncols < 1 || ncols > 4000)
	fatal("ncols has an implausible value");
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Read the class definitions */
    if (debug >= 1)
	fprintf(stderr, "Reading class definitions...");
    if (classfn == 0)
	fatal("No classes file was specified");
    clsin = fopen(classfn, "r");
    if (clsin == 0)
	fatal("Unable to open class defn file %s", classfn);
    for (nclass = 0; ; nclass++) {
	if (nclass >= MAXCLS)
	    fatal("Too many class centroids found");
	ndat = cdata_read(clsin, &data, &cmt);
	if (ndat == EOF)  break;
	if (ndat < 2)
	    fatal("Too few values for class centroid");
	clid[nclass] = atoi(data[0]);
	cent[nclass] = atoi(data[1]);
	cdata_free(ndat, data, cmt);
    }
    fclose(clsin);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Open and read the image file */
    if (debug >= 1)
	fprintf(stderr, "Reading input image...");
    if (imagefn == 0)
	fatal("No input image file was specified");
    imgin = fopen(imagefn, "r");
    if (imgin == 0)
	fatal("Unable to open image file %s", imagefn);
    image = vxl_create_array();
    vxl_read_uchar_data(image, imgin, rank, dimen);
    fclose(imgin);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Zero the boundary voxels in the input image */
    if (debug >= 1)
        fprintf(stderr, "Identifying interior voxels...");
    inter = vxl_create_array();
    vxl_zerbv_uchar3d(inter, image, thresh);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Optionally output an image with boundary voxels zeroed */
    if (debug >= 3) {
	FILE *mskout;
	fprintf(stderr, "Writing mask image to file skmiv_mask.img...");
	mskout = fopen("skmiv_mask.img", "w");
	if (mskout == 0)
	    fatal("Unable to open mask file for output");
	vxl_write_uchar_data(mskout, inter);
	fclose(mskout);
	fprintf(stderr, "done.\n");
    }

    /* Compute histogram of the interior voxels */
    if (debug >= 1)
        fprintf(stderr, "Computing histogram of interior voxels...");
    hist = vxl_create_array();
    vxl_histo_uchar(hist, inter, 0);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Optionally output the non-zero bins of the histogram */
    if (debug >= 3) {
	unsigned long *hdata;
	FILE *hstout;
	hstout = fopen("skmiv_hist.dat", "w");
	if (hstout == 0)
	    fatal("Unable to open histogram output file");
	hdata = hist->data;
	for (i = 0; i <= UCHAR_MAX; i++)
	    if (hdata[i] > 0)
		fprintf(hstout, "%4d  %8lu\n", i, hdata[i]);
	fclose(hstout);
    }

    /* Run K-means to find improved class centroids */
    if (debug >= 1)
        fprintf(stderr, "Computing new class centroids...");
    vxl_kmeans1_histo(hist, nclass, cent);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Optionally report the class centroids */
    if (debug >= 2) {
	fprintf(stderr, "Class Centroids:\nclid  cent\n");
	for (i = 0; i < nclass; i++)
	    fprintf(stderr, "%4d  %4d\n", (int)clid[i], (int)cent[i]);
    }

    /* Segment the original image using the class centroids */
    if (debug >= 1)
        fprintf(stderr, "Segmenting the original image...");
    segmnt = vxl_create_array();
    vxl_nnclass1_uchar(segmnt, image, nclass, clid, cent);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Write out the processed image */
    if (debug >= 1)
        fprintf(stderr, "Writing the segmented image to a file...");
    if (imoutfn == 0)
	fatal("No imgout file was specified");
    imgout = fopen(imoutfn, "w");
    if (imgout == 0)
	fatal("Unable to open output file %s", imoutfn);
    vxl_write_uchar_data(imgout, segmnt);
    fclose(imgout);
    if (debug >= 1)
	fprintf(stderr, "done.\n");

    /* Clean up and exit */
    if (segmnt != 0) vxl_destroy_array(segmnt);
    if (hist   != 0) vxl_destroy_array(hist);
    if (inter  != 0) vxl_destroy_array(inter);
    if (image  != 0) vxl_destroy_array(image);

    exit(EXIT_SUCCESS);
}
