/****************************************************************************
swab.c - Swap byte order according to a pattern

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program takes input from a file or standard input, swaps bytes
according to a specified pattern, and writes the result to standard
output.  See the man page for a full description.
*****************************************************************************/

static const char rcsid[] = "$Id: swab.c,v 1.5 2000/12/17 20:14:12 hughett Exp $ Release $Name:  $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "exim.h"

#define MAXOUT  35      /* Maximum number of bytes per output block */
#define NBLOCK  1000    /* Nmbr blocks to read at a time */        


int
main(int argc, char *argv[])
{
    FILE *input = stdin;           /* File to get the input string from */
    int nbin = 0;                  /* Nmbr bytes per input block */
    int nbout = 2;                 /* Nmbr bytes per output block */
    int inbyte[MAXOUT+1]           /* Where to get each output byte */
	 = {1, 0};
    unsigned char *ibuf, *obuf;    /* Input and output buffers */
    long ilen, olen;               /* Nmbr chars to read or write */
    long len;                      /* Actual number of chars */
    long nblock;                   /* Actual number of blocks read */
    int i, k, v;
    char *p;
    
    /* Process the command line arguments */
    for (i = 1; i < argc; i++) {
	if (strncmp(argv[i], "-i", 2) == 0) {
	    if (strlen(argv[i]) == 2)  p = argv[++i];
	    else                       p = argv[i] + 2;
	    nbin = atoi(p);
	    if (nbin <= 0) {
		fprintf(stderr, "Invalid argument to -i option\n");
		exit(EXIT_FAILURE); }
	}
	else if (strncmp(argv[i], "-o", 2) == 0) {
	    if (strlen(argv[i]) == 2) p = argv[++i];
	    else                      p = argv[i] + 2;
	    nbout = strlen(p);
	    if (nbout > MAXOUT) {
		fprintf(stderr, "Can't handle so many output bytes\n");
		exit(EXIT_FAILURE); }
	    for (k = 0; k < nbout; k++) {
		/* exim_swap_bytes counts bytes from 0 rather than 1 */
		v = p[k];
		if ('0' <= v && v <= '9')
		    inbyte[k] = v - '0' - 1;
		else if ('a' <= v && v <= 'z')
		    inbyte[k] = v - 'a' + 9;
		else if ('A' <= v && v <= 'Z')
		    inbyte[k] = v - 'A' + 9;
		else {
		    fprintf(stderr, "Invalid argument to -o option\n");
		    exit(EXIT_FAILURE); }
	    }
	}
	else if (strncmp(argv[i], "-", 1) == 0) {
	    fprintf(stderr, "Invalid option on command line\n");
	    exit(EXIT_FAILURE); }
	else  { /* Assume it's a file name */
	    input = fopen(argv[i], "rb");
	    if (input == 0) {
		fprintf(stderr, "Error opening the input file.\n");
		exit(EXIT_FAILURE); }
	}
    }

    /* Set nbin default from value of inbytes */
    if (nbin == 0) 
	for (k = 0; k < nbout; k++) 
	    if (nbin <= inbyte[k]) nbin = inbyte[k] + 1;

    /* Allocate input and output buffers */
    ilen = (long)nbin *NBLOCK;
    olen = (long)nbout*NBLOCK;
    ibuf = malloc(ilen);
    obuf = malloc(olen);
    if (ibuf == 0 || obuf == 0) {
	fprintf(stderr, "Unable to allocate i/o buffers\n");
	exit(EXIT_FAILURE); }

    /* Process by batches of blocks */
    while (!feof(input)) {

	/* Read a batch of input */
	len = fread(ibuf, 1, ilen, input);
	if (len == 0) break;

	/* Compute actual number of blocks; fill last with zeros */
	nblock = (len + nbin - 1) / nbin;
	if (len < nblock*nbin)
	    for (k = len; k < nblock*nbin; k++)  ibuf[k] = 0;
	
	/* Transform input */
	exim_swap_bytes(nblock, obuf, ibuf, nbout, nbin, inbyte, 0);

	/* Write to standard output */
	len = fwrite(obuf, nbout, nblock, stdout);
	if (len < nblock) {
	    fprintf(stderr, "Error writing output file\n");
	    exit(EXIT_FAILURE); }
    }

    return EXIT_SUCCESS;
}
