/****************************************************************************
usb2uc.c - Convert unsigned short big-endian data to unsigned char

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program reads data from a file or standard input as a sequence of
unsigned short (16-bit) big-endian values, rescales them to fit inside
an unsigned char (8-bit), and writes the scaled values to a file or
standard output.

See the man page for a full description.

This program has been written for simplicity rather than speed; disk
i/o is likely to be the bottleneck in any case.
*****************************************************************************/

static const char rcsid[] = "$Id: usb2uc.c,v 1.2 2000/01/24 21:30:41 hughett Exp $ Release $Name:  $";

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>       /* For getopt and friends */

#include "errm.h"


int
main(int argc, char *argv[])
{
    FILE *input = stdin;            /* File to read data from */
    FILE *output = stdout;          /* File to write data to */
    unsigned int wmin = 0;          /* Value to convert to 0 */
    unsigned int wmax = 4095;       /* Value to convert to 255 */
    unsigned int umin = USHRT_MAX;  /* Actual minimum value in input */
    unsigned int umax = 0;          /* Actual maximum value in input */
    unsigned int inval;             /* Input value */
    double outval;                  /* Scaled output value */
    unsigned char iv[2];            /* Two bytes of the input value */
    unsigned char ov;               /* One byte for output */
    double scale;                   /* Scale factor for rescaling */
    int n, c;
    char *s;
    unsigned int r1;

    /* Get the command options, if any */
    while ( (c = getopt(argc, argv, "w:")) != EOF )
	switch (c) {
	case 'w':
	    r1 = strtoul(optarg, &s, 0);
	    if (*s == 0 || isspace((int)*s)) {
		wmax = r1;
		break; }
	    if (*s != ':' && *s != ',')
		fatal("Valid format of -w option is -w[min:]max");
	    wmin = r1;
	    wmax = strtoul(s+1, 0, 0);
	    break;
	}

    /* DEBUG:  Show options */
    if (0) {
	fprintf(stderr, "Intensity window is %u..%u\n", wmin, wmax);
    }

    /* Get the input file name and open it */
    if (argc-optind >= 1) {
	input = fopen(argv[optind], "rb");
	if (input == 0)
	    fatal("Unable to open input file %s", argv[optind]);
    }

    /* Get the output file name (if any) and open it */
    if (argc-optind >= 2) {
	output = fopen(argv[optind+1], "wb");
	if (output == 0)
	    fatal("Unable to open output file %s", argv[optind+1]);
    }

    /* Rescale and output values */
    scale = 255.0/(double)(wmax-wmin);
    while ( 1 ) {
	n = fread(iv, 1, 2, input);
	if (feof(input)) break;
	if (n != 2)
	    fatal("Read error, or odd number of bytes in file.");
	inval = (iv[0] << 8) + iv[1];
	if (inval > umax) umax = inval;
	if (inval < umin) umin = inval;
	outval = scale*((double)inval-wmin);
	if      (outval < 0.0)   ov = 0;
	else if (outval > 255.0) ov = 255;
	else                     ov = (unsigned char)outval;
	n = fwrite(&ov, 1, 1, output);
	if (n != 1)
	    fatal("Write error on output file");
    }

    /* Show the actual range of values in the input */
    fprintf(stderr, "Range of unsigned values in input was %u..%u\n",
	    umin, umax);

    return EXIT_SUCCESS;
}
