/**************************************************************************
vibihist.c - Compute bivariate histogram of two volume images

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program computes the bivariate histogram of a pair of unsigned
char volume images.  The first image defines row numbers in the
bivariate histogram; the second determines column numbers.  The
results are written to standard output.  The first line of the output
is a space-delimited list of the column numbers (i.e. gray level
values in the second image) for which there is at least one voxel.
There is one additional line for each gray level in the first image
for which there is at least one voxel; that line contains the gray
level value, followed by the pixel counts for the appropriate
combination of row and column number (that is, gray levels in the
first and second images).

The parameters are taken from the param file named on the command line.
The following parameters are understood:

   image1  = Filename of the first (row) image
   image2  = Filename of the second (column) image
   nslices = Number of slices in the image
   nrows   = Number of rows in the image
   ncols   = Number of columns in the image

***************************************************************************/

static const char rcsid[] = "$Id: vibihist.c,v 1.5 2001/11/14 16:30:07 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "errm.h"
#include "param.h"
#include "voxel.h"


int
main(int argc, char **argv)
{
    Param_list params;             /* Contents of param file */
    long nslices, nrows, ncols;    /* Image dimensions */
    char *stage = "vibihist";      /* Default stage name for params */
    const char *imag1fn = 0;       /* Filename for the first image */
    const char *imag2fn = 0;       /* Filename for the second image */
    const char *imoutfn = 0;       /* Filename for the histogram as an image */
    FILE *imag1in;                 /* Stream for the first image */
    FILE *imag2in;                 /* Stream for the second image */
    FILE *imgout;                  /* Stream for the histogram as an image */
    voxel_array *imag1 = 0;        /* Voxel array for first image */
    voxel_array *imag2 = 0;        /* Voxel array for second image */
    voxel_array *hist = 0;         /* Bivariate histogram, as voxel array */
    voxel_array *mar1 = 0;         /* Marginal histogram for first image */
    voxel_array *mar2 = 0;         /* Marginal histogram for second image */
    voxel_array *histimg = 0;      /* Histogram as an image */
    unsigned long *hdata;          /* Actual histogram data */
    unsigned long *m1, *m2;        /* Actual marginal histogram data */
    int rank = 3;                  /* Number of dimensions in images */
    long dimen[3];                 /* Array of the dimensions */
    int irow, icol;                /* Row and column of the histogram */
    const char *s;

    /* Determine the parameters for this run */
    if (argc < 2)
	fatal("No param file specified");
    params   = read_param_file(argv[1]);
    imag1fn  = get_param(params, stage, "image1");
    imag2fn  = get_param(params, stage, "image2");
    imoutfn  = get_param(params, stage, "imgout");
    nslices  = ( s = get_param(params, stage, "nslices") ) ? atol(s) :  256;
    nrows    = ( s = get_param(params, stage, "nrows")   ) ? atol(s) :  256;
    ncols    = ( s = get_param(params, stage, "ncols")   ) ? atol(s) :  256;
    dimen[0] = nslices;
    dimen[1] = nrows;
    dimen[2] = ncols;

    /* Do some consistency checking on parameters */
    if (nslices < 1 || nslices > 4000)
	fatal("nslices has an implausible value");
    if (nrows < 1 || nrows > 4000)
	fatal("nrows has an implausible value");
    if (ncols < 1 || ncols > 4000)
	fatal("ncols has an implausible value");

    /* Open and read the first image file */
    if (imag1fn == 0)
	fatal("No image1 file was specified");
    imag1in = fopen(imag1fn, "r");
    if (imag1in == 0)
	fatal("Unable to open image1 file %s", imag1fn);
    imag1 = vxl_create_array();
    vxl_read_uchar_data(imag1, imag1in, rank, dimen);
    fclose(imag1in);

    /* Open and read the second image file */
    if (imag2fn == 0)
	fatal("No image2 file was specified");
    imag2in = fopen(imag2fn, "r");
    if (imag2in == 0)
	fatal("Unable to open image2 file %s", imag2fn);
    imag2 = vxl_create_array();
    vxl_read_uchar_data(imag2, imag2in, rank, dimen);
    fclose(imag2in);

    /* Compute the histogram */
    hist = vxl_create_array();
    mar1 = vxl_create_array();
    mar2 = vxl_create_array();
    vxl_bihist_uchar(hist, mar1, mar2, imag1, imag2, 0);

    /* Write out the column numbers with non-zero bins */
    hdata = hist->data;     /* FIXME:  Private info? */
    m1    = mar1->data;
    m2    = mar2->data;
    printf("    .  ");
    for (icol = 0; icol <= UCHAR_MAX; icol++)
	if (m2[icol] > 0) printf("%10d  ", icol);
    printf("\n");

    /* Now write out the row numbers and bin counts */
    for (irow = 0; irow <= UCHAR_MAX; irow++) {
	if (m1[irow] > 0) {
	    printf("%5d  ", irow);
	    for (icol = 0; icol <= UCHAR_MAX; icol++)
		if (m2[icol] > 0)
		    printf("%10lu  ", hdata[irow*(UCHAR_MAX+1)+icol]);
	    printf("\n");
	}
    }

    /* If needed, write out the histogram as a raw image file */
    if (imoutfn != 0) {
	imgout = fopen(imoutfn, "w");
	if (imgout == 0)
	    fatal("Unable to open histo image output file %s", imoutfn);
	histimg = vxl_create_array();
	vxl_logcomp_ulong(histimg, hist);
	vxl_write_uchar_data(imgout, histimg);
	vxl_destroy_array(histimg);
	fclose(imgout);
    }

    /* Clean up and exit */
    if (hist  != 0) vxl_destroy_array(hist);
    if (mar1  != 0) vxl_destroy_array(mar1);
    if (mar2  != 0) vxl_destroy_array(mar2);
    if (imag1 != 0) vxl_destroy_array(imag1);
    if (imag2 != 0) vxl_destroy_array(imag2);

    exit(EXIT_SUCCESS);
}
