/**************************************************************************
vihist.c - Compute histogram of a volume image

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program computes the univariate histogram of an unsigned char
volume image.  If coordinate spacings are provided in physical units,
it also computes the volume represented by each intensity value; this
is useful for computing tissue volumes in a segmented image.  The results
are written to standard output, one line per intensity level with a
non-zero count; the values provided are intensity level, count, and
volume.  Any physical spacing not specified is assumed to be 1.0.

The parameters are taken from the param file named on the command line.
The following parameters are understood:

   image   = Filename of the image
   nslices = Number of slices in the image
   nrows   = Number of rows in the image
   ncols   = Number of columns in the image
   zspace  = Physical spacing between slices
   yspace  = Physical spacing between rows
   xspace  = Physical spacing between columns

***************************************************************************/

static const char rcsid[] = "$Id: vihist.c,v 1.6 2001/11/14 16:30:07 hughett Exp $ Release $Name:  $";

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "errm.h"
#include "param.h"
#include "voxel.h"


int
main(int argc, char **argv)
{
    Param_list params;             /* Contents of param file */
    long nslices, nrows, ncols;    /* Image dimensions */
    char *stage = "vihist";        /* Default stage name for params */
    const char *imagfn = 0;        /* Filename for the image */
    FILE *imagin;                  /* Stream for the image */
    voxel_array *imag = 0;         /* The image to be histogrammed */
    voxel_array *hist = 0;         /* Histogram, as voxel array */
    unsigned long *hdata;          /* Actual histogram data */
    int rank = 3;                  /* Number of dimensions in mask image */
    long dimen[3];                 /* Array of the dimensions */
    double space[3] =              /* Physical spacing of the voxels */
        {1.0, 1.0, 1.0};  
    double pixvol;                 /* Volume of a voxel */
    int i;
    const char *s;

    /* Determine the parameters for this run */
    if (argc < 2)
	fatal("No param file specified");
    params   = read_param_file(argv[1]);
    imagfn   = get_param(params, stage, "image");
    nslices  = ( s = get_param(params, stage, "nslices") ) ? atol(s) :  256;
    nrows    = ( s = get_param(params, stage, "nrows")   ) ? atol(s) :  256;
    ncols    = ( s = get_param(params, stage, "ncols")   ) ? atol(s) :  256;
    space[0] = ( s = get_param(params, stage, "zspace")  ) ? atof(s) :  1.0;
    space[1] = ( s = get_param(params, stage, "yspace")  ) ? atof(s) :  1.0;
    space[2] = ( s = get_param(params, stage, "xspace")  ) ? atof(s) :  1.0;
    dimen[0] = nslices;
    dimen[1] = nrows;
    dimen[2] = ncols;

    /* Do some consistency checking on parameters */
    if (nslices < 1 || nslices > 4000)
	fatal("nslices has an implausible value");
    if (nrows < 1 || nrows > 4000)
	fatal("nrows has an implausible value");
    if (ncols < 1 || ncols > 4000)
	fatal("ncols has an implausible value");

    /* Open and read the image file */
    if (imagfn == 0)
	fatal("No image file was specified");
    imagin = fopen(imagfn, "r");
    if (imagin == 0)
	fatal("Unable to open image file %s", imagfn);
    imag = vxl_create_array();
    vxl_read_uchar_data(imag, imagin, rank, dimen);
    fclose(imagin);

    /* Compute the histogram */
    hist = vxl_create_array();
    vxl_histo_uchar(hist, imag, 0);

    /* Write out the non-zero bins of the histogram */
    pixvol = space[0] * space[1] * space[2];
    hdata = hist->data;     /* FIXME:  Private info? */
    for (i = 0; i <= UCHAR_MAX; i++)
	if (hdata[i] > 0)
	    printf("%4d  %8lu  %12f\n",
		   i, hdata[i], pixvol*(double)hdata[i]);

    /* Clean up and exit */
    if (hist != 0) vxl_destroy_array(hist);
    if (imag != 0) vxl_destroy_array(imag);
    exit(EXIT_SUCCESS);
}
