/****************************************************************************
voxel.c - Voxel Kit routines for volume image processing

Author:  Paul Hughett

Copyright (c) 1998-2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular files defines the core operations, plus
various non-core operations that don't yet fit anywhere else.

FIXME:  This file is getting large enough that it might make sense to split
it into separate files.  In any case, some rational order should be imposed
on the functions herein.

FIXME: Clarify exactly when data lengths and locations are specified in
bytes and when in elements.

FIXME:  Identify common validity checking steps and isolate as functions
or macros.  Q:  What are the efficiency tradeoffs between the two?

FIXME: Add voxel kit functions for compound assignments, or updating
in place, to follow the BIPS functions for this purpose.  Maybe also
extend upwards into Pyvox as well (since they will save storage in
many applications).
*****************************************************************************/

static const char rcsid[] = "$Id: voxel.c,v 1.93 2002/02/25 16:53:49 hughett Exp $ Release $Name:  $";

#include <float.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"
#include "bips.h"



/**************************************************************************
                     Internal Utility Functions

                 For use within the Voxel Kit itself
**************************************************************************/


/*-------------------------------------------------------------------------
cmp_uchar - Compare two unsigned char values for qsort or bsearch

FIXME:  Consider promoting this to a general-purpose library function
somewhere.
-------------------------------------------------------------------------*/

static int
cmp_uchar(
    const void *v1,
    const void *v2)
{
    const unsigned char *c1 = v1, *c2 = v2;
    if      (*c1 < *c2) return -1;
    else if (*c1 > *c2) return +1;
    else                return  0;
}
   
    



/**************************************************************************
                    Array Structure Manipulations
**************************************************************************/


/*-------------------------------------------------------------------------
vxl_create_array - Create and initialize a voxel array

This function allocates memory for a voxel array, initializes it to
have no dimensions and to contain nothing, and returns a pointer to
the newly created voxel array.  The function vxl_destroy_array should
be used to destroy the voxel array and contents once it is no longer
useful.
---------------------------------------------------------------------------*/

voxel_array *
vxl_create_array()
{
    voxel_array *dest;

    dest = mallock(sizeof(voxel_array));
    vxl_init_array(dest);

    return dest;
}



/*-------------------------------------------------------------------------
vxl_create_scalar - Create a scalar voxel array

This function allocates memory for a voxel array, initializes it to
contain a rank zero array (that is, a scalar array) containing a
specified scalar value of some given type, and returns a pointer to
that voxel array.  The function vxl_destroy_array should be used to
destroy the voxel array and contents once it is no longer useful.
---------------------------------------------------------------------------*/

voxel_array *
vxl_create_scalar(
    int type,                  /* Type of the scalar */
    double value)              /* Value to initialize to */
{
    voxel_array *dest;

    dest = vxl_create_array();
    vxl_alloc_array(dest, type, 0, NULL);
    exim_set_value(dest->data, type, 0, value);

    return dest;
}




/*--------------------------------------------------------------------------
vxl_destroy_array - Destroy a dynamically allocated voxel array

This function destroys a voxel array created using vxl_create_array,
returning both it and any contents to the heap.  If you attempt to
destroy a voxel array not created by vxl_create_array or to access one
which has been destroyed, the results are unpredictable; if you are
lucky, your program will crash before it destroys any of your data.
---------------------------------------------------------------------------*/

void 
vxl_destroy_array(
    voxel_array *dest)          /* Voxel array to be destroyed */
{
    /* If pointer is zero, just return silently */
    if (dest == 0)  return;

    if (dest->magic != VXL_MAGIC)
	fatal("vxl_destroy_array: This doesn't look like a voxel array!");

    if (dest->length > 0 && dest->data != 0)
	free(dest->data);
    free(dest);

    return;
}



/*-------------------------------------------------------------------------
vxl_init_array - Initialize a voxel array

Given a voxel_array structure, this function initializes its parameters
to default values, including the auxinfo, usertype, and userinfo fields, 
but does not allocate any memory for its contents.  Any previous contents
of the voxel array are lost forever.

This function differs from vxl_create_array in that the voxel array to be
initialized must already exist; it typically has been declared as voxel_array
in a program.  It is not yet clear which approach is the better.
---------------------------------------------------------------------------*/

void
vxl_init_array(
    voxel_array *dest)         /* Voxel array to initialize */
{
    int i;

    dest->magic    = VXL_MAGIC;
    dest->rank     = 0;
    dest->error    = 0;
    dest->type     = 0;
    dest->nbytes   = 0;
    dest->length   = 0;
    dest->auxinfo  = 0;
    dest->usertype = 0;
    dest->userinfo = 0;
    dest->data     = 0;

    for (i = 0; i < VXL_MAX_RANK; i++) {
	dest->dimen[i] = 0;
        dest->origin [i] = 0.0;
	dest->spacing[i] = 1.0;
    }

    return;
}



/*-------------------------------------------------------------------------
vxl_store_scalar - Store a scalar value as a voxel array

Given a voxel_array structure, this function allocates or reallocates
storage to contain a single scalar of some given type and initializes
it to contain that scalar value.  The rank, type, nbytes, length, and
data fields are set appropriately.  The error field is set to zero.
The auxinfo, usertype, and userinfo fields are unchanged.  Note that
the voxel array must already have been initialized via vxl_create_array
or vxl_init_array before it is possible to store into it.
---------------------------------------------------------------------------*/

void
vxl_store_scalar(
    voxel_array *dest,         /* Voxel array to initialize */
    int type,                  /* Type of the scalar */
    double value)              /* Value to initialize to */
{
    /* Initialize the voxel array to default values */
    vxl_alloc_array(dest, type, 0, NULL);

    /* Store the scalar value */
    exim_set_value(dest->data, type, 0, value);

    return;
}



/*-------------------------------------------------------------------------
vxl_alloc_array - Allocate a voxel array

Given a voxel_array structure, this function allocates or reallocates
the necessary data storage for an array of a given type and
dimensions.  The previous data, if any, is lost.  The rank, type,
dimen, nbytes, length, and data fields are set appropriately.  The
error field is set to zero.  The origin and spacing are set to zero
and one, thus making physical coordinates the same as index
coordinates; if other values are wanted, the caller must set them
separately.  The auxinfo, usertype, and userinfo fields are unchanged.

The case of a rank 0 array is handled correctly.  That is, space is
allocated for a single element of the given type; the dimen array is
ignored.

It is (currently) a fatal error if the necessary storage cannot be
allocated.

FIXME:  Does it make sense that, if dest is zero, this function should
create and allocate a new voxel array and return that?

FIXME:  If realloc fails, then the old data is lost forever.
---------------------------------------------------------------------------*/


void
vxl_alloc_array(
    voxel_array *dest,         /* Voxel array to initialize */
    int type,                  /* Desired type of array */
    int rank,                  /* Desired rank of array */
    long *dimen)               /* Array of dimensions */
{
    long len;                  /* Total length in bytes */
    int nbytes;                /* Bytes per datum */
    int i;

    /* Check for valid voxel array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("vxl_alloc_array: Invalid voxel array");

    /* Allocate the necessary data area */
    nbytes = exim_sizeof_type(type);
    if (nbytes == 0)
	fatal("vxl_alloc_array: invalid voxel type requested");
    dest->nbytes = nbytes;
    len = nbytes;
    for (i = 0; i < rank; i++) {
	dest->dimen[i] = dimen[i];
	len *= dimen[i];
        dest->origin [i] = 0.0;
	dest->spacing[i] = 1.0;
    }
    dest->length = len;
    dest->data = reallock(dest->data, len);
    if (dest->data == 0)
	fatal("vxl_alloc_array: Failed to allocate sufficient memory");

    /* Initialize other fields in the vxl_array structure */
    dest->rank  = rank;
    dest->error = 0;
    dest->type  = type;

    return;
}




/*--------------------------------------------------------------------------
vxl_free_array - Free the contents of a voxel array

This function frees the contents of a voxel array, making that storage
available for other needs; the voxel_array itself is NOT freed, so
this function is suitable for statically allocated voxel arrays.  It
generates a fatal error if it appears that that voxel_array has never been
initialized.  But if the voxel array is merely empty, then it is left
empty.

Prior to 19990318, this function would generate a fatal error if no
data area had been previously allocated; that is, if dest->data were
zero.
---------------------------------------------------------------------------*/

void
vxl_free_array(
    voxel_array *dest)        /* Voxel array to be freed */
{
    /* Check for a valid (properly initialized) voxel array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("vxl_free_array: Invalid or uninitialized voxel array");

    /* Free the data area, if any */
    if (dest->data != 0)   free(dest->data);

    /* Mark this voxel array as empty or unallocated */
    dest->error = 0;
    dest->data  = 0;

    return;
}



/*-------------------------------------------------------------------------
vxl_copy_array - Create copy of a voxel array

This function creates an exact copy of a voxel array, the attributes
and contents of which are identical to the original array but which
shares no storage with it.  No error checking is done, except to
verify that the source has the right magic number; if the source
contains garbage, the copy will contain the same garbage.  If the
pointer to the source is zero, then a zero pointer will be returned.

There is another function vxl_copy which copies the contents of a
voxel array into another voxel array but leaves the attributes of the
destination unchanged.

FIXME: The values of auxinfo and userinfo are just copied, which is
probably not the right thing to do.  As far as I remember, these fields
are never used, so consider removing them.
--------------------------------------------------------------------------*/

voxel_array *
vxl_copy_array(
    voxel_array *src)        /* Voxel array to be copied */
{
    voxel_array *dest;       /* Copy of the voxel array struct */
    int flag, i;

    /* Just copy a NULL pointer */
    if (src == 0)
	return 0;

    /* Verify that the source has the right magic number */
    if (src->magic != VXL_MAGIC)
	fatal("Attempted to copy something not a voxel array");

    /* Create a new array struct and copy the old one into it */
    dest = mallock(sizeof(voxel_array));
    dest->magic    = VXL_MAGIC;
    dest->rank     = src->rank;
    dest->error    = src->error;
    dest->type     = src->type;
    dest->nbytes   = src->nbytes;
    dest->length   = src->length;
    dest->auxinfo  = src->auxinfo;
    dest->usertype = src->usertype;
    dest->userinfo = src->userinfo;

    for (i = 0; i < VXL_MAX_RANK; i++) {
	dest->dimen  [i] = src->dimen[i];
        dest->origin [i] = src->origin[i];
	dest->spacing[i] = src->spacing[i];
    }

    /* Copy the data, if any */
    dest->data = 0;
    if (src->data != 0) {
	dest->data = mallock(src->length);
        flag = bips_copy(src->length, INTP_UCHAR, dest->data, 1,
			 src->data, 1);
	if (flag != 0)
	    fatal("Error calling BIPS");
    }

    /* Return a pointer to the copy */
    return dest;
}



/*--------------------------------------------------------------------------
vxl_reshape_array - Change rank and dimensions of a voxel array

This function changes the rank and dimensions of a voxel array while
keeping the data itself unchanged.  (The type cannot be portably
changed, so this is not permitted.)  It is a fatal error if the total
number of elements is not conserved.
---------------------------------------------------------------------------*/

void
vxl_reshape_array(
    voxel_array *dest,    /* Destination voxel array */
    int rank,             /* New rank */
    long *dimen)          /* New dimensions */
{
    int i;
    long len;

    /* Check validity of arguments */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("vxl_reshape_array: Invalid voxel array as destination");
    if (dest->data == 0)
	fatal("vxl_reshape_array: Attempt to reshape empty array");
    if (rank < 0 || rank > VXL_MAX_RANK)
	fatal("vxl_reshape_array: Invalid rank requested");
   
    /* Check that new length equals old length */
    len = dest->nbytes;
    for (i = 0; i < rank; i++) {
	if (dimen[i] <= 0)
	    fatal("vxl_reshape_array: invalid dimension requested");
	len *= dimen[i];
    }
    if (len != dest->length)
	fatal("vxl_reshape_array: Attempt to change number of elements");

    /* Change the shape as requested */
    dest->rank = rank;
    for (i = 0; i < rank; i++)
	dest->dimen[i] = dimen[i];

    return;
}



/**************************************************************************
                        Voxel Access Functions
**************************************************************************/



/*--------------------------------------------------------------------------
vxl_get_voxel - Get the value of a specified voxel (as a double)

This function returns the contents of the voxel at specified
coordinates as a double value; it is a fatal error if the given
coordinates are not valid.  A double value should suffice to
hold an actual value of any non-complex type, except on really
bizarre platforms; since you should use this function to examine
isolated values, any loss of efficiency should be minor.

FIXME: There is no check that the coords array has the same length as
the rank of the voxel array.

FIXME: Consider providing an option to return a default value if the
indices are outside the image.
---------------------------------------------------------------------------*/

double
vxl_get_voxel(
    voxel_array *src,              /* Voxel array to examine */
    long *coords)                  /* Point coordinates */
{
    long offset;                   /* Position in vectorized array */

    /* Compute location of desired element; vxli_offset returns a 
       negative value if the coordinates are invalid. */
    offset = vxli_offset(src, coords, 0);
    if (offset < 0)
        fatal("vxl_get_voxel: Subscript is out of range");

    /* Extract and return the right element */
    return exim_get_value(src->data, src->type, offset);
}



/*--------------------------------------------------------------------------
vxl_set_voxel - Set the value of a specified voxel (from a double)

This function sets the value of a voxel at specified coordinates to
a value given as a double and converted to the type of the voxel array
(using the usual C conversions).  It is a fatal error if the given
coordinates are outside the array.

FIXME:  There is no check on the length of the coords array.
---------------------------------------------------------------------------*/

void
vxl_set_voxel(
    voxel_array *dst,             /* Voxel array to modify */
    long *coords,                 /* Location to modify */
    double value)                 /* Value to convert and set */
{
    long offset;                  /* Position in vectorized array */

    /* Compute location of desired element; vxli_offset returns a 
       negative value if the coordinates are invalid. */
    offset = vxli_offset(dst, coords, 0);
    if (offset < 0)
        fatal("vxl_get_voxel: Subscript is out of range");

    /* Set the right element */
    exim_set_value(dst->data, dst->type, offset, value);
}




/*************************************************************************
                    Reading and Writing Slices

This section includes functions for reading or writing slices of a
voxel array, where a slice is defined by taking a subset of the
possible index values along each axis.  Possible options for each axis
include taking all possible index values, taking a set of specified
values (possibly only a single value), or taking an arithmetic
sequence of index values specified by an initial value, a limit value,
and a step.  Following the usual convention in C and Python (but not
Fortran, Matlab, or Octave), the indices start at 0 and the limit
value is approached but never actually attained.  It is a fatal error
if no indices are selected along any axis or if the step argument in
an arithmetic sequence is zero.

The nidx argument specifies the number of index values to take along
each axis, and the index argument specifies the actual index values.
There are two special values for nidx[i]: The value 0 means to take
all the indices present along that axis; the corresponding index array
is ignored.  The value -1 means to take an arithmetic sequence of
indices; the corresponding index must contain the starting value,
limit value, and step in that order.  It is permitted  to specify a
slice which contains only a single element, but it may not be efficient
(and often requires special handling in the algorithms).

This code is optimized for the case of large arrays; it will not be
particularly efficient at handling many small arrays.

FIXME: Establish a consistent nomenclature for rank, dimension, axis,
index, coordinate, etc.

The functions vxli_prep_slice and vxli_next_slice are wholly owned
subsidiaries of the public functions vxl_get_slice and vxl_set_slice;
their purpose is to encapsulate in one place some operations
common to the public functions, to make maintenance a little easier.
There is some rather messy code in these functions, to handle all the
possible cases without sacrificing too much efficiency; take care if
you decide to modify these functions.

ALGORITHMS: All of the actual data copying is done in one of the
optimized BIPS functions bips_copy, bips_gather, or bips_scatter; all
of the code that you see here is just to divide the labor into
suitable chunks for BIPS.  For efficiency, the algorithm divides the
axes of the source array into three sets.  The prime axis is the last
axis with more than one index value, and is processed within BIPS for
efficiency.  The leading axes are those which precede the prime axis,
and are handled by looping through the possibilities.  The trailing
axes are those which follow the prime axis and have only a single
index value to extract.  The sets of leading and trailing axes may be
empty.  If every axis selects only a single index, then the prime axis
is taken as the first axis.

Considerable care has been taken to ensure that the following special
cases are handled correctly: The source has rank 0; every axis has
only one index value selected; the set of leading or trailing axes is
empty; there is only one element in the selected slice.  If you modify
this code, make sure that it still handles these special cases.
*************************************************************************/


/*------------------------------------------------------------------------
vxli_prep_slice - Prepare a slice description for processing

This function translates a more or less human-readable slice description
(nidx and index) into the form which permits rapid execution using the
BIPS functions.  It is used by vxl_get_slice and vxl_put_slice and is
probably not useful anywhere else.  Fatal errors are reported by calling
fatal() for appropriate handling.  This code assumes that the array to
be sliced has already been checked for validity; the nidx and index are
checked here.  If only one index value is specified on each axis or the
array being sliced is rank 0, then prime is set to -1 to indicate this; 
note that this case often requires special handling in a fast, row-oriented 
algorithm.

FIXME: Make this inline for a little added performance?
------------------------------------------------------------------------*/

#define PAPRIM  10           /* Pre-allocated space for prime axis */

static long *
vxli_prep_slice(
    voxel_array *slc,        /* The voxel array to be sliced */
    long *nidx,              /* What to do with each axis */
    long **index,            /* The indices for each axis */
    long *nmbr,              /* (out) Nmbr elements to xfer for each axis */
    long *init,              /* (out) Initial index for each axis */
    long *step,              /* (out) Index increment for each axis */
    long *scor,              /* (out) Initial coords for sliced array */
    long *ucor,              /* (out) Initial coords for unsliced array */
    int  *prime,             /* (out) Prime axis for BIPS */
    long *sidx)              /* (in/out) Offsets along prime axis */
{
    int rank;                /* Number of dimensions in sliced array */
    int iaxis;               /* Index over axes */
    int paxis;               /* Prime axis */
    long nidxp;              /* The value of nidx(*prime) */
    long ss;                 /* Stride along prime axis */
    long i;

    /* Handle rank 0 as a special case */
    rank = slc->rank;
    if (rank == 0) {
        *prime = -1;
        return sidx; }

    /* Validate arguments */
    if (nidx == 0)
        fatal("The nidx argument provided was zero");
    if (index == 0)
        fatal("The index argument provided was zero");
    for (iaxis = 0; iaxis < rank; iaxis++)
        if (nidx[iaxis] != 0 && index[iaxis] == 0)
            fatal("Null pointer found in the index array");
    if (sidx == 0)
        fatal("Internal error: no preallocated storage was provided");

    /* Compute dimensions of the slice */
    for (iaxis = 0; iaxis < rank; iaxis++) {
        if ( nidx[iaxis] > 0 ) {
            /* Desired indices are given explicitly */
            nmbr[iaxis] = nidx[iaxis];
            init[iaxis] = index[iaxis][0];
            step[iaxis] = 1;
            for (i = 0; i < nmbr[iaxis]; i++)
                if ( index[iaxis][i] < 0 ||
                     index[iaxis][i] >= slc->dimen[iaxis] )
                    fatal("Index outside sliced array");
        }
        else if ( nidx[iaxis] == 0 ) {
            /* Desired indices constitute the entire axis */
            nmbr[iaxis] = slc->dimen[iaxis];
            init[iaxis] = 0;
            step[iaxis] = 1;
        }
        else {
            /* Desired indices given by an arithmetic sequence */
	    /* NOTE that division of negative integers is not portable. */
	    long frst, last, delt;
            frst = init[iaxis] = index[iaxis][0];
            delt = step[iaxis] = index[iaxis][2];
	    last = index[iaxis][1];
	    if (delt > 0 && last >= frst)
		nmbr[iaxis] = (last - frst + delt - 1) / delt;
	    else if (delt < 0 && last <= frst)
		nmbr[iaxis] = (frst - last - delt - 1) / (-delt);
	    else
		fatal("Slice object defines infinite sequence");
            if (init[iaxis] < 0 || index[iaxis][1] > slc->dimen[iaxis])
                fatal("Index outside sliced array");
        }
        if ( nmbr[iaxis] < 1 )
            fatal("Some slice axis has non-positive count");
    }

    /* Set initial coordinates for sliced and unsliced arrays */
    for (iaxis = 0; iaxis < rank; iaxis++) {
        scor[iaxis] = init[iaxis];
        ucor[iaxis] = 0;
    }

    /* Identify the prime axis */
    for ( paxis = rank - 1;
          paxis >= 0 && nmbr[paxis] == 1;
          paxis-- ) {};
    *prime = paxis;
    if ( paxis < 0 )
        return sidx;

    /* Precompute offsets along prime axis for BIPS */
    nidxp = nidx[paxis];
    if (nidxp > 0) {
        if (nidxp > PAPRIM) 
            sidx = mallock(nidxp*sizeof(long));
        ss = vxli_step(slc, paxis);
        for (i = 0; i < nidxp; i++)
            sidx[i] = ss * index[paxis][i];
    }
    else {
        sidx[0] = vxli_step(slc, paxis) * step[paxis];
    }

    return sidx;
}



/*------------------------------------------------------------------------
vxli_next_slice - Step to the next set of leading indices

This function increments the current values of the leading coordinates
in the sliced and unsliced arrays; it returns 1 until the last set has
been processed, whereupon it returns 0.  This function does not handle
the case prime == -1, which is normally handled as a special case
outside the loop over rows.

FIXME: Make this inline for a little added performance?

NOTE that this code has been verified to work correctly for the
special cases: prime == 0; and nmbr[prime] == 1.
------------------------------------------------------------------------*/

static int
vxli_next_slice(
    long *nidx,              /* How to scan each axis */
    long **index,            /* Indices for each axis */
    long *nmbr,              /* Number of indices on each axis */
    long *init,              /* Initial index on each axis */
    long *step,              /* Step size on each axis */
    long *scor,              /* Current indices for sliced array */
    long *ucor,              /* Current indices for unsliced array */
    int  prime)              /* Prime axis */
{
    int ilead;               /* Index over leading axes */

    for (ilead = prime-1; ilead >= 0; ilead--) {
        ucor[ilead] += 1;
        if ( ucor[ilead] < nmbr[ilead] ) {
            if ( nidx[ilead] > 0 )
                scor[ilead] = index[ilead][ucor[ilead]];
            else
                scor[ilead] += step[ilead];
            break; }
        else {
            ucor[ilead] = 0;
            scor[ilead] = init[ilead]; }
    }

    return ( ilead > 0 || ucor[0] != 0 );
}


/*------------------------------------------------------------------------
vxli_describe_slice - Dump a slice description for debugging

This function prints out (on stderr) a human-readable description of
the slice defined by a given nidx and index.
-------------------------------------------------------------------------*/

static void
vxli_describe_slice(
    voxel_array *slc,           /* Voxel array being sliced */
    long *nidx,                 /* What to do on each axis */
    long **index)               /* Indices for each axis */
{
    int i, j;

    for (i = 0; i < slc->rank; i++) {
        fprintf(stderr, "%ld:", nidx[i]);
        if (nidx[i] > 0) {
            for (j = 0; j < nidx[i]; j++)
                fprintf(stderr, "  %ld", index[i][j]);
        }
        else if (nidx[i] < 0) {
            for (j = 0; j < 3; j++)
                fprintf(stderr, "  %ld", index[i][j]);
        }
        else /* if (nidx[i] == 0) */
            {}
        fprintf(stderr, "\n");
    }

    return;
}



/*------------------------------------------------------------------------
vxl_get_slice - Extract a slice from a voxel array

This function creates a new voxel array containing a subset of the
data from some existing voxel array; the data extracted is specified
by giving the index values to be included along each dimension.  The
rank of the output array is the same as the rank of the input, even if
some of the dimensions have length 1.  

Selecting only one element along an axis does not eliminate that axis
and reduce the rank of the result.  In particular, it is permissible
to select only one index value along each axis, but the result is
still an array of rank equal to the source.

If the source array has rank zero, then the destination array is
simply a copy of the source; the slice selection parameters are
ignored.
------------------------------------------------------------------------*/

void
vxl_get_slice(
    voxel_array *dst,           /* Destination array */
    voxel_array *src,           /* Source array */
    long *nidx,                 /* Number of indices along each axis */
    long **index)               /* Actual indices to take along each axis */
{
    int type;                   /* Data type of source and destination */
    int rank;                   /* Rank of source and destination */
    int prime;                  /* Prime axis */
    long nmbr[VXL_MAX_RANK];    /* Dimensions of the destination array */
    long dcor[VXL_MAX_RANK];    /* Current index along each axis in dest */
    long init[VXL_MAX_RANK];    /* Initial value along each source axis */
    long step[VXL_MAX_RANK];    /* Increment along each source axis */
    long scor[VXL_MAX_RANK];    /* Current index along each axis in source */
    long pasidx[PAPRIM];        /* Pre-allocation for a short sidx */
    long *sidx;                 /* Desired indices times appropriate spacing */
    unsigned char *drow;        /* Start of current row in destination */
    unsigned char *srow;        /* Start of current row in source */
    long ddel;                  /* Row-to-row increment in destination */
    long pnidx, pnmbr;          /* nidx[prime] and nmbr[prime] */

    /* (DEBUG) Print out the slice description */
    if (0)
        vxli_describe_slice(src, nidx, index);

    /* Validate the source and destination arrays */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    type = src->type;
    rank = src->rank;
    if (dst == 0 || dst->magic != VXL_MAGIC)
        fatal("Invalid destination array");

    /* Prepare slice description for fast processing */
    sidx = vxli_prep_slice(src, nidx, index, nmbr, init, step, 
                           scor, dcor, &prime, pasidx);
    pnidx = nidx[prime];
    pnmbr = nmbr[prime];

    /* Allocate the destination array */
    vxl_alloc_array(dst, type, rank, nmbr);

    /* Handle a single-element slice as a special case */
    if ( prime == -1 ) {
        srow = vxli_locate(src, scor, 1);
        drow = dst->data;
        bips_copy(1, type, drow, 1, srow, 1);
        if (sidx != pasidx) free(sidx);
    }

    /* Setup for the loop over leading indices */
    drow = dst->data;
    ddel = pnmbr*exim_sizeof_intype(type);

    /* Loop over the possible values of the leading indices */
    do {
        /* Copy values using BIPS */
        /* FIXME: Already checked that indices are within the array */
        srow  = vxli_locate(src, scor, 1);
        if ( pnidx > 0 ) {
            bips_gather(pnmbr, type, drow, 1, srow, sidx); }
        else {
            bips_copy  (pnmbr, type, drow, 1, srow, sidx[0]); }
        drow += ddel;
    }
    while (vxli_next_slice(nidx, index, nmbr, init, step, scor, dcor, prime));

    /* Release temporary array, if used */
    if (sidx != pasidx) 
        free(sidx);

    return;
}



/*------------------------------------------------------------------------
vxl_set_slice - Assign values to a slice in a voxel array

This function assigns values from some voxel array to a slice within
another voxel array; to elements to be assigned to are specified
data from some existing voxel array; the data extracted is specified
by giving the index values to be included along each dimension.  The source
may be either a scalar array, in which case the single value contained therein
is broadcast to all the destination elements, or a non-scalar array,
in which case it must contain exactly the number of elements needed by
the destination slice.  The rank and shape of a non-scalar source array
are ignored; the rank and shape of the destination are unchanged.

The nidx argument specifies the number of index values to take along
each axis, and the index argument specifies the actual index values.
There are two special values for nidx[i]: The value 0 means to take
all the indices present along that axis; the corresponding index array
is ignored.  The value -1 means to take an arithmetic sequence of
indices; the corresponding index must contain the starting value,
limit value, and step in that order.  NOTE that the given limit value
is never actually selected; this is consistent with the usual
convention in C and Python but not with Matlab or Octave.  NOTE also
that the first element has index 0, rather than 1.
------------------------------------------------------------------------*/

void
vxl_set_slice(
    voxel_array *dst,           /* Destination array */
    long *nidx,                 /* Number of indices along each axis */
    long **index,               /* Actual indices to take along each axis */
    voxel_array *src)           /* Source array */
{
    int  stype, dtype;          /* Data type of source and destination */
    int  srank, drank;          /* Rank of source and destination */
    long slen, dlen;            /* Length of source and destination */
    int  prime;                 /* Prime axis */
    long nmbr[VXL_MAX_RANK];    /* Number of elements to take on each axis */
    long dcor[VXL_MAX_RANK];    /* Current index along each axis in dest */
    long init[VXL_MAX_RANK];    /* Initial value along each source axis */
    long step[VXL_MAX_RANK];    /* Increment along each source axis */
    long scor[VXL_MAX_RANK];    /* Current index along each axis in source */
    long padidx[PAPRIM];        /* Pre-allocation for a short didx */
    long *didx = 0;             /* Element offsets within destination row */
    unsigned char *drow;        /* Start of current row in destination */
    unsigned char *srow;        /* Start of current row in source */
    long ss = 1, ds = 1;        /* Stride within a source or dest row */
    long sdel = 0;              /* Row-to-row increment in source array */
    unsigned char *buff = 0;    /* Row buffer for type conversions */
    long pnidx, pnmbr;          /* nidx[prime] and nmbr[prime] */
    int mode;                   /* Mask for: scatter => +2; cast => +1 */
    int i;

    /* (DEBUG) Print out the slice description */
    if (0)
        vxli_describe_slice(dst, nidx, index);

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;
    srank = src->rank;
    slen  = 1;
    for (i = 0; i < srank; i++)
        slen *= src->dimen[i];

    /* Validate the destination array */
    if (dst == 0 || dst->magic != VXL_MAGIC || dst->data == 0)
        fatal("Invalid or empty destination array");
    dtype = dst->type;
    drank = dst->rank;

    /* Prepare slice description for fast processing */
    didx = vxli_prep_slice(dst, nidx, index, nmbr, init, step, 
                           dcor, scor, &prime, padidx);

    /* Handle single element destination as a special case */
    if (prime == -1 && slen == 1) {
        drow = vxli_locate(dst, dcor, 1);
        srow = src->data;
        bips_cast(1, dtype, drow, 1, stype, srow, 1);
        return; 
    }
    else if (prime == -1)
        fatal("Length mismatch between source and destination slice");

    /* Verify length compatibility; set source stride */
    for (dlen  = 1, i = 0; i < drank; i++)
        dlen *= nmbr[i];
    if (slen == 1) {
        ss = 0; 
        sdel = 0; }
    else if (slen == dlen) {
        ss = 1;
        sdel = nmbr[prime]*exim_sizeof_intype(stype);
    }
    else
        fatal("Length mismatch between source and destination slice");

    /* Allocate buffer for type conversions, if needed */
    if (dtype != stype)
        buff = mallock(nmbr[prime]*exim_sizeof_intype(dtype));

    /* Set up remaining parameters for the loop */
    pnidx = nidx[prime];
    pnmbr = nmbr[prime];
    ds    = didx[0];
    mode  = (stype != dtype) ? 1 : 0;
    mode += (pnidx > 0) ? 2 : 0;

    /* Loop over the possible values of the leading indices */
    srow  = src->data;
    do {
        /* Copy values using BIPS */
        /* FIXME: Already checked that indices are within dest */
        drow = vxli_locate(dst, dcor, 1);
        switch (mode) {
        case 0:    /* No scatter, no casting */
            bips_copy(pnmbr, stype, drow, ds, srow, ss);
            break;
        case 1:    /* No scatter, type casting */
            bips_cast(pnmbr, dtype, drow, ds, stype, srow, ss);
            break;
        case 2:    /* Scatter, no casting */
            bips_scatter(pnmbr, stype, drow, didx, srow, ss);
            break;
        case 3:    /* Scatter and type casting */
            /* FIXME: Inefficient if ss == 0 */
            bips_cast(pnmbr, dtype, buff, 1, stype, srow, ss);
            bips_scatter(pnmbr, dtype, drow, didx, buff, 1);
            break;
        default:
            panic("Internal error: invalid mode");
        }
        srow += sdel;
    }
    while (vxli_next_slice(nidx, index, nmbr, init, step, dcor, scor, prime));

    /* Release temporary array, if used */
    if (didx != padidx && didx != 0) 
        free(didx);
    if (buff != 0)
        free(buff);

    return;
}




/***************************************************************************
                           Image Read and Write
***************************************************************************/


/*---------------------------------------------------------------------------
vxl_read_uchar_data - Read external uchar data and convert to a voxel array

Given a stream, this function reads unsigned char data from that stream
and converts it to a voxel array.  The rank of the voxel array and the
size of each dimension must be specified by the caller.  It is a fatal
error if the stream runs out of data before the voxel array is filled.

In most applications, it will be preferable to use vxl_read_raw_data,
since that function automatically converts the data from external to
internal formats; this function is appropriate only when the data are
known to be in the unsigned char format, or when the external-to-internal
conversion must be done after the data are read.
----------------------------------------------------------------------------*/

void
vxl_read_uchar_data(
    voxel_array *dest,      /* Destination for voxel data */
    FILE *src,              /* Stream to read data from */
    int rank,               /* Rank (number of dimensions) */
    long *dimen)            /* Array of dimensions */
{
    long count;

    /* Initialize the vxl_array structure; allocate memory */
    vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);

    /* Read the data (no conversion is needed) */
    count = fread(dest->data, 1, dest->length, src);
    if (count < dest->length)
	fatal("vxl_read_uchar_data: Premature end of data");    

    return;
}



/*------------------------------------------------------------------------
vxl_write_uchar_data - Write voxel array as external uchar data stream

Given an unsigned char volume image, this function writes the voxels of
that image as a stream of unsigned characters.

In most applications, it will be preferable to use vxl_write_raw_data,
since that function automatically converts the data from internal to
external formats; this function is appropriate only when the data are
known to be in the unsigned char format.
------------------------------------------------------------------------*/

void
vxl_write_uchar_data(
    FILE *dest,          /* Stream to write to */
    voxel_array *src)    /* Volume image to write */
{
    long count;          /* Number of voxels in the image */

    /* Validate arguments */
    if (dest == 0)
	fatal("vxl_write_uchar_data: Invalid output stream");
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_UCHAR)
	fatal("vxl_write_uchar_data: Invalid or empty source array");

    /* Write out the data */
    count = fwrite(src->data, 1, src->length, dest);
    if (count < src->length)
	fatal("vxl_write_uchar_data: Error while writing data");    
    
    return;
}



/*------------------------------------------------------------------------
vxl_read_raw_data - Read raw data file and convert to voxel array

This function reads raw image data from a specified stream in some
specified external format, converts it on the fly to a specified
internal format (or the default "natural" format corresponding to the
external format), and stores it as a voxel array.  The rank and
dimensions of the data are specified by the caller.  If the internal
type is given as INTP_NONE, then the "natural" internal type
corresponding to the external type will be used.  It is a fatal error
if the stream runs out of data before the voxel array is filled.
------------------------------------------------------------------------*/

void
vxl_read_raw_data(
    voxel_array *dest,              /* Where to store the data input */
    FILE *src,                      /* File to read the data from */
    int rank,                       /* Number of dimensions */
    long *dimen,                    /* Dimensions themselves */
    int intype,                     /* Desired internal format */
    int extype,                     /* Format of the external data */
    int bigend)                     /* Big-endian? */
{
    int exbytes;                    /* Bytes per external datum */
    int inbytes;                    /* Bytes per internal datum */
    size_t count;                   /* Number of data to read */
    size_t nbuff = 1024;            /* Number of data per buffer */
    unsigned char *buff;            /* Buffer for input data */
    size_t nget;                    /* Number of data to get next */
    size_t ngot;                    /* Number of data actually gotten */
    unsigned char *data;            /* Data array in destination */
    int i;

    /* Verify the external type; find number of bytes per datum */
    exbytes = exim_sizeof_type(extype);
    if (exbytes == 0)
	fatal("Invalid external type");

    /* Determine the internal type; find number of bytes per datum */
    if (intype == INTP_NONE)
	intype = exim_natural_intype(extype);
    inbytes = exim_sizeof_type(intype);
    if (inbytes == 0)
	fatal("Invalid internal type");

    /* Initialize the voxel array; allocate memory */
    vxl_alloc_array(dest, intype, rank, dimen);
    buff = mallock(nbuff*exbytes);

    /* Read and convert the data one bufferfull at a time */
    count = vxli_count(dest);
    data = dest->data;
    while (count > 0) {
	nget = (count < nbuff) ? count : nbuff;
	ngot = fread(buff, exbytes, nget, src);
	if (ngot < nget)
	    fatal("Premature end of data, or I/O error");
	i = exim_import(nget, data, intype, 1, buff, extype, exbytes, bigend);
	if (i != 0)
	    fatal("Error while converting from external to internal format");
	count -= nget;
	data  += nget*inbytes;
    }

    free(buff);
    return;
}



/*-----------------------------------------------------------------------
vxl_write_raw_data - Convert voxels to external format and write to file

This function takes the contents of a voxel array, converts them to
some specified external data format, and writes them to a file stream.
------------------------------------------------------------------------*/

void
vxl_write_raw_data(
    FILE *dest,                   /* Where to write the data */
    int extype,                   /* External data format to use */
    int bigend,                   /* Write as big endian? */
    voxel_array *src)             /* The voxel array to write */
{
    int exbytes;                    /* Bytes per external datum */
    int intype;                     /* Type of internal data */
    int inbytes;                    /* Bytes per internal datum */
    size_t count;                   /* Number of data to write */
    size_t nbuff = 1024;            /* Number of data per buffer */
    unsigned char *buff;            /* Buffer for output data */
    size_t nput;                    /* Number of data to write next */
    size_t nact;                    /* Number of data actually written */
    unsigned char *data;            /* Data array in source */
    int flag;                       /* Error flag */

    /* Verify the external type; find number of bytes per datum */
    exbytes = exim_sizeof_type(extype);
    if (exbytes == 0)
	fatal("Invalid external type");

    /* Determine the internal type; find number of bytes per datum */
    intype  = src->type;
    inbytes = exim_sizeof_intype(intype);
    if (inbytes == 0)
	fatal("Invalid internal type");

    /* Allocate a buffer for i/o */
    buff = mallock(nbuff*exbytes);

    /* Convert and write the data one bufferfull at a time */
    count = vxli_count(src);
    data  = src->data;
    while (count > 0) {
	nput = (count < nbuff) ? count : nbuff;
	flag = exim_export(nput, buff, extype, exbytes, bigend, 
			   data, intype, 1);
	if ( flag != 0)
	    fatal("Error converting from internal to external format");
	nact = fwrite(buff, exbytes, nput, dest);
	if (nact < nput)
	    fatal("I/O error, or disk full");
	count -= nput;
	data  += nput*inbytes;
    }
    
    /* Free buffer and return */
    free(buff);
    return;
}



/***************************************************************************
                    Type Conversion Operations
***************************************************************************/



/*------------------------------------------------------------------------
vxl_uchar - Convert to unsigned char

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned char.
-------------------------------------------------------------------------*/

void
vxl_uchar(
    voxel_array *dest,          /* Destination array; always uchar */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    unsigned char *ddata;       /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_uchar(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_schar - Convert to signed char

This function converts a voxel array of any (non-complex) type into a
voxel array of type signed char.
-------------------------------------------------------------------------*/

void
vxl_schar(
    voxel_array *dest,          /* Destination array; always signed char */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    signed char *ddata;         /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_SCHAR, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_schar(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_ushort - Convert to unsigned short

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned short.
-------------------------------------------------------------------------*/

void
vxl_ushort(
    voxel_array *dest,          /* Destination array; always ushort */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    unsigned short *ddata;      /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_USHORT, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_ushort(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_short - Convert to short

This function converts a voxel array of any (non-complex) type into a
voxel array of type short.
-------------------------------------------------------------------------*/

void
vxl_short(
    voxel_array *dest,          /* Destination array; always short */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    short *ddata;               /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_SHORT, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_short(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_uint - Convert to unsigned int

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned int.
-------------------------------------------------------------------------*/

void
vxl_uint(
    voxel_array *dest,          /* Destination array; always uint */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    unsigned int *ddata;        /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_UINT, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_uint(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_int - Convert to type int

This function converts a voxel array of any (non-complex) type into a
voxel array of type int.
-------------------------------------------------------------------------*/

void
vxl_int(
    voxel_array *dest,          /* Destination array; always int */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    int *ddata;                 /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_INT, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_int(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_ulong - Convert to unsigned long

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned long.
-------------------------------------------------------------------------*/

void
vxl_ulong(
    voxel_array *dest,          /* Destination array; always ulong */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    unsigned long *ddata;       /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_ULONG, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_ulong(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_long - Convert to type long

This function converts a voxel array of any (non-complex) type into a
voxel array of type long.
-------------------------------------------------------------------------*/

void
vxl_long(
    voxel_array *dest,          /* Destination array; always long */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    long *ddata;                /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_LONG, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_long(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_float - Convert to type float

This function converts a voxel array of any (non-complex) type into a
voxel array of type float.
-------------------------------------------------------------------------*/

void
vxl_float(
    voxel_array *dest,          /* Destination array; always float */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    float *ddata;               /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_FLOAT, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_float(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_double - Convert to type double

This function converts a voxel array of any (non-complex) type into a
voxel array of type double.
-------------------------------------------------------------------------*/

void
vxl_double(
    voxel_array *dest,          /* Destination array; always double */
    voxel_array *src)           /* Source array */
{
    long count;                 /* Total number of voxels */
    double *ddata;              /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_DOUBLE, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    err = bips_double(count, ddata, 1, stype, sdata, 1);
    if (err != 0)
	fatal("Error in type conversion");

    return;
}



/*------------------------------------------------------------------------
vxl_cast - Convert to any specified type

This function converts a voxel array of any (non-complex) type into a
voxel array of any type specified by the caller.
-------------------------------------------------------------------------*/

void
vxl_cast(
    voxel_array *dest,          /* Destination array */
    int dtype,                  /* Destination type */
    voxel_array *src)           /* Source array */
{
    /* Switch according to the destination type */
    switch (dtype) {
	
    case INTP_UCHAR:
	vxl_uchar(dest, src);
	return;

    case INTP_SCHAR:
	vxl_schar(dest, src);
	return;

    case INTP_USHORT:
	vxl_ushort(dest, src);
	return;

    case INTP_SHORT:
	vxl_short(dest, src);
	return;

    case INTP_UINT:
	vxl_uint(dest, src);
	return;

    case INTP_INT:
	vxl_int(dest, src);
	return;

    case INTP_ULONG:
	vxl_ulong(dest, src);
	return;

    case INTP_LONG:
	vxl_long(dest, src);
	return;

    case INTP_FLOAT:
	vxl_float(dest, src);
	return;

    case INTP_DOUBLE:
	vxl_double(dest, src);
	return;

    default:
	fatal("Invalid destination type specified");

    }

    return;    /* Keeps compiler happy */

}



/*-------------------------------------------------------------------------
vxl_scale_to_uchar - Scale and convert to unsigned char

This function scales a image of any non-complex type and converts it
to an unsigned character image.  The caller specifies lower and upper
limits which are converted to 0 and 255 respectively; input values
outside these limits are silently converted to 0 and 255.
-------------------------------------------------------------------------*/

void
vxl_scale_to_uchar(
    voxel_array *dest,          /* Destination array; always uchar */
    voxel_array *src,           /* Source array; any non-complex type */
    double minval,              /* Input value corresponding to 0 */
    double maxval)              /* Input value corresponding to 255 */
{
    double offset, gain;        /* Scaling parameters for bips */
    long count;                 /* Total number of voxels */
    unsigned char *ddata;       /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination */
    vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    offset = -minval;
    gain   = 255.0 / (maxval - minval);
    err = bips_scale_to_uchar(count, ddata, 1, stype, sdata, 1, offset, gain);
    if (err != 0)
	fatal("Error converting to unsigned char");

    return;
}



/***************************************************************************
                      Unary Point Operations
***************************************************************************/


/*-------------------------------------------------------------------------
vxl_const - Fill voxel array with a constant value

This function sets all elements of a voxel array to a constant value.
The rank, dimensions, and type of the array must have been set previously.
The value is specified as a double and will be converted to the type
of the array.
--------------------------------------------------------------------------*/

void
vxl_const(
    voxel_array *dest,               /* Destination */
    double value)                    /* Value to set to */
{
    int type = 0;                    /* Type of the destination array */
    long count;                      /* Number of elements in destination */
    int nbytes = 0;                  /* Number of bytes per element */
    anytype val;                     /* Value converted to dest type */

    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 ||
	 (nbytes = exim_sizeof_intype((type = dest->type))) == 0 )
	fatal("Invalid destination array");
    count = vxli_count(dest);

    /* Convert value to the destination type */
    exim_set_value(val, type, 0, value);

    /* Copy the converted value to each element of the destination */
    bips_copy(count, type, dest->data, 1, val, 0);

    return;
}



/*------------------------------------------------------------------------
vxl_ramp - Fill voxel array with a linear ramp

This function sets each element of a voxel array equal to one of its
coordinates.  If the coordinate cannot be represented in the voxel
type, it will be converted from (signed) long to the voxel type
according to the usual C conversion rules.  This operation is not
permitted for empty or rank-0 arrays.  (It could be, but it's not
particularly useful and would have to be treated as a special case.)
-------------------------------------------------------------------------*/

void 
vxl_ramp(
    voxel_array *dest,         /* Destination array */
    int axis)                  /* Coordinate axis to store */
{
    int rank;                  /* Rank of the array */
    int type = 0;              /* Type of the destination */
    long *dimen;               /* Dimensions of the array */
    long nelem;                /* Number of elements per row */
    long coord[VXL_MAX_RANK];  /* Starting voxel of each row */
    unsigned char *row;        /* Location of a row */
    anytype val;               /* Axis coord converted to dest type */
    int err;
    
    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 ||
	 dest->rank == 0 ||  vxli_count(dest) == 0 ||
	 exim_sizeof_intype((type = dest->type)) == 0 )
	fatal("Invalid destination array");
    rank  = dest->rank;
    dimen = dest->dimen;
    nelem = dimen[rank-1];

    /* Check validity of the axis */
    if ( axis < 0 || axis >= rank )
	fatal("Invalid axis specified");

    /* Loop over the rows of the voxel array */
    vxli_firstrow(rank, coord);
    do {
	row = vxli_locate(dest, coord, 1);

	/* Case 1: Ramping on the last axis */
	if (axis == rank-1) {
	    err = bips_ramp(nelem, type, row, 1);
	}

	/* Case 2: Ramping on any other axis */
	else {
	    err = bips_cast(1, type, val, 0, INTP_LONG, &coord[axis], 0);
	    err = bips_copy(nelem, type, row, 1, val, 0);
	}
	
	if (err)
	    fatal("Unexpected error in BIPS");

    } 
    while ( vxli_nextrow(rank, dimen, coord) );

    return;
}



/*-------------------------------------------------------------------------
vxli_standard_unop - Common core for a "standard" unary point operation

A "standard" unary point operation in BIPS is one whose calling sequence
is consistent with the bips_unop prototype shown in bips.h; the following
function is used to invoke a standard unary point operation to do the
defined operation on two voxel_arrays.  For an example of its use, see
the function vxl_not and its friends below

FIXME:  This approach might also be the right way to implement the binary
point operations in standard form.
-------------------------------------------------------------------------*/

static inline void
vxli_standard_unop(
    bips_unop *bips_fcn,           /* Standard BIPS function to use */
    voxel_array *dest,             /* Destination array */
    voxel_array *src1)             /* Source array */
{
    int type = 0;                  /* Internal type code for data */
    long ds = 1, ss1 = 1;          /* Stride in each array */
    long count;                    /* Total number of elements to process */
    int err;

    /* Check validity of source array; count number of elements */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0
	 || exim_sizeof_intype((type = src1->type)) == 0 )
	fatal("Invalid source array");
    count = vxli_count(src1);

    /* Allocate the destination */
    vxl_alloc_array(dest, type, src1->rank, src1->dimen);

    /* Call the specified BIPS routine to do the actual computation */
    err = bips_fcn(count, type, dest->data, ds, src1->data, ss1);
    if (err != 0)
	fatal("Error computing on voxel array");

    return;
}


/*-------------------------------------------------------------------------
Standard unary point operations on voxel arrays

These definitions could be reduced to one line apiece using an appropriate
macro, but that would make them too hard for etags (and the ordinary
programmer) to find and understand.
--------------------------------------------------------------------------*/

/* FIXME: This is never used and is arguably obviated by vxl_copy_array */
void vxl_copy(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_copy, dest, src1); return;
}

void vxl_not(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_not, dest, src1); return;
}

void vxl_minus(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_minus, dest, src1); return;
}

void vxl_abs(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_abs, dest, src1); return;
}

void vxl_floor(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_floor, dest, src1); return;
}

void vxl_ceil(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_ceil, dest, src1); return;
}

void vxl_sqrt(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_sqrt, dest, src1); return;
}

void vxl_exp(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_exp, dest, src1); return;
}

void vxl_log(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_log, dest, src1); return;
}

void vxl_log10(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_log10, dest, src1); return;
}

void vxl_sin(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_sin, dest, src1); return;
}

void vxl_cos(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_cos, dest, src1); return;
}

void vxl_tan(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_tan, dest, src1); return;
}

void vxl_asin(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_asin, dest, src1); return;
}

void vxl_acos(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_acos, dest, src1); return;
}

void vxl_atan(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_atan, dest, src1); return;
}

void vxl_sinh(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_sinh, dest, src1); return;
}

void vxl_cosh(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_cosh, dest, src1); return;
}

void vxl_tanh(voxel_array *dest, voxel_array *src1)
{   vxli_standard_unop(bips_tanh, dest, src1); return;
}



/***************************************************************************
                      Binary Point Operations
***************************************************************************/



/*------------------------------------------------------------------------
vxli_setup_binpt - Setup a binary point operation on voxel arrays

This function does some setup processing that is common to all binary
point operations on voxel arrays (well, at least those that will
actually be done within bips), and is intended to make it easier to
permit one of the operands to be a scalar.  Specifically, this
function verifies that both operands are valid, non-empty, and have
the same type and that they either have the same shape or that one is
a scalar (rank 0) voxel array; it further determines the total count
of elements to be processed and determines the stride in each source
array, the rank of the result (of operating on the two arrays), and
the dimensions of the result.  The value returned is the total number
of elements to process, or zero if the arrays are invalid or
incompatible.  The case of two scalars is supported for uniformity,
but not efficient compared to the direct C calculation.
-------------------------------------------------------------------------*/

static long
vxli_setup_binpt(                    /* Returns nmbr of elements, or zero */
    voxel_array *src1,               /* First array */
    voxel_array *src2,               /* Second array */
    int  *rank,                      /* (Out) Rank of result */
    long **dimen,                    /* (Out) Dimensions of result */
    long *ss1,                       /* (Out) Stride in first array */
    long *ss2)                       /* (Out) Stride in second array */
{
    /* Check validity of first source array */
    if (src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0)
	fatal("Invalid or empty source array src1");

    /* Check validity of second source array */
    if (src2 == 0 || src2->magic != VXL_MAGIC || src2->data == 0)
	fatal("Invalid or empty source array src2");

    /* Check that source types are the same */
    if (src1->type != src2->type)
	fatal("Source types are different");

    /* Check that shapes are compatible and determine strides */
    if (src1->rank == 0 && src2->rank ==0) {
	*rank  = 0;
	*dimen = 0;     /* Should never be dereferenced */
	*ss1   = 0;
	*ss2   = 0;
	return 1; }
    else if (src1->rank == 0) {
	*rank  = src2->rank;
	*dimen = src2->dimen;
	*ss1   = 0;
	*ss2   = 1;
	return vxli_count(src2); }
    else if (src2->rank == 0) {
	*rank  = src1->rank;
	*dimen = src1->dimen;
	*ss1   = 1;
	*ss2   = 0;
	return vxli_count(src1); }
    else if (vxli_same_shape(src1, src2)) {
	*rank  = src1->rank;
	*dimen = src1->dimen;
	*ss1   = 1;
	*ss2   = 1;
	return vxli_count(src1); }
    else
	return 0;
}


/*-------------------------------------------------------------------------
vxl_compare - Compare two voxel arrays

This function compares two voxel arrays element by element and generates
an usigned char voxel array of the same shape, each element of which is
one of three given values according to whether an element of the first
array is less than, equal to, or greater than the corresponding element
of the second array.  The two arrays may be of any non-complex type; 
either one may be a scalar, in which case that scalar is used in all the
comparisons.

FIXME: Write a more general function which allows lsval, eqval, and
gtval to be arrays.
------------------------------------------------------------------------*/

void
vxl_compare(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2,             /* Second source array */
    unsigned char lsval,           /* Output value if less than */
    unsigned char eqval,           /* Output value if equal */
    unsigned char gtval)           /* Output value if greater than */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);

    /* Do the actual comparison */
    err = bips_compare(count, dest->data, ds, 
		       src1->type, src1->data, ss1, src2->data, ss2,
		       lsval, eqval, gtval);
    if (err != 0)
	fatal("Error comparing voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_and - Bitwise AND of two voxel arrays

This function computes the elementwise bitwise AND of two voxel arrays
of the same integer type and generates an output array of the same size
and type.  Either array may be a scalar array, in which case that scalar
value is used for all computations.
------------------------------------------------------------------------*/

void
vxl_and(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_and(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_or - Bitwise OR of two voxel arrays

This function computes the elementwise bitwise OR of two voxel arrays
of the same integer type and generates an output array of the same size
and type.  Either array may be a scalar array, in which case that scalar
value is used for all computations.
------------------------------------------------------------------------*/

void
vxl_or(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_or(count, type, dest->data, ds, 
		  src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_xor - Bitwise XOR of two voxel arrays

This function computes the elementwise bitwise XOR of two voxel arrays
of the same integer type and generates an output array of the same size
and type.  Either array may be a scalar array, in which case that scalar
value is used for all computations.
------------------------------------------------------------------------*/

void
vxl_xor(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_xor(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_min - Elementwise minimum of two voxel arrays

This function computes the elementwise minimum of two voxel arrays of the
same type and generates an output array of the same size and type.
Either array may be a scalar array, in which case that scalar value is
used for all computations.
------------------------------------------------------------------------*/

void
vxl_min(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_min(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_max - Elementwise maximum of two voxel arrays

This function computes the elementwise maximum of two voxel arrays of the
same type and generates an output array of the same size and type.
Either array may be a scalar array, in which case that scalar value is
used for all computations.
------------------------------------------------------------------------*/

void
vxl_max(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_max(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_add - Addition of two voxel arrays

This function computes the elementwise sum of two voxel arrays of the
same type and generates an output array of the same size and type.
Either array may be a scalar array, in which case that scalar value is
used for all computations.
------------------------------------------------------------------------*/

void
vxl_add(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_add(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_sub - Difference of two voxel arrays

This function subtracts the second voxel array from the first voxel by
voxel and generates an output array of the same size and type.  The
two arrays must be the same numeric type.  Either array may be a scalar
array, in which case that scalar value is used for all computations.
------------------------------------------------------------------------*/

void
vxl_sub(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_sub(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_mul - Product of two voxel arrays

This function computes the elementwise product of two voxel arrays of
the same numeric type and generates an output array of the same size
and type.  Either array may be a scalar array, in which case that
scalar value is used for all computations.
------------------------------------------------------------------------*/

void
vxl_mul(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_mul(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_div - Quotient of two voxel arrays

This function computes the elementwise quotient of two voxel arrays of
the same numeric type and generates an output array of the same size
and type.  Either array may be a scalar array, in which case that
scalar value is used for all computations.  The handling of division
by zero is inherited from the underlying C implementation.
------------------------------------------------------------------------*/

void
vxl_div(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_div(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_pow - Exponentiation of two voxel arrays

This function computes the elementwise power function (or exponential)
of two voxel arrays of the same floating point type and generates an
output array of the same size and type; that is, dest = pow(src1,
src2) is computed voxel by voxel.  Either array may be a scalar array,
in which case that scalar value is used for all computations.  The
handling of domain and range errors is inherited from the underlying
C implementation.
------------------------------------------------------------------------*/

void
vxl_pow(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_pow(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/*-------------------------------------------------------------------------
vxl_atan2 - Two-argument arc tangent of two voxel arrays

This function computes the elementwise two-argument arc tangent of two
voxel arrays of the same floating point type and generates an output
array of the same size and type; that is, dest = atan2(src1, src2) is
computed for each voxel.  Either array may be a scalar array, in which
case that scalar value is used for all computations.  The handling of
domain and range errors is inherited from the underlying C
implementation.
------------------------------------------------------------------------*/

void
vxl_atan2(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen;                   /* Dimensions of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	fatal("Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &ss1, &ss2);
    if (count <= 0)
	fatal("Source arrays are incompatible");

    /* Allocate the destination image */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Do the actual comparison */
    err = bips_atan2(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)
	fatal("Error computing on voxel arrays");

    return;
}



/************************************************************************
                       Graphical Operations
************************************************************************/


/*----------------------------------------------------------------------
vxl_fill_2d_contour - Fill contour in a 2d image

This function fills the interior of a contour or polyline on a 2-D
image with a specified value.  The given fill value is cast to
whatever type the array is.  The contour is defined by the vertices
given in the array yx, with y and x values appearing alternately.  If
the last coordinate pair is not identical to the first, then a segment
from the last to the first is assumed to close the contour.  The
interior of the contour is defined by the even/odd rule: if a ray to
infinity from a point crosses the boundary contour an odd number of
times, then that point is inside the boundary; otherwise it is
outside.

Integer-valued coordinates are assumed to correspond to the center of
a voxel.  The ambiguity that occurs when a voxel center lies exactly
on a boundary vertex or segment is handled by pretending that the
voxel center is actually located at the given coordinates plus a
positive epsilon; epsilon is chosen large enough to avoid the
ambiguity but small enough to avoid creating another ambiguity.
-----------------------------------------------------------------------*/

#define MAX_INTER  256          /* Maximum intersections of row and segment */

void
vxl_fill_2d_contour(
    voxel_array *canvas,        /* 2D array to draw into */
    long npoint,                /* Number of points in contour */
    double *yx,                 /* (y,x) pairs defining contour vertices */
    double value)               /* Value to fill with */
{
    int erank;                  /* Effective rank = # of non-unit dimensions */
    long dimen[VXL_MAX_RANK];   /* Effective dimensions of the canvas */
    int  type;                  /* Data type of the canvas */
    long coord[2];              /* Index over rows of the canvas */
    unsigned char *drow;        /* Current row of canvas */
    double ymin, ymax;          /* Vertical extent of contour */
    double xmin, xmax;          /* Horizontal extent of canvas */
    double xint[MAX_INTER];     /* Sorted list of intersections */
    int  nint;                  /* Number of intersections in xint */
    long nseg;                  /* Number of segments in boundary */
    long iseg;                  /* Index over segments of contour */
    double y1, x1, y2, x2;      /* Endpoints of current boundary segment */
    double yi, xi;              /* Coordinates of an intersection */
    long first, limit;          /* A range of columns within contour */
    int flag, i, j;

    /* Validate the array to draw into; get useful parameters */
    if ( canvas == 0 || canvas->magic != VXL_MAGIC ||
	 canvas->data == 0 )
	fatal("Invalid or empty canvas");
    erank = 0;
    for (i = 0; i < canvas->rank; i++)
	if (canvas->dimen[i] > 1)
	    dimen[erank++] = canvas->dimen[i];
    if (erank != 2)
	fatal("Canvas is not two-dimensional");
    type  = canvas->type;
    xmin = 0.0;
    xmax = dimen[1] - 1;

    /* Determine vertical extent of the contour */
    ymin = dimen[0];
    ymax = 0;
    for (iseg = 0; iseg < npoint; iseg++) {
	if (yx[2*iseg] < ymin)  ymin = yx[2*iseg];
	if (yx[2*iseg] > ymax)  ymax = yx[2*iseg];
    }

    /* Determine number of segments in the boundary */
    if (yx[0] == yx[2*npoint-2] && yx[1] == yx[2*npoint-1])
	nseg = npoint - 1;
    else
	nseg = npoint;
    if (nseg < 3)
	warning("vxl_fill_2d_contour: Fewer than 3 segments in the contour");

    /* Loop over the rows that intersect the contour */
    for (yi = 1 + floor(ymin); yi <= ceil(ymax); yi++) {
	coord[1] = 0;
	coord[0] = yi;

	/* Loop over the segments in the contour and find intersections */
	nint = 0;
	for (iseg = 0; iseg < nseg; iseg++) {

	    /* Find endpoints of the current segment */
	    y1 = yx[2*iseg];
	    x1 = yx[2*iseg+1];
	    if (iseg < npoint-1) {
		y2 = yx[2*iseg+2];
		x2 = yx[2*iseg+3];
	    }
	    else {
		y2 = yx[0];
		x2 = yx[1];
	    }

	    /* Test for an intersection with the current row */
	    if ( (y1 < yi && yi <= y2) ||
		 (y2 < yi && yi <= y1 ) ) {

		/* If found, clip to left/right edges of array */
		xi = x1 + (x2-x1)*(yi-y1)/(y2-y1);
		if (xi < xmin) xi = xmin;
		if (xi > xmax) xi = xmax;

		/* And add to sorted list of intersections */
		if (nint == MAX_INTER)
		    fatal("Too many intersections");
		for (i = 0; i < nint && xi >= xint[i]; i++) 
		    /* Do nothing */ ;
		for (j = nint; j > i; j--)
		    xint[j] = xint[j-1];
		xint[i] = xi;
		nint++;
	    }
	    
	}

	/* Fill portions of row within the contour */
	if (nint % 2 != 0)
	    fatal("Odd number of intersections");
        drow = (unsigned char *)canvas->data +
	       coord[0]*dimen[1]*(canvas->nbytes);
	for (i = 0; i < nint; i += 2) {
	    first = 1 + floor(xint[i]);
	    limit = 1 +  ceil(xint[i+1]);
	    flag  = bips_fill_range(first, limit, type, drow, 1, value);
	    if (flag)
		fatal("Error calling BIPS");
	}
    }

    return;
}



/************************************************************************
                 Miscellaneous Simple Operations
************************************************************************/


/*-------------------------------------------------------------------------
vxl_logcomp_ulong - Log-compress an unsigned long volume image

This function compresses the dynamic range of an unsigned long volume
image using the transformation y = A*log(1+x), where A is chosen such
that the largest voxel value actually present in the image is converted
to the value 255; the compressed image is saved as unsigned char.
--------------------------------------------------------------------------*/

void
vxl_logcomp_ulong(
    voxel_array *dest,         /* Destination array; always uchar */
    voxel_array *src)          /* Source array; always ulong */
{
    unsigned long minv, maxv;  /* Range of values in source */
    double A;                  /* Scaling factor for compressed values */
    unsigned char *ddata;      /* Actual array of destination values */
    unsigned long *sdata;      /* Actual array of source values */
    long count;                /* Number of voxels */
    long i;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_ULONG)
	fatal("vxl_logcomp_ulong: Invalid or empty source array");

    /* Allocate the destination argument */
    vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));

    /* Determine the appropriate value of A */
    vxl_minmax_ulong(&minv, &maxv, src);
    A = 255.99/log(1.0+maxv);

    /* Transform the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    for (i = 0; i < count; i++)
	ddata[i] = floor(A*log(1.0001+(double)sdata[i]));

    return;
}


/*------------------------------------------------------------------------
vxl_lookup - Process volume image through a lookup table

This function takes each voxel in the source image and uses it as the
index into a lookup table, saving the result of the lookup as the
destination image.  The source image must be either unsigned char or
unsigned short but may be of any shape.  The table image must be rank
1 and contain at least as many entries as the largest voxel value
within the source image; but it (and thus the destination) may be of
any type.
-------------------------------------------------------------------------*/

void
vxl_lookup(
    voxel_array *dest,            /* Destination array */
    voxel_array *src,             /* Source array */
    voxel_array *table)           /* Lookup table */
{
    long count;                   /* Total number of voxels in source */
    int err;

    /* Validate the source argument */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    if ( src->type != INTP_UCHAR && src->type != INTP_USHORT )
	fatal("Source must be unsigned char or unsigned short");

    /* Validate the table array */
    if ( table == 0 || table->magic != VXL_MAGIC ||
	 table->data == 0 || table->rank != 1)
	fatal("Invalid or empty lookup table");

    /* Allocate the destination and process using BIPS */
    count = vxli_count(src);
    vxl_alloc_array(dest, table->type, src->rank, src->dimen);
    err = bips_lookup(count, table->type, src->type, dest->data, 1, 
		      src->data, 1, table->data, table->dimen[0]);
    if(err != 0)
	fatal("Call to BIPS returned an error");

    return;
}



/*-------------------------------------------------------------------------
vxl_scale - Scale and offset an image of any type

This function casts each element of a voxel array to double,
multiplies it by a double constant, adds another double constant, and
casts it back to the original type, rounding and limiting as
necessary.  The algorithm used here makes better use of cache and is
more efficient than the separate calculations for large arrays of
integral type.
---------------------------------------------------------------------------*/

void
vxl_scale(
    voxel_array *dest,              /* Destination array */
    voxel_array *src,               /* Source array */
    double gain,                    /* Constant to multiply by */
    double bias)                    /* Constant to add */
{
    int type;                       /* Type of source data */
    int rank;                       /* Rank of source */
    long *dimen;                    /* Dimensions of source */
    long coord[VXL_MAX_RANK];       /* Coordinates of current row */
    long nelem;                     /* Number of elements in a row */
    unsigned char *srow;            /* Addr of current row in source */
    unsigned char *drow;            /* Addr of current row in destination */
    double *buffer;                 /* Working memory for one row */
    int flag;                       /* Error flag */

    /* Validate the source argument */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	fatal("Invalid or empty source array");
    type  = src->type;
    rank  = src->rank;
    dimen = src->dimen;
    nelem = dimen[rank-1];

    /* Allocate the destination */
    vxl_alloc_array(dest, type, rank, dimen);

    /* Allocate buffer for one row of doubles */
    buffer = mallock(nelem*sizeof(double));

    /* Loop over rows of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate source and destination rows */
	srow = vxli_locate(src,  coord, 1);
	drow = vxli_locate(dest, coord, 1);

	/* Cast source to double */
	flag = bips_double(nelem, buffer, 1, type, srow, 1);
	if (flag != 0)
	    fatal("Error calling BIPS");

	/* Multiply by gain and add bias */
	if (gain != 1.0) {
	    flag = bips_mul_set(nelem, INTP_DOUBLE, buffer, 1, &gain, 0);
	    if (flag != 0)
		fatal("Error calling BIPS");
	}
	if (bias != 0.0) {
	    flag = bips_add_set(nelem, INTP_DOUBLE, buffer, 1, &bias, 0);
	    if (flag != 0)
		fatal("Error calling BIPS");
	}

	/* Cast back to original type */
	flag = bips_round_cast(nelem, type, drow, 1, INTP_DOUBLE, buffer, 1);
	if (flag != 0)
	    fatal("Error calling BIPS");

    }
    while( vxli_nextrow(rank, dimen, coord) );

    /* Free working memory and return */
    free(buffer);
    return;
}



/*************************************************************************
                 Statistical Processing of Images

Includes voxel classification algorithms.  Because of their similarity
to the other reduction operations, moment calculations are considered
to be reduction operations and are included in reduce.d, though a user-
oriented classification would probably still consider them as statistical
operations.
*************************************************************************/



/*-------------------------------------------------------------------------
vxl_histo_uchar - Compute histogram of an unsigned char volume image

This function computes the univariate histogram of an unsigned char
volume image and stores it as a one-dimensional volume image.  An
optional weight array may be provided, and defines the value to be
added to the histogram for each corresponding voxel; specifying a NULL
pointer indicates a weight of 1.  The source and weight arrays must be
unsigned char.  The histogram entries are NOT checked for overflow, so
this function should not be used with large arrays and weights.
-------------------------------------------------------------------------*/

void
vxl_histo_uchar(
    voxel_array *hist,       /* Place to store the histogram */
    voxel_array *src,        /* Image to take histogram of */
    voxel_array *weight)     /* Weight image */
{
    long dimen[1];           /* Dimensions of the histogram */
    unsigned long *hdata;    /* Histogram data itself */
    unsigned char *sdata;    /* Image data itself */
    unsigned char *wdata;    /* Weight data itself */
    long count;              /* Count of voxels in the image */
    long i;

    /* Validate the arguments provided */
    if (hist == 0 || hist->magic != VXL_MAGIC)
	fatal("vxl_histo_uchar: Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_UCHAR)
	fatal("vxl_histo_uchar: Invalid or empty source array");

    /* Validate the weight argument, if present */
    if (weight != 0) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     weight->type != INTP_UCHAR )
	    fatal("vxl_histo_uchar: Invalid or empty weight array");
	if ( !vxli_same_shape(src, weight) )
	    fatal("vxl_histo_uchar: "
		  "Source and weight images do not have same shape");
    }

    /* Set up the destination histogram; clear to zero */
    dimen[0] = UCHAR_MAX + 1;
    vxl_alloc_array(hist, INTP_ULONG, 1, dimen);
    hdata = hist->data;
    for (i = 0; i <= UCHAR_MAX; i++)
	hdata[i] = 0;

    /* Compute the histogram */
    sdata = src->data;
    count = vxli_count(src);
    if (weight != 0) {
	wdata = weight->data;
	for (i = 0; i < count; i++)
	    hdata[sdata[i]] += wdata[i];
    }
    else {
	for (i = 0; i < count; i++)
	    hdata[sdata[i]] += 1;
    }

    return;
}



/*-------------------------------------------------------------------------
vxl_bihist_uchar - Compute bivariate histogram of two unsigned char images

This function computes the bivariate histogram of a pair of unsigned
char images and stores it as a two-dimensional unsigned long image; it
also computes the two marginal histograms and stores them as
one-dimensional unsigned long images.  The first source image defines
the row number of the bivariate histogram; the second the column
number; the weight array, if present, determines the count added to
the histogram for the corresponding voxel.  The source and weight
arrays must all be unsigned char; the weight array may be replaced by
a NULL pointer to specify a constant weight 1.  The histogram entries
are NOT checked for overflow, so this function should not be used with
large arrays and weights.
-------------------------------------------------------------------------*/

void
vxl_bihist_uchar(
    voxel_array *hist,       /* Place to store the bivariate histogram */
    voxel_array *mar1,       /* Histogram of first image */
    voxel_array *mar2,       /* Histogram of second image */
    voxel_array *src1,       /* First image to take histogram of */
    voxel_array *src2,       /* Second image to take histogram of */
    voxel_array *weight)     /* Weight image, or zero */
{
    long dimen[2];           /* Dimensions of the histogram */
    unsigned long *hdata;    /* Data array of bivariate histogram */
    long hlen;               /* Total elements in bivar histogram */
    long hsp;                /* Spacing between rows of bivar hist */
    unsigned long *m1, *m2;  /* Data array of marginal histograms */
    unsigned char *s1, *s2;  /* Data array of source images */
    unsigned char *w;        /* Data array of weight image */
    unsigned char wi;        /* Weight at a given voxel */
    long count;              /* Count of voxels in the image */
    long i;

    /* Validate the destination arguments provided */
    if (hist == 0 || hist->magic != VXL_MAGIC)
	fatal("vxl_bihist_uchar: Invalid destination array hist");
    if (mar1 == 0 || mar1->magic != VXL_MAGIC)
	fatal("vxl_bihist_uchar: Invalid destination array mar1");
    if (mar2 == 0 || mar2->magic != VXL_MAGIC)
	fatal("vxl_bihist_uchar: Invalid destination array mar2");

    /* Validate the source arguments provided */
    if (src1 == 0 || src1->magic != VXL_MAGIC || 
        src1->data == 0 || src1->type != INTP_UCHAR)
	fatal("vxl_bihist_uchar: Invalid or empty source array src1");
    if (src2 == 0 || src2->magic != VXL_MAGIC || 
        src2->data == 0 || src2->type != INTP_UCHAR)
	fatal("vxl_bihist_uchar: Invalid or empty source array src2");
    if ( !vxli_same_shape(src1, src2) )
	 fatal("vxl_bihist_uchar: Source images do not have same shape");

    /* Validate the weight array, if provided */
    if (weight != 0) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     weight->type != INTP_UCHAR )
	    fatal("vxl_bihist_uchar: Invalid or empty weight array");
	if ( !vxli_same_shape(src1, weight) )
	    fatal("vxl_bihist_uchar: "
		  "Source and weight images do not have same shape");
    }

    /* Set up the destination histograms; clear to zero */
    dimen[0] = UCHAR_MAX + 1;
    dimen[1] = UCHAR_MAX + 1;
    hlen = dimen[0]*dimen[1];
    hsp  = dimen[1];
    vxl_alloc_array(hist, INTP_ULONG, 2, dimen);
    vxl_alloc_array(mar1, INTP_ULONG, 1, dimen);
    vxl_alloc_array(mar2, INTP_ULONG, 1, dimen);
    hdata = hist->data;
    m1    = mar1->data;
    m2    = mar2->data;
    for (i = 0; i <= UCHAR_MAX; i++)
	m1[i] = m2[i] = 0;
    for (i = 0; i < hlen; i++)
	hdata[i] = 0;

    /* Compute the histogram */
    s1 = src1->data;
    s2 = src2->data;
    count = vxli_count(src1);
    if (weight != 0) {
	w = weight->data;
	for (i = 0; i < count; i++) {
	    wi = w[i];
	    m1[s1[i]] += wi;
	    m2[s2[i]] += wi;
	    hdata[s1[i]*hsp+s2[i]] += wi;
	}
    }
    else {
	for (i = 0; i < count; i++) {
	    m1[s1[i]] += 1;
	    m2[s2[i]] += 1;
	    hdata[s1[i]*hsp+s2[i]] += 1;
	}
    }

    return;
}



/*--------------------------------------------------------------------------
vxl_kmeans1_histo - Univariate kmeans from a histogram

Given the univariate histogram of an unsigned char image and initial
guesses at class centroids, this function uses the K-means algorithm
to refine the class centroids.  If no initial guesses are available for
the class centroids, they may be initialized to random (but distinct)
values; but note that the results might possibly depend on the particular
random values chosen.  The final centroids will be sorted into ascending
order; no attempt will be made to remove duplicates.

FIXME: Generalize this to allow the histogram to contain any number of
bins?

FIXME:  There seem to be at least some cases in which setting the initial
centroids to all zeros yields two or more final centroids that are zero.
This might be related to the fact that empty centroids are not changed.

FIXME:  It's not clear how to make this work consistently when the
initial means are unknown; perhaps they should be changed to fall
at uniform intervals within the gray level range?

FIXME:  If a class is empty, then its centroid is left unchanged.  Is
this the right thing to do?
--------------------------------------------------------------------------*/

void 
vxl_kmeans1_histo(
    voxel_array *hist,       /* Histogram as 1d voxel_array */
    int nclass,              /* Nmbr of classes = length of cent array */
    unsigned char *cent)     /* Initial/final centroids */
{
    unsigned long *hd;       /* Actual data array for the histogram */
    long *cnt;               /* Count of gray levels by class */
    double *sum;             /* Sum of gray levels by class */
    int changed = 1;         /* Any class mean changed in last iteration? */
    int i;                   /* Count of iterations */
    int ig;                  /* Index over gray levels in histogram */
    int r, rb;               /* Distance to centroid, and best so far */
    int ic, icb;             /* Index over classes, and best so far */
    double mean;             /* Mean gray value of the current class */

    /* Validate histogram argument */
    if (hist == 0 || hist->magic != VXL_MAGIC || 
        hist->data == 0 || hist->type != INTP_ULONG ||
	hist->rank != 1 || hist->dimen[0] != UCHAR_MAX+1)
	fatal("vxl_kmeans1_histo: Invalid or empty histogram array");
    hd = (unsigned long *)hist->data;

    /* Validate centroid arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound 
       is more or less arbitrary but might catch some blunders. */
    if (nclass <= 1 || nclass > 512)
	fatal("vxl_kmeans1_histo: Invalid number of classes");

    /* Allocate working space for class averages */
    cnt = mallock(nclass*sizeof(long));
    sum = mallock(nclass*sizeof(double));    

    /* Iterate until class means converge.  Never do more than 100
       iterations to prevent a possible endless loop; the value 100
       is just a quick guess unsupported by analysis. */
    changed = 1;
    for (i = 0; changed && i < 100; i++) {

	/* Clear counters for class averages */
	for (ic = 0; ic < nclass; ic++) {
	    cnt[ic] = 0;   sum[ic] = 0.0; }

	/* Assign each gray level to the nearest class centroid */
	for (ig = 0; ig < UCHAR_MAX; ig++) {

	    /* Find the best class for this gray level.  The test for
	       the best centroid is "<=" so that a gray value will be
	       assigned to the last class found if or more are equally
	       good; this should mean that the final class centroids
	       are in ascending order when the initial values are all
	       zero.  */
	    rb  = INT_MAX;
	    icb = -1;
	    for (ic = 0; ic < nclass; ic++) {
		r = ig - (int)cent[ic];
		if (r < 0)  r = -r;
		if (r <= rb) {icb = ic;  rb = r; }
	    }
	    if (icb < 0 || icb >= nclass)
		fatal("vxl_kmeans1_histo (internal): Failed to find class");

	    /* Increment sum for the best class */
	    cnt[icb] += hd[ig];
	    sum[icb] += hd[ig] * ig;
	}

	/* Update class means; check if means have changed.  The new
	   means are rounded from a double computation.  NOTE that
	   roundoff error might cause the new mean to be slightly
	   larger than UCHAR_MAX, so allow for it; it should be
	   impossible for the new mean to be less than zero. */
	changed = 0;
	for (ic = 0; ic < nclass; ic++) {
	    if (cnt[ic] > 0) {
		mean = sum[ic] / (double)cnt[ic];
		if (mean < 0.0 || mean > (double)UCHAR_MAX+0.1)
		    fatal("vxl_kmeans1_histo (internal): Found invalid mean");
		if (mean != cent[ic]) {
		    changed = 1;
		    cent[ic] = (unsigned char)floor(mean+0.5); }
	    }
	}
    }

    /* Sort the centroids into ascending order */
    qsort(cent, nclass, sizeof(unsigned char), cmp_uchar);

    /* Free allocated storage and exit */
    free(cnt);
    free(sum);

    return;
}



/*-------------------------------------------------------------------------
vxl_nnclass1_uchar - Univariate nearest neighbor classification

Given an unsigned char volume image and a set of class centroids, this
function finds the nearest class centroid for each voxel position and
outputs them as a class image.  Different centroids may be assigned to
the same class number.
-------------------------------------------------------------------------*/

void
vxl_nnclass1_uchar(
    voxel_array *dest,      /* Classification image */
    voxel_array *src,       /* First feature image */
    int nclas,              /* Nmbr of classes = length of class array */
    unsigned char *clas,    /* Class id to assign for each centroid */
    unsigned char *cent)    /* Class centroids */
{
    unsigned char *sd;         /* Actual data for the source images */
    unsigned char *dd;         /* Actual data for the destination image */
    long count;                /* Number of voxels in each image */
    long i;                    /* Index over voxels */
    int ic, icb;               /* Index over classes & best so far */
    int r, rb;                 /* Distance to centroid & best so far */

    /* Check validity of source image */
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_UCHAR)
	fatal("vxl_nnclass1_uchar: Invalid or empty source array");

    /* Check validity of other arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound 
       is more or less arbitrary but should catch some blunders. */
    if (nclas <= 1 || nclas > 512)
	fatal("vxl_nnclass1_uchar: Invalid number of classes");

    /* Allocate space for the destination */
    vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));

    /* Loop over voxels in the images */
    count = vxli_count(src);
    sd    = src->data;
    dd    = dest->data;
    for (i = 0; i < count; i++) {

	/* Find the nearest class and assign it.  If two or more classes
	   are equally close, the first found is chosen.  */
	rb  = INT_MAX;
	icb = -1;
	for (ic = 0; ic < nclas; ic++) {
	    r = (int)sd[i]-(int)cent[ic];
	    if (r < 0) r = -r;
	    if (r < rb) {
	        icb = ic;
		rb  = r; }
	}
	if (icb < 0 || icb >= nclas)
	    fatal("vxl_nnclass1_uchar (internal):  Failed to find class");
	dd[i] = clas[icb];
    }

    return;
}



/*-------------------------------------------------------------------------
vxl_nnclass2_uchar - Bivariate nearest neighbor classification

Given two unsigned char volume images of the same shape and a set of
class centroids, this function finds the nearest class centroid for
each voxel position and outputs that as a class image.  Different
centroids may be assigned to the same class number.
-------------------------------------------------------------------------*/

void
vxl_nnclass2_uchar(
    voxel_array *dest,      /* Classification image */
    voxel_array *src1,      /* First feature image */
    voxel_array *src2,      /* Second feature image */
    int nclas,              /* Nmbr of classes = length of class array */
    unsigned char *clas,    /* Class id to assign for each centroid */
    unsigned char *cent1,   /* Class centroid in first feature image */
    unsigned char *cent2)   /* Class centroid in second feature image */
{
    unsigned char *sd1, *sd2;  /* Actual data for the source images */
    unsigned char *dd;         /* Actual data for the destination image */
    long count;                /* Number of voxels in each image */
    long i;                    /* Index over voxels */
    int ic, icb;               /* Index over classes & best so far */
    double r, rb;              /* Distance to centroid & best so far */
    double r1, r2;

    /* Check validity and compatibility of source images */
    if (src1 == 0 || src1->magic != VXL_MAGIC || 
        src1->data == 0 || src1->type != INTP_UCHAR)
	fatal("vxl_nnclass2_uchar: Invalid or empty source array src1");
    if (src2 == 0 || src2->magic != VXL_MAGIC || 
        src2->data == 0 || src2->type != INTP_UCHAR)
	fatal("vxl_nnclass2_uchar: Invalid or empty source array src2");
    if ( !vxli_same_shape(src1, src2) )
	 fatal("vxl_nnclass2_uchar: Source images do not have same shape");

    /* Check validity of other arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound is
       more or less arbitrary but should catch some blunders. */
    if (nclas <= 1 || nclas > 512)
	fatal("vxl_nnclass2_uchar: Invalid number of classes");

    /* Allocate space for the destination */
    vxl_alloc_array(dest, INTP_UCHAR, src1->rank, &(src1->dimen[0]));

    /* Loop over voxels in the images */
    count = vxli_count(src1);
    sd1   = src1->data;
    sd2   = src2->data;
    dd    = dest->data;
    for (i = 0; i < count; i++) {

	/* Find the nearest class and assign it */
	rb  = HUGE_VAL;
	icb = -1;
	for (ic = 0; ic < nclas; ic++) {
	    r1 = (double)(sd1[i]-cent1[ic]);
	    r2 = (double)(sd2[i]-cent2[ic]);
	    r  = r1*r1 + r2*r2;
	    if (r < rb) {
	        icb = ic;
		rb  = r; }
	}
	if (icb < 0 || icb >= nclas)
	    fatal("vxl_nnclass2_uchar (internal):  Failed to find class");
	dd[i] = clas[icb];
    }

    return;
}



/***************************************************************************
                     Image Display Operations
***************************************************************************/


/*-------------------------------------------------------------------------
vxl_2d_section - Extract a full 2D slice from a voxel array

Given a voxel array and two coordinate axes, this function extracts
the 2d slice of the voxel array specified by dim0 and dim1 and stores
it as a two-dimensional voxel array.  This is useful for image display.

The slice argument is an array of length equal to the rank of the
source; each element corresponds to one source dimension and gives the
desired coordinate value in that dimension.  The elements
corresponding to the values of dim0 and dim1 are ignored.

FIXME:  An error ought to set the error field in the destination.

FIXME:  No attempt has been made to optimize the cache hit ratio.
--------------------------------------------------------------------------*/

void
vxl_2d_section(
    voxel_array *dest,            /* Where to put the 2D section */
    voxel_array *src,             /* Array to extract section from */
    int dim0,                     /* Src dimen to become dest dimen 0 */
    int dim1,                     /* Src dimen to become dest dimen 1 */
    long *slice)                  /* Coordinates in other src dimens */
{
    long sbuff[VXL_MAX_RANK];     /* Working copy of dimensions/slice coords */
    long soff;                    /* Offset to start of data in src */
    long n0, n1;                  /* Nmbr elements in dest dimens 0 and 1 */
    long sd0, sd1;                /* Step sizes in source */
    long dd0;                     /* Step size in destination */
    long i0, i1;                  /* Index over voxels in destination */
    int i;

    /* Verify the arguments provided.  Note that the validity of the
       values in slice will be verified in vxli_offset. */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("vxl_2d_section: Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("vxl_2d_section: Invalid or empty source array");
    if (src->rank < 2)
	fatal("vxl_2d_section: Rank of source is less than 2");
    if (dim0 < 0 || dim0 >= src->rank)
	fatal("vxl_2d_section: dim0 specifies invalid slice");
    if (dim1 < 0 || dim1 >= src->rank)
	fatal("vxl_2d_section: dim1 specifies invalid slice");

    /* Set up the destination voxel array */
    n0 = sbuff[0] = src->dimen[dim0];
    n1 = sbuff[1] = src->dimen[dim1];
    vxl_alloc_array(dest, src->type, 2, sbuff);

    /* Compute location of (0,0) and step sizes for indexing */
    for (i = 0; i < src->rank; i++)
	sbuff[i] = slice[i];
    sbuff[dim0] = 0;
    sbuff[dim1] = 0;
    soff = vxli_offset(src, sbuff, 1);
    sd0  = vxli_step(src, dim0);
    sd1  = vxli_step(src, dim1);
    dd0   = dest->dimen[1];

    /* The following code is the inner loop of vxl_2d_section.  It
       is written as a macro, since it must be repeated for each
       data type.  The do {...} while (0) construction is an ugly
       artifice that lets the macro look like a function call in
       any context.  FIXME: Move outside the function, a la BIPS */
    #define vxl_2d_section_loop(type)   \
    do {  \
	type *dbase, *sbase;    /* Location of (0,0) */  \
	dbase = (type *)dest->data;  \
	sbase = (type *)src->data + soff;  \
	for (i0 = 0; i0 < n0; i0++)   \
	    for (i1 = 0; i1 < n1; i1++)   \
	        dbase[i0*dd0 + i1] = sbase[i0*sd0+i1*sd1];  \
    } while (0)

    /* Copy the requested slice, using the correct C type for data */
    switch (dest->type) {

    case INTP_UCHAR: 
	vxl_2d_section_loop(unsigned char);  
	break;

    case INTP_USHORT: 
	vxl_2d_section_loop(unsigned short);
	break;

    default:
	fatal("vxl_2d_section: Voxel type %s not yet implemented",
	      exim_nameof_type(dest->type));

    }   /* End of switch by voxel type */

    return;
}



/*-------------------------------------------------------------------------
vxl_2d_sectwin - Extract and intensity window a 2D slice from a voxel array

Given a voxel array and two coordinate axes, this function extracts
the 2d slice of the voxel array specified by dim0 and dim1, rescales
the intensity window given by gmin and gmax to the range [0..255], and
stores the result as a two-dimensional unsigned char voxel array.  The
last dimension of the voxel array runs over the bytes of a voxel; this
is to permit on-the-fly conversion from unsigned char, unsigned short
big-endian, and unsigned short little-endian in a display program
without forcing the user to know the data format in advance.

The slice argument is an array of length equal to the rank of the
source; each element corresponds to one source dimension and gives the
desired coordinate value in that dimension.  The elements
corresponding to the values of dim0 and dim1 are ignored.

FIXME:  An error ought to set the error field in the destination.

FIXME:  No attempt has been made to optimize the cache hit ratio.
--------------------------------------------------------------------------*/

void
vxl_2d_sectwin(
    voxel_array *dest,            /* Where to put the 2D section */
    voxel_array *src,             /* Array to extract section from */
    int dim0,                     /* Src dimen to become dest dimen 0 */
    int dim1,                     /* Src dimen to become dest dimen 1 */
    long *slice,                  /* Coordinates in other src dimens */
    int bigend,                   /* 1 => big-endian, 0 => little-endian */
    int gmin,                     /* Gray level that maps to 0 */
    int gmax)                     /* Gray level that maps to 255 */
{
    long sbuff[VXL_MAX_RANK];     /* Working copy of dimensions/slice coords */
    long soff;                    /* Offset to start of data in src */
    long n0, n1;                  /* Nmbr elements in dest dimens 0 and 1 */
    long sd0, sd1;                /* Step sizes in source */
    long dd0;                     /* Step size in destination */
    long i0, i1;                  /* Index over voxels in destination */
    unsigned char *dbase;         /* Address of (0,0) in the destination */
    unsigned char *sbase;         /* Address of that voxel in the source */
    int ldim;                     /* Index of last (byte) dimension */
    int nbytes;                   /* Number of bytes per voxel */
    unsigned int  sgray;          /* Intensity of a source voxel */
    unsigned char dgray;          /* Intensity of a destination voxel */
    float scale;                  /* Source-to-dest conversion factor */
    int i;

    /* Verify the arguments provided.  Note that the validity of the
       values in slice will be verified in vxli_offset. */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("vxl_2d_sectwin: Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("vxl_2d_sectwin: Invalid or empty source array");
    if (src->rank < 3)
	fatal("vxl_2d_sectwin: Rank of source is less than 3");
    if (dim0 < 0 || dim0 >= src->rank)
	fatal("vxl_2d_sectwin: dim0 specifies invalid slice");
    if (dim1 < 0 || dim1 >= src->rank)
	fatal("vxl_2d_sectwin: dim1 specifies invalid slice");
    ldim = src->rank - 1;
    nbytes = src->dimen[ldim];
    if (nbytes < 1 || nbytes > 2)
	fatal("vxl_2d_sectwin: nbytes is neither 1 nor 2");

    /* Set up the destination voxel array */
    n0 = sbuff[0] = src->dimen[dim0];
    n1 = sbuff[1] = src->dimen[dim1];
    vxl_alloc_array(dest, INTP_UCHAR, 2, sbuff);

    /* Compute location of (0,0) and step sizes for indexing */
    for (i = 0; i < src->rank; i++)
	sbuff[i] = slice[i];
    sbuff[dim0] = 0;
    sbuff[dim1] = 0;
    sbuff[ldim] = 0;        /* Ignore any byte coord from user */
    soff = vxli_offset(src, sbuff, 1);
    sd0  = vxli_step(src, dim0);
    sd1  = vxli_step(src, dim1);
    dd0   = dest->dimen[1];
    dbase = (unsigned char *)dest->data;
    sbase = (unsigned char *)src->data + soff;
    scale = 255.0 / (gmax - gmin);

    /* Copy and window the requested slice, if source is char */
    if (nbytes == 1) {
	for (i0 = 0; i0 < n0; i0++)
	    for (i1 = 0; i1 < n1; i1++) {
	        sgray = sbase[i0*sd0+i1*sd1];
		if (sgray <= gmin)
		    dgray = 0;
		else if (sgray >= gmax)
		    dgray = 255;
		else
		    dgray = scale*(sgray - gmin) + 0.5;
		dbase[i0*dd0 + i1] = dgray;
	    }
    }

    /* Copy and window the requested slice, if source is big-endian */
    else if (nbytes == 2 && bigend) {
	for (i0 = 0; i0 < n0; i0++)
	    for (i1 = 0; i1 < n1; i1++) {
	        sgray  = sbase[i0*sd0+i1*sd1] << 8;
		sgray += sbase[i0*sd0+i1*sd1+1];
		if (sgray <= gmin)
		    dgray = 0;
		else if (sgray >= gmax)
		    dgray = 255;
		else
		    dgray = scale*(sgray - gmin) + 0.5;
		dbase[i0*dd0 + i1] = dgray;
	    }
    }

    /* Copy and window the requested slice, if source is little-endian */
    else  {  /* nbytes == 2 && !bigend */
	for (i0 = 0; i0 < n0; i0++)
	    for (i1 = 0; i1 < n1; i1++) {
	        sgray  = sbase[i0*sd0+i1*sd1];
		sgray += sbase[i0*sd0+i1*sd1+1] << 8;
		if (sgray <= gmin)
		    dgray = 0;
		else if (sgray >= gmax)
		    dgray = 255;
		else
		    dgray = scale*(sgray - gmin) + 0.5;
		dbase[i0*dd0 + i1] = dgray;
	    }
    }


    return;
}



/*------------------------------------------------------------------------
vxl_init_raw_format - Initialize a raw format struct with default values

This function initializes a raw format struct with the default values
for each field.  This makes it easier to add new fields later without
having to update every user of this structure (assuming, of course,
that all users are wise enough to use this function rather than doing
it themselves).  The default rank is 0 for an empty array, so you must
at least specify the rank and dimensions to get a sensible display.
------------------------------------------------------------------------*/

void 
vxl_init_raw_format(
    struct vxl_raw_format *fmt)    /* Format struct to be initialized */
{
    int i;

    if (fmt == 0)
	fatal("vxl_init_raw_format: Null pointer to struct to be initialized");

    fmt->magic  = VXL_RAWFORM;
    fmt->error  = 0;
    fmt->type   = INTP_UCHAR;
    fmt->bigend = 1;
    fmt->rank   = 0;
    for (i = 0; i < VXL_MAX_RANK; i++) {
	fmt->dimen[i]   = 0;
	fmt->current[i] = 0;
	fmt->direct[i]  = 0; }
    for (i = 0; i < VXL_MAX_ECHO; i++) {
	fmt->minint[i]   =   0;
	fmt->maxint[i]   = 255;
	fmt->colormap[i] =   0; }
    fmt->xdim  = 0;
    fmt->ydim  = 1;
    fmt->edim  = 2;
    fmt->recho = 0;
    fmt->gecho = 0;
    fmt->becho = 0;

    return;
}


/*-------------------------------------------------------------------------
vxl_raw_to_rgb2d - Convert external (raw) image format to 2d rgb section

This function takes image data in its external (raw) format and
extracts the red, green, and blue channels of an unsigned char 2d
section, using intensity windowing to adjust the intensity range of
the raw data to the intensity range of the display.  The primary
purpose of this function is to make it easy to write an image display
program that does not have to commit to an image data format before it
reads the data into memory but that can let the user interactively try
different data formats while trying to determine the real format of an
image data file abandoned on his doorstep.


FIXME:  This code makes some probably non-portable assumptions about
the results of converting various pointers to unsigned char pointers,
doing arithmetic on those pointers, and converting back to the original
pointer types.  Unfortunately, I don't currently see any portable way
of doing the right thing without going all the way around the barn.

FIXME:  Provide color lookup tables for each channel.  Allow reversal
of specified coordinate directions.
-------------------------------------------------------------------------*/

void
vxl_raw_to_rgb2d(
    voxel_array *dst,               /* Output image, as rgb2d data */
    voxel_array *src,               /* Input  image, as raw data */
    struct vxl_raw_format *fmt)     /* Descriptor of external format */
{
    long xdim, ydim, edim;          /* Coordinates for x, y, echo */
    long sbuff[VXL_MAX_RANK];       /* Working copy of dimens/slice coords */
    int st;                         /* Type of the source data */
    int ft = INTP_FLOAT;            /* Type of the intermediate data */
    unsigned char *sdata;           /* Location of the raw (source) data */
    int snbytes;                    /* Number of bytes per source datum */
    float *fbuff;                   /* Buffer for intermediate data */
    int   bigend;                   /* Is source data bigendian? */
    unsigned char   *dst0;          /* Start of a destination scanline */
    unsigned char   *rsrc0 = 0,  
        *gsrc0 = 0, *bsrc0 = 0;     /* Start of a source scanline, for RGB */
    double ro, go, bo;              /* Intensity offsets for RGB */
    double rm, gm, bm;              /* Intensity multipliers for RGB */
    int    rerr=0, gerr=0, berr=0;  /* Error flags */
    long   nrows, ncols;            /* Size of the section */
    long   ssx, ssy;                /* Source strides in dest's x and y */
    long   dsx, dsy;                /* Destination strides in x and y */
    long   i;

    /* Verify the arguments */
    if (dst == 0 || dst->magic != VXL_MAGIC)
	fatal("vxl_raw_to_rgb2d: Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("vxl_raw_to_rgb2d: Invalid or empty source array");
    if (fmt == 0 || fmt->magic != VXL_RAWFORM)
	fatal("vxl_raw_to_rgb2d: Missing or invalid raw format specifier");
    if (fmt->rank < 2)
	fatal("vxl_raw_to_rgb2d: Assumed rank of source is less than 2");

    /* Get and verify the transformation parameters */
    bigend = fmt->bigend;
    st   = fmt->type;
    xdim = fmt->xdim;
    ydim = fmt->ydim;
    edim = fmt->edim;
    if (xdim < 0 || xdim >= fmt->rank)
	fatal("vxl_raw_to_rgb2d: Invalid choice of destination X coordinate");
    if (ydim < 0 || ydim >= fmt->rank)
	fatal("vxl_raw_to_rgb2d: Invalid choice of destination Y coordinate");
    if (edim < 0 || edim > fmt->rank)  /* NOTE: edim MAY equal rank */
	fatal("vxl_raw_to_rgb2d: Invalid choice of destination echo coord");

    /* FIXME:  Should probably also check here that the type and dimensions
       are reasonable and that xdim, ydim, and edim are distinct. */

    /* Set up the destination voxel array */
    nrows = sbuff[0] = fmt->dimen[ydim];
    ncols = sbuff[1] = fmt->dimen[xdim];
    sbuff[2] = 3;   /* RGB is 3 channels */
    vxl_alloc_array(dst, INTP_UCHAR, 3, sbuff);
    dst0 = dst->data;
    dsx = 3;
    dsy = 3*ncols;

    /* Set up an intermediate float buffer */
    fbuff = mallock(3*ncols*sizeof(float));

    /* Set up indexing for the source array */
    snbytes = exim_sizeof_type(fmt->type);
    ssx = snbytes * vxli_step2(fmt->rank, fmt->dimen, xdim);
    ssy = snbytes * vxli_step2(fmt->rank, fmt->dimen, ydim);
    sdata = src->data;
    for (i = 0; i < VXL_MAX_RANK; i++)
	sbuff[i] = fmt->current[i];
    sbuff[xdim] = 0;
    sbuff[ydim] = 0;
    if (0 <= edim && edim < fmt->rank)  {                   
	/* Echo is an explicit dimension */
	sbuff[edim] = fmt->recho;
	rsrc0 = sdata + snbytes * 
	    vxli_offset2(fmt->rank, fmt->dimen, sbuff, 1);
	sbuff[edim] = fmt->gecho;
	gsrc0 = sdata + snbytes * 
	    vxli_offset2(fmt->rank, fmt->dimen, sbuff, 1);
	sbuff[edim] = fmt->becho;
	bsrc0 = sdata + snbytes * 
	    vxli_offset2(fmt->rank, fmt->dimen, sbuff, 1);
    }
    else {
	/* Echo is an implicit dimension */ 
	rsrc0 = gsrc0 = bsrc0 = sdata + snbytes * 
	    vxli_offset2(fmt->rank, fmt->dimen, sbuff, 1);
    }

    /* DEBUG:  Print out some possibly useful values */
    if (0) {
	fprintf(stderr, "rank = %ld,  xdim = %ld  ydim = %ld  edim = %ld\n",
		fmt->rank, xdim, ydim, edim);
	fprintf(stderr, "dimen =  %ld, %ld, %ld, %ld\n",
		fmt->dimen[0], fmt->dimen[1], fmt->dimen[2], fmt->dimen[3]);
	fprintf(stderr, "sbuff = %ld, %ld, %ld, %ld\n",
		sbuff[0], sbuff[1], sbuff[2], sbuff[3]);
	fprintf(stderr, "nbytes = %d ssx = %ld  ssy = %ld  rsrc0 = %d\n", 
		snbytes, ssx, ssy, rsrc0 - (unsigned char *)(sdata));
    }

    /* Set up intensity windowing for each color channel */
    ro = fmt->minint[fmt->recho];
    go = fmt->minint[fmt->gecho];
    bo = fmt->minint[fmt->becho];
    rm = 255.0 / (fmt->maxint[fmt->recho] - fmt->minint[fmt->recho]);
    gm = 255.0 / (fmt->maxint[fmt->gecho] - fmt->minint[fmt->gecho]);
    bm = 255.0 / (fmt->maxint[fmt->becho] - fmt->minint[fmt->becho]);
    
    /* Loop over scanlines in the section */
    for (i = 0; i < nrows; i++) {

	/* Convert each (rgb) channel from raw to float */
	rerr = exim_import(ncols, fbuff,   ft, 3, rsrc0, st, ssx, bigend);
	gerr = exim_import(ncols, fbuff+1, ft, 3, gsrc0, st, ssx, bigend);
	berr = exim_import(ncols, fbuff+2, ft, 3, bsrc0, st, ssx, bigend);
	if (rerr || gerr || berr)
	    fatal("vxl_raw_to_rgb2d: Can't import image data");

	/* Scale each channel from float to unsigned char */
	rerr = bips_scale_to_uchar(ncols, dst0  , dsx, ft, fbuff,   3, ro, rm);
	gerr = bips_scale_to_uchar(ncols, dst0+1, dsx, ft, fbuff+1, 3, go, gm);
	berr = bips_scale_to_uchar(ncols, dst0+2, dsx, ft, fbuff+2, 3, bo, bm);
	if (rerr || gerr || berr)
	    fatal("vxl_raw_to_rgb2d: Can't scale data to uchar range");

	/* Find beginning of next source and destination scanline */
	dst0  += dsy;
	rsrc0 += ssy;
	gsrc0 += ssy;
	bsrc0 += ssy;
    }

    /* Free the intermediate buffer */
    free(fbuff);

    return;
}



/*-------------------------------------------------------------------------
vxl_write_ppm_image - Write a rgb2d image as a ppm file

This function takes an rgb2d image as input and writes it to a file as
raw bits Portable Pixel Map (ppm) image.  The image must be a
two-dimensional rgb unsigned char image.
-------------------------------------------------------------------------*/

void
vxl_write_ppm_image(
    const char *fname,          /* Name of the new ppm file to write */
    voxel_array *src)           /* The voxel image to write */
{
    FILE *file;                 /* Stream to write the ppm to */
    long width, height;         /* Dimensions of voxel array */
    unsigned char *data;        /* Data array in voxel array */
    size_t count;
    long i;

    /* Verify image properties */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	fatal("Invalid or empty source array");
    if (src->type != INTP_UCHAR)
	fatal("Only unsigned char can be written to a ppm file");
    if (src->rank != 3 || src->dimen[2] != 3)
	fatal("Only a 2d rgb image can be written to a ppm file");

    /* Open the output file */
    file = fopen(fname, "wb");
    if (file == 0)
	fatal("Unable to open ppm output file");

    /* Write the header */
    width  = src->dimen[1];
    height = src->dimen[0];
    fprintf(file, "P6 %ld %ld 255\n", width, height);

    /* Write the pixels themselves; each pixel is three bytes */
    data = src->data;
    for (i = 0; i < height; i++) {
	count = fwrite(data + i*3*width, 1, 3*width, file);
	if ( count < 3*width )
	    fatal("I/O error while writing ppm file");
    }

    fclose(file);
    return;
}



/***************************************************************************
                     Image Scanning Operations

Image scanning operations scan an image in some data-dependent or
complex fashion and may generate either a new image or scalar results.
Some examples are object extraction, finding the bounding box, and
distance transforms.
***************************************************************************/


/*-------------------------------------------------------------------------
vxl_find_objects_uchar3d - Extract the objects in a volume image

This function scans a 3d unsigned char volume image and separates it
into objects (which are defined as maximal connected regions of
non-zero voxels).  It returns the number of objects found (including
the background as object 0); it also sets a pointer to a table
containing information on the objects found.  The algorithm used here
sometimes assigns two or more object ids to what is later found to be
a single object; rather than go back and change the old idents (which
would require an additional and perhaps unnecessary pass over the
image), it simply chooses the smaller id as canonical, and indicates
in the object table the canonical id of the redundant ids.

The algorithm used here is a reference implementation which is
intended to be (relatively!) easy to understand and verify rather than
fast; it will eventually be replaced by a faster algorithm which can
then be tested against this reference implementation.  This algorithm
is to scan the image in raster order, considering each voxel in turn.
If the voxel is zero, then it is always assigned to object 0, which is
background.  Otherwise the 13 neighbors of that voxel which preceded
it in the scan are examined.  If all the neighbors are zero, then a
new object id is allocated and assigned to the current voxel.  If all
the non-zero neighbors have been assigned the same object id, then the
current voxel is assigned that same id.  If the non-zero neighbors
have been assigned two or more distinct object ids, then the smaller
of one of those object ids is taken as canonical and the others are
consolidated into it; then the current voxel is assigned the canonical
id.  The object table is initially allocated to contain a small number
of objects and reallocated when necessary.  The count of voxels in
each object is updated as each voxel is processed; the counts of the
non-canonical objects are merged into the canonical counts in a pass
over the object table once the image has been completely scanned.

FIXME: The handling of voxels at the boundary of the image is probably
correct but rather inefficient.  How do I handle voxels at the
boundary of the image without excessively complicating the algorithm
or losing performance?  One possible approach is to move the loop over
the neighbors outside the innermost loop over the voxels in a
scanline; then testing for the boundary of the image can be done once
per scanline rather than once per voxel.  A more radical approach (but
not suitable for a reference implementation) might be to focus on
object boundaries as the places where the output object id must
change; the object boundaries can probably be found much faster.

FIXME:  Can this be generalized to arbitrary dimension and structuring
elements using the same methods used for convolution et al?
-------------------------------------------------------------------------*/

int                              /* Number of objects found, or 0 on error */
vxl_find_objects_uchar3d(
    voxel_array *dest,           /* Volume image containing object labels */
    struct voxel_object_data     /* Table of objects found */
        *oblist[],
    voxel_array *src)            /* Volume image to extract objects from */
{
    unsigned char  *sd;          /* Actual data array for source */
    unsigned short *dd;          /* Actual data array for destination */
    long nz, ny, nx;             /* Number of slices, rows, columns */
    long iz, iy, ix;             /* Indices over slices, rows, columns */
    long dz, dy, dx;             /* Stride in slices, rows, columns */
    struct voxel_object_data *obl;     /* Current object list */
    int lobl;                    /* Allocated length of object table */
    int nobs;                    /* Number of objects found so far */
    long dn[13];                 /* Offsets to neighbors of current voxel */
    int in;                      /* Index over neighbors */
    int nbr;                     /* Object id of a neighbor */
    int nbr1, nbr2;              /* First and second ob ids in a neghborhood */
    int can1, can2;              /* Canonical ids for nbr1 and nbr2 */
    long nbytes;                 /* Number of bytes needed for object list */
    long total;                  /* Total number of voxels in all objects */
    int n, nn;

    /* Binary code table indicating which neighbors to take when
       iz == 0 (4), iy == 0 (2), or ix == 0 (1) */
    int take[13] = {0, 1, 1, 2, 3, 3, 2, 3, 3, 4, 5, 5, 6};

    /* Check validity of source array; get the dimensions. */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 || 
	src->type != INTP_UCHAR || src->rank != 3)
	fatal("Invalid or empty source array");
    sd = src->data;
    nz = src->dimen[0];
    ny = src->dimen[1];
    nx = src->dimen[2];

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	fatal("Invalid destination array");
    vxl_alloc_array(dest, INTP_USHORT, 3, &(src->dimen[0]));
    dd = dest->data;

    /* Allocate initial object table; define the background object */
    nobs =   1;
    lobl = 256;
    obl  = mallock(lobl*sizeof(struct voxel_object_data));
    obl[0].ident = 0;
    obl[0].canon = 0;
    obl[0].count = 0;

    /* Compute step sizes in each dimension */
    dz = vxli_step(src, 0);
    dy = vxli_step(src, 1);
    dx = vxli_step(src, 2);

    /* Compute offsets to neighbors */
    dn[ 0] = -dz-dy-dx;
    dn[ 1] = -dz-dy;
    dn[ 2] = -dz-dy+dx;
    dn[ 3] = -dz   -dx;
    dn[ 4] = -dz;
    dn[ 5] = -dz   +dx;
    dn[ 6] = -dz+dy-dx;
    dn[ 7] = -dz+dy;
    dn[ 8] = -dz+dy+dx;
    dn[ 9] =    -dy-dx;
    dn[10] =    -dy;
    dn[11] =    -dy+dx;
    dn[12] =       -dx;

    /* Loop over the voxels in the image */
    /* FIXME:  Probably faster to loop over neighbors outside columns */
    for (iz = 0; iz < nz; iz++)
	for (iy = 0; iy < ny; iy++)
	    for (ix = 0; ix < nx; ix++) {

		nbr1 = 0;

		/* If voxel is non-zero, consider as possible object */
		if (*sd != 0) {


		    /* Loop over the neighbors (omitting those that
		       fall outside the edges of the image) and record
		       the object id in nbr1.  If two different object
		       ids are seen in the neighborhood, mark them as
		       the same in the object list, taking the smaller
		       id as canonical. */
		    for (in = 0; in < 13; in++)
			if ( ( iz > 0 || take[in] & 4 ) &&
			     ( iy > 0 || take[in] & 2 ) && 
			     ( ix > 0 || take[in] & 1 ) &&
			     ((nbr = dd[dn[in]]) != 0 ) ) {

			    /* Check for broken algorithm */
			    if (nbr < 1 || nbr >= nobs)
				panic("Error in object finding algorithm");

			    /* Case 1: The first non-zero neighbor
			     need only be recorded. */
			    if ( nbr1 == 0 ) {
				nbr1 = nbr;
			    }
			    
			    /* Case 2: There are two apparently
			       distinct neighbor ids.  Trace down the
			       canonical ids.  If they are different,
			       make the smaller id canonical. In any
			       case, record the canonical ids in the
			       actual ids */
			    else if ( nbr != nbr1 ) {

				for ( n = 0, can1 = nbr1;
				      n < nobs && can1 != obl[can1].canon;
				      n++, can1 = obl[can1].canon ) /* NIL */;
				if (n >= nobs)
				    panic("Infinite loop in object finder");

				for ( n = 0, can2 = nbr2 = nbr;
				      n < nobs && can2 != obl[can2].canon;
				      n++, can2 = obl[can2].canon ) /* NIL */;
				if (n >= nobs)
				    panic("Infinite loop in object finder");

				if (can2 < can1) {
				    obl[can1].canon = can2;
				    obl[nbr1].canon = can2;
				    obl[nbr2].canon = can2; }
				else if (can1 < can2) {
				    obl[can2].canon = can1;
				    obl[nbr1].canon = can1;
				    obl[nbr2].canon = can1; }
				else {
				    obl[nbr1].canon = can1;
				    obl[nbr2].canon = can2; }
			    }

			}

		    /* If all neighbors are zero, assign new object number,
		       enlarging the object list if necessary. */
		    if (nbr1 == 0) {
			if (nobs >= USHRT_MAX)
			    panic("Found too many objects for algorithm\n");
			nbr1 = nobs++;
			if (0) 
			    fprintf(stderr, "Creating object #%d\n", nbr1);
			if (nobs > lobl) {
			    lobl *= 2;
			    nbytes = lobl*sizeof(struct voxel_object_data);
			    obl = reallock(obl, nbytes);
			}
			obl[nbr1].ident  = nbr1;
			obl[nbr1].canon  = nbr1;
			obl[nbr1].count  = 0;
			obl[nbr1].point[0] = iz;
			obl[nbr1].point[1] = iy;
			obl[nbr1].point[2] = ix;
		    }

		}

		/* Capture the coordinates of the first zero voxel seen */
		else if ( *sd == 0 && obl[0].count == 0 ) {
		    obl[0].point[2] = ix;
		    obl[0].point[1] = iy;
		    obl[0].point[0] = iz;
		}

		/* Set label in destination and update voxel count */
		obl[nbr1].count += 1;
		*dd = nbr1;
		
		/* Step to the next voxel in source and destination */
		sd++;
		dd++;
	    }

    /* Clean up the object list */
    for (n = 0; n < nobs; n++) {

	/* Set each canon field to its ultimate ancestor */
	for ( nn = 0, can1 = n;
	      nn < nobs && can1 != obl[can1].canon;
	      nn++, can1 = obl[can1].canon ) /* NIL */;
	if (nn >= nobs)
	    panic("Infinite loop in object list");
	obl[n].canon = can1;

	/* Add counts to the canonical object; zero non-canonical counts */
	if (obl[n].canon != n) {
	    can1 = obl[n].canon;
	    obl[can1].count += obl[n].count;
	    obl[n].count = 0;
	}
    }

    /* Do some inexpensive consistency checks: The ident field is
       equal to the index; every canon field is less than or equal to
       the ident field; the sum of the count fields is equal to the
       total number of voxels in the image. */
    total = 0;
    for (n = 0; n < nobs; n++) {
	if (n != obl[n].ident)
	    panic("Invalid ident field in object finder");
	if (obl[n].canon > obl[n].ident)
	    panic("Invalid canon field in object finder");
	if (n > 0 && obl[n].canon == n && obl[n].count <= 0)
	    panic("Zero count for a (non-background) canonical object");
	if (obl[n].canon != n && obl[n].count != 0)
	    panic("Positive count for non-canonical object");
	total += obl[n].count;
    }
    if (total != nz*ny*nx)
	panic("Objects do not account for all voxels");

    /* Trim the object table before returning to caller */
    if (nobs < lobl) {
	nbytes = nobs*sizeof(struct voxel_object_data);
	obl = reallock(obl, nbytes);
    }
    *oblist = obl;

    return nobs;
}



/*--------------------------------------------------------------------------
vxl_chamfer - Compute chamfer distance image

This function computes the chamfer distance transform of a given
image.  The chamfer distance image is zero wherever the original image
is non-zero and is otherwise equal to the taxicab (L1) distance to the
nearest non-zero point in the original image.  The type of the output
image may be specified by the caller to either unsigned char or
unsigned short, or allowed to default to whichever type is sufficient
to contain the maximum possible distance.

FIXME: Verify that this works for dimensions that are 0 or 1.
----------------------------------------------------------------------------*/

void
vxl_chamfer(
    voxel_array *dest,           /* Destination image */
    voxel_array *src,            /* Source image */
    int dtype)                   /* Desired destination type, or zero */
{
    int rank;                    /* Rank of the src and dest arrays */
    int stype;                   /* Type of the source array */
    long *dimen;                 /* Dimensions of the two arrays */
    long coord[VXL_MAX_RANK];    /* Coordinates of current row */
    int idim;                    /* Current scan dimension */
    long nelem;                  /* Number of elements in row for last dimen */
    long nelem1;                 /* Number of elements, minus 1 */
    unsigned char *drow;         /* Start of a row in destination */
    unsigned char *srow;         /* Start of a row in source */
    long doff;                   /* Destination offset to previous voxel */
    long md;                     /* Maximum possible taxicab distance */
    int flag;

    /* Verify the source image and get its properties */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 ||
	 src->rank < 1 )
	fatal("Source array is invalid");
    rank   = src->rank;
    stype  = src->type;
    dimen  = src->dimen;
    nelem  = dimen[rank-1];
    nelem1 = nelem - 1;

    /* Chose default destination type, if necessary */
    if (dtype == 0) {
        md = 0;
	for (idim = 0; idim < rank; idim++)
	    md += dimen[idim];
	if (md > UCHAR_MAX)
	    dtype = INTP_USHORT;
	else
	    dtype = INTP_UCHAR;
    }
    else if (dtype != INTP_USHORT && dtype != INTP_UCHAR)
	fatal("Invalid destination type in chamfer transform");

    /* Allocate the destination array */
    vxl_alloc_array(dest, dtype, rank, dimen);

    /* Initialize destination to largest possible distance and scan
       innermost dimension in positive direction */
    doff = -1;
    vxli_firstrow(rank, coord);
    do {

        /* Initialize the destination row, starting with the first voxel */
        coord[rank-1] = 0;
        drow = vxli_locate(dest, coord, 1);
        flag = bips_fill_max(nelem, dtype, drow, 1);
	if (flag)
	    fatal("Error calling BIPS");

	/* Setup and execute the chamfer inner loop, starting with the
           second voxel in the row */
        coord[rank-1] = 1;
	drow = vxli_locate(dest, coord, 1);
	srow = vxli_locate(src,  coord, 1);
	flag = bips_chamfer(nelem1, dtype, drow, 1, doff, stype, srow, 1);
	if (flag)
	    fatal("Error calling BIPS");

    } while ( vxli_nextrow(rank, dimen, coord) );

    /* Scan outer dimensions in positive direction by taking rows in
       forward order */
    for (idim = rank-2; idim >= 0; idim--) {

	/* Loop over the rows along the current direction */
	doff = -vxli_step(dest, idim);
	vxli_firstrow(rank, coord);
	do {

	    /* Skip first row in outer dimensions */
	    if ( coord[idim] == 0 )
		continue;

	    /* Locate starting pixel in the current row */
	    drow  = vxli_locate(dest, coord, 1);
            srow  = vxli_locate(src,  coord, 1);

	    /* Execute the chamfer inner loop */
	    flag = bips_chamfer(nelem, dtype, drow, 1, doff, stype, srow, 1);
	    if (flag)
		fatal("Error calling BIPS");

	} while ( vxli_nextrow(rank, dimen, coord) );

    }

    /* Scan innermost dimension in negative direction */
    doff = 1;
    vxli_lastrow(rank, dimen, coord);
    do {

	/* Locate starting pixel at the end of the current row */
        coord[rank-1] = dimen[rank-1] - 2;
	drow = vxli_locate(dest, coord, 1);
	srow = vxli_locate(src,  coord, 1);

	/* Execute the chamfer inner loop */
	flag = bips_chamfer(nelem1, dtype, drow, -1, doff, stype, srow, -1);
	if (flag)
	    fatal("Error calling BIPS");

    } while ( vxli_prevrow(rank, dimen, coord) );

    /* Scan outer dimensions in negative direction by taking rows in
       reverse order */
    for (idim = rank-2; idim >= 0; idim--) {

	/* Initialize dimension-specific parameters */
	doff  = vxli_step(dest, idim);

	/* Loop over the rows along the current direction */
	vxli_lastrow(rank, dimen, coord);
	do {

	    /* Skip last row in outer dimensions */
	    if ( coord[idim] == dimen[idim] - 1 )
		continue;

	    /* Locate starting pixel in the current row */
	    coord[rank-1] = 0;
	    drow  = vxli_locate(dest, coord, 1);
            srow  = vxli_locate(src,  coord, 1);

	    /* Execute the chamfer inner loop */
	    flag = bips_chamfer(nelem, dtype, drow, 1, doff, stype, srow, 1);
	    if (flag)
		fatal("Error calling BIPS");

	} while ( vxli_prevrow(rank, dimen, coord) );
    }

    return;
}
