/****************************************************************************
vxli.c - Callable internal functions for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  It normally contains various functions for internal use
that are called rather than inlined; on a compiler that does not
support inlining, it also contains callable versions of the functions
that are normally inlined.

FIXME:  Identify common validity checking steps and isolate as functions
or macros.  Q:  What are the efficiency tradeoffs between the two?
*****************************************************************************/

static const char rcsid[] = "$Id: vxli.c,v 1.1 2001/10/10 21:02:08 hughett Exp $ Release $Name:  $";

#include "config.h"
#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"



/*---------------------------------------------------------------------------
Compile inlineable functions as callable functions on unsupportive compilers
----------------------------------------------------------------------------*/

#ifndef HAVE_INLINE
#include "vxli2.c"
#endif



/*------------------------------------------------------------------------
vxli_count - Compute number of voxels in a voxel array

Given an voxel array, this function computes the total number of voxel
elements that it contains.  Little validity checking is done, so this
version is suitable only for use within the voxel kit itself.
--------------------------------------------------------------------------*/

long
vxli_count(
    voxel_array *src)        /* Voxel array to define dimensions */
{
    long count;              /* Count of elements */
    int i;
    
    /* Check for valid voxel array */
    if (src == 0 || src->magic != VXL_MAGIC)
	fatal("vxli_count: Invalid voxel array.");

    /* Compute the desired count */
    count = 1;
    for (i = src->rank - 1; i >= 0; i--) 
	count *= src->dimen[i];

    return count;
}



/*------------------------------------------------------------------------
vxli_delta - Compute element stride for given subscript deltas

Given an voxel array, and subscript deltas for each coordinate, this
function computes the distance in the array which corresponds to the
given change in the coordinates, measured in elements (not bytes).
The length of the coords array must equal the rank of the voxel array.
Little validity checking is done, so this version is suitable only for
use within the Voxel Kit itself.

The related routine vxli_offset computes the location of an element
relative to the beginning of the array, AND verifies that the given
coordinates are within the array.  The routine vxli_step and its friends
computes the delta for changing a single coordinate.
--------------------------------------------------------------------------*/

long
vxli_delta(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate deltas */
{
    long step;               /* Step size for current coordinate */
    long delta;              /* Delta computed so far */
    int i;
    
    /* Check for valid voxel array */
    if (src == 0 || src->magic != VXL_MAGIC)
	fatal("vxli_delta: Invalid voxel array.");

    /* Compute the desired delta */
    delta = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	delta += coords[i]*step;
	step *= src->dimen[i];
    }

    return delta;
}



/*------------------------------------------------------------------------
vxli_offset - Compute element number for given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes the offset or element number selected by the given
coordinate values, measured in elements (not bytes).  The length of
the coords array must equal the rank of the voxel array.  If the given
coordinates are outside the voxel array, then a fatal error is
reported if the flag errifout is set; otherwise a negative offset is
returned.  Little validity checking is done, so this version is
suitable only for use within the voxel kit itself.
--------------------------------------------------------------------------*/

long
vxli_offset(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords,            /* Coordinate values */
    int errifout)            /* What to do if out of range */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    int i;
    
    /* Check for valid voxel array */
    if (src == 0 || src->magic != VXL_MAGIC)
	fatal("vxli_offset: Invalid voxel array.");

    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= src->dimen[i]) {
	    if (errifout)
		fatal("vxli_offset: Voxel array subscript out of range");
	    else
		return -1;
	}
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    return offset;
}



/*------------------------------------------------------------------------
vxli_offset2 - Compute element number for given subscripts

Given the rank, dimensions, and subscript values for each coordinate,
this variant of the vxli_offset function computes the offset or
element number selected by the given coordinate values, measured in
elements (not bytes).  The length of the coords array must equal the
rank of the voxel array.  If the given coordinates are outside the
voxel array, then a fatal error is reported if the flag errifout is
set; otherwise a negative offset is returned.  Little validity
checking is done, so this version is suitable only for use within the
voxel kit itself.
--------------------------------------------------------------------------*/

long
vxli_offset2(
    long rank,               /* Number of dimensions */
    long *dimen,             /* Number of elements along each dimension */
    long *coords,            /* Coordinate values for each dimension */
    int errifout)            /* What to do if out of range */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    long i;
    
    /* Check for valid arguments */
    if (rank < 0 || dimen == NULL || coords == NULL)
	fatal("vxli_offset2: Invalid rank, dimen, or coords");

    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= dimen[i]) {
	    if (errifout)
		fatal("vxli_offset2: Voxel array subscript out of range");
	    else
		return -1;
	}
	offset += coords[i]*step;
	step *= dimen[i];
    }

    return offset;
}



/*------------------------------------------------------------------------
vxli_locate - Compute pointer to element with given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes an unsigned char pointer to the element selected by
the given coordinate values.  The length of the coords array must
equal the rank of the voxel array.  If the given coordinates are
outside the voxel array, then a fatal error is reported if the flag
errifout is set; otherwise a null pointer is returned.  Little
validity checking is done, so this version is suitable only for use
within the voxel kit itself.  In fact, the most common use of this
function is to compute the base address of a scanline or slice to hand
off to a bips function.

FIXME: Do two versions of this function; one with checking and one
without, for applications where the coordinates have already been tested.
Ditto for vxli_offset.  Maybe also inline the unchecked version.

FIXME:  This function computes the desired point as an unsigned char
pointer, using the element byte size given in the header.  It is not
entirely clear that this is truly portable.  It is also likely to
break if someone forgets to set the byte size; perhaps we should get
the byte size from the type instead?
--------------------------------------------------------------------------*/

unsigned char *
vxli_locate(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords,            /* Coordinate values */
    int errifout)            /* What to do if out of range */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    unsigned char *point;    /* Pointer to the desired element */
    int i;
    
    /* Check for valid voxel array */
    if (src == 0 || src->magic != VXL_MAGIC)
	fatal("vxli_locate: Invalid voxel array.");

    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= src->dimen[i]) {
	    if (errifout)
		fatal("vxli_locate: Voxel array subscript out of range");
	    else
		return NULL;
	}
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    point = (unsigned char *)src->data + (src->nbytes)*offset;
    return point;
}



/*-------------------------------------------------------------------------
vxli_step - Compute step size in a given dimension

Given a voxel array, this function computes the step size or stride
that corresponds to an increment of 1 in a given dimension of the
array, measured in elements (not bytes).  Only the most basic error
checking is done.

FIXME:  Perhaps check at least that the dimensions are positive?  Or is
it better to do that once when the whole array is checked for validity?
--------------------------------------------------------------------------*/

long
vxli_step(
    voxel_array *src,       /* Voxel array to compute for */
    int dim)                /* Dimension to compute for (0-base) */
{
    long step = 1;          /* Desired step size, so far */
    int i;

    /* Basic validity check */
    if (src == 0 || src->magic != VXL_MAGIC)
	fatal("vxli_step: Invalid voxel array as argument");

    /* Compute the desired step size */
    for (i = src->rank - 1; i > dim; i--)
	step *= src->dimen[i];

    return step;
}



/*-------------------------------------------------------------------------
vxli_step2 - Compute step size in a given dimension

Given the rank and dimensions of a voxel array, this variant of the
vxli_step function computes the step size or stride that corresponds
to an increment of 1 in a given dimension of the array, measured in
elements (not bytes).  Only the most basic error checking is done.

FIXME:  Perhaps check at least that the dimensions are positive?  Or is
it better to do that once when the whole array is checked for validity?
--------------------------------------------------------------------------*/

long
vxli_step2(
    long rank,              /* Number of dimensions */
    long *dimen,            /* Number of elements along each dimension */
    int dim)                /* Dimension to compute for (0-base) */
{
    long step = 1;          /* Desired step size, so far */
    int i;

    /* Basic validity check */
    if (rank < 0 || dimen == NULL || dim < 0 || dim >= rank)
	fatal("vxli_step2: Invalid arguments");

    /* Compute the desired step size */
    for (i = rank - 1; i > dim; i--)
	step *= dimen[i];

    return step;
}



/*-------------------------------------------------------------------------
vxli_same_shape - Check whether rank and dimensions match

Given two voxel arrays, this function returns 1 if they have the same
rank and dimensions; it returns zero otherwise.  The type is not checked,
and it is assumed that the arrays have already be checked for existence
and validity; it is assumed that other necessary checking will be done
at a higher level, so that this routine is not suitable to be called
directly by the Voxel Kit user.
------------------------------------------------------------------------*/

int
vxli_same_shape(
    voxel_array *src1,           /* First array */
    voxel_array *src2)           /* Second array */
{
    int i;

    /* If rank or dimen doesn't match, return 0 */
    if (src1->rank != src2->rank) return 0;
    for (i = 0; i < src1->rank; i++)
	if (src1->dimen[i] != src2->dimen[i])
	    return 0;

    /* Otherwise return 1 */
    return 1;
}
