/****************************************************************************
vxli2.c - Inlineable internal functions for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 2001 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  It contains various functions for internal use that
will be inlined in the compiler supports that keyword; to this end, it
will be either included in the using program via the vxli.h header or
included into the file vxli.c containing the compiled functions.

FIXME: Identify common validity checking steps and isolate as
(inline?) functions or macros.  Q: What are the efficiency tradeoffs
between the two?
*****************************************************************************/

static const char rcsid2[] = "$Id: vxli2.c,v 1.2 2001/11/13 18:56:28 hughett Exp $ Release $Name:  $";

/* Assumes that config.h and vxli.h have already been included */



/*------------------------------------------------------------------------
vxli_offset_nocheck - Compute element number for given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes the offset or element number selected by the given
coordinate values, measured in elements (not bytes).  The length of
the coords array must equal the rank of the voxel array.  If the given
coordinates are outside the voxel array, the results are unpredictable,
up to and including a segmentation fault; it is assumed that the
source array and coordinates have already been checked for validity.
--------------------------------------------------------------------------*/

inlineable long
vxli_offset_nocheck(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate values */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    int i;
    
    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    return offset;
}



/*------------------------------------------------------------------------
vxli_firstvoxel - Initialize iteration over voxels of an image

This function initializes a coordinate array prior to iterating over
the voxels of an image with vxli_nextvoxel; voxels are selected in
memory order, with the last coordinate varying most rapidly.  Only
elementary error checking is done in vxli_firstvoxel and
vxli_nextvoxel; these functions should be called only in contexts
where the array has already been checked for validity.  Note further
that in most applications, it is better to iterate over rows and use
the optimized BIPS functions to process voxels within a row; this
function is intended primarily for reference implementations and
similar contexts where clarity is more important than speed.

Assuming that the special case of an empty array has been excluded, 
the canonical loop over voxels looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of first voxel in row

vxli_firstvoxel(array->rank, coord);
do {
    ...operate on a voxel...
} 
while (vxli_nextvoxel(array->rank, array->dimen, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_firstvoxel(
    int rank,                   /* Rank of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = 0;
    return;
}



/*-------------------------------------------------------------------------
vxli_nextvoxel - Compute coordinates of next row in an image

Given a voxel array and a coordinate array pointing to some particular
voxel in an image, this function updates the coordinate array to point
to the next voxel.  This function returns zero when all rows have been
seen.  See vxli_firstvoxel for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_nextvoxel(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-1; i >= 0; i--) {
	if ( ++coord[i] < dimen[i] )
	    return 1;
	else
	    coord[i] = 0;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_firstrow - Initialize iteration over rows of an image

This function initializes a coordinate array prior to iterating over
the rows of an image with vxli_nextrow, where a row is defined here as
a sequence of voxels all having their first (N-1) coordinates the same
and differing only in their last coordinate.  The rows are guaranteed
to be generated in forward order, with coordinate values increasing;
the index of the last dimension is always set to zero.  Only elementary
error checking is done in vxli_firstrow and vxli_nextrow; these
functions should be called only in contexts where the array has
already been checked for validity.

Iterating over the rows of an image is an important operation in the
Voxel Kit, because most loop nests are organized with the last
dimension as the innermost loop and the outer dimensions as the outer
loops, with other loops put between.  See the implementation of
convolution as an example.  

Assuming that the special case of an empty array has been excluded, 
the canonical loop over rows looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of first voxel in row

vxli_firstrow(array->rank, coord);
do {
    ...operate on a row...
} 
while (vxli_nextrow(array->rank, array->dimen, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_firstrow(
    int rank,                   /* Rank of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = 0;
    return;
}



/*-------------------------------------------------------------------------
vxli_nextrow - Compute coordinates of next row in an image

Given a voxel array and a coordinate array pointing to some particular
row in an image, this function updates the coordinate array to point
to the next row.  This function returns zero when all rows
have been seen.  See vxli_firstrow for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_nextrow(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-2; i >= 0; i--) {
	if ( ++coord[i] < dimen[i] )
	    return 1;
	else
	    coord[i] = 0;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_lastrow - Initialize reversed iteration over rows of an image

This function initializes a coordinate array prior to iterating over
the rows of an image in reverse order using vxli_prevrow.  A row is
defined here as a sequence of voxels all having their first (N-1)
coordinates the same and differing only in their last coordinate.  It
is similar to vxli_firstrow and vxli_nextrow but generates the rows in
reverse order; that is, with coordinate values decreasing.  The index
of the last dimension is set to the last pixel in that row.  Only
elementary error checking is done in vxli_lastrow and vxli_prevrow;
these functions should be called only in contexts where the array has
already been checked for validity.

Assuming that the special case of an empty array has been excluded, 
the canonical loop over rows looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of last voxel in row

vxli_lastrow(array->rank, array->dimen, coord);
do {
    ...operate on a row...
} 
while (vxli_prevrow(array->rank, array->dimen, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_lastrow(
    int rank,                   /* Rank of the array */
    long *dimen,                /* Dimensions of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = dimen[i] - 1;
    return;
}



/*-------------------------------------------------------------------------
vxli_prevrow - Compute coordinates of previous row in an image

Given a voxel array and a coordinate array pointing to some particular
row in an image, this function updates the coordinate array to point
to the previous row.  This function returns zero when all rows
have been seen.  See vxli_lastrow for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_prevrow(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-2; i >= 0; i--) {
	if ( --coord[i] >= 0 )
	    return 1;
	else
	    coord[i] = dimen[i] - 1;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_firstrow2 - Initialize iteration over generalized rows of an image

This function initializes a coordinate array prior to iterating over
the generalized rows of an image with vxli_nextrow2.  A generalized
row is defined here as a sequence of voxels which have N-1 of their
coordinates the same and differing only in the remaining coordinate.
It generalizes the concept of a row to any dimension of an array,
rather than always the last dimension.  Generalized rows are used much
less often than plain rows but are necessary for some operations,
notably the chamfer distance transform.  The generalized rows are
guaranteed to be generated in ascending order by coordinates.  Only
elementary error checking is done in vxli_firstrow2 and vxli_nextrow2;
these functions should be called only in contexts where the array has
already been checked for validity.

Assuming that the special case of an empty array has been excluded,
the canonical loop over the generalized rows along the ith dimension
looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of first voxel in row

vxli_firstrow2(array->rank, coord);
...initialize the ith coordinate, if not zero...
do {
    ...operate on a generalized row...
} 
while (vxli_nextrow2(array->rank, array->dimen, i, coord));

FIXME: This code (including vxli_nextrow2) has never been used or tested,
so use it with caution.
--------------------------------------------------------------------------*/

inlineable void
vxli_firstrow2(
    int rank,                   /* Rank of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = 0;
    return;
}



/*-------------------------------------------------------------------------
vxli_nextrow2 - Compute coordinates of next row in an image

Given a voxel array and a coordinate array pointing to some particular
row in an image, this function updates the coordinate array to point
to the next generalized row along the idim-th direction.  This
function returns zero when all rows have been seen.  See vxli_firstrow2
for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_nextrow2(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    int idim,                 /* Dimension to generate rows along */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-1; i >= 0; i--) {
	if ( i == idim )
	    continue;
	else if ( ++coord[i] < dimen[i] )
	    return 1;
	else
	    coord[i] = 0;
    }

    return 0;
}
