/*****************************************************************************
test_rand.c - Do some simple tests on the functions in rand.c

Author:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program does some elementary statistical tests on the functions
defined in rand.c, with a distinct bias toward those tests that I
think most relevant to my immediate needs for random numbers.  Note
that the functions are based on a published algorithm that is believed
to be strong, so these tests are mostly to insure against blunders in
the implementation.  I don't expect that these tests will be run very
often, so you have to change the source code and recompile to modify
the parameters of the tests.  Similarly, the statistics are computed
for you, but you have to look up the P values yourself.
******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "rand.h"
#include "errm.h"        /* For fatal, etc. */


void ran1_long_test1  (void);
void ran1_choose_test1(void);
void ran1_choose_test2(void);
void ran1_float_test1 (void);


int main()
{
    ran1_long_test1();

    ran1_choose_test1();

    ran1_choose_test2();

    ran1_float_test1();

    return 0;

}



/* Chi^2 test on ran1_long with 256 bins, 100000 variates */
void
ran1_long_test1()
{
    long ntries=100000;       /* How many variates to collect */
    int nbins = 100;          /* Number of bins to use */
    long *bin;                /* Count in each bin */
    double freq;              /* Number expected per bin */
    double chi2;              /* Chi-squared value */
    double z;                 /* Approximate Z statistic */
    int dof;                  /* Number of degrees of freedom */
    int nseed = 50, iseed;    /* Number of different seeds to try */
    long seed=3;              /* Seed for random generator */
    void *state;              /* State of random generator */
    long itry, x;
    int i;

    printf("\nChi-squared test of ran1_long:\n");
    printf("   ntries = %ld\n", ntries);
    printf("   nbins  = %d\n",  nbins);
    dof = nbins - 1;
    printf("   dof    = %d\n\n",  dof);

    /* Allocate space for histogram */
    bin = malloc(nbins*sizeof(long));

    /* Try with multiple different seeds */
    for (iseed = 0; iseed < nseed; iseed++) {

	/* Initialize random number generator; zero histogram */
	seed = iseed + 100000;
	printf("   seed = %10ld", seed);
	state = ran1_init(seed);
	for (i = 0; i < nbins; i++)
	    bin[i] = 0;

	/* Generate and histogram the random numbers */
	for (itry = 0; itry < ntries; itry++) {
	    x = ran1_long(state);
	    i = nbins*((x-1)/(double)RAN1_MAX);
	    if (i < 0 || i >= nbins)
		fatal("ran1_choose() is out of range ");
	    bin[i] += 1;
	}

#if 0
	/* Show the histogram */
	for (i = 0; i < nbins; i++)
	    printf("   bin[%4d] = %6ld\n", i, bin[i]);
#endif

	/* Compute and display chi-squared and DOF */
	freq = (double)ntries / nbins;
	chi2 = 0.0;
	for (i = 0; i < nbins; i++)
	    chi2 += (bin[i]-freq)*(bin[i]-freq)/freq;
	z = (chi2-dof)/sqrt(2*dof);
	printf("   chi^2 = %10f", chi2);
	printf("   z = %10f\n", z);
    }

    return;
}


/* Chi^2 test on ran1_choose with 256 bins, 10000 variates */
void
ran1_choose_test1()
{
    long ntries=100000;       /* How many variates to collect */
    int nbins = 256;          /* Number of bins to use */
    long *bin;                /* Count in each bin */
    double freq;              /* Number expected per bin */
    double chi2;              /* Chi-squared value */
    double z;                 /* Approximate Z statistic */
    int dof;                  /* Number of degrees of freedom */
    int nseed = 50, iseed;    /* Number of different seeds to try */
    long seed=3;              /* Seed for random generator */
    void *state;              /* State of random generator */
    long itry;
    int i;

    printf("\nChi-squared test of ran1_choose:\n");
    printf("   ntries = %ld\n", ntries);
    printf("   nbins  = %d\n",  nbins);
    dof = nbins - 1;
    printf("   dof    = %d\n\n",  dof);

    /* Allocate space for histogram */
    bin = malloc(nbins*sizeof(long));

    /* Try with multiple different seeds */
    for (iseed = 0; iseed < nseed; iseed++) {

	/* Initialize random number generator; zero histogram */
	seed = iseed + 100000;
	printf("   seed = %10ld", seed);
	state = ran1_init(seed);
	for (i = 0; i < nbins; i++)
	    bin[i] = 0;

	/* Generate and histogram the random numbers */
	for (itry = 0; itry < ntries; itry++) {
	    i = ran1_choose(state, nbins);
	    if (i < 0 || i >= nbins)
		fatal("ran1_choose() is out of range ");
	    bin[i] += 1;
	}

#if 0
	/* Show the histogram */
	for (i = 0; i < nbins; i++)
	    printf("   bin[%4d] = %6ld\n", i, bin[i]);
#endif

	/* Compute and display chi-squared and DOF */
	freq = (double)ntries / nbins;
	chi2 = 0.0;
	for (i = 0; i < nbins; i++)
	    chi2 += (bin[i]-freq)*(bin[i]-freq)/freq;
	z = (chi2-dof)/sqrt(2*dof);
	printf("   chi^2 = %10f", chi2);
	printf("   z = %10f\n", z);

    }

    return;
}




/* Chi^2 test on sequential triples from  ran1_choose */
void
ran1_choose_test2()
{
    long ntries = 100000;     /* How many variates to collect */
    int nrows = 256;          /* Voxels per dimension */
    int nbd = 16;             /* Number of bins per dimension */
    int nrb = nrows/nbd;      /* Divisor to get bin number from coord */
    int nbins = nbd*nbd*nbd;  /* Total number of bins */
    long *bin;                /* Count in each bin */
    double freq;              /* Number expected per bin */
    double chi2;              /* Chi-squared value */
    double z;                 /* Approximate Z statistic */
    int dof;                  /* Number of degrees of freedom */
    int nseed = 50, iseed;    /* Number of different seeds to try */
    long seed;                /* Seed for random generator */
    void *state;              /* State of random generator */
    long itry;
    int ix, iy, iz;           /* A random coordinate */
    int i;

    printf("\nChi-squared test of triples from ran1_choose:\n");
    printf("   ntries = %ld\n", ntries);
    printf("   nrows  = %d\n",  nrows);
    printf("   nbd    = %d\n",  nbd);
    dof = nbins - 1;
    printf("   dof    = %d\n\n",  dof);

    /* Allocate space for histogram */
    bin = malloc(nbins*sizeof(long));

    /* Try with multiple different seeds */
    for (iseed = 0; iseed < nseed; iseed++) {

	/* Initialize random number generator; zero histogram */
	seed = iseed + 100000;
	printf("   seed = %10ld", seed);
	state = ran1_init(seed);
	for (i = 0; i < nbins; i++)
	    bin[i] = 0;

	/* Generate and histogram triples of random numbers */
	for (itry = 0; itry < ntries; itry++) {
	    ix = ran1_choose(state, nrows);
	    iy = ran1_choose(state, nrows);
	    iz = ran1_choose(state, nrows);
	    i  = ix/nrb + nbd*(iy/nrb) + nbd*nbd*(iz/nrb);
	    if (i < 0 || i >= nbins)
		fatal("ran1_choose() is out of range ");
	    bin[i] += 1;
	}

#if 0
	/* Show the histogram */
	for (i = 0; i < nbins; i++)
	    printf("   bin[%4d] = %6ld\n", i, bin[i]);
#endif

	/* Compute and display chi-squared and DOF */
	freq = (double)ntries / nbins;
	chi2 = 0.0;
	for (i = 0; i < nbins; i++)
	    chi2 += (bin[i]-freq)*(bin[i]-freq)/freq;
	z = (chi2-dof)/sqrt(2*dof);
	printf("   chi^2 = %10f", chi2);
	printf("   z = %10f\n", z);

    }

    return;
}


/* Really quick and dirty test for ran1_float */
void ran1_float_test1
    (void)
{
    long i;
    void *state;       /* A random state */

    state = ran1_init(5);
    for (i = 0; i < 100; i++)
        printf("random float = %7.4f\n", ran1_float(state));
    return;
}
